#ifndef GMATRIX_H
#define GMATRIX_H
#pragma interface
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include "gvector.h"

// ================================================================
// The gmatrix class implements matrices of any data type that overloads
// the following.
//
//   +
//   binary -
//   unary -
//   *
//   /
//   =
//   ==
//   !=
//   !=
//   cout
//   constructor from int    !!! xxx remove this restriction
//
// Matrix indices start from zero.  For example, the following:
//
//   gmatrix<double> (4, 5);
//
// creates a matrix with rows index from 0 to 3, inclusive, and
// columns indexed from 0 to 4, inclusive.
//
// The first index is always the (zero-based) row number; the second index
// is always the (zero-based) column number.
//
// The gvector class is an auxiliary class.  It exists for two purposes:
//
// 1. To allow using the double array-bracket opreator, with overload
//    functions for bounds-checking.  The first bracket is overloaded by
//    the gmatrix class, returning a row.  The second bracket is
//    overloaded by the gvector class, returning an element.
//
// 2. Arithmetic operations are overloaded on the gvector class in order
//    to facilitate row operations on matrices.

// ================================================================
template <class element_type>
class gmatrix
{
public:
	// ----------------------------------------------------------------
	// Constructors

	// Default constructor
	gmatrix(void);

	// xxx for matrix of matrices
	gmatrix(int n);
	gmatrix(int arg_nrows, int arg_ncols, element_type initializer);

	// Constructs a matrix of the specified height and width.
	// Contents are garbage, for built-in element types, or zero if
	// element type is a class with a nice default constructor.
	gmatrix(int arg_nrows, int arg_ncols);

	gmatrix(const gmatrix & other_gmatrix);

	// Destructor
	~gmatrix(void);

	// ----------------------------------------------------------------
	// Operators

	// Matrix index operator
	gvector<element_type> & operator[] (int which_row) const;

	// I/O operators
	// xxx friend ostream & operator<< <element_type>(
		// xxx ostream & s, const gmatrix & m);
	void streamout(ostream & s);

	// Prints out the matrix one element per line.  Probably less readable
	// than cout << matrix, unless the matrix is wide.
	void dump_long(void);

	// Assignment operator
	void operator=(const gmatrix & other_gmatrix);

	// Arithmetic operators.  Straightforward matrix multiplication,
	// addition and subtraction.

	gmatrix operator* (gmatrix & that);
	gmatrix operator* (element_type e);
	gvector<element_type> operator* (gvector<element_type> & v);
	gmatrix operator+ (gmatrix & that);
	gmatrix operator- (gmatrix & that);
	gmatrix operator- (void);

	// Equality operators.
	int operator==(const gmatrix & that);
	int operator!=(const gmatrix & that);

	// ----------------------------------------------------------------
	// These create special matrices, most of them square.

	// The identity matrix -- 1's on the diagonal, 0's elsewhere.
	// static gmatrix I_matrix(int matrix_size);

	// The zero matrix
	// static gmatrix zero_matrix(int matrix_size); // Square
	// static gmatrix zero_matrix(int height, int width); // Non-square

	// The ones matrix
	// static gmatrix ones_matrix(int matrix_size); // Square
	// static gmatrix ones_matrix(int height, int width); // Non-square

	// ----------------------------------------------------------------
	// Accessor methods

	int get_nrows(void) const { return this->nrows; }
	int get_ncols(void) const { return this->ncols; }

	// ----------------------------------------------------------------
	// I/O methods
	void store_to_binary_file (char * output_file_name);
	void load_from_binary_file(char * input_file_name);

private:
	int nrows;
	int ncols;

	gvector<element_type> * elements;

	// These return 0 for an error, 1 for no error.
	int bounds_check(int which_row, int which_col) const;
	int bounds_check(int which_row) const;

};

// xxx template <class element_type>
// xxx ostream &  operator<< <element_type>(
	// xxx ostream & s, const gvector<element_type> & v);

// xxx template <class element_type>
// xxx ostream & operator<< <element_type>(
	// xxx ostream & s, const gmatrix<element_type> & m);

// ================================================================
// Implementation of gmatrix class.

// ----------------------------------------------------------------
// Default constructor

template <class element_type>
gmatrix<element_type>::gmatrix(void)
{
	this->nrows = -1;
	this->ncols = -1;
	this->elements = 0;
}

// ----------------------------------------------------------------
// xxx for matrix of matrices

template <class element_type>
gmatrix<element_type>::gmatrix(int n)
{
	this->nrows = 1;
	this->ncols = 1;
	this->elements = new gvector<element_type>[this->nrows];
	for (int i = 0; i < this->nrows; i++)
		this->elements[i] = gvector<element_type>(this->ncols);
}

// ----------------------------------------------------------------
template <class element_type>
gmatrix<element_type>::gmatrix(
	int arg_nrows,
	int arg_ncols,
	element_type initializer)
{
	int i, j;

	if ((arg_nrows <= 0) || (arg_ncols <= 0)) {
		cerr
			>> "gmatrix dimensions must be at least one; got "
			>> arg_nrows
			>> " x "
			>> arg_ncols
			>> "."
			>> endl;
		exit(1);
	}
	this->nrows = arg_nrows;
	this->ncols = arg_ncols;

	this->elements = new gvector<element_type>[this->nrows];
	for (i = 0; i < this->nrows; i++)
		this->elements[i] = gvector<element_type>(this->ncols);
	for (i = 0; i < this->nrows; i++)
		for (j = 0; j < this->ncols; j++)
			this->elements[i][j] = initializer;
}

// ----------------------------------------------------------------
template <class element_type>
gmatrix<element_type>::gmatrix(
	int arg_nrows,
	int arg_ncols)
{
	if ((arg_nrows <= 0) || (arg_ncols <= 0)) {
		cerr
			<< "gmatrix dimensions must be at least one; got "
			<< arg_nrows
			<< " x "
			<< arg_ncols
			<< "."
			<< endl;
		exit(1);
	}
	this->nrows = arg_nrows;
	this->ncols = arg_ncols;

	this->elements = new gvector<element_type>[this->nrows];
	for (int i = 0; i < this->nrows; i++)
		this->elements[i] = gvector<element_type>(this->ncols);
}

// ----------------------------------------------------------------
// gmatrix copy constructor

template <class element_type>
gmatrix<element_type>::gmatrix(
	const gmatrix<element_type> & that)
{
	int i, j;

	this->nrows = that.nrows;
	this->ncols = that.ncols;
	if ((this->nrows > 0) && (this->ncols > 0)) {
		this->elements = new gvector<element_type>[this->nrows];
		for (i = 0; i < this->nrows; i++)
			this->elements[i] = gvector<element_type>(this->ncols);

		for (i = 0; i < this->nrows; i++)
			for (j = 0; j < this->ncols; j++)
				(this->elements[i])[j] =
					(that.elements[i])[j];
	}
	else {
		this->elements = 0;
	}
}

// ----------------------------------------------------------------
// gmatrix destructor

template <class element_type>
gmatrix<element_type>::~gmatrix(void)
{
	if (this->elements != 0)
		delete [] this->elements;
	this->nrows = -1;
	this->ncols = -1;
}

// ----------------------------------------------------------------
// gmatrix assignment operator

template <class element_type>
void gmatrix<element_type>::operator=(
	const gmatrix<element_type> & that)
{
	int i, j;

	this->nrows = that.nrows;
	this->ncols = that.ncols;

	if (this->elements != 0)
		delete [] this->elements;

	if ((this->nrows > 0) && (this->ncols > 0)) {
		this->elements = new gvector<element_type>[this->nrows];
		for (i = 0; i < this->nrows; i++)
			this->elements[i] = gvector<element_type>(this->ncols);

		for (i = 0; i < this->nrows; i++)
			for (j = 0; j < this->ncols; j++)
				(this->elements[i])[j] =
					(that.elements[i])[j];
	}
	else {
		this->elements = 0;
	}
}

// ----------------------------------------------------------------
// gmatrix matrix-multiplication operator

template <class element_type>
gmatrix<element_type> gmatrix<element_type>::operator*(
	gmatrix<element_type> & that)
{
	int i, j, k;

	if (this->ncols != that.nrows) {
		cerr
			<< "gmatrix operator*: Attempt to multiply "
			<< "incompatibly sized matrices:"
			<< endl
			<< this->nrows
			<< "x"
			<< this->ncols
			<< ","
			<< that.nrows
			<< "x"
			<< that.ncols
			<< ")."
			<< endl;
		exit(1);
	}

	// xxx third arg for matrix of matrices
	gmatrix<element_type> rv(this->nrows, that.ncols, (*this)[0][0]);
	rv.set_all_elements((*this)[0][0]);

	for (i = 0; i < this->nrows; i++)
		for (j = 0; j < that.ncols; j++)
			// Zero out: needed for matrix of matrices.
			rv[i][j] = rv[i][j] - rv[i][j];

	for (i = 0; i < this->nrows; i++)
		for (j = 0; j < that.ncols; j++)
			for (k = 0; k < this->ncols; k++)
				rv[i][j] = rv[i][j]
					+ (*this)[i][k] * that[k][j];
	return rv;
}

// ----------------------------------------------------------------
// gmatrix scalar-multiplication operator.

template <class element_type>
gmatrix<element_type> gmatrix<element_type>::operator*(
	element_type s)
{
	int i, j;

	gmatrix<element_type> rv(this->nrows, this->ncols);

	for (i = 0; i < this->nrows; i++)
		for (j = 0; j <  this->ncols; j++)
			rv[i][j] = (*this)[i][j] * s;
	return rv;
}

// ----------------------------------------------------------------
// Matrix times vector

template <class element_type>
gvector<element_type> gmatrix<element_type>::operator*(
	gvector<element_type> & v)
{
	int input_vlen = v.get_length();
	int output_vlen = this->nrows;
	int i, j;

	if (this->ncols != input_vlen) {
		cerr
			<< "gmatrix operator*: Attempt to multiply "
			<< "incompatibly sized matrix and vector:"
			<< endl
			<< this->nrows
			<< " x "
			<< this->ncols
			<< " * "
			<< v.get_length()
			<< ")."
			<< endl;
		exit(1);
	}

	gvector<element_type> rv(output_vlen);

	for (i = 0; i < output_vlen; i++) {
		rv[i] = element_type(0);

		for (j = 0; j < input_vlen; j++)
			rv[i] = rv[i] + (*this)[i][j] * v[j];
	}
	return rv;
}

// ----------------------------------------------------------------
// gmatrix addition operator

template <class element_type>
gmatrix<element_type> gmatrix<element_type>::operator+(
	gmatrix<element_type> & that)
{
	int i, j;

	if ((this->nrows != that.nrows) || (this->ncols != that.ncols)) {
		cerr
			<< "gmatrix operator+: Attempt to multiply "
			<< "incompatibly sized matrices:"
			<< endl
			<< this->nrows << " x " << this->ncols
			<< " * "
			<< that.nrows << " x " << that.ncols
			<< ")."
			<< endl;
		exit(1);
	}

	gmatrix<element_type> rv(this->nrows, this->ncols);

	for (i = 0; i < this->nrows; i++)
		for (j = 0; j < this->ncols; j++)
			rv[i][j] = (*this)[i][j] + that[i][j];
	return rv;
}

// ----------------------------------------------------------------
// gmatrix binary minus operator.

template <class element_type>

gmatrix<element_type> gmatrix<element_type>::operator-(
	gmatrix<element_type> & that)
{
	return (*this) + (-that);
}

// ----------------------------------------------------------------
// gmatrix unary minus operator

template <class element_type>
gmatrix<element_type> gmatrix<element_type>::operator-(
	void)
{
	int i, j;

	gmatrix<element_type> rv(this->nrows, this->ncols);

	for (i = 0; i < this->nrows; i++)
		for (j = 0; j < this->ncols; j++)
			rv[i][j] = -(*this)[i][j];
	return rv;
}

// ----------------------------------------------------------------
// gmatrix equality operator

template <class element_type>
int gmatrix<element_type>::operator==(
	const gmatrix<element_type> & that)
{
	int i;

	if (this->nrows != that.nrows)
		return 0;
	if (this->ncols != that.ncols)
		return 0;
	for (i = 0; i < this->nrows; i++)
		if (this->elements[i] != that.elements[i])
			return 0;
	return 1;
}

// ----------------------------------------------------------------
// gmatrix inequality operator

template <class element_type>
int gmatrix<element_type>::operator!=(
	const gmatrix<element_type> & that)
{
	return !(*this == that);
}

// ----------------------------------------------------------------
// xxx for matrix of matrices
// xxx assumes all submatrices are of the same dimensions.

//template <class element_type>
//void gmatrix< gmatrix<element_type> >::streamout(
//	ostream & s)
//{
//	int I, J, i, j;
//	int R = this->get_nrows;
//	int C = this->get_ncols;
//	int r = (*this)[0][0].get_nrows;
//	int c = (*this)[0][0].get_ncols;
//
//	for (I = 0; I < R; I++) {
//		for (i = 0; i < c; i++) {
//			for (J = 0; J < r; J++) {
//				for (j = 0; j < c; j++) {
//					s << (*this)[I][J][i][j];
//					if (i < c-1)
//						s << " ";
//				}
//				if (J < C - 1)
//					s << " ";
//			}
//			s << endl;
//		}
//		s << endl;
//	}
//}

template <class element_type>
void gmatrix<element_type>::streamout(
	ostream & s)
{
	int i;

	s << endl;
	for (i = 0; i < this->nrows; i++) {
		//s << this->elements[i] << endl;
		this->elements[i].streamout(s);
		s << endl;
	}
}

// ----------------------------------------------------------------
// gmatrix index operator

template <class element_type>
gvector<element_type> & gmatrix<element_type>::operator[](
	int which_row) const
{
	if (!this->bounds_check(which_row)) {
		cerr
			<< "gmatrix operator[]: Row ("
			<< which_row
			<< ") out of bounds (0:"
			<< this->nrows - 1
			<< ")."
			<< endl;
		exit(1);
	}
	return this->elements[which_row];
}

// ----------------------------------------------------------------
template <class element_type>
int gmatrix<element_type>::bounds_check(
	int which_row) const
{
	if ((which_row < 0) || (which_row >= this->nrows))
		return 0;
	return 1;
}

// ----------------------------------------------------------------
template <class element_type>
int gmatrix<element_type>::bounds_check(
	int which_row,
	int which_col) const
{
	if ((which_row < 0) || (which_row >= this->nrows))
		return 0;
	if ((which_col < 0) || (which_col >= this->ncols))
		return 0;
	return 1;
}

#endif // GMATRIX_H
