#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "group.h"
#include "group_check.h"
#include "group_file.h"

#include "triv.h"
#include "testing.h"
#include "v4.h"
#include "mod.h"
#include "z2ipolymod.h"
#include "zppolymod.h"
#include "zmx.h"
#include "miquat.h"
#include "fpell3.h"
#include "ftell5.h"

#include "q8.h"
#include "dih.h"
#include "metacyc.h"
#include "genquat.h"
#include "pmt.h"
#include "tetra.h"
#include "cube.h"
#include "icos.h"
#include "lgf.h"

static group_spec_t * pgroup_specs[] = {

	&triv_group_spec,
	&testing_group_spec,
	&mod_add_group_spec,
	&mod_mul_group_spec,
	&smod_add_group_spec,
	&smod_mul_group_spec,
	&zmx_add_group_spec,
	&zmx_mul_group_spec,
	&fpell3_group_spec,
	&ftell5_group_spec,

	&miquat_add_group_spec,
	&miquat_mul_group_spec,
	&smiquat_add_group_spec,
	&smiquat_mul_group_spec,

	&z2ipolymodadd_group_spec,
	&z2ipolymodmul_group_spec,
	&zppolymod_add_group_spec,
	&zppolymod_mul_group_spec,

	&v4_group_spec,
	&q8_group_spec,
	&dn_group_spec,
	&metacyc_group_spec,
	&genquat_group_spec,
	&sn_group_spec,
	&an_group_spec,
	&snc_group_spec,
	&anc_group_spec,

	&tetra_group_spec,
	&cube_group_spec,
	&icos_group_spec,

	&slgf_group_spec,
	&glgf_group_spec,
	&dgf_group_spec,

	0 // Must be null-terminated.
};

// ----------------------------------------------------------------
int find_group_spec(
	char * desc,
	group_spec_t ** ppspec)
{
	int i;
	for (i = 0; pgroup_specs[i]; i++) {
		if (strcmp(pgroup_specs[i]->desc, desc) == 0) {
			*ppspec = pgroup_specs[i];
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
void list_group_specs(
	void)
{
	int i;
	int point;
	int curlen;
	int term_width = 80;

	fprintf(stderr, "-- Group specs:\n");
	point = 0;
	for (i = 0; pgroup_specs[i]; i++) {

		curlen = strlen(pgroup_specs[i]->desc);
		if (point != 0)
			curlen++;

		if (point + curlen >= term_width) {
			fprintf(stderr, "\n");
			fprintf(stderr, "%s", pgroup_specs[i]->desc);
			point = curlen;
		}
		else {
			if (point != 0)
				fprintf(stderr, " ");
			fprintf(stderr, "%s", pgroup_specs[i]->desc);
			point += curlen;
		}
	}
	if (point != 0)
		fprintf(stderr, "\n");
}



// ----------------------------------------------------------------
int make_group(
	group_t * pgroup,           // input
	group_spec_t * pgroup_spec, // input
	void    * pvaux,            // input
	char    * elements_file,    // input -- null for default elts.
	void    * elements_buffer,  // output
	int       max_group_bytes)  // output
{
	pgroup->pvelements = elements_buffer;
	pgroup->pvaux = pvaux;
	pgroup->pspec = pgroup_spec;

	if (elements_file) {
		return fill_group_from_file(pgroup, pgroup_spec, pvaux,
			elements_file, elements_buffer, max_group_bytes);
	}
	else {

		// xxx check max group bytes
		return pgroup_spec->pfill_group(
			pgroup->pvelements,
			pgroup->pvaux,
			max_group_bytes,
			&pgroup->num_elements);
	}
}

// ----------------------------------------------------------------
void describe_group(
	group_t * pgroup)
{
	printf("%s", pgroup->pspec->desc);
	if (pgroup->pspec->pvauxprint) {
		printf(" with parameter ");
		pgroup->pspec->pvauxprint(pgroup->pvaux);
	}
	printf(":\n");
}

// ----------------------------------------------------------------
void print_group_elements(
	group_t * pgroup)
{
	int i;
	void * pa;

	for (i = 0; i < pgroup->num_elements; i++) {
		pa = VGROUP_ELEMENT(pgroup, i);
		printf("  ");
		pgroup->pspec->pvprint(pa, pgroup->pvaux);
		printf("\n");
	}
}

// ----------------------------------------------------------------
// xxx should use log2 algm.
void group_element_power(
	group_t * pgroup,
	void    * pvin,
	int       power,
	void    * pvout)
{
	int i;

	// xxx to lib func.
	if (pgroup->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pgroup->pspec->element_size);
		exit(1);
	}

	while (power <= 0)
		power += pgroup->num_elements;
	while (power > pgroup->num_elements)
		power -= pgroup->num_elements;

	memcpy(pvout, pvin, pgroup->pspec->element_size);
	for (i = 1; i < power; i++) {
		pgroup->pspec->pvop(pvout, pvout, pvin, pgroup->pvaux);
	}
}

// ----------------------------------------------------------------
void group_element_inverse(
	group_t * pgroup,
	void    * pvin,
	void    * pvout)
{
	// For any finite group G, for all x in G, x^|G| = e.
	// Therefore x^(|G|-1) = x^(-1).
	group_element_power(pgroup, pvin, pgroup->num_elements - 1, pvout);
}

// ----------------------------------------------------------------
int group_element_order(
	group_t * pgroup,
	void    * pa)
{
	unsigned char apower[VOP_MAX_ELEMENT_SIZE];
	int order;

	// xxx to lib func.
	if (pgroup->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pgroup->pspec->element_size);
		exit(1);
	}

	memcpy(apower, pa, pgroup->pspec->element_size);
	pgroup->pspec->pvop(apower, apower, pa, pgroup->pvaux);
	order = 1;

	while (!pgroup->pspec->pvequal(pa, apower, pgroup->pvaux)) {

		if (order > pgroup->num_elements) {
			printf("group_find_orders: "
				"non-invertible element ");
			pgroup->pspec->pvprint(pa, pgroup->pvaux);
			printf(".\n");
			exit(1);
		}

		pgroup->pspec->pvop(apower, apower, pa, pgroup->pvaux);
		order++;
	}

	return order;
}

// ----------------------------------------------------------------
int group_append_element(
	group_t * pgroup,
	void    * psrc,
	int       max_group_bytes)
{
	int max_num_elements = max_group_bytes / pgroup->pspec->element_size;
	void * pdst;

	if (pgroup->num_elements >= max_num_elements) {
		fprintf(stderr, "fill_group_from_file: "
		"too many elements; max %d.\n",
		max_num_elements);
		return 0;
	}
	pdst = VGROUP_ELEMENT(pgroup, pgroup->num_elements);
	memcpy(pdst, psrc, pgroup->pspec->element_size);
	pgroup->num_elements++;
	return 1;
}

// ----------------------------------------------------------------
int group_maybe_add_element(
	group_t * pgroup,
	void    * psrc,
	int       max_group_bytes)
{
	int index;
	if (group_find_index(pgroup, psrc, &index))
		return 1;
	else
		return group_append_element(pgroup, psrc, max_group_bytes);
}
