#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "malloc_check.h"
#include "group.h"
#include "group_check.h"
#include "group_cosets.h"
#include "group_ascendant.h"

// ----------------------------------------------------------------
// make_ascendant_subgroup:  Computes eta^-1(Z(G/Zi(G))).
//
// * Zi(G) is given
//
// * G/Zi(G) is obtained by calling make_cosets()
//
// * Z(G/Zi(G)):  First consider computing the center of a non-quotient
//   group K.  Z(K) = {z in K : zk = kz for all k in K}.  So Z(G/Zi)
//   is {gj Zi in G/Zi : gj Zi gk Zi = gk Zi gj Zi for all gk Zi in G/Zi}.
//
//     gj Zi gk Zi = gk Zi gj Zi -- when?  Note Zi's are normal in G, so
//     gj gk Zi Zi = gk gj Zi Zi
//     gj gk Zi = gk gj Zi
//     gj gk (gk gj)^-1 in Zi
//     gj gk gj^-1 gk^-1 in Zi
//
//   where gi, gj loop over coset representatives, i.e. 1 <= i, j <= [G:Zi].
//
// * eta^-1 of a set of cosets is just a "flattening" or "brace erase":
//   E.g. if G = {1, -1, i, -i, j, -j, k, -k} and H = {1, -1}, with eta
//   sending an element g of G to its coset gH in G/H, suppose
//   S = {{1, -1}, {i, -i}} = {1H, iH}.  Then eta^-1(G/H) is simply
//   {1, -1, i, -i}.
//
// So, in practice:
// * Given G, Zi
// * Form G/Zi
// * For gj looping over coset reps
//     If [gj,gk] is in Zi for gk looping over coset reps
//       then append all elements of gj Zi to the ascendant.

int make_ascendant_subgroup(
	group_t  * pG,
	group_t  * pZi,  // Z_i(G)
	group_t  * pZi1, // Z_{i+1}(G)
	void     * Zi1_elements_buffer,
	int        max_Zi1_bytes,
	int        verbosity)
{
	group_cosets_t G_mod_Zi;
	unsigned char * quot_buf;
	unsigned quot_bytes;
	int oG, oZi, i_G_Zi;
	int j, k;
	void * pgj, * pgk;
	unsigned char gjinv[VOP_MAX_ELEMENT_SIZE];
	unsigned char gkinv[VOP_MAX_ELEMENT_SIZE];
	unsigned char gj_gk[VOP_MAX_ELEMENT_SIZE];
	unsigned char gj_gk_gjinv[VOP_MAX_ELEMENT_SIZE];
	unsigned char gj_gk_gjinv_gkinv[VOP_MAX_ELEMENT_SIZE];
	int dummy, commutator_in_Zi, all_commute;

	oG  = pG->num_elements;
	oZi = pZi->num_elements;
	if ((oG % oZi) != 0) {
		fprintf(stderr, "make_ascendant:  Order of sub (%u) does not "
			"divide order of group (%u).\n",
			oZi, oG);
		return 0;
	}
	i_G_Zi = oG / oZi;

	// xxx to lib func.
	if (pG->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pG->pspec->element_size);
		exit(1);
	}

	pZi1->pvelements = Zi1_elements_buffer;
	pZi1->num_elements = 0; // increment below
	pZi1->pvaux = pG->pvaux;
	pZi1->pspec = pG->pspec;

	quot_bytes = pG->pspec->element_size * pG->num_elements;
	quot_buf = malloc_check(quot_bytes);

	// Compute G/Zi(G)
	if (!make_cosets(pG, pZi, LEFT_COSETS, &G_mod_Zi, quot_buf,
		quot_bytes, verbosity))
	{
		free(quot_buf);
		fprintf(stderr,
		"make_ascendant_subgroup:  Couldn't compute G/Zi(G).\n");
		return 0;
	}

	if (verbosity) {
		printf("\n");
		printf("make_ascendant:\n");

		printf("G:\n");
		print_group_elements(pG);
		printf("\n");

		printf("Zi:\n");
		print_group_elements(pZi);
		printf("\n");

		printf("G/Zi(G):\n");
		group_print_cosets(&G_mod_Zi);
		printf("\n");
	}

	// Compute Z(G/Zi(G))
	for (j = 0; j < i_G_Zi; j++) {
		pgj = VCOSETS_ELEMENT(&G_mod_Zi, j, 0);
		group_element_inverse(pG, pgj, gjinv);

		all_commute = 1;
		for (k = 0; k < i_G_Zi; k++) {
			pgk = VCOSETS_ELEMENT(&G_mod_Zi, k, 0);
			group_element_inverse(pG, pgk, gkinv);

			pG->pspec->pvop(gj_gk, pgj, pgk, pG->pvaux);
			pG->pspec->pvop(gj_gk_gjinv, gj_gk, gjinv,
				pG->pvaux);
			pG->pspec->pvop(gj_gk_gjinv_gkinv, gj_gk_gjinv,
				gkinv, pG->pvaux);
			commutator_in_Zi = group_find_index(pZi,
				gj_gk_gjinv_gkinv, &dummy);

			if (verbosity >= 3) {
				printf("    j=%d k=%d", j, k);
				printf(" gj=");
				pG->pspec->pvprint(pgj, pG->pvaux);
				printf(" gk=");
				pG->pspec->pvprint(pgk, pG->pvaux);
				printf(" gjinv=");
				pG->pspec->pvprint(gjinv, pG->pvaux);
				printf(" gkinv=");
				pG->pspec->pvprint(gkinv, pG->pvaux);
				printf(" [gj,gk]=");
				pG->pspec->pvprint(gj_gk_gjinv_gkinv,
					pG->pvaux);
				if (commutator_in_Zi)
					printf(" *");
				printf("\n");
			}

			if (!commutator_in_Zi) {
				all_commute = 0;
				break;
			}
		}
		if (verbosity >= 3) {
			printf("\n");
		}
		if (all_commute) {
			// append all elements of this coset to the
			// ascendant.
			for (k = 0; k < oZi; k++) {
				pgk = VCOSETS_ELEMENT(&G_mod_Zi, j, k);
				if (!group_append_element(pZi1, pgk,
					max_Zi1_bytes))
				{
					fprintf(stderr,
						"make_ascendant:  Zi1 full.\n");
					return 0;
				}
				if (verbosity >= 2) {
					printf("  append ");
					pG->pspec->pvprint(pgk, pG->pvaux);
					printf("\n");
				}
			}
			if (verbosity >= 2) {
				printf("\n");
			}
		}
	}

	if (verbosity) {

		printf("Zi1:\n");
		print_group_elements(pZi1);
		printf("\n");
	}

	free(quot_buf);
	return 1;
}
