#ifndef GVECTOR_H
#define GVECTOR_H
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>

// ================================================================
// The gvector class implements arrays of any data type that overloads
// the following:
//
//   +
//   binary -
//   unary -
//   *
//   /
//   =
//   ==
//   !=
//   !=
//   cout
//   constructor from int
//
// Array indices start from zero.  For example, the following:
//
//   gmatrix<double> (4);
//
// creates an array with elements indexed from 0 to 3, inclusive.

// ================================================================
template <class element_type>
class gvector
{
public:
	// Default constructor
	gvector();

	// Constructs a vector of the specified length.
	// Contents are garbage, for built-in element types, or zero if
	// element type is a class with a nice default constructor.
	gvector (int ncols);

	// xxx for matrix of mods
	gvector (int ncols, element_type initializer);

	// Copy constructor
	gvector(const gvector & that);

	// Destructor.
	~gvector(void);

	// Operators:  index, output, assignment and arithmetic.

	element_type &    operator[] (int which_column) const;
	void              operator=  (const gvector & that);
	gvector           operator+  (const gvector & that);
	gvector           operator-  (const gvector & that);
	gvector           operator-  (void);
	gvector           operator*  (element_type s);
	//gvector           operator*  (element_type s, int dummy);
	gvector           operator/  (element_type s);
	int               operator== (const gvector & that);
	int               operator!= (const gvector & that);
	// xxx friend ostream &  operator<< <element_type> (
		// xxx ostream & s, const gvector<element_type> & v);
	void streamout(ostream & s);

	// Accessor methods
	int get_length() { return this->length; }

	// I/O methods
	void store_to_binary_file  (char * output_file_name);
	void store_to_binary_stream(ofstream & output_stream,
		char * file_name_for_error_messages);
	void load_from_binary_file (char * input_file_name);
	void load_from_binary_stream(ifstream & input_stream,
		char * file_name_for_error_messages);

	int contains_element(element_type e);

private:
	int length;
	element_type * row_elements;
};

// ================================================================
// Implementation of gvector class.

// ----------------------------------------------------------------
template <class element_type>
gvector<element_type>::gvector(void)
{
	this->length = -1;
	this->row_elements = 0;
}

// ----------------------------------------------------------------
// gvector constructor with uninitialized data
template <class element_type>
gvector<element_type>::gvector(
	int ncols)
{
	this->length = ncols;
	this->row_elements = new element_type[this->length];
}

// ----------------------------------------------------------------
// gvector constructor with initialized data
template <class element_type>
gvector<element_type>::gvector(
	int ncols,
	element_type initializer)
{
	this->length = ncols;
	this->row_elements = new element_type[this->length];
	for (int i = 0; i < this->length; i++)
		this->row_elements[i] = initializer;
}

// ----------------------------------------------------------------
// gvector copy constructor
template <class element_type>
gvector<element_type>::gvector(
	const gvector<element_type> & that)
{
	this->length = that.length;
	if (this->length > 0) {
		this->row_elements = new element_type[this->length];
		for (int i = 0; i < this->length; i++)
			this->row_elements[i] = that.row_elements[i];
	}
	else {
		this->row_elements = 0;
	}
}

// ----------------------------------------------------------------
// gvector destructor
template <class element_type>
gvector<element_type>::~gvector(
	void)
{
	if (this->row_elements != 0)
		delete [] this->row_elements;
	this->row_elements = 0;
	this->length = -1;
}

// ----------------------------------------------------------------
// gvector index operator
template <class element_type>
element_type &
gvector<element_type>::operator[](
	int which_column) const
{
	if ((which_column < 0) || (which_column >= this->length)) {
		cerr
			<< "gvector operator[]: Index "
			<< which_column
			<< " out of bounds (0:"
			<< this->length - 1
			<< ")."
			<< endl;
		exit(1);
	}
	return this->row_elements[which_column];
}

// ----------------------------------------------------------------
// gvector assignment operator
template <class element_type>
void
gvector<element_type>::operator=(
	const gvector<element_type> & that)
{
	if (this->row_elements != 0)
		delete [] this->row_elements;
	this->length = that.length;
	if (this->length != 0) {
		this->row_elements = new element_type[this->length];
		for (int i = 0; i < this->length; i++)
			this->row_elements[i] = that.row_elements[i];
	}
	else {
		this->row_elements = 0;
	}
}

// ----------------------------------------------------------------
// gvector addition operator
template <class element_type>
gvector<element_type>
gvector<element_type>::operator+(
	const gvector<element_type> & that)
{
	if (this->length != that.length) {
		cerr
			<< "gvector operator+(): Attempt to add "
			<< "incompatiblty sized rows ("
			<< this->length
			<< ", "
			<< that.length
			<< ")."
			<< endl;
		exit(1);
	}
	gvector<element_type> rv(this->length);
	for (int i = 0; i < this->length; i++)
		rv.row_elements[i] =
			this->row_elements[i] + that.row_elements[i];
	return rv;
}

// ----------------------------------------------------------------
// gvector binary minus operator
template <class element_type>
gvector<element_type>
gvector<element_type>::operator-(
	const gvector<element_type> & that)
{
	if (this->length != that.length) {
		cerr
			<< "gvector operator-(): Attempt to add "
			<< "incompatiblty sized rows ("
			<< this->length
			<< ", "
			<< that.length
			<< ")."
			<< endl;
		exit(1);
	}
	gvector<element_type> rv(this->length);
	for (int i = 0; i < this->length; i++)
		rv.row_elements[i] =
			this->row_elements[i] - that.row_elements[i];
	return rv;
}

// ----------------------------------------------------------------
// gvector unary minus operator
template <class element_type>
gvector<element_type>
gvector<element_type>::operator-(
	void)
{
	gvector<element_type> rv(this->length);
	for (int i = 0; i < this->length; i++)
		rv.row_elements[i] = -this->row_elements[i];
	return rv;
}

// ----------------------------------------------------------------
// gvector scalar multiplication operator
template <class element_type>
gvector<element_type>
gvector<element_type>::operator*(
	element_type s)
{
	gvector<element_type> rv(this->length);
	for (int i = 0; i < this->length; i++)
		rv.row_elements[i] = this->row_elements[i] * s;
	return rv;
}

// ----------------------------------------------------------------
// gvector scalar division operator
template <class element_type>
gvector<element_type>
gvector<element_type>::operator/(
	element_type s)
{
	gvector<element_type> rv(this->length);
	for (int i = 0; i < this->length; i++)
		rv.row_elements[i] = this->row_elements[i] / s;
	return rv;
}

// ----------------------------------------------------------------
// gvector equality operator
template <class element_type>
int
gvector<element_type>::operator==(
	const gvector<element_type> & that)
{
	if (this->length != that.length)
		return 0;
	for (int i = 0; i < this->length; i++)
		if (this->row_elements[i] != that.row_elements[i])
			return 0;
	return 1;
}

// ----------------------------------------------------------------
// gvector inequality operator
template <class element_type>
int
gvector<element_type>::operator!=(
	const gvector<element_type> & that)
{
	return !(*this == that);
}

// ----------------------------------------------------------------
template <class element_type>
void
gvector<element_type>::streamout(
	ostream & s)
{
	for (int i = 0; i < this->length; i++) {
		s << this->row_elements[i];
		if (i < (this->length - 1))
			s << " ";
	}
}

#endif // GVECTOR_H
