#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ring.h"
#include "ring_check.h"
#include "ring_file.h"

//#include "triv.h"
//#include "testing.h"
#include "mod.h"
//#include "z2ipolymod.h"
//#include "zppolymod.h"
//#include "zmx.h"
//#include "miquat.h"

static ring_spec_t * pring_specs[] = {

	//&triv_ring_spec,
	//&testing_ring_spec,
	&mod_ring_spec,
	&smod_ring_spec,
//	&zmx_ring_spec,

//	&miquat_ring_spec,
//	&smiquat_ring_spec,

//	&z2ipolymod_ring_spec,
//	&zppolymod_ring_spec,

	//&slgf_ring_spec,
	//&glgf_ring_spec,
	//&dgf_ring_spec,

	0 // Must be null-terminated.
};

// ----------------------------------------------------------------
int find_ring_spec(
	char * desc,
	ring_spec_t ** ppspec)
{
	int i;
	for (i = 0; pring_specs[i]; i++) {
		if (strcmp(pring_specs[i]->desc, desc) == 0) {
			*ppspec = pring_specs[i];
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
void list_ring_specs(
	void)
{
	int i;
	int point;
	int curlen;
	int term_width = 80;

	fprintf(stderr, "-- Ring specs:\n");
	point = 0;
	for (i = 0; pring_specs[i]; i++) {

		curlen = strlen(pring_specs[i]->desc);
		if (point != 0)
			curlen++;

		if (point + curlen >= term_width) {
			fprintf(stderr, "\n");
			fprintf(stderr, "%s", pring_specs[i]->desc);
			point = curlen;
		}
		else {
			if (point != 0)
				fprintf(stderr, " ");
			fprintf(stderr, "%s", pring_specs[i]->desc);
			point += curlen;
		}
	}
	if (point != 0)
		fprintf(stderr, "\n");
}



// ----------------------------------------------------------------
int make_ring(
	ring_t * pring,           // input
	ring_spec_t * pring_spec, // input
	void    * pvaux,            // input
	char    * elements_file,    // input -- null for default elts.
	void    * elements_buffer,  // output
	int       max_ring_bytes)  // output
{
	pring->pvelements = elements_buffer;
	pring->pvaux = pvaux;
	pring->pspec = pring_spec;

	if (elements_file) {
		return fill_ring_from_file(pring, pring_spec, pvaux,
			elements_file, elements_buffer, max_ring_bytes);
	}
	else {

		// xxx check max ring bytes
		return pring_spec->pfill_ring(
			pring->pvelements,
			pring->pvaux,
			max_ring_bytes,
			&pring->num_elements);
	}
}

// ----------------------------------------------------------------
void describe_ring(
	ring_t * pring)
{
	printf("%s", pring->pspec->desc);
	if (pring->pspec->pvauxprint) {
		printf(" with parameter ");
		pring->pspec->pvauxprint(pring->pvaux);
	}
	printf(":\n");
}

// ----------------------------------------------------------------
void print_ring_elements(
	ring_t * pring)
{
	int i;
	void * pa;

	for (i = 0; i < pring->num_elements; i++) {
		pa = VRING_ELEMENT(pring, i);
		printf("  ");
		pring->pspec->pvprint(pa, pring->pvaux);
		printf("\n");
	}
}

// ----------------------------------------------------------------
// xxx should use log2 algm.
void ring_element_power(
	ring_t * pring,
	void    * pvin,
	int       power,
	void    * pvout)
{
	int i;

	// xxx to lib func.
	if (pring->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pring->pspec->element_size);
		exit(1);
	}

	while (power <= 0)
		power += pring->num_elements;
	while (power > pring->num_elements)
		power -= pring->num_elements;

	memcpy(pvout, pvin, pring->pspec->element_size);
	for (i = 1; i < power; i++) {
		pring->pspec->pvadd(pvout, pvout, pvin, pring->pvaux);
	}
}

// ----------------------------------------------------------------
void ring_element_inverse(
	ring_t * pring,
	void    * pvin,
	void    * pvout)
{
	// For any finite ring G, for all x in G, x^|G| = e.
	// Therefore x^(|G|-1) = x^(-1).
	ring_element_power(pring, pvin, pring->num_elements - 1, pvout);
}

// ----------------------------------------------------------------
int ring_element_order(
	ring_t * pring,
	void    * pa)
{
	unsigned char apower[VOP_MAX_ELEMENT_SIZE];
	int order;

	// xxx to lib func.
	if (pring->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pring->pspec->element_size);
		exit(1);
	}

	memcpy(apower, pa, pring->pspec->element_size);
	pring->pspec->pvadd(apower, apower, pa, pring->pvaux);
	order = 1;

	while (!pring->pspec->pvequal(pa, apower, pring->pvaux)) {

		if (order > pring->num_elements) {
			printf("ring_find_orders: "
				"non-invertible element ");
			pring->pspec->pvprint(pa, pring->pvaux);
			printf(".\n");
			exit(1);
		}

		pring->pspec->pvadd(apower, apower, pa, pring->pvaux);
		order++;
	}

	return order;
}

// ----------------------------------------------------------------
int ring_append_element(
	ring_t * pring,
	void    * psrc,
	int       max_ring_bytes)
{
	int max_num_elements = max_ring_bytes / pring->pspec->element_size;
	void * pdst;

	if (pring->num_elements >= max_num_elements) {
		fprintf(stderr, "fill_ring_from_file: "
		"too many elements; max %d.\n",
		max_num_elements);
		return 0;
	}
	pdst = VRING_ELEMENT(pring, pring->num_elements);
	memcpy(pdst, psrc, pring->pspec->element_size);
	pring->num_elements++;
	return 1;
}

// ----------------------------------------------------------------
int ring_maybe_add_element(
	ring_t * pring,
	void    * psrc,
	int       max_ring_bytes)
{
	int index;
	if (ring_find_index(pring, psrc, &index))
		return 1;
	else
		return ring_append_element(pring, psrc, max_ring_bytes);
}
