// ----------------------------------------------------------------
// Input ((Z_10, +) in this example):
//
//   0 1 2 3 4 5 6 7 8 9
//
// ----------------------------------------------------------------
// Compute orders:
//
//   |0| =  1
//   |1| = 10
//   |2| =  5
//   |3| = 10
//   |4| =  5
//   |5| =  2
//   |6| =  5
//   |7| = 10
//   |8| =  5
//   |9| = 10
//
// ----------------------------------------------------------------
// Sort elements by increasing orders:
//
//   |0| =  1
//   |5| =  2
//   |2| =  5
//   |4| =  5
//   |6| =  5
//   |8| =  5
//   |1| = 10
//   |3| = 10
//   |7| = 10
//   |9| = 10
//
// ----------------------------------------------------------------
// Identify subsets of same orders:
//
//     Start index = 0  count = 1  order = 1: 0
//     Start index = 1  count = 1  order = 2: 5
//     Start index = 2  count = 4  order = 5: 2 4 6 8
//     Start index = 6  count = 4  order =10: 1 3 7 9
//
// ----------------------------------------------------------------
// Permute only within same-order subsets:
//
//     Count = 1! =  1   1: 0
//     Count = 1! =  1   2: 5
//     Count = 4! = 24   5: 2 4 6 8
//     Count = 4! = 24   10: 1 3 7 9
//
// Total number of permutations: 1*1*24*24 = 576.  A lot better than testing
// all 9! = 362,880 identity-fixing permutations.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ring.h"
#include "pmt.h"
#include "ring_check.h"
#include "ring_orders.h"
#include "ring_aut_ring.h"

#include "factorial.h"

// ----------------------------------------------------------------
// Description of a same-order subset.
typedef struct _sos_desc_t {
	int start_index;
	int count;
	int order;
	int npmts;
} sos_desc_t;

// ----------------------------------------------------------------
static void show_elements_and_orders(
	ring_t * pring,
	pmt_t    * psort_map,
	int     * orders);

static void sort_elements_by_order(
	ring_t * pring,
	pmt_t    * psort_map,
	int     * orders);

static int count_same_order_subsets(
	int     * orders,
	int       num_elements);

static void populate_sos_descs(
	int        * orders,
	int          num_elements,
	sos_desc_t * sos_descs,
	int          num_sos,
	int        * pnum_pmts);

static void print_sos_descs(
	sos_desc_t * sos_descs,
	int          num_sos);

static void decompose_pmt_indices(
	sos_desc_t * sos_descs,
	int          num_sos,
	int          whichp,
	int        * pmt_indices);

static void compute_permutation(
	int  * pmt_indices,
	sos_desc_t * sos_descs,
	int    num_sos,
	int    ring_num_elements,
	pmt_t * pid,
	pmt_t * pout);

static void compute_subpermutation(
	pmt_t * pin,
	pmt_t * ptemp,
	pmt_t * pout,
	int    start,
	int    count,
	int    nfact,
	int    whichp);

static void print_elements_with_sos(
	ring_t    * pring,
	pmt_t       * ppmt,
	sos_desc_t * sos_descs,
	int          num_sos);

int is_automorphism(
	ring_t * pring,
	pmt_t    * ppmt,
	int       verbosity);

// ----------------------------------------------------------------
void make_aut_ring(
	ring_t * pring,
	ring_t * paut_ring,
	void    * aut_buffer, // xxx pass pmt_t * instead?
	int       max_aut_ring_bytes,
	int       verbosity)
{
	int    max_aut_ring_size;
	pmt_t * pSns;
	pmt_t   curr_pmt;
	int  * orders;
	int    ring_num_bytes;
	pmt_t   id_map;
	pmt_t   sort_map;
	pmt_t   sort_map_inv;
	pmt_t   sc;
	void * sorted_elements;
	//void * temp;
	//void * pmt;
	ring_t sort_ring;
	//ring_t pmt_ring;
	int    num_sos;
	sos_desc_t * sos_descs;
	int  * pmt_indices;
	int    whichp;
	int    num_pmts;

	if (verbosity >= 1) {
		printf("make_aut_ring: Group size %d.\n",
			pring->num_elements);
	}

// Allocate some arrays
	ring_num_bytes = pring->num_elements * pring->pspec->element_size;
	orders = malloc(pring->num_elements * sizeof(int));
	sorted_elements = malloc(ring_num_bytes);
	//temp = malloc(ring_num_bytes);
	//pmt  = malloc(ring_num_bytes);
	//if ((orders == 0) || (sorted_elements == 0) || (temp == 0)
		//|| (pmt == 0))
	if ((orders == 0) || (sorted_elements == 0)) {
		fprintf(stderr, "malloc(%d) failed.\n",
			pring->num_elements * sizeof(int));
		exit(1);
	}

// xxx to lib func.
	if (pring->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pring->pspec->element_size);
		exit(1);
	}

// xxx cmt here
	paut_ring->pvelements = aut_buffer;
	paut_ring->num_elements = 0; // Increment below.
	paut_ring->pvaux = &pring->num_elements;
	paut_ring->pspec = &sn_ring_spec;
	pSns = (pmt_t *)aut_buffer;
	max_aut_ring_size = max_aut_ring_bytes / sizeof(pmt_t);

	sort_ring.pvelements = sorted_elements;
	sort_ring.num_elements = pring->num_elements;
	sort_ring.pvaux = pring->pvaux;
	sort_ring.pspec = pring->pspec;

	//pmt_ring.pvelements = pmt;
	//pmt_ring.num_elements = pring->num_elements;
	//pmt_ring.pvaux = pring->pvaux;
	//pmt_ring.pspec = pring->pspec;

// Compute orders of elements
	// xxx to lib func.
	ring_find_orders(pring, orders, verbosity);

	// xxx cmt here
	pmt_make_id(&id_map,   pring->num_elements);
	pmt_make_id(&sort_map, pring->num_elements);

	if (verbosity >= 2) {
		printf("Elements and orders:\n");
		show_elements_and_orders(pring, &sort_map, orders);
		fflush(stdout);
	}

// Sort elements by increasing orders
	memcpy(sorted_elements, pring->pvelements, ring_num_bytes);
	sort_elements_by_order(&sort_ring, &sort_map, orders);

	if (verbosity >= 2) {
		printf("Elements and orders, sorted by increasing order:\n");
		show_elements_and_orders(&sort_ring, &id_map, orders);
		fflush(stdout);
	}

	pmt_invert(&sort_map, &sort_map_inv,
		pring->num_elements);

// Identify subsets of same orders;
// count permutations within each same-order subset.
// Total number of permutations is the product of the number of
// permutations of each same-order subset.
	num_sos = count_same_order_subsets(orders, pring->num_elements);
	sos_descs = (sos_desc_t *)malloc(num_sos * sizeof(sos_desc_t));
	pmt_indices = (int *)malloc(num_sos * sizeof(int));
	if ((sos_descs == 0) || (pmt_indices == 0)) {
		fprintf(stderr, "malloc(%d) failed.\n",
			num_sos * sizeof(sos_desc_t));
		exit(1);
	}

	populate_sos_descs(orders, pring->num_elements, sos_descs, num_sos,
		&num_pmts);
	if (verbosity >= 3)
		print_sos_descs(sos_descs, num_sos);

	if (verbosity >= 1) {
		printf("make_aut_ring: %d permutation%s to test.\n",
			num_pmts,
			num_pmts == 1 ? "" : "s");
	}

// Permute only within same-order subsets

	for (whichp = 0; whichp < num_pmts; whichp++) {

		decompose_pmt_indices(sos_descs, num_sos, whichp,
			pmt_indices);

		compute_permutation(pmt_indices, sos_descs, num_sos,
			pring->num_elements, &id_map, &curr_pmt);

		if (verbosity >= 3) {
			int pi;
			printf("-- Testing permutation %7d of %7d:\n",
				whichp, num_pmts);

			printf("-- Subpmts: ");
			for (pi = 0; pi < num_sos; pi++)
				printf(" %3d/%3d",
					pmt_indices[pi],
					sos_descs[pi].npmts);
			printf("\n");

			print_elements_with_sos(&sort_ring,
				&id_map,
				sos_descs, num_sos);
			print_elements_with_sos(&sort_ring, &curr_pmt,
				sos_descs, num_sos);
		}

		if (is_automorphism(&sort_ring, &curr_pmt, verbosity)) {
			//xxx need to conjugate by the sort map!!
			if (verbosity >= 3) {
				printf("Is an automorphism.\n");
			}
			else if (verbosity >= 2) {
				print_elements_with_sos(&sort_ring,
					&curr_pmt, sos_descs, num_sos);
			}

			if (paut_ring->num_elements >= max_aut_ring_size) {
				fprintf(stderr, "make_aut_ring: out of space "
					"(%d automorphism%s).\n",
					max_aut_ring_size,
					max_aut_ring_size == 1 ? "" : "s");
				exit(1);
			}

			//pSns[paut_ring->num_elements] = curr_pmt;

			// xxx cmt conjugation
			sc = pmt_mul(&sort_map, &curr_pmt,
				pring->num_elements);
			pSns[paut_ring->num_elements] = pmt_mul(
				&sc, &sort_map_inv, pring->num_elements);

			paut_ring->num_elements++;
		}
		else {
			if (verbosity >= 3)
				printf("Not an automorphism.\n");
		}
	}

	if (verbosity >= 1) {
		printf("make_aut_ring: %d automorphism%s found.\n",
			paut_ring->num_elements,
			paut_ring->num_elements == 1 ? "" : "s");
	}

	free(orders);
	free(sorted_elements);
	//free(temp);
	//free(pmt);
	free(sos_descs);
	free(pmt_indices);
}

// ----------------------------------------------------------------
// We are passed a permutation of the ring, so the mapping is already
// one-to-one and onto.  We need only to check that the mapping is also
// a homomorphism, i.e. phi(a) phi(b) = phi(a b) for all a, b in the ring.

// xxx to take pring & ppmt instead.
int is_automorphism(
	ring_t * pring,
	pmt_t    * ppmt,
	int       verbosity)
{
	int i, j;
	void * pa, * pb;
	void * pphi_a, * pphi_b;
	unsigned char a_b[VOP_MAX_ELEMENT_SIZE];
	void * pphi_a_b;
	unsigned char phi_a_phi_b[VOP_MAX_ELEMENT_SIZE];
	int ab_index;
	int match;

	// xxx to lib func
	if (pring->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr, "%s:  element size %d too big for %d. "
			"Please adjust and recompile.\n",
			__FILE__,
			pring->pspec->element_size,
			VOP_MAX_ELEMENT_SIZE);
		exit(1);
	}

	for (i = 0; i < pring->num_elements; i++) {
		pa = VRING_ELEMENT(pring, i);

		// Compute phi(a)
		pphi_a = VRING_ELEMENT(pring, ppmt->images[i]);

		for (j = 0; j < pring->num_elements; j++) {
			pb = VRING_ELEMENT(pring, j);

			// Compute phi(b)
			pphi_b = VRING_ELEMENT(pring, ppmt->images[j]);

			// Compute a*b
			pring->pspec->pvop(a_b, pa, pb, pring->pvaux);

			// Compute phi(a*b)

			if (!ring_find_index(pring, a_b, &ab_index)) {
				fprintf(stderr,
					"is_automorphism: ring not closed.\n");
				exit(1);
			}

			pphi_a_b = VRING_ELEMENT(pring,
				ppmt->images[ab_index]);
			// Compute phi(a) * phi(b)
			pring->pspec->pvop(phi_a_phi_b, pphi_a, pphi_b,
				pring->pvaux);

			// Check if phi(a) phi(b) = phi(ab)
			match = pring->pspec->pvequal(pphi_a_b,
				phi_a_phi_b, pring->pvaux);

			if (verbosity >= 4) {
				printf("---- [%2d/%2d,%2d/%2d]",
					i, pring->num_elements,
					j, pring->num_elements);
				printf(" a=");
				pring->pspec->pvprint(pa,
					pring->pvaux);

				printf(" b=");
				pring->pspec->pvprint(pb,
					pring->pvaux);

				printf(" ab=");
				pring->pspec->pvprint(a_b,
					pring->pvaux);

				printf(" phi(a)=");
				pring->pspec->pvprint(pphi_a,
					pring->pvaux);

				printf(" phi(b)=");
				pring->pspec->pvprint(pphi_b,
					pring->pvaux);

				printf(" phi(ab)=");
				pring->pspec->pvprint(pphi_a_b,
					pring->pvaux);

				printf(" %s ", match ? "==" : "!=");

				printf(" phi(a)phi(b)=");
				pring->pspec->pvprint(phi_a_phi_b,
					pring->pvaux);

				printf("\n");
			}

			// Check if phi(a) phi(b) = phi(ab)
			if (!match)
				return 0;
		}
	}
	return 1;
}

// ----------------------------------------------------------------
static void show_elements_and_orders(
	ring_t * pring,
	pmt_t    * psort_map,
	int     * orders)
{
	int i;
	void * pa;

	printf("Orders:\n");
	for (i = 0; i < pring->num_elements; i++) {
		pa = VRING_ELEMENT(pring, psort_map->images[i]);
		printf("  ");
		pring->pspec->pvprint(pa, pring->pvaux);
		printf(": %3d\n", orders[psort_map->images[i]]);
	}
	printf("\n");
}

// ----------------------------------------------------------------
// xxx to ring_order.c?
static void sort_elements_by_order(
	ring_t * pring,
	pmt_t    * psort_map,
	int     * orders)
{
	unsigned char temp_element[VOP_MAX_ELEMENT_SIZE];
	int temp_order;
	pmt_elt_t temp_elt;
	int i, top;
	void * pa, * pb;
	void * pt = &temp_element[0];
	int size = pring->pspec->element_size;

	for (top = pring->num_elements - 1; top > 0; top--) {
		for (i = 0; i < top; i++) {
			if (orders[i] > orders[i+1]) {

				pa = VRING_ELEMENT(pring, i);
				pb = VRING_ELEMENT(pring, i+1);
				memcpy(pt, pa, size);
				memcpy(pa, pb, size);
				memcpy(pb, pt, size);

				temp_order  = orders[i];
				orders[i]   = orders[i+1];
				orders[i+1] = temp_order;

				temp_elt = psort_map->images[i];
				psort_map->images[i] = psort_map->images[i+1];
				psort_map->images[i+1] = temp_elt  ;
			}
		}
	}
}

// ----------------------------------------------------------------
// Sample input: 1 2 5 5 5 5 10 10 10 10
// Output: 4
static int count_same_order_subsets(
	int  * orders,
	int    num_elements)
{
	int rv = 1;
	int i;
	int last = orders[0];

	for (i = 0; i < num_elements; i++) {
		if (orders[i] != last) {
			last = orders[i];
			rv++;
		}
	}
	return rv;
}

// ----------------------------------------------------------------
// Sample input: 1 2 5 5 5 5 10 10 10 10
// Output:
//   [0]  start_index = 0   count = 1  order = 1
//   [1]  start_index = 0   count = 1  order = 1
//   [2]  start_index = 0   count = 1  order = 1
//   [3]  start_index = 0   count = 1  order = 1

static void populate_sos_descs(
	int        * orders,
	int          num_elements,
	sos_desc_t * sos_descs,
	int          num_sos,
	int        * pnum_pmts)
{
	int i;
	int last = orders[0];
	int didx = 0;

	sos_descs[didx].start_index = 0;
	sos_descs[didx].count = 0;
	sos_descs[didx].order = orders[0];

	for (i = 0; i < num_elements; i++) {
		if (orders[i] != last) {
			didx++;
			sos_descs[didx].start_index = i;
			sos_descs[didx].count = 1;
			sos_descs[didx].order = orders[i];
			last = orders[i];
		}
		else {
			sos_descs[didx].count++;
		}
	}

	*pnum_pmts = 1;
	for (didx = 0; didx < num_sos; didx++) {
		if (sos_descs[didx].count > 10) {
			fprintf(stderr, "autck:  too many permutations.\n");
			exit(1);
		}
		sos_descs[didx].npmts = calc_factorial(
			sos_descs[didx].count);
		*pnum_pmts *= sos_descs[didx].npmts;
		if (*pnum_pmts > 10000000) {
			fprintf(stderr, "autck:  too many permutations.\n");
			exit(1);
		}
	}
}

// ----------------------------------------------------------------
static void print_sos_descs(
	sos_desc_t * sos_descs,
	int          num_sos)
{
	int didx;
	printf("Same-order subsets:\n");
	for (didx = 0; didx < num_sos; didx++) {
		printf("  [%d] si=%2d  ct=%2d  ord=%2d npmt=%6d\n",
			didx,
			sos_descs[didx].start_index,
			sos_descs[didx].count,
			sos_descs[didx].order,
			sos_descs[didx].npmts);
	}
}

// ----------------------------------------------------------------
static void decompose_pmt_indices(
	sos_desc_t * sos_descs,
	int          num_sos,
	int          whichp,
	int        * pmt_indices)
{
	int i;

	for (i = num_sos - 1; i >= 0; i--) {
		pmt_indices[i] = whichp % sos_descs[i].npmts;
		whichp /= sos_descs[i].npmts;
	}
}

// ----------------------------------------------------------------
static void compute_permutation(
	int  * pmt_indices,
	sos_desc_t * sos_descs,
	int    num_sos,
	int    ring_num_elements,
	pmt_t * pid,
	pmt_t * pout)
{
	pmt_t temp;
	int sosi;

	for (sosi = 0; sosi < num_sos; sosi++) {
		compute_subpermutation(pid, &temp, pout,
			sos_descs[sosi].start_index, sos_descs[sosi].count,
			sos_descs[sosi].npmts, pmt_indices[sosi]);
	}
}

// ----------------------------------------------------------------
static void compute_subpermutation(
	pmt_t * pin,
	pmt_t * ptemp,
	pmt_t * pout,
	int    start,
	int    count,
	int    nfact,
	int    whichp)
{
	int     pos, q, r, i, ni;
	int     nifact = nfact;

	for (i = 0; i < count; i++)
		ptemp->images[start + i] = pin->images[start + i];

	ni = count;
	for (pos = 0; pos < count; pos++) {
		nifact /= ni;
		r = whichp % nifact;
		q = whichp / nifact;
		whichp = r;
		pout->images[start+pos] = ptemp->images[start+q];
		for (i = q; i < ni; i++)
			ptemp->images[start+i] = ptemp->images[start+i+1];
		ni--;
	}
}

// ----------------------------------------------------------------
static void print_elements_with_sos(
	ring_t    * pring,
	pmt_t       * ppmt,
	sos_desc_t * sos_descs,
	int          num_sos)
{
	int i;
	int sosi = 0;
	void * pa;

	printf("-- | ");
	for (i = 0; i < pring->num_elements; i++) {
		pa = VRING_ELEMENT(pring, ppmt->images[i]);
		if (i != 0) {
			printf(" ");
			for (sosi = 0; sosi < num_sos; sosi++) {
				if (sos_descs[sosi].start_index == i)
					printf("| ");
			}
		}
		pring->pspec->pvprint(pa, pring->pvaux);
	}
	printf(" |\n");
}
