#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "vmatrix.h"

// ----------------------------------------------------------------
void vmatrix_print(
	void     * pvmat,
	void     * pvaux,
	vfops_t  * pvfops,
	int        nrows,
	int        ncols,
	int        lda)
{
	int i, j;
	void * pa;

	for (i = 0; i < nrows; i++) {
		for (j = 0; j < ncols; j++) {
			if (j > 0)
				printf(" ");
			pa = VMATRIX_ELEMENT(pvmat, i, j, lda,
				pvfops->element_size);
			pvfops->pvprint(pa, pvaux);
		}
		printf("\n");
	}
	printf("\n");
}

// ----------------------------------------------------------------
// Results are undefined if pout points to the same space as pvin1 or pvin2.
void vmatrix_add(
	void     * pvin1,
	void     * pvin2,
	void     * pvout,
	void     * pvaux,
	vfops_t  * pvfops,
	int        nrows,
	int        ncols,
	int        ldain1,
	int        ldain2,
	int        ldaout)
{
	int i, j;
	void * pa, * pb, * pc;

	for (i = 0; i < nrows; i++) {
		for (j = 0; j < ncols; j++) {
			pa = VMATRIX_ELEMENT(pvin1, i, j, ldain1,
				pvfops->element_size);
			pb = VMATRIX_ELEMENT(pvin2, i, j, ldain2,
				pvfops->element_size);
			pc = VMATRIX_ELEMENT(pvout, i, j, ldaout,
				pvfops->element_size);
			pvfops->pvadd(pa, pb, pc, pvaux);
		}
	}
}

// ----------------------------------------------------------------
// Results are undefined if pout points to the same space as pvin1 or pvin2.
void vmatrix_mul(
	void     * pvin1,
	void     * pvin2,
	void     * pvout,
	void     * pvaux,
	vfops_t  * pvfops,
	int        nrows1,
	int        ncols1,
	int        nrows2,
	int        ncols2,
	int        ldain1,
	int        ldain2,
	int        ldaout)
{
	int i, j, k;
	void * pa, * pb, * pc;
	unsigned char temp[VOP_MAX_ELEMENT_SIZE];

	if (ncols1 != nrows2) {
		fprintf(stderr,
		"vmatrix_mul:  Incompatible dimensions %d,%d x %d,%d\n",
			nrows1, ncols1, nrows2, ncols2);
		exit(1);
	}

	// xxx to lib func.
	if (pvfops->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: element size %d too big.\n",
				__FILE__, pvfops->element_size);
		exit(1);
	}

	for (i = 0; i < nrows1; i++) {
		for (j = 0; j < ncols2; j++) {
			pc = VMATRIX_ELEMENT(pvout,i,j,ldaout,
				pvfops->element_size);
			// out[i][j] = 0
			pvfops->pvmake_add_id(pc, pvaux);

			for (k = 0; k < ncols1; k++) {
				// out[i][j] += in1[i][k] + in2[k][j]
				pa = VMATRIX_ELEMENT(pvin1,i,k,ldain1,
					pvfops->element_size);
				pb = VMATRIX_ELEMENT(pvin2,k,j,ldain2,
					pvfops->element_size);
				pvfops->pvmul(pa, pb, temp, pvaux);
				pvfops->pvadd(pc, temp, pc, pvaux);
			}
		}
	}
}

// ----------------------------------------------------------------
void vmatrix_det(
	void     * pvmat, // In
	void     * pvdet, // Out
	void     * pvaux,
	vfops_t  * pvfops,
	int        dim,
	int        lda)
{
	void * temp;
	void * pa, * pt;
	int i, j;
	int nswaps;
	unsigned char zero[VOP_MAX_ELEMENT_SIZE];

	// xxx to lib func.
	if (pvfops->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: element size %d too big.\n",
				__FILE__, pvfops->element_size);
		exit(1);
	}

	temp = malloc(dim * dim * pvfops->element_size);
	if (temp == 0) {
		fprintf(stderr, "malloc(%d) failed.\n", dim * dim *
			pvfops->element_size);
		exit(1);
	}

	for (i = 0; i < dim; i++) {
		for (j = 0; j < dim; j++) {
			pa = VMATRIX_ELEMENT(pvmat, i, j, lda,
				pvfops->element_size);
			pt = VMATRIX_ELEMENT(temp , i, j, dim,
				pvfops->element_size);
			memcpy(pt, pa, pvfops->element_size);
		}
	}

	vmatrix_row_reduce_above(temp, pvaux, pvfops, dim, dim, dim, &nswaps);

	// det starts with 1.  Then, multiply in all diagonal elements of
	// the row-reduced matrix.
	pvfops->pvmake_mul_id(pvdet, pvaux);
	for (i = 0; i < dim; i++) {
		pt = VMATRIX_ELEMENT(temp, i, i, dim, pvfops->element_size);
		pvfops->pvmul(pvdet, pt, pvdet, pvaux);
	}
	if (nswaps & 1) {
		pvfops->pvmake_add_id(zero, pvaux);
		pvfops->pvsub(zero, pvdet, pvdet, pvaux);
	}

	free(temp);
}

// ----------------------------------------------------------------
static void demote_leading_zero_rows(
	void     * pvmat,
	void     * pvaux,
	vfops_t  * pvfops,
	int        nrows,
	int        ncols,
	int        lda,
	int        column,
	void     * ptemp_row,
	int        row_bytes,
	int      * pnswap)
{
	int i, top;
	void * plead;
	void * pa, * pb, * pt = ptemp_row;
	unsigned char zero[VOP_MAX_ELEMENT_SIZE];
	int nswap = *pnswap;

	// xxx to lib func.
	if (pvfops->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: element size %d too big.\n",
				__FILE__, pvfops->element_size);
		exit(1);
	}

	pvfops->pvmake_add_id(zero, pvaux);

	for (top = nrows - 1; top > column; top--) {
		for (i = column; i < top; i++) {
			plead = VMATRIX_ELEMENT(pvmat, i, column, lda,
				pvfops->element_size);
			if (pvfops->pvequal(plead, zero, pvaux)) {
				nswap++;
				pa = VMATRIX_ELEMENT(pvmat,i,  0, lda,
					pvfops->element_size);
				pb = VMATRIX_ELEMENT(pvmat,i+1,0, lda,
					pvfops->element_size);

				memcpy(pt, pa, row_bytes);
				memcpy(pa, pb, row_bytes);
				memcpy(pb, pt, row_bytes);
			}
		}
	}
	*pnswap = nswap;
}

// ----------------------------------------------------------------
void vmatrix_row_reduce_above(
	void     * pvmat,
	void     * pvaux,
	vfops_t  * pvfops,
	int        nrows,
	int        ncols,
	int        lda,
	int      * pnswaps)
{
	int i, j, k;
	void * temp_row;
	void * puplead;
	void * pdownlead;
	void * pupelt;
	void * pdownelt;
	int row_bytes = lda * pvfops->element_size;
	unsigned char zero[VOP_MAX_ELEMENT_SIZE];
	unsigned char clear_factor[VOP_MAX_ELEMENT_SIZE];
	unsigned char temp[VOP_MAX_ELEMENT_SIZE];

	*pnswaps = 0;

	// xxx to lib func.
	if (pvfops->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: element size %d too big.\n",
				__FILE__, pvfops->element_size);
		exit(1);
	}
	pvfops->pvmake_add_id(zero, pvaux);

	temp_row = malloc(row_bytes);
	if (temp_row == 0) {
		fprintf(stderr, "malloc(%d) failed.\n", row_bytes);
		exit(1);
	}
#ifdef VMATRIX_TRACE
	printf("\n");
	printf("row red start:\n");
	vmatrix_print(pvmat, pvaux, pvfops, nrows, ncols, lda);
#endif //VMATRIX_TRACE

	for (j = 0; j < ncols-1; j++) {

#ifdef VMATRIX_TRACE
		printf("\n");
		printf("pre demote col %d:\n", j);
		vmatrix_print(pvmat, pvaux, pvfops, nrows, ncols, lda);
#endif //VMATRIX_TRACE

		// Put all leading-zero rows at the bottom, and
		// non-leading-zero rows before.
		demote_leading_zero_rows(pvmat, pvaux, pvfops,
			nrows, ncols, lda, j, temp_row, row_bytes, pnswaps);

#ifdef VMATRIX_TRACE
		printf("\n");
		printf("post demote col %d:\n", j);
		vmatrix_print(pvmat, pvaux, pvfops, nrows, ncols, lda);
#endif //VMATRIX_TRACE

		puplead = VMATRIX_ELEMENT(pvmat, j, j, lda,
			pvfops->element_size);
		if (pvfops->pvequal(puplead, zero, pvaux))
			continue;
#ifdef VMATRIX_TRACE
		printf("uplead   %d,%d = ", j, j);
		pvfops->pvprint(puplead, pvaux);
		printf("\n");
#endif //VMATRIX_TRACE

		for (i = j+1; i < nrows; i++) {
			pdownlead = VMATRIX_ELEMENT(pvmat, i, j, lda,
				pvfops->element_size);
			if (pvfops->pvequal(pdownlead, zero, pvaux))
				continue;
			pvfops->pvdiv(pdownlead, puplead, clear_factor, pvaux);

#ifdef VMATRIX_TRACE
			printf("downlead %d,%d = ", i, j);
			pvfops->pvprint(pdownlead, pvaux);
			printf("\n");

			printf("subtract ");
			pvfops->pvprint(clear_factor, pvaux);
			printf(" times row %d from row %d\n", j, i);
#endif //VMATRIX_TRACE

			for (k = 0; k < ncols; k++) {
				pupelt = VMATRIX_ELEMENT(pvmat, j, k, lda,
					pvfops->element_size);
				pdownelt = VMATRIX_ELEMENT(pvmat, i, k, lda,
					pvfops->element_size);
				pvfops->pvmul(pupelt, clear_factor,
					temp, pvaux);
				pvfops->pvsub(pdownelt, temp,
					pdownelt, pvaux);
			}
#ifdef VMATRIX_TRACE
			printf("\n");
			printf("post subtraction:\n");
			vmatrix_print(pvmat, pvaux, pvfops, nrows, ncols, lda);
#endif // VMATRIX_TRACE

			// row[i] := row[i] - (row[j] / row[j][j] * row[i][j])
			// row[i] := row[i] - (row[j] / uplead * downlead)
		}
	}

	free(temp_row);
}
