// Symmetries of the cube (same as symmetries of the octahedron when
// faces are swapped with vertices):

//    E---F
//   /|  /|
//  / G-/-H
// A---B /
// |/  |/
// C---D

// These can be viewed as a constrained subset of S_8.  Alternatively,
// they are generated by pitch (P), roll (R) and yaw (Y), each taken to
// be clockwise as viewed from the right, front and top, respectively.

//    E---F     B---D     C---G        D---C     A---E     F---H
//   /|  /|    /|  /|    /|  /|       /|  /|    /|  /|    /|  /|
//  / G-/-H   / F-/-H   / D-/-H      / H-/-G   / C-/-G   / E-/-G
// A---B /   A---C /   A---E /      B---A/    B---F /   B---D /
// |/  |/    |/  |/    |/  |/       |/  |/    |/  |/    |/  |/
// C---D     E---G     B---F        F---E     D---H     A---C
//
// p0y0r0    p?y?r?    p?y?r?       p?y?r?    p?y?r?    p?y?r?
//
//    G---E     D---H     A---B        H---G     C---A     B---F
//   /|  /|    /|  /|    /|  /|       /|  /|    /|  /|    /|  /|
//  / H-/-F   / B-/-F   / E-/-F      / F-/-E   / G-/-E   / A-/-E
// C---A /   C---G /   C---D /      D---C /   D---B /   D---H /
// |/  |/    |/  |/    |/  |/       |/  |/    |/  |/    |/  |/
// D---B     A---E     G---H        B---A     H---F     C---G
//
// p?y?r?    p?y?r?    p1y0r0       p?y?r?    p?y?r?    p?y?r?
//
//    F---B     A---C     G---H        H---G     H---D     E---G
//   /|  /|    /|  /|    /|  /|       /|  /|    /|  /|    /|  /|
//  / H-/-D   / B-/-D   / C-/-D      / D-/-C   / G-/-C   / A-/-C
// E---A /   E---G /   E---F /      F---E /   F---B /   F---H/
// |/  |/    |/  |/    |/  |/       |/  |/    |/  |/    |/  |/
// G---C     F---H     A---B        B---A     E---A     B---D
//
// p?y?r?    p?y?r?    p?y?r?       p?y?r?    p?y?r?    p?y?r?
//
//    E---A     H---F     C---D        D---B     D---C     G---C
//   /|  /|    /|  /|    /|  /|       /|  /|    /|  /|    /|  /|
//  / F-/-B   / D-/-B   / A-/-B      / C-/-A   / B-/-A   / E-/-A
// G---C /   G---E/    G---H /      H---F /   H---G /   H---D /
// |/  |/    |/  |/    |/  |/       |/  |/    |/  |/    |/  |/
// H---D     C---A     E---F        G---E     F---E     F---B
//
// p?y?r?    p?y?r?    p?y?r?       p?y?r?    p?y?r?    p?y?r?


//   p0y0r0 p0y0r1 p0y0r2 p0y0r3
//   p0y1r0 p0y1r1 p0y1r2 p0y1r3
//   p0y2r0 p0y2r1 p0y2r2 p0y2r3
//   p0y3r0 p0y3r1 p0y3r2 p0y3r3

//   p1y0r0 p1y0r1 p1y0r2 p1y0r3
//   p1y1r0 p1y1r1 p1y1r2 p1y1r3
//   p1y2r0 p1y2r1 p1y2r2 p1y2r3
//   p1y3r0 p1y3r1 p1y3r2 p1y3r3

//   p2y0r0 p2y0r1 p2y0r2 p2y0r3
//   p2y1r0 p2y1r1 p2y1r2 p2y1r3
//   p2y2r0 p2y2r1 p2y2r2 p2y2r3
//   p2y3r0 p2y3r1 p2y3r2 p2y3r3

//   p3y0r0 p3y0r1 p3y0r2 p3y0r3
//   p3y1r0 p3y1r1 p3y1r2 p3y1r3
//   p3y2r0 p3y2r1 p3y2r2 p3y2r3
//   p3y3r0 p3y3r1 p3y3r2 p3y3r3

// This gives me 64, not 24, so there must be dups.
// What is the simplification rule?

// Identity:    Pitch:       Roll:        Yaw:
//    E---F        A---B        G---E        A---E
//   /|  /|       /|  /|       /|  /|       /|  /|
//  / G-/-H      / E-/-F      / H-/-F      / C-/-G
// A---B /      C---D /      C---A /      B---F /
// |/  |/       |/  |/       |/  |/       |/  |/
// C---D        G---H        D---B        D---H
//
// ABCDEFGH     CDGHABEF     CADBGEHF     BFDHAECG

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cube.h"
#include "pmt.h"

// ----------------------------------------------------------------
static void cube_vprint(
	void * pvin,
	void * pvaux)
{
	pmt_t * ps = (pmt_t *)pvin;
	int i;

	for (i = 0; i < 8; i++) {
		printf("%c", ps->images[i] + 'A');
	}
}

// ----------------------------------------------------------------
int cube_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	pmt_t * ppmt = (pmt_t *)pvout;
	int i;

	if (sizeof(pmt_t) > max_elt_bytes) {
		fprintf(stderr, "cube:  Need %d bytes for element.\n",
			sizeof(pmt_t));
		exit(1);
	}

	if (strlen(pin) != 8) {
		fprintf(stderr,
			"cube element input \"%s\" must have length 8.\n",
			pin);
		return 0;
	}

	for (i = 0; i < 8; i++) {
		if ((pin[i] >= 'A') && (pin[i] <= 'Z'))
			ppmt->images[i] = pin[i] - 'A';
		else if ((pin[i] >= 'a') && (pin[i] <= 'z'))
			ppmt->images[i] = pin[i] - 'a';
		else if ((pin[i] >= '0') && (pin[i] <= '9'))
			ppmt->images[i] = pin[i] - '0';
		else {
			fprintf(stderr,
				"Malformed cube element \"%s\".\n",
				pin);
			return 0;
		}
	}

	if (!pmt_validate_element(ppmt, 8)) {
		fprintf(stderr,
			"cube element \"%s\" is not a permuation.\n",
			pin);
		return 0;
	}

	return 1;
}


// ----------------------------------------------------------------
int cube_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	if (argc != 0)
		return 0;
	*(int *)pvaux = 8;
	return 1;
}

// ----------------------------------------------------------------
int  cube_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	pmt_t * pelements = (pmt_t *) pvelements;
	unsigned char * psrc = 0, * pdst = 0;
	int i, j;

	static char * strings[] = {
//"ABCDEFGH", "ACEGBDFH", "AEBFCGDH", "BAFEDCHG", "BFDHAECG", "BDACFHEG",
//"CADBGEHF", "CGAEDHBF", "CDGHABEF", "DCBAHGFE", "DBHFCAGE", "DHCGBFAE",
//"EAGCFBHD", "EGFHACBD", "EFABGHCD", "FEBAHGDC", "FBEAHDGC", "FHBDEGAC",
//"GCHDEAFB", "GECAHFDB", "GHEFCDAB", "HFGEDBCA", "HGFEDCBA", "HDFBGCEA",


	"ABCDEFGH", "ACEGBDFH", "AEBFCGDH", "BAFEDCHG", "BDACFHEG", "BFDHAECG",
	"CADBGEHF", "CDGHABEF", "CGAEDHBF", "DBHFCAGE", "DCBAHGFE", "DHCGBFAE",
	"EAGCFBHD", "EFABGHCD", "EGFHACBD", "FBEAHDGC", "FEHGBADC", "FHBDEGAC",
	"GCHDEAFB", "GECAHFDB", "GHEFCDAB", "HDFBGCEA", "HFGEDBCA", "HGDCFEBA"

	};

	*pnum_elements = 24;

	if (*pnum_elements > max_group_bytes/sizeof(pmt_t))
		return 0;

	for (i = 0; i < 24; i++) {
		psrc = strings[i];
		pdst = pelements[i].images;
		for (j = 0; j < 8; j++)
			pdst[j] = psrc[j] - 'A';
	}

	return 1;
}

// ================================================================
group_spec_t cube_group_spec = {
	element_size  : sizeof(pmt_t),
	naux          : 0,
	pvop          : pmt_vmul,
	pvequal       : pmt_vequal,
	pvless        : pmt_vless,
	pvprint       : cube_vprint,
	pvauxprint     : 0,
	pvscan        : cube_vscan,
	pvauxscan      : cube_vauxscan,
	pfill_group   : cube_fill_group,
	desc          : "cube"
};
