// Computation of cyclotomic polynomials:
//
// x^n - 1 = product over d dividing n of Phi_d(n).
// E.g. x^8 - 1 = Phi_1(x) * Phi_2(x) * Phi_4(x) * Phi_8(x).
// So, Phi_n(x) = (x^n - 1) / product over d dividing n and d<n of Phi_d(n).
// E.g.  Phi_8(x) = (x^8 - 1) / (Phi_1(x) * Phi_2(x) * Phi_4(x)).

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cyclopoly.h"

#define CYCLO_MEMO_MAX 16
static int     zcyclo_memo_flags[CYCLO_MEMO_MAX+1] = {0};
static zpoly_t zcyclo_memo_data [CYCLO_MEMO_MAX+1];

static int     qcyclo_memo_flags[CYCLO_MEMO_MAX+1] = {0};
static qpoly_t qcyclo_memo_data [CYCLO_MEMO_MAX+1];

// ----------------------------------------------------------------
zpoly_t zpoly_find_nth_cyclo(int n)
{
	int d;
	zpoly_t rv;
	zpoly_t dth;
	zpoly_t rem;

	if (n < 1) {
		fprintf(stderr, "zpoly_find_nth_cyclo: argument must be "
			"positive; got %d.\n", n);
		exit(1);
	}

	if ((n <= CYCLO_MEMO_MAX) && (zcyclo_memo_flags[n]))
		return zcyclo_memo_data[n];

	if (n == 1) {
		rv.degree = 1;
		rv.coeffs[1] =  1;
		rv.coeffs[0] = -1;

		// update memo
		if (n <= CYCLO_MEMO_MAX) {
			zcyclo_memo_data[n] = rv;
			zcyclo_memo_flags[n] = 1;
		}

		return rv;
	}

	// x^n - 1
	rv = zpoly_binomial_from_coeffs(1, n, -1, 0);

	for (d = 1; d < n; d++) {
		if ((n % d) != 0)
			continue;

		if ((d <= CYCLO_MEMO_MAX) && (zcyclo_memo_flags[d]))
			dth = zcyclo_memo_data[d];
		else
			dth = zpoly_find_nth_cyclo(d);

		zpoly_quot_and_rem(&rv, &dth, &rv, &rem);
		if (!zpoly_is_zero(&rem)) {
			fprintf(stderr,
				"zpoly_find_nth_cyclo: internal error.\n");
			exit(1);
		}
	}

	// update memo
	if (n <= CYCLO_MEMO_MAX) {
		zcyclo_memo_data[n] = rv;
		zcyclo_memo_flags[n] = 1;
	}

	return rv;
}

// ----------------------------------------------------------------
qpoly_t qpoly_find_nth_cyclo(int n)
{
	int d;
	qpoly_t rv;
	qpoly_t dth;
	qpoly_t rem;

	if (n < 1) {
		fprintf(stderr, "qpoly_find_nth_cyclo: argument must be "
			"positive; got %d.\n", n);
		exit(1);
	}

	if ((n <= CYCLO_MEMO_MAX) && (qcyclo_memo_flags[n]))
		return qcyclo_memo_data[n];

	if (n == 1) {
		rv.degree = 1;
		rv.coeffs[1] = rat_from_int( 1);
		rv.coeffs[0] = rat_from_int(-1);

		// update memo
		if (n <= CYCLO_MEMO_MAX) {
			qcyclo_memo_data[n] = rv;
			qcyclo_memo_flags[n] = 1;
		}

		return rv;
	}

	// x^n - 1
	rv = qpoly_binomial_from_coeffs(
		rat_from_int(1), n,
		rat_from_int(-1), 0);

	for (d = 1; d < n; d++) {
		if ((n % d) != 0)
			continue;

		if ((d <= CYCLO_MEMO_MAX) && (qcyclo_memo_flags[d]))
			dth = qcyclo_memo_data[d];
		else
			dth = qpoly_find_nth_cyclo(d);

		qpoly_quot_and_rem(&rv, &dth, &rv, &rem);
		if (!qpoly_is_zero(&rem)) {
			fprintf(stderr,
				"qpoly_find_nth_cyclo: internal error.\n");
			exit(1);
		}
	}

	// update memo
	if (n <= CYCLO_MEMO_MAX) {
		qcyclo_memo_data[n] = rv;
		qcyclo_memo_flags[n] = 1;
	}

	return rv;
}
