#include <stdio.h>
#include <stdlib.h>
#include "dih.h"

#include "phi.h"
#include "gcd.h"

// ----------------------------------------------------------------
// Elements of the dihedral group D_n are generated by a rotation r
// (of order n) and a flip f (of order 2):
//
//   D_n = { e, r, r^2, ..., r^(n-1),   f, r f, r^2 f, ..., r^(n-1) f }
//
// Simplification/symmetry rule:
//
//   r^k f = f r^(n-k)
//
// So, composition may be implemented as follows:
//
//   (r^a f^i) * (r^b f^j)
//     = r^a (f^i * r^b) f^j
//   Case i = 0:
//     = r^(a+b) f^j
//   Case i = 1:
//     = r^a (r^(n-b) * f^i) f^j
//     = r^(n+a-b) * f^(i XOR j)

dih_t dih_mul(dih_t a, dih_t b, int n)
{
	dih_t c;

	if (a.flip)
		c.rot = n + a.rot - b.rot;
	else
		c.rot = a.rot + b.rot;
	if (c.rot >= n)
		c.rot -= n;
	c.flip = a.flip ^ b.flip;
	return c;
}

// ----------------------------------------------------------------
int dih_equal(dih_t a, dih_t b, int n)
{
	if (a.flip != b.flip)
		return 0;
	if (a.rot != b.rot)
		return 0;
	return 1;
}

// ================================================================
static void dih_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	dih_t a = *(dih_t *)pvin1;
	dih_t b = *(dih_t *)pvin2;
	int  n = *(int  *)pvaux;
	dih_t c = dih_mul(a, b, n);
	*(dih_t *)pvout = c;
}

// ----------------------------------------------------------------
static int dih_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	dih_t a = *(dih_t *)pvin1;
	dih_t b = *(dih_t *)pvin2;
	int n = *(int *)pvaux;
	return dih_equal(a, b, n);
}

// ----------------------------------------------------------------
static int dih_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	dih_t a = *(dih_t *)pvin1;
	dih_t b = *(dih_t *)pvin2;
	if (a.flip < b.flip)
		return 1;
	else if (a.flip > b.flip)
		return 0;
	else if (a.rot < b.rot)
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
static void dih_vprint(
	void * pvin,
	void * pvaux)
{
	dih_t x = *(dih_t *)pvin;
	int n = *(int *)pvaux;
	if (n > 50)
		printf("%3d:%d", x.rot, x.flip);
		//printf("%3d",  x.rot  + n * x.flip);
	else if (n > 5)
		printf("%2d:%d", x.rot, x.flip);
		//printf("%2d",  x.rot  + n * x.flip);
	else
		printf("%d:%d",  x.rot, x.flip);
		//printf("%d",  x.rot  + n * x.flip);
}

// ----------------------------------------------------------------
static void dih_vauxprint(
	void * pvaux)
{
	printf("%d", *(int *)pvaux);
}

// ----------------------------------------------------------------
int dih_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	dih_t * pout = (dih_t *)pvout;

	if (sizeof(dih_t) > max_elt_bytes) {
		fprintf(stderr, "dn:  Need %d bytes for element.\n",
			sizeof(dih_t));
		exit(1);
	}

	return sscanf(pin, "%d:%d", &pout->rot, &pout->flip);
}

// ----------------------------------------------------------------
int dih_vauxscan(
	void * pvaux,
	int    argc,
	char **argv,
	int    max_aux_bytes)
{
	int modulus;

	if (max_aux_bytes < sizeof(int)) {
		fprintf(stderr, "dn:  Need %d bytes for aux.\n",
			sizeof(int));
		exit(1);
	}

	if (argc != 1) {
		fprintf(stderr, "dn:  Need 1 aux; got %d.\n",
			argc);
		return 0;
	}
	if (sscanf(argv[0], "%d", &modulus) != 1) {
		fprintf(stderr, "Couldn't scan \"%s\" as int.\n",
			argv[0]);
		exit(1);
	}
	if (modulus < 1) {
		fprintf(stderr, "Modulus must be >= 1; got %d.\n",
			modulus);
		exit(1);
	}
	*(int *)pvaux = modulus;
	return 1;
}

// ----------------------------------------------------------------
int  dih_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	dih_t * pelements = (dih_t *) pvelements;
	int n = *(int *)pvaux;
	int i;

	*pnum_elements = 2 * n;
	if (*pnum_elements > max_group_bytes/sizeof(dih_t))
		return 0;

	for (i = 0; i < n; i++) {
		pelements[i].rot  = i;
		pelements[i].flip = 0;
	}
	for (i = 0; i < n; i++) {
		pelements[i+n].rot  = i;
		pelements[i+n].flip = 1;
	}

	return 1;
}

// ================================================================
group_spec_t dn_group_spec = {
	element_size  : sizeof(dih_t),
	naux          : 1,
	pvop          : dih_vmul,
	pvequal       : dih_vequal,
	pvless        : dih_vless,
	pvprint       : dih_vprint,
	pvauxprint    : dih_vauxprint,
	pvscan        : dih_vscan ,
	pvauxscan     : dih_vauxscan,
	pfill_group   : dih_fill_group,
	desc          : "dn"
};
