#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fpell3.h"
#include "posmod.h"
#include "mod.h"
#include "malloc_check.h"

// Computations are done projectively; I/O may be affine if desired.  If so,
// the point at infinity is labelled "O".
#define USE_AFFINE_IO

typedef struct _fpell3_pt_t {
	int x;
	int y;
	int z;
} fpell3_pt_t;

typedef struct _fpell3_aux_t {
	int a;
	int b;
	int c;
	int p;
} fpell3_aux_t;

// ----------------------------------------------------------------
// Affine:     y^2  = x^3 + ax^2 + bx + c
// Projective: y^2z = x^3 + ax^2z + bxz^2 + cz^3

static int on_curve(int x, int y, int z, int a, int b, int c, int p)
{
	int lhs = posmod(y*y*z, p);
	int rhs = posmod(x*x*x, p) + posmod(a*x*x*z, p) + posmod(b*x*z*z, p)
		+ posmod(c*z*z*z, p);
	rhs = posmod(rhs, p);
	if (lhs == rhs)
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
static void check_point(fpell3_pt_t * ppt, fpell3_aux_t * paux, char * desc)
{
	if (!on_curve(ppt->x, ppt->y, ppt->z, paux->a, paux->b, paux->c, paux->p)) {
		fprintf(stderr,
	"fpell3:  the %s point [%d,%d,%d] is not on the curve y^2 = x^3 + %dx^2 + %dx + %d mod %d.\n",
			desc, ppt->x, ppt->y, ppt->z, paux->a, paux->b, paux->c, paux->p);
		exit(1);
	}
}

// ================================================================
static void fpell3_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	fpell3_pt_t  * p1   = (fpell3_pt_t  *)pvin1;
	fpell3_pt_t  * p2   = (fpell3_pt_t  *)pvin2;
	fpell3_pt_t  * p3   = (fpell3_pt_t  *)pvout;
	fpell3_aux_t * paux = (fpell3_aux_t *)pvaux;

	int a  = paux->a;
	int b  = paux->b;
	int p  = paux->p;

	int x1 = p1->x;
	int y1 = p1->y;
	int z1 = p1->z;
	int x2 = p2->x;
	int y2 = p2->y;
	int z2 = p2->z;
	int x3, y3, z3;
	int lambda, nu;

	if (x1 == 0 && y1 == 1 && z1 == 0) {
		// P1 is O
		x3 = x2;
		y3 = y2;
		z3 = z2;
	}
	else if (x2 == 0 && y2 == 1 && z2 == 0) {
		// P2 is O
		x3 = x1;
		y3 = y1;
		z3 = z1;
	}
	else if (x1 == x2 && y1 == y2 && z1 == z2) {
		if (y1 == 0LL) {
			x3 = 0LL;
			y3 = 1LL;
			z3 = 0LL;
		}
		else {
			// Duplication formula
			// lambda = (x1*x1*3 + a*x1*2 + b) / (y1*2);
			// nu = y1 - lambda * x1;
			// x3 = lambda*lambda - a - x1 - x1;
			// y3 = -(lambda*x3 + nu);
			// z3 = 1;
			lambda = posmod((posmod(x1*x1*3, p) + posmod(a*x1*2 + b, p))
				* mod_recip(posmod(y1*2,p), p), p);
			nu = posmod(y1 - lambda * x1, p);
			x3 = posmod(posmod(lambda*lambda,p) - a - x1 - x1, p);
			y3 = posmod(-(lambda*x3 + nu), p);
			z3 = 1;
		}
	}
	else if (x1 == x2) {
		// Vertical line
		x3 = 0;
		y3 = 1;
		z3 = 0;
	}
	else {
		// Product formula
		// lambda = (y2 - y1) / (x2 - x1);
		// nu = y1 - lambda * x1;
		// x3 = lambda*lambda - a - x1 - x2;
		// y3 = -(lambda*x3 + nu);
		// z3 = 1;
		lambda = posmod(posmod(y2 - y1, p) * mod_recip(x2 - x1, p), p);
		nu = posmod(y1 - lambda * x1, p);
		x3 = posmod(posmod(lambda*lambda, p) - a - x1 - x2, p);
		y3 = posmod(-lambda*x3 - nu, p);
		z3 = 1;
	}

	p3->x = x3;
	p3->y = y3;
	p3->z = z3;
	check_point(p3, paux, "computed");
}

// ----------------------------------------------------------------
static int fpell3_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	fpell3_pt_t * p1 = (fpell3_pt_t *)pvin1;
	fpell3_pt_t * p2 = (fpell3_pt_t *)pvin2;
	if (p1->x == p2->x && p1->y == p2->y && p1->z == p2->z)
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
static int fpell3_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	fpell3_pt_t * p1 = (fpell3_pt_t *)pvin1;
	fpell3_pt_t * p2 = (fpell3_pt_t *)pvin2;

	// The point at infinity comes first.
	if (p1->x == 0 && p1->y == 1 && p1->z == 0) {
		if (p2->x == 0 && p2->y == 1 && p2->z == 0)
			return 0;
		else
			return 1;
	}
	else if (p1->x < p2->x)
		return 1;
	else if (p1->x > p2->x)
		return 0;
	else if (p1->y < p2->y)
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
static void fpell3_vprint(
	void * pvin,
	void * pvaux)
{
	fpell3_pt_t  * ppt = (fpell3_pt_t  *)pvin;

#ifdef USE_AFFINE_IO
	if (ppt->x == 0 && ppt->y == 1 && ppt->z == 0)
		printf("O");
	else
		printf("%d,%d", ppt->x, ppt->y);
#else
	printf("%d,%d,%d", ppt->x, ppt->y, ppt->z);
#endif
}

// ----------------------------------------------------------------
static void fpell3_vauxprint(
	void * pvaux)
{
	fpell3_aux_t * paux = (fpell3_aux_t *)pvaux;
	printf("%d,%d,%d", paux->a, paux->b, paux->c);
}

// ----------------------------------------------------------------
int fpell3_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	fpell3_pt_t  * pout = (fpell3_pt_t  *)pvout;
	fpell3_aux_t * paux = (fpell3_aux_t *)pvaux;

	if (sizeof(fpell3_pt_t) > max_elt_bytes) {
		fprintf(stderr, "fpell3:  Need %d bytes for element.\n",
			sizeof(fpell3_pt_t));
		exit(1);
	}

#ifdef USE_AFFINE_IO
	if (strcmp(pin, "O") == 0) {
		pout->x = 0;
		pout->y = 1;
		pout->z = 0;
	}
	else if (sscanf(pin, "%d,%d", &pout->x, &pout->y) == 2) {
		pout->z = 1;
	}
	else {
		return 0;
	}
#else
	if (sscanf(pin, "%d,%d,%d", &pout->x, &pout->y, &pout->z) != 3)
		return 0;
	pout->x = posmod(pout->x, paux->p);
	pout->y = posmod(pout->y, paux->p);
	pout->z = posmod(pout->z, paux->p);
#endif

	check_point(pout, paux, "input");

	return 1;
}

// ----------------------------------------------------------------
int fpell3_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	fpell3_aux_t aux;

	if (max_aux_bytes < sizeof(fpell3_aux_t)) {
		fprintf(stderr, "fpell3:  Need %d bytes for aux.\n",
			sizeof(fpell3_aux_t));
		exit(1);
	}

	if (argc != 2) {
		fprintf(stderr, "fpell3_vauxscan: need argc == 2; got %d.\n", argc);
		return 0;
	}
	if (sscanf(argv[0], "%d", &aux.p) != 1) {
		fprintf(stderr, "fpell3_vauxscan: couldn't scan p from \"%s\".\n", argv[0]);
		return 0;
	}
	if (sscanf(argv[1], "%d,%d,%d", &aux.a, &aux.b, &aux.c) != 3) {
		fprintf(stderr, "fpell3_vauxscan: couldn't scan a,b,c from \"%s\".\n", argv[1]);
		return 0;
	}
	aux.a = posmod(aux.a, aux.p);
	aux.b = posmod(aux.b, aux.p);
	aux.c = posmod(aux.c, aux.p);
	*(fpell3_aux_t *)pvaux = aux;
	return 1;
}

// ----------------------------------------------------------------
int fpell3_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	fpell3_pt_t * pelements = (fpell3_pt_t *) pvelements;
	fpell3_aux_t * paux = (fpell3_aux_t *)pvaux;
	int maxn = max_group_bytes / sizeof(fpell3_pt_t);
	int n = 0;
	int x, y;
	int p = paux->p;
	int a = paux->a;
	int b = paux->b;
	int c = paux->c;

	// Point at infinity.
	pelements[n].x = 0;
	pelements[n].y = 1;
	pelements[n].z = 0;
	n++;

	// Sanity check
	check_point(&pelements[0], paux, "infinite");

	// Affine points.
	for (x = 0; x < p; x++) {
		for (y = 0; y < p; y++) {
			if (on_curve(x, y, 1, a, b, c, p)) {
				if (n > maxn)
					return 0;
				pelements[n].x = x;
				pelements[n].y = y;
				pelements[n].z = 1;
				n++;
			}
		}
	}
	*pnum_elements = n;
	return 1;
}

// ================================================================
group_spec_t fpell3_group_spec = {
	element_size  : sizeof(fpell3_pt_t),
	naux          : 2,
	pvop          : fpell3_vadd,
	pvequal       : fpell3_vequal,
	pvless        : fpell3_vless,
	pvprint       : fpell3_vprint,
	pvauxprint    : fpell3_vauxprint,
	pvscan        : fpell3_vscan,
	pvauxscan     : fpell3_vauxscan,
	pfill_group   : fpell3_fill_group,
	desc          : "fpell3"
};
