#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ftell5.h"
#include "z2ipolymod.h"
#include "malloc_check.h"

// Computations are done projectively; I/O may be affine if desired.  If so,
// the point at infinity is labelled "O".
#define USE_AFFINE_IO

typedef struct _ftell5_pt_t {
	z2ipoly_t x;
	z2ipoly_t y;
	z2ipoly_t z;
} ftell5_pt_t;

typedef struct _ftell5_aux_t {
	z2ipoly_t a1;
	z2ipoly_t a2;
	z2ipoly_t a3;
	z2ipoly_t a4;
	z2ipoly_t a6;
	z2ipoly_t m;
} ftell5_aux_t;

// ----------------------------------------------------------------
#define mul(a,b,m) z2ipolymod_mul((a),(b),(m))
#define div(a,b,m) z2ipolymod_div((a),(b),(m))

static int on_curve(
	z2ipoly_t x,  z2ipoly_t y,  z2ipoly_t z,
	z2ipoly_t a1, z2ipoly_t a2, z2ipoly_t a3, z2ipoly_t a4, z2ipoly_t a6,
	z2ipoly_t m)
{
	// y^2 + a1 x y + a3 y + x^3 + a2 x^2 + a4 x + a6
	// y^2 z + a1 x y z + a3 y z^2 + x^3 + a2 x^2 z + a4 x z^2 + a6 z^3

	z2ipoly_t f =
		mul(mul(y, y, m), z, m) ^
		mul(mul(a1, x, m), mul(y, z, m), m) ^
		mul(mul(a3, y, m), mul(z, z, m), m) ^
		mul(mul(x, x, m), x, m) ^
		mul(mul(a2, x, m), mul(x, z, m), m) ^
		mul(mul(a4, x, m), mul(z, z, m), m) ^
		mul(mul(a6, z, m), mul(z, z, m), m);

	if (f == 0LL)
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
static void check_point(ftell5_pt_t * ppt, ftell5_aux_t * paux, char * desc)
{
	int oc = on_curve(ppt->x, ppt->y, ppt->z, paux->a1, paux->a2, paux->a3,
		paux->a4, paux->a6, paux->m);
	if (!oc) {
		fprintf(stderr,
	"ftell5:  the %s point [%llx,%llx,%llx] is not on the curve "
	"y^2 + %llx x y + %llx y= x^3 + %llx x^2 + %llx x + %llx mod %llx.\n",
			desc, ppt->x, ppt->y, ppt->z,
			paux->a1, paux->a3, paux->a2, paux->a4, paux->a6, paux->m);
		exit(1);
	}
}

// ================================================================
static void ftell5_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	ftell5_pt_t  * p1   = (ftell5_pt_t  *)pvin1;
	ftell5_pt_t  * p2   = (ftell5_pt_t  *)pvin2;
	ftell5_pt_t  * p3   = (ftell5_pt_t  *)pvout;
	ftell5_aux_t * paux = (ftell5_aux_t *)pvaux;

	z2ipoly_t a1 = paux->a1;
	z2ipoly_t a2 = paux->a2;
	z2ipoly_t a3 = paux->a3;
	z2ipoly_t a4 = paux->a4;
	z2ipoly_t m  = paux->m;

	z2ipoly_t x1 = p1->x;
	z2ipoly_t y1 = p1->y;
	z2ipoly_t z1 = p1->z;
	z2ipoly_t x2 = p2->x;
	z2ipoly_t y2 = p2->y;
	z2ipoly_t z2 = p2->z;
	z2ipoly_t x3, y3, z3;
	z2ipoly_t lambda, nu;

	if (x1 == 0LL && y1 == 1LL && z1 == 0LL) {
		// P1 is O
		x3 = x2;
		y3 = y2;
		z3 = z2;
	}
	else if (x2 == 0LL && y2 == 1LL && z2 == 0LL) {
		// P2 is O
		x3 = x1;
		y3 = y1;
		z3 = z1;
	}
	else if (x1 == x2 && y1 == y2 && z1 == z2) {
		// Duplication formula:  Silverman pp. 58-59.
		z2ipoly_t x12 = mul(x1, x1, m);
		lambda = div((x12 ^ a4 ^ mul(a1,y1,m)), (mul(a1,x1,m) ^ a3), m);
		nu = div(
			(mul(x1,x12,m) ^ mul(a4,x1,m) ^ mul(a3,y1,m)),
			mul(a1,x1,m) ^ a3,
			m);

		x3 = mul(lambda,lambda,m) ^ mul(a1,lambda,m) ^ a2 ^ x1 ^ x2;
		y3 = mul((lambda ^ a1),x3,m) ^ nu ^ a3;
		z3 = 1LL;
	}
	else if (x1 == x2) {
		// Vertical line
		x3 = 0LL;
		y3 = 1LL;
		z3 = 0LL;
	}
	else {
		// Product formula:  Silverman pp. 58-59.
		lambda = div((y2 ^ y1), (x2 ^ x1), m);
		nu = div((mul(y1,x2,m) ^ mul(y2,x1,m)), (x2 ^ x1), m);

		x3 = mul(lambda,lambda,m) ^ mul(a1,lambda,m) ^ a2 ^ x1 ^ x2;
		y3 = mul((lambda ^ a1),x3,m) ^ nu ^ a3;
		z3 = 1LL;
	}

	p3->x = x3;
	p3->y = y3;
	p3->z = z3;
	check_point(p3, paux, "computed");
}

// ----------------------------------------------------------------
static int ftell5_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	ftell5_pt_t * p1 = (ftell5_pt_t *)pvin1;
	ftell5_pt_t * p2 = (ftell5_pt_t *)pvin2;
	if (p1->x == p2->x && p1->y == p2->y && p1->z == p2->z)
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
static int ftell5_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	ftell5_pt_t * p1 = (ftell5_pt_t *)pvin1;
	ftell5_pt_t * p2 = (ftell5_pt_t *)pvin2;

	// The point at infinity comes first.
	if (p1->x == 0LL && p1->y == 1LL && p1->z == 0LL) {
		if (p2->x == 0LL && p2->y == 1LL && p2->z == 0LL)
			return 0;
		else
			return 1;
	}
	else if (p1->x < p2->x)
		return 1;
	else if (p1->x > p2->x)
		return 0;
	else if (p1->y < p2->y)
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
static void ftell5_vprint(
	void * pvin,
	void * pvaux)
{
	ftell5_pt_t  * ppt = (ftell5_pt_t  *)pvin;

#ifdef USE_AFFINE_IO
	if (ppt->x == 0LL && ppt->y == 1LL && ppt->z == 0LL)
		printf("O");
	else
		printf("%llx,%llx", ppt->x, ppt->y);
#else
	printf("%llx,%llx,%llx", ppt->x, ppt->y, ppt->z);
#endif
}

// ----------------------------------------------------------------
static void ftell5_vauxprint(
	void * pvaux)
{
	ftell5_aux_t * paux = (ftell5_aux_t *)pvaux;
	printf("%llx,%llx,%llx,%llx,%llx",
		paux->a1, paux->a2, paux->a3, paux->a4, paux->a6);
}

// ----------------------------------------------------------------
int ftell5_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	ftell5_pt_t  * pout = (ftell5_pt_t  *)pvout;
	ftell5_aux_t * paux = (ftell5_aux_t *)pvaux;

	if (sizeof(ftell5_pt_t) > max_elt_bytes) {
		fprintf(stderr, "ftell5:  Need %d bytes for element.\n",
			sizeof(ftell5_pt_t));
		exit(1);
	}

#ifdef USE_AFFINE_IO
	if (strcmp(pin, "O") == 0) {
		pout->x = 0LL;
		pout->y = 1LL;
		pout->z = 0LL;
	}
	else if (sscanf(pin, "%llx,%llx", &pout->x, &pout->y) == 2) {
		pout->z = 1LL;
	}
	else {
		return 0;
	}
#else
	if (sscanf(pin, "%llx,%llx,%llx", &pout->x, &pout->y, &pout->z) != 3)
		return 0;
	pout->x = z2ipolymod(pout->x, paux->p);
	pout->y = z2ipolymod(pout->y, paux->p);
	pout->z = z2ipolymod(pout->z, paux->p);
#endif

	check_point(pout, paux, "input");

	return 1;
}

// ----------------------------------------------------------------
int ftell5_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	ftell5_aux_t aux;

	if (max_aux_bytes < sizeof(ftell5_aux_t)) {
		fprintf(stderr, "ftell5:  Need %d bytes for aux.\n",
			sizeof(ftell5_aux_t));
		exit(1);
	}

	if (argc != 2) {
		fprintf(stderr, "ftell5_vauxscan: need argc == 2; got %d.\n", argc);
		return 0;
	}
	if (sscanf(argv[0], "%llx", &aux.m) != 1) {
		fprintf(stderr,
			"ftell5_vauxscan: couldn't scan m from \"%s\".\n",
			argv[0]);
		return 0;
	}
	if (sscanf(argv[1], "%llx,%llx,%llx,%llx,%llx",
		&aux.a1, &aux.a2, &aux.a3, &aux.a4, &aux.a6) != 5)
	{
		fprintf(stderr,
			"ftell5_vauxscan: couldn't scan a1,a2,a3,a4,a6 from \"%s\".\n",
			argv[1]);
		return 0;
	}
	aux.a1 = z2ipoly_mod(aux.a1, aux.m);
	aux.a2 = z2ipoly_mod(aux.a2, aux.m);
	aux.a3 = z2ipoly_mod(aux.a3, aux.m);
	aux.a4 = z2ipoly_mod(aux.a4, aux.m);
	aux.a6 = z2ipoly_mod(aux.a6, aux.m);
	*(ftell5_aux_t *)pvaux = aux;
	return 1;
}

// ----------------------------------------------------------------
int ftell5_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	ftell5_pt_t * pelements = (ftell5_pt_t *) pvelements;
	ftell5_aux_t * paux = (ftell5_aux_t *)pvaux;
	int maxn = max_group_bytes / sizeof(ftell5_pt_t);
	int n = 0;
	z2ipoly_t q = 1LL << z2ipoly_get_degree(paux->m);
	z2ipoly_t x, y;
	z2ipoly_t m = paux->m;
	z2ipoly_t a1 = paux->a1;
	z2ipoly_t a2 = paux->a2;
	z2ipoly_t a3 = paux->a3;
	z2ipoly_t a4 = paux->a4;
	z2ipoly_t a6 = paux->a6;

	// Point at infinity.
	pelements[n].x = 0LL;
	pelements[n].y = 1LL;
	pelements[n].z = 0LL;
	n++;

	// Sanity check
	check_point(&pelements[0], paux, "infinite");

	// Affine points.
	for (x = 0LL; x < q; x++) {
		for (y = 0LL; y < q; y++) {
			if (on_curve(x, y, 1LL, a1, a2, a3, a4, a6, m)) {
				if (n > maxn)
					return 0;
				pelements[n].x = x;
				pelements[n].y = y;
				pelements[n].z = 1LL;
				n++;
			}
		}
	}
	*pnum_elements = n;
	return 1;
}

// ================================================================
group_spec_t ftell5_group_spec = {
	element_size  : sizeof(ftell5_pt_t),
	naux          : 2,
	pvop          : ftell5_vadd,
	pvequal       : ftell5_vequal,
	pvless        : ftell5_vless,
	pvprint       : ftell5_vprint,
	pvauxprint    : ftell5_vauxprint,
	pvscan        : ftell5_vscan,
	pvauxscan     : ftell5_vauxscan,
	pfill_group   : ftell5_fill_group,
	desc          : "ftell5"
};
