#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "genquat.h"

// ----------------------------------------------------------------
// Presentation:
// < a, b | a^2m = 1, b^2 = a^m, ab = ba^-1 >
//
// Expressions simplify to a^i b^j for i=0,1,..,m-1 and j=0,1,2,3:
// * Powers on a don't exceed 2m.
// * Powers on b don't exceed 4.
// * Powers on a of m .. 2m-1 may have have a^m replaced with b^2.
// * The quasi-commutator ab=ba^-1 leads to ba = a^-1b which allows
//   all powers of a, and all powers of b, to be collected together.
//
// That is:
// * ba = a^-1 b
// * b a^i = a^-i b
// * b^j a = a^s b^j  where s = (-1)^j
//
// Thus
//   b^j a^i = b^j-1 b a^i
//           = b^j-2 a^i  b^2
//           = b^j-3 a^-i b^3
//           = b^j-4 a^i  b^4
//           = ...
//           = a^i   b^j  if  j even
//           = a^-i  b^j  if  j odd
//           = a^si  b^j  where s = (-1)^j

// Thus
//   a^i b^j a^k b^l = a^i (b^j a^k) b^l
//                   = a^i (a^sk b^j) b^l
//                   = a^(i+sk) b^(j+l)

genquat_t genquat_mul(genquat_t a, genquat_t b, int m)
{
	genquat_t c;
	int i = a.i;
	int j = a.j;
	int k = b.i;
	int l = b.j;
	int twom = m+m;

	// a^i b^j a^k b^l = a^(i+sk) b^(j+l)

	if (j & 1) {
		c.i = (i-k+twom) % twom;
	}
	else {
		c.i = (i+k) % twom;
	}
	c.j = (j+l) & 3;
	if (c.i >= m) {
		c.i -= m;
		c.j += 2;
		c.j &= 3;
	}
	return c;
}

// ----------------------------------------------------------------
int genquat_equal(genquat_t a, genquat_t b)
{
	if (a.i != b.i)
		return 0;
	if (a.j != b.j)
		return 0;
	return 1;
}

// ================================================================
static void genquat_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	genquat_t a = *(genquat_t *)pvin1;
	genquat_t b = *(genquat_t *)pvin2;
	int m = *(int *)pvaux;
	genquat_t c = genquat_mul(a, b, m);
	*(genquat_t *)pvout = c;
}

// ----------------------------------------------------------------
static int genquat_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	genquat_t a = *(genquat_t *)pvin1;
	genquat_t b = *(genquat_t *)pvin2;
	return genquat_equal(a, b);
}

// ----------------------------------------------------------------
static int genquat_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	genquat_t a = *(genquat_t *)pvin1;
	genquat_t b = *(genquat_t *)pvin2;
	if (a.i < b.i)
		return 1;
	else if (a.i > b.i)
		return 0;
	else if (a.j < b.j)
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
static void genquat_vprint(
	void * pvin,
	void * pvaux)
{
	genquat_t x = *(genquat_t *)pvin;
	int m = *(int *)pvaux;
	if (m < 10)
		printf("%d%d",  x.i, x.j);
	else
		printf("%d:%d",  x.i, x.j);
}

// ----------------------------------------------------------------
static void genquat_vauxprint(
	void * pvaux)
{
	int m = *(int *)pvaux;
	printf("%d", m);
}

// ----------------------------------------------------------------
int genquat_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	genquat_t * pout = (genquat_t *)pvout;
	int m = *(int *)pvaux;

	if (sizeof(genquat_t) > max_elt_bytes) {
		fprintf(stderr, "genquat:  Need %d bytes for element.\n",
			sizeof(genquat_t));
		exit(1);
	}

	if ((m < 10) && (strlen(pin) == 2)
		&& isdigit(pin[0]) && isdigit(pin[1]))
	{
		pout->i = pin[0] - '0';
		pout->j = pin[1] - '0';
		return 1;
	}
	else {
		return sscanf(pin, "%d:%d", &pout->i, &pout->j);
	}
}

// ----------------------------------------------------------------
int genquat_vauxscan(
	void * pvaux,
	int    argc,
	char **argv,
	int    max_aux_bytes)
{
	int * pm = pvaux;

	if (max_aux_bytes < sizeof(*pm)) {
		fprintf(stderr, "genquat:  Need %d bytes for aux.\n",
			sizeof(int));
		exit(1);
	}

	if (argc != 1) {
		fprintf(stderr, "genquat:  Need 1 aux (m); got %d.\n", argc);
		return 0;
	}
	if (sscanf(argv[0], "%d", pm) != 1) {
		fprintf(stderr, "Couldn't scan \"%s\" as int.\n",
			argv[0]);
		return 0;
	}

	return 1;
}

// ----------------------------------------------------------------
int  genquat_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	genquat_t * pelements = (genquat_t *) pvelements;
	int m = *(int *)pvaux;
	int i, j, k;

	*pnum_elements = m * 4;
	if (*pnum_elements > max_group_bytes/sizeof(genquat_t))
		return 0;

	for (k = 0, i = 0; i < m; i++) {
		for (j = 0; j < 4; j++, k++) {
			pelements[k].i = i;
			pelements[k].j = j;
		}
	}

	return 1;
}

// ================================================================
group_spec_t genquat_group_spec = {
	element_size  : sizeof(genquat_t),
	naux          : 1,
	pvop          : genquat_vmul,
	pvequal       : genquat_vequal,
	pvless        : genquat_vless,
	pvprint       : genquat_vprint,
	pvauxprint    : genquat_vauxprint,
	pvscan        : genquat_vscan ,
	pvauxscan     : genquat_vauxscan,
	pfill_group   : genquat_fill_group,
	desc          : "genquat"
};
