// Symmetries of the icosahedron (same as symmetries of the dodecahedron).
// Per mathworld.wolfram.com, this is ~= A5 x Z2.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "icos.h"
#include "pmt.h"

// ----------------------------------------------------------------
void icos_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	icos_t * pa = (icos_t *)pvin1;
	icos_t * pb = (icos_t *)pvin2;
	icos_t c;

	c.a5part = pmt_mul(&pa->a5part, &pb->a5part, 5);
	c.z2part = pa->z2part ^ pb->z2part;
	*(icos_t *)pvout = c;
}

// ----------------------------------------------------------------
int icos_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	icos_t * pa = (icos_t *)pvin1;
	icos_t * pb = (icos_t *)pvin2;
	if (pa->z2part != pb->z2part)
		return 0;
	return pmt_equal(&pa->a5part, &pb->a5part, 5);
}

// ----------------------------------------------------------------
int icos_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	icos_t * pa = (icos_t *)pvin1;
	icos_t * pb = (icos_t *)pvin2;
	if (pa->z2part < pb->z2part)
		return 1;
	else if (pa->z2part > pb->z2part)
		return 0;
	else
		return pmt_less(&pa->a5part, &pb->a5part, 5);
}

// ----------------------------------------------------------------
static void icos_vprint(
	void * pvin,
	void * pvaux)
{
	icos_t * pih = (icos_t *)pvin;
	int i;

	for (i = 0; i < 5; i++) {
		printf("%c", pih->a5part.images[i] + 'A');
	}
	printf("%d", pih->z2part);
}

// ----------------------------------------------------------------
int icos_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	icos_t * pih = (icos_t *)pvout;
	int i;

	if (sizeof(icos_t) > max_elt_bytes) {
		fprintf(stderr, "icos:  Need %d bytes for element.\n",
			sizeof(icos_t));
		exit(1);
	}


	if (strlen(pin) != 6) {
		fprintf(stderr,
			"icos element input \"%s\" must have length 6.\n",
			pin);
		return 0;
	}

	for (i = 0; i < 5; i++) {
		if ((pin[i] >= 'A') && (pin[i] <= 'Z'))
			pih->a5part.images[i] = pin[i] - 'A';
		else if ((pin[i] >= 'a') && (pin[i] <= 'z'))
			pih->a5part.images[i] = pin[i] - 'a';
		else if ((pin[i] >= '0') && (pin[i] <= '9'))
			pih->a5part.images[i] = pin[i] - '0';
		else {
			fprintf(stderr, "Malformed icos element \"%s\".\n",
				pin);
			return 0;
		}
	}
	if (pin[5] == '1') {
		pih->z2part = 1;
	}
	else if (pin[5] == '0') {
		pih->z2part = 0;
	}
	else {
		fprintf(stderr, "Malformed icos element \"%s\".\n",
			pin);
		return 0;
	}

	if (!pmt_validate_element(&pih->a5part, 5)) {
		fprintf(stderr,
			"icos element \"%s\" is not an even permuation.\n",
			pin);
		return 0;
	}

	return 1;
}


// ----------------------------------------------------------------
int icos_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	if (argc != 0)
		return 0;
	*(int *)pvaux = 5;
	return 1;
}

// ----------------------------------------------------------------
int  icos_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	icos_t * pelements = (icos_t *) pvelements;
	pmt_t a5s[60];
	int num_a5s;
	int n = 5;
	int i;

	*pnum_elements = 120;

	if (*pnum_elements > max_group_bytes/sizeof(icos_t))
		return 0;
	if (!an_fill_group(a5s, &n, sizeof(a5s), &num_a5s))
		return 0;
	if (num_a5s != 60) {
		fprintf(stderr,
			"icos_fill_group:  Expect |A5| = 60; got %d.\n",
			num_a5s);
		return 0;
	}

	for (i = 0; i < 60; i++) {
		pelements[i   ].z2part = 0;
		pelements[i+60].z2part = 1;
		pelements[i   ].a5part = a5s[i];
		pelements[i+60].a5part = a5s[i];
	}

	return 1;
}

// ================================================================
group_spec_t icos_group_spec = {
	element_size  : sizeof(icos_t),
	naux          : 0,
	pvop          : icos_vmul,
	pvequal       : icos_vequal,
	pvless        : icos_vless,
	pvprint       : icos_vprint,
	pvauxprint     : 0,
	pvscan        : icos_vscan,
	pvauxscan      : icos_vauxscan,
	pfill_group   : icos_fill_group,
	desc          : "icos"
};
