#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "metacyc.h"

#include "mod.h"

typedef struct _pqt_t {
	int p;
	int q;
	int t;
} pqt_t;

// ----------------------------------------------------------------
// a^i b^j  .  a^k b^l  =  a ^ {i + k . t^j} . b ^ {j + l}
// t   != 1 mod p
// t^q == 1 mod p

metacyc_t metacyc_mul(metacyc_t a, metacyc_t b, int p, int q, int t)
{
	metacyc_t c;
	int i = a.ppow;
	int j = a.qpow;
	int k = b.ppow;
	int l = b.qpow;

	c.ppow = (i + k * mod_power(t, j, p)) % p;
	c.qpow = (j + l) % q;
	return c;
}

// ----------------------------------------------------------------
int metacyc_equal(metacyc_t a, metacyc_t b)
{
	if (a.ppow != b.ppow)
		return 0;
	if (a.qpow != b.qpow)
		return 0;
	return 1;
}

// ----------------------------------------------------------------
static int find_t(
	int   p,
	int   q,
	int * pt)
{
	int t;
	for (t = 2; t < p; t++) {
		if (mod_power(t, q, p) == 1) {
			*pt = t;
			return 1;
		}
	}
	return 0;
}

// ================================================================
static void metacyc_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	metacyc_t a = *(metacyc_t *)pvin1;
	metacyc_t b = *(metacyc_t *)pvin2;
	pqt_t * ppqt = pvaux;
	metacyc_t c = metacyc_mul(a, b, ppqt->p, ppqt->q, ppqt->t);
	*(metacyc_t *)pvout = c;
}

// ----------------------------------------------------------------
static int metacyc_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	metacyc_t a = *(metacyc_t *)pvin1;
	metacyc_t b = *(metacyc_t *)pvin2;
	return metacyc_equal(a, b);
}

// ----------------------------------------------------------------
static int metacyc_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	metacyc_t a = *(metacyc_t *)pvin1;
	metacyc_t b = *(metacyc_t *)pvin2;
	if (a.ppow < b.ppow)
		return 1;
	else if (a.ppow > b.ppow)
		return 0;
	else if (a.qpow < b.qpow)
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
static void metacyc_vprint(
	void * pvin,
	void * pvaux)
{
	metacyc_t x = *(metacyc_t *)pvin;
	pqt_t * ppqt = pvaux;
	if ((ppqt->p < 10) && (ppqt->q < 10))
		printf("%d%d",  x.ppow, x.qpow);
	else
		printf("%d:%d",  x.ppow, x.qpow);
}

// ----------------------------------------------------------------
static void metacyc_vauxprint(
	void * pvaux)
{
	pqt_t * ppqt = pvaux;
	printf("%d/%d/%d", ppqt->p, ppqt->q, ppqt->t);
}

// ----------------------------------------------------------------
int metacyc_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	metacyc_t * pout = (metacyc_t *)pvout;
	pqt_t * ppqt = pvaux;

	if (sizeof(metacyc_t) > max_elt_bytes) {
		fprintf(stderr, "metacyc:  Need %d bytes for element.\n",
			sizeof(metacyc_t));
		exit(1);
	}

	if ((ppqt->p < 10) && (ppqt->q < 10) && (strlen(pin) == 2)
		&& isdigit(pin[0]) && isdigit(pin[1]))
	{
		pout->ppow = pin[0] - '0';
		pout->qpow = pin[1] - '0';
		return 1;
	}
	else {
		return sscanf(pin, "%d:%d", &pout->ppow, &pout->qpow);
	}
}

// ----------------------------------------------------------------
int metacyc_vauxscan(
	void * pvaux,
	int    argc,
	char **argv,
	int    max_aux_bytes)
{
	pqt_t * ppqt = pvaux;

	if (max_aux_bytes < sizeof(pqt_t)) {
		fprintf(stderr, "metacyc:  Need %d bytes for aux.\n",
			sizeof(int));
		exit(1);
	}

	if (argc != 1) {
		fprintf(stderr,
		"metacyc:  Need 1 aux (p,q without whitespace); got %d.\n",
			argc);
		return 0;
	}
	if (sscanf(argv[0], "%d,%d", &ppqt->p, &ppqt->q) != 2) {
		fprintf(stderr, "Couldn't scan \"%s\" as int, int.\n",
			argv[0]);
		return 0;
	}

	//if (sscanf(argv[2], "%d", &ppqt->t) != 1) {
	//	fprintf(stderr, "Couldn't scan \"%s\" as int.\n",
	//		argv[2]);
	//	return 0;
	//}
	//if (ppqt->t == 0) {
	if (!find_t(ppqt->p, ppqt->q, &ppqt->t)) {
		fprintf(stderr,
		"metacyc:  Couldn't find t for p=%d, q=%d.\n",
			ppqt->p, ppqt->q);
		return 0;
	}
	//}

	return 1;
}

// ----------------------------------------------------------------
int  metacyc_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	metacyc_t * pelements = (metacyc_t *) pvelements;
	pqt_t * ppqt = pvaux;
	int i, j, k;

	*pnum_elements = ppqt->p * ppqt->q;
	if (*pnum_elements > max_group_bytes/sizeof(metacyc_t))
		return 0;

	for (k = 0, i = 0; i < ppqt->p; i++) {
		for (j = 0; j < ppqt->q; j++, k++) {
			pelements[k].ppow = i;
			pelements[k].qpow = j;
		}
	}

	return 1;
}

// ================================================================
group_spec_t metacyc_group_spec = {
	element_size  : sizeof(metacyc_t),
	naux          : 1,
	pvop          : metacyc_vmul,
	pvequal       : metacyc_vequal,
	pvless        : metacyc_vless,
	pvprint       : metacyc_vprint,
	pvauxprint    : metacyc_vauxprint,
	pvscan        : metacyc_vscan ,
	pvauxscan     : metacyc_vauxscan,
	pfill_group   : metacyc_fill_group,
	desc          : "metacyc"
};
