#include <stdio.h>
#include <stdlib.h>
#include "mod.h"
#include "posmod.h"

#include "phi.h"
#include "gcd.h"

typedef struct _smod_aux_t {
	int modulus;
	int stride;
} smod_aux_t;

// ----------------------------------------------------------------
int mod_add(int a, int b, int m)
{
	int c;
	c = a + b;
	if (c >= m)
		c -= m;
	return c;
}

// ----------------------------------------------------------------
int mod_sub(int a, int b, int m)
{
	int c;
	c = (m + a) - b;
	if (c >= m)
		c -= m;
	return c;
}

// ----------------------------------------------------------------
int mod_mul(int a, int b, int m)
{
	int prod = (a * b) % m;
	if (prod < 0) // Silly C mod
		prod += m;
	return prod;
}

// ----------------------------------------------------------------
int mod_div(int a, int b, int m)
{
	int bi = mod_recip(b, m);
	return (a * bi) % m;
}

// ----------------------------------------------------------------
int mod_recip(int x, int m)
{
	static int mlast = 3;
	static int phi = 2;
	int rv, check;

	// Cache
	if (m != mlast) {
		mlast = m;
		phi = calc_phi_signed(m);
	}

	rv = mod_power(x, phi-1, m);
	check = (x * rv) % m;
	if (check != 1) {
		fprintf(stderr,
			"mod_recip():  Can't reciprocate %d mod %d.\n",
			x, m);
		exit(1);
	}

	return rv;
}

// ----------------------------------------------------------------
int mod_power(int x, int e, int m)
{
	int xp = x % m;
	int rv = 1;

	while (e != 0) {
		if (e & 1) {
			rv = (rv * xp) % m;
		}
		e = (unsigned)e >> 1;
		xp = (xp * xp) % m;
	}
	return (int)rv;
}

// ----------------------------------------------------------------
long long mod64_div(long long a, long long b, long long m)
{
	long long bi = mod64_recip(b, m);
	return (a * bi) % m;
}

// ----------------------------------------------------------------
long long mod64_recip(long long x, long long m)
{
	static long long mlast = 3;
	static long long phi = 2;
	long long rv, check;

	// Cache
	if (m != mlast) {
		mlast = m;
		phi = calc_phi_signed_ll(m);
	}

	rv = mod64_power(x, phi-1, m);
	check = (x * rv) % m;
	if (check != 1) {
		fprintf(stderr,
			"mod64_recip():  Can't reciprocate %lld mod %lld.\n",
			x, m);
		exit(1);
	}

	return rv;
}

// ----------------------------------------------------------------
long long mod64_power(long long x, long long e, long long m)
{
	long long xp = x % m;
	long long rv = 1;

	while (e != 0) {
		if (e & 1) {
			rv = (rv * xp) % m;
		}
		e = (unsigned long long)e >> 1;
		xp = (xp * xp) % m;
	}
	return rv;
}

// ----------------------------------------------------------------
long long mod64_mul_order(long long x, long long m)
{
	long long i, xp;

	x %= m;
	if (x < 0LL)
		x += m;
	if (x == 0LL)
		return 0LL;
	xp = x;

	for (i = 1LL; i <= m; i++) {
		if (xp == 1LL)
			return i;
		xp *= x;
		xp %= m;
	}
	printf("Yow! x=%lld m=%lld\n", x, m);
	exit(1);
	return 0;
}

// ----------------------------------------------------------------
int mod_mul_order(int x, int m)
{
	int i, xp;

	x %= m;
	if (x < 0)
		x += m;
	if (x == 0)
		return 0;
	xp = x;

	for (i = 1; i <= m; i++) {
		if (xp == 1)
			return i;
		xp *= x;
		xp %= m;
	}
	printf("Yow! x=%d m=%d\n", x, m);
	exit(1);
	return 0;
}

// ----------------------------------------------------------------
int mod_equal(int a, int b, int m)
{
	return a == b;
}

// ================================================================
static void mod_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a = *(int *)pvin1;
	int b = *(int *)pvin2;
	int m = *(int *)pvaux;
	int c = mod_add(a, b, m);
	*(int *)pvout = c;
}

// ================================================================
static void mod_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a = *(int *)pvin1;
	int b = *(int *)pvin2;
	int m = *(int *)pvaux;
	int c = mod_sub(a, b, m);
	*(int *)pvout = c;
}

// ----------------------------------------------------------------
static void mod_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a = *(int *)pvin1;
	int b = *(int *)pvin2;
	int m = *(int *)pvaux;
	int c = mod_mul(a, b, m);
	*(int *)pvout = c;
}

// ----------------------------------------------------------------
static void mod_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a = *(int *)pvin1;
	int b = *(int *)pvin2;
	int m = *(int *)pvaux;
	int c = mod_div(a, b, m);
	*(int *)pvout = c;
}

// ----------------------------------------------------------------
static int mod_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a = *(int *)pvin1;
	int b = *(int *)pvin2;
	//int m = *(int *)pvaux;
	//return mod_equal(a, b, m);
	return (a == b);
}

// ----------------------------------------------------------------
static int mod_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a = *(int *)pvin1;
	int b = *(int *)pvin2;
	return (a < b);
}

// ----------------------------------------------------------------
void mod_make_add_id(
	void * pvin,
	void * pvaux)
{
	*(int *)pvin = 0;
}

// ----------------------------------------------------------------
void mod_make_mul_id(
	void * pvin,
	void * pvaux)
{
	*(int *)pvin = 1;
}

// ----------------------------------------------------------------
static void mod_vprint(
	void * pvin,
	void * pvaux)
{
	int x = *(int *)pvin;
	int m = *(int *)pvaux;
	if (m > 100)
		printf("%3d", x % m);
	else if (m > 10)
		printf("%2d", x % m);
	else
		printf("%d", x % m);
}

// ----------------------------------------------------------------
int mod_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	int * pout = (int *)pvout;

	if (sizeof(int) > max_elt_bytes) {
		fprintf(stderr, "mod:  Need %d bytes for element.\n",
			sizeof(int));
		exit(1);
	}


	return sscanf(pin, "%d", pout);
}

// ----------------------------------------------------------------
static void mod_vauxprint(
	void * pvaux)
{
	printf("%d", *(int *)pvaux);
}

// ----------------------------------------------------------------
static int mod_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	int modulus;

	if (max_aux_bytes < sizeof(int)) {
		fprintf(stderr, "mod:  Need %d bytes for aux.\n",
			sizeof(int));
		exit(1);
	}

	if (argc != 1)
		return 0;
	if (sscanf(argv[0], "%d", &modulus) != 1) {
		fprintf(stderr, "Couldn't scan \"%s\" as int.\n",
			argv[0]);
		exit(1);
	}
	if (modulus < 1) {
		fprintf(stderr, "Modulus must be >= 1; got %d.\n",
			modulus);
		exit(1);
	}
	*(int *)pvaux = modulus;
	return 1;
}

// ----------------------------------------------------------------
int  modadd_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	int * pelements = (int *) pvelements;
	int m = *(int *)pvaux;
	int i;

	if (m > max_group_bytes/sizeof(int))
		return 0;
	for (i = 0; i < m; i++)
		pelements[i] = i;
	*pnum_elements = m;
	return 1;
}

// ----------------------------------------------------------------
int  modmul_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	int * pelements = (int *) pvelements;
	int m = *(int *)pvaux;
	int zi, zmi;

	*pnum_elements = calc_phi_signed(m);
	if (*pnum_elements > max_group_bytes)
		return 0;

	for (zi = 0, zmi = 0; zi < m; zi++) {
		if (calc_gcd_signed(zi, m) == 1) {
			pelements[zmi] = zi;
			zmi++;
		}
	}

	return 1;
}

// ----------------------------------------------------------------
static void smod_vauxprint(
	void * pvaux)
{
	smod_aux_t * paux = pvaux;
	printf("%d:%d", paux->modulus, paux->stride);
}

// ----------------------------------------------------------------
static int smod_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	smod_aux_t * paux = pvaux;

	if (max_aux_bytes < sizeof(smod_aux_t)) {
		fprintf(stderr, "mod:  Need %d bytes for aux.\n",
			sizeof(smod_aux_t));
		exit(1);
	}

	if (argc != 1)
		return 0;
	if (sscanf(argv[0], "%d:%d", &paux->modulus, &paux->stride) == 2) {
	}
	else if (sscanf(argv[0], "%d", &paux->modulus) == 1) {
		paux->stride = 1;
	}
	else {
		fprintf(stderr, "smod: Couldn't scan \"%s\".\n",
			argv[0]);
		exit(1);
	}
	if (paux->modulus < 1) {
		fprintf(stderr, "Modulus must be >= 1; got %d.\n",
			paux->modulus);
		exit(1);
	}
	if (paux->stride < 1) {
		fprintf(stderr, "Stride must be >= 1; got %d.\n",
			paux->stride);
		exit(1);
	}
	return 1;
}

// ----------------------------------------------------------------
int smodadd_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	int * pelements = (int *) pvelements;
	smod_aux_t * paux = pvaux;
	int i, v;
	int num_elements = paux->modulus / paux->stride;

	if (num_elements > max_group_bytes/sizeof(int))
		return 0;
	for (i = 0, v = 0; i < num_elements; i++, v += paux->stride)
		pelements[i] = v;
	*pnum_elements = num_elements;
	return 1;
}

// ----------------------------------------------------------------
// E.g. m = 60, s = 5, m/s = 12.  Z/(12)^* has 4 elements: 1, 5, 7, 11.
// So 5Z/(60) has the 4 elements 5, 25, 35, 55.

int smodmul_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	int * pelements = (int *) pvelements;
	smod_aux_t * paux = pvaux;
	int m = paux->modulus;
	int s = paux->stride;
	int ms = m/s;
	int zi, didx;
	int num_elements = calc_phi_signed(ms);

	if ((num_elements * sizeof(pelements[0])) > max_group_bytes)
		return 0;
	for (zi = 0, didx = 0; zi < ms; zi++) {
		if (calc_gcd_signed(zi, ms) == 1) {
			pelements[didx] = zi * s;
			didx++;
		}
	}

	*pnum_elements = num_elements;
	return 1;
}

// ================================================================
group_spec_t mod_add_group_spec = {
	element_size  : sizeof(int),
	naux          : 1,
	pvop          : mod_vadd,
	pvequal       : mod_vequal,
	pvless        : mod_vless,
	pvprint       : mod_vprint,
	pvscan        : mod_vscan,
	pvauxprint    : mod_vauxprint,
	pvauxscan     : mod_vauxscan,
	pfill_group   : modadd_fill_group,
	desc          : "modadd"
};

group_spec_t mod_mul_group_spec = {
	element_size  : sizeof(int),
	naux          : 1,
	pvop          : mod_vmul,
	pvequal       : mod_vequal,
	pvless        : mod_vless,
	pvprint       : mod_vprint,
	pvscan        : mod_vscan,
	pvauxprint    : mod_vauxprint,
	pvauxscan     : mod_vauxscan,
	pfill_group   : modmul_fill_group,
	desc          : "modmul"
};

group_spec_t smod_add_group_spec = {
	element_size  : sizeof(int),
	naux          : 1,
	pvop          : mod_vadd,
	pvequal       : mod_vequal,
	pvless        : mod_vless,
	pvprint       : mod_vprint,
	pvscan        : mod_vscan,
	pvauxprint    : smod_vauxprint,
	pvauxscan     : smod_vauxscan,
	pfill_group   : smodadd_fill_group,
	desc          : "smodadd"
};

group_spec_t smod_mul_group_spec = {
	element_size  : sizeof(int),
	naux          : 1,
	pvop          : mod_vmul,
	pvequal       : mod_vequal,
	pvless        : mod_vless,
	pvprint       : mod_vprint,
	pvscan        : mod_vscan,
	pvauxprint    : smod_vauxprint,
	pvauxscan     : smod_vauxscan,
	pfill_group   : smodmul_fill_group,
	desc          : "smodmul"
};

ring_spec_t mod_ring_spec = {
	element_size  : sizeof(int),
	naux          : 1,
	pvadd         : mod_vadd,
	pvmul         : mod_vmul,
	pvequal       : mod_vequal,
	pvless        : mod_vless,
	pvprint       : mod_vprint,
	pvauxprint    : mod_vauxprint,
	pvscan        : mod_vscan,
	pvauxscan     : mod_vauxscan,
	pfill_ring    : modadd_fill_group,
	desc          : "mod"
};

ring_spec_t smod_ring_spec = {
	element_size  : sizeof(int),
	naux          : 1,
	pvadd         : mod_vadd,
	pvmul         : mod_vmul,
	pvequal       : mod_vequal,
	pvless        : mod_vless,
	pvprint       : mod_vprint,
	pvauxprint    : mod_vauxprint,
	pvscan        : mod_vscan,
	pvauxscan     : mod_vauxscan,
	pfill_ring    : smodadd_fill_group,
	desc          : "smod"
};

vfops_t mod_vfops = {
	element_size  : sizeof(int),
	pvadd         : mod_vadd,
	pvsub         : mod_vsub,
	pvmul         : mod_vmul,
	pvdiv         : mod_vdiv,
	pvequal       : mod_vequal,
	pvless        : mod_vless,
	pvprint       : mod_vprint,
	pvmake_add_id : mod_make_add_id,
	pvmake_mul_id : mod_make_mul_id
};
