#include <stdio.h>
#include <stdlib.h>

typedef unsigned long long u64;

typedef struct _P_t {
	u64 Z;
	u64 N;
} P_t;

P_t calc_M(u64 qm, P_t p0, P_t p1);
P_t calc_S(u64 qm, P_t p0, P_t p1);
P_t calc_P_nM(u64 qm, u64 n);
P_t calc_P_nSk(u64 qm, u64 n, u64 k);
P_t calc_P(u64 qm, u64 n);

// ----------------------------------------------------------------
P_t calc_M(u64 qm, P_t p0, P_t p1)
{
	P_t p2;

	//z_2 = z_0z_1 + (q^m-1)z_0n_1 + (q^m-1)z_1n_0
	//n_2 = (q^m-1)^2 n_0n_1

	p2.Z = p0.Z * p1.Z
		+ (qm - 1) * p0.Z * p1.N
		+ (qm - 1) * p1.Z * p0.N;
	p2.N = (qm-1) * p0.N * p1.N;

#ifdef TRACE
	printf("    M((%llu, %llu), (%llu, %llu)) = (%llu, %llu)\n",
		p0.Z, p0.N,
		p1.Z, p1.N,
		p2.Z, p2.N);
#endif
	return p2;
}

// ----------------------------------------------------------------
P_t calc_S(u64 qm, P_t p0, P_t p1)
{
	P_t p2;

	//z_2 = z_0z_1 + (q^m-1)n_0n_1
	//n_2 = (q^m-1) z_1n_0 + (q^m-1) z_0n_1 + (q^m-2)^2 n_0n_1

	p2.Z = p0.Z * p1.Z + (qm - 1) * p0.N * p1.N;
	p2.N =  p1.Z * p0.N +
		p0.Z * p1.N +
		(qm-2) * p0.N * p1.N;
#ifdef TRACE
	printf("    S((%llu, %llu), (%llu, %llu)) = (%llu, %llu)\n",
		p0.Z, p0.N,
		p1.Z, p1.N,
		p2.Z, p2.N);
#endif

	return p2;
}

// ----------------------------------------------------------------
P_t calc_P_nM(
	u64 qm,
	u64 n)
{
	P_t rv;
	if (n > 1) {
		rv = calc_M(
			qm,
			calc_P(qm,   1),
			calc_P(qm, n-1));
	}
	else {
		fprintf(stderr, "P_1M not defined; we shouln't be here.\n");
		exit(1);
	}
#ifdef TRACE
	printf("P_%lluM:  (%16llu, %16llu)\n", n, rv.Z, rv.N);
#endif
	return rv;
}

// ----------------------------------------------------------------
P_t calc_P_nSk(
	u64 qm,
	u64 n,
	u64 k)
{
	P_t rv;
	if (k > 1) {
		rv = calc_S(
			qm,
			calc_P_nSk(qm, n, k-1),
			calc_P_nSk(qm, n,   1));
	}
	else {
		rv = calc_P_nM(qm, n);
	}
#ifdef TRACE
	printf("P_%lluS%llu: (%16llu, %16llu)\n", n, k, rv.Z, rv.N);
#endif
	return rv;
}

// ----------------------------------------------------------------
P_t calc_P(
	u64 qm,
	u64 n)
{
	P_t rv;

	if (n == 1) {
		rv.Z = 1;
		rv.N = 1;
	}
	else {
		rv = calc_P_nSk(qm, n, n);
	}
#ifdef TRACE
	printf("P_%llu:   (%16llu, %16llu)\n", n, rv.Z, rv.N);
#endif

	return rv;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	u64 n = 3, q = 2, m = 1;
	u64 qm;
	u64 i;
	P_t Pn;
	u64 check;
	u64 qmn2;

	if (argc == 1) {
		;
	}
	else if (argc == 3) {
		(void)sscanf(argv[1], "%llu", &n);
		(void)sscanf(argv[2], "%llu", &q);
		m = 1;
	}
	else if (argc == 4) {
		(void)sscanf(argv[1], "%llu", &n);
		(void)sscanf(argv[2], "%llu", &q);
		(void)sscanf(argv[3], "%llu", &m);
	}
	else {
		fprintf(stderr, "Usage: %s {n} {q} {m}\n", argv[0]);
		exit(1);
	}

	// Compute q^m.
	qm = 1;
	for (i = 0; i < m; i++)
		qm *= q;

	// Compute P(n).
	Pn = calc_P(qm, n);

	printf("%llux%llu matrices over GF(%llu",
		n, n, q);
	if (m > 1)
		printf("^%llu", m);
	printf("):\n");

	printf("  %16llu             zero determinants\n", Pn.Z);
	printf("  %16llu of each non-zero determinant\n", Pn.N);

	check = Pn.Z + (qm-1) * Pn.N;

	// Compute (q^m).
	qmn2 = 1;
	for (i = 0; i < n*n; i++)
		qmn2 *= qm;

	if (check == qmn2)
		printf("OK\n");
	else
		printf("(q^m)^(n^2) = %llu, but check = %llu: not good.\n",
			qmn2, check);

	return 0;
}
