#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pmt.h"
#include "factorial.h"
#include "tokenize.h"
#include "malloc_check.h"

static void kth_pmt(
	pmt_elt_t * in,
	pmt_elt_t * temp,
	pmt_elt_t * out,
	int             n,
	int             nfact,
	int             which);

static void pmt_sort(
	pmt_t * ppmt,
	int    n,
	int  * pnswap);

// ----------------------------------------------------------------
pmt_t pmt_mul(pmt_t * pa, pmt_t * pb, int n)
{
	int i;
	pmt_t c;

	for (i = 0; i < n; i++)
		c.images[i] = pa->images[pb->images[i]];
	return c;
}

// ----------------------------------------------------------------
void pmt_make_id(pmt_t * pa, int n)
{
	int i;
	for (i = 0; i < n; i++)
		pa->images[i] = i;
}

// ----------------------------------------------------------------
void pmt_invert(pmt_t * pin, pmt_t * pout, int n)
{
	int i;
	for (i = 0; i < n; i++)
		pout->images[pin->images[i]] = i;
}

// ----------------------------------------------------------------
int pmt_parity(
	pmt_t * ppmt,
	int     n)
{
	pmt_t copy = *ppmt;
	int nswap;

	pmt_sort(&copy, n, &nswap);

	return nswap & 1;
}

// ----------------------------------------------------------------
void pmt_print(pmt_t * pa, int n)
{
	int i;

	//printf("(");
	for (i = 0; i < n; i++) {
		if (i != 0)
			printf(",");
		//if (n >= 10)
			//printf("%2d", (int)pa->images[i] + 1);
		//else
			printf("%d", (int)pa->images[i] + 1);
	}
	//printf(")");
}

// ----------------------------------------------------------------
int pmt_count_cycles(
	pmt_t * pa,
	int n)
{
	int i;
	unsigned char marks[PMT_MAX_N];
	unsigned char next;
	int rv = 0;

	for (i = 0; i < n; i++)
		marks[i] = 0;
	for (i = 0; i < n; i++) {
		if (marks[i])
			continue;

		rv++;
		next = i;
		marks[next] = 1;

		while (1) {
			next = pa->images[next];
			if (next == i)
				break;
			marks[next] = 1;
		}
	}
	return rv;
}

// ----------------------------------------------------------------
void pmt_print_cd(
	pmt_t * pa,
	int n)
{
	pmt_print_cd_style(pa, n, PMT_CD_COMMA_COLON);
}

// ----------------------------------------------------------------
void pmt_print_cd_style(
	pmt_t * pa,
	int n,
	int style)
{
	int i;
	unsigned char marks[PMT_MAX_N];
	unsigned char next;
	int ncyc_printed = 0;
	int n_in_cyc;

	for (i = 0; i < n; i++)
		marks[i] = 0;
	for (i = 0; i < n; i++) {
		if (marks[i])
			continue;

		n_in_cyc = 0;
		next = i;
		marks[next] = 1;
		if (pa->images[next] != next) {

			if (style == PMT_CD_PARENS) {
				if (ncyc_printed > 0)
					printf(" ");
				printf("(");
			}
			else {
				if (ncyc_printed > 0)
					printf(":");
			}

			printf("%d", next + 1);
			ncyc_printed++;
			n_in_cyc++;
		}

		while (1) {
			next = pa->images[next];
			if (next == i)
				break;
			marks[next] = 1;
			if (style == PMT_CD_PARENS)
				printf(" ");
			else
				printf(",");
			printf("%d", next + 1);
			n_in_cyc++;
		}
		if (style == PMT_CD_PARENS) {
			if (n_in_cyc > 0)
				//printf(")<%d>", n_in_cyc);
				printf(")");
		}
	}

	if (ncyc_printed == 0) {
		printf("1");
		return;
	}
}

// ----------------------------------------------------------------
static int uchar_qcmp(const void * pv1, const void * pv2)
{
	const unsigned char * p1 = pv1;
	const unsigned char * p2 = pv2;
	if (*p1 < *p2)
		return -1;
	if (*p1 > *p2)
		return 1;
	return 0;
}

void pmt_print_cycle_type(
	pmt_t * pa,
	int n)
{
	int i;
	unsigned char marks[PMT_MAX_N];
	unsigned char next;
	unsigned char lengths[PMT_MAX_N];
	int ncyc = 0;

	for (i = 0; i < n; i++)
		lengths[i] = 0;

	for (i = 0; i < n; i++)
		marks[i] = 0;

	for (i = 0; i < n; i++) {
		if (marks[i])
			continue;

		lengths[ncyc] = 0;
		next = i;
		marks[next] = 1;
		if (pa->images[next] != next) {
			lengths[ncyc]++;
		}

		while (1) {
			next = pa->images[next];
			if (next == i)
				break;
			marks[next] = 1;
			lengths[ncyc]++;
		}
		if (lengths[ncyc] > 0)
			ncyc++;
	}

	qsort(lengths, ncyc, sizeof(lengths[0]), uchar_qcmp);
	for (i = 0; i < ncyc; i++) {
		if (i > 0)
			printf(",");
		printf("%d", (int)lengths[i]);
	}
	if (ncyc == 0)
		printf("1");
}

// ----------------------------------------------------------------
int pmt_equal(pmt_t * pa, pmt_t * pb, int n)
{
	int i;
	// xxx bounds-check on n

	for (i = 0; i < n; i++)
		if (pa->images[i] != pb->images[i])
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int pmt_less(pmt_t * pa, pmt_t * pb, int n)
{
	int i;

	for (i = 0; i < n; i++) {
		if (pa->images[i] < pb->images[i])
			return 1;
		if (pa->images[i] > pb->images[i])
			return 0;
	}
	return 0;
}

// ----------------------------------------------------------------
static void kth_pmt(
	pmt_elt_t * in,
	pmt_elt_t * temp,
	pmt_elt_t * out,
	int         n,
	int         nfact,
	int         which)
{
	int pos, q, r, i;
	int ni, nifact;

	nifact = nfact;

	for (i = 0; i < n; i++)
		temp[i] = in[i];

	ni = n;
	for (pos = 0; pos < n; pos++) {
		nifact /= ni;
		r = which % nifact;
		q = which / nifact;
		which = r;

		out[pos] = temp[q];
		for (i = q; i < ni; i++)
			temp[i] = temp[i+1];

		ni--;
	}
}

// ================================================================
void pmt_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	pmt_t * pa = (pmt_t *)pvin1;
	pmt_t * pb = (pmt_t *)pvin2;
	int  n = *(int  *)pvaux;
	pmt_t c = pmt_mul(pa, pb, n);
	*(pmt_t *)pvout = c;
}

// ----------------------------------------------------------------
int pmt_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	pmt_t * pa = (pmt_t *)pvin1;
	pmt_t * pb = (pmt_t *)pvin2;
	int n = *(int *)pvaux;
	return pmt_equal(pa, pb, n);
}

// ----------------------------------------------------------------
int pmt_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	pmt_t * pa = (pmt_t *)pvin1;
	pmt_t * pb = (pmt_t *)pvin2;
	int n = *(int *)pvaux;

	return pmt_less(pa, pb, n);
}

// ----------------------------------------------------------------
int pmt_validate_element(
	pmt_t * ppmt,
	int    n)
{
	pmt_t copy = *ppmt;
	int nswap;
	int i;

	pmt_sort(&copy, n, &nswap);

	for (i = 0; i < n; i++)
		if (copy.images[i] != i)
			return 0;

	return 1;
}

// ----------------------------------------------------------------
int an_validate_element(
	pmt_t * ppmt,
	int     n)
{
	if (!pmt_validate_element(ppmt, n))
		return 0;
	if (pmt_parity(ppmt, n) != 0)
		return 0;
	return 1;
}

// ----------------------------------------------------------------
static void pmt_sort(
	pmt_t * ppmt,
	int     n,
	int   * pnswap)
{
	pmt_elt_t temp;
	int i, top;
	int nswap = 0;

	for (top = n - 1; top > 0; top--) {
		for (i = 0; i < top; i++) {
			if (ppmt->images[i] > ppmt->images[i+1]) {
				temp = ppmt->images[i];
				ppmt->images[i] = ppmt->images[i+1];
				ppmt->images[i+1] = temp;
				nswap++;
			}
		}
	}

	*pnswap = nswap;
}

// ----------------------------------------------------------------
static void pmt_vprint(
	void * pvin,
	void * pvaux)
{
	pmt_t * ppmt = (pmt_t *)pvin;
	int n = *(int *)pvaux;
	pmt_print(ppmt, n);
}

// ----------------------------------------------------------------
static void pmt_vprint_cd(
	void * pvin,
	void * pvaux)
{
	pmt_t * ppmt = (pmt_t *)pvin;
	int n = *(int *)pvaux;
	pmt_print_cd(ppmt, n);
}

// ----------------------------------------------------------------
static void pmt_vauxprint(
	void * pvaux)
{
	printf("%d", *(int *)pvaux);
}

// ----------------------------------------------------------------
// xxx temp hack hack hack
int pmt_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	pmt_t * ppmt = (pmt_t *)pvout;
	int    n = *(int *)pvaux;
	int    i;
#define MAX_ARGC 100
	int    argc;
	char * argv[MAX_ARGC];
	char   temp[256];
	unsigned u;

	if (sizeof(pmt_t) > max_elt_bytes) {
		fprintf(stderr, "sn:  Need %d bytes for element.\n",
			sizeof(pmt_t));
		exit(1);
	}

	strcpy(temp, pin);

	argc = tokenize(temp, ",", argv, MAX_ARGC);
	if (argc != n) {
		fprintf(stderr,
		"pmt_vscan:  Needed %d item%s but got %d in \"%s\".\n",
			n, n == 1 ? "" : "s", argc, pin);
		return 0;
	}

	for (i = 0; i < argc; i++) {
		if (sscanf(argv[i], "%u", &u) != 1) {
			fprintf(stderr, "pmt_vscan:  Couldn't parse element "
				"(\"%s\") of \"%s\".\n",
				argv[i], pin);
			return 0;
		}
		ppmt->images[i] = u - 1;
	}
	if (!pmt_validate_element(ppmt, n)) {
		fprintf(stderr,
		"pmt_vscan: \"%s\" is not a permutation in S%d\n",
			pin, n);
		return 0;
	}
	return 1;
}

// ----------------------------------------------------------------
// xxx temp hack hack hack

// 0 1 2 3 4
// 0 1 2 3 4

// temp = 1,2,4:3,5
// outer_argv = 1,2,4   3,5

// inner_argc = 1  2  4
// u[0] = 0  u[1] = 1  u[2] = 3

// 0 1 2 3 4
// 1 3 2 0 4

// inner_argc = 3  5
// u[0] = 2  u[1] = 4

// 0 1 2 3 4
// 1 3 4 0 2

// 0 1 2 3 4
// 1 3 2 0 4

// This function is a hack job.  It doesn't correctly see that 1,3:1,3
// is the identity permutation -- rather, it gets 1,3.

int pmt_vscan_cd(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	pmt_t * ppmt = (pmt_t *)pvout;
	pmt_t  single;
	int    n = *(int *)pvaux;
	int    i, j;
#define MAX_ARGC 100
	int    outer_argc;
	char * outer_argv[MAX_ARGC];
	int    inner_argc;
	char * inner_argv[MAX_ARGC];
	char   temp[256];
	unsigned * u = 0;

	if (sizeof(pmt_t) > max_elt_bytes) {
		fprintf(stderr, "sn:  Need %d bytes for element.\n",
			sizeof(pmt_t));
		exit(1);
	}

	for (i = 0; i < PMT_MAX_N; i++)
		ppmt->images[i] = i;

	strcpy(temp, pin);

	outer_argc = tokenize(temp, ":", outer_argv, MAX_ARGC);
	for (i = 0; i < outer_argc; i++) {
		inner_argc = tokenize(outer_argv[i], ",", inner_argv, MAX_ARGC);
		u = malloc_check(inner_argc * sizeof(unsigned));
		for (j = 0; j < inner_argc; j++) {
			if (sscanf(inner_argv[j], "%u", &u[j]) != 1) {
				fprintf(stderr,
				"pmt_vscan_cd:  Couldn't parse element "
					"(\"%s\") of \"%s\".\n",
					inner_argv[i], pin);
				return 0;
			}
			if ((u[j] < 1) || (u[j] > PMT_MAX_N)) {
				fprintf(stderr,
				"pmt_vscan_cd:  element %d out of range 1-%d\n",
					u[j], n);
				return 0;
			}
			u[j]--;
		}

		pmt_make_id(&single, n);
		for (j = 1; j < inner_argc; j++) {
			single.images[u[j-1]] = u[j];
		}
		single.images[u[inner_argc - 1]] = u[0];
		*ppmt = pmt_mul(ppmt, &single, n);
		free(u);
	}

	if (!pmt_validate_element(ppmt, n)) {
		printf("pmt_vscan: \"%s\" is not a permutation in S%d: got ",
			pin, n);
		pmt_print(ppmt, n);
		printf("\n");
		//return 0;
		exit(1);
	}
	return 1;
}

// ----------------------------------------------------------------
int pmt_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	int n;

	if (max_aux_bytes < sizeof(int)) {
		fprintf(stderr, "sn:  Need %d bytes for aux.\n",
			sizeof(int));
		exit(1);
	}

	if (argc != 1)
		return 0;
	if (sscanf(argv[0], "%d", &n) != 1) {
		fprintf(stderr, "Couldn't scan \"%s\" as int.\n",
			argv[0]);
		exit(1);
	}
	if (n < 1) {
		fprintf(stderr, "n must be >= 1; got %d.\n",
			n);
		exit(1);
	}
	*(int *)pvaux = n;
	return 1;
}

// ----------------------------------------------------------------
// Symmetric group
int  sn_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	pmt_t * pelements = (pmt_t *) pvelements;
	int n = *(int *)pvaux;
	int i;
	pmt_t identity;
	pmt_t temp;

	if (n > 12)
		// The factorial calculation itself will overflow a
		// signed 32-bit integer.
		return 0;

	*pnum_elements = calc_factorial(n);
	if (*pnum_elements > max_group_bytes/sizeof(pmt_t))
		return 0;

	for (i = 0; i < n; i++)
		identity.images[i] = i;

	for (i = 0; i < *pnum_elements; i++) {
		kth_pmt(identity.images, temp.images, pelements[i].images,
			n, *pnum_elements, i);
	}

	return 1;
}

// ----------------------------------------------------------------
// Alternating group
int  an_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	pmt_t * pelements = (pmt_t *) pvelements;
	int n = *(int *)pvaux;
	int si;
	int ai;
	pmt_t identity;
	pmt_t temp;

	if (n > 12)
		// The factorial calculation itself will overflow a
		// signed 32-bit integer.
		return 0;

	*pnum_elements = calc_factorial(n) / 2;
	if (*pnum_elements > max_group_bytes/sizeof(pmt_t))
		return 0;

	for (ai = 0; ai < n; ai++)
		identity.images[ai] = ai;

	for (si = 0, ai = 0; ai < *pnum_elements; si++) {
		kth_pmt(identity.images, temp.images, pelements[ai].images,
			n, *pnum_elements * 2, si);
		if (pmt_parity(&pelements[ai], n) == 0)
			ai++;
	}

	return 1;
}

// ================================================================
group_spec_t sn_group_spec = {
	// Symmetric group with image-list I/O
	element_size  : sizeof(pmt_t),
	naux          : 1,
	pvop          : pmt_vmul,
	pvequal       : pmt_vequal,
	pvless        : pmt_vless,
	pvprint       : pmt_vprint,
	pvscan        : pmt_vscan,
	pvauxprint    : pmt_vauxprint,
	pvauxscan     : pmt_vauxscan,
	pfill_group   : sn_fill_group,
	desc          : "sn"
};

group_spec_t an_group_spec = {
	// Alternating group with image-list I/O
	element_size  : sizeof(pmt_t),
	naux          : 1,
	pvop          : pmt_vmul,
	pvequal       : pmt_vequal,
	pvless        : pmt_vless,
	pvprint       : pmt_vprint,
	pvscan        : pmt_vscan,
	pvauxprint    : pmt_vauxprint,
	pvauxscan     : pmt_vauxscan,
	pfill_group   : an_fill_group,
	desc          : "an"
};

group_spec_t snc_group_spec = {
	// Symmetric group with cycle-decomposition I/O
	element_size  : sizeof(pmt_t),
	naux          : 1,
	pvop          : pmt_vmul,
	pvequal       : pmt_vequal,
	pvless        : pmt_vless,
	pvprint       : pmt_vprint_cd,
	pvscan        : pmt_vscan_cd,
	pvauxprint    : pmt_vauxprint,
	pvauxscan     : pmt_vauxscan,
	pfill_group   : sn_fill_group,
	desc          : "snc"
};

group_spec_t anc_group_spec = {
	// Alternating group with cycle-decomposition I/O
	element_size  : sizeof(pmt_t),
	naux          : 1,
	pvop          : pmt_vmul,
	pvequal       : pmt_vequal,
	pvless        : pmt_vless,
	pvprint       : pmt_vprint_cd,
	pvscan        : pmt_vscan_cd,
	pvauxprint    : pmt_vauxprint,
	pvauxscan     : pmt_vauxscan,
	pfill_group   : an_fill_group,
	desc          : "anc"
};
