#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "qpoly.h"
#include "tokenize.h"

static int qpoly_find_degree(
	qpoly_t * pin,
	int       max_deg);

static void qpoly_check_degree(
	int degree);

static qpoly_t qpoly_scalar_mul(
	qpoly_t * pin1,
	rat_t     in2);

// ----------------------------------------------------------------
qpoly_t qpoly_add(
	qpoly_t * pin1,
	qpoly_t * pin2)
{
	qpoly_t out;
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++)
			out.coeffs[i] = rat_add(
				pin1->coeffs[i], pin2->coeffs[i]);
		for (i = pin2->degree + 1; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i];
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++)
			out.coeffs[i] = rat_add(
				pin1->coeffs[i], pin2->coeffs[i]);
		for (i = pin1->degree + 1; i <= pin2->degree; i++)
			out.coeffs[i] = pin2->coeffs[i];
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were added.
	out.degree = qpoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
qpoly_t qpoly_sub(
	qpoly_t * pin1,
	qpoly_t * pin2)
{
	qpoly_t out;
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++)
			out.coeffs[i] = rat_sub(
				pin1->coeffs[i], pin2->coeffs[i]);
		for (i = pin2->degree + 1; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i];
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++)
			out.coeffs[i] = rat_sub(
				pin1->coeffs[i], pin2->coeffs[i]);
		for (i = pin1->degree + 1; i <= pin2->degree; i++)
			out.coeffs[i] = rat_usub(pin2->coeffs[i]);
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were
	// subtracted.
	out.degree = qpoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
qpoly_t qpoly_usub(
	qpoly_t * pin)
{
	qpoly_t out;
	int i;

	for (i = 0; i <= pin->degree; i++)
		out.coeffs[i] = rat_usub(pin->coeffs[i]);
	out.degree = pin->degree;
	return out;
}

// ----------------------------------------------------------------
qpoly_t qpoly_mul(
	qpoly_t * pin1,
	qpoly_t * pin2)
{
	qpoly_t  out;
	int i, j;
	rat_t prod;

	out.degree = pin1->degree + pin2->degree;
	qpoly_check_degree(out.degree);

	for (i = 0; i <= out.degree; i++)
		out.coeffs[i] = rat_from_int(0);

	for (i = 0; i <= pin1->degree; i++) {
		for (j = 0; j <= pin2->degree; j++) {
			prod = rat_mul(pin1->coeffs[i], pin2->coeffs[j]);
			out.coeffs[i+j] = rat_add(out.coeffs[i+j], prod);
		}
	}

	// E.g. if one polynomial was zero.  My convention is that the zero
	// polynomial has "degree" zero -- correct as an array index, but
	// mathematically incorrect.
	out.degree = qpoly_find_degree(&out, out.degree);

	return out;
}

// ----------------------------------------------------------------
qpoly_t qpoly_div(
	qpoly_t * pin1,
	qpoly_t * pin2)
{
	qpoly_t quot, rem;
	qpoly_quot_and_rem(pin1, pin2, &quot, &rem);
	return quot;
}

// ----------------------------------------------------------------
qpoly_t qpoly_mod(
	qpoly_t * pin1,
	qpoly_t * pin2)
{
	qpoly_t quot, rem;
	qpoly_quot_and_rem(pin1, pin2, &quot, &rem);
	return rem;
}

// ----------------------------------------------------------------
void qpoly_quot_and_rem(
	qpoly_t * pdividend,
	qpoly_t * pdivisor,
	qpoly_t * pquot,
	qpoly_t * prem)
{
	qpoly_t quot;
	qpoly_t rem;
	int i;
	int max_shift;
	int shift;
	rat_t rem_leader;
	rat_t divisor_leader;
	rat_t dlinv;
	rat_t multiplier;
	qpoly_t divisor_with_mul;
	int stop;

	max_shift = pdividend->degree - pdivisor->degree;
	if (pdividend->degree < pdivisor->degree) {
		quot.coeffs[0] = rat_from_int(0);
		quot.degree = 0;
		rem = *pdividend;

		*pquot = quot;
		*prem  = rem;
		return;
	}

	quot.degree = pdividend->degree - pdivisor->degree;
	for (i = 0; i <= quot.degree; i++)
		quot.coeffs[i] = rat_from_int(0);
	rem = *pdividend;

	divisor_leader = pdivisor->coeffs[pdivisor->degree];
	dlinv = rat_recip(divisor_leader);
	for (shift = max_shift; shift >= 0; shift--) {
		rem_leader = rem.coeffs[shift + pdivisor->degree];
		multiplier = rat_mul(rem_leader, dlinv);
		divisor_with_mul = qpoly_scalar_mul(pdivisor, multiplier);
		quot.coeffs[shift] = multiplier;

		stop = shift + divisor_with_mul.degree;
		for (i = shift; i <= stop; i++) {
			rem.coeffs[i] = rat_sub(rem.coeffs[i],
				divisor_with_mul.coeffs[i-shift]);
		}
	}

	quot.degree = qpoly_find_degree(&quot, quot.degree);
	rem.degree  = qpoly_find_degree(&rem,  rem.degree);

	*pquot = quot;
	*prem  = rem;
}

// ----------------------------------------------------------------
// Input: a, b
// Output: GCD g, and m, n such that am + bn = d

qpoly_t qpoly_ext_gcd(
	qpoly_t * pa, // 1st input.
	qpoly_t * pb, // 2nd input.
	qpoly_t * pm, // Output coefficient on a
	qpoly_t * pn) // Output coefficient on b
{
	qpoly_t mprime, nprime, m, n, c, q, r, t, qm, qn;
	qpoly_t d;   // Return value.

	// Initialize
	mprime = qpoly_from_coeffs0(rat_from_int(1));
	n      = qpoly_from_coeffs0(rat_from_int(1));
	m      = qpoly_from_coeffs0(rat_from_int(0));
	nprime = qpoly_from_coeffs0(rat_from_int(0));
	c = *pa;
	d = *pb;


	while (1) {

		// Divide
		// q = c / d, r = c % d
		qpoly_quot_and_rem(&c, &d, &q, &r);
		// Note:  now c = qd + r and 0 <= r < d

		// Remainder zero?
		//if (r == zero)
		if ((r.degree == 0) && rat_is_zero(r.coeffs[0])) {
			break;
		}

		// Recycle
		c = d;
		d = r;

		t = mprime;
		mprime = m;
		//qm = q * m;
		qm = qpoly_mul(&q, &m);
		//m = t - qm;
		m = qpoly_sub(&t, &qm);

		t = nprime;
		nprime = n;
		//qn = q * n;
		qn = qpoly_mul(&q, &n);
		//n = t - qn;
		n = qpoly_sub(&t, &qn);
	}
	*pm = m;
	*pn = n;
	return d;
}

// ----------------------------------------------------------------
// Input: a, b
// Output: GCD g, and m, n such that am + bn = d

qpoly_t qpoly_gcd(
	qpoly_t * pa, // 1st input.
	qpoly_t * pb) // 2nd input.
{
	qpoly_t c, q, r;
	qpoly_t d;   // Return value.

	if (qpoly_is_zero(pa))
		return *pb;
	if (qpoly_is_zero(pb))
		return *pa;

	c = *pa;
	d = *pb;

	while (1) {

		// Divide
		// q = c / d, r = c % d
		qpoly_quot_and_rem(&c, &d, &q, &r);
		// Note:  now c = qd + r and 0 <= r < d

		// Remainder zero?
		//if (r == zero)
		if ((r.degree == 0) && rat_is_zero(r.coeffs[0])) {
			break;
		}

		// Recycle
		c = d;
		d = r;
	}
	return d;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
qpoly_t qpoly_power(
	qpoly_t * px,
	unsigned  power)
{
	qpoly_t x2 = *px;
	qpoly_t rv;

	rv = qpoly_from_coeffs0(rat_from_int(1));

	while (power != 0) {
		if (power & 1)
			rv = qpoly_mul(&x2, &rv);
		power >>= 1;
		x2 = qpoly_mul(&x2, &x2);
	}
	return rv;
}

// ----------------------------------------------------------------
int qpoly_equal(
	qpoly_t * pin1,
	qpoly_t * pin2)
{
	int i;
	if (pin1->degree != pin2->degree)
		return 0;
	for (i = 0; i <= pin1->degree; i++)
		if (!rat_equal(pin1->coeffs[i], pin2->coeffs[i]))
			return 0;
	return 1;
}

// ----------------------------------------------------------------
static qpoly_t qpoly_scalar_mul(
	qpoly_t * pin1,
	rat_t     in2)
{
	qpoly_t  out;
	int i;

	for (i = 0; i <= pin1->degree; i++)
		out.coeffs[i] = rat_mul(pin1->coeffs[i], in2);

	if (rat_is_zero(in2))
		out.degree = 0;
	else
		out.degree = pin1->degree;
	return out;
}

// ----------------------------------------------------------------
qpoly_t qpoly_from_coeffs(
	int     degree,
	rat_t * coeffs)
{
	int i;
	qpoly_t rv;

	qpoly_check_degree(degree);
	rv.degree = degree;
	for (i = 0; i <= degree; i++)
		rv.coeffs[i] = coeffs[i];
	rv.degree = qpoly_find_degree(&rv, rv.degree);
	return rv;
}

qpoly_t qpoly_from_coeffs0(
	rat_t coeff0)
{
	qpoly_t rv;
	rv.degree = 0;
	qpoly_check_degree(rv.degree);
	rv.coeffs[0] = coeff0;
	rv.degree = qpoly_find_degree(&rv, rv.degree);
	return rv;
}

qpoly_t qpoly_from_coeffs1(
	rat_t coeff1,
	rat_t coeff0)
{
	qpoly_t rv;
	int i;
	rv.degree = 1;
	qpoly_check_degree(rv.degree);
	// xxx to common func
	for (i = 0; i <= QPOLY_MAX_DEGREE; i++)
		rv.coeffs[i] = rat_from_int(0);
	rv.coeffs[1] = coeff1;
	rv.coeffs[0] = coeff0;
	rv.degree = qpoly_find_degree(&rv, rv.degree);
	return rv;
}

qpoly_t qpoly_from_coeffs2(
	rat_t coeff2,
	rat_t coeff1,
	rat_t coeff0)
{
	qpoly_t rv;
	rv.degree = 2;
	qpoly_check_degree(rv.degree);
	rv.coeffs[2] = coeff2;
	rv.coeffs[1] = coeff1;
	rv.coeffs[0] = coeff0;
	rv.degree = qpoly_find_degree(&rv, rv.degree);
	return rv;
}

qpoly_t qpoly_binomial_from_coeffs(
	rat_t coeffa,
	int   dega,
	rat_t coeffb,
	int   degb)
{
	qpoly_t rv;
	int i;
	qpoly_check_degree(dega);
	qpoly_check_degree(degb);
	if (dega > degb)
		rv.degree = dega;
	else
		rv.degree = degb;
	for (i = 0; i <= rv.degree; i++)
		rv.coeffs[i] = rat_from_int(0);
	rv.coeffs[dega] = coeffa;
	rv.coeffs[degb] = coeffb;
	rv.degree = qpoly_find_degree(&rv, rv.degree);
	return rv;
}

// ----------------------------------------------------------------
void qpoly_print(
	qpoly_t * pin)
{
	qpoly_comma_print(pin);
}

// ----------------------------------------------------------------
void qpoly_comma_print(
	qpoly_t * pin)
{
	int i;

	for (i = pin->degree; i >= 0; i--) {
		if (i < pin->degree)
			printf(",");
		rat_print(pin->coeffs[i]);
	}
}

// ----------------------------------------------------------------
void qpoly_x_print(
	qpoly_t * pin)
{
	int i;

	for (i = pin->degree; i >= 0; i--) {
		if (i < pin->degree)
			printf(",");
		rat_print(pin->coeffs[i]);
	}
}

// ----------------------------------------------------------------
static int qpoly_list_scan(
	qpoly_t * pout,
	char * pin)
{
	int    i;
	char * coeff_stringv[QPOLY_MAX_DEGREE + 1];
	char   temp_line[1024];

	pout->degree = 0;
	for (i = 0; i <= QPOLY_MAX_DEGREE; i++)
		pout->coeffs[i] = rat_from_int(0);

	strcpy(temp_line, pin);
	pout->degree = tokenize(temp_line, ",", coeff_stringv,
		QPOLY_MAX_DEGREE + 1) - 1;
	if (pout->degree < 0) {
		fprintf(stderr, "qpoly_list_scan:  Couldn't scan \"%s\".\n",
			pin);
		return 0;
	}

	for (i = 0; i <= pout->degree; i++) {
		if (!rat_sscan(coeff_stringv[pout->degree - i],
			&pout->coeffs[i]))
		{
			fprintf(stderr,
				"qpoly:  Couldn't scan coefficient \"%s\".\n",
				coeff_stringv[pout->degree - i]);
			return 0;
		}
	}

	// Check for leading zero coefficients, which would reduce the
	// degree.
	pout->degree = qpoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
static int qpoly_exp_scan(
	qpoly_t * pout,
	char * pin)
{
	int    i;
	int    num_pieces;
	char * coeff_stringv[QPOLY_MAX_DEGREE + 1];
	char * coeff_and_exp[2];
	int    num_sub_pieces;
	char   temp_line[1024];
	char   temp_piece[128];
	int    exp;

	pout->degree = 0;
	for (i = 0; i <= QPOLY_MAX_DEGREE; i++)
		pout->coeffs[i] = rat_from_int(0);

	strcpy(temp_line, pin);
	num_pieces = tokenize(temp_line, ",", coeff_stringv,
		QPOLY_MAX_DEGREE + 1);
	if (num_pieces <= 0) {
		fprintf(stderr, "qpoly_exp_scan:  Couldn't scan \"%s\".\n",
			pin);
		return 0;
	}

	for (i = 0; i < num_pieces; i++) {
		// Split on "^"; expect ints on both sides.
		// Bounds-check the exponent.
		strcpy(temp_piece, coeff_stringv[i]);
		num_sub_pieces = tokenize(temp_piece, "^", coeff_and_exp, 2);
		if (num_sub_pieces < 1) {
			fprintf(stderr,
				"qpoly:  Couldn't scan coefficient \"%s\".\n",
				temp_piece);
			return 0;
		}
		else if (num_sub_pieces == 1) {
			exp = 0;
		}
		else {
			// sscanf and bounds-check the exp
			if (sscanf(coeff_and_exp[1], "%d", &exp) != 1) {
				fprintf(stderr,
				"qpoly:  Couldn't scan coefficient \"%s\".\n",
					temp_piece);
				return 0;
			}
			if ((exp < 0) || (exp > QPOLY_MAX_DEGREE)) {
				fprintf(stderr,
				"qpoly:  Exponent %d out of range 0-%d.\n",
					exp, QPOLY_MAX_DEGREE);
				return 0;
			}
		}

		// Remember highest exponent encountered.
		if (exp > pout->degree)
			pout->degree = exp;

		// sscanf and bounds-check the coeff
		if (!rat_sscan(coeff_and_exp[0], &pout->coeffs[exp])) {
			fprintf(stderr,
				"qpoly:  Couldn't scan coefficient \"%s\".\n",
				temp_piece);
			return 0;
		}
	}

	// Check for leading zero coefficients, which would reduce the
	// degree.
	pout->degree = qpoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
int qpoly_scan(
	qpoly_t  * pout,
	char     * pin)
{
	if (strchr(pin, '^'))
		return qpoly_exp_scan(pout, pin);
	else
		return qpoly_list_scan(pout, pin);
}

// ----------------------------------------------------------------
//     a_4 x^4 + a_3  x^3 + a_2  x^2 + a_1  x + a_0
//    (a_4 x^3 + a_3  x^2 + a_2  x   + a_1) x + a_0
//   ((a_4 x^2 + a_3  x   + a_2) x   + a_1) x + a_0
//  (((a_4 x   + a_3) x   + a_2) x   + a_1) x + a_0

rat_t qpoly_eval(qpoly_t * pz, rat_t x)
{
	rat_t rv;
	int d;

	rv = pz->coeffs[pz->degree];
	for (d = pz->degree - 1; d >= 0; d--) {
		rv = rat_mul(rv, x);
		rv = rat_add(rv, pz->coeffs[d]);
	}
	return rv;
}

// ----------------------------------------------------------------
int qpoly_is_zero(qpoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (!rat_is_zero(pz->coeffs[0]))
		return 0;
	return 1;
}

// ----------------------------------------------------------------
int qpoly_is_one(qpoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (!rat_is_one(pz->coeffs[0]))
		return 0;
	return 1;
}

// ----------------------------------------------------------------
static int qpoly_find_degree(
	qpoly_t * pin,
	int       max_deg)
{
	int i;

	for (i = max_deg; i >= 0; i--)
		if (!rat_is_zero(pin->coeffs[i]))
			return i;
	return 0;
}

// ----------------------------------------------------------------
static void qpoly_check_degree(int degree)
{
	if (degree < 0) {
		fprintf(stderr,
			"qpoly:  degree must be non-negative; got %d.\n",
			degree);
		exit(1);
	}
	if (degree > QPOLY_MAX_DEGREE) {
		fprintf(stderr,
			"qpoly:  degree must be <= %d; got %d.\n",
			QPOLY_MAX_DEGREE, degree);
		exit(1);
	}
}
