#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "qpolymod.h"
#include "tokenize.h"

// ----------------------------------------------------------------
// qpolymod_mul
// qpolymod_rarmul
// qpolymod_rermul
//
// Multiplication may be done (at least) two ways:  (1) Take the ring product
// of the two polynomials, then reduce mod r; (2) reduce mod r at every step.

// ----------------------------------------------------------------
// Sample of the ring-and-reduce method:
//
// Use GF(7^4), with reductor 1,0,1,3,5.  Multiply 1,2,3,4 by 2,4,6,1.
//
// First, multiply in the ring GF(7)[x]:
//
//       1 2 3 4
//       2 4 6 1
//      =========
//       1 2 3 4
//     6 5 4 3
//   4 1 5 2
// 2 4 6 1
// ==============
// 2 1 6 5 1 6 4
//
// Second, reduce the product mod the reductor (i.e. divide and take the
// remainder):
//
//                     2 1 4
//           +---------------
// 1 0 1 3 5 | 2 1 6 5 1 6 4
//           | 2 0 2 6 3
//           +---------------
//           |   1 4 6 5 6 4
//           |   1 0 1 3 5
//           +---------------
//           |     4 5 2 1 4
//           |     4 0 4 5 6
//           +---------------
//           |       5 5 3 5

static qpoly_t qpolymod_rarmul(
	qpoly_t * pin1,
	qpoly_t * pin2,
	qpoly_t * paux)
{
	qpoly_t   rprod, quot, out;

	// Take the polynomial product (degree may be higher than that
	// of the reductor).
	rprod = qpoly_mul(pin1, pin2);

	// Reduce the polynomial product modulo the reductor.
	qpoly_quot_and_rem(&rprod, paux, &quot, &out);
	return out;
}

qpoly_t qpolymod_mul(
	qpoly_t * pin1,
	qpoly_t * pin2,
	qpoly_t * paux)
{
	//if (...)
		//qpolymod_rermul(pout, pin1, pin2, paux);
	//else
		return qpolymod_rarmul(pin1, pin2, paux);
}

// ----------------------------------------------------------------
qpoly_t qpolymod_div(
	qpoly_t * pin1,
	qpoly_t * pin2,
	qpoly_t * paux)
{
	qpoly_t   binv;

	// 1/b mod r
	binv = qpolymod_recip(pin2, paux);

	// a * (1/b) mod r
	return qpolymod_mul(pin1, &binv, paux);
}

// ----------------------------------------------------------------
qpoly_t qpolymod_recip(
	qpoly_t * pin,
	qpoly_t * paux)
{
	qpoly_t g, m, n;
	rat_t c0i;
	int   i;

	g = qpoly_ext_gcd(paux, pin, &m, &n);

	// Ext. GCD actually returns a scalar multiple of 1.  Divide this out.
	c0i = rat_recip(g.coeffs[0]);
	for (i = 0; i <= n.degree; i++)
		n.coeffs[i] = rat_mul(n.coeffs[i], c0i);
	return n;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
qpoly_t qpolymod_power(
	qpoly_t * px,
	unsigned  power,
	qpoly_t * paux)
{
	qpoly_t x2 = *px;
	qpoly_t rv;

	rv = qpoly_from_coeffs0(rat_from_int(1));

	while (power != 0) {
		if (power & 1)
			rv = qpolymod_mul(&x2, &rv, paux);
		power >>= 1;
		x2 = qpolymod_mul(&x2, &x2, paux);
	}
	return rv;
}
