#include <stdio.h>
#include <stdlib.h>
#include "gcd.h"
#include "rat.h"
#include "cmps.h"

static void rat_simplify(rat_t * p);

// ----------------------------------------------------------------
//  a     c     ad     bc     ad + bc
// --- + --- = ---- + ---- = ---------
//  b     d     bd     bd       bd

rat_t rat_add(rat_t a, rat_t b)
{
	rat_t c;
	int  g, l, r;

	g = calc_gcd_signed(a.denom, b.denom);
	l = b.denom / g;
	r = a.denom / g;

	c.numer = a.numer * l + b.numer * r;
	c.denom = a.denom * l;

	rat_simplify(&c);
	return c;
}

// ----------------------------------------------------------------
//  a     c     ad     bc     ad - bc
// --- - --- = ---- - ---- = ---------
//  b     d     bd     bd       bd

rat_t rat_sub(rat_t a, rat_t b)
{
	rat_t c;
	int g, l, r;

	g = calc_gcd_signed(a.denom, b.denom);
	l = b.denom / g;
	r = a.denom / g;

	c.numer = a.numer * l - b.numer * r;
	c.denom = a.denom * l;

	rat_simplify(&c);
	return c;
}

// ----------------------------------------------------------------
rat_t rat_usub(rat_t a)
{
	rat_t rv;
	rv.numer = -a.numer;
	rv.denom =  a.denom;
	return rv;
}

// ----------------------------------------------------------------
rat_t rat_mul(rat_t a, rat_t b)
{
	rat_t c;

	c.numer = a.numer * b.numer;
	c.denom = a.denom * b.denom;

	rat_simplify(&c);
	return c;
}

// ----------------------------------------------------------------
//  a     c     a     d     ad
// --- / --- = --- * --- = ----
//  b     d     b     c     bc

rat_t rat_div(rat_t a, rat_t b)
{
	rat_t c;

	c.numer = a.numer * b.denom;
	c.denom = a.denom * b.numer;

	rat_simplify(&c);
	return c;
}

// ----------------------------------------------------------------
rat_t rat_recip(rat_t a)
{
	rat_t c;

	c.numer = a.denom;
	c.denom = a.numer;

	rat_simplify(&c); // Checks for divide by zero
	return c;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
rat_t rat_power(rat_t a, int e)
{
	rat_t a2, apower;

	if (e < 0) {
		// Breaks for MININT; I don't care.
		e = -e;
		a = rat_recip(a);
	}
	a2 = a;
	apower = rat_from_int(1);

	while (e != 0) {
		if (e & 1)
			apower = rat_mul(a2, apower);
		e = (unsigned)e >> 1;
		a2 = rat_mul(a2, a2);
	}
	return apower;
}

// ----------------------------------------------------------------
int rat_equal(rat_t a, rat_t b)
{
	if (a.numer != b.numer)
		return 0;
	if (a.denom != b.denom)
		return 0;
	return 1;
}

// ----------------------------------------------------------------
int rat_cmp(rat_t a, int cmp, rat_t b)
{
	int direction = 0; // -1 = less, 0 = equal, +1 = greater;
	int an_bd = a.numer * b.denom;
	int ad_bn = a.denom * b.numer;

	if (an_bd < ad_bn) {
		direction = -1;
	}
	else if (an_bd > ad_bn) {
		direction =  1;
	}

	switch (cmp) {
	case CMP_LE: return (direction <= 0); break;
	case CMP_LT: return (direction <  0); break;
	case CMP_EQ: return (direction == 0); break;
	case CMP_NE: return (direction != 0); break;
	case CMP_GE: return (direction >= 0); break;
	case CMP_GT: return (direction >  0); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); return 0; break;
	}
}

// ----------------------------------------------------------------
int rat_lt(rat_t a, rat_t b)
{
	return rat_cmp(a, CMP_LT, b);
}
int rat_le(rat_t a, rat_t b)
{
	return rat_cmp(a, CMP_LE, b);
}
int rat_gt(rat_t a, rat_t b)
{
	return rat_cmp(a, CMP_GT, b);
}
int rat_ge(rat_t a, rat_t b)
{
	return rat_cmp(a, CMP_GE, b);
}

// ----------------------------------------------------------------
// For use with qsort(3).
int rat_qcmp(
	const void * pv1,
	const void * pv2)
{
	rat_t r1 = *(const rat_t *)pv1;
	rat_t r2 = *(const rat_t *)pv2;
	if (rat_lt(r1, r2))
		return -1;
	if (rat_gt(r1, r2))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
rat_t rat_from_int(int numer)
{
	rat_t rv;
	rv.numer = numer;
	rv.denom = 1;
	return rv;
}

// ----------------------------------------------------------------
rat_t rat_from_ints(int numer,int denom)
{
	rat_t rv;
	rv.numer = numer;
	rv.denom = denom;
	rat_simplify(&rv);
	return rv;
}

// ----------------------------------------------------------------
int rat_is_zero(rat_t a)
{
	if (a.numer == 0)
		return 1;
	return 0;
}

// ----------------------------------------------------------------
int rat_is_one(rat_t a)
{
	if ((a.numer == 1) && (a.denom == 1))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
void rat_print(rat_t a)
{
	printf("%d", a.numer);
	if (a.denom != 1)
		printf("/%d", a.denom);
}

// ----------------------------------------------------------------
int rat_sscan(
	char  * pin,
	rat_t * pout)
{
	if (sscanf(pin, "%d/%d", &pout->numer, &pout->denom) == 2) {
		rat_simplify(pout);
		return 1;
	}
	pout->denom = 1;
	if (sscanf(pin, "%d", &pout->numer) == 1)
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static void rat_simplify(rat_t * p)
{
	int g;
	if (p->denom == 0) {
		fprintf(stderr, "rat: Divide by zero.\n");
		exit(1);
	}
	if (p->denom < 0) {
		p->numer = -p->numer;
		p->denom = -p->denom;
	}
	g = calc_gcd_signed(p->numer, p->denom);
	p->numer /= g;
	p->denom /= g;
}
