// Symmetries of the tetrahedron:
//
//   A      A      A     //
//  /|\    /|\    /|\    //
// D-|-B  C-|-D  B-|-C   //
//  \|/    \|/    \|/    //
//   C      B      D     //
//                       //
//   D      D      D     //
//  /|\    /|\    /|\    //
// A-|-C  B-|-A  C-|-B   //
//  \|/    \|/    \|/    //
//   B      C      A     //
//                       //
//   B      B      B     //
//  /|\    /|\    /|\    //
// A-|-D  C-|-A  D-|-C   //
//  \|/    \|/    \|/    //
//   C      D      A     //
//                       //
//   C      C      C     //
//  /|\    /|\    /|\    //
// L-|-A  B-|-L  A-|-B   //
//  \|/    \|/    \|/    //
//   B      A      L     //

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tetra.h"
#include "pmt.h"

// ----------------------------------------------------------------
static void tetra_vprint(
	void * pvin,
	void * pvaux)
{
	pmt_t * ps = (pmt_t *)pvin;
	int i;

	for (i = 0; i < 4; i++) {
		printf("%c", ps->images[i] + 'A');
	}
}

// ----------------------------------------------------------------
int tetra_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	pmt_t * ppmt = (pmt_t *)pvout;
	int i;

	if (sizeof(pmt_t) > max_elt_bytes) {
		fprintf(stderr, "tetra:  Need %d bytes for element.\n",
			sizeof(pmt_t));
		exit(1);
	}

	if (strlen(pin) != 4) {
		fprintf(stderr,
			"tetra element input \"%s\" must have length 4.\n",
			pin);
		return 0;
	}
	for (i = 0; i < 4; i++) {
		if ((pin[i] >= 'A') && (pin[i] <= 'Z'))
			ppmt->images[i] = pin[i] - 'A';
		else if ((pin[i] >= 'a') && (pin[i] <= 'z'))
			ppmt->images[i] = pin[i] - 'a';
		else if ((pin[i] >= '0') && (pin[i] <= '9'))
			ppmt->images[i] = pin[i] - '0';
		else {
			fprintf(stderr,
				"Malformed tetra element \"%s\".\n",
				pin);
			return 0;
		}
	}

	if (!pmt_validate_element(ppmt, 4)) {
		fprintf(stderr,
			"tetra element \"%s\" is not a permuation.\n",
			pin);
		return 0;
	}

	return 1;
}

// ----------------------------------------------------------------
int tetra_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	if (argc != 0)
		return 0;
	*(int *)pvaux = 4;
	return 1;
}

// ----------------------------------------------------------------
int  tetra_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	pmt_t * pelements = (pmt_t *) pvelements;
	unsigned char * psrc = 0, * pdst = 0;
	int i, j;

	static char * strings[] = {
		"ADBC", "ACDB", "ABCD",
		"DACB", "DBAC", "DCBA",
		"BADC", "BCAD", "BDCA",
		"CDAB", "CBDA", "CABD"
	};

	*pnum_elements = 12;

	if (*pnum_elements > max_group_bytes/sizeof(pmt_t))
		return 0;

	for (i = 0; i < 12; i++) {
		psrc = strings[i];
		pdst = pelements[i].images;
		for (j = 0; j < 4; j++)
			pdst[j] = psrc[j] - 'A';
	}

	return 1;
}

// ================================================================
group_spec_t tetra_group_spec = {
	element_size  : sizeof(pmt_t),
	naux          : 0,
	pvop          : pmt_vmul,
	pvequal       : pmt_vequal,
	pvless        : pmt_vless,
	pvprint       : tetra_vprint,
	pvauxprint     : 0,
	pvscan        : tetra_vscan,
	pvauxscan      : tetra_vauxscan,
	pfill_group   : tetra_fill_group,
	desc          : "tetra"
};
