#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "ubig.h"
#include "log2.h"

static void ubig_accum_pprod(
	ubig * pc,
	ubig_part_t pprod,
	int idx);

static void ubig_quot_rem(
	ubig * pquot,
	ubig * prem,
	ubig * pdividend,
	ubig * pdivisor);

static int  ubig_find_most_significant_one(
	ubig * pc,
	int  * pidx);

static void ubig_rsh1(ubig * pc);

// ----------------------------------------------------------------
static void scan_add_digit(ubig * pa, unsigned base, unsigned digit)
{
	int i;
	ubig_pprod_t pprod;
	ubig_part_t  carry;

	carry = 0;
	for (i = 0; i < UBIG_NUM_PARTS; i++) {
		pprod = carry + pa->parts[i] * base;
		carry = pprod >> 16;
		pa->parts[i] = pprod & 0xffff;
	}

	carry = digit;
	for (i = 0; i < UBIG_NUM_PARTS; i++) {
		pprod = pa->parts[i] + carry;
		carry = pprod >> 16;
		pa->parts[i] = pprod & 0xffff;
		if (carry == 0)
			break;
	}
}

// ----------------------------------------------------------------
//        0 1d20 8a5a
//   +---------------
// a | 0001 2345 6789
//   | 0000
//   + --------------
//   | 0001 2345      (12345 / a = 1d20; 12345 % a = 5)
//   | 0001 2340      (a * 1d20 = 12340)
//   + --------------
//   |         5 6789 (56789 / a = 8a5a; 56789 % a = 5)
//   |         5 6784 (a * 8a5a = 56784)
//   + --------------
//                  5
//
// So, quotient is 0000_1d20_8a5a and remainder is 5.

static void print_del_digit(ubig * pa, unsigned base, unsigned * pdigit,
	int * pis_zero)
{
	int i;
	ubig_pprod_t partial_dividend;
	ubig_pprod_t partial_quotient;
	ubig_pprod_t partial_remainder;
	ubig_pprod_t subtrahend;

	*pis_zero = 1;
	partial_dividend = pa->parts[UBIG_NUM_PARTS-1];
	for (i = UBIG_NUM_PARTS-1; i >= 0; i--) {
		partial_quotient  = partial_dividend / base;
		partial_remainder = partial_dividend % base;
		subtrahend = partial_quotient * base;
		if (partial_quotient & 0x10000) {
			fprintf(stderr, "Ouch!\n");
			exit(1);
		}
		pa->parts[i] = partial_quotient;
		if (partial_quotient != 0)
			*pis_zero = 0;
		partial_dividend -= subtrahend;
		if (i > 0)
			partial_dividend =
				(partial_dividend << 16) | pa->parts[i-1];
	}
	*pdigit = partial_remainder;
}

static int ubig_fscan_hex(FILE * fp, ubig * pb, ubig_opts_t * pubig_opts)
{
	char ch;
	int i;
	unsigned xdigit;

	for (i = UBIG_NUM_PARTS-1; i >= 0; i--)
		pb->parts[i] = 0;

	ch = getc(fp);
	while (isxdigit(ch)) {
		if (isdigit(ch))
			xdigit = ch - '0';
		else if (islower(ch))
			xdigit = ch - 'a' + 10;
		else
			xdigit = ch - 'A' + 10;
		scan_add_digit(pb, 0x10, xdigit);
		ch = getc(fp);
	}

	if (ch != EOF)
		ungetc(ch, fp);

	return 1;
}

static int ubig_fscan_dec(FILE * fp, ubig * pb, ubig_opts_t * pubig_opts)
{
	char ch;
	int i;

	for (i = UBIG_NUM_PARTS-1; i >= 0; i--)
		pb->parts[i] = 0;

	ch = getc(fp);
	while (isdigit(ch)) {
		scan_add_digit(pb, 10, ch - '0');
		ch = getc(fp);
	}

	if (ch != EOF)
		ungetc(ch, fp);

	return 1;
}

int  ubig_fscan (FILE * fp, ubig * pc, ubig_opts_t * pubig_opts)
{
	if (pubig_opts->hex_in)
		return ubig_fscan_hex(fp, pc, pubig_opts);
	else
		return ubig_fscan_dec(fp, pc, pubig_opts);
}

// ----------------------------------------------------------------
static int ubig_sscan_hex(char * pin, ubig * pb, ubig_opts_t * pubig_opts)
{
	char * p;
	int i;
	unsigned xdigit;

	for (i = UBIG_NUM_PARTS-1; i >= 0; i--)
		pb->parts[i] = 0;

	for (p = pin; *p; p++) {
		if (isdigit(*p)) {
			xdigit = *p - '0';
		}
		else if (isxdigit(*p)) {
			if (islower(*p))
				xdigit = *p - 'a' + 10;
			else
				xdigit = *p - 'A' + 10;
		}
		else {
			fprintf(stderr,
			"ubig_sscan_hex():  Non-hex input '%c' in \"%s\".\n",
				*p, pin);
			exit(1);
		}
		scan_add_digit(pb, 0x10, xdigit);
	}

	return 1;
}

static int ubig_sscan_dec(char * pin, ubig * pb, ubig_opts_t * pubig_opts)
{
	char * p;
	int i;
	unsigned digit;

	for (i = UBIG_NUM_PARTS-1; i >= 0; i--)
		pb->parts[i] = 0;

	for (p = pin; *p; p++) {
		if (isdigit(*p)) {
			digit = *p - '0';
		}
		else {
			fprintf(stderr, "ubig_sscan_dec():  Non-decimal "
				"input '%c' in \"%s\".\n",
				*p, pin);
			exit(1);
		}
		scan_add_digit(pb, 10, digit);
	}

	return 1;
}

int ubig_sscan(char * pin, ubig * pc, ubig_opts_t * pubig_opts)
{
	if (pubig_opts->hex_in)
		return ubig_sscan_hex(pin, pc, pubig_opts);
	else
		return ubig_sscan_dec(pin, pc, pubig_opts);
}

// ----------------------------------------------------------------
int ubig_sdscan(char * pin, ubig * pc)
{
	ubig_opts_t opts;
	opts.hex_in = 0;
	return ubig_sscan_dec(pin, pc, &opts);
}

// ----------------------------------------------------------------
static void ubig_print_hex(FILE * fp, ubig * pc, ubig_opts_t * pubig_opts)
{
	int i;
	int got_non_zero = 0;
	char * fmt_first = 0;
	char * fmt_rest = 0;
	char * fmt;

	if (pubig_opts->zero_fill) {
		fmt_first = "%04hx";
		fmt_rest  = "%04hx";
	}
	else {
		fmt_first = "%hx";
		fmt_rest  = "%04hx";
	}
	fmt = fmt_first;

	// 0000 0123 4560 8abc
	// nz=0 nz=1

	for (i = UBIG_NUM_PARTS-1; i >= 0; i--) {
		if (pc->parts[i])
			got_non_zero = 1;
		if (i == 0)
			got_non_zero = 1;
		if (got_non_zero || pubig_opts->zero_fill)
			fprintf(fp, fmt, pc->parts[i]);
		if (pubig_opts->underscore_out)
			if (i != 0)
				fprintf(fp, "_");
		if (got_non_zero)
			fmt = fmt_rest;
	}
}

static void ubig_print_dec(FILE * fp, ubig * pc, ubig_opts_t * pubig_opts)
{
	char rev_chars[2 * UBIG_NUM_NYBBLES];
	char fwd_chars[2 * UBIG_NUM_NYBBLES];
	int  revi;
	int  fwdi;
	int  is_zero = 0;
	unsigned digit;
	ubig temp;

	temp = *pc;
	revi = 0;
	while (!is_zero) {
		print_del_digit(&temp, 10, &digit, &is_zero);
		rev_chars[revi]   = digit + '0';
		rev_chars[revi+1] = 0;
		revi++;
	}

	fwdi = 0;
	revi--;
	while (revi >= 0) {
		fwd_chars[fwdi++] = rev_chars[revi--];
	}
	fwd_chars[fwdi] = 0;
	fprintf(fp, "%s", fwd_chars);
}

void ubig_print(FILE * fp, ubig * pc, ubig_opts_t * pubig_opts)
{
	if (pubig_opts->hex_out)
		ubig_print_hex(fp, pc, pubig_opts);
	else
		ubig_print_dec(fp, pc, pubig_opts);
}

// ----------------------------------------------------------------
void ubig_xprint(ubig * pc)
{
	ubig_opts_t o;
	o.hex_out = 1;
	o.zero_fill = 1;
	o.underscore_out = 1;
	ubig_print(stdout, pc, &o);
}

void ubig_dprint(ubig * pc)
{
	ubig_opts_t o;
	o.hex_out = 0;
	o.zero_fill = 0;
	o.underscore_out = 0;
	ubig_print(stdout, pc, &o);
}

void ubig_bprint(ubig *pc)
{
	int parti;
	int biti;
	for (parti = UBIG_NUM_PARTS-1; parti >= 0; parti--) {
		for (biti = 15; biti >= 0; biti--) {
			if (pc->parts[parti] & (1 << biti))
				printf("1");
			else
				printf("0");
			if ((biti & 3) == 0)
				if (biti != 0)
					printf(" ");
		}
		if (parti != 0)
			printf("  ");
	}
	printf("\n");
}

void ubig_beq(void)
{
	int parti;
	int biti;
	for (parti = UBIG_NUM_PARTS-1; parti >= 0; parti--) {
		for (biti = 15; biti >= 0; biti--) {
			printf("=");
			if ((biti & 3) == 0)
				if (biti != 0)
					printf(" ");
		}
		if (parti != 0)
			printf("  ");
	}
	printf("\n");
}

// ----------------------------------------------------------------
ubig ubig_or(ubig * pa, ubig * pb)
{
	ubig c;
	int i;
	for (i = 0; i < UBIG_NUM_PARTS; i++)
		c.parts[i] = pa->parts[i] | pb->parts[i];
	return c;
}

// ----------------------------------------------------------------
ubig ubig_xor(ubig * pa, ubig * pb)
{
	ubig c;
	int i;
	for (i = 0; i < UBIG_NUM_PARTS; i++)
		c.parts[i] = pa->parts[i] ^ pb->parts[i];
	return c;
}

// ----------------------------------------------------------------
ubig ubig_and(ubig * pa, ubig * pb)
{
	ubig c;
	int i;
	for (i = 0; i < UBIG_NUM_PARTS; i++)
		c.parts[i] = pa->parts[i] & pb->parts[i];
	return c;
}

// ----------------------------------------------------------------
ubig ubig_lsh(ubig * pa, ubig * pb)
{
	return ubig_ilsh(pa, pb->parts[0]);
}

ubig ubig_rsh(ubig * pa, ubig * pb)
{
	return ubig_irsh(pa, pb->parts[0]);
}

// ----------------------------------------------------------------
ubig ubig_ilsh(ubig * pa, unsigned bits_shift)
{
	ubig c;
	ubig temp;
	unsigned parts_shift;
	unsigned parts_diff;
	unsigned in_part_shift;
	unsigned in_part_diff;
	int srci;
	int dsti;
	int i;

	if (bits_shift >= UBIG_NUM_BITS) {
		memset(&c, 0, sizeof(ubig));
		return c;
	}

	parts_shift   = bits_shift >> UBIG_PART_LOG;
	in_part_shift = bits_shift &  UBIG_PART_MASK;
	parts_diff    = UBIG_NUM_PARTS - parts_shift;
	in_part_diff  = UBIG_BITS_PER_PART - in_part_shift;

	for (dsti = 0; dsti < parts_shift; dsti++)
		temp.parts[dsti] = 0;
	for (dsti = parts_shift, srci = 0; srci < parts_diff; srci++, dsti++)
		temp.parts[dsti] = pa->parts[srci];

	for (i = UBIG_NUM_PARTS-1; i >= 0; i--) {
		c.parts[i] = temp.parts[i] << in_part_shift;
		if (i != 0) {
			if (in_part_diff != UBIG_BITS_PER_PART)
				c.parts[i] |= temp.parts[i-1] >> in_part_diff;
		}
	}
	return c;
}

// ----------------------------------------------------------------
ubig ubig_irsh(ubig * pa, unsigned bits_shift)
{
	ubig c;
	ubig temp;
	unsigned parts_shift;
	unsigned parts_diff;
	unsigned in_part_shift;
	unsigned in_part_diff;
	int srci;
	int dsti;
	int i;

	if (bits_shift >= UBIG_NUM_BITS) {
		memset(&c, 0, sizeof(ubig));
		return c;
	}
	parts_shift   = bits_shift >> UBIG_PART_LOG;
	in_part_shift = bits_shift &  UBIG_PART_MASK;
	parts_diff    = UBIG_NUM_PARTS - parts_shift;
	in_part_diff  = UBIG_BITS_PER_PART - in_part_shift;

	for (srci = parts_shift, dsti = 0; dsti < parts_diff; dsti++, srci++)
		temp.parts[dsti] = pa->parts[srci];
	for (dsti = parts_diff; dsti < UBIG_NUM_PARTS; dsti++)
		temp.parts[dsti] = 0;

	for (i = UBIG_NUM_PARTS-1; i >= 0; i--) {
		c.parts[i] = temp.parts[i] >> in_part_shift;
		if (i != UBIG_NUM_PARTS-1) {
			if (in_part_diff != UBIG_BITS_PER_PART)
				c.parts[i] |= temp.parts[i+1] << in_part_diff;
		}
	}
	return c;
}

// ----------------------------------------------------------------
static void ubig_rsh1(ubig * pc)
{
	int i;
	for (i = 0; i < UBIG_NUM_PARTS; i++) {
		pc->parts[i] >>= 1;
		if (i != UBIG_NUM_PARTS-1)
			pc->parts[i] |= pc->parts[i+1]
				<< (UBIG_BITS_PER_PART-1);
	}
}

// ----------------------------------------------------------------
void ubig_set_bit(
	ubig * pc,
	unsigned pos)
{
	int parti = pos >> UBIG_PART_LOG;
	int biti  = pos &  UBIG_PART_MASK;
	pc->parts[parti] |= (1 << biti);
}

// ----------------------------------------------------------------
unsigned ubig_get_bit(
	ubig * pc,
	unsigned pos)
{
	int parti = pos >> UBIG_PART_LOG;
	int biti  = pos &  UBIG_PART_MASK;
	return 1 & (pc->parts[parti] >> biti);
}

// ----------------------------------------------------------------
unsigned ubig_find_msb(ubig * pc)
{
	int parti;

	for (parti = UBIG_NUM_PARTS-1; parti >= 0; parti--) {
		if (pc->parts[parti] != 0) {
			return find_msb_16(pc->parts[parti])
				+ parti * UBIG_BITS_PER_PART;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
ubig ubig_add(ubig * pa, ubig * pb)
{
	ubig c;
	int i;
	ubig_part_t   carry = 0;
	ubig_part_t * paparts = &pa->parts[0];
	ubig_part_t * pbparts = &pb->parts[0];
	ubig_part_t * pcparts = &c.parts[0];
	ubig_part_t   msba, msbb, msbc;
	int abmspart = -1;

	memset(&c, 0, sizeof(ubig));
	for (i = UBIG_NUM_PARTS-1; i >= 0; i--) {
		if ((paparts[i] != 0) || (pbparts[i] != 0)) {
			abmspart = i;
			break;
		}
	}
	if (abmspart == -1) {
		return c;
	}

	for (i = 0; i < UBIG_NUM_PARTS; i++) {
		pcparts[i] = paparts[i] + pbparts[i] + carry;
		msba = paparts[i] & UBIG_PART_MSB;
		msbb = pbparts[i] & UBIG_PART_MSB;
		msbc = pcparts[i] & UBIG_PART_MSB;

		carry = 0;
		if (msba & msbb) {
			carry = 1;
		}
		else if (msba | msbb) {
			if (!msbc)
				carry = 1;
		}
		if ((carry == 0) && (i > abmspart))
			break;
	}
	return c;
}

// ----------------------------------------------------------------
ubig ubig_sub(ubig * pa, ubig * pb)
{
	ubig bneg = ubig_usub(pb);
	return ubig_add(pa, &bneg);
}

// ----------------------------------------------------------------
//                         +--+--+--+--+--+--+--+--+
//                         |a0|a1|a2|a3|a4|a5|a6|a7|
//                         +--+--+--+--+--+--+--+--+
//                         *
//                         +--+--+--+--+--+--+--+--+
//                         |b0|b1|b2|b3|b4|b5|b6|b7|
//                         +--+--+--+--+--+--+--+--+
// ==================================================
// : 0. 1: 2. 3: 4. 5: 6. 7| 8. 9:10.11:12.13:14.15:
//                         | 0. 1: 2. 3: 4. 5: 6. 7:
// ==================================================

// xxx update ASCII art to match reversed index

// :  .  :  .  :  .  :  .  |  .  :  .  :  .  :a7*b7:
// :  .  :  .  :  .  :  .  |  .  :  .  :  .a6*b7.  :
// :  .  :  .  :  .  :  .  |  .  :  .  :a5*b7:  .  :
// :  .  :  .  :  .  :  .  |  .  :  .a4*b7.a7*b6.  :
// :  .  :  .  :  .  :  .  |  .  :a3*b7:a6*b6:  .  :
// :  .  :  .  :  .  :  .  |  .a2*b7.a5*b6.  :  .  :
// :  .  :  .  :  .  :  .  |a1*b7:a4*b6:a7*b5:  .  :
// :  .  :  .  :  .  :  .a0*b7.a3*b6.a6*b5.  :  .  :
// :  .  :  .  :  .  :  .  |a2*b6:a5*b5:  .  :  .  :
// :  .  :  .  :  .  :  .a1*b6.a4*b5.a7*b4.  :  .  :
// :  .  :  .  :  .  :a0*b6|a3*b5:a6*b4:  .  :  .  :
// :  .  :  .  :  .  :  .a2*b5.a5*b4.  .  .  :  .  :
// :  .  :  .  :  .  :a1*b5|a4*b4:a7*b3:  .  :  .  :
// :  .  :  .  :  .a0*b5.a3*b4.a6*b3.  .  .  :  .  :
// :  .  :  .  :  .  :a2*b4|a5*b3:  .  :  .  :  .  :
// :  .  :  .  :  .a1*b4.a4*b3.a7*b2.  .  .  :  .  :
// :  .  :  .  :a0*b4:a3*b3|a6*b2:  .  :  .  :  .  :
// :  .  :  .  :  .a2*b3.a5*b2.  .  .  :  .  :  .  :
// :  .  :  .  :a1*b3:a4*b2|a7*b1:  .  :  .  :  .  :
// :  .  :  .a0*b3.a3*b2.a6*b1.  .  .  :  .  :  .  :
// :  .  :  .  :a2*b2:a5*b1|  .  :  .  :  .  :  .  :
// :  .  :  .a1*b2.a4*b1.a7*b0.  .  .  :  .  :  .  :
// :  .  :a0*b2:a3*b1:a6*b0|  .  :  .  :  .  :  .  :
// :  .  :  .a2*b1.a5*b0.  |  .  :  .  :  .  :  .  :
// :  .  :a1*b1:a4*b0:  .  |  .  :  .  :  .  :  .  :
// :  .a0*b1.a3*b0.  .  .  |  .  :  .  :  .  :  .  :
// :  .  :a2*b0:  .  :  .  |  .  :  .  :  .  :  .  :
// :  .a1*b0.  .  .  :  .  |  .  :  .  :  .  :  .  :
// :a0*b0:  .  :  .  :  .  |  .  :  .  :  .  :  .  :
// :  .  :  .  :  .  :  .  |  .  :  .  :  .  :  .  :

ubig ubig_mul(ubig * pa, ubig * pb)
{
	int i, j, k;
	ubig_pprod_t pprod;
	ubig_part_t  pphi, pplo;
	ubig c;
	int amspart = -1, bmspart = -1;

	memset(&c, 0, sizeof(ubig));

	for (i = UBIG_NUM_PARTS-1; i >= 0; i--) {
		if (pa->parts[i] != 0) {
			amspart = i;
			break;
		}
	}
	for (j = UBIG_NUM_PARTS-1; j >= 0; j--) {
		if (pb->parts[j] != 0) {
			bmspart = j;
			break;
		}
	}
	if ((amspart == -1) || (bmspart == -1)) {
		return c;
	}

	for (i = 0; i <= amspart; i++) {
		for (j = 0; j <= bmspart; j++) {
			k = i + j;
			if (k >= UBIG_NUM_PARTS)
				continue;

			pprod = (ubig_pprod_t)pa->parts[i]
				* (ubig_pprod_t)pb->parts[j];
			pphi = pprod >> 16;
			pplo = pprod & 0xffff;
			if (pphi)
				ubig_accum_pprod(&c, pphi, k+1);
			if (pplo)
				ubig_accum_pprod(&c, pplo, k);
		}
	}
	return c;
}

// ----------------------------------------------------------------
static void ubig_accum_pprod(ubig * pc, ubig_part_t pprod, int idx)
{
	int i;
	ubig_pprod_t temp;
	ubig_part_t  carry;

	temp  = pc->parts[idx] + pprod;
	carry = (temp & UBIG_CARRY_BIT) ? 1 : 0;
	pc->parts[idx] = temp;

	for (i = idx + 1; i < UBIG_NUM_PARTS; i++) {
		temp = pc->parts[i] + carry;
		carry = (temp & UBIG_CARRY_BIT) ? 1 : 0;
		pc->parts[i] = temp;
		if (!carry)
			break;
	}
}

// ----------------------------------------------------------------
ubig ubig_div(ubig * pa, ubig * pb)
{
	ubig quot, rem;
	ubig_quot_rem(&quot, &rem, pa, pb);
	return quot;
}

// ----------------------------------------------------------------
ubig ubig_mod(ubig * pa, ubig * pb)
{
	ubig quot, rem;
	ubig_quot_rem(&quot, &rem, pa, pb);
	return rem;
}

// ----------------------------------------------------------------
static void ubig_quot_rem(
	ubig * pquot,
	ubig * prem,
	ubig * pdividend,
	ubig * pdivisor)
{
	int dividend_l1_pos;
	int divisor_l1_pos;
	int l1_diff;
	ubig shift_divisor; // Make a copy so we don't modify argument
	int check_pos;
	int quot_pos;
	ubig quot, rem;

	memset(&quot, 0, sizeof(ubig));
	memset(&rem,  0, sizeof(ubig));

	if (!ubig_find_most_significant_one(pdividend, &dividend_l1_pos)) {
		// Dividend is zero.
		*pquot = quot;
		*prem  = rem;
		return;
	}
	if (!ubig_find_most_significant_one(pdivisor, &divisor_l1_pos)) {
		// Divisor is zero.
		fprintf(stderr, "ubig:  Divide by zero.\n");
		exit(1);
	}
	l1_diff = dividend_l1_pos - divisor_l1_pos;

	if (ubig_cmp(pdividend, CMP_LT, pdivisor)) {
		*pquot = quot;
		*prem  = *pdividend;
		return;
	}

	if (l1_diff < 0) {
		// Should not happen, since dividend is already known
		// be >= divisor.
		fprintf(stderr, "Coding error file %s line %d.\n",
			__FILE__,__LINE__);
		exit(1);
	}

	shift_divisor = ubig_ilsh(pdivisor, l1_diff);

	rem = *pdividend;
	for (
		check_pos = dividend_l1_pos, quot_pos = l1_diff;
		check_pos >= divisor_l1_pos;
		check_pos--, quot_pos--)
	{
		if (ubig_cmp(&rem, CMP_GE, &shift_divisor)) {
			rem = ubig_sub(&rem, &shift_divisor);
			ubig_set_bit(&quot, quot_pos);
		}
		ubig_rsh1(&shift_divisor);
	}
	*pquot = quot;
	*prem  = rem;
}

// ----------------------------------------------------------------
static int ubig_find_most_significant_one(ubig * pc, int * pidx)
{
	int parti;
	int biti;

	*pidx = UBIG_NUM_BITS - 1;

	for (parti = UBIG_NUM_PARTS-1; parti >= 0; parti--)
		for (biti = UBIG_BITS_PER_PART - 1; biti >= 0; biti--, (*pidx)--)
			if (pc->parts[parti] & (1 << biti))
				return 1;

	return 0;
}

// ----------------------------------------------------------------
static int ubig_raw_cmp(ubig * pa, ubig * pb)
{
	int i;

	for (i = UBIG_NUM_PARTS-1; i >= 0; i--) {
		if (pa->parts[i] < pb->parts[i]) {
			return -1;
			break;
		}
		else if (pa->parts[i] > pb->parts[i]) {
			return 1;
			break;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
int ubig_cmp(ubig * pa, int cmp, ubig * pb)
{
	int direction = 0; // -1 = less, 0 = equal, +1 = greater;

	direction = ubig_raw_cmp(pa, pb);

	switch (cmp) {
	case CMP_LE: return (direction <= 0); break;
	case CMP_LT: return (direction <  0); break;
	case CMP_EQ: return (direction == 0); break;
	case CMP_NE: return (direction != 0); break;
	case CMP_GE: return (direction >= 0); break;
	case CMP_GT: return (direction >  0); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); return 0; break;
	}
}

int  ubig_lt(ubig * pa, ubig * pb)
{
	return ubig_cmp(pa, CMP_LT, pb);
}
int  ubig_le(ubig * pa, ubig * pb)
{
	return ubig_cmp(pa, CMP_LE, pb);
}
int  ubig_gt(ubig * pa, ubig * pb)
{
	return ubig_cmp(pa, CMP_GT, pb);
}
int  ubig_ge(ubig * pa, ubig * pb)
{
	return ubig_cmp(pa, CMP_GE, pb);
}
int  ubig_eq(ubig * pa, ubig * pb)
{
	return ubig_cmp(pa, CMP_EQ, pb);
}
int  ubig_ne(ubig * pa, ubig * pb)
{
	return ubig_cmp(pa, CMP_NE, pb);
}

// ----------------------------------------------------------------
int ubig_qcmp(const void * pv1, const void * pv2)
{
	const ubig * p1 = pv1;
	const ubig * p2 = pv2;
	return ubig_raw_cmp((ubig *)p1, (ubig *)p2);
}

// ----------------------------------------------------------------
ubig ubig_ipower(ubig * pa, unsigned e)
{
	ubig c;
	ubig p2;
	int i;

	c = ubig_from_unsigned(1);

	p2 = *pa;
	for (i = 0; i < 32; i++) {
		if (e & 1) {
			c = ubig_mul(&c, &p2);
		}
		e >>= 1;
		p2 = ubig_mul(&p2, &p2);
	}
	return c;
}

// ----------------------------------------------------------------
ubig ubig_power(ubig * pa, ubig * pb)
{
	// xxx need to use all bits of b.
	return ubig_ipower(pa, pb->parts[0]);
}

// ----------------------------------------------------------------
ubig ubig_uadd(ubig * pa)
{
	return *pa;
}

// ----------------------------------------------------------------
ubig ubig_usub(ubig * pa)
{
	ubig c = ubig_unot(pa);
	ubig_incr(&c);
	return c;
}

// ----------------------------------------------------------------
ubig ubig_unot(ubig * pa)
{
	ubig c;
	int i;
	for (i = 0; i < UBIG_NUM_PARTS; i++)
		c.parts[i] = ~pa->parts[i];
	return c;
}

// ----------------------------------------------------------------
void ubig_incr(ubig * pc)
{
	int i;
	ubig_part_t carry = 1;
	ubig_part_t * pcparts = &pc->parts[0];

	for (i = 0; i < UBIG_NUM_PARTS; i++) {
		pcparts[i] = pcparts[i] + carry;
		if ((pcparts[i] == 0) && carry)
			carry = 1;
		else
			carry = 0;
	}
}

// ----------------------------------------------------------------
void ubig_incr2(ubig * pc)
{
	int i;
	ubig_part_t carry = 2;
	ubig_part_t * pcparts = &pc->parts[0];

	for (i = 0; i < UBIG_NUM_PARTS; i++) {
		pcparts[i] = pcparts[i] + carry;
		if ((pcparts[i] == 0) && carry)
			carry = 1;
		else
			carry = 0;
	}
}

// ----------------------------------------------------------------
void ubig_decr(ubig * pc)
{
	int i, borrow;
	ubig_part_t * pcparts = &pc->parts[0];

	for (i = 0; i < UBIG_NUM_PARTS; i++) {
		borrow = (pcparts[i] == 0) ? 1 : 0;
		pcparts[i]--;
		if (!borrow)
			break;
	}
}

// ----------------------------------------------------------------
ubig ubig_from_unsigned(unsigned ival)
{
	int i;
	ubig rv;
	for (i = 0; i < UBIG_NUM_PARTS; i++)
		rv.parts[i] = 0;
	rv.parts[0] = ival;
#if UBIG_NUM_PARTS >= 2
	rv.parts[1] = ival >> 16;
#endif
	return rv;
}

// ----------------------------------------------------------------
int ubig_equals_unsigned(ubig * pa, unsigned ival)
{
	unsigned short val0 = ival & 0xffff;
	unsigned short val1 = ival >> 16;
	int i;

	for (i = 2; i < UBIG_NUM_PARTS; i++)
		if (pa->parts[i] != 0)
			return 0;
#if UBIG_NUM_PARTS >= 2
	if (val1 != pa->parts[1])
		return 0;
#endif
	if (val0 != pa->parts[0])
		return 0;
	return 1;
}
