#include <stdio.h>
#include "z2ilinalg.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
void z2ila_transpose(
	z2ipoly_t * Mout,
	z2ipoly_t * Min,
	int n)
{
	int row, col;

	for (row = 0; row < n; row++) {
		Mout[row] = 0;
	}
	for (row = 0; row < n; row++) {
		for (col = 0; col < n; col++)
			if (Min[row] & (1LL << col))
				Mout[col] |= 1LL << row;
	}
}

// ----------------------------------------------------------------
void z2ila_row_echelon_form(
	z2ipoly_t * M,
	int n)
{
	z2ila_row_reduce_below(M, n);
	z2ila_row_reduce_above(M, n);
}

// ----------------------------------------------------------------
static int find_row_with_leader_at_col(
	z2ipoly_t * M,
	int         n,
	int         clear_col,
	int         leader_row,
	int       * pswap_row)
{
	int row;
	z2ipoly_t clear_bit = 1LL << clear_col;
	for (row = leader_row; row < n; row++) {
		if (M[row] & clear_bit) {
			*pswap_row = row;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
void z2ila_row_reduce_below(
	z2ipoly_t * M,
	int         n)
{
	int clear_col, leader_row, swap_row, clear_row;
	z2ipoly_t temp;

	for (clear_col = n-1, leader_row = 0; clear_col >= 0; clear_col--) {
		if (!find_row_with_leader_at_col(M, n, clear_col, leader_row,
			&swap_row))
		{
			continue;
		}
		if (leader_row != swap_row) {
			temp          = M[swap_row];
			M[swap_row]   = M[leader_row];
			M[leader_row] = temp;
		}

		for (clear_row = leader_row + 1; clear_row < n; clear_row++) {
			if (M[clear_row] & (1LL << clear_col))
				M[clear_row] ^= M[leader_row];
		}
		leader_row++;
	}
}

// ----------------------------------------------------------------
// Assumes row_reduce_below has already been done.

void z2ila_row_reduce_above(
	z2ipoly_t * M,
	int n)
{
	int row, row2, row2_deg;
	z2ipoly_t bit;

	for (row = 0; row < n; row++) {
		if (M[row] == 0)
			break;

		for (row2 = row+1; row2 < n; row2++) {
			row2_deg = z2ipoly_get_degree(M[row2]);
			bit = 1LL << row2_deg;
			if (M[row] & bit)
				M[row] ^= M[row2];
		}
		//  11111111    10000010
		//  00111111    01100110
		//  00011010    00010100
		//  00001110    00001110
		//  00000100    00000100
		//  00000001    00000001
		//  00000000    00000000
		//  00000000    00000000

	}
}

// ----------------------------------------------------------------
int z2ila_find_rank(
	z2ipoly_t * M,
	int n)
{
	int rv = 0;
	int i;

	for (i = 0; i < n; i++)
		if (M[i])
			rv++;
	return rv;
}

// ----------------------------------------------------------------
// Assumes the matrix is already in row-echelon form.
void z2ila_find_nullspace_basis(
	z2ipoly_t * BIT,
	int         n,
	z2ipoly_t * nullspace_basis,
	int         dimker)
{
	int i, j, deg;
	unsigned char free_flags[Z2IPOLY_NUM_BITS];
	unsigned char free_indices[Z2IPOLY_NUM_BITS];
	int nfree = 0; // == dimker but I'll compute it anyway
	int rank = n - dimker;
	int dep_pos;
	z2ipoly_t free_bit, dep_bit;

	for (i = 0; i < n; i++) {
		free_flags[i] = 1;
	}
	for (i = 0; i < rank; i++) {
		deg = z2ipoly_get_degree(BIT[i]);
		free_flags[deg] = 0;
	}

	for (i = 0; i < n; i++) {
		if (free_flags[i])
			free_indices[nfree++] = i;
	}

	// For each free coefficient:
	//   Let that free coefficient be 1 and the rest be zero.
	//   Also set any dependent coefficients which depend on that
	//   free coefficient.
	for (i = 0; i < dimker; i++) {
		free_bit = 1LL << free_indices[i];
		nullspace_basis[i] = free_bit;

		for (j = 0; j < rank; j++) {
			if (BIT[j] & free_bit) {
				dep_pos = z2ipoly_get_degree(BIT[j]);
				dep_bit = 1LL << dep_pos;
				nullspace_basis[i] |= dep_bit;
			}
		}
	}
}

// ----------------------------------------------------------------
// The caller must free the return value.

// Example:  r = 0x13 = 10011

// Galois-configuration matrix:
//   [ 0 0 0 1 ]
//   [ 1 0 0 1 ]
//   [ 0 1 0 0 ]
//   [ 0 0 1 0 ]
// As z2ipolys:
//   0x01
//   0x09
//   0x04
//   0x02

// Fibonacci-configuration matrix:
// Matrix:
//   [ 0 0 1 1 ]
//   [ 1 0 0 0 ]
//   [ 0 1 0 0 ]
//   [ 0 0 1 0 ]
// As z2ipolys:
//   0x03
//   0x08
//   0x04
//   0x02

z2ipoly_t * z2ipoly_companion_matrix(
	z2ipoly_t r,
	int do_galois_cfg)
{
	int n = z2ipoly_get_degree(r);
	int i;
	z2ipoly_t * rv = malloc_check(n * sizeof(z2ipoly_t));

	for (i = 0; i < n; i++) {
		if (i > 0)
			rv[i] = 1LL << (n-i);
		else
			rv[i] = 0LL;
		if (do_galois_cfg) {
			if (r & (1LL << i))
				rv[i] |= 1LL;
		}
	}
	if (!do_galois_cfg) {
		rv[0] ^= r ^ (1LL << n);
	}

	return rv;
}

// ----------------------------------------------------------------
int z2ila_mx_times_vr(
	z2ipoly_t * M,
	int         n,
	z2ipoly_t   x)
{
	z2ipoly_t y = 0LL;
	z2ipoly_t bit = 0LL;
	int parity, i;

	bit = 1LL << (n-1);
	for (i = 0; i < n; i++) {
		parity = z2ipoly_count_bits(M[i] & x);
		if (parity & 1)
			y |= bit;
		bit >>= 1;
	}

	return y;
}

// ----------------------------------------------------------------
//     j             k             j
// i [ . . . . ] i [ . . . . ] k [ . . . . ]
//   [ . . . . ]   [ . . . . ]   [ . . . . ]
//   [ . . . . ] = [ . . . . ] * [ . . . . ]
//   [ . . . . ]   [ . . . . ]   [ . . . . ]

// xxx hack hack hack

void z2ila_sq_mx_times_sq_mx(
	z2ipoly_t * C, // output
	z2ipoly_t * A, // input 1
	z2ipoly_t * B, // input 2
	int         n)
{
	// Allows C to safely point to A or B:
	z2ipoly_t * prod = malloc_check(n * sizeof(z2ipoly_t));
	int i, j, k;
	int ri, rj, rk;

	for (i = 0; i < n; i++)
		prod[i] = 0LL;

	for (i = 0, ri = n-1; i < n; i++, ri--) {
		for (j = 0, rj = n-1; j < n; j++, rj--) {
			for (k = 0, rk = n-1; k < n; k++, rk--) {
				//printf("[%lld%lld%lld]",
					//(A[i] >> k) & 1,
					//(B[k] >> j) & 1,
					//(A[i] >> k) & (B[k] >> j) & 1);
				prod[i] ^=
				((A[i] >> rk) & (B[k] >> rj) & 1) << rj;
			}
		}
	}

// 0 0 1 1    0 0 1 1
// 1 0 0 0    1 0 0 0
// 0 1 0 0    0 1 0 0
// 0 0 1 0    0 0 1 0

	for (i = 0; i < n; i++)
		C[i] = prod[i];
	free(prod);
}
