#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipoly.h"
#include "factorize.h"
#include "log2.h"
#include "count_bits.h"
#include "malloc_check.h"

static void complain_about_overflow(int deg);

// ----------------------------------------------------------------
z2ipoly_t z2ipoly_add(
	z2ipoly_t a,
	z2ipoly_t b)
{
	return a ^ b;
}

// ----------------------------------------------------------------
z2ipoly_t z2ipoly_sub(
	z2ipoly_t a,
	z2ipoly_t b)
{
	return a ^ b;
}

// ----------------------------------------------------------------
z2ipoly_t z2ipoly_mul(z2ipoly_t a, z2ipoly_t b)
{
	int adeg, bdeg, cdeg;
	z2ipoly_t ashift = a;
	z2ipoly_t rv = 0;
	int j;

	adeg = z2ipoly_get_degree(a);
	bdeg = z2ipoly_get_degree(b);
	cdeg = adeg + bdeg;

	if (cdeg > Z2IPOLY_MAX_DEGREE)
		complain_about_overflow(cdeg);

	for (j = 0; j <= bdeg; j++) {
		if (Z2IPOLY_BIT_AT(b, j))
			rv ^= ashift;
		ashift <<= 1;
	}
	return rv;
}

// ----------------------------------------------------------------
z2ipoly_t z2ipoly_div(z2ipoly_t a, z2ipoly_t b)
{
	z2ipoly_t quot, rem;
	z2ipoly_quot_and_rem(a, b, &quot, &rem);
	return quot;
}

// ----------------------------------------------------------------
z2ipoly_t z2ipoly_mod(z2ipoly_t a, z2ipoly_t b)
{
	z2ipoly_t quot, rem;
	z2ipoly_quot_and_rem(a, b, &quot, &rem);
	return rem;
}

// ----------------------------------------------------------------
void z2ipoly_quot_and_rem(
	z2ipoly_t   dividend,
	z2ipoly_t   divisor,
	z2ipoly_t * pquot,
	z2ipoly_t * prem)
{
	int dividend_l1_pos;
	int divisor_l1_pos;
	int l1_diff;
	z2ipoly_t shift_divisor; // Make a copy so we don't modify argument
	int check_pos;
	int quot_pos;

	if (divisor == 0) {
		// Divisor is zero.
		fprintf(stderr, "z2ipoly_quot_and_rem:  Divide by zero.\n");
		exit(1);
	}
	divisor_l1_pos = z2ipoly_get_degree(divisor);

	if (dividend == 0) {
		*pquot = 0;
		*prem  = 0;
		return;
	}
	dividend_l1_pos = z2ipoly_get_degree(dividend);

	l1_diff = dividend_l1_pos - divisor_l1_pos;

	if (l1_diff < 0) {
		// Dividend has lower degree than divisor.
		*pquot = 0;
		*prem = dividend;
		return;
	}

	shift_divisor = divisor << l1_diff;

	*prem = dividend;
	*pquot = 0;
	for (
		check_pos = dividend_l1_pos,
			quot_pos = l1_diff;
		check_pos >= divisor_l1_pos;
		check_pos--, quot_pos--)
	{
		if (Z2IPOLY_BIT_AT(*prem, check_pos)) {
			*prem ^= shift_divisor;
			Z2IPOLY_SET_BIT(*pquot, quot_pos);
		}
		shift_divisor >>= 1;
	}
}

// ----------------------------------------------------------------
z2ipoly_t z2ipoly_gcd(z2ipoly_t a, z2ipoly_t b)
{
	z2ipoly_t c, q, r, d;

	if (a == 0)
		return b;
	if (b == 0)
		return a;

	// Initialize
	c = a;
	d = b;

	while (1) {
		// Divide
		// q = c / d, r = c % d
		z2ipoly_quot_and_rem(c, d, &q, &r);
		// Note:  now c = qd + r and 0 <= r < d
		// Remainder zero?
		if (r == 0) {
			break;
		}
		// Recycle
		c = d;
		d = r;
	}
	return d;
}

// ----------------------------------------------------------------
// Input: a, b
// Output: GCD g, and m, n such that am + bn = d

// xxx describe Blankinship alg'm here.

z2ipoly_t z2ipoly_ext_gcd(
	z2ipoly_t   a,
	z2ipoly_t   b,
	z2ipoly_t * pm,
	z2ipoly_t * pn)
{
	z2ipoly_t m, n, mprime, nprime, c, q, r, t, d;

	// Initialize
	mprime = 1;
	n = 1;
	m = 0;
	nprime = 0;
	c = a;
	d = b;

	while (1) {

		// Divide
		// q = c / d, r = c % d
		z2ipoly_quot_and_rem(c, d, &q, &r);
		// Note:  now c = qd + r and 0 <= r < d

		// Remainder zero?
		if (r == 0) {
			break;
		}

		// Recycle
		c = d;
		d = r;

		t = mprime;
		mprime = m;
		m = t ^ z2ipoly_mul(q, m); // m = t - qm;

		t = nprime;
		nprime = n;
		n = t ^ z2ipoly_mul(q, n); // n = t - qn;
	}
	*pm = m;
	*pn = n;
	return d;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
z2ipoly_t z2ipoly_power(
	z2ipoly_t a,
	unsigned  power)
{
	z2ipoly_t a2 = a;
	z2ipoly_t apower = 1LL;

	while (power != 0) {
		if (power & 1)
			apower = z2ipoly_mul(a2, apower);
		power >>= 1;
		a2 = z2ipoly_mul(a2, a2);
	}
	return apower;
}

// ----------------------------------------------------------------
// Assumes the input is already known to be a square.  Relies on the
// fact that f(x^p) = f^p(x) over Fp[x].
//
// in  = a4 x^4 + a2 x^2 + a0
// out = a4 x^2 + a2 x   + a0

z2ipoly_t z2ipoly_square_root(
	z2ipoly_t in)
{
	int si, di;
	z2ipoly_t out = 0LL;
	z2ipoly_t inbit = 1LL, outbit = 1LL;

	for (si = 0, di = 0; si < Z2IPOLY_NUM_BITS; si+=2, di++) {
		if (in & inbit)
			out |= outbit;
		inbit <<= 1;
		if (in & inbit) {
			printf("z2ipoly_square_root: %llx is not square.\n",
				in);
			exit(1);
		}
		inbit <<= 1;
		outbit <<= 1;
	}

	return out;
}

// ----------------------------------------------------------------
void z2ipoly_print(
	z2ipoly_t a)
{
	int width = (z2ipoly_get_degree(a) + 3) >> 2;
	printf("%0*llx", width, a);
}

// ----------------------------------------------------------------
void z2ipoly_dprint(
	z2ipoly_t a,
	int       min_deg)
{
	int width = (min_deg >> 2) + 1;
	printf("%0*llx", width, a);
}

// ----------------------------------------------------------------
void z2ipoly_binary_print(
	z2ipoly_t a)
{
	z2ipoly_binary_width_print(a, z2ipoly_get_degree(a));
}

// ----------------------------------------------------------------
void z2ipoly_binary_width_print(
	z2ipoly_t a,
	int width)
{
	int degree = z2ipoly_get_degree(a);
	int i;
	if (width < degree)
		width = degree;
	for (i = width; i >= 0; i--)
		printf("%c", ((unsigned) (a >> i) & 1) + '0');
}

// ----------------------------------------------------------------
void z2ipoly_pretty_print(
	z2ipoly_t a)
{
	int d = z2ipoly_get_degree(a);
	int i;
	int got_first = 0;

	if (a == 0) {
		printf("0");
		return;
	}

	for (i = d; i >= 0; i--) {
		if (Z2IPOLY_BIT_AT(a, i)) {
			if (got_first)
				printf("+");
			if (i == 1)
				printf("x");
			else if (i == 0)
				printf("1");
			else
				printf("x^%d", i);
			got_first = 1;
		}
	}
}

// ----------------------------------------------------------------
int z2ipoly_scan(
	z2ipoly_t * pz,
	char    * pin)
{
	return sscanf(pin, "%llx", pz);
}

// ----------------------------------------------------------------
int z2ipoly_eval(z2ipoly_t r, int x)
{
	if (x & 1)
		return z2ipoly_count_bits(r);
	else
		return r & 1;
}

// ----------------------------------------------------------------
z2ipoly_t z2ipoly_compose(
	z2ipoly_t r,
	z2ipoly_t z)
{
	z2ipoly_t rz = 0LL;
	z2ipoly_t zp = 1LL;
	z2ipoly_t rsh;
	int n = z2ipoly_get_degree(r);
	int d;

	for (d = 0, rsh = r; d <= n; d++, rsh >>= 1) {
		if (rsh & 1LL)
			rz ^= zp;
		zp = z2ipoly_mul(zp, z);
	}

	return rz;
}

// ----------------------------------------------------------------
z2ipoly_t z2ipoly_deriv(
	z2ipoly_t z)
{
	return (z >> 1) & 0x5555555555555555LL;
}

// ----------------------------------------------------------------
int z2ipoly_is_rootless(z2ipoly_t r)
{
	if (z2ipoly_eval(r, 0) == 0)
		return 0;
	if (z2ipoly_eval(r, 1) == 0)
		return 0;
	return 1;
}

// ----------------------------------------------------------------
int z2ipoly_count_bits(z2ipoly_t z)
{
	return 1 & count_one_bits((unsigned char *)&z, sizeof(z));
}

// ----------------------------------------------------------------
int z2ipoly_get_degree(z2ipoly_t z)
{
	return find_msb_64(z);
}

// ----------------------------------------------------------------
void z2ipoly_check_degree(
	int degree)
{
	if (degree < 0) {
		fprintf(stderr,
			"z2ipoly:  degree must be non-negative; got %d.\n",
			degree);
		exit(1);
	}
	if (degree > Z2IPOLY_MAX_DEGREE) {
		fprintf(stderr,
			"z2ipoly:  degree must be <= %d; got %d.\n",
			Z2IPOLY_MAX_DEGREE, degree);
		exit(1);
	}
}

// ----------------------------------------------------------------
static void complain_about_overflow(int degree)
{
	fprintf(stderr, "z2ipoly_t overflow:  degree %d, max %d.\n",
		degree, Z2IPOLY_MAX_DEGREE);
	exit(1);
}

// ----------------------------------------------------------------
int z2ipoly_qcmp(const void * pv1, const void * pv2)
{
	const z2ipoly_t * p1 = pv1;
	const z2ipoly_t * p2 = pv2;
	if (*p1 < *p2)
		return -1;
	if (*p1 > *p2)
		return 1;
	return 0;
}
