#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "malloc_check.h"
#include "isqrt.h"

static int z2ipolymod_element_order_irr(z2ipoly_t z, z2ipoly_t r);
static int z2ipolymod_element_order_red(z2ipoly_t z, z2ipoly_t r);

// ----------------------------------------------------------------
// z2ipolymod_mul
//
// Multiplication may be done (at least) two ways:  (1) Take the ring product
// of the two polynomials, then reduce mod r; (2) reduce mod r at every step.
// Testing has shown that the latter method (reduce en route) is faster than
// the former (ring and reduce).  Furthermore, the ring-and-reduce method
// requires storage space for degree up to 2d-2, where d is the degree of the
// reductor.  E.g. with 64-bit arithmetic, we can't multiply 40-bit polynomials
// using ring-and-reduce, but we can if we use reduce-en-route.
//
// The ring-and-reduce method would be this:
//
//     return z2ipoly_mod(z2ipoly_mul(a, b), r);

z2ipoly_t z2ipolymod_mul(z2ipoly_t a, z2ipoly_t b, z2ipoly_t r)
{
	z2ipoly_t temp, out;
	int i, power, degb, degr;
	z2ipoly_t bit_at_deg;

	a = z2ipoly_mod(a, r);
	b = z2ipoly_mod(b, r);
	degb = z2ipoly_get_degree(b);
	degr = z2ipoly_get_degree(r);
	bit_at_deg = (z2ipoly_t)1 << degr;

	out = 0;

	for (i = 0; i <= degb; i++) {
		// Multiply a by scalar coefficient on current term of b.
		// For Z2, this is either a 0 or 1.  0 means this partial
		// product is zero.  1 means we need to multiply by x^k.
		if (!Z2IPOLY_BIT_AT(b, i))
			continue;
		temp = a;

		// Now need to multiply by the power of x on the current term
		// of b, reducing mod r en route.
		for (power = 1; power <= i; power++) {
			temp <<= 1;
			if (temp & bit_at_deg)
				temp ^= r;
		}
		// Add in this partial product.
		out ^= temp;
	}

	return out;
}

// ----------------------------------------------------------------
z2ipoly_t z2ipolymod_div(z2ipoly_t a, z2ipoly_t b, z2ipoly_t r)
{
	z2ipoly_t binv;
	binv = z2ipolymod_recip(b, r);
	return z2ipolymod_mul(a, binv, r);
}

// ----------------------------------------------------------------
// 1/b mod r.
//
// Finite-field division uses extended Euclidean GCD, with ring division
// (quotient and remainder) as a substep.
//
// a / b is a * 1/b.  To compute 1/b, find the GCD d.  (Since the reductor
// is irreducible -- although we don't test this on every division -- and
// if b is monic, the GCD had better be 1).  Then, compute 1 = m*b + n*r.
// Since r = 0 mod r, 1 = m*b and therefore m = 1/b.  The extended Euclidean
// algorithm (see Knuth vol. 1) computes d, m and n all at once.
//
// An alternate reciprocation algorithm uses the fact that a finite field
// GF(p^m) has its multiplicative group of order p^m-1.  So, by Euler's
// criterion, x^(p^n-2) = x^(-1) for all non-zero elements of the field.
//
// Testing has shown the extended-GCD reciprocation is about 2x faster.

z2ipoly_t z2ipolymod_recip(z2ipoly_t b, z2ipoly_t r)
{
#if 1
	z2ipoly_t m, n, g;
	if (b == 0LL) {
		fprintf(stderr, "z2ipolymod_recip:  divide by zero.\n");
		exit(1);
	}
	g = z2ipoly_ext_gcd(b, r, &m, &n);
	if (g != 1) {
		fprintf(stderr,
			"z2ipolymod_recip:  %llx is reducible "
			"(GCD(%llx, %llx) = %llx, != 1).\n",
			r, b, r, g);
		exit(1);
	}
	return m;
#else
	unsigned long long one = 1;
	unsigned long long two = 2;
	unsigned long long pm2 = (one << z2ipoly_get_degree(r)) - two;
	return z2ipolymod_power64(b, pm2, r);
#endif
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
z2ipoly_t z2ipolymod_power(
	z2ipoly_t  z,
	unsigned   power,
	z2ipoly_t  r)
{
	z2ipoly_t z2 = z;
	z2ipoly_t rv = 1;

	while (power != 0) {
		if (power & 1)
			rv = z2ipolymod_mul(rv, z2, r);
		power >>= 1;
		z2 = z2ipolymod_mul(z2, z2, r);
	}
	return rv;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
z2ipoly_t z2ipolymod_power64(
	z2ipoly_t  z,
	unsigned long long power64,
	z2ipoly_t  r)
{
	z2ipoly_t z2 = z;
	z2ipoly_t rv = 1;
	unsigned long long zero = 0;
	unsigned long long one  = 1;

	while (power64 != zero) {
		if (power64 & one)
			rv = z2ipolymod_mul(rv, z2, r);
		power64 >>= 1;
		z2 = z2ipolymod_mul(z2, z2, r);
	}
	return rv;
}

// ----------------------------------------------------------------
z2ipoly_t z2ipolymod_compose(
	z2ipoly_t r,
	z2ipoly_t z,
	z2ipoly_t m)
{
	z2ipoly_t rz = 0LL;
	z2ipoly_t zp = 1LL;
	z2ipoly_t rsh;
	int n = z2ipoly_get_degree(r);
	int d;

	for (d = 0, rsh = r; d <= n; d++, rsh >>= 1) {
		if (rsh & 1LL)
			rz ^= zp;
		zp = z2ipolymod_mul(zp, z, m);
	}

	return rz;
}


// ----------------------------------------------------------------
int z2ipolymod_element_order(z2ipoly_t z, z2ipoly_t r)
{
	int is_irr, rv;
	// Cache:
	static z2ipoly_t last_r = 0x13;
	static int last_i = 1;

	if (z == 0LL)
		return 0;
	// x mod x, the latter being irreducible.
	if ((z == 2LL) && (r == 2LL))
		return 0;

	if (r == last_r) {
		is_irr = last_i;
	}
	else {
		is_irr = z2ipoly_is_irreducible(r);
		last_i = is_irr;
		last_r = r;
	}

	if (is_irr)
		rv = z2ipolymod_element_order_irr(z, r);
	else
		rv = z2ipolymod_element_order_red(z, r);

	return rv;
}

static int z2ipolymod_element_order_irr(z2ipoly_t z, z2ipoly_t r)
{
	int m = z2ipoly_get_degree(r);
	unsigned long long pm = (unsigned long long)1 << m;
	unsigned long long pm1 = pm - 1;
	unsigned nd, k;
	unsigned long long * all_divisors = 0;
	unsigned long long f = 0;
	z2ipoly_t zpower;
	int got_it = 0;

	all_divisors = get_all_divisors_ull(pm1, &nd);

	for (k = 0; k < nd; k++) {
		f = all_divisors[k];
		zpower = z2ipolymod_power64(z, f, r);
		if (zpower == 1) {
			got_it = 1;
			break;
		}
	}

	free(all_divisors);
	if (got_it) {
		return f;
	}
	else {
		fprintf(stderr, "z2ipoly_t:  Internal error:  order of "
			"%llx mod %llx not found.\n", z, r);
		exit(1);
		return 0; // Not reached.
	}
}

// When r is reducible, nonetheless the invertible elements (units) of the
// polymod ring do form a multiplicative group.  An efficient order calculator
// would find the order of that multiplicative group, then try raising z only
// to powers dividing that order.  This function, though, is a brute-force
// search.
//
// How does one compute a sort of totient function for Fp[x] without factoring
// r?  Or, maybe it's worthwhile to go ahead and factor r.

static int z2ipolymod_element_order_red(z2ipoly_t z, z2ipoly_t r)
{
	int rv;
	int max = 1LL << z2ipoly_get_degree(r);
	z2ipoly_t zpower = z;

	for (rv = 1; rv < max; rv++) {
		if (zpower == 1LL)
			return rv;
		zpower = z2ipolymod_mul(zpower, z, r);
	}

	fprintf(stderr, "z2ipoly_t:  Internal error:  order of "
		"%llx mod %llx not found.\n", z, r);
	exit(1);
	return 0; // Not reached.
}

// ----------------------------------------------------------------
z2ipoly_t z2ipolymod_find_generator(
	z2ipoly_t r)
{
	int n = z2ipoly_get_degree(r);
	int pn1 = (1LL << n) - 1;
	z2ipoly_t glo = 2LL;
	z2ipoly_t ghi = pn1;
	z2ipoly_t g = 0LL;

	if (ghi < glo)
		ghi = glo; // linear fields

	for (g = glo; g <= ghi; g++)
		if (z2ipolymod_element_order(g, r) == pn1)
			return g;

	printf("Generator for %llx not found.\n", r);
	exit(1);
	return 0; // not reached
}

// ----------------------------------------------------------------
// Find e such that g^e == z mod r

int z2ipolymod_naive_log(
	z2ipoly_t z, // Input
	z2ipoly_t g, // Base, assumed (!) primitive mod r.
	z2ipoly_t r) // Modulus for FF arithmetic
{
	int log;
	z2ipoly_t gp = 1LL;
	int deg = z2ipoly_get_degree(r);
	int pm = (1 << deg);

	if (z == 0LL)
		return 0;

	for (log = 0; log <= pm; log++) {
		if (z == gp)
			return log;
		gp = z2ipolymod_mul(gp, g, r);
	}

	printf("z2ipoly_log: %llx is not primitive mod %llx\n", g, r);
	exit(1);
}

// ----------------------------------------------------------------
// This is Shanks' baby-steps-giant-steps technique, as described in Odlyzko's
// 1999 survey paper "Discrete logarithms:  The past and the future".
//
// h = g^(i + jm), where m = ceil sqrt(n), n is the group size, and
// 0 <= i, j <= m - 1.  Compute hg^(-i) and g^(jm); look for a match.
//
// This cuts the naive search (which is O(n)) down to a search of O(sqrt(n)).

typedef struct _z2idx_t {
	z2ipoly_t z;
	int       idx;
} z2idx_t;

static int z2idx_qcmp(const void * pv1, const void * pv2)
{
	const z2idx_t * p1 = pv1;
	const z2idx_t * p2 = pv2;
	if (p1->z < p2->z)
		return -1;
	if (p1->z > p2->z)
		return 1;
	return 0;
}

int z2ipolymod_shanks_log(
	z2ipoly_t h, // Input
	z2ipoly_t g, // Base, assumed (!) primitive mod r.
	z2ipoly_t r) // Modulus for FF arithmetic
{
	int log = -1;
	unsigned long long n = 1LL << z2ipoly_get_degree(r);
	unsigned m = (unsigned)calc_isqrt_64(n, ICEIL);
	z2idx_t * hgni = malloc_check(m * sizeof(z2idx_t));
	z2idx_t * gmj  = malloc_check(m * sizeof(z2idx_t));
	z2ipoly_t ginv = z2ipolymod_recip(g, r);
	z2ipoly_t gm   = z2ipolymod_power(g, m, r);
	int i, j;

	if (h == 0LL) {
		free(hgni);
		free(gmj);
		return 0;
	}

	hgni[0].z = h;
	hgni[0].idx = 0;
	gmj [0].z = 1LL;
	gmj [0].idx = 0;

	for (i = 1; i < m; i++) {
		hgni[i].z = z2ipolymod_mul(hgni[i-1].z, ginv, r);
		hgni[i].idx = i;
	}
	for (j = 1; j < m; j++) {
		gmj[j].z = z2ipolymod_mul(gmj[j-1].z, gm, r);
		gmj[j].idx = j;
	}

	qsort(hgni, m, sizeof(z2idx_t), z2idx_qcmp);
	qsort(gmj,  m, sizeof(z2idx_t), z2idx_qcmp);

	for (i = 0, j = 0; i < m && j < m; ) {
		if (hgni[i].z == gmj[j].z) {
			// xxx could overflow 32-bit int; fix me.
			log = hgni[i].idx + gmj[j].idx * m;
			log %= (n - 1);
			break;
		}
		else if (hgni[i].z < gmj[j].z) {
			i++;
		}
		else {
			j++;
		}
	}

	if (log == -1) {
		printf("z2ipolymod_shanks_log: %llx is not primitive mod %llx\n", g, r);
		exit(1);
	}

	free(hgni);
	free(gmj);

	return log;
}

// ----------------------------------------------------------------
int z2ipolymod_log(
	z2ipoly_t z, // Input
	z2ipoly_t g, // Base, assumed (!) primitive mod r.
	z2ipoly_t r) // Modulus for FF arithmetic
{
	//return z2ipolymod_naive_log(z, g, r);
	return z2ipolymod_shanks_log(z, g, r);
}

// ----------------------------------------------------------------
z2ipoly_t z2ipolymod_add(z2ipoly_t a, z2ipoly_t b, z2ipoly_t r)
{
	return z2ipoly_add(a, b);
}

// ----------------------------------------------------------------
z2ipoly_t z2ipolymod_sub(z2ipoly_t a, z2ipoly_t b, z2ipoly_t r)
{
	return z2ipoly_sub(a, b);
}

// ----------------------------------------------------------------
// Sum of conjugates.

z2ipoly_t z2ipolymod_trace(
	z2ipoly_t a,
	z2ipoly_t r)
{
	z2ipoly_t rv = 0LL;
	z2ipoly_t ap = a;
	int i;
	int n = z2ipoly_get_degree(r);

	for (i = 0; i < n; i++) {
		rv ^= ap;
		ap = z2ipolymod_power(ap, 2, r);
		if (ap == a)
			break;
	}
	return rv;
}

// ----------------------------------------------------------------
// Product of conjugates.

z2ipoly_t z2ipolymod_norm(
	z2ipoly_t a,
	z2ipoly_t r)
{
	z2ipoly_t rv = 1LL;
	z2ipoly_t ap = a;
	int i;
	int n = z2ipoly_get_degree(r);

	for (i = 0; i < n; i++) {
		rv = z2ipolymod_mul(rv, ap, r);
		ap = z2ipolymod_power(ap, 2, r);
		if (ap == a)
			break;
	}
	return rv;
}

// ----------------------------------------------------------------
void z2ipolymod_print(
	z2ipoly_t a,
	z2ipoly_t r)
{
	int width = (z2ipoly_get_degree(r) + 3) >> 2;
	printf("%0*llx", width, a);
}

// ----------------------------------------------------------------
int z2ipolymod_auxscan(
	z2ipoly_t * pr,
	char    * pin)
{
	z2ipoly_t r;

	// If it doesn't start with ":", parse it as a user-specified reductor.
	// Else, parse as int m and search for a reductor of that degree.
	if (pin[0] != ':') {
		if (!z2ipoly_scan(&r, pin)) {
			fprintf(stderr, "Couldn't scan \"%s\" as hex.\n",
				pin);
			exit(1);
		}
		if (r < 2) {
			fprintf(stderr, "r must be >= 2; got %llx.\n", r);
			exit(1);
		}
	}
	else {
		int degree;
		if (sscanf(pin, ":%d", &degree) != 1) {
			fprintf(stderr,
				"z2ipoly_t:  Couldn't scan \"%s\" as degree.\n",
				pin);
			return 0;
		}
		if (degree > Z2IPOLY_MAX_DEGREE) {
			fprintf(stderr,
				"z2ipoly_t:  degree must be <= %d; got %d.\n",
				Z2IPOLY_MAX_DEGREE, degree);
			return 0;
		}
		if (degree < 1) {
			fprintf(stderr,
				"z2ipoly_t:  degree must be >= 1; got %d.\n",
				degree);
			return 0;
		}
		r = z2ipoly_find_irreducible(degree);
	}

	*pr = r;
	return 1;
}

// ----------------------------------------------------------------
z2ipoly_t z2ipoly_find_irreducible(int m)
{
	z2ipoly_t i, pm, tpm;

	// xxx need to sanity-check m.
	pm = (z2ipoly_t)1 << m;
	tpm = pm << 1;
	for (i = pm; i < tpm; i++) {
		if (z2ipoly_is_irreducible(i))
			if (z2ipoly_is_primitive(i))
				return i;
	}
	fprintf(stderr,
	"z2ipoly_t: coding error: couldn't find any irreds in GF(%d^%d).\n",
		2, m);
	exit(1);
	return 0;
}

// ----------------------------------------------------------------
// X^(p^n) - X factors into all irreducibles of degree d for all d dividing
// n.  Therefore, if r (the input, of degree n) has an irreducible factor
// of degree < n, that factor will divide X^(p^d) - X for some d < n.
// If r has GCD = 1 with X^(p^d) - X for all d < n, then r is irreducible.
// Reduction mod r is taken at each step to avoid overflow, with no effect
// on the GCD.

int z2ipoly_is_irreducible(z2ipoly_t r)
{
	z2ipoly_t x = 2;
	z2ipoly_t u = 2;
	z2ipoly_t ux;
	z2ipoly_t g;
	int  d = z2ipoly_get_degree(r);
	int i;

	for (i = 1; i < d; i++) {
		// u = u**2 mod r; test GCD(u - x, r)
		u = z2ipolymod_mul(u, u, r);
		ux = u^x;
		g = z2ipoly_gcd(r, ux);

		if (g != 1)
			return 0;
	}
	return 1;
}

// ----------------------------------------------------------------
// Assumes r is known to be irreducible.
// See if the order of x mod r is p^m-1.  Certainly, x^(p^m-1) will be 1.  But
// check x to all powers dividing p^m-1 to make sure they are *not* 1.

int z2ipoly_is_primitive(z2ipoly_t r)
{
	int m = z2ipoly_get_degree(r);
	unsigned long long pm = 1LL << m;
	unsigned long long pm1 = pm - 1;
	factor_info_ull_t finfo;

	calc_factors_ull(pm1, &finfo);
	return z2ipoly_is_primitive_f(r, pm1, &finfo);
}

// ----------------------------------------------------------------
int z2ipoly_is_primitive_f(
	z2ipoly_t r,
	unsigned long long pm1,
	factor_info_ull_t * pfinfo)
{
	z2ipoly_t x = 2;
	z2ipoly_t xpower;
	unsigned nf, f, k;

	if (r == 2LL)
		return 0;

	nf = num_divisors_ull(pfinfo);

	for (k = 0; k < nf; k++) {
		f = kth_divisor_ull(pfinfo, k);
		xpower = z2ipolymod_power(x, f, r);
		if ((f != pm1) && (xpower == 1)) {
			return 0;
		}
		if ((f == pm1) && (xpower != 1)) {
			fprintf(stderr, "z2ipoly_is_primitive_f abend!\n");
			fprintf(stderr, "r=%llx nf=%d f=%u pm1=%llu\n",
				r, nf, f, pm1);
			exit(1);
		}
	}
	return 1;
}

// ----------------------------------------------------------------
void z2ipoly_find_dirreds(
	unsigned    n,
	z2ipoly_t * pirred_factors,
	unsigned  * pnum_irred_factors,
	unsigned    max_irred_factors)
{
	unsigned num_irred_factors = 0;
	unsigned d;
	z2ipoly_t lo, hi, f;

	for (d = 1; d <= n; d++) {
		if ((n % d) != 0)
			continue;

		lo = 1LL << d;
		hi = lo << 1;
		for (f = lo; f < hi; f++) {
			if (!z2ipoly_is_irreducible(f))
				continue;

			if (num_irred_factors >= max_irred_factors) {
				fprintf(stderr, "Too many factors.\n");
				exit(1);
			}
			pirred_factors[num_irred_factors] = f;
			num_irred_factors++;

		}
	}

	*pnum_irred_factors = num_irred_factors;
}

// ----------------------------------------------------------------
z2ipoly_t * z2ipolymod_get_elements(
	int n,
	int * pnel)
{
	int nel = 1 << n;
	z2ipoly_t * rv = malloc_check(nel * sizeof(z2ipoly_t));
	z2ipoly_t z;

	for (z = 0; z < nel; z++)
		rv[z] = z;

	*pnel = nel;
	return rv;
}

// ================================================================
static void z2ipolymod_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvred)
{
	z2ipoly_t * pin1 = (z2ipoly_t *)pvin1;
	z2ipoly_t * pin2 = (z2ipoly_t *)pvin2;
	z2ipoly_t * pout = (z2ipoly_t *)pvout;
	*pout = *pin1 ^ *pin2;
}

// ----------------------------------------------------------------
static void z2ipolymod_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvred)
{
	z2ipoly_t * pin1 = (z2ipoly_t *)pvin1;
	z2ipoly_t * pin2 = (z2ipoly_t *)pvin2;
	z2ipoly_t * pout = (z2ipoly_t *)pvout;
	z2ipoly_t * pred = (z2ipoly_t *)pvred;
	*pout = z2ipolymod_mul(*pin1, *pin2, *pred);
}

// ----------------------------------------------------------------
static int z2ipolymod_vequal(
	void * pin1,
	void * pin2,
	void * pred)
{
	return (*(z2ipoly_t *)pin1 == *(z2ipoly_t *)pin2);
}

// ----------------------------------------------------------------
static int z2ipolymod_vless(
	void * pin1,
	void * pin2,
	void * pred)
{
	return (*(z2ipoly_t *)pin1 < *(z2ipoly_t *)pin2);
}

// ----------------------------------------------------------------
static void z2ipolymod_vprint(
	void * pin,
	void * pred)
{
	z2ipoly_t a = *(z2ipoly_t *)pin;
	z2ipoly_t r = *(z2ipoly_t *)pred;
	z2ipolymod_print(a, r);
}

// ----------------------------------------------------------------
static void z2ipolymod_vauxprint(
	void * pred)
{
	z2ipoly_t r = *(z2ipoly_t *)pred;
	z2ipoly_print(r);
}

// ----------------------------------------------------------------
static int z2ipolymod_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	z2ipoly_t * pout = pvout;
	if (sizeof(z2ipoly_t) > max_elt_bytes) {
		fprintf(stderr, "z2ipoly_t:  Need %d bytes for element.\n",
			sizeof(z2ipoly_t));
		exit(1);
	}
	return z2ipoly_scan(pout, pin);
}

// ----------------------------------------------------------------
static int z2ipolymod_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	if (max_aux_bytes < sizeof(z2ipoly_t)) {
		fprintf(stderr, "dn:  Need %d bytes for aux.\n",
			sizeof(z2ipoly_t));
		exit(1);
	}

	if (argc != 1)
		return 0;

	return z2ipolymod_auxscan(pvaux, argv[0]);
}

// ----------------------------------------------------------------
int  z2ipolymodadd_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	z2ipoly_t * pelements = (z2ipoly_t *) pvelements;
	z2ipoly_t r = *(z2ipoly_t *)pvaux;
	int d = z2ipoly_get_degree(r);
	z2ipoly_t a;
	z2ipoly_t hi = 1LL << d;
	int group_size = hi;

	if (group_size > max_group_bytes/sizeof(z2ipoly_t))
		return 0;
	for (a = 0LL; a < hi; a++)
		pelements[a] = a;
	*pnum_elements = group_size;
	return 1;
}

// ----------------------------------------------------------------
int  z2ipolymodmul_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	z2ipoly_t * pelements = (z2ipoly_t *) pvelements;
	z2ipoly_t r = *(z2ipoly_t *)pvaux;
	int d = z2ipoly_get_degree(r);
	z2ipoly_t a;
	z2ipoly_t hi = 1LL << d;
	int group_size = 0;

	for (a = 1LL; a < hi; a++) {
		if (z2ipoly_gcd(a, r) != 1)
			continue;

		if (group_size > max_group_bytes/sizeof(z2ipoly_t))
			return 0;
		pelements[group_size++] = a;
	}
	*pnum_elements = group_size;
	return 1;
}

// ================================================================
group_spec_t z2ipolymodadd_group_spec = {
	element_size  : sizeof(z2ipoly_t),
	naux          : 1,
	pvop          : z2ipolymod_vadd,
	pvequal       : z2ipolymod_vequal,
	pvless        : z2ipolymod_vless,
	pvprint       : z2ipolymod_vprint,
	pvauxprint    : z2ipolymod_vauxprint,
	pvscan        : z2ipolymod_vscan,
	pvauxscan     : z2ipolymod_vauxscan,
	pfill_group   : z2ipolymodadd_fill_group,
	desc          : "z2ipolymodadd"
};

group_spec_t z2ipolymodmul_group_spec = {
	element_size  : sizeof(z2ipoly_t),
	naux          : 1,
	pvop          : z2ipolymod_vmul,
	pvequal       : z2ipolymod_vequal,
	pvless        : z2ipolymod_vless,
	pvprint       : z2ipolymod_vprint,
	pvauxprint    : z2ipolymod_vauxprint,
	pvscan        : z2ipolymod_vscan,
	pvauxscan     : z2ipolymod_vauxscan,
	pfill_group   : z2ipolymodmul_fill_group,
	desc          : "z2ipolymodmul"
};
