#include "z2linalg.h"

// ----------------------------------------------------------------
void z2la_transpose(
	z2poly_t * Mout,
	z2poly_t * Min,
	int n)
{
	int row, col;

	for (row = 0; row < n; row++) {
		Mout[row] = z2poly_from_base_rep(0);
	}
	for (row = 0; row < n; row++) {
		for (col = 0; col < n; col++)
			if (z2poly_bit_at(&Min[row], col))
				z2poly_set_bit(&Mout[col], row);
	}
}

// ----------------------------------------------------------------
void z2la_row_echelon_form(
	z2poly_t * M,
	int n)
{
	z2la_row_reduce_below(M, n);
	z2la_row_reduce_above(M, n);
}

// ----------------------------------------------------------------
static int find_row_with_leader_at_col(
	z2poly_t * M,
	int        n,
	int        clear_col,
	int        leader_row,
	int      * pswap_row)
{
	int row;
	for (row = leader_row; row < n; row++) {
		if (z2poly_bit_at(&M[row], clear_col)) {
			*pswap_row = row;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
void z2la_row_reduce_below(
	z2poly_t * M,
	int        n)
{
	int clear_col, leader_row, swap_row, clear_row;
	z2poly_t temp;

	for (clear_col = n-1, leader_row = 0; clear_col >= 0; clear_col--) {
		if (!find_row_with_leader_at_col(M, n, clear_col,
			leader_row, &swap_row))
		{
			continue;
		}
		if (leader_row != swap_row) {
			temp          = M[swap_row];
			M[swap_row]   = M[leader_row];
			M[leader_row] = temp;
		}

		for (clear_row = leader_row + 1; clear_row < n; clear_row++) {
			if (z2poly_bit_at(&M[clear_row], clear_col))
				M[clear_row] = z2poly_add(&M[clear_row],
					&M[leader_row]);
		}
		leader_row++;
	}
}

// ----------------------------------------------------------------
// Assumes row_reduce_below has already been done.

void z2la_row_reduce_above(
	z2poly_t * M,
	int n)
{
	int row, row2, row2_deg;

	for (row = 0; row < n; row++) {
		if (z2poly_equals_unsigned(&M[row], 0))
			break;

		for (row2 = row+1; row2 < n; row2++) {
			row2_deg = z2poly_get_degree(&M[row2]);
			if (z2poly_bit_at(&M[row], row2_deg))
				M[row] = z2poly_add(&M[row], &M[row2]);
		}
		//  11111111    10000010
		//  00111111    01100110
		//  00011010    00010100
		//  00001110    00001110
		//  00000100    00000100
		//  00000001    00000001
		//  00000000    00000000
		//  00000000    00000000

	}
}

// ----------------------------------------------------------------
int z2la_find_rank(
	z2poly_t * M,
	int n)
{
	int rv = 0;
	int i;

	for (i = 0; i < n; i++)
		if (!z2poly_equals_unsigned(&M[i], 0))
			rv++;
	return rv;
}

// ----------------------------------------------------------------
// Assumes the matrix is already in row-echelon form.
void z2la_find_nullspace_basis(
	z2poly_t * BIT,
	int        n,
	z2poly_t * nullspace_basis,
	int        dimker)
{
	int i, j, deg;
	unsigned char free_flags[Z2POLY_NUM_BITS];
	unsigned char free_indices[Z2POLY_NUM_BITS];
	int nfree = 0; // == dimker but I'll compute it anyway
	int rank = n - dimker;
	int dep_pos;

	for (i = 0; i < n; i++) {
		free_flags[i] = 1;
	}
	for (i = 0; i < rank; i++) {
		deg = z2poly_get_degree(&BIT[i]);
		free_flags[deg] = 0;
	}

	for (i = 0; i < n; i++) {
		if (free_flags[i])
			free_indices[nfree++] = i;
	}

	// For each free coefficient:
	//   Let that free coefficient be 1 and the rest be zero.
	//   Also set any dependent coefficients which depend on that
	//   free coefficient.
	for (i = 0; i < dimker; i++) {
		nullspace_basis[i] = z2poly_from_base_rep(0);
		z2poly_set_bit(&nullspace_basis[i], free_indices[i]);

		for (j = 0; j < rank; j++) {
			if (z2poly_bit_at(&BIT[j], free_indices[i])) {
				dep_pos = z2poly_get_degree(&BIT[j]);
				z2poly_set_bit(&nullspace_basis[i], dep_pos);
			}
		}
	}
}
