#include "z2nilinalg.h"
#include "mod.h"
#include "malloc_check.h"

static int find_row_leader_pos(
	z2ipoly_t * prow,
	int         ncols,
	int       * ppos);

// ----------------------------------------------------------------
z2ipoly_t ** z2nila_alloc_matrix(
	int  nrows,
	int  ncols)
{
	z2ipoly_t ** rv;
	int i;

	rv = malloc_check(nrows * sizeof(z2ipoly_t *));
	for (i = 0; i < nrows; i++)
		rv[i] = malloc_check(ncols * sizeof(z2ipoly_t));
	return rv;
}

// ----------------------------------------------------------------
void z2nila_free_matrix(
	z2ipoly_t ** M,
	int    nrows,
	int    ncols)
{
	int i;
	for (i = 0; i < nrows; i++)
		free(M[i]);
	free(M);
}

// ----------------------------------------------------------------
void z2nila_print_matrix(
	z2ipoly_t ** M,
	int    nrows,
	int    ncols,
	z2ipoly_t r)
{
	int i, j;
	for (i = 0; i < nrows; i++) {
		printf(" ");
#if 1
		for (j = 0; j < ncols; j++)
#else
		for (j = ncols-1; j >= 0; j--)
#endif
		{
			printf(" ");
			z2ipolymod_print(M[i][j], r);
		}
		printf("\n");
	}
}

// ----------------------------------------------------------------
void z2nila_print_vector(
	z2ipoly_t * v,
	int    n,
	z2ipoly_t r)
{
	int i;
	int count = 0;

#if 1
	for (i = 0; i < n; i++)
#else
	for (i = n-1; i >= 0; i--)
#endif
	{
		if (count++ > 0)
			printf(" ");
		z2ipolymod_print(v[i], r);
	}
}

// ----------------------------------------------------------------
void z2nila_matrix_from_z2nipolys(
	z2ipoly_t  ** M,
	int         nrows,
	int         ncols,
	z2nipoly_t * polys)
{
	int i;
	for (i = 0; i < nrows; i++)
		z2nila_row_from_z2nipoly(M[i], &polys[i], ncols);
}

// ----------------------------------------------------------------
void z2nila_z2nipolys_from_matrix(
	z2nipoly_t  * polys,
	z2ipoly_t ** M,
	int        nrows,
	int        ncols)
{
	int i, j;
	for (i = 0; i < nrows; i++) {
		for (j = 0; j < ncols; j++) {
			polys[i].coeffs[j] = M[i][j];
			polys[i].degree = z2nipoly_find_degree(
				&polys[i], ncols - 1);
		}
	}
}

// ----------------------------------------------------------------
void z2nila_row_from_z2nipoly(
	z2ipoly_t  * prow,
	z2nipoly_t * pzpoly,
	int          ncols)
{
	int i;
	int d = pzpoly->degree;
	for (i = 0; i <= d; i++)
		prow[i] = pzpoly->coeffs[i];
	for (i = d+1; i < ncols; i++)
		prow[i] = 0LL;
}

// ----------------------------------------------------------------
void z2nila_transpose(
	z2ipoly_t ** Mout,
	z2ipoly_t ** Min,
	int        n)
{
	int row, col;

	for (row = 0; row < n; row++)
		for (col = 0; col < n; col++)
			Mout[row][col] = Min[col][row];
}

// ----------------------------------------------------------------
void z2nila_row_echelon_form(
	z2ipoly_t ** M,
	z2ipoly_t  r,
	int        nrows,
	int        ncols)
{
	z2nila_row_reduce_below (M, r, nrows, ncols);
	z2nila_row_reduce_above (M, r, nrows, ncols);
	z2nila_make_matrix_monic(M, r, nrows, ncols);
}

// ----------------------------------------------------------------
static int find_row_with_leader_at_col(
	z2ipoly_t ** M,
	int    nrows,
	int    ncols,
	int    clear_col,
	int    leader_row,
	int  * pswap_row)
{
	int row;
	for (row = leader_row; row < nrows; row++) {
		if (M[row][clear_col] != 0LL) {
			*pswap_row = row;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
static void swap_rows(
	z2ipoly_t ** M,
	int    nrows,
	int    ncols,
	int    leader_row,
	int    swap_row)
{
	int col;
	z2ipoly_t temp;

	for (col = 0; col < ncols; col++) {
		temp               = M[leader_row][col];
		M[leader_row][col] = M[swap_row][col];
		M[swap_row][col]   = temp;
	}
}

// ----------------------------------------------------------------
void z2nila_make_row_monic(
	z2ipoly_t * row,
	int   ncols,
	int   leader_col,
	z2ipoly_t r)
{
	int col;
	z2ipoly_t cinv = z2ipolymod_recip(row[leader_col], r);
	for (col = leader_col; col < ncols; col++) {
		row[col] = z2ipolymod_mul(row[col], cinv, r);
	}
}

// ----------------------------------------------------------------
void z2nila_row_reduce_below(
	z2ipoly_t ** M,
	z2ipoly_t  r,
	int        nrows,
	int        ncols)
{
	int clear_col, leader_row, swap_row, clear_row;

	for (clear_col = 0, leader_row = 0; clear_col < ncols; clear_col++) {
		if (!find_row_with_leader_at_col(M, nrows, ncols, clear_col,
			leader_row, &swap_row))
		{
			continue;
		}
		if (leader_row != swap_row)
			swap_rows(M, nrows, ncols, leader_row, swap_row);
		if (M[leader_row][clear_col] != 1LL)
			z2nila_make_row_monic(M[leader_row], ncols,
				clear_col, r);
		for (clear_row = leader_row+1; clear_row < nrows; clear_row++) {
			int col;
			z2ipoly_t clear_leader, clear_leader_inv;
			clear_leader = M[clear_row][clear_col];
			if (clear_leader == 0LL)
				continue;
			clear_leader_inv = z2ipolymod_recip(clear_leader, r);
			for (col = clear_col; col < ncols; col++) {
				M[clear_row][col] = z2ipolymod_mul(
					M[clear_row][col],
					clear_leader_inv, r);
				M[clear_row][col] = z2ipolymod_sub(
					M[clear_row][col],
					M[leader_row][col], r);
			}
		}
		leader_row++;
	}
}

// ----------------------------------------------------------------
// Assumes row_reduce_below has already been done.

void z2nila_row_reduce_above(
	z2ipoly_t ** M,
	z2ipoly_t  r,
	int        nrows,
	int        ncols)
{
	int row, row2, row2_leader_pos, col;
	z2ipoly_t row_leader_val, row2_leader_val, mul;

	for (row = 0; row < nrows; row++) {
		for (row2 = row+1; row2 < nrows; row2++) {
			if (!find_row_leader_pos(M[row2], ncols,
				&row2_leader_pos))
			{
				break;
			}
			row2_leader_val = M[row2][row2_leader_pos];
			row_leader_val  = M[row][row2_leader_pos];
			if (row_leader_val == 0LL)
				continue;

			mul = z2ipolymod_div(row_leader_val, row2_leader_val,
				r);
			for (col = 0; col < ncols; col++) {
				z2ipoly_t prod = z2ipolymod_mul(M[row2][col],
					mul, r);
				M[row][col] = z2ipolymod_sub(M[row][col],
					prod, r);
			}
		}
	}
}

// ----------------------------------------------------------------
// Assumes row_reduce_below has already been done.

void z2nila_make_matrix_monic(
	z2ipoly_t ** M,
	z2ipoly_t  r,
	int        nrows,
	int        ncols)
{
	int row, row_leader_pos, col;
	z2ipoly_t row_leader_val;

	for (row = 0; row < nrows; row++) {
		if (!find_row_leader_pos(M[row], ncols, &row_leader_pos))
			break;
		row_leader_val = M[row][row_leader_pos];

		for (col = 0; col < ncols; col++) {
			M[row][col] = z2ipolymod_div(
				M[row][col], row_leader_val, r);
		}
	}
}

// ----------------------------------------------------------------
static int find_row_leader_pos(
	z2ipoly_t * prow,
	int         ncols,
	int       * ppos)
{
	int col;

	for (col = 0; col < ncols; col++) {
		if (prow[col] != 0LL) {
			*ppos = col;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
// Assumes the matrix is already row-reduced.
int z2nila_find_rank(
	z2ipoly_t ** M,
	int    nrows,
	int    ncols)
{
	int rv = 0;
	int row;

	for (row = 0; row < nrows; row++) {
		if (!z2nila_row_is_zero(M[row], ncols))
			rv++;
	}
	return rv;
}

// ----------------------------------------------------------------
int z2nila_row_is_zero(
	z2ipoly_t  * prow,
	int          ncols)
{
	int col;
	for (col = 0; col < ncols; col++)
		if (prow[col] != 0LL)
			return 0;
	return 1;
}

// ----------------------------------------------------------------
// Assumes the matrix is already in row-echelon form.

void z2nila_find_nullspace_basis(
	z2ipoly_t ** BIT,
	int        nrows,
	int        ncols,
	z2ipoly_t  r,
	z2nipoly_t  * nullspace_basis,
	int        dimker)
{
	int i, j;
	unsigned char * free_flags;
	unsigned char * free_indices;
	int nfree = 0; // == dimker but I'll compute it anyway
	int rank = nrows - dimker;
	int dep_pos;

	for (i = 0; i < dimker; i++) {
		for (j = 0; j < ncols; j++)
			nullspace_basis[i].coeffs[j] = 0LL;
		nullspace_basis[i].degree = ncols - 1;
	}

	free_flags   = malloc_check(nrows);
	free_indices = malloc_check(nrows);

	for (i = 0; i < nrows; i++) {
		free_flags[i] = 1;
	}

	for (i = 0; i < rank; i++) {
		if (find_row_leader_pos(BIT[i], ncols, &dep_pos))
			free_flags[dep_pos] = 0;
	}

	for (i = 0; i < nrows; i++) {
		if (free_flags[i])
			free_indices[nfree++] = i;
	}

	// For each free coefficient:
	//   Let that free coefficient be 1 and the rest be zero.
	//   Also set any dependent coefficients which depend on that
	//   free coefficient.
	for (i = 0; i < dimker; i++) {
		nullspace_basis[i].coeffs[free_indices[i]] = 1LL;

		// (B-I)^t in row echelon form:
		//
		// 0210     c0 = ??      c0 = 1  c0 = 0
		// 1000     c1 = -2 c2   c1 = 0  c1 = 5
		// 0000     c2 = ??      c2 = 0  c2 = 1
		// 0000     c3 = 0       c3 = 0  c3 = 0

		// j  = 0,1
		// fi = 0,2

		// i = 0:
		//   j = 0  row 0 fi 0 = row 0 c0 = 0
		//   j = 1  row 1 fi 0 = row 1 c0 = 0
		// i = 1:
		//   j = 0  row 0 fi 1 = row 0 c2 = 2 dep_pos = 1
		//   j = 1  row 1 fi 1 = row 1 c2 = 0

		// 0001
		// 01?0

		for (j = 0; j < rank; j++) {
			if (BIT[j][free_indices[i]] == 0LL)
				continue;

			(void)find_row_leader_pos(BIT[j], ncols, &dep_pos);
			nullspace_basis[i].coeffs[dep_pos] =
				//z2ipolymod_sub -- not needed in Z2
				BIT[j][free_indices[i]];
		}
		nullspace_basis[i].degree = z2nipoly_find_degree(
			&nullspace_basis[i], ncols - 1);
	}
	free(free_flags);
	free(free_indices);
}

// ----------------------------------------------------------------
// Square matrix times vector
void z2nila_mx_times_vr(
	z2ipoly_t  ** M,
	z2ipoly_t   * vin,
	z2ipoly_t   * vout,
	int           n,
	z2ipoly_t     r)
{
	int i, j;
	z2ipoly_t prod;

	for (i = 0; i < n; i++) {
		vout[i] = 0LL;
		for (j = 0; j < n; j++) {
			prod = z2ipolymod_mul(M[i][j], vin[j], r);
			vout[i] ^= prod;
		}
	}
}
