#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "z2nipolymod.h"
#include "mod.h"
#include "power.h"
#include "tokenize.h"
#include "cmps.h"

// ----------------------------------------------------------------
void z2nipoly_check_degree(
	int degree)
{
	if (degree < 0) {
		fprintf(stderr,
			"z2nipoly:  degree must be non-negative; got %d.\n",
			degree);
		exit(1);
	}
	if (degree > Z2NIPOLY_MAX_DEGREE) {
		fprintf(stderr,
			"z2nipoly:  degree must be <= %d; got %d.\n",
			Z2NIPOLY_MAX_DEGREE, degree);
		exit(1);
	}
}

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_add(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2ipoly_t    r)
{
	z2nipoly_t   out;
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++)
			out.coeffs[i] = z2ipoly_add(pin1->coeffs[i],
				pin2->coeffs[i]);
		for (i = pin2->degree + 1; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i];
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++)
			out.coeffs[i] = z2ipoly_add(pin1->coeffs[i],
				pin2->coeffs[i]);
		for (i = pin1->degree + 1; i <= pin2->degree; i++)
			out.coeffs[i] = pin2->coeffs[i];
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were added.
	out.degree = z2nipoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_sub(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2ipoly_t    r)
{
	z2nipoly_t   out;
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++)
			out.coeffs[i] = z2ipoly_sub(pin1->coeffs[i],
				pin2->coeffs[i]);
		for (i = pin2->degree + 1; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i];
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++) {
			out.coeffs[i] = z2ipoly_sub(pin1->coeffs[i],
				pin2->coeffs[i]);
		}
		for (i = pin1->degree + 1; i <= pin2->degree; i++) {
			// out.coeffs[i] = negative of pin2->coeffs[i]
			// No-op for p=2
			out.coeffs[i] = pin2->coeffs[i];
		}
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were added.
	out.degree = z2nipoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_mul(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2ipoly_t    r)
{
	z2nipoly_t   out;
	int          i, j;
	z2ipoly_t    prod;

	out.degree = pin1->degree + pin2->degree;
	z2nipoly_check_degree(out.degree);
	for (i = 0; i <= out.degree; i++)
		out.coeffs[i] = 0LL;

	for (i = 0; i <= pin1->degree; i++) {
		for (j = 0; j <= pin2->degree; j++) {
			prod = z2ipolymod_mul(pin1->coeffs[i],
				pin2->coeffs[j], r);
			out.coeffs[i+j] = z2ipoly_add(out.coeffs[i+j], prod);
		}
	}
	// E.g. if one polynomial was zero.  My convention is that the zero
	// polynomial has "degree" zero -- correct as an array index, but
	// mathematically incorrect.
	out.degree = z2nipoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
int z2nipoly_is_zero(z2nipoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (pz->coeffs[0] != 0LL)
		return 0;
	return 1;
}

// ----------------------------------------------------------------
int z2nipoly_is_one(z2nipoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (pz->coeffs[0] != 1LL)
		return 0;
	return 1;
}

// ----------------------------------------------------------------
void z2nipoly_make_monic(
	z2nipoly_t * pz,
	z2ipoly_t    r)
{
	int i;
	z2ipoly_t c, cinv;

	c = pz->coeffs[pz->degree];
	if (c == 0LL)
		return;
	cinv = z2ipolymod_recip(c, r);

	for (i = 0; i <= pz->degree; i++) {
		pz->coeffs[i] = z2ipolymod_mul(pz->coeffs[i], cinv, r);
	}
}

// ----------------------------------------------------------------
static void z2nipoly_smul(
	z2nipoly_t * pout,
	z2nipoly_t * pin, // Poly
	z2ipoly_t    scalar, // Scalar
	z2ipoly_t    r)
{
	z2nipoly_t  out;
	int i;

	for (i = 0; i <= pin->degree; i++)
		out.coeffs[i] = z2ipolymod_mul(pin->coeffs[i], scalar, r);

	if (scalar == 0LL)
		out.degree = 0;
	else
		out.degree = pin->degree;
	*pout = out;
}

// ----------------------------------------------------------------
int z2nipoly_find_degree(
	z2nipoly_t * pin,
	int          max_deg)
{
	int i;

	for (i = max_deg; i >= 0; i--)
		if (pin->coeffs[i] != 0LL)
			return i;
	return 0;
}

// ----------------------------------------------------------------
// E.g.
// dividend = 1,2,3,4 (1 + 2x + 3x^2 + 4x^3)
// divisor  = 1,1,2   (1 +  x + 2x^2)
// modulus = 7
//
//             q=4,2   r = 4,3
//        +----------
// 1,1,2  |  1,2,3,4
//        |    2,2,4 shift = 1.  4/2 mod 7 = 2.  1,1,2 * 2 = 2,2,4.
//        +----------
//        |  1 0 1
//        |  4 4 1   shift = 0.  1/2 mod 7 = 4   1,1,2 * 4 = 4,4,1
//        +----------
//        |  4 3

void z2nipoly_quot_and_rem(
	z2nipoly_t * pdividend,
	z2nipoly_t * pdivisor,
	z2nipoly_t * pquot,
	z2nipoly_t * prem,
	z2ipoly_t    r)
{
	z2nipoly_t quot;
	z2nipoly_t rem;
	int i;
	int max_shift;
	int shift;
	z2ipoly_t rem_leader;
	z2ipoly_t divisor_leader;
	z2ipoly_t dlinv;
	z2ipoly_t multiplier;
	z2nipoly_t divisor_with_mul;
	int stop;

	max_shift = pdividend->degree - pdivisor->degree;
	if (pdividend->degree < pdivisor->degree) {
		quot.coeffs[0] = 0LL;
		quot.degree = 0;
		rem = *pdividend;

		*pquot = quot;
		*prem  = rem;
		return;
	}

	quot.degree = pdividend->degree - pdivisor->degree;
	for (i = 0; i <= quot.degree; i++)
		quot.coeffs[i] = 0LL;
	rem = *pdividend;

	divisor_leader = pdivisor->coeffs[pdivisor->degree];
	dlinv = z2ipolymod_recip(divisor_leader, r);
	for (shift = max_shift; shift >= 0; shift--) {
		rem_leader = rem.coeffs[shift + pdivisor->degree];
		multiplier = z2ipolymod_mul(rem_leader, dlinv, r);
		z2nipoly_smul(&divisor_with_mul, pdivisor, multiplier, r);
		quot.coeffs[shift] = multiplier;
		stop = shift + divisor_with_mul.degree;
		for (i = shift; i <= stop; i++) {
			rem.coeffs[i] = z2ipoly_sub(rem.coeffs[i],
				divisor_with_mul.coeffs[i-shift]);
		}
	}

	quot.degree = z2nipoly_find_degree(&quot, quot.degree);
	rem.degree  = z2nipoly_find_degree(&rem,  rem.degree);

	*pquot = quot;
	*prem  = rem;
}

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_div(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2ipoly_t    r)
{
	z2nipoly_t quot, rem;
	z2nipoly_quot_and_rem(pin1, pin2, &quot, &rem, r);
	return quot;
}

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_mod(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2ipoly_t    r)
{
	z2nipoly_t quot, rem;
	z2nipoly_quot_and_rem(pin1, pin2, &quot, &rem, r);
	return rem;
}

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_gcd(
	z2nipoly_t * pa,
	z2nipoly_t * pb,
	z2ipoly_t    fr)
{
	z2nipoly_t c, quot, rem;
	z2nipoly_t d;   // Return value.

	if (z2nipoly_is_zero(pa))
		return *pb;
	if (z2nipoly_is_zero(pb))
		return *pa;

	c = *pa;
	d = *pb;

	while (1) {

		// Divide
		// quot = c / d, rem = c % d
		z2nipoly_quot_and_rem(&c, &d, &quot, &rem, fr);
		// Note:  now c = qd + rem and 0 <= rem < d

		// Remainder zero?
		if (z2nipoly_is_zero(&rem))
			break;

		// Recycle
		c = d;
		d = rem;
	}
	return d;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
z2nipoly_t z2nipoly_power(
	z2nipoly_t * pin,
	int          power,
	z2ipoly_t    r)
{
	z2nipoly_t in2 = *pin;
	z2nipoly_t inpower = z2nipoly_from_coeffs0(1LL);

	while (power != 0) {
		if (power & 1)
			inpower = z2nipoly_mul(&in2, &inpower, r);
		power = (unsigned)power >> 1;
		in2 = z2nipoly_mul(&in2, &in2, r);
	}
	return inpower;
}

// ----------------------------------------------------------------
// Assumes the input is already known to be a square.

z2nipoly_t z2nipoly_square_root(
	z2nipoly_t * pin,
	z2ipoly_t    r)
{
	int si, di;
	z2nipoly_t out;

	for (di = 0; di <= pin->degree; di++) {
		out.coeffs[di] = 0LL;
	}

	for (si = 0, di = 0; si <= pin->degree; si += 2, di++) {
		out.coeffs[di] = pin->coeffs[si];
		if (si+1 > pin->degree)
			break;
		if (pin->coeffs[si+1] != 0LL) {
			printf("z2nipoly_square_root: ");
			z2nipoly_print(pin, r);
			printf(" is not a square.\n");
			exit(1);
		}
	}
	out.degree = z2nipoly_find_degree(&out, pin->degree);

	return out;
}

// ----------------------------------------------------------------
int z2nipoly_equal(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2ipoly_t    r)
{
	int i;
	if (pin1->degree != pin2->degree)
		return 0;
	for (i = 0; i <= pin1->degree; i++)
		if (pin1->coeffs[i] != pin2->coeffs[i])
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int z2nipoly_lt(z2nipoly_t * pa, z2nipoly_t * pb)
{
	return z2nipoly_cmp(pa, CMP_LT, pb);
}
int z2nipoly_le(z2nipoly_t * pa, z2nipoly_t * pb)
{
	return z2nipoly_cmp(pa, CMP_LE, pb);
}
int z2nipoly_gt(z2nipoly_t * pa, z2nipoly_t * pb)
{
	return z2nipoly_cmp(pa, CMP_GT, pb);
}
int z2nipoly_ge(z2nipoly_t * pa, z2nipoly_t * pb)
{
	return z2nipoly_cmp(pa, CMP_GE, pb);
}
int z2nipoly_eq(z2nipoly_t * pa, z2nipoly_t * pb)
{
	return z2nipoly_cmp(pa, CMP_EQ, pb);
}
int z2nipoly_ne(z2nipoly_t * pa, z2nipoly_t * pb)
{
	return z2nipoly_cmp(pa, CMP_NE, pb);
}

// ----------------------------------------------------------------
int z2nipoly_cmp(
	z2nipoly_t * pa,
	int          cmp,
	z2nipoly_t * pb)
{
	int direction = 0; // -1 = less, 0 = equal, +1 = greater;
	int i;

	if (pa->degree > pb->degree) {
		direction =  1;
	}
	else if (pa->degree < pb->degree) {
		direction = -1;
	}
	else {
		for (i = pa->degree; i >= 0; i--) {
			if (pa->coeffs[i] < pb->coeffs[i]) {
				direction = -1;
				break;
			}
			else if (&pa->coeffs[i] > &pb->coeffs[i]) {
				direction =  1;
				break;
			}
		}
	}

	switch (cmp) {
	case CMP_LE: return (direction <= 0); break;
	case CMP_LT: return (direction <  0); break;
	case CMP_EQ: return (direction == 0); break;
	case CMP_NE: return (direction != 0); break;
	case CMP_GE: return (direction >= 0); break;
	case CMP_GT: return (direction >  0); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); return 0; break;
	}
}

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_from_coeffs0(
	z2ipoly_t c0)
{
	z2nipoly_t rv;
	rv.degree = 0;
	rv.coeffs[0] = c0;
	return rv;
}

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_from_coeffs1(
	z2ipoly_t c1,
	z2ipoly_t c0)
{
	z2nipoly_t rv;
	rv.coeffs[1] = c1;
	rv.coeffs[0] = c0;
	rv.degree = z2nipoly_find_degree(&rv, 1);
	return rv;
}

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_from_coeffs2(
	z2ipoly_t c2,
	z2ipoly_t c1,
	z2ipoly_t c0)
{
	z2nipoly_t rv;
	rv.coeffs[2] = c2;
	rv.coeffs[1] = c1;
	rv.coeffs[0] = c0;
	rv.degree = z2nipoly_find_degree(&rv, 2);
	return rv;
}

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_from_coeffs3(
	z2ipoly_t c3,
	z2ipoly_t c2,
	z2ipoly_t c1,
	z2ipoly_t c0)
{
	z2nipoly_t rv;
	rv.coeffs[3] = c3;
	rv.coeffs[2] = c2;
	rv.coeffs[1] = c1;
	rv.coeffs[0] = c0;
	rv.degree = z2nipoly_find_degree(&rv, 3);
	return rv;
}

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_from_base_rep(
	int rep,
	int intermed_degree)
{
	int mask;
	z2nipoly_t rv;

	if (rep < 0) {
		fprintf(stderr, "z2nipoly_from_base_rep: negative input %d.\n",
			rep);
		exit(1);
	}

	if (intermed_degree < 1) {
		fprintf(stderr,
			"z2nipoly_from_base_rep: need deg >= 1; got %d.\n",
			intermed_degree);
		exit(1);
	}
	mask = (1 << intermed_degree) - 1;

	rv.degree = 0;
	rv.coeffs[0] = 0;
	while (rep) {
		rv.coeffs[rv.degree] = rep & mask;
		rep >>= intermed_degree;
		if (rv.degree >= Z2NIPOLY_MAX_DEGREE) {
			fprintf(stderr, "z2nipoly_from_base_rep: overflow.\n");
			exit(1);
		}
		rv.degree++;
	}
	if (rv.degree > 0)
		rv.degree--;

	return rv;
}

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_extend_z2ipoly(
	z2ipoly_t    zp)
{
	z2nipoly_t zpn;
	int i;

	zpn.degree = z2ipoly_get_degree(zp);
	for (i = 0; i <= zpn.degree; i++)
		zpn.coeffs[i] = Z2IPOLY_BIT_AT(zp, i);
	return zpn;
}

// ----------------------------------------------------------------
z2ipoly_t z2ipoly_unextend_z2nipoly(
	z2nipoly_t * pznp,
	z2ipoly_t    r)
{
	int i;
	z2ipoly_t rv = 0LL;

	for (i = 0; i <= pznp->degree; i++) {
		if (z2ipoly_get_degree(pznp->coeffs[i]) > 0) {
			printf("z2npoly:  Can't unextend ");
			z2nipoly_print(pznp, r);
			printf("\n");
			exit(1);
		}
		if (pznp->coeffs[i])
			rv |= 1LL << i;
	}

	return rv;
}

// ----------------------------------------------------------------
// Example:
// * a_red   = a^2 + a + 1 = 0x07 which defines an F_2^2.
// * b_red   = b^6 + b + 1 = 0x43 which defines an F_2^6.
// * ab_poly = x^3 + x^2 + (a+1) x + a = 1:1:3:2
//   which is a cubic polynomial in F_2^2[x].
// * ab_exp  = 21 since (2^6-1) / (2^2-1) = 21.
//
// This function substitutes x with b and a with b^exp, then reduces mod
// b_red.  The above example ab_poly reduces to 0 mod b^6 + b + 1.

z2ipoly_t z2nipoly_flatten(
	z2nipoly_t   * pab_poly,
	z2ipoly_t      a_red,
	z2ipoly_t      b_red,
	int            ab_exp)
{
	z2ipoly_t flat_poly = 0LL;
	int in_deg, j, i, out_deg;

	for (j = 0; j <= pab_poly->degree; j++) {
		in_deg = z2ipoly_get_degree(pab_poly->coeffs[j]);
		for (i = 0; i <= in_deg; i++) {
			if ((pab_poly->coeffs[j] >> i) & 1LL) {
				out_deg = j + ab_exp * i;

				//z2ipoly_check_degree(out_deg);
				//flat_poly ^= 1LL << out_deg;
				flat_poly ^= z2ipolymod_power(2LL,
					out_deg, b_red);
			}
		}
	}
	flat_poly = z2ipoly_mod(flat_poly, b_red);

	return flat_poly;
}

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_deriv(
	z2nipoly_t * pz,
	z2ipoly_t    r)
{
	z2nipoly_t rv;
	int i;

	if (pz->degree == 0) {
		return z2nipoly_from_coeffs0(0LL);
	}

	for (i = 1; i <= pz->degree; i++) {
		if (i & 1)
			rv.coeffs[i-1] = pz->coeffs[i];
		else
			rv.coeffs[i-1] = 0LL;
	}
	rv.degree = z2nipoly_find_degree(&rv, pz->degree - 1);

	return rv;
}

// ----------------------------------------------------------------
z2ipoly_t z2nipoly_eval(
	z2ipoly_t    a,
	z2nipoly_t * poutr,
	z2ipoly_t    innr)
{
	z2ipoly_t rv;
	int d;

	rv = poutr->coeffs[poutr->degree];
	for (d = poutr->degree - 1; d >= 0; d--) {
		rv = z2ipolymod_mul(rv, a, innr);
		rv = z2ipolymod_add(rv, poutr->coeffs[d], innr);
	}
	return rv;
}

// ----------------------------------------------------------------
int z2nipoly_is_rootless(
	z2nipoly_t * poutr,
	z2ipoly_t    innr)
{
	z2ipoly_t a;
	z2ipoly_t pn = 1LL << z2ipoly_get_degree(innr);

	for (a = 0LL; a < pn; a++)
		if (z2nipoly_eval(a, poutr, innr) == 0LL)
			return 0;
	return 1;
}

// ----------------------------------------------------------------
// Assumes outr is known to be irreducible.
// See if the order of x mod r is p^m-1.  Certainly, x^(p^m-1) will be 1.  But
// check x to all powers dividing p^m-1 to make sure they are *not* 1.

int z2nipoly_is_primitive(
	z2nipoly_t * poutr,
	z2ipoly_t    innr)
{
	unsigned m = poutr->degree;
	unsigned pm = 1 << m;
	unsigned pm1 = pm - 1;
	factor_info_t finfo;

	calc_factors(pm1, &finfo);
	return z2nipoly_is_primitive_f(poutr, innr, pm1, &finfo);
}

// ----------------------------------------------------------------
int z2nipoly_is_primitive_f(
	z2nipoly_t * poutr,
	z2ipoly_t    innr,
	unsigned     pm1,
	factor_info_t * pfinfo)
{
	z2nipoly_t x   = z2nipoly_from_coeffs1(1LL, 0LL);
	z2nipoly_t one = z2nipoly_from_coeffs0(1LL);
	z2nipoly_t xpower;
	unsigned nf, f, k;
	int is_one;

	if (z2nipoly_equal(poutr, &x, innr))
		return 0;

	nf = num_divisors(pfinfo);

	for (k = 0; k < nf; k++) {
		f = kth_divisor(pfinfo, k);
		xpower = z2nipolymod_power(&x, f, poutr, innr);
		is_one = z2nipoly_equal(&xpower, &one, innr);
		if ((f != pm1) && is_one) {
			return 0;
		}
		if ((f == pm1) && !is_one) {
			fprintf(stderr, "ABEND!\n");
			exit(1);
		}
	}
	return 1;
}

// ----------------------------------------------------------------
// Example with p=2 and hexadecimal notation:
//
//   mod 13:
//   02 =  0 . . . [  0]
//   03 =  1 . . . [  1]
//   07 =  6 7 . . [  3]
//   13 =  2 3 4 5 [ 15]
//   19 =  9 b d e [ 15]
//   1f =  8 a c f [  5]
//
// 13 is the reductor r.  b is the input z.  19 is the desired minimal
// polynomial.  Fact:  If b is a root of 19, so are all the elements
// of its Frobenius orbit mod r, e.g. b^2, b^4, b^8.  These powers yield
// b, 9, d and e.  Then, (X-b)(X-9)(X-d)(X-e) is the desired minimal
// polynomial with coefficients in Fpn[x], 11001 .  Un-extend to Fp[x]
// to get 19 (in hex).

//         1 b
// *       1 9
// ===========
//         9 c
// +     1 b
// ===========
//       1 2 c
// *       1 d
// ===========
//       d 9 3
// +   1 2 c
// ===========
//     1 f 5 3
// *       1 e
// ===========
//     e 5 3 1
// + 1 f 5 3
// ===========
//   1 1 0 0 1 = 19 in hex.

z2ipoly_t z2ipoly_get_min_poly(
	z2ipoly_t z,
	z2ipoly_t r)
{
	int n = z2ipoly_get_degree(r);
	int i;
	z2nipoly_t extrv = z2nipoly_extend_z2ipoly(1LL);
	z2ipoly_t z2ipower = z;
	z2nipoly_t factor;
	z2ipoly_t rv;

	for (i = 0; i < n; i++) {
		factor.degree = 1;
		factor.coeffs[1] = 1LL;
		//factor.coeffs[0] = z2ipoly_usub(&z2ipower);
		//Z2 self-negates
		factor.coeffs[0] = z2ipower;
		extrv = z2nipoly_mul(&extrv, &factor, r);
		z2ipower = z2ipolymod_power(z2ipower, 2, r);
		if (z2ipower == z)
			break;
	}
	rv = z2ipoly_unextend_z2nipoly(&extrv, r);
	return rv;
}

// ----------------------------------------------------------------
void z2nipoly_print(
	z2nipoly_t * pin,
	z2ipoly_t    r)
{
	int i;

	for (i = pin->degree; i >= 0; i--) {
		if (i < pin->degree)
			printf(":");
		z2ipolymod_print(pin->coeffs[i], r);
	}
}

// ----------------------------------------------------------------
int z2nipoly_list_scan(
	z2nipoly_t * pout,
	char       * pin,
	z2ipoly_t    r)
{
	int    i;
	char * coeff_stringv[Z2NIPOLY_MAX_DEGREE + 1];
	char   temp_line[1024];

	pout->degree = 0;
	for (i = 0; i <= Z2NIPOLY_MAX_DEGREE; i++)
		pout->coeffs[i] = 0LL;

	strcpy(temp_line, pin);
	pout->degree = tokenize(temp_line, ":", coeff_stringv,
		Z2NIPOLY_MAX_DEGREE + 1) - 1;
	if (pout->degree < 0) {
		fprintf(stderr, "z2nipoly_list_scan:  Couldn't scan \"%s\".\n",
			pin);
		return 0;
	}

	for (i = 0; i <= pout->degree; i++) {
		if (!z2ipoly_scan(&pout->coeffs[pout->degree - i],
			coeff_stringv[i]))
		{
			fprintf(stderr,
				"z2nipoly:  Couldn't scan coefficient \"%s\".\n",
				coeff_stringv[pout->degree - i]);
			return 0;
		}
	}

	// Check for leading zero coefficients, which would reduce the
	// degree.
	pout->degree = z2nipoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
int z2nipoly_exp_scan(
	z2nipoly_t * pout,
	char       * pin,
	z2ipoly_t    r)
{
	int    i;
	int    num_pieces;
	char * coeff_stringv[Z2NIPOLY_MAX_DEGREE + 1];
	char * coeff_and_exp[2];
	int    num_sub_pieces;
	char   temp_line[1024];
	char   temp_piece[128];
	int    exp;

	pout->degree = 0;
	for (i = 0; i <= Z2NIPOLY_MAX_DEGREE; i++)
		pout->coeffs[i] = 0LL;

	strcpy(temp_line, pin);
	num_pieces = tokenize(temp_line, ":", coeff_stringv,
		Z2NIPOLY_MAX_DEGREE + 1);
	if (num_pieces <= 0) {
		fprintf(stderr, "z2nipoly_exp_scan:  Couldn't scan \"%s\".\n",
			pin);
		return 0;
	}

	for (i = 0; i < num_pieces; i++) {
		// Split on "^"; expect z2ipoly on left and int on right.
		// Bounds-check the exponent.
		strcpy(temp_piece, coeff_stringv[i]);
		num_sub_pieces = tokenize(temp_piece, "^", coeff_and_exp, 2);
		if (num_sub_pieces < 1) {
			fprintf(stderr,
				"z2nipoly:  Couldn't scan coefficient \"%s\".\n",
				temp_piece);
			return 0;
		}
		else if (num_sub_pieces == 1) {
			exp = 0;
		}
		else {
			// sscanf and bounds-check the exp
			if (sscanf(coeff_and_exp[1], "%d", &exp) != 1) {
				fprintf(stderr,
				"z2nipoly:  Couldn't scan coefficient \"%s\".\n",
					temp_piece);
				return 0;
			}
			if ((exp < 0) || (exp > Z2NIPOLY_MAX_DEGREE)) {
				fprintf(stderr,
				"z2nipoly:  Exponent %d out of range 0-%d.\n",
					exp, Z2NIPOLY_MAX_DEGREE);
				return 0;
			}
		}

		// Remember highest exponent encountered.
		if (exp > pout->degree)
			pout->degree = exp;

		// scan and bounds-check the coeff
		if (!z2ipoly_scan(&pout->coeffs[exp], coeff_and_exp[0])) {
			fprintf(stderr,
				"z2nipoly:  Couldn't scan coefficient \"%s\".\n",
				temp_piece);
			return 0;
		}
	}

	// Check for leading zero coefficients, which would reduce the
	// degree.
	pout->degree = z2nipoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
int z2nipoly_scan(
	z2nipoly_t * pout,
	char       * pin,
	z2ipoly_t    r)
{
	if (strchr(pin, '^'))
		return z2nipoly_exp_scan(pout, pin, r);
	else
		return z2nipoly_list_scan(pout, pin, r);
}
