// Elements of the ring F_2^n[x] (polynomials with coefficients in a
// Galois field).

#ifndef Z2NIPOLY_H
#define Z2NIPOLY_H

#include "group.h"
#include "ring.h"
#include "z2ipolymod.h"

// ----------------------------------------------------------------
#define Z2NIPOLY_MAX_DEGREE Z2IPOLY_MAX_DEGREE
typedef struct _z2nipoly_t {
	z2ipoly_t coeffs[Z2NIPOLY_MAX_DEGREE+1];
	int degree;
} z2nipoly_t;

int z2nipoly_is_zero(
	z2nipoly_t * pz);

int z2nipoly_is_one(
	z2nipoly_t * pz);

void z2nipoly_make_monic(
	z2nipoly_t * pz,
	z2ipoly_t    r);

z2nipoly_t z2nipoly_add(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2ipoly_t    r);

z2nipoly_t z2nipoly_sub(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2ipoly_t    r);

z2nipoly_t z2nipoly_mul(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2ipoly_t    r);

void z2nipoly_quot_and_rem(
	z2nipoly_t * pdividend,
	z2nipoly_t * pdivisor,
	z2nipoly_t * pquot,
	z2nipoly_t * prem,
	z2ipoly_t    r);

z2nipoly_t z2nipoly_div(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2ipoly_t    r);

z2nipoly_t z2nipoly_mod(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2ipoly_t    r);

z2nipoly_t z2nipoly_gcd(
	z2nipoly_t * pa,
	z2nipoly_t * pb,
	z2ipoly_t    r);

z2nipoly_t z2nipoly_power(
	z2nipoly_t * pin,
	int          power,
	z2ipoly_t    r);

// Assumes the input is already known to be a square.
z2nipoly_t z2nipoly_square_root(
	z2nipoly_t * pin,
	z2ipoly_t    r);

// ----------------------------------------------------------------
int z2nipoly_equal(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2ipoly_t    r);

// Fpn[x] is *not* totally ordered.  Nonetheless, these functions are
// handy for looping and sorting.
int z2nipoly_lt(z2nipoly_t * pa, z2nipoly_t * pb);
int z2nipoly_gt(z2nipoly_t * pa, z2nipoly_t * pb);
int z2nipoly_le(z2nipoly_t * pa, z2nipoly_t * pb);
int z2nipoly_ge(z2nipoly_t * pa, z2nipoly_t * pb);
int z2nipoly_eq(z2nipoly_t * pa, z2nipoly_t * pb);
int z2nipoly_ne(z2nipoly_t * pa, z2nipoly_t * pb);

int z2nipoly_cmp(
	z2nipoly_t * pa,
	int           cmp, // see cmps.h
	z2nipoly_t * pb);

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_from_coeffs0(
	z2ipoly_t c0);

z2nipoly_t z2nipoly_from_coeffs1(
	z2ipoly_t c1,
	z2ipoly_t c0);

z2nipoly_t z2nipoly_from_coeffs2(
	z2ipoly_t c2,
	z2ipoly_t c1,
	z2ipoly_t c0);

z2nipoly_t z2nipoly_from_coeffs3(
	z2ipoly_t c3,
	z2ipoly_t c2,
	z2ipoly_t c1,
	z2ipoly_t c0);

z2nipoly_t z2nipoly_extend_z2ipoly(
	z2ipoly_t    zp);

z2ipoly_t z2ipoly_unextend_z2nipoly(
	z2nipoly_t * pznp,
	z2ipoly_t    r);

// Example:
// * a_red   = a^2 + a + 1 = 0x07 which defines an F_2^2.
// * b_red   = b^6 + b + 1 = 0x43 which defines an F_2^6.
// * ab_poly = x^3 + x^2 + (a+1) x + a = 1:1:3:2
//   which is a cubic polynomial in F_2^2[x].
// * ab_exp  = 21 since (2^6-1) / (2^2-1) = 21.
//
// This function substitutes x with b and a with b^exp, then reduces mod
// b_red.  The above example ab_poly reduces to 0 mod b^6 + b + 1.

z2ipoly_t z2nipoly_flatten(
	z2nipoly_t   * pab_poly,
	z2ipoly_t      a_red,
	z2ipoly_t      b_red,
	int            ab_exp);

z2nipoly_t z2nipoly_from_base_rep(
	int rep,
	int intermed_degree);

// ----------------------------------------------------------------
z2nipoly_t z2nipoly_deriv(
	z2nipoly_t * pz,
	z2ipoly_t    r);

z2ipoly_t z2nipoly_eval(
	z2ipoly_t    a,
	z2nipoly_t * poutr,
	z2ipoly_t    innr);

int z2nipoly_is_rootless(
	z2nipoly_t * poutr,
	z2ipoly_t    innr);

int z2nipoly_is_primitive(
	z2nipoly_t * poutr,
	z2ipoly_t    innr);

int z2nipoly_is_primitive_f(
	z2nipoly_t * poutr,
	z2ipoly_t    innr,
	unsigned     pm1,
	factor_info_t * pfinfo);

//z2ipoly_t z2ipoly_compose(
//	z2ipoly_t r,
//	z2ipoly_t z);

//z2ipoly_t z2ipoly_deriv(
//	z2ipoly_t z);

z2ipoly_t z2ipoly_get_min_poly(
	z2ipoly_t z,
	z2ipoly_t r);

// ----------------------------------------------------------------
// For general use, please just examine pin->degree.  This function is
// used to recompute the degree upon polynomial multiplication.
int z2nipoly_find_degree(
	z2nipoly_t * pin,
	int        max_deg);

// Used to detect overflow in polynomial multiplication.
void z2nipoly_check_degree(
	int degree);

// ----------------------------------------------------------------
void z2nipoly_print(z2nipoly_t * pin, z2ipoly_t r);
int  z2nipoly_scan(z2nipoly_t * pout, char * pin, z2ipoly_t r);

#endif // Z2NIPOLY_H
