#include <stdio.h>
#include <stdlib.h>
#include "malloc_check.h"
#include "power.h"
#include "z2nipolymod.h"
#include "z2nipoly_factor.h"
#include "z2nilinalg.h"

static void gather_multiple_factors(
	z2nipoly_t  * pirred_factors,
	z2ipoly_t     r,
	int           num_irred_factors,
	z2nipoly_factor_info_t * pfinfo);

static void sort_up(
	z2nipoly_t  * polys,
	int           nrows);

static void pre_berlekamp2ni(
	z2nipoly_t  * pz,
	z2ipoly_t     r,
	z2nipoly_t  * pirred_factors,
	int         * pnum_irred_factors);

static void berlekamp2ni(
	z2nipoly_t  * pz,
	z2ipoly_t     r,
	z2nipoly_t  * pirred_factors,
	int         * pnum_irred_factors);

// ----------------------------------------------------------------
void z2nipoly_factor(
	z2nipoly_t  * pz,
	z2ipoly_t     r,
	z2nipoly_factor_info_t * pfinfo)
{
	int num_irred_factors, i;
	z2nipoly_t * irred_factors = malloc_check(
		Z2NIPOLY_MAX_NUM_FACTORS * sizeof(z2nipoly_t));
	z2ipoly_t leader;

	num_irred_factors = 0;
	pre_berlekamp2ni(pz, r, irred_factors, &num_irred_factors);

	for (i = 0; i < num_irred_factors; i++) {
		z2nipoly_make_monic(&irred_factors[i], r);
	}

	leader = pz->coeffs[pz->degree];
	if ((leader != 0LL) && (leader != 1LL)) {
		irred_factors[num_irred_factors++] =
			z2nipoly_from_coeffs0(leader);
	}

	sort_up(irred_factors, num_irred_factors);
	gather_multiple_factors(irred_factors, r, num_irred_factors, pfinfo);

	free(irred_factors);
}

// ----------------------------------------------------------------
static void gather_multiple_factors(
	z2nipoly_t * pirred_factors,
	z2ipoly_t    r,
	int          num_irred_factors,
	z2nipoly_factor_info_t * pfinfo)
{
	int i;
	z2nipoly_t  * plast;
	int nd = 0;

	if (num_irred_factors == 0) {
		pfinfo->num_distinct = 0;
		return;
	}

	plast = &pirred_factors[0];
	pfinfo->factor_counts[nd].f = *plast;
	pfinfo->factor_counts[nd].count = 0;

	for (i = 0; i < num_irred_factors; i++) {
		if (!z2nipoly_equal(&pirred_factors[i], plast, r)) {
			nd++;
			pfinfo->factor_counts[nd].count = 1;
			pfinfo->factor_counts[nd].f = pirred_factors[i];
			plast = &pirred_factors[i];
		}
		else {
			pfinfo->factor_counts[nd].count++;
		}
	}
	pfinfo->num_distinct = nd + 1;
}

// ----------------------------------------------------------------
static void pre_berlekamp2ni(
	z2nipoly_t  * pz,
	z2ipoly_t     r,
	z2nipoly_t  * pirred_factors,
	int         * pnum_irred_factors)
{
	z2nipoly_t d, g;

	d = z2nipoly_deriv(pz, r);
	g = z2nipoly_gcd(pz, &d, r);

	if (z2nipoly_is_zero(&g)) {
		if (!z2nipoly_is_zero(pz)) {
			fprintf(stderr, "CODING ERROR file %s\n", __FILE__);
			exit(1);
		}
		pirred_factors[*pnum_irred_factors] = *pz;
		(*pnum_irred_factors)++;
		return;
	}
	else if (g.degree == 0) {
		// Input is squarefree.
		berlekamp2ni(pz, r, pirred_factors, pnum_irred_factors);
	}
	else if (z2nipoly_is_zero(&d)) {
		// Input is a perfect square.
		// Multiplicity is p only if degree is > 0.
		z2nipoly_t s = z2nipoly_square_root(pz, r);
		int i;
		int multiplicity = pz->degree > 0 ? 2 : 1;

		for (i = 0; i < multiplicity; i++) {
			pre_berlekamp2ni(&s, r, pirred_factors,
				pnum_irred_factors);
		}
	}
	else {
		z2nipoly_t q;
		q = z2nipoly_div(pz, &g, r);
		pre_berlekamp2ni(&g, r, pirred_factors, pnum_irred_factors);
		pre_berlekamp2ni(&q, r, pirred_factors, pnum_irred_factors);
	}
}

// ----------------------------------------------------------------
static void berlekamp2ni(
	z2nipoly_t  * pz,
	z2ipoly_t     r,
	z2nipoly_t  * pirred_factors,
	int         * pnum_irred_factors)
{
	int n = pz->degree;
	int q = 1 << z2ipoly_get_degree(r);
	z2nipoly_t x  = z2nipoly_from_coeffs1(1LL, 0LL);
	z2nipoly_t xq;
	z2nipoly_t xqi = z2nipoly_from_coeffs0(1LL);
	int row, rank, dimker;
	z2ipoly_t c;
	int got_it = 0;

	z2ipoly_t    ** BI;
	z2ipoly_t    ** BIT;
	z2nipoly_t * nullspace_basis;
	z2nipoly_t   h, f1, f2;

	if (n < 2) {
		pirred_factors[*pnum_irred_factors] = *pz;
		(*pnum_irred_factors)++;
		return;
	}

	BI  = z2nila_alloc_matrix(n, n);
	BIT = z2nila_alloc_matrix(n, n);
	nullspace_basis = malloc_check(n * sizeof(z2nipoly_t));

	xq = z2nipolymod_power(&x, q, pz, r);

	// Populate the B matrix with row i equal to x^pi mod z
	for (row = 0; row < n; row++) {
		z2nila_row_from_z2nipoly(BI[row], &xqi, n);
		xqi = z2nipolymod_mul(&xqi, &xq, pz, r);
	}

	// Replace B with B - I
	for (row = 0; row < n; row++) {
		BI[row][row] = z2ipolymod_sub(BI[row][row], 1LL, r);
	}

	z2nila_transpose(BIT, BI, n);
	z2nila_row_echelon_form(BIT, r, n, n);
	rank = z2nila_find_rank(BIT, n, n);
	dimker = n - rank;

	if (dimker == 1) {
		pirred_factors[*pnum_irred_factors] = *pz;
		(*pnum_irred_factors)++;
		return;
	}

	z2nila_find_nullspace_basis(BIT, n, n, r, nullspace_basis, dimker);

#if 1
	// Check h^q = h.  The algorithm guarantees this, *if* implemented
	// correctly.  This computation, while not strictly necessary, is a
	// nice sanity check.
	for (row = 0; row <  dimker; row++) {
		z2nipoly_t * ph = &nullspace_basis[row];
		z2nipoly_t hp;
		hp = z2nipolymod_power(ph, q, pz, r);
		if (!z2nipoly_equal(ph, &hp, r)) {
			fprintf(stderr, "CODING ERROR\n");
			exit(1);
		}
	}
#endif

	got_it = 0;
	for (row = 0; row <  dimker && !got_it; row++) {
		if (z2nipoly_is_one(&nullspace_basis[row]))
			continue;

		for (c = 0; c < (z2ipoly_t)q; c++) {
			z2nipoly_t cpoly = z2nipoly_from_coeffs0(c);
			h = z2nipoly_sub(&nullspace_basis[row], &cpoly, r);
			f1 = z2nipoly_gcd(pz, &h, r);

			if ((0 < f1.degree) && (f1.degree < pz->degree)) {
				got_it = 1;
				break;
			}
		}
	}

	if (!got_it) {
		printf("CODING ERROR: no non-trivial factors found: ");
		z2nipoly_print(pz, r);
		printf("\n");
		exit(1);
	}

	f2 = z2nipoly_div(pz, &f1, r);

	// The nullity of B-I is the number of irreducible
	// factors of r.  If the nullity is 2, we have a
	// pair of factors which are both irreducible and
	// so we don't need to recurse.
	if (dimker == 2) {
		pirred_factors[*pnum_irred_factors] = f1;
		(*pnum_irred_factors)++;
		pirred_factors[*pnum_irred_factors] = f2;
		(*pnum_irred_factors)++;
	}
	else {
		pre_berlekamp2ni(&f1, r, pirred_factors, pnum_irred_factors);
		pre_berlekamp2ni(&f2, r, pirred_factors, pnum_irred_factors);
	}

	z2nila_free_matrix(BI,  n, n);
	z2nila_free_matrix(BIT, n, n);
	free(nullspace_basis);
}

// ----------------------------------------------------------------
// Ascending bubble sort
static void sort_up(
	z2nipoly_t * polys,
	int          nrows)
{
	int i, top;
	z2nipoly_t  temp;

	for (top = nrows-1; top > 0; top--) {
		for (i = 0; i < top; i++) {
			if (!z2nipoly_lt(&polys[i], &polys[i+1])) {
				temp       = polys[i];
				polys[i]   = polys[i+1];
				polys[i+1] = temp;
			}
		}
	}
}

// ----------------------------------------------------------------
// Given the prime factorization p1^m1 ... pk^mk of n, how to enumerate all
// factors of n?
//
// Example 72 = 2^3 * 3^2.  Exponent on 2 is one of 0, 1, 2, 3.  Exponent on
// 3 is one of 0, 1, 2.  Number of possibilities:  product over i of (mi + 1).
// Factors are:
//
//	2^0 3^0    1
//	2^0 3^1    3
//	2^0 3^2    9
//	2^1 3^0    2
//	2^1 3^1    6
//	2^1 3^2   18
//	2^2 3^0    4
//	2^2 3^1   12
//	2^2 3^2   36
//	2^3 3^0    8
//	2^3 3^1   24
//	2^3 3^2   72

unsigned z2nipoly_num_divisors(
	z2nipoly_factor_info_t * pfinfo)
{
	unsigned rv;
	int i;

	if ((pfinfo->num_distinct < 0)
		|| (pfinfo->num_distinct > Z2NIPOLY_MAX_NUM_FACTORS))
	{
		fprintf(stderr, "z2nipoly_num_divisors:  num distinct (%d) "
			"out of bounds 0-%d\n",
			pfinfo->num_distinct, Z2NIPOLY_MAX_NUM_FACTORS);
		exit(1);
	}

	rv = 1;
	for (i = 0; i < pfinfo->num_distinct; i++)
		rv *= pfinfo->factor_counts[i].count + 1;
	return rv;
}

// ----------------------------------------------------------------
// See comments to the above.  k is treated as a multibase representation
// over the bases mi+1.

z2nipoly_t z2nipoly_kth_divisor(
	z2nipoly_factor_info_t * pfinfo,
	z2ipoly_t r,
	unsigned  k)
{
	z2nipoly_t rv;
	z2nipoly_t fp;
	int        i;
	unsigned   base;
	unsigned   power;

	rv = z2nipoly_from_coeffs0(1LL);
	for (i = 0; i < pfinfo->num_distinct; i++) {
		base = pfinfo->factor_counts[i].count + 1;
		power = k % base;
		k = k / base;

		fp = z2nipoly_power(&pfinfo->factor_counts[i].f, power, r);
		rv = z2nipoly_mul(&rv, &fp, r);
	}
	return rv;
}

// ----------------------------------------------------------------
static int z2nipoly_qcmp(const void * pv1, const void * pv2)
{
	if (z2nipoly_lt((z2nipoly_t*)pv1, (z2nipoly_t*)pv2))
		return -1;
	if (z2nipoly_gt((z2nipoly_t*)pv1, (z2nipoly_t*)pv2))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
// The caller must free the return value.

z2nipoly_t * z2nipoly_get_all_divisors(
	z2nipoly_t  * pz,
	z2ipoly_t     r,
	unsigned    * pnum_divisors)
{
	z2nipoly_factor_info_t finfo;
	unsigned nf, k;
	z2nipoly_t * all_divisors;

	z2nipoly_factor(pz, r, &finfo);
	nf = z2nipoly_num_divisors(&finfo);
	*pnum_divisors = nf;

	all_divisors = (z2nipoly_t *)malloc_check(nf * sizeof(z2nipoly_t));
	for (k = 0; k < nf; k++) {
		all_divisors[k] = z2nipoly_kth_divisor(&finfo, r, k);
	}
	qsort(all_divisors, nf, sizeof(z2nipoly_t), z2nipoly_qcmp);

	return all_divisors;
}

// ----------------------------------------------------------------
// The caller must free the return value.
z2ipoly_t * z2ipoly_get_roots(
	z2ipoly_t z,
	z2ipoly_t r)
{
	z2nipoly_t zx = z2nipoly_extend_z2ipoly(z);
	z2nipoly_factor_info_t finfo;
	int n = z2ipoly_get_degree(z);
	int i;
	z2ipoly_t * rv = malloc_check(n * sizeof(z2ipoly_t));

	z2nipoly_factor(&zx, r, &finfo);

	if (finfo.num_distinct != n) {
		printf("z2ipoly_get_roots: coding error: split failure\n");
		exit(1);
	}

	for (i = 0; i < n; i++) {
		if (finfo.factor_counts[i].count != 1) {
			printf("z2ipoly_get_roots: coding error: "
				"split failure\n");
			exit(1);
		}
		if (finfo.factor_counts[i].f.degree != 1) {
			printf("z2ipoly_get_roots: coding error: "
				"split failure\n");
			exit(1);
		}
		// Z2 is self-negating
		rv[i] = finfo.factor_counts[i].f.coeffs[0];
	}

	return rv;
}

// ----------------------------------------------------------------
// The caller must free the return value.
z2ipoly_t * z2ipolymod_get_elements_frobord(
	z2ipoly_t r,
	int * pnel)
{
	int n = z2ipoly_get_degree(r);
	int nel = 1 << n;
	z2ipoly_t * rv = malloc_check(nel * sizeof(z2ipoly_t));
	z2ipoly_t * dirreds = malloc_check(nel * sizeof(z2ipoly_t));
	z2ipoly_t * roots, f0, fcurr;
	int nirred;
	int i, didx = 0;

	// Factor X^(p^n) - X into irreducible factors over Fp[x].
	z2ipoly_find_dirreds(n, dirreds, &nirred, nel);

	// For each irreducible factor of X^(p^n) - X, get its roots.
	for (i = 0; i < nirred; i++) {
		roots = z2ipoly_get_roots(dirreds[i], r);

		// Get frobit of each root.
		f0 = roots[0];
		fcurr = f0;
		do {
			if (didx >= nel) {
				fprintf(stderr, "FROBORD ABEND!\n");
				exit(1);
			}
			rv[didx++] = fcurr;
			fcurr = z2ipolymod_power(fcurr, 2, r);
		} while (fcurr != f0);
	}

	free(dirreds);
	*pnel = nel;
	return rv;
}
