#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "z2nipolymod.h"

// ----------------------------------------------------------------
// z2nipolymod_mul
// z2nipolymod_rarmul
// z2nipolymod_rermul
//
// Multiplication may be done (at least) two ways:  (1) Take the ring product
// of the two polynomials, then reduce mod r; (2) reduce mod r at every step.
// Testing has shown that the latter method (reduce en route) is faster
// for larger p and smaller n; the former method (ring and reduce) is faster
// for smaller p and larger n.  Furthermore, the ring-and-reduce method
// requires storage space for degree up to 2d-2, where d is the degree of the
// reductor.  E.g. with maximum degree 100, we can't multiply two degree-60
// polynomials using ring-and-reduce, but we can if we use reduce-en-route.

// ----------------------------------------------------------------
// Sample of the ring-and-reduce method:
//
// Use GF(7^4), with reductor 1,0,1,3,5.  Multiply 1,2,3,4 by 2,4,6,1.
//
// First, multiply in the ring GF(7)[x]:
//
//       1 2 3 4
//       2 4 6 1
//      =========
//       1 2 3 4
//     6 5 4 3
//   4 1 5 2
// 2 4 6 1
// ==============
// 2 1 6 5 1 6 4
//
// Second, reduce the product mod the reductor (i.e. divide and take the
// remainder):
//
//                     2 1 4
//           +---------------
// 1 0 1 3 5 | 2 1 6 5 1 6 4
//           | 2 0 2 6 3
//           +---------------
//           |   1 4 6 5 6 4
//           |   1 0 1 3 5
//           +---------------
//           |     4 5 2 1 4
//           |     4 0 4 5 6
//           +---------------
//           |       5 5 3 5

static z2nipoly_t z2nipolymod_rarmul(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2nipoly_t * pr,
	z2ipoly_t    fr)
{
	z2nipoly_t rprod;
	z2nipoly_t quot;
	z2nipoly_t out;

	// Take the polynomial product (degree may be higher than that
	// of the reductor).
	rprod = z2nipoly_mul(pin1, pin2, fr);

	// Reduce the polynomial product modulo the reductor.
	z2nipoly_quot_and_rem(&rprod, pr, &quot, &out, fr);
	return out;
}

// ----------------------------------------------------------------
// Sample of the reduce-en-route method:
//
// Use GF(7^4), with reductor 1,0,1,3,5.  Multiply 1,2,3,4 by 2,4,6,1.
//
//       1 2 3 4
//       2 4 6 1
//      =========
//       1 2 3 4   1,2,3,4 times 1
//
//     6 5 4 3 0   1,2,3,4 times 6x
//     6 0 6 4 2   reduce mod r
//    -----------
//       5 5 6 5
//
//   4 1 5 2 0     1,2,3,4 times 4x
//   4 0 4 5 6     reduce mod r
//  -------------
//     1 1 4 1 0   times x again
//     1 0 1 3 5   reduce mod r
//  -------------
//       1 3 5 2
//
// 2 4 6 1         1,2,3,4 times 2x
// 2 0 2 6 3       reduce mod r
// --------------
//   4 4 2 4       times x again
//   4 0 4 5 6     reduce mod r
// --------------
//     4 5 6 1     times x again
//     4 0 4 5 6   reduce mod r
// --------------
//       5 2 3 1
// ==============
//       1 2 3 4   four partial products
//       5 5 6 5
//       1 3 5 2
//       5 2 3 1
// ==============
//       5 5 3 5   final product is the sum of the four partial products
//
// ----------------------------------------------------------------
// The reduce-en-route method can be optimized a little, to not have
// to shift the coefficients when multiplying a partial product by x.
// Example:
//
//       1 2 3 4
//       2 4 6 1
//      =========
//       1 2 3 4     in1 * in2[0]
//
//
//       6 5 4 3(0)  in1 * in2[1]
//      / / / / /
//     6 0 6 4 2     mul    = in1[3]
//                   in1[3] = in1[2] - red[3] * mul
//                   in1[2] = in1[1] - red[2] * mul
//                   in1[1] = in1[0] - red[1] * mul
//                   in1[0] =     0  - red[0] * mul
//    -----------
//       5 5 6 5
//
//
//       4 1 5 2(0)  in1 * in2[2]
//      / / / / /
//     4 0 4 5 6
//    -----------
//       1 1 4 1(0)
//      / / / / /
//     1 0 1 3 5
//    -----------
//       1 3 5 2
//
//
//       2 4 6 1(0)  in1 * in2[3]
//      / / / / /
//     2 0 2 6 3
//    -----------
//       4 4 2 4(0)
//      / / / / /
//     4 0 4 5 6
//    -----------
//       4 5 6 1(0)
//      / / / / /
//     4 0 4 5 6
//    -----------
//       5 2 3 1
//
//
// ==============
//       1 2 3 4
//       5 5 6 5
//       1 3 5 2
//       5 2 3 1
// ==============
//       5 5 3 5

static z2nipoly_t z2nipolymod_rermul(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2nipoly_t * pr,
	z2ipoly_t    fr)
{
	z2nipoly_t  temp, out;
	int i, j, power, d1;
	z2ipoly_t * p1c, * p2c, * ptc, * prc;
	z2nipoly_t in1red, in2red, rmonic;
	z2ipoly_t elim, prod;

	z2nipoly_mod(&in1red, pin1, fr);
	z2nipoly_mod(&in2red, pin2, fr);
	rmonic = *pr;
	z2nipoly_make_monic(&rmonic, fr);

	p1c = &in1red.coeffs[0];
	p2c = &in2red.coeffs[0];
	ptc = &temp.coeffs[0];
	prc = &rmonic.coeffs[0];
	d1  = pr->degree - 1;

	out.degree = d1;
	z2nipoly_check_degree(out.degree);

	for (i = 0; i <= out.degree; i++)
		out.coeffs[i] = 0LL;

	for (i = 0; i <= pin2->degree; i++) {
		// Multiply in1 by scalar coefficient on current term of in2.
		for (j = 0; j <= in1red.degree; j++)
			ptc[j] = z2ipolymod_mul(p1c[j], p2c[i], fr);
		for (j = in1red.degree + 1; j < pr->degree; j++)
			ptc[j] = 0LL;
		temp.degree = d1;

		// Now need to multiply by the power of x on the current term
		// of in2, reducing mod r en route.
		for (power = 1; power <= i; power++) {
			elim = ptc[d1];
			for (j = d1 - 1; j >= 0; j--) {
				prod = z2ipolymod_mul(prc[j+1], elim, fr);
				ptc[j+1] = z2ipoly_sub(ptc[j], prod);
			}
			//prod = z2ipolymod_mul(prc[0], elim, fr);
			//ptc[0] = z2ipoly_usub(prod, paux->p);
			// Z2[x] is self-negating
			ptc[0] = z2ipolymod_mul(prc[0], elim, fr);
		}
		temp.degree = z2nipoly_find_degree(&temp, temp.degree);
		out = z2nipoly_add(&out, &temp, fr);
	}

	// E.g. if one polynomial was zero.  My convention is that the zero
	// polynomial has "degree" zero -- correct as an array index, but
	// mathematically incorrect.
	out.degree = z2nipoly_find_degree(&out, out.degree);
	return out;
}

z2nipoly_t z2nipolymod_mul(
	z2nipoly_t * pin1,
	z2nipoly_t * pin2,
	z2nipoly_t * pr,
	z2ipoly_t    fr)
{
	// Reduce-en-route allows us to use bigger polys w/o overflow.
	//if ((pin1->degree + pin2->degree) > ZPNPOLY_MAX_DEGREE)
	//if (1)
	if (0)
		return z2nipolymod_rermul(pin1, pin2, pr, fr);
	else
		return z2nipolymod_rarmul(pin1, pin2, pr, fr);
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
z2nipoly_t z2nipolymod_power(
	z2nipoly_t * px,
	unsigned     power,
	z2nipoly_t * pr,
	z2ipoly_t    fr)
{
	z2nipoly_t x2 = *px;
	z2nipoly_t rv = z2nipoly_from_coeffs0(1LL);

	while (power != 0) {
		if (power & 1)
			rv = z2nipolymod_mul(&x2, &rv, pr, fr);
		power >>= 1;
		x2 = z2nipolymod_mul(&x2, &x2, pr, fr);
	}
	return rv;
}
