#include "z2nlinalg.h"
#include "mod.h"
#include "malloc_check.h"

static int find_row_leader_pos(
	z2poly_t * prow,
	int         ncols,
	int       * ppos);

// ----------------------------------------------------------------
z2poly_t ** z2nla_alloc_matrix(
	int  nrows,
	int  ncols)
{
	z2poly_t ** rv;
	int i;

	rv = malloc_check(nrows * sizeof(z2poly_t *));
	for (i = 0; i < nrows; i++)
		rv[i] = malloc_check(ncols * sizeof(z2poly_t));
	return rv;
}

// ----------------------------------------------------------------
void z2nla_free_matrix(
	z2poly_t ** M,
	int    nrows,
	int    ncols)
{
	int i;
	for (i = 0; i < nrows; i++)
		free(M[i]);
	free(M);
}

// ----------------------------------------------------------------
void z2nla_print_matrix(
	z2poly_t ** M,
	int    nrows,
	int    ncols,
	z2poly_t * pr)
{
	int i, j;
	for (i = 0; i < nrows; i++) {
		printf(" ");
#if 1
		for (j = 0; j < ncols; j++)
#else
		for (j = ncols-1; j >= 0; j--)
#endif
		{
			printf(" ");
			z2polymod_print(&M[i][j], pr);
		}
		printf("\n");
	}
}

// ----------------------------------------------------------------
void z2nla_matrix_from_z2npolys(
	z2poly_t ** M,
	int         nrows,
	int         ncols,
	z2npoly_t * polys)
{
	int i;
	for (i = 0; i < nrows; i++)
		z2nla_row_from_z2npoly(M[i], &polys[i], ncols);
}

// ----------------------------------------------------------------
void z2nla_z2npolys_from_matrix(
	z2npoly_t  * polys,
	z2poly_t ** M,
	int        nrows,
	int        ncols)
{
	int i, j;
	for (i = 0; i < nrows; i++) {
		for (j = 0; j < ncols; j++) {
			polys[i].coeffs[j] = M[i][j];
			polys[i].degree = z2npoly_find_degree(
				&polys[i], ncols - 1);
		}
	}
}

// ----------------------------------------------------------------
void z2nla_row_from_z2npoly(
	z2poly_t  * prow,
	z2npoly_t * pzpoly,
	int          ncols)
{
	int i;
	int d = pzpoly->degree;
	z2poly_t zero = z2poly_from_base_rep(0);
	for (i = 0; i <= d; i++)
		prow[i] = pzpoly->coeffs[i];
	for (i = d+1; i < ncols; i++)
		prow[i] = zero;
}

// ----------------------------------------------------------------
void z2nla_transpose(
	z2poly_t ** Mout,
	z2poly_t ** Min,
	int        n)
{
	int row, col;

	for (row = 0; row < n; row++)
		for (col = 0; col < n; col++)
			Mout[row][col] = Min[col][row];
}

// ----------------------------------------------------------------
void z2nla_row_echelon_form(
	z2poly_t ** M,
	z2poly_t  * pr,
	int        nrows,
	int        ncols)
{
	z2nla_row_reduce_below (M, pr, nrows, ncols);
	z2nla_row_reduce_above (M, pr, nrows, ncols);
	z2nla_make_matrix_monic(M, pr, nrows, ncols);
}

// ----------------------------------------------------------------
static int find_row_with_leader_at_col(
	z2poly_t ** M,
	int    nrows,
	int    ncols,
	int    clear_col,
	int    leader_row,
	int  * pswap_row)
{
	int row;
	for (row = leader_row; row < nrows; row++) {
		if (!z2poly_is_zero(&M[row][clear_col])) {
			*pswap_row = row;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
static void swap_rows(
	z2poly_t ** M,
	int    nrows,
	int    ncols,
	int    leader_row,
	int    swap_row)
{
	int col;
	z2poly_t temp;

	for (col = 0; col < ncols; col++) {
		temp               = M[leader_row][col];
		M[leader_row][col] = M[swap_row][col];
		M[swap_row][col]   = temp;
	}
}

// ----------------------------------------------------------------
void z2nla_make_row_monic(
	z2poly_t * row,
	int   ncols,
	int   leader_col,
	z2poly_t * pr)
{
	int col;
	z2poly_t cinv = z2polymod_recip(&row[leader_col], pr);
	for (col = leader_col; col < ncols; col++) {
		row[col] = z2polymod_mul(&row[col], &cinv, pr);
	}
}

// ----------------------------------------------------------------
void z2nla_row_reduce_below(
	z2poly_t ** M,
	z2poly_t  * pr,
	int        nrows,
	int        ncols)
{
	int clear_col, leader_row, swap_row, clear_row;

	for (clear_col = 0, leader_row = 0; clear_col < ncols; clear_col++) {
		if (!find_row_with_leader_at_col(M, nrows, ncols, clear_col,
			leader_row, &swap_row))
		{
			continue;
		}
		if (leader_row != swap_row)
			swap_rows(M, nrows, ncols, leader_row, swap_row);
		if (!z2poly_is_one(&M[leader_row][clear_col]))
			z2nla_make_row_monic(M[leader_row], ncols,
				clear_col, pr);
		for (clear_row = leader_row+1; clear_row < nrows; clear_row++) {
			int col;
			z2poly_t clear_leader, clear_leader_inv;
			clear_leader = M[clear_row][clear_col];
			if (z2poly_is_zero(&clear_leader))
				continue;
			clear_leader_inv = z2polymod_recip(&clear_leader, pr);
			for (col = clear_col; col < ncols; col++) {
				M[clear_row][col] = z2polymod_mul(
					&M[clear_row][col],
					&clear_leader_inv, pr);
				M[clear_row][col] = z2poly_sub(
					&M[clear_row][col],
					&M[leader_row][col]);
			}
		}
		leader_row++;
	}
}

// ----------------------------------------------------------------
// Assumes row_reduce_below has already been done.

void z2nla_row_reduce_above(
	z2poly_t ** M,
	z2poly_t  * pr,
	int        nrows,
	int        ncols)
{
	int row, row2, row2_leader_pos, col;
	z2poly_t row_leader_val, row2_leader_val, mul;

	for (row = 0; row < nrows; row++) {
		for (row2 = row+1; row2 < nrows; row2++) {
			if (!find_row_leader_pos(M[row2], ncols,
				&row2_leader_pos))
			{
				break;
			}
			row2_leader_val = M[row2][row2_leader_pos];
			row_leader_val  = M[row][row2_leader_pos];
			if (z2poly_is_zero(&row_leader_val))
				continue;

			mul = z2polymod_div(&row_leader_val, &row2_leader_val,
				pr);
			for (col = 0; col < ncols; col++) {
				z2poly_t prod = z2polymod_mul(&M[row2][col],
					&mul, pr);
				M[row][col] = z2poly_sub(&M[row][col], &prod);
			}
		}
	}
}

// ----------------------------------------------------------------
// Assumes row_reduce_below has already been done.

void z2nla_make_matrix_monic(
	z2poly_t ** M,
	z2poly_t  * pr,
	int        nrows,
	int        ncols)
{
	int row, row_leader_pos, col;
	z2poly_t row_leader_val;

	for (row = 0; row < nrows; row++) {
		if (!find_row_leader_pos(M[row], ncols, &row_leader_pos))
			break;
		row_leader_val = M[row][row_leader_pos];

		for (col = 0; col < ncols; col++) {
			M[row][col] = z2polymod_div(
				&M[row][col], &row_leader_val, pr);
		}
	}
}

// ----------------------------------------------------------------
static int find_row_leader_pos(
	z2poly_t * prow,
	int        ncols,
	int      * ppos)
{
	int col;

	for (col = 0; col < ncols; col++) {
		if (!z2poly_is_zero(&prow[col])) {
			*ppos = col;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
// Assumes the matrix is already row-reduced.
int z2nla_find_rank(
	z2poly_t ** M,
	int    nrows,
	int    ncols)
{
	int rv = 0;
	int row;

	for (row = 0; row < nrows; row++) {
		if (!z2nla_row_is_zero(M[row], ncols))
			rv++;
	}
	return rv;
}

// ----------------------------------------------------------------
int z2nla_row_is_zero(
	z2poly_t  * prow,
	int          ncols)
{
	int col;
	for (col = 0; col < ncols; col++)
		if (!z2poly_is_zero(&prow[col]))
			return 0;
	return 1;
}

// ----------------------------------------------------------------
// Assumes the matrix is already in row-echelon form.

void z2nla_find_nullspace_basis(
	z2poly_t ** BIT,
	int        nrows,
	int        ncols,
	z2poly_t  * pr,
	z2npoly_t  * nullspace_basis,
	int        dimker)
{
	int i, j;
	unsigned char * free_flags;
	unsigned char * free_indices;
	int nfree = 0; // == dimker but I'll compute it anyway
	int rank = nrows - dimker;
	int dep_pos;
	z2poly_t zero = z2poly_from_base_rep(0);
	z2poly_t one  = z2poly_from_base_rep(1);

	for (i = 0; i < dimker; i++) {
		for (j = 0; j < ncols; j++)
			nullspace_basis[i].coeffs[j] = zero;
		nullspace_basis[i].degree = ncols - 1;
	}

	free_flags   = malloc_check(nrows);
	free_indices = malloc_check(nrows);

	for (i = 0; i < nrows; i++) {
		free_flags[i] = 1;
	}

	for (i = 0; i < rank; i++) {
		if (find_row_leader_pos(BIT[i], ncols, &dep_pos))
			free_flags[dep_pos] = 0;
	}

	for (i = 0; i < nrows; i++) {
		if (free_flags[i])
			free_indices[nfree++] = i;
	}

	// For each free coefficient:
	//   Let that free coefficient be 1 and the rest be zero.
	//   Also set any dependent coefficients which depend on that
	//   free coefficient.
	for (i = 0; i < dimker; i++) {
		nullspace_basis[i].coeffs[free_indices[i]] = one;

		// (B-I)^t in row echelon form:
		//
		// 0210     c0 = ??      c0 = 1  c0 = 0
		// 1000     c1 = -2 c2   c1 = 0  c1 = 5
		// 0000     c2 = ??      c2 = 0  c2 = 1
		// 0000     c3 = 0       c3 = 0  c3 = 0

		// j  = 0,1
		// fi = 0,2

		// i = 0:
		//   j = 0  row 0 fi 0 = row 0 c0 = 0
		//   j = 1  row 1 fi 0 = row 1 c0 = 0
		// i = 1:
		//   j = 0  row 0 fi 1 = row 0 c2 = 2 dep_pos = 1
		//   j = 1  row 1 fi 1 = row 1 c2 = 0

		// 0001
		// 01?0

		for (j = 0; j < rank; j++) {
			if (z2poly_equal(&BIT[j][free_indices[i]], &zero))
				continue;

			(void)find_row_leader_pos(BIT[j], ncols, &dep_pos);
			nullspace_basis[i].coeffs[dep_pos] =
				//z2polymod_sub -- not needed in Z2
				BIT[j][free_indices[i]];
		}
		nullspace_basis[i].degree = z2npoly_find_degree(
			&nullspace_basis[i], ncols - 1);
	}
	free(free_flags);
	free(free_indices);
}
