#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "z2npoly.h"
#include "mod.h"
#include "power.h"
#include "tokenize.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
void z2npoly_check_degree(int degree)
{
	if (degree < 0) {
		fprintf(stderr,
			"z2npoly:  degree must be non-negative; got %d.\n",
			degree);
		exit(1);
	}
	if (degree > Z2NPOLY_MAX_DEGREE) {
		fprintf(stderr,
			"z2npoly:  degree must be <= %d; got %d.\n",
			Z2NPOLY_MAX_DEGREE, degree);
		exit(1);
	}
}

// ----------------------------------------------------------------
z2npoly_t z2npoly_add(
	z2npoly_t * pin1,
	z2npoly_t * pin2,
	z2poly_t  * pr)
{
	z2npoly_t   out;
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++)
			out.coeffs[i] = z2poly_add(&pin1->coeffs[i],
				&pin2->coeffs[i]);
		for (i = pin2->degree + 1; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i];
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++)
			out.coeffs[i] = z2poly_add(&pin1->coeffs[i],
				&pin2->coeffs[i]);
		for (i = pin1->degree + 1; i <= pin2->degree; i++)
			out.coeffs[i] = pin2->coeffs[i];
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were added.
	out.degree = z2npoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
z2npoly_t z2npoly_sub(
	z2npoly_t * pin1,
	z2npoly_t * pin2,
	z2poly_t * pr)
{
	z2npoly_t   out;
	z2poly_t    zero = z2poly_from_base_rep(0);
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++)
			out.coeffs[i] = z2poly_sub(&pin1->coeffs[i],
				&pin2->coeffs[i]);
		for (i = pin2->degree + 1; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i];
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++) {
			out.coeffs[i] = z2poly_sub(&pin1->coeffs[i],
				&pin2->coeffs[i]);
		}
		for (i = pin1->degree + 1; i <= pin2->degree; i++) {
			// out.coeffs[i] = negative of pin2->coeffs[i]
			out.coeffs[i] = z2poly_sub(&zero, &pin2->coeffs[i]);
		}
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were added.
	out.degree = z2npoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
z2npoly_t z2npoly_mul(
	z2npoly_t * pin1,
	z2npoly_t * pin2,
	z2poly_t * pr)
{
	z2npoly_t out;
	int i, j;
	z2poly_t prod;

	out.degree = pin1->degree + pin2->degree;
	z2npoly_check_degree(out.degree);
	for (i = 0; i <= out.degree; i++)
		out.coeffs[i] = z2poly_from_base_rep(0);

	for (i = 0; i <= pin1->degree; i++) {
		for (j = 0; j <= pin2->degree; j++) {
			prod = z2polymod_mul(&pin1->coeffs[i], &pin2->coeffs[j],
				pr);
			out.coeffs[i+j] = z2poly_add(&out.coeffs[i+j], &prod);
		}
	}
	// E.g. if one polynomial was zero.  My convention is that the zero
	// polynomial has "degree" zero -- correct as an array index, but
	// mathematically incorrect.
	out.degree = z2npoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
z2npoly_t z2npoly_power(
	z2npoly_t * pin,
	int       power,
	z2poly_t * pr)
{
	z2npoly_t in2 = *pin;
	z2npoly_t inpower = z2npoly_from_int(1);

	while (power != 0) {
		if (power & 1)
			inpower = z2npoly_mul(&in2, &inpower, pr);
		power = (unsigned)power >> 1;
		in2 = z2npoly_mul(&in2, &in2, pr);
	}
	return inpower;
}

// ----------------------------------------------------------------
// Assumes the input is already known to be a square.

z2npoly_t z2npoly_square_root(
	z2npoly_t * pin,
	z2poly_t * pr)
{
	int si, di;
	z2npoly_t out;
	z2poly_t zero = z2poly_from_base_rep(0);

	for (di = 0; di <= pin->degree; di++) {
		out.coeffs[di] = zero;
	}

	for (si = 0, di = 0; si <= pin->degree; si += 2, di++) {
		out.coeffs[di] = pin->coeffs[si];
		if (si+1 > pin->degree)
			break;
		if (!z2poly_is_zero(&pin->coeffs[si+1])) {
			printf("z2npoly_square_root: ");
			z2npoly_print(pin, pr);
			printf(" is not a square.\n");
			exit(1);
		}
	}
	out.degree = z2npoly_find_degree(&out, pin->degree);

	return out;
}

// ----------------------------------------------------------------
// Input: a, b
// Output: GCD g, and m, n such that am + bn = d

z2npoly_t z2npoly_gcd(
	z2npoly_t * pa, // 1st input.
	z2npoly_t * pb, // 2nd input.
	z2poly_t * pr)
{
	z2npoly_t c, q, r;
	z2npoly_t d;   // Return value.

	if (z2npoly_is_zero(pa))
		return *pb;
	if (z2npoly_is_zero(pb))
		return *pa;

	c = *pa;
	d = *pb;

	while (1) {

		// Divide
		// q = c / d, r = c % d
		z2npoly_quot_and_rem(&c, &d, &q, &r, pr);
		// Note:  now c = qd + r and 0 <= r < d

		// Remainder zero?
		//if (r == zero)
		if (z2npoly_is_zero(&r))
			break;

		// Recycle
		c = d;
		d = r;
	}
	return d;
}

// ----------------------------------------------------------------
int z2npoly_is_zero(z2npoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (!z2poly_is_zero(&pz->coeffs[0]))
		return 0;
	return 1;
}

// ----------------------------------------------------------------
int z2npoly_is_one(z2npoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (!z2poly_is_one(&pz->coeffs[0]))
		return 0;
	return 1;
}

// ----------------------------------------------------------------
void z2npoly_make_monic(
	z2npoly_t * pz,
	z2poly_t * pr)
{
	int i;
	z2poly_t one = z2poly_from_base_rep(1);
	z2poly_t c   = pz->coeffs[pz->degree];
	z2poly_t cinv;

	if (z2poly_is_zero(&c))
		return;
	cinv = z2polymod_div(&one, &c, pr);

	for (i = 0; i <= pz->degree; i++) {
		pz->coeffs[i] = z2polymod_mul(&pz->coeffs[i], &cinv, pr);
	}
}

// ----------------------------------------------------------------
static z2npoly_t z2npoly_smul(
	z2npoly_t * pin, // Poly
	z2poly_t  * pscalar, // Scalar
	z2poly_t * pr)
{
	z2npoly_t  out;
	int i;

	for (i = 0; i <= pin->degree; i++)
		out.coeffs[i] = z2polymod_mul(&pin->coeffs[i], pscalar, pr);

	if (z2poly_is_zero(pscalar))
		out.degree = 0;
	else
		out.degree = pin->degree;
	return out;
}

// ----------------------------------------------------------------
int z2npoly_find_degree(
	z2npoly_t * pin,
	int        max_deg)
{
	int i;

	for (i = max_deg; i >= 0; i--)
		if (!z2poly_is_zero(&pin->coeffs[i]))
			return i;
	return 0;
}

// ----------------------------------------------------------------
// E.g.
// dividend = 1,2,3,4 (1 + 2x + 3x^2 + 4x^3)
// divisor  = 1,1,2   (1 +  x + 2x^2)
// modulus = 7
//
//             q=4,2   r = 4,3
//        +----------
// 1,1,2  |  1,2,3,4
//        |    2,2,4 shift = 1.  4/2 mod 7 = 2.  1,1,2 * 2 = 2,2,4.
//        +----------
//        |  1 0 1
//        |  4 4 1   shift = 0.  1/2 mod 7 = 4   1,1,2 * 4 = 4,4,1
//        +----------
//        |  4 3

void z2npoly_quot_and_rem(
	z2npoly_t * pdividend,
	z2npoly_t * pdivisor,
	z2npoly_t * pquot,
	z2npoly_t * prem,
	z2poly_t  * pr)
{
	z2npoly_t quot;
	z2npoly_t rem;
	int i;
	int max_shift;
	int shift;
	z2poly_t rem_leader;
	z2poly_t divisor_leader;
	z2poly_t dlinv;
	z2poly_t multiplier;
	z2npoly_t divisor_with_mul;
	int stop;

	max_shift = pdividend->degree - pdivisor->degree;
	if (pdividend->degree < pdivisor->degree) {
		quot.coeffs[0] = z2poly_from_base_rep(0);
		quot.degree = 0;
		rem = *pdividend;

		*pquot = quot;
		*prem  = rem;
		return;
	}

	quot.degree = pdividend->degree - pdivisor->degree;
	for (i = 0; i <= quot.degree; i++)
		quot.coeffs[i] = z2poly_from_base_rep(0);
	rem = *pdividend;

	divisor_leader = pdivisor->coeffs[pdivisor->degree];
	dlinv = z2polymod_recip(&divisor_leader, pr);
	for (shift = max_shift; shift >= 0; shift--) {
		rem_leader = rem.coeffs[shift + pdivisor->degree];
		multiplier = z2polymod_mul(&rem_leader, &dlinv, pr);
		divisor_with_mul = z2npoly_smul(pdivisor, &multiplier, pr);
		quot.coeffs[shift] = multiplier;
		stop = shift + divisor_with_mul.degree;
		for (i = shift; i <= stop; i++) {
			rem.coeffs[i] = z2poly_sub(
				&rem.coeffs[i],
				&divisor_with_mul.coeffs[i-shift]);
		}
	}

	quot.degree = z2npoly_find_degree(&quot, quot.degree);
	rem.degree  = z2npoly_find_degree(&rem,  rem.degree);

	*pquot = quot;
	*prem  = rem;
}

// ----------------------------------------------------------------
z2npoly_t z2npoly_div(
	z2npoly_t * pin1,
	z2npoly_t * pin2,
	z2poly_t * pr)
{
	z2npoly_t quot, rem;
	z2npoly_quot_and_rem(pin1, pin2, &quot, &rem, pr);
	return quot;
}

// ----------------------------------------------------------------
z2npoly_t z2npoly_mod(
	z2npoly_t * pin1,
	z2npoly_t * pin2,
	z2poly_t * pr)
{
	z2npoly_t quot, rem;
	z2npoly_quot_and_rem(pin1, pin2, &quot, &rem, pr);
	return rem;
}

// ----------------------------------------------------------------
// Example with p=2 and hexadecimal notation:
//
//   mod 13:
//   02 =  0 . . . [  0]
//   03 =  1 . . . [  1]
//   07 =  6 7 . . [  3]
//   13 =  2 3 4 5 [ 15]
//   19 =  9 b d e [ 15]
//   1f =  8 a c f [  5]
//
// 13 is the reductor r.  b is the input z.  19 is the desired minimal
// polynomial.  Fact:  If b is a root of 19, so are all the elements
// of its Frobenius orbit mod r, e.g. b^2, b^4, b^8.  These powers yield
// b, 9, d and e.  Then, (X-b)(X-9)(X-d)(X-e) is the desired minimal
// polynomial with coefficients in Fpn[x], 11001 .  Un-extend to Fp[x]
// to get 19 (in hex).

//         1 b
// *       1 9
// ===========
//         9 c
// +     1 b
// ===========
//       1 2 c
// *       1 d
// ===========
//       d 9 3
// +   1 2 c
// ===========
//     1 f 5 3
// *       1 e
// ===========
//     e 5 3 1
// + 1 f 5 3
// ===========
//   1 1 0 0 1 = 19 in hex.

z2poly_t z2poly_get_min_poly(
	z2poly_t * pz,
	z2poly_t * pr)
{
	int n = z2poly_get_degree(pr);
	int i;
	z2poly_t one = z2poly_from_base_rep(1);
	z2npoly_t extrv = z2npoly_extend_z2poly(&one);
	z2poly_t z2power = *pz;
	z2npoly_t factor;
	z2poly_t rv;

	for (i = 0; i < n; i++) {
		factor.degree = 1;
		factor.coeffs[1] = z2poly_from_base_rep(1);
		//factor.coeffs[0] = z2poly_usub(&z2power);
		//Z2 self-negates
		factor.coeffs[0] = z2power;
		extrv = z2npoly_mul(&extrv, &factor, pr);
		z2power = z2polymod_power(&z2power, 2, pr);
		if (z2poly_equal(&z2power, pz))
			break;
	}
	rv = z2poly_unextend_z2npoly(&extrv, pr);
	return rv;
}

// ----------------------------------------------------------------
int z2npoly_equal(
	z2npoly_t * pin1,
	z2npoly_t * pin2,
	z2poly_t * pr)
{
	int i;
	if (pin1->degree != pin2->degree)
		return 0;
	for (i = 0; i <= pin1->degree; i++)
		if (!z2poly_equal(&pin1->coeffs[i], &pin2->coeffs[i]))
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int z2npoly_lt(z2npoly_t * pa, z2npoly_t * pb)
{
	return z2npoly_cmp(pa, CMP_LT, pb);
}
int z2npoly_le(z2npoly_t * pa, z2npoly_t * pb)
{
	return z2npoly_cmp(pa, CMP_LE, pb);
}
int z2npoly_gt(z2npoly_t * pa, z2npoly_t * pb)
{
	return z2npoly_cmp(pa, CMP_GT, pb);
}
int z2npoly_ge(z2npoly_t * pa, z2npoly_t * pb)
{
	return z2npoly_cmp(pa, CMP_GE, pb);
}
int z2npoly_eq(z2npoly_t * pa, z2npoly_t * pb)
{
	return z2npoly_cmp(pa, CMP_EQ, pb);
}
int z2npoly_ne(z2npoly_t * pa, z2npoly_t * pb)
{
	return z2npoly_cmp(pa, CMP_NE, pb);
}

// ----------------------------------------------------------------
int z2npoly_cmp(
	z2npoly_t * pa,
	int         cmp,
	z2npoly_t * pb)
{
	int direction = 0; // -1 = less, 0 = equal, +1 = greater;
	int i;

	if (pa->degree > pb->degree) {
		direction =  1;
	}
	else if (pa->degree < pb->degree) {
		direction = -1;
	}
	else {
		for (i = pa->degree; i >= 0; i--) {
			if (z2poly_lt(&pa->coeffs[i], &pb->coeffs[i])) {
				direction = -1;
				break;
			}
			else if (z2poly_gt(&pa->coeffs[i], &pb->coeffs[i])) {
				direction =  1;
				break;
			}
		}
	}

	switch (cmp) {
	case CMP_LE: return (direction <= 0); break;
	case CMP_LT: return (direction <  0); break;
	case CMP_EQ: return (direction == 0); break;
	case CMP_NE: return (direction != 0); break;
	case CMP_GE: return (direction >= 0); break;
	case CMP_GT: return (direction >  0); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); return 0; break;
	}
}

// ----------------------------------------------------------------
z2npoly_t z2npoly_extend_z2poly(
	z2poly_t   * pz2)
{
	int i;
	z2npoly_t rv;

	rv.degree = z2poly_get_degree(pz2);
	for (i = 0; i <= rv.degree; i++) {
		if (z2poly_bit_at(pz2, i))
			rv.coeffs[i] = z2poly_from_base_rep(1);
		else
			rv.coeffs[i] = z2poly_from_base_rep(0);
	}
	return rv;
}

// ----------------------------------------------------------------
z2poly_t z2poly_unextend_z2npoly(
	z2npoly_t * pznp,
	z2poly_t * pr)
{
	int i;
	z2poly_t rv = z2poly_from_base_rep(0);

	for (i = 0; i <= pznp->degree; i++) {
		if (z2poly_get_degree(&pznp->coeffs[i]) > 0) {
			printf("z2npoly:  Can't unextend ");
			z2npoly_print(pznp, pr);
			printf("\n");
			exit(1);
		}
		if (z2poly_is_one(&pznp->coeffs[i]))
			z2poly_set_bit(&rv, i);
	}

	return rv;
}

// ----------------------------------------------------------------
// Example:
// * a_red   = a^2 + a + 1 = 0x07 which defines an F_2^2.
// * b_red   = b^6 + b + 1 = 0x43 which defines an F_2^6.
// * ab_poly = x^3 + x^2 + (a+1) x + a = 1:1:3:2
//   which is a cubic polynomial in F_2^2[x].
// * ab_exp  = 21 since (2^6-1) / (2^2-1) = 21.
//
// This function substitutes x with b and a with b^exp, then reduces mod
// b_red.  The above example ab_poly reduces to 0 mod b^6 + b + 1.

z2poly_t z2npoly_flatten(
	z2npoly_t * pab_poly,
	z2poly_t  * pa_red,
	z2poly_t  * pb_red,
	int         ab_exp)
{
	z2poly_t flat_poly = z2poly_from_base_rep(0);
	int in_deg, j, i, out_deg;
	z2poly_t term;
	z2poly_t b = z2poly_from_base_rep(2);

	for (j = 0; j <= pab_poly->degree; j++) {
		in_deg = z2poly_get_degree(&pab_poly->coeffs[j]);
		for (i = 0; i <= in_deg; i++) {
			if (z2poly_bit_at(&pab_poly->coeffs[j], i)) {
				out_deg = j + ab_exp * i;

				//z2poly_check_degree(out_deg);
				//z2poly_toggle_bit(&flat_poly, out_deg);
				term = z2polymod_power(&b, out_deg, pb_red);
				flat_poly = z2poly_add(&flat_poly, &term);
			}
		}
	}
	flat_poly = z2poly_mod(&flat_poly, pb_red);

	return flat_poly;
}

// ----------------------------------------------------------------
z2npoly_t z2npoly_from_coeffs0(
	z2poly_t * pcoeff0)
{
	z2npoly_t rv;
	rv.degree = 0;
	rv.coeffs[0] = *pcoeff0;
	return rv;
}

// ----------------------------------------------------------------
z2npoly_t z2npoly_from_coeffs1(
	z2poly_t * pcoeff1,
	z2poly_t * pcoeff0)
{
	z2npoly_t rv;
	rv.coeffs[1] = *pcoeff1;
	rv.coeffs[0] = *pcoeff0;
	rv.degree = z2npoly_find_degree(&rv, 1);
	return rv;
}

// ----------------------------------------------------------------
z2npoly_t z2npoly_from_int(
	int a)
{
	z2npoly_t rv;
	rv.degree = 0;
	rv.coeffs[0] = z2poly_from_base_rep(a);
	return rv;
}

// ----------------------------------------------------------------
z2npoly_t z2npoly_from_base_rep(
	int rep,
	int intermediate_degree)
{
	int mask;
	z2npoly_t rv;

	if (rep < 0) {
		fprintf(stderr, "z2npoly_from_base_rep: negative input %d.\n",
			rep);
		exit(1);
	}

	if (intermediate_degree < 1) {
		fprintf(stderr,
			"z2npoly_from_base_rep: need deg >= 1; got %d.\n",
			intermediate_degree);
		exit(1);
	}
	mask = (1 << intermediate_degree) - 1;

	rv.degree = 0;
	rv.coeffs[0] = z2poly_from_base_rep(0);
	while (rep) {
		rv.coeffs[rv.degree] = z2poly_from_base_rep(rep & mask);
		rep >>= intermediate_degree;
		if (rv.degree >= Z2NPOLY_MAX_DEGREE) {
			fprintf(stderr, "z2npoly_from_base_rep: overflow.\n");
			exit(1);
		}
		rv.degree++;
	}
	if (rv.degree > 0)
		rv.degree--;

	return rv;
}

// ----------------------------------------------------------------
z2npoly_t z2npoly_from_two_ints(
	int a1,
	int a0)
{
	z2npoly_t rv;
	rv.coeffs[1] = z2poly_from_base_rep(a1);
	rv.coeffs[0] = z2poly_from_base_rep(a0);
	rv.degree = z2npoly_find_degree(&rv, 1);
	return rv;
}

// ----------------------------------------------------------------
z2npoly_t z2npoly_deriv(
	z2npoly_t * pz,
	z2poly_t * pr)
{
	z2npoly_t rv;
	int i;
	z2poly_t zero = z2poly_from_base_rep(0);

	if (pz->degree == 0) {
		return z2npoly_from_int(0);
	}

	for (i = 1; i <= pz->degree; i++) {
		if (1 & 1)
			rv.coeffs[i-1] = pz->coeffs[i];
		else
			rv.coeffs[i-1] = zero;
	}
	rv.degree = z2npoly_find_degree(&rv, pz->degree - 1);

	return rv;
}

// ----------------------------------------------------------------
void z2npoly_print(
	z2npoly_t * pin,
	z2poly_t  * pr)
{
	int i;

	for (i = pin->degree; i >= 0; i--) {
		if (i < pin->degree)
			printf(":");
		z2polymod_print(&pin->coeffs[i], pr);
	}
}

// ----------------------------------------------------------------
int z2npoly_list_scan(
	z2npoly_t * pout,
	char      * pin,
	z2poly_t  * pr)
{
	int    i;
	char * coeff_stringv[Z2NPOLY_MAX_DEGREE + 1];
	char   temp_line[1024];

	pout->degree = 0;
	for (i = 0; i <= Z2NPOLY_MAX_DEGREE; i++)
		pout->coeffs[i] = z2poly_from_base_rep(0);

	strcpy(temp_line, pin);
	pout->degree = tokenize(temp_line, ":", coeff_stringv,
		Z2NPOLY_MAX_DEGREE + 1) - 1;
	if (pout->degree < 0) {
		fprintf(stderr, "z2npoly_list_scan:  Couldn't scan \"%s\".\n",
			pin);
		return 0;
	}

	for (i = 0; i <= pout->degree; i++) {
		if (!z2poly_scan(&pout->coeffs[pout->degree - i],
			coeff_stringv[i]))
		{
			fprintf(stderr,
				"z2npoly:  Couldn't scan coefficient \"%s\".\n",
				coeff_stringv[pout->degree - i]);
			return 0;
		}
	}

	// Check for leading zero coefficients, which would reduce the
	// degree.
	pout->degree = z2npoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
int z2npoly_exp_scan(
	z2npoly_t * pout,
	char      * pin,
	z2poly_t  * pr)
{
	int    i;
	int    num_pieces;
	char * coeff_stringv[Z2NPOLY_MAX_DEGREE + 1];
	char * coeff_and_exp[2];
	int    num_sub_pieces;
	char   temp_line[1024];
	char   temp_piece[128];
	int    exp;

	pout->degree = 0;
	for (i = 0; i <= Z2NPOLY_MAX_DEGREE; i++)
		pout->coeffs[i] = z2poly_from_base_rep(0);

	strcpy(temp_line, pin);
	num_pieces = tokenize(temp_line, ":", coeff_stringv,
		Z2NPOLY_MAX_DEGREE + 1);
	if (num_pieces <= 0) {
		fprintf(stderr, "z2npoly_exp_scan:  Couldn't scan \"%s\".\n",
			pin);
		return 0;
	}

	for (i = 0; i < num_pieces; i++) {
		// Split on "^"; expect z2poly on left and int on right.
		// Bounds-check the exponent.
		strcpy(temp_piece, coeff_stringv[i]);
		num_sub_pieces = tokenize(temp_piece, "^", coeff_and_exp, 2);
		if (num_sub_pieces < 1) {
			fprintf(stderr,
				"z2npoly:  Couldn't scan coefficient \"%s\".\n",
				temp_piece);
			return 0;
		}
		else if (num_sub_pieces == 1) {
			exp = 0;
		}
		else {
			// sscanf and bounds-check the exp
			if (sscanf(coeff_and_exp[1], "%d", &exp) != 1) {
				fprintf(stderr,
				"z2npoly:  Couldn't scan coefficient \"%s\".\n",
					temp_piece);
				return 0;
			}
			if ((exp < 0) || (exp > Z2NPOLY_MAX_DEGREE)) {
				fprintf(stderr,
				"z2npoly:  Exponent %d out of range 0-%d.\n",
					exp, Z2NPOLY_MAX_DEGREE);
				return 0;
			}
		}

		// Remember highest exponent encountered.
		if (exp > pout->degree)
			pout->degree = exp;

		// scan and bounds-check the coeff
		if (!z2poly_scan(&pout->coeffs[exp], coeff_and_exp[0])) {
			fprintf(stderr,
				"z2npoly:  Couldn't scan coefficient \"%s\".\n",
				temp_piece);
			return 0;
		}
	}

	// Check for leading zero coefficients, which would reduce the
	// degree.
	pout->degree = z2npoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
int z2npoly_scan(
	z2npoly_t * pout,
	char      * pin,
	z2poly_t  * pr)
{
	if (strchr(pin, '^'))
		return z2npoly_exp_scan(pout, pin, pr);
	else
		return z2npoly_list_scan(pout, pin, pr);
}
