#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2poly.h"
#include "factorize.h"
#include "log2.h"
#include "count_bits.h"

static void complain_about_overflow(
	int deg);

//static void z2poly_clear_bit(
//	z2poly_t * pa,
//	int        pos);

//static void z2poly_bounds_check_bit(
//	int        pos);

// ----------------------------------------------------------------
z2poly_t z2poly_add(
	z2poly_t * pin1,
	z2poly_t * pin2)
{
	z2poly_t out;
	int i;
	for (i = 0; i < Z2POLY_NUM_PARTS; i++)
		out.parts[i] = pin1->parts[i] ^ pin2->parts[i];
	return out;
}

// ----------------------------------------------------------------
z2poly_t z2poly_sub(
	z2poly_t * pin1,
	z2poly_t * pin2)
{
	z2poly_t out;
	int i;
	for (i = 0; i < Z2POLY_NUM_PARTS; i++)
		out.parts[i] = pin1->parts[i] ^ pin2->parts[i];
	return out;
}

// ----------------------------------------------------------------
z2poly_t z2poly_mul(
	z2poly_t * pa,
	z2poly_t * pb)
{
	int adeg, bdeg, cdeg;
	z2poly_t ashift = *pa;
	z2poly_t out = z2poly_from_base_rep(0);
	int j, k;

	adeg = z2poly_get_degree(pa);
	bdeg = z2poly_get_degree(pb);
	cdeg = adeg + bdeg;

	if (cdeg > Z2POLY_MAX_DEGREE)
		complain_about_overflow(cdeg);

	for (j = 0; j <= bdeg; j++) {
		if (z2poly_bit_at(pb, j)) {
			for (k = 0; k < Z2POLY_NUM_PARTS; k++)
				out.parts[k] ^= ashift.parts[k];
		}
		z2poly_promote_1(&ashift);
	}
	return out;
}

// ----------------------------------------------------------------
z2poly_t z2poly_div(
	z2poly_t * pin1,
	z2poly_t * pin2)
{
	z2poly_t quot, rem;
	z2poly_quot_and_rem(pin1, pin2, &quot, &rem);
	return quot;
}

// ----------------------------------------------------------------
z2poly_t z2poly_mod(
	z2poly_t * pin1,
	z2poly_t * pin2)
{
	z2poly_t quot, rem;
	z2poly_quot_and_rem(pin1, pin2, &quot, &rem);
	return rem;
}

// ----------------------------------------------------------------
void z2poly_quot_and_rem(
	z2poly_t * pdividend,
	z2poly_t * pdivisor,
	z2poly_t * pquot,
	z2poly_t * prem)
{
	int dividend_l1_pos;
	int divisor_l1_pos;
	int l1_diff;
	z2poly_t shift_divisor, quot, rem;
	int check_pos;
	int quot_pos;

	quot = z2poly_from_base_rep(0);
	rem  = z2poly_from_base_rep(0);

	if (z2poly_equals_unsigned(pdividend, 0)) {
		*pquot = quot;
		*prem  = rem;
		return;
	}

	dividend_l1_pos = z2poly_get_degree(pdividend);
	if (z2poly_equals_unsigned(pdivisor, 0)) {
		// Divisor is zero.
		fprintf(stderr, "z2poly_quot_and_rem:  Divide by zero.\n");
		exit(1);
	}
	divisor_l1_pos = z2poly_get_degree(pdivisor);
	l1_diff = dividend_l1_pos - divisor_l1_pos;

	if (l1_diff < 0) {
		// Dividend has lower degree than divisor.
		*pquot = quot;
		*prem = *pdividend;
		return;
	}

	shift_divisor = z2poly_promote_n(pdivisor, l1_diff);

	rem = *pdividend;
	for (
		check_pos = dividend_l1_pos, quot_pos = l1_diff;
		check_pos >= divisor_l1_pos;
		check_pos--, quot_pos--)
	{
		if (z2poly_bit_at(&rem, check_pos)) {
			rem = z2poly_sub(&rem, &shift_divisor);
			z2poly_set_bit(&quot, quot_pos);
		}
		z2poly_demote_1(&shift_divisor);
	}
	*pquot = quot;
	*prem  = rem;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
z2poly_t z2poly_power(
	z2poly_t * px,
	unsigned   power)
{
	z2poly_t x2 = *px;
	z2poly_t rv;

	rv = z2poly_from_base_rep(1);

	while (power != 0) {
		if (power & 1)
			rv = z2poly_mul(&x2, &rv);
		power >>= 1;

		if (power == 0)
			break;
		x2 = z2poly_mul(&x2, &x2);
	}
	return rv;
}

// ----------------------------------------------------------------
// Assumes the input is already known to be a square.  Relies on the
// fact that f(x^p) = f^p(x) over Fp[x].
//
// in  = a4 x^4 + a2 x^2 + a0
// out = a4 x^2 + a2 x   + a0

z2poly_t z2poly_square_root(
	z2poly_t * pin)
{
	int si, di;
	z2poly_t out = z2poly_from_base_rep(0);

	for (si = 0, di = 0; si < Z2POLY_NUM_BITS; si+=2, di++) {
		if (z2poly_bit_at(pin, si))
			z2poly_set_bit(&out, di);
		if (z2poly_bit_at(pin, si + 1)) {
			printf("z2poly_square_root: ");
			z2poly_print(pin);
			printf(" is not square.\n");
			printf("\n");
			exit(1);
		}
	}

	return out;
}

// ----------------------------------------------------------------
z2poly_t z2poly_gcd(
	z2poly_t * pa,
	z2poly_t * pb)
{
	z2poly_t c, q, r, d;

	if (z2poly_equals_unsigned(pa, 0))
		return *pb;
	if (z2poly_equals_unsigned(pb, 0))
		return *pa;

	// Initialize
	c = *pa;
	d = *pb;

	while (1) {
		// Divide
		// q = c / d, r = c % d
		z2poly_quot_and_rem(&c, &d, &q, &r);
		// Note:  now c = qd + r and 0 <= r < d
		// Remainder zero?

		if (z2poly_equals_unsigned(&r, 0)) {
			break;
		}

		// Recycle
		c = d;
		d = r;
	}
	return d;
}

// ----------------------------------------------------------------
// Input: a, b
// Output: GCD g, and m, n such that am + bn = d

// xxx describe Blankinship alg'm here.

z2poly_t z2poly_ext_gcd(
	z2poly_t * pa,
	z2poly_t * pb,
	z2poly_t * pm,
	z2poly_t * pn)
{
	z2poly_t m, n, mprime, nprime, c, q, r, t, d, prod;

	// Initialize
	mprime = z2poly_from_base_rep(1);
	n      = z2poly_from_base_rep(1);
	m      = z2poly_from_base_rep(0);
	nprime = z2poly_from_base_rep(0);
	c      = *pa;
	d      = *pb;

	while (1) {

		// Divide
		// q = c / d, r = c % d
		z2poly_quot_and_rem(&c, &d, &q, &r);
		// Note:  now c = qd + r and 0 <= r < d

		// Remainder zero?
		if (z2poly_equals_unsigned(&r, 0)) {
			break;
		}

		// Recycle
		c = d;
		d = r;

		t = mprime;
		mprime = m;
		// m = t - q*m;
		prod = z2poly_mul(&q, &m);
		m = z2poly_sub(&t, &prod);

		t = nprime;
		nprime = n;
		// n = t - qn;
		prod = z2poly_mul(&q, &n);
		n = z2poly_sub(&t, &prod);
	}
	*pm = m;
	*pn = n;
	return d;
}

// ----------------------------------------------------------------
int z2poly_equal(
	z2poly_t * pa,
	z2poly_t * pb)
{
	int i;
	for (i = 0; i < Z2POLY_NUM_PARTS; i++)
		if (pa->parts[i] != pb->parts[i])
			return 0;
	return 1;
}

int  z2poly_lt(z2poly_t * pa, z2poly_t * pb)
{
	return z2poly_cmp(pa, CMP_LT, pb);
}
int  z2poly_le(z2poly_t * pa, z2poly_t * pb)
{
	return z2poly_cmp(pa, CMP_LE, pb);
}
int  z2poly_gt(z2poly_t * pa, z2poly_t * pb)
{
	return z2poly_cmp(pa, CMP_GT, pb);
}
int  z2poly_ge(z2poly_t * pa, z2poly_t * pb)
{
	return z2poly_cmp(pa, CMP_GE, pb);
}
int  z2poly_eq(z2poly_t * pa, z2poly_t * pb)
{
	return z2poly_cmp(pa, CMP_EQ, pb);
}
int  z2poly_ne(z2poly_t * pa, z2poly_t * pb)
{
	return z2poly_cmp(pa, CMP_NE, pb);
}

// ----------------------------------------------------------------
int z2poly_cmp(
	z2poly_t * pa,
	int        cmp,
	z2poly_t * pb)
{
	int direction = 0; // -1 = less, 0 = equal, +1 = greater;
	int i;

	for (i = Z2POLY_NUM_PARTS-1; i >= 0; i--) {
		if (pa->parts[i] < pb->parts[i]) {
			direction = -1;
			break;
		}
		else if (pa->parts[i] > pb->parts[i]) {
			direction =  1;
			break;
		}
	}

	switch (cmp) {
	case CMP_LE: return (direction <= 0); break;
	case CMP_LT: return (direction <  0); break;
	case CMP_EQ: return (direction == 0); break;
	case CMP_NE: return (direction != 0); break;
	case CMP_GE: return (direction >= 0); break;
	case CMP_GT: return (direction >  0); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); return 0; break;
	}
}

// ----------------------------------------------------------------
z2poly_t z2poly_from_base_rep(
	unsigned val)
{
	z2poly_t rv;
	int i;
	for (i = 1; i < Z2POLY_NUM_PARTS; i++)
		rv.parts[i] = 0;
	rv.parts[0] = val;
	return rv;
}

// ----------------------------------------------------------------
z2poly_t z2poly_from_base_rep_64(
	unsigned long long val)
{
	z2poly_t rv;
	int i;
	for (i = 2; i < Z2POLY_NUM_PARTS; i++)
		rv.parts[i] = 0;
	rv.parts[1] = val >> 32;
	rv.parts[0] = val;
	return rv;
}

// ----------------------------------------------------------------
int z2poly_equals_unsigned(
	z2poly_t * pa,
	unsigned   val)
{
	int i;
	for (i = 1; i < Z2POLY_NUM_PARTS; i++)
		if (pa->parts[i] != 0)
			return 0;
	return (val == pa->parts[0]);
}

// ----------------------------------------------------------------
int z2poly_is_zero(
	z2poly_t * pa)
{
	return z2poly_equals_unsigned(pa, 0);
}

// ----------------------------------------------------------------
int z2poly_is_one(
	z2poly_t * pa)
{
	return z2poly_equals_unsigned(pa, 1);
}

// ----------------------------------------------------------------
// +---+---+---+---+
// |   |   |   |   | Array of 16-bit integers
// +---+---+---+---+
// +-------+-------+
// |       |       | Array of 32-bit integers
// +-------+-------+

// +---+---+---+---+
// |   |   |   |   |
// +---+---+---+---+
// +-------+-------+-------+
// |       |       |       |
// +-------+-------+-------+

// +---+---+---+---+---+---+
// |   |   |   |   |   |   |
// +---+---+---+---+---+---+
// +-------+-------+
// |       |       |
// +-------+-------+

z2poly_t z2poly_from_ubig(
	ubig * pu)
{
	z2poly_t rv;
	int srci, dsti;

	for (
		srci  = 0, dsti = 0;
		srci < UBIG_NUM_PARTS - 1 && dsti < Z2POLY_NUM_PARTS;
		srci += 2, dsti++)
	{
		rv.parts[dsti] = pu->parts[srci] |
			((unsigned)pu->parts[srci+1] << 16);
	}

	for ( ; dsti < Z2POLY_NUM_PARTS; dsti++)
		rv.parts[dsti] = 0;

	return rv;
}

// ----------------------------------------------------------------
void z2poly_print(
	z2poly_t * pa)
{
	int i;
	int leader_pos = 0;

	for (i = Z2POLY_NUM_PARTS - 1; i >= 0; i--) {
		if ((pa->parts[i]) || (i == 0)) {
			leader_pos = i;
			break;
		}
	}

	for (i = Z2POLY_NUM_PARTS - 1; i >= 0; i--) {
		if (i > leader_pos)
			;
		else if (i == leader_pos)
			printf("%x", pa->parts[i]);
		else
			printf("_%08x", pa->parts[i]);
	}

}

// ----------------------------------------------------------------
void z2poly_pretty_print(
	z2poly_t * pa)
{
	int d = z2poly_get_degree(pa);
	int i;
	int got_first = 0;

	if (z2poly_equals_unsigned(pa, 0)) {
		printf("0");
		return;
	}

	for (i = d; i >= 0; i--) {
		if (z2poly_bit_at(pa, i)) {
			if (got_first)
				printf("+");
			if (i == 1)
				printf("x");
			else if (i == 0)
				printf("1");
			else
				printf("x^%d", i);
			got_first = 1;
		}
	}
}

// ----------------------------------------------------------------
int z2poly_scan(
	z2poly_t * pz,
	char     * pin)
{
	char   nybbles[Z2POLY_NUM_NYBBLES+1];
	char   zf_nybbles[Z2POLY_NUM_NYBBLES+1];
	char * pn;
	int    num_nybbles_recvd;
	int    num_zf;

	int    srci;
	int    dsti;
	int    parti;
	char   part_nybbles[Z2POLY_NYBBLES_PER_PART+1];

	memset(nybbles, 0, sizeof(nybbles));
	*pz = z2poly_from_base_rep(0);

	num_nybbles_recvd = 0;
	for (pn = nybbles; *pin; pin++) {
		if (num_nybbles_recvd > Z2POLY_NUM_NYBBLES) {
			nybbles[Z2POLY_NUM_NYBBLES] = 0;
			fprintf(stderr, "z2poly:  input overflow at \"%s\".\n",
				nybbles);
			exit(1);
		}
		if (*pin != '_') {
			*pn = *pin;
			pn++;
			num_nybbles_recvd++;
		}
	}
	*pn = 0;

	num_zf = Z2POLY_NUM_NYBBLES - num_nybbles_recvd;
	for (dsti = 0; dsti < num_zf; dsti++)
		zf_nybbles[dsti] = '0';
	for (srci = 0, dsti = num_zf; dsti < Z2POLY_NUM_NYBBLES; srci++, dsti++)
		zf_nybbles[dsti] = nybbles[srci];
	zf_nybbles[dsti] = 0;

	for (srci = 0, parti = Z2POLY_NUM_PARTS-1; parti >= 0; parti--) {
		for (dsti = 0; dsti < Z2POLY_NYBBLES_PER_PART; dsti++, srci++) {
			part_nybbles[dsti] = zf_nybbles[srci];
		}
		part_nybbles[dsti] = 0;
		(void)sscanf(part_nybbles, "%x", &pz->parts[parti]);
	}

	return 1;
}

// ----------------------------------------------------------------
int z2poly_eval(
	z2poly_t * pa,
	int        x)
{
	if (x & 1)
		return z2poly_count_bits(pa);
	else
		return pa->parts[0] & 1;
}

// ----------------------------------------------------------------
z2poly_t z2poly_compose(
	z2poly_t * pr,
	z2poly_t * pz)
{
	z2poly_t rz = z2poly_from_base_rep(0);
	z2poly_t zp = z2poly_from_base_rep(1);
	int n = z2poly_get_degree(pr);
	int d;

	for (d = 0; d <= n; d++) {
		if (z2poly_bit_at(pr, d))
			rz = z2poly_add(&rz, &zp);
		zp = z2poly_mul(&zp, pz);
	}

	return rz;
}

// ----------------------------------------------------------------
z2poly_t z2poly_deriv(
	z2poly_t * pz)
{
	int i;
	z2poly_t rv = *pz;
	z2poly_demote_1(&rv);
	for (i = 0; i < Z2POLY_NUM_PARTS; i++)
		rv.parts[i] &= 0x55555555;
	return rv;
}

// ----------------------------------------------------------------
int z2poly_is_rootless(
	z2poly_t * pa)
{
	if (z2poly_eval(pa, 0) == 0)
		return 0;
	if (z2poly_eval(pa, 1) == 0)
		return 0;
	return 1;
}

// ----------------------------------------------------------------
int z2poly_count_bits(
	z2poly_t * pa)
{
	return 1 & count_one_bits((unsigned char *)pa, sizeof(*pa));
}

// ----------------------------------------------------------------
int z2poly_get_degree(
	z2poly_t * pa)
{
	int j;

	for (j = Z2POLY_NUM_PARTS - 1; j >= 0; j--) {
		if (pa->parts[j])
			return find_msb_32(pa->parts[j])
				+ (j << Z2POLY_PART_LOG);
	}
	return 0;
}

// ----------------------------------------------------------------
void z2poly_check_degree(
	int degree)
{
	if (degree < 0) {
		fprintf(stderr,
			"z2poly:  degree must be non-negative; got %d.\n",
			degree);
		exit(1);
	}
	if (degree > Z2POLY_MAX_DEGREE) {
		fprintf(stderr,
			"z2poly:  degree must be <= %d; got %d.\n",
			Z2POLY_MAX_DEGREE, degree);
		exit(1);
	}
}

// ----------------------------------------------------------------
static void complain_about_overflow(int degree)
{
	fprintf(stderr, "z2poly_t overflow:  degree %d, max %d.\n",
		degree, Z2POLY_MAX_DEGREE);
	exit(1);
}

//// ----------------------------------------------------------------
//// Multiplies in place by x (i.e. promotes degree by 1).
//void z2poly_promote_1(
//	z2poly_t * pc)
//{
//	int i;
//	int shamt = Z2POLY_BITS_PER_PART - 1;
//	for (i = Z2POLY_NUM_PARTS - 1; i >= 0; i--) {
//		pc->parts[i] <<= 1;
//		if (i != 0)
//			pc->parts[i] |= pc->parts[i-1] >> shamt;
//	}
//}

// ----------------------------------------------------------------
// Example: shift by 72
// parts[3] parts[2] parts[1] parts[0]
// 00000000 00000000 00000000 00000001 <-- input
// 00000000 00000100 00000000 00000000 <-- desired output
// 72 = 2 whole-word ("part") shifts, 8 bit ("in-part") shifts.
// diffs = part 2, in-part 24.
//
// 00000000 00000000 00000000 00000001
//                   00000000 00000001 00000000 00000000
//                   00000000 00000100 00000000 00000000

z2poly_t z2poly_promote_n(
	z2poly_t * pin,
	unsigned bits_shift)
{
	z2poly_t out, temp;
	unsigned parts_shift;
	unsigned parts_diff;
	unsigned in_part_shift;
	unsigned in_part_diff;
	int srci;
	int dsti;
	int i;

	out  = z2poly_from_base_rep(0);
	temp = z2poly_from_base_rep(0);
	if (bits_shift >= Z2POLY_NUM_BITS) {
		return out;
	}

	parts_shift   = bits_shift >> Z2POLY_PART_LOG;
	in_part_shift = bits_shift &  Z2POLY_PART_MASK;
	parts_diff    = Z2POLY_NUM_PARTS - parts_shift;
	in_part_diff  = Z2POLY_BITS_PER_PART - in_part_shift;

	for (dsti = 0; dsti < parts_shift; dsti++)
		temp.parts[dsti] = 0;
	for (dsti = parts_shift, srci = 0; srci < parts_diff; srci++, dsti++)
		temp.parts[dsti] = pin->parts[srci];

	for (i = Z2POLY_NUM_PARTS - 1; i >= 0; i--) {
		out.parts[i] = temp.parts[i] << in_part_shift;
		if (i != 0) {
			if (in_part_diff != Z2POLY_BITS_PER_PART)
				out.parts[i] |= temp.parts[i-1] >> in_part_diff;
		}
	}
	return out;
}

// ----------------------------------------------------------------
int z2poly_qcmp(const void * pv1, const void * pv2)
{
	if (z2poly_lt((z2poly_t*)pv1, (z2poly_t*)pv2))
		return -1;
	if (z2poly_gt((z2poly_t*)pv1, (z2poly_t*)pv2))
		return 1;
	return 0;
}

//// ----------------------------------------------------------------
//unsigned z2poly_bit_at(
//	z2poly_t * pin,
//	int        pos)
//{
//	int which_part = pos >> Z2POLY_PART_LOG;
//	int which_bit  = pos &  Z2POLY_PART_MASK;
//
//	z2poly_bounds_check_bit(pos);
//	return (pin->parts[which_part] >> which_bit) & 1;
//}

//// ----------------------------------------------------------------
//void z2poly_set_bit(
//	z2poly_t * pa,
//	int        pos)
//{
//	int which_part = pos >> Z2POLY_PART_LOG;
//	int which_bit  = pos &  Z2POLY_PART_MASK;
//
//	z2poly_bounds_check_bit(pos);
//	pa->parts[which_part] |= (1 << which_bit);
//}

// ----------------------------------------------------------------
//static void z2poly_clear_bit(
//	z2poly_t * pa,
//	int        pos)
//{
//	int which_part = pos >> Z2POLY_PART_LOG;
//	int which_bit  = pos &  Z2POLY_PART_MASK;
//
//	z2poly_bounds_check_bit(pos);
//	pa->parts[which_part] &= ~(1 << which_bit);
//}

// ----------------------------------------------------------------
//static void z2poly_bounds_check_bit(
//	int pos)
//{
//	if ((pos < 0) || (pos >= Z2POLY_NUM_BITS)) {
//		fprintf(stderr,
//		"z2poly: bit position %d out of range 0-%d.\n",
//			pos, Z2POLY_NUM_BITS - 1);
//		exit(1);
//	}
//}
