// Elements of the ring Z_2[x], implemented using multiple 32-bit integers.

#ifndef Z2POLY_H
#define Z2POLY_H

#include "factorize.h"
#include "ubig.h"
#include "cmps.h"

// ----------------------------------------------------------------
#define Z2POLY_NUM_PARTS  8//20//80//16

#define Z2POLY_PART_MASK  31
#define Z2POLY_PART_LOG    5

#define Z2POLY_NYBBLES_PER_PART (sizeof(unsigned) * 2)
#define Z2POLY_NUM_NYBBLES (Z2POLY_NUM_PARTS * Z2POLY_NYBBLES_PER_PART)

#define Z2POLY_BITS_PER_PART    (sizeof(unsigned) * 8)
#define Z2POLY_NUM_BITS    (Z2POLY_NUM_PARTS * Z2POLY_BITS_PER_PART)

#define Z2POLY_MAX_DEGREE (Z2POLY_NUM_BITS - 1)

typedef struct _z2poly_t {
	unsigned parts[Z2POLY_NUM_PARTS];
} z2poly_t;

// ----------------------------------------------------------------
//#define Z2POLY_NUM_BITS (8 * sizeof(z2poly_t))
//#define Z2POLY_MAX_DEGREE (Z2POLY_NUM_BITS - 1)

// ----------------------------------------------------------------
z2poly_t z2poly_add(
	z2poly_t * pin1,
	z2poly_t * pin2);

z2poly_t z2poly_sub(
	z2poly_t * pin1,
	z2poly_t * pin2);

z2poly_t z2poly_mul(
	z2poly_t * pin1,
	z2poly_t * pin2);

z2poly_t z2poly_div(
	z2poly_t * pin1,
	z2poly_t * pin2);

z2poly_t z2poly_mod(
	z2poly_t * pin1,
	z2poly_t * pin2);

void z2poly_quot_and_rem(
	z2poly_t * pa,
	z2poly_t * pb,
	z2poly_t * pquot,
	z2poly_t * prem);

z2poly_t z2poly_power(
	z2poly_t * px,
	unsigned  power);

// Assumes the input is already known to be a square.
z2poly_t z2poly_square_root(
	z2poly_t * pin);

z2poly_t z2poly_gcd(
	z2poly_t * pin1,
	z2poly_t * pin2);

z2poly_t z2poly_ext_gcd(
	z2poly_t * pa,
	z2poly_t * pb,
	z2poly_t * pm,
	z2poly_t * pn);

int z2poly_equal(
	z2poly_t * pa,
	z2poly_t * pb);

// Fp[x] is *not* totally ordered.  Nonetheless, these functions are
// handy for looping and sorting.
int z2poly_lt(z2poly_t * pa, z2poly_t * pb);
int z2poly_gt(z2poly_t * pa, z2poly_t * pb);
int z2poly_le(z2poly_t * pa, z2poly_t * pb);
int z2poly_ge(z2poly_t * pa, z2poly_t * pb);
int z2poly_eq(z2poly_t * pa, z2poly_t * pb);
int z2poly_ne(z2poly_t * pa, z2poly_t * pb);

int z2poly_cmp(
	z2poly_t * pa,
	int        cmp, // see cmps.h
	z2poly_t * pb);

// ----------------------------------------------------------------
// Construction

z2poly_t z2poly_from_base_rep(
	unsigned val);

z2poly_t z2poly_from_base_rep_64(
	unsigned long long val);

int z2poly_equals_unsigned(
	z2poly_t * pa,
	unsigned   val);

int z2poly_is_zero(
	z2poly_t * pa);

int z2poly_is_one(
	z2poly_t * pa);

z2poly_t z2poly_from_ubig(
	ubig * pu);

// ----------------------------------------------------------------
// I/O

void z2poly_print(
	z2poly_t * pa);

void z2poly_pretty_print(
	z2poly_t * pa);

int z2poly_scan(
	z2poly_t * pz,
	char     * pin);

// ----------------------------------------------------------------
// Factorization, evaluation and classification.

int z2poly_eval(
	z2poly_t * pa,
	int        x);

z2poly_t z2poly_compose(
	z2poly_t * pr,
	z2poly_t * pz);

z2poly_t z2poly_deriv(
	z2poly_t * pz);

int z2poly_is_rootless(
	z2poly_t * pa);

int z2poly_count_bits(
	z2poly_t * pa);

int z2poly_get_degree(
	z2poly_t * pa);

// Used to check for overflow.
void z2poly_check_degree(
	int degree);

// ----------------------------------------------------------------
// Internal functions, shared between z2poly and z2polymod.

//void z2poly_promote_1(
//	z2poly_t * pc);

// Multiplies in place by x (i.e. promotes degree by 1).
static inline void z2poly_promote_1(
	z2poly_t * pc)
{
	int i;
	int shamt = Z2POLY_BITS_PER_PART - 1;
	for (i = Z2POLY_NUM_PARTS - 1; i >= 0; i--) {
		pc->parts[i] <<= 1;
		if (i != 0)
			pc->parts[i] |= pc->parts[i-1] >> shamt;
	}
}

static inline void z2poly_demote_1(
	z2poly_t * pc)
{
	int i;
	int shamt = Z2POLY_BITS_PER_PART - 1;
	int np1 = Z2POLY_NUM_PARTS - 1;
	unsigned * pp = &pc->parts[0];

	for (i = 0; i < np1; i++) {
		//pc->parts[i] >>= 1;
		//pc->parts[i] |= pc->parts[i+1] << shamt;
		*pp >>= 1;
		*pp |= pp[1] << shamt;
		pp++;
	}

	pc->parts[np1] >>= 1;
}

z2poly_t z2poly_promote_n(
	z2poly_t * pin,
	unsigned bits_shift);

#define z2poly_bit_at(pz,pos) \
	(((pz)->parts[(pos)>>Z2POLY_PART_LOG] >> ((pos)&Z2POLY_PART_MASK)) & 1)

#define z2poly_set_bit(pz,pos) \
	((pz)->parts[(pos)>>Z2POLY_PART_LOG] |= 1 << ((pos)&Z2POLY_PART_MASK))

#define z2poly_toggle_bit(pz,pos) \
	((pz)->parts[(pos)>>Z2POLY_PART_LOG] ^= 1 << ((pos)&Z2POLY_PART_MASK))

// ----------------------------------------------------------------
// For use with qsort(3).
int z2poly_qcmp(
	const void * pv1,
	const void * pv2);

#endif // Z2POLY_H
