#include <stdio.h>
#include <stdlib.h>
#include "log2.h"
#include "malloc_check.h"
#include "z2polymod.h"
#include "z2linalg.h"
#include "z2poly_factor.h"

static void pre_berlekamp2(
	z2poly_t * pr,
	z2poly_t * pirred_factors,
	int      * pnum_irred_factors);

static void berlekamp2(
	z2poly_t * pr,
	z2poly_t * pirred_factors,
	int      * pnum_irred_factors);

static void gather_multiple_factors(
	z2poly_t * pirred_factors,
	int         num_irred_factors,
	z2poly_factor_info_t * pfinfo);

static void sort_up(
	z2poly_t * M,
	int n);

// ----------------------------------------------------------------
void z2poly_factor(
	z2poly_t * pr,
	z2poly_factor_info_t * pfinfo)
{
	int num_irred_factors;
	z2poly_t irred_factors[Z2POLY_MAX_NUM_FACTORS];

	num_irred_factors = 0;
	pre_berlekamp2(pr, irred_factors, &num_irred_factors);
	sort_up(irred_factors, num_irred_factors);
	gather_multiple_factors(irred_factors, num_irred_factors, pfinfo);
}

// ----------------------------------------------------------------
static void pre_berlekamp2(
	z2poly_t * pr,
	z2poly_t * pirred_factors,
	int      * pnum_irred_factors)
{
	z2poly_t d, g;

	d = z2poly_deriv(pr);
	g = z2poly_gcd(pr, &d);

	if (z2poly_equals_unsigned(&g, 0)) {
		if (!z2poly_equals_unsigned(pr, 0)) {
			fprintf(stderr, "CODING ERROR file %s\n", __FILE__);
			exit(1);
		}
		pirred_factors[*pnum_irred_factors] = *pr;
		(*pnum_irred_factors)++;
		return;
	}
	else if (z2poly_equals_unsigned(&g, 1)) {
		// Input is squarefree.
		berlekamp2(pr, pirred_factors, pnum_irred_factors);
	}
	else if (z2poly_equals_unsigned(&d, 0)) {
		// Input is a perfect square
		z2poly_t s = z2poly_square_root(pr);

		// Multiplicity is p only if degree is > 0.
		pre_berlekamp2(&s, pirred_factors, pnum_irred_factors);
		if (z2poly_get_degree(pr) > 0)
			pre_berlekamp2(&s, pirred_factors, pnum_irred_factors);
	}
	else {
		z2poly_t q;
		q = z2poly_div(pr, &g);
		pre_berlekamp2(&g, pirred_factors, pnum_irred_factors);
		pre_berlekamp2(&q, pirred_factors, pnum_irred_factors);
	}
}

//// ----------------------------------------------------------------
//// Berlekamp's algorithm requires squarefree input.  In the literature,
//// this is often described as being done by factoring out GCD(r(x), r'(x)).
//// This function takes a different approach to find squareful inputs, and also
//// often produces a complete factorization:
////
//// X^(p^n) - X factors into all irreducibles of degree d for all d dividing
//// n.  Therefore, if r (the input, of degree n) has an irreducible factor
//// of degree < n, that factor will divide X^(p^d) - X for some d < n.
//// If r has GCD = 1 with X^(p^d) - X for all d < n, then r is irreducible.
//// Reduction mod r is taken at each step to avoid overflow, with no effect
//// on the GCD.
//
//static void pre_berlekamp2(
//	z2poly_t * pr,
//	z2poly_t * pirred_factors,
//	int      * pnum_irred_factors)
//{
//	z2poly_t x = z2poly_from_base_rep(2);
//	z2poly_t u = z2poly_from_base_rep(2);
//	z2poly_t ux;
//	z2poly_t g;
//	int d = z2poly_get_degree(pr);
//	int i;
//
//	for (i = 1; i < d; i++) {
//		// u = u**2 mod r; test GCD(u - x, r)
//		z2polymod_mul(&u, &u, &u, pr);
//		z2poly_add(&ux, &u, &x);
//		g = z2poly_gcd(pr, &ux);
//
//		if (z2poly_equals_unsigned(&ux, 0)) {
//			// If we got here, r is known to be reducible but this
//			// algorithm can't distinguish its factors.  However,
//			// it is known that r is squarefree, suitable for input
//			// to Berlekamp's algorithm.  (Incidentally, all of r's
//			// factors are of the same degree, but that fact won't
//			// be needed.)
//			berlekamp2(pr, pirred_factors, pnum_irred_factors);
//			return;
//		}
//
//		if (!z2poly_equals_unsigned(&g, 1)) {
//			// If we got here, r is known to be reducible and g
//			// is a factor.
//			z2poly_t q;
//			z2poly_div(&q, pr, &g);
//			pre_berlekamp2(&g, pirred_factors, pnum_irred_factors);
//			pre_berlekamp2(&q, pirred_factors, pnum_irred_factors);
//			return;
//		}
//	}
//	// If we got here, r is known to be irreducible.
//	pirred_factors[*pnum_irred_factors] = *pr;
//	(*pnum_irred_factors)++;
//}

// ----------------------------------------------------------------
// Worked example: r = 70e = 2 * 3 * 7 * b * d.  Deg(r) = 10.
//
// By explicit search, we can find the following polynomials h of degree
// < 10 such that h**2 = h mod r (Lidl and Niederreiter call these "f-reducing
// polynomials"):
//
// 000 001 07e 07f 096 097 0e8 0e9 102 103 17c 17d 194 195 1ea 1eb 212 213
// 26c 26d 284 285 2fa 2fb 310 311 36e 36f 386 387 3f8 3f9
//
// The rest of the example is interspersed through the function:  Berlekamp's
// algorithm uses linear algebra to produce f-reducing h's, without recourse
// to a brute-force search for them.  We will use one of them to produce a
// non-trivial factorization of r.

static void berlekamp2(
	z2poly_t * pr,
	z2poly_t * pirred_factors,
	int      * pnum_irred_factors)
{
	int n  = z2poly_get_degree(pr);
	z2poly_t x = z2poly_from_base_rep(2);
	z2poly_t x2;
	z2poly_t x2i = z2poly_from_base_rep(1);
	z2poly_t f1, f2;
	int row, rank, dimker;

	//z2poly_t BI[Z2POLY_NUM_BITS];
	//z2poly_t BIT[Z2POLY_NUM_BITS];
	//z2poly_t nullspace_basis[Z2POLY_NUM_BITS];

	z2poly_t * BI = malloc_check(
		Z2POLY_NUM_BITS * sizeof(z2poly_t));
	z2poly_t * BIT = malloc_check(
		Z2POLY_NUM_BITS * sizeof(z2poly_t));
	z2poly_t * nullspace_basis = malloc_check(
		Z2POLY_NUM_BITS * sizeof(z2poly_t));

	x2 = z2polymod_power(&x, 2, pr);

	// Input: r = 70e = 2 * 3 * 7 * b * d.
	// Since r has degree 10, B is a 10x10 matrix with row i equal to
	// x^2i mod r.  Note:  using the usual arithmetic leads to a backward
	// matrix, i.e. x^0 is on the right and x^9 is on the left.

	//  9876543210
	//  ----------
	//  0000000001         1  | 0
	//  0000000100         4  | 1
	//  0000010000        10  | 2
	//  0001000000        40  | 3
	//  0100000000       100  | 4
	//  1100001110       30e  | 5
	//  1000100100       224  | 6
	//  0110000010       182  | 7
	//  0100000110       106  | 8
	//  1100010110       316  | 9

	if (n < 2) {
		pirred_factors[*pnum_irred_factors] = *pr;
		(*pnum_irred_factors)++;
		free(BI);
		free(BIT);
		free(nullspace_basis);
		return;
	}

	// Populate the B matrix with row i equal to x^2i mod r
	for (row = 0; row < n; row++) {
		BI[row] = x2i;
		x2i = z2polymod_mul(&x2i, &x2, pr);
	}

	// Now form B - I:
	//  0000000000
	//  0000000110
	//  0000010100
	//  0001001000
	//  0100010000
	//  1100101110
	//  1001100100
	//  0100000010
	//  0000000110
	//  0100010110

	// Replace B with B - I
	for (row = 0; row < n; row++) {
		z2poly_toggle_bit(&BI[row], row);
	}

	// Following Lidl and Niederreiter, we want to column-reduce the
	// B-I matrix.  However, it's convenient in C to be able to use
	// word-size arithmetic, e.g. XOR'ing two rows together in a single
	// operation.  So, we'll transpose B-I, and row-reduce it:

	//  0000000000
	//  1110100010
	//  1101100110
	//  0000101000
	//  1000010100
	//  0001100000
	//  0001001000
	//  0000000000
	//  1010110000
	//  0001100000

	z2la_transpose(BIT, BI, n);
	z2la_row_echelon_form(BIT, n);

	// Now the transposed B-I matrix is in echelon form:
	//
	//  1000010100       214
	//  0100010010       112
	//  0010001100        8c
	//  0001001000        48
	//  0000101000        28
	//  0000000000         0
	//  0000000000         0
	//  0000000000         0
	//  0000000000         0
	//  0000000000         0
	//
	// Clearly, its rank is 5 and its nullity is also 5.  x^0 through
	// x^4 are free; x^5 through x^9 are bound.

	rank = z2la_find_rank(BIT, n);
	dimker = n - rank;

	if (dimker == 1) {
		pirred_factors[*pnum_irred_factors] = *pr;
		(*pnum_irred_factors)++;
		free(BI);
		free(BIT);
		free(nullspace_basis);
		return;
	}

	// Now we want to find a basis for the nullspace of B-I.
	// There are 5 free coefficients (c0 through c4).  For each, in
	// turn, set ci = 1 and the rest = 0.
	//
	// Transposing the matrix back, we have:
	//
	//  0000000000
	//  0100000000
	//  1010000000
	//  0011100000
	//  1100000000
	//  0000100000
	//  0001000000
	//  0010000000
	//  0100000000
	//  1000000000
	//
	// With ci = 1 and the rest = 0, for i = 0-4, in rows 0-4, we have
	// the following.  I use * to mark new 1's.

	//  000000000*         1
	//  01000000*0       102
	//  1010000*00       284
	//  001110*000        e8
	//  11000*0000       310
	//  0000100000        20
	//  0001000000        40
	//  0010000000        80
	//  0100000000       100
	//  1000000000       200

	// Rows with asterisks in them are elements of the basis of the
	// nullspace of B-I.

	//Nullspace basis:
	//  0000000001       001
	//  0100000010       102
	//  1010000100       284
	//  0011101000       0e8
	//  1100010000       310

	// Find a basis for the nullspace of B - I.
	z2la_find_nullspace_basis(BIT, n, nullspace_basis, dimker);

	// For each h in the nullspace basis, form
	//   f1 = gcd(r, h)
	//   f2 = gcd(r, h-1)
	// Now, the polynomial h=1 is always in the nullspace, in which case
	// f1 = 1 and f2 = r, which results in a non-trivial factorization.
	// Any of the other h's will work fine, producing a non-trivial
	// factorization of r into two factors.  (Note that either or both
	// of them may be reducible, in which we'll need to apply this
	// algorithm recursively until we're down to irreducible factors.)
	// Here, for the sake of illustration, is what happens with all the
	// h's, even though we need only one of them:

	//h=001
	//  f1: 001
	//  f2: 70e = 2 3 7 b d
	//h=102
	//  f1: 102 = 2 3 b d
	//  f2: 007
	//h=284
	//  f1: 00e = 2 7
	//  f2: 081 = 3 b d
	//h=0e8
	//  f1: 03a = 2 3 b
	//  f2: 023 = 7 d
	//h=310
	//  f1: 062 = 2 7 b
	//  f2: 017 = 3 d

	for (row = 0; row <  dimker; row++) {
		z2poly_t * ph, hc;
		ph = &nullspace_basis[row];
		hc = *ph;
		z2poly_toggle_bit(&hc, 0);

		f1 = z2poly_gcd(pr, ph);
		f2 = z2poly_gcd(pr, &hc);

		if (z2poly_equals_unsigned(&f1, 1) ||
			z2poly_equals_unsigned(&f2, 1))
		{
			continue;
		}

		// The nullity of B-I is the number of irreducible
		// factors of r.  If the nullity is 2, we have a
		// pair of factors which are both irreducible and
		// so we don't need to recurse.
		if (dimker == 2) {
			pirred_factors[*pnum_irred_factors] = f1;
			(*pnum_irred_factors)++;
			pirred_factors[*pnum_irred_factors] = f2;
			(*pnum_irred_factors)++;
		}
		else {
			pre_berlekamp2(&f1, pirred_factors, pnum_irred_factors);
			pre_berlekamp2(&f2, pirred_factors, pnum_irred_factors);
		}
		free(BI);
		free(BIT);
		free(nullspace_basis);
		return;
	}
	printf("CODING ERROR: no non-trivial factors found: ");
	z2poly_print(pr);
	printf("\n");
	free(BI);
	free(BIT);
	free(nullspace_basis);
	exit(1);
}

// ----------------------------------------------------------------
static void gather_multiple_factors(
	z2poly_t * pirred_factors,
	int         num_irred_factors,
	z2poly_factor_info_t * pfinfo)
{
	int i;
	z2poly_t * plast;
	int nd = 0;

	if (num_irred_factors == 0) {
		pfinfo->num_distinct = 0;
		return;
	}

	plast = &pirred_factors[0];
	pfinfo->factor_counts[nd].f = *plast;
	pfinfo->factor_counts[nd].count = 0;

	for (i = 0; i < num_irred_factors; i++) {
		if (!z2poly_equal(&pirred_factors[i], plast)) {
			nd++;
			pfinfo->factor_counts[nd].count = 1;
			pfinfo->factor_counts[nd].f = pirred_factors[i];
			plast = &pirred_factors[i];
		}
		else {
			pfinfo->factor_counts[nd].count++;
		}
	}
	pfinfo->num_distinct = nd + 1;
}

// ----------------------------------------------------------------
// Ascending bubble sort
static void sort_up(
	z2poly_t * M,
	int n)
{
	int i, top;
	z2poly_t temp;

	for (top = n-1; top > 0; top--) {
		for (i = 0; i < top; i++) {
			if (z2poly_gt(&M[i], &M[i+1])) {
				temp   = M[i];
				M[i]   = M[i+1];
				M[i+1] = temp;
			}
		}
	}
}

// ----------------------------------------------------------------
// Given the prime factorization p1^m1 ... pk^mk of n, how to enumerate all
// factors of n?
//
// Example 72 = 2^3 * 3^2.  Exponent on 2 is one of 0, 1, 2, 3.  Exponent on
// 3 is one of 0, 1, 2.  Number of possibilities:  product over i of (mi + 1).
// Factors are:
//
//	2^0 3^0    1
//	2^0 3^1    3
//	2^0 3^2    9
//	2^1 3^0    2
//	2^1 3^1    6
//	2^1 3^2   18
//	2^2 3^0    4
//	2^2 3^1   12
//	2^2 3^2   36
//	2^3 3^0    8
//	2^3 3^1   24
//	2^3 3^2   72

unsigned z2poly_num_divisors(
	z2poly_factor_info_t * pfinfo)
{
	unsigned rv;
	int i;

	if ((pfinfo->num_distinct < 0)
		|| (pfinfo->num_distinct > Z2POLY_MAX_NUM_FACTORS))
	{
		fprintf(stderr, "z2poly_num_divisors:  num distinct (%d) "
			"out of bounds 0-%d\n",
			pfinfo->num_distinct, Z2POLY_MAX_NUM_FACTORS);
		exit(1);
	}

	rv = 1;
	for (i = 0; i < pfinfo->num_distinct; i++)
		rv *= pfinfo->factor_counts[i].count + 1;
	return rv;
}

// ----------------------------------------------------------------
// See comments to the above.  k is treated as a multibase representation
// over the bases mi+1.

z2poly_t z2poly_kth_divisor(
	z2poly_factor_info_t * pfinfo,
	unsigned k)
{
	z2poly_t rv;
	z2poly_t fp;
	int i;
	unsigned base;
	unsigned power;

	rv = z2poly_from_base_rep(1);
	for (i = 0; i < pfinfo->num_distinct; i++) {
		base = pfinfo->factor_counts[i].count + 1;
		power = k % base;
		k = k / base;

		fp = z2poly_power(&pfinfo->factor_counts[i].f, power);
		rv = z2poly_mul(&rv, &fp);
	}
	return rv;
}

// ----------------------------------------------------------------
// The caller must free the return value.

z2poly_t * z2poly_get_all_divisors(
	z2poly_t * pr,
	unsigned * pnum_divisors)
{
	z2poly_factor_info_t finfo;
	unsigned nf, k;
	z2poly_t * all_divisors;

	z2poly_factor(pr, &finfo);
	nf = z2poly_num_divisors(&finfo);
	*pnum_divisors = nf;

	all_divisors = (z2poly_t *)malloc_check(nf * sizeof(z2poly_t));
	for (k = 0; k < nf; k++) {
		all_divisors[k] = z2poly_kth_divisor(&finfo, k);
	}
	qsort(all_divisors, nf, sizeof(z2poly_t), z2poly_qcmp);

	return all_divisors;
}
