#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2polymod.h"
#include "ubig.h"
#include "ubig_factorize.h"
#include "ubig_f2n1.h"

// ----------------------------------------------------------------
// z2polymod_mul
//
// Multiplication may be done (at least) two ways:  (1) Take the ring product
// of the two polynomials, then reduce mod r; (2) reduce mod r at every step.
// Testing has shown that the latter method (reduce en route) is faster than
// the former (ring and reduce).  Furthermore, the ring-and-reduce method
// requires storage space for degree up to 2d-2, where d is the degree of the
// reductor.  E.g. with 64-bit arithmetic, we can't multiply 40-bit polynomials
// using ring-and-reduce, but we can if we use reduce-en-route.
//
// The ring-and-reduce method would be this:
//
//     temp = z2poly_mul(pa, pb);
//     return z2poly_mod(&temp, paux);

z2poly_t z2polymod_mul(
	z2poly_t * pa,
	z2poly_t * pb,
	z2poly_t * paux)
{
	z2poly_t temp, out;
	int i, power, degb, degr;
	z2poly_t ared, bred;

	ared = z2poly_mod(pa, paux);
	bred = z2poly_mod(pb, paux);

	degb = z2poly_get_degree(&bred);
	degr = z2poly_get_degree(paux);

	out = z2poly_from_base_rep(0);

	for (i = 0; i <= degb; i++) {
		// Multiply a by scalar coefficient on current term of b.
		// For Z2, this is either a 0 or 1.  0 means this partial
		// product is zero.  1 means we need to multiply by x^k.
		if (!z2poly_bit_at(&bred, i))
			continue;
		temp = ared;

		// Multiply by the power of x on the current term
		// of b, reducing mod r en route.
		for (power = 1; power <= i; power++) {
			z2poly_promote_1(&temp);
			if (z2poly_bit_at(&temp, degr))
				temp = z2poly_sub(&temp, paux);
		}
		// Add in this partial product.
		out =z2poly_add(&out, &temp);
	}
	return out;
}

// ----------------------------------------------------------------
z2poly_t z2polymod_div(
	z2poly_t * pin1,
	z2poly_t * pin2,
	z2poly_t * paux)
{
	z2poly_t binv;
	binv = z2polymod_recip(pin2, paux);
	return z2polymod_mul(pin1, &binv, paux);
}

// ----------------------------------------------------------------
// 1/b mod r.
//
// Finite-field division uses extended Euclidean GCD, with ring division
// (quotient and remainder) as a substep.
//
// a / b is a * 1/b.  To compute 1/b, find the GCD d.  (Since the reductor
// is irreducible -- although we don't test this on every division -- and
// if b is monic, the GCD had better be 1).  Then, compute 1 = m*b + n*r.
// Since r = 0 mod r, 1 = m*b and therefore m = 1/b.  The extended Euclidean
// algorithm (see Knuth vol. 1) computes d, m and n all at once.
//
// An alternate reciprocation algorithm uses the fact that a finite field
// GF(p^m) has its multiplicative group of order p^m-1.  So, by Euler's
// criterion, x^(p^n-2) = x^(-1) for all non-zero elements of the field.
//
// Testing has shown the extended-GCD reciprocation is about 2x faster.

z2poly_t z2polymod_recip(
	z2poly_t * pin,
	z2poly_t * paux)
{
	z2poly_t m, n, g;
	g = z2poly_ext_gcd(pin, paux, &m, &n);
	if (!z2poly_equals_unsigned(&g, 1)) {
		fprintf(stderr,
			"z2polymod_recip:  reductor is reducible.\n");
		exit(1);
	}
	return m;
}

// ----------------------------------------------------------------
z2poly_t z2polymod_compose(
	z2poly_t * pr,
	z2poly_t * pz,
	z2poly_t * pm)
{
	z2poly_t rz = z2poly_from_base_rep(0);
	z2poly_t zp = z2poly_from_base_rep(1);
	int n = z2poly_get_degree(pr);
	int d;

	for (d = 0; d <= n; d++) {
		if (z2poly_bit_at(pr, d))
			rz = z2poly_add(&rz, &zp);
		zp = z2polymod_mul(&zp, pz, pm);
	}

	return rz;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
z2poly_t z2polymod_power(
	z2poly_t * pin,
	unsigned   power,
	z2poly_t * paux)
{
	z2poly_t z2 = *pin;
	z2poly_t out = z2poly_from_base_rep(1);

	while (power != 0) {
		if (power & 1)
			out = z2polymod_mul(&out, &z2, paux);
		power >>= 1;
		z2 = z2polymod_mul(&z2, &z2, paux);
	}
	return out;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
z2poly_t z2polymod_power64(
	z2poly_t * pin,
	unsigned long long power64,
	z2poly_t * paux)
{
	z2poly_t z2 = *pin;
	z2poly_t out = z2poly_from_base_rep(1);
	unsigned long long zero = 0;
	unsigned long long one  = 1;

	while (power64 != zero) {
		if (power64 & one)
			out = z2polymod_mul(&out, &z2, paux);
		power64 >>= 1;
		z2 = z2polymod_mul(&z2, &z2, paux);
	}
	return out;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
z2poly_t z2polymod_power_ubig(
	z2poly_t * pr,
	ubig     * pbigpower,
	z2poly_t * paux)
{
	z2poly_t z2 = *pr;
	z2poly_t out = z2poly_from_base_rep(1);
	int msbpos = ubig_find_msb(pbigpower);
	int pos;

	for (pos = 0; pos <= msbpos; pos++) {
		if (ubig_get_bit(pbigpower, pos))
			out = z2polymod_mul(&out, &z2, paux);
		z2 = z2polymod_mul(&z2, &z2, paux);
	}

	return out;
}

// ----------------------------------------------------------------
// It suffices to check only factors of p^m-1.

// xxx this only works correctly when r is irreducible.
// xxx need to code for reducible case, with zero-divisor check, and
// phi(?) rather than p^m-1.

int z2polymod_element_order(
	z2poly_t * pz,
	z2poly_t * pr)
{
	int m;
	unsigned long long pm, pm1;
	unsigned nd, k;
	unsigned long long * all_divisors = 0;
	unsigned long long f = 0;
	z2poly_t zpower;
	int got_it = 0;

	m = z2poly_get_degree(pr);
	if (m >= 64) {
		fprintf(stderr, "Need ubig factorization here.  Kode me up!\n");
		exit(1);
		return 999;
	}
	pm = (unsigned long long)1 << m;
	pm1 = pm - 1;

	if (z2poly_is_zero(pz))
		return 0;

	// x mod x, the latter being irreducible.
	if (z2poly_equals_unsigned(pz, 2) && z2poly_equals_unsigned(pr, 2))
		return 0;

	all_divisors = get_all_divisors_ull(pm1, &nd);

	for (k = 0; k < nd; k++) {
		f = all_divisors[k];
		zpower = z2polymod_power64(pz, f, pr);
		if (z2poly_is_one(&zpower)) {
			got_it = 1;
			break;
		}
	}

	free(all_divisors);
	if (got_it) {
		return f;
	}
	else {
		printf("z2poly_t:  Internal error:  order of ");
		z2poly_print(pz);
		printf(" mod ");
		z2poly_print(pr);
		printf(" not found.\n");
		exit(1);
		return 0; // Not reached.
	}
}

// ----------------------------------------------------------------
void z2polymod_print(
	z2poly_t * pin,
	z2poly_t * paux)
{
// xxx temp
	z2poly_print(pin);
//	int width = (z2poly_get_degree(r) + 3) >> 2;
//	printf("%0*llx", width, a);
}

// ----------------------------------------------------------------
int z2polymod_auxscan(
	z2poly_t * pout,
	char     * pin)
{
	z2poly_t r;

	// If it doesn't start with ":", parse it as a user-specified reductor.
	// Else, parse as int m and search for a reductor of that degree.
	if (pin[0] != ':') {
		if (!z2poly_scan(&r, pin)) {
			fprintf(stderr, "Couldn't scan \"%s\" as hex.\n",
				pin);
			exit(1);
		}
		if (z2poly_get_degree(&r) < 1) {
			fprintf(stderr,
				"deg(r) must be >= 1; got %d.\n",
					z2poly_get_degree(&r));
			exit(1);
		}
	}
	else {
		int degree;
		if (sscanf(pin, ":%d", &degree) != 1) {
			fprintf(stderr,
				"z2poly_t:  Couldn't scan \"%s\" as degree.\n",
				pin);
			return 0;
		}
		if (degree > Z2POLY_MAX_DEGREE) {
			fprintf(stderr,
				"z2poly_t:  degree must be <= %d; got %d.\n",
				Z2POLY_MAX_DEGREE, degree);
			return 0;
		}
		if (degree < 1) {
			fprintf(stderr,
				"z2poly_t:  degree must be >= 1; got %d.\n",
				degree);
			return 0;
		}
		r = z2poly_find_irreducible(degree);
	}

	*pout = r;
	return 1;
}

// ----------------------------------------------------------------
z2poly_t z2poly_find_irreducible(
	int m)
{
	ubig i, pm, tpm;
	z2poly_t r;

	// xxx need to sanity-check m.
	pm = ubig_from_unsigned(1);
	pm = ubig_ilsh(&pm, m);
	tpm = ubig_ilsh(&pm, 1);
	ubig_incr(&pm);
	for (i = pm; ubig_lt(&i, &tpm); ubig_incr2(&i)) {
		r = z2poly_from_ubig(&i);
		if (z2poly_is_irreducible(&r))
			if (z2poly_is_primitive(&r))
				return r;
	}
	fprintf(stderr,
	"z2poly_t: coding error: couldn't find any irreds in GF(%d^%d).\n",
		2, m);
	exit(1);
}

// ----------------------------------------------------------------
// X^(p^n) - X factors into all irreducibles of degree d for all d dividing
// n.  Therefore, if r (the input, of degree n) has an irreducible factor
// of degree < n, that factor will divide X^(p^d) - X for some d < n.
// If r has GCD = 1 with X^(p^d) - X for all d < n, then r is irreducible.
// Reduction mod r is taken at each step to avoid overflow, with no effect
// on the GCD.

int z2poly_is_irreducible(
	z2poly_t * pr)
{
	z2poly_t u = z2poly_from_base_rep(2);
	z2poly_t ux;
	z2poly_t g;
	int  d = z2poly_get_degree(pr);
	int i;

	for (i = 1; i < d; i++) {
		u = z2polymod_mul(&u, &u, pr);
		ux = u;
		ux.parts[0] ^= 2;
		g = z2poly_gcd(pr, &ux);
		if (!z2poly_equals_unsigned(&g, 1))
			return 0;
	}
	return 1;
}

// ----------------------------------------------------------------
// Assumes z is known to be irreducible.
// See if the order of z is p^m-1.  Certainly, z^(p^m-1) will be 1.  But
// check z to all powers dividing p^m-1 to make sure they are *not* 1.

int z2poly_is_primitive(
	z2poly_t * pr)
{
	int m;
	ubig_factor_info_t * pfinfo;
	unsigned nf, k;
	ubig pm1, f, u1 = ubig_from_unsigned(1);
	z2poly_t z1 = z2poly_from_base_rep(1);
	z2poly_t xpower, x = z2poly_from_base_rep(2);
	int f_equals_pm1, xpower_equals_z1;

	if (z2poly_equals_unsigned(pr, 2))
		return 0;

	m = z2poly_get_degree(pr);
	pm1 = ubig_ilsh(&u1, m);
	ubig_decr(&pm1);
	pfinfo = ubig_fac2n1_get(m);
	nf = ubig_num_divisors(pfinfo);

	for (k = 0; k < nf; k++) {
		f = ubig_kth_divisor(pfinfo, k);

		xpower = z2polymod_power_ubig(&x, &f, pr);
		f_equals_pm1 = ubig_eq(&f, &pm1);
		xpower_equals_z1 = z2poly_eq(&xpower, &z1);

		if (!f_equals_pm1 && xpower_equals_z1) {
			return 0;
		}
		if (f_equals_pm1 && !xpower_equals_z1) {
			printf("z2poly_is_primitive_f abend!\n");
			printf("r=");
			z2poly_print(pr);
			printf(" nf=%u f=", nf);
			ubig_dprint(&f);
			printf(" pm1=");
			ubig_dprint(&pm1);
			printf("\n");
			exit(1);
		}
	}
	return 1;
}

// ----------------------------------------------------------------
z2poly_t z2polymod_find_generator(
	z2poly_t * pr)
{
	int n = z2poly_get_degree(pr);
	int pn1 = (1 << n) - 1;
	int gi;
	int gilo = 2;
	int gihi = pn1;
	z2poly_t g;

	if (gihi < gilo)
		gihi = gilo; // linear fields

	for (gi = gilo; gi <= gihi; gi++) {
		g = z2poly_from_base_rep(gi);
		if (z2polymod_element_order(&g, pr) == pn1)
			return g;
	}

	printf("Generator for ");
	z2poly_print(pr);
	printf(" not found.\n");
	exit(1);
}

// ----------------------------------------------------------------
int z2polymod_log(
	z2poly_t * pz, // Input
	z2poly_t * pg, // Base, assumed (!) primitive mod r.
	z2poly_t * pr) // Modulus for FF arithmetic
{
	int log;
	z2poly_t gp = *pg;
	int deg = z2poly_get_degree(pr);
	int pm = 1 << deg;

	// find e such that g^e == z mod r
	if (z2poly_is_zero(pz))
		return 0;

	for (log = 1; log <= pm+2; log++) {
		if (z2poly_equal(pz, &gp))
			return log;
		gp = z2polymod_mul(&gp, pg, pr);
	}

	printf("z2poly_log: ");
	z2poly_print(pg);
	printf(" is not primitive mod ");
	z2poly_print(pr);
	printf("\n");
	exit(1);
	return 0; // not reached
}

