#include "z3linalg.h"
#include "malloc_check.h"

static int find_row_leader_pos(
	unsigned char * prow,
	int       ncols,
	int     * ppos);

// ----------------------------------------------------------------
unsigned char ** z3la_alloc_matrix(
	int  nrows,
	int  ncols)
{
	unsigned char ** rv;
	int i;

	rv = malloc_check(nrows * sizeof(unsigned char *));
	for (i = 0; i < nrows; i++)
		rv[i] = malloc_check(ncols * sizeof(unsigned char));
	return rv;
}

// ----------------------------------------------------------------
void z3la_free_matrix(
	unsigned char ** M,
	int    nrows,
	int    ncols)
{
	int i;
	for (i = 0; i < nrows; i++)
		free(M[i]);
	free(M);
}

// ----------------------------------------------------------------
void z3la_print_matrix(
	unsigned char ** M,
	int    nrows,
	int    ncols)
{
	int i, j;
	for (i = 0; i < nrows; i++) {
#if 0
		for (j = 0; j < ncols; j++)
#else
		for (j = ncols-1; j >= 0; j--)
#endif
		{
			printf("%u", (unsigned)M[i][j]);
		}
		printf("\n");
	}
}

// ----------------------------------------------------------------
void z3la_matrix_from_z3polys(
	unsigned char ** M,
	int        nrows,
	int        ncols,
	z3poly_t  * polys)
{
	int i;
	for (i = 0; i < nrows; i++)
		z3la_row_from_z3poly(M[i], &polys[i], ncols);
}

// ----------------------------------------------------------------
void z3la_z3polys_from_matrix(
	z3poly_t  * polys,
	unsigned char ** M,
	int        nrows,
	int        ncols)
{
	int i, j;
	for (i = 0; i < nrows; i++) {
		for (j = 0; j < ncols; j++)
			polys[i].coeffs[j] = M[i][j];
		for (j = ncols; j <= Z3POLY_MAX_DEGREE; j++)
			polys[i].coeffs[j] = 0;
	}
}

// ----------------------------------------------------------------
void z3la_row_from_z3poly(
	unsigned char * prow,
	z3poly_t * pz3poly,
	int        ncols)
{
	int i;
	int d = z3poly_get_degree(pz3poly);
	for (i = 0; i <= d; i++)
		prow[i] = pz3poly->coeffs[i];
	for (i = d+1; i < ncols; i++)
		prow[i] = 0;
}

// ----------------------------------------------------------------
void z3la_transpose(
	unsigned char ** Mout,
	unsigned char ** Min,
	int       n)
{
	int row, col;

	for (row = 0; row < n; row++)
		for (col = 0; col < n; col++)
			Mout[row][col] = Min[col][row];
}

// ----------------------------------------------------------------
void z3la_row_echelon_form(
	unsigned char ** M,
	int        nrows,
	int        ncols)
{
	z3la_row_reduce_below (M, nrows, ncols);
	z3la_row_reduce_above (M, nrows, ncols);
	z3la_make_matrix_monic(M, nrows, ncols);
}

// ----------------------------------------------------------------
static int find_row_with_leader_at_col(
	unsigned char ** M,
	int    nrows,
	int    ncols,
	int    clear_col,
	int    leader_row,
	int  * pswap_row)
{
	int row;
	for (row = leader_row; row < nrows; row++) {
		if (M[row][clear_col] != 0) {
			*pswap_row = row;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
static void swap_rows(
	unsigned char ** M,
	int    nrows,
	int    ncols,
	int    leader_row,
	int    swap_row)
{
	int col, temp;

	for (col = 0; col < ncols; col++) {
		temp               = M[leader_row][col];
		M[leader_row][col] = M[swap_row][col];
		M[swap_row][col]   = temp;
	}
}

// ----------------------------------------------------------------
void z3la_make_row_monic(
	unsigned char * row,
	int   ncols,
	int   leader_col)
{
	int col;
	//int cinv = mod_recip(row[leader_col]);
	int cinv = row[leader_col]; // Z3 is self-reciprocating
	for (col = leader_col; col < ncols; col++) {
		row[col] = Z3_MUL(row[col], cinv);
	}
}

// ----------------------------------------------------------------
void z3la_row_reduce_below(
	unsigned char ** M,
	int        nrows,
	int        ncols)
{
	int clear_col, leader_row, swap_row, clear_row;

	for (clear_col = 0, leader_row = 0; clear_col < ncols; clear_col++) {
		if (!find_row_with_leader_at_col(M, nrows, ncols, clear_col,
			leader_row, &swap_row))
		{
			continue;
		}
		if (leader_row != swap_row)
			swap_rows(M, nrows, ncols, leader_row, swap_row);
		if (M[leader_row][clear_col] != 1)
			z3la_make_row_monic(M[leader_row], ncols, clear_col);
		for (clear_row = leader_row+1; clear_row < nrows; clear_row++) {
			int col, clear_leader, clear_leader_inv;
			clear_leader = M[clear_row][clear_col];
			if (clear_leader == 0)
				continue;
			//clear_leader_inv = mod_recip(clear_leader);
			// Z3 is self-reciprocating
			clear_leader_inv = clear_leader;
			for (col = clear_col; col < ncols; col++) {
				M[clear_row][col] = Z3_MUL(
					M[clear_row][col], clear_leader_inv);
				M[clear_row][col] = Z3_SUB(
					M[clear_row][col], M[leader_row][col]);
			}
		}
		leader_row++;
	}
}

// ----------------------------------------------------------------
// Assumes row_reduce_below has already been done.

void z3la_row_reduce_above(
	unsigned char ** M,
	int        nrows,
	int        ncols)
{
	int row, row2, row2_leader_pos;
	int row_leader_val, row2_leader_val, mul, col;

	for (row = 0; row < nrows; row++) {
		for (row2 = row+1; row2 < nrows; row2++) {
			if (!find_row_leader_pos(M[row2], ncols,
				&row2_leader_pos))
			{
				break;
			}
			row2_leader_val = M[row2][row2_leader_pos];
			row_leader_val  = M[row][row2_leader_pos];
			if (row_leader_val == 0)
				continue;

			mul = Z3_DIV(row_leader_val, row2_leader_val);
			for (col = 0; col < ncols; col++)
				// row = row - mul * row2
				M[row][col] = Z3_SUB(M[row][col],
					Z3_MUL(M[row2][col], mul));
		}
	}
}

// ----------------------------------------------------------------
// Assumes row_reduce_below has already been done.

void z3la_make_matrix_monic(
	unsigned char ** M,
	int        nrows,
	int        ncols)
{
	int row, row_leader_pos, row_leader_val, col;

	for (row = 0; row < nrows; row++) {
		if (!find_row_leader_pos(M[row], ncols, &row_leader_pos))
			break;
		row_leader_val = M[row][row_leader_pos];

		for (col = 0; col < ncols; col++) {
			M[row][col] = Z3_DIV(
				M[row][col], row_leader_val);
		}
	}
}

// ----------------------------------------------------------------
static int find_row_leader_pos(
	unsigned char * prow,
	int       ncols,
	int     * ppos)
{
	int col;

	for (col = 0; col < ncols; col++) {
		if (prow[col] != 0) {
			*ppos = col;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
// Assumes the matrix is already row-reduced.
int z3la_find_rank(
	unsigned char ** M,
	int    nrows,
	int    ncols)
{
	int rv = 0;
	int row;

	for (row = 0; row < nrows; row++) {
		if (!z3la_row_is_zero(M[row], ncols))
			rv++;
	}
	return rv;
}

// ----------------------------------------------------------------
int z3la_row_is_zero(
	unsigned char * prow,
	int        ncols)
{
	int col;
	for (col = 0; col < ncols; col++)
		if (prow[col] != 0)
			return 0;
	return 1;
}

// ----------------------------------------------------------------
// Assumes the matrix is already in row-echelon form.

void z3la_find_nullspace_basis(
	unsigned char ** BIT,
	int        nrows,
	int        ncols,
	z3poly_t  * nullspace_basis,
	int        dimker)
{
	int i, j;
	unsigned char * free_flags;
	unsigned char * free_indices;
	int nfree = 0; // == dimker but I'll compute it anyway
	int rank = nrows - dimker;
	int dep_pos;

	for (i = 0; i < dimker; i++) {
		for (j = 0; j < ncols; j++)
			nullspace_basis[i].coeffs[j] = 0;
		for (j = ncols; j <= Z3POLY_MAX_DEGREE; j++)
			nullspace_basis[i].coeffs[j] = 0;
	}

	free_flags   = malloc_check(nrows);
	free_indices = malloc_check(nrows);

	for (i = 0; i < nrows; i++) {
		free_flags[i] = 1;
	}

	for (i = 0; i < rank; i++) {
		if (find_row_leader_pos(BIT[i], ncols, &dep_pos))
			free_flags[dep_pos] = 0;
	}

	for (i = 0; i < nrows; i++) {
		if (free_flags[i])
			free_indices[nfree++] = i;
	}

	// For each free coefficient:
	//   Let that free coefficient be 1 and the rest be zero.
	//   Also set any dependent coefficients which depend on that
	//   free coefficient.
	for (i = 0; i < dimker; i++) {
		nullspace_basis[i].coeffs[free_indices[i]] = 1;

		// (B-I)^t in row echelon form:
		//
		// 0210     c0 = ??      c0 = 1  c0 = 0
		// 1000     c1 = -2 c2   c1 = 0  c1 = 5
		// 0000     c2 = ??      c2 = 0  c2 = 1
		// 0000     c3 = 0       c3 = 0  c3 = 0

		// j  = 0,1
		// fi = 0,2

		// i = 0:
		//   j = 0  row 0 fi 0 = row 0 c0 = 0
		//   j = 1  row 1 fi 0 = row 1 c0 = 0
		// i = 1:
		//   j = 0  row 0 fi 1 = row 0 c2 = 2 dep_pos = 1
		//   j = 1  row 1 fi 1 = row 1 c2 = 0

		// 0001
		// 01?0

		for (j = 0; j < rank; j++) {
			if (BIT[j][free_indices[i]] == 0)
				continue;

			(void)find_row_leader_pos(BIT[j], ncols, &dep_pos);
			nullspace_basis[i].coeffs[dep_pos] =
				Z3_SUB(0, BIT[j][free_indices[i]]);
		}
	}
	free(free_flags);
	free(free_indices);
}
