#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "z3npoly.h"
#include "mod.h"
#include "power.h"
#include "tokenize.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
void z3npoly_check_degree(int degree)
{
	if (degree < 0) {
		fprintf(stderr,
			"z3npoly:  degree must be non-negative; got %d.\n",
			degree);
		exit(1);
	}
	if (degree > Z3NPOLY_MAX_DEGREE) {
		fprintf(stderr,
			"z3npoly:  degree must be <= %d; got %d.\n",
			Z3NPOLY_MAX_DEGREE, degree);
		exit(1);
	}
}

// ----------------------------------------------------------------
z3npoly_t z3npoly_add(
	z3npoly_t * pin1,
	z3npoly_t * pin2,
	z3poly_t  * pr)
{
	z3npoly_t   out;
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++)
			out.coeffs[i] = z3poly_add(&pin1->coeffs[i],
				&pin2->coeffs[i]);
		for (i = pin2->degree + 1; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i];
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++)
			out.coeffs[i] = z3poly_add(&pin1->coeffs[i],
				&pin2->coeffs[i]);
		for (i = pin1->degree + 1; i <= pin2->degree; i++)
			out.coeffs[i] = pin2->coeffs[i];
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were added.
	out.degree = z3npoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
z3npoly_t z3npoly_sub(
	z3npoly_t * pin1,
	z3npoly_t * pin2,
	z3poly_t * pr)
{
	z3npoly_t   out;
	z3poly_t    zero = z3poly_from_base_rep(0);
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++)
			out.coeffs[i] = z3poly_sub(&pin1->coeffs[i],
				&pin2->coeffs[i]);
		for (i = pin2->degree + 1; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i];
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++) {
			out.coeffs[i] = z3poly_sub(&pin1->coeffs[i],
				&pin2->coeffs[i]);
		}
		for (i = pin1->degree + 1; i <= pin2->degree; i++) {
			// out.coeffs[i] = negative of pin2->coeffs[i]
			out.coeffs[i] = z3poly_sub(&zero, &pin2->coeffs[i]);
		}
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were added.
	out.degree = z3npoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
z3npoly_t z3npoly_mul(
	z3npoly_t * pin1,
	z3npoly_t * pin2,
	z3poly_t * pr)
{
	z3npoly_t out;
	int i, j;
	z3poly_t prod;

	out.degree = pin1->degree + pin2->degree;
	z3npoly_check_degree(out.degree);
	for (i = 0; i <= out.degree; i++)
		out.coeffs[i] = z3poly_from_base_rep(0);

	for (i = 0; i <= pin1->degree; i++) {
		for (j = 0; j <= pin2->degree; j++) {
			prod = z3polymod_mul(&pin1->coeffs[i], &pin2->coeffs[j],
				pr);
			out.coeffs[i+j] = z3poly_add(&out.coeffs[i+j], &prod);
		}
	}
	// E.g. if one polynomial was zero.  My convention is that the zero
	// polynomial has "degree" zero -- correct as an array index, but
	// mathematically incorrect.
	out.degree = z3npoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
z3npoly_t z3npoly_power(
	z3npoly_t * pin,
	int       power,
	z3poly_t * pr)
{
	z3npoly_t in2 = *pin;
	z3npoly_t inpower = z3npoly_from_int(1);

	while (power != 0) {
		if (power & 1)
			inpower = z3npoly_mul(&in2, &inpower, pr);
		power = (unsigned)power >> 1;
		in2 = z3npoly_mul(&in2, &in2, pr);
	}
	return inpower;
}

// ----------------------------------------------------------------
// Assumes the input is already known to be a cube.

z3npoly_t z3npoly_cube_root(
	z3npoly_t * pin,
	z3poly_t * pr)
{
	int si, di;
	int j;
	z3npoly_t out;
	z3poly_t zero = z3poly_from_base_rep(0);

	for (di = 0; di <= pin->degree; di++) {
		out.coeffs[di] = zero;
	}

	for (si = 0, di = 0; si <= pin->degree; si += 3, di++) {
		out.coeffs[di] = pin->coeffs[si];
		for (j = 1; j < 3; j++) {
			if (si+j > pin->degree)
				break;
			if (!z3poly_is_zero(&pin->coeffs[si+j])) {
				printf("z3npoly_cube_root: ");
				z3npoly_print(pin, pr);
				printf(" is not a cube.\n");
				exit(1);
			}
		}
	}
	out.degree = z3npoly_find_degree(&out, pin->degree);

	return out;
}

// ----------------------------------------------------------------
// Input: a, b
// Output: GCD g, and m, n such that am + bn = d

z3npoly_t z3npoly_gcd(
	z3npoly_t * pa, // 1st input.
	z3npoly_t * pb, // 2nd input.
	z3poly_t * pr)
{
	z3npoly_t c, q, r;
	z3npoly_t d;   // Return value.

	if (z3npoly_is_zero(pa))
		return *pb;
	if (z3npoly_is_zero(pb))
		return *pa;

	c = *pa;
	d = *pb;

	while (1) {

		// Divide
		// q = c / d, r = c % d
		z3npoly_quot_and_rem(&c, &d, &q, &r, pr);
		// Note:  now c = qd + r and 0 <= r < d

		// Remainder zero?
		//if (r == zero)
		if (z3npoly_is_zero(&r))
			break;

		// Recycle
		c = d;
		d = r;
	}
	return d;
}

// ----------------------------------------------------------------
int z3npoly_is_zero(z3npoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (!z3poly_is_zero(&pz->coeffs[0]))
		return 0;
	return 1;
}

// ----------------------------------------------------------------
int z3npoly_is_one(z3npoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (!z3poly_is_one(&pz->coeffs[0]))
		return 0;
	return 1;
}

// ----------------------------------------------------------------
void z3npoly_make_monic(
	z3npoly_t * pz,
	z3poly_t * pr)
{
	int i;
	z3poly_t one = z3poly_from_base_rep(1);
	z3poly_t c   = pz->coeffs[pz->degree];
	z3poly_t cinv;

	if (z3poly_is_zero(&c))
		return;
	cinv = z3polymod_div(&one, &c, pr);

	for (i = 0; i <= pz->degree; i++) {
		pz->coeffs[i] = z3polymod_mul(&pz->coeffs[i], &cinv, pr);
	}
}

// ----------------------------------------------------------------
static z3npoly_t z3npoly_smul(
	z3npoly_t * pin, // Poly
	z3poly_t  * pscalar, // Scalar
	z3poly_t * pr)
{
	z3npoly_t  out;
	int i;

	for (i = 0; i <= pin->degree; i++)
		out.coeffs[i] = z3polymod_mul(&pin->coeffs[i], pscalar, pr);

	if (z3poly_is_zero(pscalar))
		out.degree = 0;
	else
		out.degree = pin->degree;
	return out;
}

// ----------------------------------------------------------------
int z3npoly_find_degree(
	z3npoly_t * pin,
	int        max_deg)
{
	int i;

	for (i = max_deg; i >= 0; i--)
		if (!z3poly_is_zero(&pin->coeffs[i]))
			return i;
	return 0;
}

// ----------------------------------------------------------------
// E.g.
// dividend = 1,2,3,4 (1 + 2x + 3x^2 + 4x^3)
// divisor  = 1,1,2   (1 +  x + 2x^2)
// modulus = 7
//
//             q=4,2   r = 4,3
//        +----------
// 1,1,2  |  1,2,3,4
//        |    2,2,4 shift = 1.  4/2 mod 7 = 2.  1,1,2 * 2 = 2,2,4.
//        +----------
//        |  1 0 1
//        |  4 4 1   shift = 0.  1/2 mod 7 = 4   1,1,2 * 4 = 4,4,1
//        +----------
//        |  4 3

void z3npoly_quot_and_rem(
	z3npoly_t * pdividend,
	z3npoly_t * pdivisor,
	z3npoly_t * pquot,
	z3npoly_t * prem,
	z3poly_t  * pr)
{
	z3npoly_t quot;
	z3npoly_t rem;
	int i;
	int max_shift;
	int shift;
	z3poly_t rem_leader;
	z3poly_t divisor_leader;
	z3poly_t dlinv;
	z3poly_t multiplier;
	z3npoly_t divisor_with_mul;
	int stop;

	max_shift = pdividend->degree - pdivisor->degree;
	if (pdividend->degree < pdivisor->degree) {
		quot.coeffs[0] = z3poly_from_base_rep(0);
		quot.degree = 0;
		rem = *pdividend;

		*pquot = quot;
		*prem  = rem;
		return;
	}

	quot.degree = pdividend->degree - pdivisor->degree;
	for (i = 0; i <= quot.degree; i++)
		quot.coeffs[i] = z3poly_from_base_rep(0);
	rem = *pdividend;

	divisor_leader = pdivisor->coeffs[pdivisor->degree];
	dlinv = z3polymod_recip(&divisor_leader, pr);
	for (shift = max_shift; shift >= 0; shift--) {
		rem_leader = rem.coeffs[shift + pdivisor->degree];
		multiplier = z3polymod_mul(&rem_leader, &dlinv, pr);
		divisor_with_mul = z3npoly_smul(pdivisor, &multiplier, pr);
		quot.coeffs[shift] = multiplier;
		stop = shift + divisor_with_mul.degree;
		for (i = shift; i <= stop; i++) {
			rem.coeffs[i] = z3poly_sub(
				&rem.coeffs[i],
				&divisor_with_mul.coeffs[i-shift]);
		}
	}

	quot.degree = z3npoly_find_degree(&quot, quot.degree);
	rem.degree  = z3npoly_find_degree(&rem,  rem.degree);

	*pquot = quot;
	*prem  = rem;
}

// ----------------------------------------------------------------
z3npoly_t z3npoly_div(
	z3npoly_t * pin1,
	z3npoly_t * pin2,
	z3poly_t * pr)
{
	z3npoly_t quot, rem;
	z3npoly_quot_and_rem(pin1, pin2, &quot, &rem, pr);
	return quot;
}

// ----------------------------------------------------------------
z3npoly_t z3npoly_mod(
	z3npoly_t * pin1,
	z3npoly_t * pin2,
	z3poly_t * pr)
{
	z3npoly_t quot, rem;
	z3npoly_quot_and_rem(pin1, pin2, &quot, &rem, pr);
	return rem;
}

// ----------------------------------------------------------------
// Example with p=2 and hexadecimal notation:
//
//   mod 13:
//   02 =  0 . . . [  0]
//   03 =  1 . . . [  1]
//   07 =  6 7 . . [  3]
//   13 =  2 3 4 5 [ 15]
//   19 =  9 b d e [ 15]
//   1f =  8 a c f [  5]
//
// 13 is the reductor r.  b is the input z.  19 is the desired minimal
// polynomial.  Fact:  If b is a root of 19, so are all the elements
// of its Frobenius orbit mod r, e.g. b^2, b^4, b^8.  These powers yield
// b, 9, d and e.  Then, (X-b)(X-9)(X-d)(X-e) is the desired minimal
// polynomial with coefficients in Fpn[x], 11001 .  Un-extend to Fp[x]
// to get 19 (in hex).

//         1 b
// *       1 9
// ===========
//         9 c
// +     1 b
// ===========
//       1 2 c
// *       1 d
// ===========
//       d 9 3
// +   1 2 c
// ===========
//     1 f 5 3
// *       1 e
// ===========
//     e 5 3 1
// + 1 f 5 3
// ===========
//   1 1 0 0 1 = 19 in hex.

z3poly_t z3poly_get_min_poly(
	z3poly_t * pz,
	z3poly_t * pr)
{
	int n = z3poly_get_degree(pr);
	int i;
	z3poly_t one = z3poly_from_base_rep(1);
	z3npoly_t extrv = z3npoly_extend_z3poly(&one);
	z3poly_t z3power = *pz;
	z3npoly_t factor;
	z3poly_t rv;

	for (i = 0; i < n; i++) {
		factor.degree = 1;
		factor.coeffs[1] = z3poly_from_base_rep(1);
		factor.coeffs[0] = z3poly_usub(&z3power);
		extrv = z3npoly_mul(&extrv, &factor, pr);
		z3power = z3polymod_power(&z3power, 3, pr);
		if (z3poly_equal(&z3power, pz))
			break;
	}
	rv = z3poly_unextend_z3npoly(&extrv, pr);
	return rv;
}

// ----------------------------------------------------------------
int z3npoly_equal(
	z3npoly_t * pin1,
	z3npoly_t * pin2,
	z3poly_t * pr)
{
	int i;
	if (pin1->degree != pin2->degree)
		return 0;
	for (i = 0; i <= pin1->degree; i++)
		if (!z3poly_equal(&pin1->coeffs[i], &pin2->coeffs[i]))
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int z3npoly_lt(z3npoly_t * pa, z3npoly_t * pb)
{
	return z3npoly_cmp(pa, CMP_LT, pb);
}
int z3npoly_le(z3npoly_t * pa, z3npoly_t * pb)
{
	return z3npoly_cmp(pa, CMP_LE, pb);
}
int z3npoly_gt(z3npoly_t * pa, z3npoly_t * pb)
{
	return z3npoly_cmp(pa, CMP_GT, pb);
}
int z3npoly_ge(z3npoly_t * pa, z3npoly_t * pb)
{
	return z3npoly_cmp(pa, CMP_GE, pb);
}
int z3npoly_eq(z3npoly_t * pa, z3npoly_t * pb)
{
	return z3npoly_cmp(pa, CMP_EQ, pb);
}
int z3npoly_ne(z3npoly_t * pa, z3npoly_t * pb)
{
	return z3npoly_cmp(pa, CMP_NE, pb);
}

// ----------------------------------------------------------------
int z3npoly_cmp(
	z3npoly_t * pa,
	int         cmp,
	z3npoly_t * pb)
{
	int direction = 0; // -1 = less, 0 = equal, +1 = greater;
	int i;

	if (pa->degree > pb->degree) {
		direction =  1;
	}
	else if (pa->degree < pb->degree) {
		direction = -1;
	}
	else {
		for (i = pa->degree; i >= 0; i--) {
			if (z3poly_lt(&pa->coeffs[i], &pb->coeffs[i])) {
				direction = -1;
				break;
			}
			else if (z3poly_gt(&pa->coeffs[i], &pb->coeffs[i])) {
				direction =  1;
				break;
			}
		}
	}

	switch (cmp) {
	case CMP_LE: return (direction <= 0); break;
	case CMP_LT: return (direction <  0); break;
	case CMP_EQ: return (direction == 0); break;
	case CMP_NE: return (direction != 0); break;
	case CMP_GE: return (direction >= 0); break;
	case CMP_GT: return (direction >  0); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); return 0; break;
	}
}

// ----------------------------------------------------------------
z3npoly_t z3npoly_extend_z3poly(
	z3poly_t   * pz3)
{
	int i;
	z3npoly_t rv;

	rv.degree = z3poly_get_degree(pz3);
	for (i = 0; i <= rv.degree; i++) {
		rv.coeffs[i] = z3poly_from_coeffs0(pz3->coeffs[i]);
	}
	return rv;
}

// ----------------------------------------------------------------
z3poly_t z3poly_unextend_z3npoly(
	z3npoly_t * pznp,
	z3poly_t * pr)
{
	int i;
	z3poly_t rv = z3poly_from_base_rep(0);

	for (i = 0; i <= pznp->degree; i++) {
		if (z3poly_get_degree(&pznp->coeffs[i]) > 0) {
			printf("z3npoly:  Can't unextend ");
			z3npoly_print(pznp, pr);
			printf("\n");
			exit(1);
		}
		rv.coeffs[i] = pznp->coeffs[i].coeffs[0];
	}

	return rv;
}

// ----------------------------------------------------------------
z3npoly_t z3npoly_from_coeffs0(
	z3poly_t * pcoeff0)
{
	z3npoly_t rv;
	rv.degree = 0;
	rv.coeffs[0] = *pcoeff0;
	return rv;
}

// ----------------------------------------------------------------
z3npoly_t z3npoly_from_coeffs1(
	z3poly_t * pcoeff1,
	z3poly_t * pcoeff0)
{
	z3npoly_t rv;
	rv.coeffs[1] = *pcoeff1;
	rv.coeffs[0] = *pcoeff0;
	rv.degree = z3npoly_find_degree(&rv, 1);
	return rv;
}

// ----------------------------------------------------------------
z3npoly_t z3npoly_from_int(
	int a)
{
	z3npoly_t rv;
	rv.degree = 0;
	rv.coeffs[0] = z3poly_from_base_rep(a);
	return rv;
}

// ----------------------------------------------------------------
z3npoly_t z3npoly_from_two_ints(
	int a1,
	int a0)
{
	z3npoly_t rv;
	rv.coeffs[1] = z3poly_from_base_rep(a1);
	rv.coeffs[0] = z3poly_from_base_rep(a0);
	rv.degree = z3npoly_find_degree(&rv, 1);
	return rv;
}

// ----------------------------------------------------------------
z3npoly_t z3npoly_deriv(
	z3npoly_t * pz,
	z3poly_t * pr)
{
	z3npoly_t rv;
	int i;

	if (pz->degree == 0) {
		return z3npoly_from_int(0);
	}

	for (i = 1; i <= pz->degree; i++)
		rv.coeffs[i-1] = z3poly_scalar_mul(&pz->coeffs[i], i % 3);
	rv.degree = z3npoly_find_degree(&rv, pz->degree - 1);

	return rv;
}

// ----------------------------------------------------------------
void z3npoly_print(
	z3npoly_t * pin,
	z3poly_t  * pr)
{
	int i;

	for (i = pin->degree; i >= 0; i--) {
		if (i < pin->degree)
			printf(":");
		z3polymod_print(&pin->coeffs[i], pr);
	}
}

// ----------------------------------------------------------------
int z3npoly_list_scan(
	z3npoly_t * pout,
	char      * pin,
	z3poly_t  * pr)
{
	int    i;
	char * coeff_stringv[Z3NPOLY_MAX_DEGREE + 1];
	char   temp_line[1024];

	pout->degree = 0;
	for (i = 0; i <= Z3NPOLY_MAX_DEGREE; i++)
		pout->coeffs[i] = z3poly_from_base_rep(0);

	strcpy(temp_line, pin);
	pout->degree = tokenize(temp_line, ":", coeff_stringv,
		Z3NPOLY_MAX_DEGREE + 1) - 1;
	if (pout->degree < 0) {
		fprintf(stderr, "z3npoly_list_scan:  Couldn't scan \"%s\".\n",
			pin);
		return 0;
	}

	for (i = 0; i <= pout->degree; i++) {
		if (!z3poly_scan(&pout->coeffs[pout->degree - i],
			coeff_stringv[i]))
		{
			fprintf(stderr,
				"z3npoly:  Couldn't scan coefficient \"%s\".\n",
				coeff_stringv[pout->degree - i]);
			return 0;
		}
	}

	// Check for leading zero coefficients, which would reduce the
	// degree.
	pout->degree = z3npoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
int z3npoly_exp_scan(
	z3npoly_t * pout,
	char      * pin,
	z3poly_t  * pr)
{
	int    i;
	int    num_pieces;
	char * coeff_stringv[Z3NPOLY_MAX_DEGREE + 1];
	char * coeff_and_exp[2];
	int    num_sub_pieces;
	char   temp_line[1024];
	char   temp_piece[128];
	int    exp;

	pout->degree = 0;
	for (i = 0; i <= Z3NPOLY_MAX_DEGREE; i++)
		pout->coeffs[i] = z3poly_from_base_rep(0);

	strcpy(temp_line, pin);
	num_pieces = tokenize(temp_line, ":", coeff_stringv,
		Z3NPOLY_MAX_DEGREE + 1);
	if (num_pieces <= 0) {
		fprintf(stderr, "z3npoly_exp_scan:  Couldn't scan \"%s\".\n",
			pin);
		return 0;
	}

	for (i = 0; i < num_pieces; i++) {
		// Split on "^"; expect z3poly on left and int on right.
		// Bounds-check the exponent.
		strcpy(temp_piece, coeff_stringv[i]);
		num_sub_pieces = tokenize(temp_piece, "^", coeff_and_exp, 2);
		if (num_sub_pieces < 1) {
			fprintf(stderr,
				"z3npoly:  Couldn't scan coefficient \"%s\".\n",
				temp_piece);
			return 0;
		}
		else if (num_sub_pieces == 1) {
			exp = 0;
		}
		else {
			// sscanf and bounds-check the exp
			if (sscanf(coeff_and_exp[1], "%d", &exp) != 1) {
				fprintf(stderr,
				"z3npoly:  Couldn't scan coefficient \"%s\".\n",
					temp_piece);
				return 0;
			}
			if ((exp < 0) || (exp > Z3NPOLY_MAX_DEGREE)) {
				fprintf(stderr,
				"z3npoly:  Exponent %d out of range 0-%d.\n",
					exp, Z3NPOLY_MAX_DEGREE);
				return 0;
			}
		}

		// Remember highest exponent encountered.
		if (exp > pout->degree)
			pout->degree = exp;

		// scan and bounds-check the coeff
		if (!z3poly_scan(&pout->coeffs[exp], coeff_and_exp[0])) {
			fprintf(stderr,
				"z3npoly:  Couldn't scan coefficient \"%s\".\n",
				temp_piece);
			return 0;
		}
	}

	// Check for leading zero coefficients, which would reduce the
	// degree.
	pout->degree = z3npoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
int z3npoly_scan(
	z3npoly_t * pout,
	char      * pin,
	z3poly_t  * pr)
{
	if (strchr(pin, '^'))
		return z3npoly_exp_scan(pout, pin, pr);
	else
		return z3npoly_list_scan(pout, pin, pr);
}
