#include <stdio.h>
#include <stdlib.h>
#include "malloc_check.h"
#include "power.h"
#include "z3npolymod.h"
#include "z3npoly_factor.h"
#include "z3nlinalg.h"

static void gather_multiple_factors(
	z3npoly_t  * pirred_factors,
	z3poly_t   * pr,
	int           num_irred_factors,
	z3npoly_factor_info_t * pfinfo);

static void sort_up(
	z3npoly_t  * polys,
	int          nrows);

static void pre_berlekamp2ni(
	z3npoly_t  * pz,
	z3poly_t   * pr,
	z3npoly_t  * pirred_factors,
	int        * pnum_irred_factors);

static void berlekamp2ni(
	z3npoly_t  * pz,
	z3poly_t   * pr,
	z3npoly_t  * pirred_factors,
	int        * pnum_irred_factors);

// ----------------------------------------------------------------
void z3npoly_factor(
	z3npoly_t  * pz,
	z3poly_t   * pr,
	z3npoly_factor_info_t * pfinfo)
{
	int num_irred_factors, i;
	z3npoly_t * irred_factors = malloc_check(
		Z3NPOLY_MAX_NUM_FACTORS * sizeof(z3npoly_t));
	z3poly_t leader;

	num_irred_factors = 0;
	pre_berlekamp2ni(pz, pr, irred_factors, &num_irred_factors);

	for (i = 0; i < num_irred_factors; i++) {
		z3npoly_make_monic(&irred_factors[i], pr);
	}

	leader = pz->coeffs[pz->degree];
	if (!z3poly_is_zero(&leader) && !z3poly_is_one(&leader)) {
		irred_factors[num_irred_factors++] =
			z3npoly_from_coeffs0(&leader);
	}

	sort_up(irred_factors, num_irred_factors);
	gather_multiple_factors(irred_factors, pr, num_irred_factors, pfinfo);

	free(irred_factors);
}

// ----------------------------------------------------------------
static void gather_multiple_factors(
	z3npoly_t * pirred_factors,
	z3poly_t   * pr,
	int          num_irred_factors,
	z3npoly_factor_info_t * pfinfo)
{
	int i;
	z3npoly_t  * plast;
	int nd = 0;

	if (num_irred_factors == 0) {
		pfinfo->num_distinct = 0;
		return;
	}

	plast = &pirred_factors[0];
	pfinfo->factor_counts[nd].f = *plast;
	pfinfo->factor_counts[nd].count = 0;

	for (i = 0; i < num_irred_factors; i++) {
		if (!z3npoly_equal(&pirred_factors[i], plast, pr)) {
			nd++;
			pfinfo->factor_counts[nd].count = 1;
			pfinfo->factor_counts[nd].f = pirred_factors[i];
			plast = &pirred_factors[i];
		}
		else {
			pfinfo->factor_counts[nd].count++;
		}
	}
	pfinfo->num_distinct = nd + 1;
}

// ----------------------------------------------------------------
static void pre_berlekamp2ni(
	z3npoly_t  * pz,
	z3poly_t   * pr,
	z3npoly_t  * pirred_factors,
	int         * pnum_irred_factors)
{
	z3npoly_t d, g;

	d = z3npoly_deriv(pz, pr);
	g = z3npoly_gcd(pz, &d, pr);

	if (z3npoly_is_zero(&g)) {
		if (!z3npoly_is_zero(pz)) {
			fprintf(stderr, "CODING ERROR file %s\n", __FILE__);
			exit(1);
		}
		pirred_factors[*pnum_irred_factors] = *pz;
		(*pnum_irred_factors)++;
		return;
	}
	else if (g.degree == 0) {
		// Input is squarefree.
		berlekamp2ni(pz, pr, pirred_factors, pnum_irred_factors);
	}
	else if (z3npoly_is_zero(&d)) {
		// Input is a perfect cube.
		// Multiplicity is p only if degree is > 0.
		z3npoly_t s = z3npoly_cube_root(pz, pr);
		int i;
		int multiplicity = pz->degree > 0 ? 3 : 1;

		for (i = 0; i < multiplicity; i++) {
			pre_berlekamp2ni(&s, pr, pirred_factors,
				pnum_irred_factors);
		}
	}
	else {
		z3npoly_t q;
		q = z3npoly_div(pz, &g, pr);
		pre_berlekamp2ni(&g, pr, pirred_factors, pnum_irred_factors);
		pre_berlekamp2ni(&q, pr, pirred_factors, pnum_irred_factors);
	}
}

// ----------------------------------------------------------------
static void berlekamp2ni(
	z3npoly_t  * pz,
	z3poly_t   * pr,
	z3npoly_t  * pirred_factors,
	int         * pnum_irred_factors)
{
	int n = pz->degree;
	int q = calc_power(3, z3poly_get_degree(pr));
	z3npoly_t x  = z3npoly_from_two_ints(1, 0);
	z3poly_t one = z3poly_from_base_rep(1);
	z3npoly_t xq;
	z3npoly_t xqi = z3npoly_from_int(1);
	int row, rank, dimker;
	int ci;
	int got_it = 0;

	z3poly_t    ** BI;
	z3poly_t    ** BIT;
	z3npoly_t * nullspace_basis;
	z3npoly_t   h, f1, f2;

	if (n < 2) {
		pirred_factors[*pnum_irred_factors] = *pz;
		(*pnum_irred_factors)++;
		return;
	}

	BI  = z3nla_alloc_matrix(n, n);
	BIT = z3nla_alloc_matrix(n, n);
	nullspace_basis = malloc_check(n * sizeof(z3npoly_t));

	xq = z3npolymod_power(&x, q, pz, pr);

	// Populate the B matrix with row i equal to x^pi mod z
	for (row = 0; row < n; row++) {
		z3nla_row_from_z3npoly(BI[row], &xqi, n);
		xqi = z3npolymod_mul(&xqi, &xq, pz, pr);
	}

	// Replace B with B - I
	for (row = 0; row < n; row++) {
		BI[row][row] = z3polymod_sub(&BI[row][row], &one, pr);
	}

	z3nla_transpose(BIT, BI, n);
	z3nla_row_echelon_form(BIT, pr, n, n);
	rank = z3nla_find_rank(BIT, n, n);
	dimker = n - rank;

	if (dimker == 1) {
		pirred_factors[*pnum_irred_factors] = *pz;
		(*pnum_irred_factors)++;
		return;
	}

	z3nla_find_nullspace_basis(BIT, n, n, pr, nullspace_basis, dimker);

#if 1
	// Check h^q = h.  The algorithm guarantees this, *if* implemented
	// correctly.  This computation, while not strictly necessary, is a
	// nice sanity check.
	for (row = 0; row <  dimker; row++) {
		z3npoly_t * ph = &nullspace_basis[row];
		z3npoly_t hp;
		hp = z3npolymod_power(ph, q, pz, pr);
		if (!z3npoly_equal(ph, &hp, pr)) {
			fprintf(stderr, "CODING ERROR\n");
			exit(1);
		}
	}
#endif

	got_it = 0;
	for (row = 0; row <  dimker && !got_it; row++) {
		if (z3npoly_is_one(&nullspace_basis[row]))
			continue;

		for (ci = 0; ci < q; ci++) {
			z3npoly_t cpoly = z3npoly_from_int(ci);
			h = z3npoly_sub(&nullspace_basis[row], &cpoly, pr);
			f1 = z3npoly_gcd(pz, &h, pr);

			if ((0 < f1.degree) && (f1.degree < pz->degree)) {
				got_it = 1;
				break;
			}
		}
	}

	if (!got_it) {
		printf("CODING ERROR: no non-trivial factors found: ");
		z3npoly_print(pz, pr);
		printf("\n");
		exit(1);
	}

	f2 = z3npoly_div(pz, &f1, pr);

	// The nullity of B-I is the number of irreducible
	// factors of pr.  If the nullity is 2, we have a
	// pair of factors which are both irreducible and
	// so we don't need to recurse.
	if (dimker == 2) {
		pirred_factors[*pnum_irred_factors] = f1;
		(*pnum_irred_factors)++;
		pirred_factors[*pnum_irred_factors] = f2;
		(*pnum_irred_factors)++;
	}
	else {
		pre_berlekamp2ni(&f1, pr, pirred_factors, pnum_irred_factors);
		pre_berlekamp2ni(&f2, pr, pirred_factors, pnum_irred_factors);
	}

	z3nla_free_matrix(BI,  n, n);
	z3nla_free_matrix(BIT, n, n);
	free(nullspace_basis);
}

// ----------------------------------------------------------------
// Ascending bubble sort
static void sort_up(
	z3npoly_t * polys,
	int          nrows)
{
	int i, top;
	z3npoly_t  temp;

	for (top = nrows-1; top > 0; top--) {
		for (i = 0; i < top; i++) {
			if (!z3npoly_lt(&polys[i], &polys[i+1])) {
				temp       = polys[i];
				polys[i]   = polys[i+1];
				polys[i+1] = temp;
			}
		}
	}
}

// ----------------------------------------------------------------
// Given the prime factorization p1^m1 ... pk^mk of n, how to enumerate all
// factors of n?
//
// Example 72 = 2^3 * 3^2.  Exponent on 2 is one of 0, 1, 2, 3.  Exponent on
// 3 is one of 0, 1, 2.  Number of possibilities:  product over i of (mi + 1).
// Factors are:
//
//	2^0 3^0    1
//	2^0 3^1    3
//	2^0 3^2    9
//	2^1 3^0    2
//	2^1 3^1    6
//	2^1 3^2   18
//	2^2 3^0    4
//	2^2 3^1   12
//	2^2 3^2   36
//	2^3 3^0    8
//	2^3 3^1   24
//	2^3 3^2   72

unsigned z3npoly_num_divisors(
	z3npoly_factor_info_t * pfinfo)
{
	unsigned rv;
	int i;

	if ((pfinfo->num_distinct < 0)
		|| (pfinfo->num_distinct > Z3NPOLY_MAX_NUM_FACTORS))
	{
		fprintf(stderr, "z3npoly_num_divisors:  num distinct (%d) "
			"out of bounds 0-%d\n",
			pfinfo->num_distinct, Z3NPOLY_MAX_NUM_FACTORS);
		exit(1);
	}

	rv = 1;
	for (i = 0; i < pfinfo->num_distinct; i++)
		rv *= pfinfo->factor_counts[i].count + 1;
	return rv;
}

// ----------------------------------------------------------------
// See comments to the above.  k is treated as a multibase representation
// over the bases mi+1.

z3npoly_t z3npoly_kth_divisor(
	z3npoly_factor_info_t * pfinfo,
	z3poly_t   * pr,
	unsigned  k)
{
	z3npoly_t rv;
	z3npoly_t fp;
	int        i;
	unsigned   base;
	unsigned   power;

	rv = z3npoly_from_int(1);
	for (i = 0; i < pfinfo->num_distinct; i++) {
		base = pfinfo->factor_counts[i].count + 1;
		power = k % base;
		k = k / base;

		fp = z3npoly_power(&pfinfo->factor_counts[i].f, power, pr);
		rv = z3npoly_mul(&rv, &fp, pr);
	}
	return rv;
}

// ----------------------------------------------------------------
static int z3npoly_qcmp(const void * pv1, const void * pv2)
{
	if (z3npoly_lt((z3npoly_t*)pv1, (z3npoly_t*)pv2))
		return -1;
	if (z3npoly_gt((z3npoly_t*)pv1, (z3npoly_t*)pv2))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
// The caller must free the return value.

z3npoly_t * z3npoly_get_all_divisors(
	z3npoly_t  * pz,
	z3poly_t   * pr,
	unsigned    * pnum_divisors)
{
	z3npoly_factor_info_t finfo;
	unsigned nf, k;
	z3npoly_t * all_divisors;

	z3npoly_factor(pz, pr, &finfo);
	nf = z3npoly_num_divisors(&finfo);
	*pnum_divisors = nf;

	all_divisors = (z3npoly_t *)malloc_check(nf * sizeof(z3npoly_t));
	for (k = 0; k < nf; k++) {
		all_divisors[k] = z3npoly_kth_divisor(&finfo, pr, k);
	}
	qsort(all_divisors, nf, sizeof(z3npoly_t), z3npoly_qcmp);

	return all_divisors;
}

// ----------------------------------------------------------------
// The caller must free the return value.
z3poly_t * z3poly_get_roots(
	z3poly_t * pz,
	z3poly_t * pr)
{
	z3npoly_t zx = z3npoly_extend_z3poly(pz);
	z3npoly_factor_info_t finfo;
	int n = z3poly_get_degree(pz);
	int i;
	z3poly_t * rv = malloc_check(n * sizeof(z3poly_t));

	z3npoly_factor(&zx, pr, &finfo);

	if (finfo.num_distinct != n) {
		printf("z3poly_get_roots: coding error: split failure\n");
		exit(1);
	}

	for (i = 0; i < n; i++) {
		if (finfo.factor_counts[i].count != 1) {
			printf("z3poly_get_roots: coding error: "
				"split failure\n");
			exit(1);
		}
		if (finfo.factor_counts[i].f.degree != 1) {
			printf("z3poly_get_roots: coding error: "
				"split failure\n");
			exit(1);
		}
		rv[i] = z3poly_usub(&finfo.factor_counts[i].f.coeffs[0]);
	}

	return rv;
}

// ----------------------------------------------------------------
// The caller must free the return value.
z3poly_t * z3polymod_get_elements_frobord(
	z3poly_t * pr,
	int * pnel)
{
	int n = z3poly_get_degree(pr);
	int nel = calc_power(3, n);
	z3poly_t * rv = malloc_check(nel * sizeof(z3poly_t));
	z3poly_t * dirreds = malloc_check(nel * sizeof(z3poly_t));
	z3poly_t * roots, f0, fcurr;
	int nirred;
	int i, didx = 0;

	// Factor X^(p^n) - X into irreducible factors over Fp[x].
	z3poly_find_dirreds(n, dirreds, &nirred, nel);

	// For each irreducible factor of X^(p^n) - X, get its roots.
	for (i = 0; i < nirred; i++) {
		roots = z3poly_get_roots(&dirreds[i], pr);

		// Get frobit of each root.
		f0 = roots[0];
		fcurr = f0;
		do {
			if (didx >= nel) {
				fprintf(stderr, "FROBORD ABEND!\n");
				exit(1);
			}
			rv[didx++] = fcurr;
			fcurr = z3polymod_power(&fcurr, 3, pr);
		} while (!z3poly_equal(&fcurr, &f0));
	}

	free(dirreds);
	*pnel = nel;
	return rv;
}
