// NOTE:  all the "x & 3" business in here is NOT intended to compute residues
// mod 3.  (In fact, it computes residues mod 4.)  Integer division is
// (relatively) expensive and is something I'm trying to avoid, replacing with
// table lookups.  The &3 is intended to protect against out-of-bounds accesses
// to those tables -- doing the &3 on table indices merely ensures that when
// this code is called with garbage data, the results will more likely be
// informative and less likely result in memory corruption.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "z3poly.h"
#include "mod.h"
#include "power.h"
#include "factorize.h"
#include "tokenize.h"
#include "posmod.h"
#include "malloc_check.h"

#define Z3POLY_NUM_TRIPLES ((Z3POLY_MAX_DEGREE + 3)/3)

#define BASE_27_IO

unsigned char z3_add_tbl[4][4] = {
	//0  1  2
	{ 0, 1, 2, 3 }, // 0
	{ 1, 2, 0, 3 }, // 1
	{ 2, 0, 1, 3 }, // 2
	{ 3, 3, 3, 3 },
};

unsigned char z3_sub_tbl[4][4] = {
	//0  1  2
	{ 0, 2, 1, 3 }, // 0
	{ 1, 0, 2, 3 }, // 1
	{ 2, 1, 0, 3 }, // 2
	{ 3, 3, 3, 3 },
};

unsigned char z3_usub_tbl[4] = {
	0, 2, 1, 3
};

unsigned char z3_mul_tbl[4][4] = {
	//0  1  2
	{ 0, 0, 0, 3 }, // 0
	{ 0, 1, 2, 3 }, // 1
	{ 0, 2, 1, 3 }, // 2
	{ 3, 3, 3, 3 },
};

unsigned char z3_div_tbl[4][4] = {
	//0  1  2
	{ 3, 0, 0, 3 }, // 0
	{ 3, 1, 2, 3 }, // 1
	{ 3, 2, 1, 3 }, // 2
	{ 3, 3, 3, 3 },
};

char z3_out3_tbl[4][4][4] = {
	{{'0' /*000*/, '1' /*001*/, '2' /*002*/, '?' },
	{ '3' /*010*/, '4' /*011*/, '5' /*012*/, '?' },
	{ '6' /*020*/, '7' /*021*/, '8' /*022*/, '?' },
	{ '?'        , '?'        , '?'        , '?' }},

	{{'a' /*100*/, 'b' /*101*/, 'c' /*102*/, '?' },
	{ 'd' /*110*/, 'e' /*111*/, 'f' /*112*/, '?' },
	{ 'g' /*120*/, 'h' /*121*/, 'i' /*122*/, '?' },
	{ '?'        , '?'        , '?'        , '?' }},

	{{'n' /*200*/, 'o' /*201*/, 'p' /*202*/, '?' },
	{ 'q' /*210*/, 'r' /*211*/, 's' /*212*/, '?' },
	{ 't' /*220*/, 'u' /*221*/, 'v' /*022*/, '?' },
	{ '?'        , '?'        , '?'        , '?' }},

	{{'?'        , '?'        , '?'        , '?' },
	{ '?'        , '?'        , '?'        , '?' },
	{ '?'        , '?'        , '?'        , '?' },
	{ '?'        , '?'        , '?'        , '?' }},
};

// ----------------------------------------------------------------
z3poly_t z3poly_add(
	z3poly_t * pin1,
	z3poly_t * pin2)
{
	z3poly_t out;
	int i;

	for (i = 0; i <= Z3POLY_MAX_DEGREE; i++)
		out.coeffs[i] = Z3_ADD(pin1->coeffs[i], pin2->coeffs[i]);
	return out;
}

// ----------------------------------------------------------------
z3poly_t z3poly_sub(
	z3poly_t * pin1,
	z3poly_t * pin2)
{
	z3poly_t out;
	int i;

	for (i = 0; i <= Z3POLY_MAX_DEGREE; i++)
		out.coeffs[i] = Z3_SUB(pin1->coeffs[i], pin2->coeffs[i]);
	return out;
}

// ----------------------------------------------------------------
z3poly_t z3poly_usub(
	z3poly_t * pin)
{
	z3poly_t out;
	int i;

	for (i = 0; i <= Z3POLY_MAX_DEGREE; i++)
		out.coeffs[i] = Z3_USUB(pin->coeffs[i]);
	return out;
}

// ----------------------------------------------------------------
z3poly_t z3poly_mul(
	z3poly_t * pin1,
	z3poly_t * pin2)
{
	z3poly_t  out;
	int i, j;
	int prod;
	int deg1 = z3poly_get_degree(pin1);
	int deg2 = z3poly_get_degree(pin2);

	z3poly_check_degree(deg1 + deg2);

	for (i = 0; i <= Z3POLY_MAX_DEGREE; i++)
		out.coeffs[i] = 0;

	for (i = 0; i <= deg1; i++) {
		for (j = 0; j <= deg2; j++) {
			prod = Z3_MUL(pin1->coeffs[i], pin2->coeffs[j]);
			out.coeffs[i+j] = Z3_ADD(out.coeffs[i+j], prod);
		}
	}
	return out;
}

// ----------------------------------------------------------------
z3poly_t z3poly_scalar_mul(
	z3poly_t * pin1,
	int       in2)
{
	z3poly_t  out;
	int i;

	for (i = 0; i <= Z3POLY_MAX_DEGREE; i++)
		out.coeffs[i] = Z3_MUL(pin1->coeffs[i], in2);
	return out;
}

// ----------------------------------------------------------------
z3poly_t z3poly_div(
	z3poly_t * pin1,
	z3poly_t * pin2)
{
	z3poly_t  quot, rem;
	z3poly_quot_and_rem(pin1, pin2, &quot, &rem);
	return quot;
}

// ----------------------------------------------------------------
z3poly_t z3poly_mod(
	z3poly_t * pin1,
	z3poly_t * pin2)
{
	z3poly_t  quot, rem;
	z3poly_quot_and_rem(pin1, pin2, &quot, &rem);
	return rem;
}

// ----------------------------------------------------------------
// E.g.
// dividend = 1,2,3,4 (1 + 2x + 3x^2 + 4x^3)
// divisor  = 1,1,2   (1 +  x + 2x^2)
// modulus = 7
//
//             q=4,2   r = 4,3
//        +----------
// 1,1,2  |  1,2,3,4
//        |    2,2,4 shift = 1.  4/2 mod 7 = 2.  1,1,2 * 2 = 2,2,4.
//        +----------
//        |  1 0 1
//        |  4 4 1   shift = 0.  1/2 mod 7 = 4   1,1,2 * 4 = 4,4,1
//        +----------
//        |  4 3

void z3poly_quot_and_rem(
	z3poly_t * pdividend,
	z3poly_t * pdivisor,
	z3poly_t * pquot,
	z3poly_t * prem)
{
	z3poly_t quot, rem;
	int dividend_degree = z3poly_get_degree(pdividend);
	int divisor_degree  = z3poly_get_degree(pdivisor);
	int max_shift;
	unsigned char divisor_leader, dlinv, rem_leader, multiplier;
	int shift, i;
	z3poly_t divisor_with_mul;
	int stop;

	quot = z3poly_from_base_rep(0);
	rem = *pdividend;
	max_shift = dividend_degree - divisor_degree;
	if (dividend_degree < divisor_degree) {
		*pquot = quot;
		*prem  = rem;
		return;
	}

	divisor_leader = pdivisor->coeffs[divisor_degree];
	dlinv = divisor_leader; // Z3 is self-reciprocating
	for (shift = max_shift; shift >= 0; shift--) {
		rem_leader = rem.coeffs[shift + divisor_degree];
		multiplier = Z3_MUL(rem_leader, dlinv);
		divisor_with_mul = z3poly_scalar_mul(pdivisor, multiplier);
		quot.coeffs[shift] = multiplier;

		stop = shift + divisor_degree;
		for (i = shift; i <= stop; i++) {
			rem.coeffs[i] = z3_sub_tbl
				[rem.coeffs[i]]
				[divisor_with_mul.coeffs[i-shift]];
		}
	}

	*pquot = quot;
	*prem  = rem;
}

// ----------------------------------------------------------------
// Input: a, b
// Output: GCD g, and m, n such that am + bn = d

z3poly_t z3poly_ext_gcd(
	z3poly_t * pa, // 1st input.
	z3poly_t * pb, // 2nd input.
	z3poly_t * pm, // Output coefficient on a
	z3poly_t * pn) // Output coefficient on b
{
	z3poly_t mprime, nprime, m, n, c, q, r, t, qm, qn;
	z3poly_t d;   // Return value.

	// Initialize
	mprime = z3poly_from_base_rep(1);
	n      = z3poly_from_base_rep(1);
	m      = z3poly_from_base_rep(0);
	nprime = z3poly_from_base_rep(0);
	c = *pa;
	d = *pb;

	while (1) {

		// Divide
		// q = c / d, r = c % d
		z3poly_quot_and_rem(&c, &d, &q, &r);
		// Note:  now c = qd + r and 0 <= r < d

		// Remainder zero?
		if (z3poly_is_zero(&r))
			break;

		// Recycle
		c = d;
		d = r;

		t = mprime;
		mprime = m;
		//qm = q * m;
		qm = z3poly_mul(&q, &m);
		//m = t - qm;
		m = z3poly_sub(&t, &qm);

		t = nprime;
		nprime = n;
		//qn = q * n;
		qn = z3poly_mul(&q, &n);
		//n = t - qn;
		n = z3poly_sub(&t, &qn);
	}
	*pm = m;
	*pn = n;
	return d;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
z3poly_t z3poly_power(
	z3poly_t * pin,
	int        power)
{
	z3poly_t in2 = *pin;
	z3poly_t inpower = z3poly_from_base_rep(1);

	while (power != 0) {
		if (power & 1)
			inpower = z3poly_mul(&in2, &inpower);
		power >>= 1;
		in2 = z3poly_mul(&in2, &in2);
	}
	return inpower;
}

// ----------------------------------------------------------------
// Assumes the input is already known to be a cube.  Uses the
// fact that f(x^p) = f^p(x) over Fp[x].

z3poly_t z3poly_cube_root(
	z3poly_t * pin)
{
	int si, di, j;
	int degin = z3poly_get_degree(pin);
	z3poly_t out = z3poly_from_base_rep(0);

	for (si = 0, di = 0; si <= degin; si += 3, di++) {
		out.coeffs[di] = pin->coeffs[si];
		for (j = 1; j < 3; j++) {
			if (si+j > degin)
				break;
			if (pin->coeffs[si+j] != 0) {
				printf("z3poly_cube_root: ");
				z3poly_print(pin);
				printf(" is not a cube.\n");
				exit(1);
			}
		}
	}

	return out;
}

// ----------------------------------------------------------------
// Input: a, b
// Output: GCD g, and m, n such that am + bn = d

z3poly_t z3poly_gcd(
	z3poly_t * pa, // 1st input.
	z3poly_t * pb) // 2nd input.
{
	z3poly_t c, q, r;
	z3poly_t d;   // Return value.

	if (z3poly_is_zero(pa))
		return *pb;
	if (z3poly_is_zero(pb))
		return *pa;

	c = *pa;
	d = *pb;

	while (1) {

		// Divide
		// q = c / d, r = c % d
		z3poly_quot_and_rem(&c, &d, &q, &r);
		// Note:  now c = qd + r and 0 <= r < d

		// Remainder zero?
		if (z3poly_is_zero(&r)) {
			break;
		}

		// Recycle
		c = d;
		d = r;
	}
	return d;
}

// ----------------------------------------------------------------
int z3poly_equal(
	z3poly_t * pin1,
	z3poly_t * pin2)
{
	int i;
	for (i = 0; i <= Z3POLY_MAX_DEGREE; i++)
		if (pin1->coeffs[i] != pin2->coeffs[i])
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int  z3poly_lt(z3poly_t * pa, z3poly_t * pb)
{
	return z3poly_cmp(pa, CMP_LT, pb);
}
int  z3poly_le(z3poly_t * pa, z3poly_t * pb)
{
	return z3poly_cmp(pa, CMP_LE, pb);
}
int  z3poly_gt(z3poly_t * pa, z3poly_t * pb)
{
	return z3poly_cmp(pa, CMP_GT, pb);
}
int  z3poly_ge(z3poly_t * pa, z3poly_t * pb)
{
	return z3poly_cmp(pa, CMP_GE, pb);
}
int  z3poly_eq(z3poly_t * pa, z3poly_t * pb)
{
	return z3poly_cmp(pa, CMP_EQ, pb);
}
int  z3poly_ne(z3poly_t * pa, z3poly_t * pb)
{
	return z3poly_cmp(pa, CMP_NE, pb);
}

// ----------------------------------------------------------------
int z3poly_cmp(
	z3poly_t * pa,
	int       cmp,
	z3poly_t * pb)
{
	int direction = 0; // -1 = less, 0 = equal, +1 = greater;
	int i;

	for (i = Z3POLY_MAX_DEGREE; i >= 0; i--) {
		if (pa->coeffs[i] < pb->coeffs[i]) {
			direction = -1;
			break;
		}
		else if (pa->coeffs[i] > pb->coeffs[i]) {
			direction =  1;
			break;
		}
	}

	switch (cmp) {
	case CMP_LE: return (direction <= 0); break;
	case CMP_LT: return (direction <  0); break;
	case CMP_EQ: return (direction == 0); break;
	case CMP_NE: return (direction != 0); break;
	case CMP_GE: return (direction >= 0); break;
	case CMP_GT: return (direction >  0); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); return 0; break;
	}
}

// ----------------------------------------------------------------
z3poly_t z3poly_from_base_rep(
	int base_rep)
{
	z3poly_t rv;
	int i;

	for (i = 0; i <= Z3POLY_MAX_DEGREE; i++)
		rv.coeffs[i] = 0;

	for (i = 0; ; i++) {
		rv.coeffs[i] = base_rep % 3;
		base_rep /= 3;
		if (base_rep == 0)
			break;
		if (i > Z3POLY_MAX_DEGREE)
			z3poly_check_degree(i);
	}

	return rv;
}

// ----------------------------------------------------------------
int z3poly_to_base_rep(
	z3poly_t * pz)
{
	int rv = 0;
	int i;

	for (i = Z3POLY_MAX_DEGREE; i >= 0; i--) {
		rv *= 3;
		rv += pz->coeffs[i];
	}

	return rv;
}

// ----------------------------------------------------------------
z3poly_t z3poly_from_coeffs(
	int   degree,
	int * coeffs)
{
	int i;
	z3poly_t rv;

	z3poly_check_degree(degree);
	for (i = 0; i <= degree; i++)
		rv.coeffs[i] = posmod(coeffs[i], 3);
	for (i = degree + 1; i <= Z3POLY_MAX_DEGREE; i++)
		rv.coeffs[i] = 0;
	return rv;
}

// ----------------------------------------------------------------
z3poly_t z3poly_from_coeffs0(
	int   coeff0)
{
	z3poly_t rv = z3poly_from_base_rep(0);
	rv.coeffs[0] = posmod(coeff0, 3);
	return rv;
}

// ----------------------------------------------------------------
z3poly_t z3poly_from_coeffs1(
	int   coeff1,
	int   coeff0)
{
	z3poly_t rv = z3poly_from_base_rep(0);
	rv.coeffs[1] = posmod(coeff1, 3);
	rv.coeffs[0] = posmod(coeff0, 3);
	return rv;
}

// ----------------------------------------------------------------
z3poly_t z3poly_from_coeffs2(
	int   coeff2,
	int   coeff1,
	int   coeff0)
{
	z3poly_t rv = z3poly_from_base_rep(0);
	rv.coeffs[2] = posmod(coeff2, 3);
	rv.coeffs[1] = posmod(coeff1, 3);
	rv.coeffs[0] = posmod(coeff0, 3);
	return rv;
}

// ----------------------------------------------------------------
z3poly_t z3poly_monomial_from_coeff(
	int   coeffa,
	int   dega)
{
	z3poly_t rv = z3poly_from_base_rep(0);

	z3poly_check_degree(dega);
	rv.coeffs[dega] = coeffa;
	return rv;
}

// ----------------------------------------------------------------
z3poly_t z3poly_binomial_from_coeffs(
	int   coeffa,
	int   dega,
	int   coeffb,
	int   degb)
{
	z3poly_t rv = z3poly_from_base_rep(0);

	z3poly_check_degree(dega);
	z3poly_check_degree(degb);
	rv.coeffs[dega] += coeffa;
	rv.coeffs[degb] += coeffb;
	return rv;
}

// ----------------------------------------------------------------
int z3poly_scan_base_27(
	z3poly_t * pout,
	char     * pin)
{
	int srci, dsti;

	for (dsti = 0; dsti <= Z3POLY_MAX_DEGREE; dsti++)
		pout->coeffs[dsti] = 0;
	for (srci = strlen(pin) - 1, dsti = 0; srci >= 0; srci--, dsti += 3) {
		if ((dsti + 2) > Z3POLY_MAX_DEGREE) {
			fprintf(stderr,
				"z3poly:  input overflow: \"%s\".\n",
				pin);
			exit(1);
		}

		// .........
		// .....1s3t
		// 876543210
		// 012345678
		// 100112022

		if (('0' <= pin[srci]) && (pin[srci] <= '8')) {
			pout->coeffs[dsti+2] = 0;
			pout->coeffs[dsti+1] = (pin[srci] - '0') / 3;
			pout->coeffs[dsti  ] = (pin[srci] - '0') % 3;
		}
		else if (('a' <= pin[srci]) && (pin[srci] <= 'i')) {
			pout->coeffs[dsti+2] = 1;
			pout->coeffs[dsti+1] = (pin[srci] - 'a') / 3;
			pout->coeffs[dsti  ] = (pin[srci] - 'a') % 3;
		}
		else if (('n' <= pin[srci]) && (pin[srci] <= 'v')) {
			pout->coeffs[dsti+2] = 2;
			pout->coeffs[dsti+1] = (pin[srci] - 'n') / 3;
			pout->coeffs[dsti  ] = (pin[srci] - 'n') % 3;
		}
		else {
			return 0;
		}
	}

	return 1;
}

// ----------------------------------------------------------------
int z3poly_scan_base_3(
	z3poly_t * pout,
	char     * pin)
{
	int srci, dsti;

	for (dsti = 0; dsti <= Z3POLY_MAX_DEGREE; dsti++)
		pout->coeffs[dsti] = 0;
	for (srci = strlen(pin) - 1, dsti = 0; srci >= 0; srci--, dsti++) {
		if (dsti > Z3POLY_MAX_DEGREE) {
			fprintf(stderr,
				"z3poly:  input overflow: \"%s\".\n",
				pin);
			exit(1);
		}

		if (('0' <= pin[srci]) && (pin[srci] <= '2')) {
			pout->coeffs[dsti] = pin[srci] - '0';
		}
		else {
			return 0;
		}
	}

	return 1;
}

// ----------------------------------------------------------------
int z3poly_scan(
	z3poly_t * pout,
	char     * pin)
{
#ifdef BASE_27_IO
	return z3poly_scan_base_27(pout, pin);
#else
	return z3poly_scan_base_3(pout, pin);
#endif
}

// ----------------------------------------------------------------
void z3poly_dprint_base_3(
	z3poly_t * pin,
	int deg)
{
	int i;
	for (i = deg; i >= 0; i--) {
		printf("%u", pin->coeffs[i]);
	}
}

// ----------------------------------------------------------------
void z3poly_dprint_base_27(
	z3poly_t * pin,
	int deg)
{
	int i;
	int trideg = (deg / 3) * 3;

	for (i = trideg; i >= 0; i -= 3) {
		printf("%c", z3_out3_tbl
			[pin->coeffs[i+2] & 3]
			[pin->coeffs[i+1] & 3]
			[pin->coeffs[i  ] & 3]);
	}
}

// ----------------------------------------------------------------
void z3poly_print_base_27(
	z3poly_t * pin)
{
	z3poly_dprint_base_27(pin, z3poly_get_degree(pin));
}

// ----------------------------------------------------------------
void z3poly_print_base_3(
	z3poly_t * pin)
{
	z3poly_dprint_base_3(pin, z3poly_get_degree(pin));
}

// ----------------------------------------------------------------
void z3poly_dprint(
	z3poly_t * pin,
	int deg)
{
#ifdef BASE_27_IO
	z3poly_dprint_base_27(pin, deg);
#else
	z3poly_dprint_base_3(pin, deg);
#endif
}

// ----------------------------------------------------------------
void z3poly_print(
	z3poly_t * pin)
{
#ifdef BASE_27_IO
	z3poly_print_base_27(pin);
#else
	z3poly_print_base_3(pin);
#endif
}

// ----------------------------------------------------------------
//     a_4 x^4 + a_3  x^3 + a_2  x^2 + a_1  x + a_0
//    (a_4 x^3 + a_3  x^2 + a_2  x   + a_1) x + a_0
//   ((a_4 x^2 + a_3  x   + a_2) x   + a_1) x + a_0
//  (((a_4 x   + a_3) x   + a_2) x   + a_1) x + a_0

int z3poly_eval(
	z3poly_t * pz,
	int        x)
{
	int rv;
	int d;
	int deg = z3poly_get_degree(pz);

	rv = pz->coeffs[deg];
	for (d = deg - 1; d >= 0; d--) {
		rv = Z3_MUL(rv, x);
		rv = Z3_ADD(rv, pz->coeffs[d]);
	}
	return rv;
}

// ----------------------------------------------------------------
// r(x) = 2x^2 + x + 1 (p = 3)
// z(x) = x + 2
// r(z(x)) = 2(x+2)^2 + (x+2) + 1

z3poly_t z3poly_compose(
	z3poly_t * pr,
	z3poly_t * pz)
{
	z3poly_t rz = z3poly_from_base_rep(0);
	z3poly_t zpower = z3poly_from_base_rep(1);
	z3poly_t czpower;
	int d;
	int degr = z3poly_get_degree(pr);

	for (d = 0; d <= degr; d++) {
		czpower = z3poly_scalar_mul(&zpower, pr->coeffs[d]);
		rz = z3poly_add(&rz, &czpower);
		zpower = z3poly_mul(&zpower, pz);
	}

	return rz;
}

// ----------------------------------------------------------------
z3poly_t z3poly_deriv(
	z3poly_t * pz)
{
	z3poly_t rv = z3poly_from_base_rep(0);
	int i;

	for (i = 1; i <= Z3POLY_MAX_DEGREE; i++)
		rv.coeffs[i-1] = Z3_MUL(pz->coeffs[i], i%3);

	return rv;
}

// ----------------------------------------------------------------
int z3poly_is_rootless(z3poly_t * pz)
{
	int x;
	for (x = 0; x < 3; x++)
		if (z3poly_eval(pz, x) == 0)
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int z3poly_count_non_zero_terms(z3poly_t * pz)
{
	int rv = 0;
	int i;

	for (i = 0; i <= Z3POLY_MAX_DEGREE; i++)
		if (pz->coeffs[i] != 0)
			rv++;
	return rv;
}

// ----------------------------------------------------------------
int z3poly_is_zero(z3poly_t * pz)
{
	int i;
	for (i = 0; i <= Z3POLY_MAX_DEGREE; i++)
		if (pz->coeffs[i] != 0)
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int z3poly_is_one(z3poly_t * pz)
{
	int i;
	if (pz->coeffs[0] != 1)
		return 0;
	for (i = 1; i <= Z3POLY_MAX_DEGREE; i++)
		if (pz->coeffs[i] != 0)
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int z3poly_is_two(z3poly_t * pz)
{
	int i;
	if (pz->coeffs[0] != 2)
		return 0;
	for (i = 1; i <= Z3POLY_MAX_DEGREE; i++)
		if (pz->coeffs[i] != 0)
			return 0;
	return 1;
}

// ----------------------------------------------------------------
void z3poly_make_monic(
	z3poly_t * pz)
{
	int i, c;
	int degz = z3poly_get_degree(pz);

	// Z3 is self-reciprocating.
	c = pz->coeffs[degz];
	if (c == 0)
		return;

	c &= 3;
	for (i = 0; i <= degz; i++)
		pz->coeffs[i] = Z3_MUL(pz->coeffs[i], c);
}

// ----------------------------------------------------------------
int z3poly_get_degree(
	z3poly_t * pin)
{
	int i;

	for (i = Z3POLY_MAX_DEGREE; i >= 0; i--)
		if (pin->coeffs[i] != 0)
			return i;
	return 0;
}

// ----------------------------------------------------------------
void z3poly_check_degree(int degree)
{
	if (degree < 0) {
		fprintf(stderr,
			"z3poly:  degree must be non-negative; got %d.\n",
			degree);
		exit(1);
	}
	if (degree > Z3POLY_MAX_DEGREE) {
		fprintf(stderr,
			"z3poly:  degree must be <= %d; got %d.\n",
			Z3POLY_MAX_DEGREE, degree);
		exit(1);
	}
}
