// Elements of the ring Z_3[x].

#ifndef Z3POLY_H
#define Z3POLY_H

#include "cmps.h"
#include "factorize.h"

// Please keep this one less than a multiple of 3.
#define Z3POLY_MAX_DEGREE 32//11

typedef struct _z3poly_t {
	unsigned char coeffs[Z3POLY_MAX_DEGREE+1];
} z3poly_t;

// ----------------------------------------------------------------
#define Z3_ADD(a,b) (z3_add_tbl[(a) & 3][(b) & 3])
#define Z3_SUB(a,b) (z3_sub_tbl[(a) & 3][(b) & 3])
#define Z3_MUL(a,b) (z3_mul_tbl[(a) & 3][(b) & 3])
#define Z3_USUB(a)  (z3_usub_tbl[(a) & 3])
#define Z3_DIV(a,b) (z3_div_tbl[(a) & 3][(b) & 3])

extern unsigned char z3_add_tbl[4][4];
extern unsigned char z3_sub_tbl[4][4];
extern unsigned char z3_usub_tbl[4];
extern unsigned char z3_mul_tbl[4][4];
extern unsigned char z3_div_tbl[4][4];

// ----------------------------------------------------------------
z3poly_t z3poly_add(
	z3poly_t * pin1,
	z3poly_t * pin2);

z3poly_t z3poly_sub(
	z3poly_t * pin1,
	z3poly_t * pin2);

z3poly_t z3poly_usub(
	z3poly_t * pin);

z3poly_t z3poly_mul(
	z3poly_t * pin1,
	z3poly_t * pin2);

z3poly_t z3poly_scalar_mul(
	z3poly_t * pin1,
	int        in2);

z3poly_t z3poly_div(
	z3poly_t * pin1,
	z3poly_t * pin2);

z3poly_t z3poly_mod(
	z3poly_t * pin1,
	z3poly_t * pin2);

void z3poly_quot_and_rem(
	z3poly_t * pdividend,
	z3poly_t * pdivisor,
	z3poly_t * pquot,
	z3poly_t * prem);

z3poly_t z3poly_ext_gcd(
	z3poly_t * pa,  // 1st input.
	z3poly_t * pb,  // 2nd input.
	z3poly_t * pm,  // Output coefficient on a
	z3poly_t * pn); // Output coefficient on b

z3poly_t z3poly_power(
	z3poly_t * pin,
	int        power);

// Assumes the input is already known to be a cube.
z3poly_t z3poly_cube_root(
	z3poly_t * pin);

z3poly_t z3poly_gcd(
	z3poly_t * pa,  // 1st input.
	z3poly_t * pb); // 2nd input.

int z3poly_equal(
	z3poly_t * pin1,
	z3poly_t * pin2);

// Fp[x] is *not* totally ordered.  Nonetheless, these functions are
// handy for looping and sorting.
int z3poly_lt(z3poly_t * pa, z3poly_t * pb);
int z3poly_gt(z3poly_t * pa, z3poly_t * pb);
int z3poly_le(z3poly_t * pa, z3poly_t * pb);
int z3poly_ge(z3poly_t * pa, z3poly_t * pb);
int z3poly_eq(z3poly_t * pa, z3poly_t * pb);
int z3poly_ne(z3poly_t * pa, z3poly_t * pb);

int z3poly_cmp(
	z3poly_t * pa,
	int       cmp, // see cmps.h
	z3poly_t * pb);

// ----------------------------------------------------------------
// Construction

z3poly_t z3poly_from_base_rep(
	int   base_rep);

int z3poly_to_base_rep(
	z3poly_t * pz);

z3poly_t z3poly_from_coeffs(
	int   degree,
	int * coeffs);

z3poly_t z3poly_from_coeffs0(
	int   coeff0);

z3poly_t z3poly_from_coeffs1(
	int   coeff1,
	int   coeff0);

z3poly_t z3poly_from_coeffs2(
	int   coeff2,
	int   coeff1,
	int   coeff0);

z3poly_t z3poly_monomial_from_coeff(
	int   coeffa,
	int   dega);

z3poly_t z3poly_binomial_from_coeffs(
	int   coeffa,
	int   dega,
	int   coeffb,
	int   degb);

// ----------------------------------------------------------------
// I/O

int z3poly_scan_base_27(
	z3poly_t * pout,
	char     * pin);

int z3poly_scan_base_3(
	z3poly_t * pout,
	char     * pin);

int z3poly_scan(
	z3poly_t * pout,
	char     * pin);

void z3poly_dprint_base_3(
	z3poly_t * pin,
	int deg);

void z3poly_dprint_base_27(
	z3poly_t * pin,
	int deg);

void z3poly_dprint(
	z3poly_t * pin,
	int deg);

void z3poly_print_base_27(
	z3poly_t * pin);

void z3poly_print_base_3(
	z3poly_t * pin);

void z3poly_print(
	z3poly_t * pin);

// ----------------------------------------------------------------
// Factorization, evaluation and classification.

int z3poly_eval(
	z3poly_t * pz,
	int   x);

z3poly_t z3poly_compose(
	z3poly_t * pr,
	z3poly_t * pz);

z3poly_t z3poly_deriv(
	z3poly_t * pz);

int z3poly_is_rootless(
	z3poly_t * pz);

int z3poly_count_non_zero_terms(
	z3poly_t * pz);

int z3poly_is_zero(
	z3poly_t * pz);

int z3poly_is_one(
	z3poly_t * pz);

int z3poly_is_two(
	z3poly_t * pz);

void z3poly_make_monic(
	z3poly_t * pz);

int z3poly_get_degree(
	z3poly_t * pin);

void z3poly_check_degree(
	int degree);

#endif // Z3POLY_H
