// xxx cmt z3poly_t uses deg & rest of coeffs[] can be garbage.
// this is for efficiency -- to avoid computing a lot of needless zeroes.
// however, here we have matrices which need to be square, including
// past-degree zero terms.  Requires using different-than-usual z3poly fcns.

#include <stdio.h>
#include <stdlib.h>
#include "malloc_check.h"
#include "z3polymod.h"
#include "z3poly_factor.h"
#include "z3linalg.h"
#include "mod.h"

static void gather_multiple_factors(
	z3poly_t * pirred_factors,
	int        num_irred_factors,
	z3poly_factor_info_t * pfinfo);

static void sort_up(
	z3poly_t * polys,
	int        nrows);

static void pre_berlekamp3(
	z3poly_t * pr,
	z3poly_t * pirred_factors,
	int      * pnum_irred_factors);

static void berlekamp3(
	z3poly_t * pr,
	z3poly_t * pirred_factors,
	int      * pnum_irred_factors);

// ----------------------------------------------------------------
void z3poly_factor(
	z3poly_t  * pr,
	z3poly_factor_info_t * pfinfo)
{
	int num_irred_factors, i;
	z3poly_t  irred_factors[Z3POLY_MAX_NUM_FACTORS];
	int leader;

	num_irred_factors = 0;
	pre_berlekamp3(pr, irred_factors, &num_irred_factors);

	for (i = 0; i < num_irred_factors; i++)
		z3poly_make_monic(&irred_factors[i]);

	leader = pr->coeffs[z3poly_get_degree(pr)];
	if (leader > 1) {
		irred_factors[num_irred_factors++] =
			z3poly_from_base_rep(leader);
	}


	sort_up(irred_factors, num_irred_factors);
	gather_multiple_factors(irred_factors, num_irred_factors, pfinfo);
}

// ----------------------------------------------------------------
static void gather_multiple_factors(
	z3poly_t * pirred_factors,
	int       num_irred_factors,
	z3poly_factor_info_t * pfinfo)
{
	int i;
	z3poly_t  * plast;
	int nd = 0;

	if (num_irred_factors == 0) {
		pfinfo->num_distinct = 0;
		return;
	}

	plast = &pirred_factors[0];
	pfinfo->factor_counts[nd].f = *plast;
	pfinfo->factor_counts[nd].count = 0;

	for (i = 0; i < num_irred_factors; i++) {
		if (!z3poly_equal(&pirred_factors[i], plast)) {
			nd++;
			pfinfo->factor_counts[nd].count = 1;
			pfinfo->factor_counts[nd].f = pirred_factors[i];
			plast = &pirred_factors[i];
		}
		else {
			pfinfo->factor_counts[nd].count++;
		}
	}
	pfinfo->num_distinct = nd + 1;
}

// ----------------------------------------------------------------
static void pre_berlekamp3(
	z3poly_t * pr,
	z3poly_t * pirred_factors,
	int      * pnum_irred_factors)
{
	z3poly_t d, g;


	d = z3poly_deriv(pr);
	g = z3poly_gcd(pr, &d);

	if (z3poly_is_zero(&g)) {
		if (!z3poly_is_zero(pr)) {
			fprintf(stderr, "CODING ERROR file %s\n", __FILE__);
			exit(1);
		}
		pirred_factors[*pnum_irred_factors] = *pr;
		(*pnum_irred_factors)++;
		return;
	}
	else if (z3poly_get_degree(&g) == 0) {
		// Input is squarefree.
		berlekamp3(pr, pirred_factors, pnum_irred_factors);
	}
	else if (z3poly_is_zero(&d)) {
		// Input is a perfect cube
		z3poly_t s = z3poly_cube_root(pr);

		// Multiplicity is 3 only if degree is > 0.
		pre_berlekamp3(&s, pirred_factors, pnum_irred_factors);
		if (z3poly_get_degree(pr) > 0) {
			pre_berlekamp3(&s, pirred_factors, pnum_irred_factors);
			pre_berlekamp3(&s, pirred_factors, pnum_irred_factors);
		}
	}
	else {
		z3poly_t q;
		q = z3poly_div(pr, &g);
		pre_berlekamp3(&g, pirred_factors, pnum_irred_factors);
		pre_berlekamp3(&q, pirred_factors, pnum_irred_factors);
	}
}

//// ----------------------------------------------------------------
//// Berlekamp's algorithm requires squarefree input.  In the literature,
//// this is often described as being done by factoring out GCD(r(x), r'(x)).
//// This function takes a different approach to find squareful inputs, and also
//// often produces a complete factorization:
////
//// X^(p^n) - X factors into all irreducibles of degree d for all d dividing
//// n.  Therefore, if r (the input, of degree n) has an irreducible factor
//// of degree < n, that factor will divide X^(p^d) - X for some d < n.
//// If r has GCD = 1 with X^(p^d) - X for all d < n, then r is irreducible.
//// Reduction mod r is taken at each step to avoid overflow, with no effect
//// on the GCD.
//
//static void pre_berlekamp3(
//	z3poly_t * pr,
//	z3poly_t * pirred_factors,
//	int      * pnum_irred_factors)
//{
//	z3poly_t x = z3poly_from_base_rep(3);
//	z3poly_t u = z3poly_from_base_rep(3);
//	z3poly_t ux, g;
//	int d = z3poly_get_degree(pr);
//	int i, br;
//
//	for (i = 1; i < d; i++) {
//
//		// u = u**p mod r; test GCD(u - x, r)
//		z3polymod_power(&u, &u, 3, pr);
//		z3poly_sub(&ux, &u, &x);
//		if (z3poly_is_zero(&ux)) {
//			// If we got here, r is known to be reducible but this
//			// algorithm can't distinguish its factors.  However,
//			// it is known that r is squarefree, suitable for input
//			// to Berlekamp's algorithm.  (Incidentally, all of r's
//			// factors are of the same degree, but that fact won't
//			// be needed.)
//			berlekamp3(pr, pirred_factors, pnum_irred_factors);
//			return;
//		}
//		g = z3poly_gcd(pr, &ux);
//		// See if the GCD is not a non-zero scalar
//		br = z3poly_to_base_rep(&g);
//		if ((br != 1) && (br != 2)) {
//			// If we got here, r is known to be reducible and g
//			// is a factor.
//			z3poly_t  q;
//			z3poly_div(&q, pr, &g);
//			pre_berlekamp3(&g, pirred_factors, pnum_irred_factors);
//			pre_berlekamp3(&q, pirred_factors, pnum_irred_factors);
//			return;
//		}
//	}
//	// If we got here, r is known to be irreducible.
//	pirred_factors[*pnum_irred_factors] = *pr;
//	(*pnum_irred_factors)++;
//}

// ----------------------------------------------------------------
static void berlekamp3(
	z3poly_t * pr,
	z3poly_t * pirred_factors,
	int      * pnum_irred_factors)
{
	int       n = z3poly_get_degree(pr);
	z3poly_t  x = z3poly_from_base_rep(3);
	z3poly_t  x3;
	z3poly_t  x3i = z3poly_from_base_rep(1);
	int row, rank, dimker, c;
	int got_it = 0;

	unsigned char ** BI;
	unsigned char ** BIT;
	z3poly_t * nullspace_basis;
	z3poly_t   h, f1, f2;

	if (n < 2) {
		pirred_factors[*pnum_irred_factors] = *pr;
		(*pnum_irred_factors)++;
		return;
	}

	BI  = z3la_alloc_matrix(n, n);
	BIT = z3la_alloc_matrix(n, n);
	nullspace_basis = malloc_check(n * sizeof(z3poly_t));

	x3 = z3polymod_power(&x, 3, pr);

	// Populate the B matrix with row i equal to x^3i mod r
	for (row = 0; row < n; row++) {
		z3la_row_from_z3poly(BI[row], &x3i, n);
		x3i = z3polymod_mul(&x3i, &x3, pr);
	}

	// Replace B with B - I
	for (row = 0; row < n; row++) {
		BI[row][row] = Z3_SUB(BI[row][row], 1);
	}

	z3la_transpose(BIT, BI, n);
	z3la_row_echelon_form(BIT, n, n);
	rank = z3la_find_rank(BIT, n, n);
	dimker = n - rank;

	if (dimker == 1) {
		pirred_factors[*pnum_irred_factors] = *pr;
		(*pnum_irred_factors)++;
		return;
	}

	z3la_find_nullspace_basis(BIT, n, n, nullspace_basis, dimker);

#if 1
	// Check h^p = h.  The algorithm guarantees this, *if* implemented
	// correctly.  This computation, while not strictly necessary, is a
	// nice sanity check.
	for (row = 0; row <  dimker; row++) {
		z3poly_t * ph = &nullspace_basis[row];
		z3poly_t hp;
		hp = z3polymod_power(ph, 3, pr);
		if (!z3poly_equal(ph, &hp)) {
			fprintf(stderr, "CODING ERROR\n");
			exit(1);
		}
	}
#endif

	got_it = 0;
	for (row = 0; row <  dimker && !got_it; row++) {
		if (z3poly_is_one(&nullspace_basis[row]))
			continue;

		for (c = 0; c < 3; c++) {
			int degf1;

			z3poly_t cpoly = z3poly_from_base_rep(c);
			h = z3poly_sub(&nullspace_basis[row], &cpoly);
			f1 = z3poly_gcd(pr, &h);
			degf1 = z3poly_get_degree(&f1);

			if ((0 < degf1) && (degf1 < n)) {
				got_it = 1;
				break;
			}
		}
	}

	if (!got_it) {
		printf("CODING ERROR: no non-trivial factors found: ");
		z3poly_print(pr);
		printf("\n");
		exit(1);
	}

	f2 = z3poly_div(pr, &f1);

	// The nullity of B-I is the number of irreducible
	// factors of r.  If the nullity is 2, we have a
	// pair of factors which are both irreducible and
	// so we don't need to recurse.
	if (dimker == 2) {
		pirred_factors[*pnum_irred_factors] = f1;
		(*pnum_irred_factors)++;
		pirred_factors[*pnum_irred_factors] = f2;
		(*pnum_irred_factors)++;
	}
	else {
		pre_berlekamp3(&f1, pirred_factors, pnum_irred_factors);
		pre_berlekamp3(&f2, pirred_factors, pnum_irred_factors);
	}

	z3la_free_matrix(BI,  n, n);
	z3la_free_matrix(BIT, n, n);
	free(nullspace_basis);
}

// ----------------------------------------------------------------
// Ascending bubble sort
static void sort_up(
	z3poly_t * polys,
	int       nrows)
{
	int i, top;
	z3poly_t  temp;

	for (top = nrows-1; top > 0; top--) {
		for (i = 0; i < top; i++) {
			if (!z3poly_lt(&polys[i], &polys[i+1])) {
				temp       = polys[i];
				polys[i]   = polys[i+1];
				polys[i+1] = temp;
			}
		}
	}
}

// ----------------------------------------------------------------
// Given the prime factorization p1^m1 ... pk^mk of n, how to enumerate all
// factors of n?
//
// Example 72 = 2^3 * 3^2.  Exponent on 2 is one of 0, 1, 2, 3.  Exponent on
// 3 is one of 0, 1, 2.  Number of possibilities:  product over i of (mi + 1).
// Factors are:
//
//	2^0 3^0    1
//	2^0 3^1    3
//	2^0 3^2    9
//	2^1 3^0    2
//	2^1 3^1    6
//	2^1 3^2   18
//	2^2 3^0    4
//	2^2 3^1   12
//	2^2 3^2   36
//	2^3 3^0    8
//	2^3 3^1   24
//	2^3 3^2   72

unsigned z3poly_num_divisors(
	z3poly_factor_info_t * pfinfo)
{
	unsigned rv;
	int i;

	if ((pfinfo->num_distinct < 0)
		|| (pfinfo->num_distinct > Z3POLY_MAX_NUM_FACTORS))
	{
		fprintf(stderr, "z3poly_num_divisors:  num distinct (%d) "
			"out of bounds 0-%d\n",
			pfinfo->num_distinct, Z3POLY_MAX_NUM_FACTORS);
		exit(1);
	}

	rv = 1;
	for (i = 0; i < pfinfo->num_distinct; i++)
		rv *= pfinfo->factor_counts[i].count + 1;
	return rv;
}

// ----------------------------------------------------------------
// See comments to the above.  k is treated as a multibase representation
// over the bases mi+1.

z3poly_t z3poly_kth_divisor(
	z3poly_factor_info_t * pfinfo,
	unsigned k)
{
	z3poly_t rv;
	z3poly_t fp;
	int i;
	unsigned base;
	unsigned power;

	rv = z3poly_from_base_rep(1);
	for (i = 0; i < pfinfo->num_distinct; i++) {
		base = pfinfo->factor_counts[i].count + 1;
		power = k % base;
		k = k / base;

		fp = z3poly_power(&pfinfo->factor_counts[i].f, power);
		rv = z3poly_mul(&rv, &fp);
	}
	return rv;
}

// ----------------------------------------------------------------
static int z3poly_qcmp(const void * pv1, const void * pv2)
{
	if (z3poly_lt((z3poly_t*)pv1, (z3poly_t*)pv2))
		return -1;
	if (z3poly_gt((z3poly_t*)pv1, (z3poly_t*)pv2))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
// The caller must free the return value.

z3poly_t * z3poly_get_all_divisors(
	z3poly_t * pr,
	unsigned * pnum_divisors)
{
	z3poly_factor_info_t finfo;
	unsigned nf, k;
	z3poly_t * all_divisors;

	z3poly_factor(pr, &finfo);
	nf = z3poly_num_divisors(&finfo);
	*pnum_divisors = nf;

	all_divisors = (z3poly_t *)malloc_check(nf * sizeof(z3poly_t));
	for (k = 0; k < nf; k++) {
		all_divisors[k] = z3poly_kth_divisor(&finfo, k);
	}
	qsort(all_divisors, nf, sizeof(z3poly_t), z3poly_qcmp);

	return all_divisors;
}
