#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "z3polymod.h"
#include "power.h"
#include "posmod.h"
#include "mod.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
// Sample of the reduce-en-route method:
//
// Use GF(7^4), with reductor 1,0,1,3,5.  Multiply 1,2,3,4 by 2,4,6,1.
//
//       1 2 3 4
//       2 4 6 1
//      =========
//       1 2 3 4   1,2,3,4 times 1
//
//     6 5 4 3 0   1,2,3,4 times 6x
//     6 0 6 4 2   reduce mod r
//    -----------
//       5 5 6 5
//
//   4 1 5 2 0     1,2,3,4 times 4x
//   4 0 4 5 6     reduce mod r
//  -------------
//     1 1 4 1 0   times x again
//     1 0 1 3 5   reduce mod r
//  -------------
//       1 3 5 2
//
// 2 4 6 1         1,2,3,4 times 2x
// 2 0 2 6 3       reduce mod r
// --------------
//   4 4 2 4       times x again
//   4 0 4 5 6     reduce mod r
// --------------
//     4 5 6 1     times x again
//     4 0 4 5 6   reduce mod r
// --------------
//       5 2 3 1
// ==============
//       1 2 3 4   four partial products
//       5 5 6 5
//       1 3 5 2
//       5 2 3 1
// ==============
//       5 5 3 5   final product is the sum of the four partial products
//
// ----------------------------------------------------------------
// The reduce-en-route method can be optimized a little, to not have
// to shift the coefficients when multiplying a partial product by x.
// Example:
//
//       1 2 3 4
//       2 4 6 1
//      =========
//       1 2 3 4     in1 * in2[0]
//
//
//       6 5 4 3(0)  in1 * in2[1]
//      / / / / /
//     6 0 6 4 2     mul    = in1[3]
//                   in1[3] = in1[2] - red[3] * mul
//                   in1[2] = in1[1] - red[2] * mul
//                   in1[1] = in1[0] - red[1] * mul
//                   in1[0] =     0  - red[0] * mul
//    -----------
//       5 5 6 5
//
//
//       4 1 5 2(0)  in1 * in2[2]
//      / / / / /
//     4 0 4 5 6
//    -----------
//       1 1 4 1(0)
//      / / / / /
//     1 0 1 3 5
//    -----------
//       1 3 5 2
//
//
//       2 4 6 1(0)  in1 * in2[3]
//      / / / / /
//     2 0 2 6 3
//    -----------
//       4 4 2 4(0)
//      / / / / /
//     4 0 4 5 6
//    -----------
//       4 5 6 1(0)
//      / / / / /
//     4 0 4 5 6
//    -----------
//       5 2 3 1
//
//
// ==============
//       1 2 3 4
//       5 5 6 5
//       1 3 5 2
//       5 2 3 1
// ==============
//       5 5 3 5

z3poly_t z3polymod_mul(
	z3poly_t * pin1,
	z3poly_t * pin2,
	z3poly_t * pr)
{
	z3poly_t  temp, out;
	int i, j, power;
	int elim;
	z3poly_t in1red, in2red, rmonic;
	unsigned char * p1c, * p2c, * ptc, * prc;
	int deg1, deg2, degr;

	in1red = z3poly_mod(pin1, pr);
	in2red = z3poly_mod(pin2, pr);
	rmonic = *pr;
	z3poly_make_monic(&rmonic);

	p1c = &in1red.coeffs[0];
	p2c = &in2red.coeffs[0];
	ptc = &temp.coeffs[0];
	prc = &rmonic.coeffs[0];

	deg1  = z3poly_get_degree(&in1red);
	deg2  = z3poly_get_degree(&in2red);
	degr  = z3poly_get_degree(pr);

	z3poly_check_degree(degr - 1);

	for (i = 0; i <= Z3POLY_MAX_DEGREE; i++)
		out.coeffs[i] = 0;

	for (i = 0; i <= deg2; i++) {
		// Multiply in1 by scalar coefficient on current term of in2.
		for (j = 0; j <= deg1; j++)
			ptc[j] = Z3_MUL(p1c[j], p2c[i]);
		for (j = deg1 + 1; j <= Z3POLY_MAX_DEGREE; j++)
			ptc[j] = 0;

		// Now need to multiply by the power of x on the current term
		// of in2, reducing mod r en route.
		for (power = 1; power <= i; power++) {
			elim = ptc[degr - 1];
			for (j = degr - 2; j >= 0; j--) {
				ptc[j+1] = Z3_SUB(ptc[j],
					Z3_MUL(prc[j+1], elim));
//printf("prc[%d]=%d\n", j+1, prc[j+1]);
				//printf("%s", "");
				ptc[j+1] = Z3_MUL(ptc[j+1], prc[degr]);
//printf("prc[%d]=%d\n", degr, prc[degr]);
			}
			ptc[0] = Z3_MUL(Z3_USUB(prc[0]), elim);
			ptc[0] = Z3_MUL(ptc[0], prc[degr]);
		}
		out = z3poly_add(&out, &temp);
	}

	return out;
}

// ----------------------------------------------------------------
z3poly_t z3polymod_div(
	z3poly_t * pin1,
	z3poly_t * pin2,
	z3poly_t * pr)
{
	z3poly_t   binv;

	// 1/b mod r
	binv = z3polymod_recip(pin2, pr);

	// a * (1/b) mod r
	return z3polymod_mul(pin1, &binv, pr);
}

// ----------------------------------------------------------------
// Finite-field division uses extended Euclidean GCD, with ring division
// (quotient and remainder) as a substep.
//
// a / b is a * 1/b.  To compute 1/b, find the GCD d.  (Since the reductor
// is irreducible -- although we don't test this on every division -- and
// if b is monic, the GCD had better be 1).  Then, compute 1 = m*b + n*r.
// Since r = 0 mod r, 1 = m*b and therefore m = 1/b.  The extended Euclidean
// algorithm (see Knuth vol. 1) computes d, m and n all at once.
//
// An alternate reciprocation algorithm uses the fact that a finite field
// GF(p^m) has its multiplicative group of order p^m-1.  So, by Euler's
// criterion, x^(p^n-2) = x^(-1) for all non-zero elements of the field.
//
// Testing has shown the extended-GCD reciprocation is faster for small p
// and large n; p^n-2 is faster for large p and small n.  (Small p and small
// n were too quick for a naive timing test; large p and large n both take
// too long to test.)

z3poly_t z3polymod_recip(
	z3poly_t * pin,
	z3poly_t * pr)
{
#if 1
	z3poly_t g, m, n;
	int c0i;
	int i;

	g = z3poly_ext_gcd(pr, pin, &m, &n);

	// Ext. GCD actually returns a scalar multiple of 1.  Divide this out.
	// Z3 is self-reciprocating.
	//c0i = mod_recip(g.coeffs[0], pr->p);
	c0i = g.coeffs[0];
	for (i = 0; i <= Z3POLY_MAX_DEGREE; i++)
		n.coeffs[i] = Z3_MUL(n.coeffs[i], c0i);
	return n;
#else
	int pm2 = calc_power(3, pr->r.degree) - 2;
	return z3polymod_power(pin, pm2, pr);
#endif
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
z3poly_t z3polymod_power(
	z3poly_t * px,
	unsigned   power,
	z3poly_t * pr)
{
	z3poly_t x2 = *px;
	z3poly_t rv = z3poly_from_base_rep(1);

	while (power != 0) {
		if (power & 1)
			rv = z3polymod_mul(&x2, &rv, pr);
		power >>= 1;
		x2 = z3polymod_mul(&x2, &x2, pr);
	}
	return rv;
}

// ----------------------------------------------------------------
// It suffices to check only factors of p^m-1.

// xxx this only works correctly when r is irreducible.
// xxx need to code for reducible case, with zero-divisor check, and
// phi(?) rather than p^m-1.

int z3polymod_element_order(
	z3poly_t * pz,
	z3poly_t * pr)
{
	int m = z3poly_get_degree(pr);
	unsigned pm = calc_power(3, m);
	unsigned pm1 = pm - 1;
	unsigned nd, k;
	unsigned * all_divisors = 0;
	unsigned f = 0;
	z3poly_t zpower;
	int got_it = 0;

	if (z3poly_to_base_rep(pz) == 0)
		return 0;
	if ((z3poly_to_base_rep(pz) == 3) && (z3poly_to_base_rep(pr) == 3))
		return 0;

	all_divisors = get_all_divisors(pm1, &nd);

	for (k = 0; k < nd; k++) {
		f = all_divisors[k];
		zpower = z3polymod_power(pz, f, pr);
		if (z3poly_to_base_rep(&zpower) == 1) {
			got_it = 1;
			break;
		}
	}

	free(all_divisors);
	if (got_it) {
		return f;
	}
	else {
		printf("z3polymod_element_order:  Internal error:  order of ");
		z3poly_print(pz);
		printf(" mod ");
		z3poly_print(pr);
		printf(" not found.\n");
		exit(1);
		return 0; // Not reached.
	}
}

// ----------------------------------------------------------------
z3poly_t z3polymod_find_generator(
	z3poly_t * pr)
{
	int n = z3poly_get_degree(pr);
	int pn1 = calc_power(3, n) - 1;
	int gi;
	int gilo = 2;
	int gihi = pn1;
	z3poly_t g;

	if (gihi < gilo)
		gihi = gilo; // linear fields

	for (gi = gilo; gi <= gihi; gi++) {
		g = z3poly_from_base_rep(gi);
		if (z3polymod_element_order(&g, pr) == pn1)
			return g;
	}

	printf("Generator for ");
	z3poly_print(pr);
	printf(" not found.\n");
	exit(1);
}

// ----------------------------------------------------------------
int z3polymod_log(
	z3poly_t * pz, // Input
	z3poly_t * pg, // Base, assumed (!) primitive mod r.
	z3poly_t * pr) // Modulus for FF arithmetic
{
	int log;
	z3poly_t gp = *pg;
	int deg = z3poly_get_degree(pr);
	int pm = calc_power(3, deg);

	// find e such that g^e == z mod r
	if (z3poly_is_zero(pz))
		return 0;

	for (log = 1; log <= pm+2; log++) {
		if (z3poly_equal(pz, &gp))
			return log;
		gp = z3polymod_mul(&gp, pg, pr);
	}

	printf("z3poly_log: ");
	z3poly_print(pg);
	printf(" is not primitive mod ");
	z3poly_print(pr);
	printf("\n");
	exit(1);
	return 0; // not reached
}

// ----------------------------------------------------------------
z3poly_t z3polymod_add(
	z3poly_t * pin1,
	z3poly_t * pin2,
	z3poly_t * pr)
{
	return z3poly_add(pin1, pin2);
}

// ----------------------------------------------------------------
z3poly_t z3polymod_sub(
	z3poly_t * pin1,
	z3poly_t * pin2,
	z3poly_t * pr)
{
	return z3poly_sub(pin1, pin2);
}

// ----------------------------------------------------------------
// We want things like Cayley tables to line up, so we want a fixed
// width for field elements.  However, when printing irreducibles
// or intermediate products (i.e. before reducing modulo r), we
// want to see all the coefficients.

void z3polymod_print(
	z3poly_t * pin,
	z3poly_t * pr)
{
	int degr, degin, degmax;

	degr  = z3poly_get_degree(pr);
	degin = z3poly_get_degree(pin);
	degmax = degr - 1;
	if (degin > degmax)
		degmax = degin;
	z3poly_dprint(pin, degmax);
}

// ----------------------------------------------------------------
// r(x) = 2x^2 + x + 1 (p = 3)
// z(x) = x + 2
// r(z(x)) = 2(x+2)^2 + (x+2) + 1

z3poly_t z3polymod_compose(
	z3poly_t * pr,
	z3poly_t * pz,
	z3poly_t * pm)
{
	z3poly_t rz = z3poly_from_base_rep(0);
	z3poly_t zpower = z3poly_from_base_rep(1);
	z3poly_t czpower;
	int d;

	for (d = 0; d <= Z3POLY_MAX_DEGREE; d++) {
		czpower = z3poly_scalar_mul(&zpower, pr->coeffs[d]);
		rz = z3poly_add(&rz, &czpower);
		zpower = z3polymod_mul(&zpower, pz, pm);
	}

	return rz;
}

// ----------------------------------------------------------------
// X^(p^n) - X factors into all irreducibles of degree d for all d dividing
// n.  Therefore, if r (the input, of degree n) has an irreducible factor
// of degree < n, that factor will divide X^(p^d) - X for some d < n.
// If r has GCD = 1 with X^(p^d) - X for all d < n, then r is irreducible.
// Reduction mod r is taken at each step to avoid overflow, with no effect
// on the GCD.

int z3poly_is_irreducible(z3poly_t * pr)
{
	z3poly_t x = z3poly_from_base_rep(3);
	z3poly_t u = z3poly_from_base_rep(3);
	z3poly_t ux, g;
	int d = z3poly_get_degree(pr);
	int i;

	for (i = 1; i < d; i++) {
		// u = u**p mod r; test GCD(u - x, r).
		u = z3polymod_power(&u, 3, pr);
		ux = z3poly_sub(&u, &x);
		if (z3poly_is_zero(&ux))
			return 0;
		g = z3poly_gcd(pr, &ux);
		// See if the GCD is not a non-zero scalar
		if (!z3poly_is_one(&g) && !z3poly_is_two(&g))
			return 0;
	}
	return 1;
}

// ----------------------------------------------------------------
// This is a brute-force search, not an efficient algorithm (see above).

// This is trial division.  Just as with factoring integers, where we
// need to try only up to sqrt(n), not n, here also we need to try factors
// only up to and including degree m/2:  e.g. if z has degree 37 and we've
// tried all factors of degree up to 18, we needn't try degree 19 and above.
// If there were a degree-19 factor, it would have a degree-18 cofactor which
// we would have already found.

// How to compute the bound using base-p representation:
//
// m odd:  up to and including degree m/2-1
// E.g. p = 2, m = 5: up to     x^2+x+1.  This is  111 = 2^3-1
// Bound in base rep is p^(m/2 + 1) - 1.
//
// m even: up to and including degree m/2
// E.g. p = 2, m = 6: up to x^3+x^2+x+1.  This is 1111 = 2^4-1
// Bound in base rep is p^(m/2 + 1) - 1.

//int z3poly_is_irreducible_brute(z3poly_t * pr, int p)
//{
//	int i;
//	z3poly_t a, q, r;
//	int pm2;
//	int m = pr->degree;
//
//	if (pr->degree <= 1)
//		return 1;
//
//	// has root => reducible
//	// i.e. not rootless => not irreducible
//	if (pr->degree > 1)
//		if (!z3poly_is_rootless(pr, p))
//			return 0;
//
//	pm2 = calc_power(p, m/2 + 1);
//	// Start at p rather than 1, since scalars will always divide evenly.
//	for (i = p; i < pm2; i++) {
//		a = z3poly_from_base_rep(p, i);
//		z3poly_quot_and_rem(pr, &a, &q, &r, p);
//		if ((r.degree == 0) && (r.coeffs[0] == 0)) {
//			if (q.degree > 0) {
//				return 0;
//			}
//		}
//	}
//	return 1;
//}

// ----------------------------------------------------------------
z3poly_t z3poly_find_irreducible(int m)
{
	z3poly_t r;
	int i;
	int pm, tpm;

	pm = calc_power(3, m);
	tpm = pm * 2;
	for (i = pm; i < tpm; i++) {
		r = z3poly_from_base_rep(i);
		if (z3poly_is_irreducible(&r))
			if (z3poly_is_primitive(&r, m))
				return r;
	}
	fprintf(stderr,
	"z3poly: coding error: couldn't find any irreds in GF(%d^%d).\n",
		3, m);
	exit(1);
}

// ----------------------------------------------------------------
// Assumes r is known to be irreducible.
//
// See if the order of x mod r is p^m-1.  Certainly, x^(p^m-1) will be 1.
// But check x to all powers dividing p^m-1 to make sure they are *not* 1.

int z3poly_is_primitive(z3poly_t * pr, int m)
{
	int pm = calc_power(3, m);
	unsigned pm1 = (unsigned)pm - 1;
	factor_info_t finfo;

	if (z3poly_get_degree(pr) == 0)
		return 0;

	calc_factors(pm1, &finfo);
	return z3poly_is_primitive_f(pr, m, pm1, &finfo);
}

// ----------------------------------------------------------------
int z3poly_is_primitive_f(
	z3poly_t * pr,
	int m,
	unsigned pm1,
	factor_info_t * pfinfo)
{
	z3poly_t x = z3poly_from_base_rep(3);
	z3poly_t xpower;
	unsigned nf, f, k;

	if (z3poly_to_base_rep(pr) == 3)
		return 0;

	nf = num_divisors(pfinfo);
	for (k = 0; k < nf; k++) {
		f = kth_divisor(pfinfo, k);
		xpower = z3polymod_power(&x, f, pr);

		if ((f != pm1) && z3poly_is_one(&xpower)) {
			return 0;
		}
		if ((f == pm1) && !z3poly_is_one(&xpower)) {
			printf("ABEND!\n");

			printf("r=");
			z3poly_print(pr);
			printf(" pm1=%u f=%u ", pm1, f);
			printf(" deg=%d ", z3poly_get_degree(pr));
			printf(" x=");
			z3poly_print(&x);
			printf(" x^f=");
			z3poly_print(&xpower);
			printf("\n");

			exit(1);
		}
	}
	return 1;
}

// ----------------------------------------------------------------
z3poly_t * z3polymod_get_elements(
	int n,
	int * pnel)
{
	int nel = calc_power(3, n);
	z3poly_t * rv = malloc_check(nel * sizeof(z3poly_t));
	int i;

	for (i = 0; i < nel; i++)
		rv[i] = z3poly_from_base_rep(i);

	*pnel = nel;
	return rv;
}

// ----------------------------------------------------------------
void z3poly_find_dirreds(
	unsigned   n,
	z3poly_t * pirred_factors,
	unsigned * pnum_irred_factors,
	unsigned   max_irred_factors)
{
	unsigned num_irred_factors = 0;
	unsigned d;
	int ilo, ihi, i;
	z3poly_t f;

	for (d = 1; d <= n; d++) {
		if ((n % d) != 0)
			continue;

		ilo = calc_power(3, d);
		ihi = ilo << 1;
		for (i = ilo; i < ihi; i++) {
			f = z3poly_from_base_rep(i);
			if (!z3poly_is_irreducible(&f))
				continue;

			if (num_irred_factors >= max_irred_factors) {
				fprintf(stderr, "Too many factors.\n");
				exit(1);
			}
			pirred_factors[num_irred_factors] = f;
			num_irred_factors++;

		}
	}

	*pnum_irred_factors = num_irred_factors;
}
