#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "zlinalg.h"
#include "mod.h"
#include "malloc_check.h"
#include "readlines.h"
#include "tokenize.h"

static void zla_row_from_zpoly(
	int     * prow,
	zpoly_t * pzpoly,
	int       ncols);

// ----------------------------------------------------------------
int ** zla_alloc_matrix(
	int  nrows,
	int  ncols)
{
	int ** rv;
	int i;

	rv = malloc_check(nrows * sizeof(int *));
	for (i = 0; i < nrows; i++)
		rv[i] = malloc_check(ncols * sizeof(int));
	return rv;
}

// ----------------------------------------------------------------
void zla_free_matrix(
	int ** M,
	int    nrows,
	int    ncols)
{
	int i;
	for (i = 0; i < nrows; i++)
		free(M[i]);
	free(M);
}

// ----------------------------------------------------------------
void zla_print_matrix(
	int ** M,
	int    nrows,
	int    ncols)
{
	int i, j;
	for (i = 0; i < nrows; i++) {
		printf(" ");
#if 1
		for (j = 0; j < ncols; j++)
#else
		for (j = ncols-1; j >= 0; j--)
#endif
		{
			printf(" ");
			printf("%d", M[i][j]);
		}
		printf("\n");
	}
}

// ----------------------------------------------------------------
// Input of the form 1010:0010:0011:1110,
// or 1,0,1,0:0,0,1,0:0,0,1,1:1,1,1,0.

int zla_matrix_from_string(
	char   * string,
	int  *** pM,
	int    * pnrows,
	int    * pncols)
{
	int      len;
	char   * mystring;
	int      num_rows;
	int      num_cols, deg;
	char  ** row_string_ptrs;
	int    * polrow;
	int      tokrc;
	zpoly_t  poly;
	int   ** M;
	int      i, j;

	len = strlen(string) + 1;
	mystring = malloc_check(len);
	strcpy(mystring, string);
	num_rows = count_tokens(string, ":");
	row_string_ptrs = malloc_check(sizeof(char*) * num_rows);
	tokrc = tokenize(mystring, ":", row_string_ptrs, num_rows);
	if (tokrc != num_rows) {
		fprintf(stderr, "coding error: tokrc %d nr %d\n",
			tokrc, num_rows);
		free(mystring);
		free(row_string_ptrs);
		return 0;
	}

	num_cols = 0;
	for (i = 0; i < num_rows; i++) {
		if (!zpoly_scan(&poly, row_string_ptrs[i])) {
			free(mystring);
			free(row_string_ptrs);
			return 0;
		}
		// xxx here, no zppoly stuff, and also require all row lengths
		// to be the same.
		deg = poly.degree + 1;
		if (deg > num_cols)
			num_cols = deg;
	}

	M = zla_alloc_matrix(num_rows, num_cols);
	polrow = malloc_check(sizeof(int) * num_cols);
	for (i = 0; i < num_rows; i++) {
		if (!zpoly_scan(&poly, row_string_ptrs[i])) {
			free(mystring);
			free(row_string_ptrs);
			return 0;
		}
		zla_row_from_zpoly(polrow, &poly, num_cols);
		for (j = 0; j < num_cols; j++)
			M[i][j] = polrow[num_cols - j - 1];
	}

	free(mystring);
	free(row_string_ptrs);

	*pM = M;
	*pnrows = num_rows;
	*pncols = num_cols;
	return 1;
}

// ----------------------------------------------------------------
// Input of the form
//   1 0 1 0
//   0 0 1 0
//   0 0 1 1
//   1 1 1 0

int zla_matrix_from_stream(
	FILE  * fp,
	int *** pM,
	int   * pnrows,
	int   * pncols)
{
	char ** line_ptrs = 0;
	char ** col_ptrs = 0;
	int     num_rows = 0;
	int     cur_cols = 0;
	int     num_cols = 0;
	char  * seps = " \t";
	int  ** M;
	int     i, j;

	if (!read_lines(fp, &line_ptrs, &num_rows)) {
		fprintf(stderr, "zla_matrix_from_stream:  couldn't slurp "
			"text input.\n");
		return 0;
	}

	if (num_rows == 0) {
		fprintf(stderr,
			"zla_matrix_from_stream:  empty text input.\n");
		return 0;
	}

	num_cols = count_tokens(line_ptrs[0], seps);
	for (i = 1; i < num_rows; i++) {
		cur_cols = count_tokens(line_ptrs[i], seps);
		if (cur_cols != num_cols) {
			fprintf(stderr, "zla_matrix_from_stream:  "
				"non-rectangular input.\n");
			return 0;
		}
	}

	if (num_cols == 0) {
		fprintf(stderr, "zla_matrix_from_stream:  no columns.\n");
		return 0;
	}

	M = zla_alloc_matrix(num_rows, num_cols);
	col_ptrs = malloc_check(num_cols * sizeof(char *));

	for (i = 0; i < num_rows; i++) {
		(void)tokenize(line_ptrs[i], seps, col_ptrs, num_cols);
		for (j = 0; j < num_cols; j++) {
			if (sscanf(col_ptrs[j], "%d", &M[i][j]) != 1) {
				fprintf(stderr,
					"Parse error at row %d, column %d.\n",
					i, j);
				return 0;
			}
		}
	}

	*pM = M;
	*pnrows = num_rows;
	*pncols = num_cols;
	free(col_ptrs);
	free_lines(line_ptrs, num_rows);
	return 1;
}

// ----------------------------------------------------------------
static void zla_row_from_zpoly(
	int     * prow,
	zpoly_t * pzpoly,
	int       ncols)
{
	int i;
	int d = pzpoly->degree;
	for (i = 0; i <= d; i++)
		prow[i] = pzpoly->coeffs[i];
	for (i = d+1; i < ncols; i++)
		prow[i] = 0;
}

// ----------------------------------------------------------------
void zla_transpose(
	int    ** Mout,
	int    ** Min,
	int       n)
{
	int row, col;

	for (row = 0; row < n; row++)
		for (col = 0; col < n; col++)
			Mout[row][col] = Min[col][row];
}

// ----------------------------------------------------------------
// For square matrices.
void zla_mats_mul(
	int     ** out,
	int     ** in1,
	int     ** in2,
	int        n)
{
	int i, j, k;
	int ** temp = zla_alloc_matrix(n, n);
	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			temp[i][j] = 0;
			for (k = 0; k < n; k++) {
				temp[i][j] += in1[i][k] * in2[k][j];
			}
		}
	}
	zla_mats_copy(out, temp, n);
	zla_free_matrix(temp, n, n);
}

// ----------------------------------------------------------------
// For square matrices.
void zla_mats_copy(
	int     ** out,
	int     ** in,
	int        n)
{
	int i, j;
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			out[i][j] = in[i][j];
}

// ----------------------------------------------------------------
// For square matrices.
int zla_is_I(
	int     ** M,
	int        n)
{
	int i, j;

	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			if (i == j) {
				if (M[i][j] != 1)
					return 0;
			}
			else {
				if (M[i][j] != 0)
					return 0;
			}
		}
	}
	return 1;
}

// ----------------------------------------------------------------
// For square matrices.
int zla_equals(
	int     ** A,
	int     ** B,
	int        n)
{
	int i, j;

	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			if (A[i][j] != B[i][j])
				return 0;
	return 1;
}
