#include <stdio.h>
#include <stdlib.h>
#include "zmx.h"
#include "phi.h"
#include "gcd.h"
#include "posmod.h"
#include "malloc_check.h"

static int zmx_validate_mods(zmx_t * pmods);

static int zmx_product(
	zmx_t * pmods);

static void zmx_vindex_to_element(
	int     which,
	zmx_t * pmods,
	zmx_t * pelement);

// ----------------------------------------------------------------
zmx_t zmx_add(zmx_t a, zmx_t b, zmx_t m)
{
	int i;
	zmx_t c;
	for (i = 0; i < m.num_parts; i++) {
		c.parts[i] = a.parts[i] + b.parts[i];
		if (c.parts[i] >= m.parts[i])
			c.parts[i] -= m.parts[i];
	}
	c.num_parts = m.num_parts;
	return c;
}

// ----------------------------------------------------------------
zmx_t zmx_mul(zmx_t a, zmx_t b, zmx_t m)
{
	int i;
	zmx_t c;
	for (i = 0; i < m.num_parts; i++) {
		c.parts[i] = (a.parts[i] * b.parts[i]) % m.parts[i];
	}
	c.num_parts = m.num_parts;
	return c;
}

// ----------------------------------------------------------------
int  zmx_equal(zmx_t a, zmx_t b, zmx_t m)
{
	int i;

	for (i = 0; i < m.num_parts; i++)
		if (a.parts[i] != b.parts[i])
			return 0;
	return 1;
}

// ================================================================
static void zmx_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zmx_t * pa = (zmx_t *)pvin1;
	zmx_t * pb = (zmx_t *)pvin2;
	zmx_t * pc = (zmx_t *)pvout;
	zmx_t * pm = (zmx_t *)pvaux;
	*pc = zmx_add(*pa, *pb, *pm);
}

// ----------------------------------------------------------------
static void zmx_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zmx_t * pa = (zmx_t *)pvin1;
	zmx_t * pb = (zmx_t *)pvin2;
	zmx_t * pc = (zmx_t *)pvout;
	zmx_t * pm = (zmx_t *)pvaux;
	*pc = zmx_mul(*pa, *pb, *pm);
}

// ----------------------------------------------------------------
static int zmx_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int i;
	zmx_t * pa = (zmx_t *)pvin1;
	zmx_t * pb = (zmx_t *)pvin2;
	zmx_t * pm = (zmx_t *)pvaux;

	for (i = 0; i < pm->num_parts; i++)
		if (pa->parts[i] != pb->parts[i])
			return 0;
	return 1;
}

// ----------------------------------------------------------------
static int zmx_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int i;
	zmx_t * pa = (zmx_t *)pvin1;
	zmx_t * pb = (zmx_t *)pvin2;
	zmx_t * pm = (zmx_t *)pvaux;

	for (i = 0; i < pm->num_parts; i++) {
		if (pa->parts[i] < pb->parts[i])
			return 1;
		else if (pa->parts[i] > pb->parts[i])
			return 0;
	}
	return 0;
}

// ----------------------------------------------------------------
static void zmx_vprint(
	void * pvin,
	void * pvaux)
{
	int i;
	zmx_t * pa = (zmx_t *)pvin;
	zmx_t * pm = (zmx_t *)pvaux;

	//printf("(");
	for (i = 0; i < pm->num_parts; i++) {
		if (i > 0)
			printf(",");
		if (pm->parts[i] > 100)
			printf("%3d", pa->parts[i]);
		else if (pm->parts[i] > 10)
			printf("%2d", pa->parts[i]);
		else
			printf("%d", pa->parts[i]);
	}
	//printf(")");
}

// ----------------------------------------------------------------
static void zmx_vauxprint(
	void * pvaux)
{
	int i;
	zmx_t * pm = (zmx_t *)pvaux;

	//printf("(");
	for (i = 0; i < pm->num_parts; i++) {
		if (i > 0)
			printf(",");
		if (pm->parts[i] > 100)
			printf("%3d", pm->parts[i]);
		else if (pm->parts[i] > 10)
			printf("%2d", pm->parts[i]);
		else
			printf("%d", pm->parts[i]);
	}
	//printf(")");
}

// ----------------------------------------------------------------
// This is a gross hack.
int zmx_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	int i;
	zmx_t * pout  = (zmx_t *)pvout;
	zmx_t * pmods = (zmx_t *)pvaux;

	if (sizeof(zmx_t) > max_elt_bytes) {
		fprintf(stderr, "zmx:  Need %d bytes for element.\n",
			sizeof(zmx_t));
		exit(1);
	}

	for (i = 0; i < ZMX_MAX_PARTS; i++)
		pout->parts[i] = 0;

	// xxx use tokenize
	pout->num_parts = sscanf(pin,
		"%d,%d,%d,%d,%d,%d,%d,%d,%d,%d",
		&pout->parts[0], &pout->parts[1], &pout->parts[2],
		&pout->parts[3], &pout->parts[4], &pout->parts[5],
		&pout->parts[6], &pout->parts[7], &pout->parts[8],
		&pout->parts[9]);
	if (pout->num_parts != pmods->num_parts) {
		fprintf(stderr, "zmx_vscan: mod #parts %d, this %d.\n",
			pmods->num_parts, pout->num_parts);
		return 0;
	}
	for (i = 0; i < pmods->num_parts; i++)
		pout->parts[i] = posmod(pout->parts[i], pmods->parts[i]);
	return 1;
}

// ----------------------------------------------------------------
// This is a gross hack.
int zmx_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	int i;
	zmx_t mods;

	if (max_aux_bytes < sizeof(zmx_t)) {
		fprintf(stderr, "zmx:  Need %d bytes for aux.\n",
			sizeof(zmx_t));
		exit(1);
	}

	if (argc != 1)
		return 0;

	for (i = 0; i < ZMX_MAX_PARTS; i++)
		mods.parts[i] = 0;

	// xxx use tokenize
	mods.num_parts = sscanf(argv[0],
		"%d,%d,%d,%d,%d,%d,%d,%d,%d,%d",
		&mods.parts[0], &mods.parts[1], &mods.parts[2],
		&mods.parts[3], &mods.parts[4], &mods.parts[5],
		&mods.parts[6], &mods.parts[7], &mods.parts[8],
		&mods.parts[9]);
	*(zmx_t *)pvaux = mods;
	return zmx_validate_mods(&mods);
}

// ----------------------------------------------------------------
static int zmx_validate_mods(zmx_t * pmods)
{
	int i;

	if ((pmods->num_parts > ZMX_MAX_PARTS) ||
		(pmods->num_parts < 1))
	{
		fprintf(stderr,
			"zmx:  num parts must be >= 1 and <= %d; got %d.\n",
			ZMX_MAX_PARTS, pmods->num_parts);
		return 0;
	}

	for (i = 0; i < pmods->num_parts; i++) {
		if (pmods->parts[i] < 2) {
			fprintf(stderr,
				"zmx:  parts must be >= 2; got %d.\n",
				pmods->parts[i]);
			return 0;
		}
	}
	return 1;
}

// ----------------------------------------------------------------
int  zmx_add_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	zmx_t * pelements = (zmx_t *) pvelements;
	zmx_t * pmods = (zmx_t *)pvaux;
	int i;
	int group_size;

	group_size = zmx_product(pmods);

	if (group_size > max_group_bytes/sizeof(int))
		return 0;
	for (i = 0; i < group_size; i++) {
		zmx_vindex_to_element(i, pmods, &pelements[i]);
	}
	*pnum_elements = group_size;
	return 1;
}

// ----------------------------------------------------------------
// Example:
// Mods = 4,5
// Phis = 2, 4
// Multiplicative groups = {1, 3} {1, 2, 3, 4}
// Desired elements:  11 12 13 14 31 32 33 34
// First compute indices into the multiplicative groups:
//   00 01 02 03 10 11 12 13
// Then, for each position, convert indices into group elements:
//
//   k = 0:
//   00 -> 10
//   01 -> 11
//   02 -> 12
//   03 -> 13
//   10 -> 30
//   11 -> 31
//   12 -> 32
//   13 -> 33
//
//   k = 1:
//   00 -> 11
//   01 -> 12
//   02 -> 13
//   03 -> 14
//   10 -> 31
//   11 -> 32
//   12 -> 33
//   13 -> 34

int  zmx_mul_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	zmx_t * pelements = (zmx_t *) pvelements;
	zmx_t * pmods = (zmx_t *)pvaux; // E.g. 3,4,5
	zmx_t   phis;                   // E.g. 2,2,4
	int     i, k, group_size;
	int   * group_elts;

	phis.num_parts = pmods->num_parts;
	group_size = 1;
	for (i = 0; i < pmods->num_parts; i++) {
		phis.parts[i] = calc_phi_signed(pmods->parts[i]);
		group_size *= phis.parts[i];
	}
	if (group_size > max_group_bytes/sizeof(int))
		return 0;

	for (i = 0; i < group_size; i++) {
		zmx_vindex_to_element(i, &phis, &pelements[i]);
	}

	for (k = 0; k < pmods->num_parts; k++) {
		int ge, count;
		group_elts = malloc_check(sizeof(int) * phis.parts[k]);
		for (ge = 1, count = 0; count < phis.parts[k]; ge++) {
			if (calc_gcd_signed(ge, pmods->parts[k]) == 1) {
				group_elts[count] = ge;
				count++;
			}
		}
		for (i = 0; i < group_size; i++) {
			pelements[i].parts[k] =
				group_elts[pelements[i].parts[k]];
		}
		free(group_elts);
	}

	*pnum_elements = group_size;
	return 1;
}

// ----------------------------------------------------------------
static void zmx_vindex_to_element(
	int     which,
	zmx_t * pmods,
	zmx_t * pelement)
{
	int i;

	pelement->num_parts = pmods->num_parts;
	for (i = pmods->num_parts - 1; i >= 0; i--) {
		pelement->parts[i] = which % pmods->parts[i];
		which /= pmods->parts[i];
	}
}

// ----------------------------------------------------------------
static int zmx_product(
	zmx_t * pmods)
{
	int i;
	int rv = 1;
	for (i = 0; i < pmods->num_parts; i++)
		rv *= pmods->parts[i];
	return rv;
}

// ================================================================
group_spec_t zmx_add_group_spec = {
	element_size  : sizeof(zmx_t),
	naux          : 1,
	pvop          : zmx_vadd,
	pvequal       : zmx_vequal,
	pvless        : zmx_vless,
	pvprint       : zmx_vprint,
	pvauxprint    : zmx_vauxprint,
	pvscan        : zmx_vscan,
	pvauxscan     : zmx_vauxscan,
	pfill_group   : zmx_add_fill_group,
	desc          : "zmxadd"
};

// ----------------------------------------------------------------
group_spec_t zmx_mul_group_spec = {
	element_size  : sizeof(zmx_t),
	naux          : 1,
	pvop          : zmx_vmul,
	pvequal       : zmx_vequal,
	pvless        : zmx_vless,
	pvprint       : zmx_vprint,
	pvauxprint    : zmx_vauxprint,
	pvscan        : zmx_vscan,
	pvauxscan     : zmx_vauxscan,
	pfill_group   : zmx_mul_fill_group,
	desc          : "zmxmul"
};
