#include <string.h>

#include "zlinalg.h"
#include "zplinalg.h"
#include "mod.h"
#include "posmod.h"
#include "malloc_check.h"
#include "tokenize.h"
#include "readlines.h"

static int find_row_leader_pos(
	int     * prow,
	int       ncols,
	int     * ppos);

// ----------------------------------------------------------------
int ** zpla_alloc_matrix(
	int  nrows,
	int  ncols)
{
	int ** rv;
	int i;

	rv = malloc_check(nrows * sizeof(int *));
	for (i = 0; i < nrows; i++)
		rv[i] = malloc_check(ncols * sizeof(int));
	return rv;
}

// ----------------------------------------------------------------
void zpla_free_matrix(
	int ** M,
	int    nrows,
	int    ncols)
{
	int i;
	for (i = 0; i < nrows; i++)
		free(M[i]);
	free(M);
}

// ----------------------------------------------------------------
void zpla_print_matrix(
	int ** M,
	int    nrows,
	int    ncols)
{
	int i, j;
	for (i = 0; i < nrows; i++) {
		printf(" ");
#if 1
		for (j = 0; j < ncols; j++)
#else
		for (j = ncols-1; j >= 0; j--)
#endif
		{
			printf(" ");
			printf("%d", M[i][j]);
		}
		printf("\n");
	}
}

// ----------------------------------------------------------------
void zpla_print_vector(
	int * v,
	int   length)
{
	int i;
	for (i = 0; i < length; i++) {
		if (i > 0)
			printf(" ");
		printf("%d", v[i]);
	}
}

// ----------------------------------------------------------------
void zpla_matrix_from_zpolys(
	int     ** M,
	int        nrows,
	int        ncols,
	zpoly_t  * polys)
{
	int i;
	for (i = 0; i < nrows; i++)
		zpla_row_from_zpoly(M[i], &polys[i], ncols);
}

// ----------------------------------------------------------------
// Input of the form 1010:0010:0011:1110,
// or 1,0,1,0:0,0,1,0:0,0,1,1:1,1,1,0.

int zpla_matrix_from_string(
	char   * string,
	int      p,
	int  *** pM,
	int    * pnrows,
	int    * pncols)
{
	int      len;
	char   * mystring;
	int      num_rows;
	int      num_cols, deg;
	char  ** row_string_ptrs;
	int    * polrow;
	int      tokrc;
	zpoly_t  poly;
	int   ** M;
	int      i, j;

	len = strlen(string) + 1;
	mystring = malloc_check(len);
	strcpy(mystring, string);
	num_rows = count_tokens(string, ":");
	row_string_ptrs = malloc_check(sizeof(char*) * num_rows);
	tokrc = tokenize(mystring, ":", row_string_ptrs, num_rows);
	if (tokrc != num_rows) {
		fprintf(stderr, "coding error: tokrc %d nr %d\n",
			tokrc, num_rows);
		free(mystring);
		free(row_string_ptrs);
		return 0;
	}

	num_cols = 0;
	for (i = 0; i < num_rows; i++) {
		if (!zppoly_scan(&poly, row_string_ptrs[i], p)) {
			free(mystring);
			free(row_string_ptrs);
			return 0;
		}
		// xxx here, no zppoly stuff, and also require all row lengths
		// to be the same.
		deg = poly.degree + 1;
		if (deg > num_cols)
			num_cols = deg;
	}

	M = zpla_alloc_matrix(num_rows, num_cols);
	polrow = malloc_check(sizeof(int) * num_cols);
	for (i = 0; i < num_rows; i++) {
		if (!zppoly_scan(&poly, row_string_ptrs[i], p)) {
			free(mystring);
			free(row_string_ptrs);
			return 0;
		}
		zpla_row_from_zpoly(polrow, &poly, num_cols);
		for (j = 0; j < num_cols; j++)
			M[i][j] = polrow[num_cols - j - 1];
	}

	free(mystring);
	free(row_string_ptrs);

	*pM = M;
	*pnrows = num_rows;
	*pncols = num_cols;
	return 1;
}

// ----------------------------------------------------------------
// Input of the form
//   1 0 1 0
//   0 0 1 0
//   0 0 1 1
//   1 1 1 0

int zpla_matrix_from_stream(
	FILE  * fp,
	int     p,
	int *** pM,
	int   * pnrows,
	int   * pncols)
{
	int ** M, i, j;

	if (!zla_matrix_from_stream(fp, pM, pnrows, pncols))
		return 0;
	M = *pM;
	for (i = 0; i < *pnrows; i++) {
		for (j = 0; j < *pncols; j++) {
			M[i][j] = posmod(M[i][j], p);
		}
	}
	return 1;
}

// ----------------------------------------------------------------
// Input of the form
//   1 0 1 0
//   0 0 1 0
//   0 0 1 1
//   1 1 1 0

int zpla_matrix_from_file(
	char  * file_name,
	int     p,
	int *** pM,
	int   * pnrows,
	int   * pncols)
{
	FILE * fp = 0;
	int rc = 0;

	if (strcmp(file_name, "-") == 0) {
		fp = stdin;
	}
	else {
		fp = fopen(file_name, "r");
		if (fp == 0) {
			perror("open");
			fprintf(stderr,
				"zpla_matrix_from_file: couldn't open "
				"\"%s\" for read.\n",
				file_name);
			return 0;
		}
	}
	rc = zpla_matrix_from_stream(fp, p, pM, pnrows, pncols);
	if (fp != stdin)
		fclose(fp);
	return rc;
}

// ----------------------------------------------------------------
void zpla_vector_from_base_rep(
	int      * v,
	int        len,
	int        p,
	int        base_rep)
{
	int i;

	for (i = len - 1; i >= 0; i--) {
		v[i] = base_rep % p;
		base_rep /= p;
	}
}

// ----------------------------------------------------------------
void zpla_zpolys_from_matrix(
	zpoly_t  * polys,
	int     ** M,
	int        nrows,
	int        ncols)
{
	int i, j;
	for (i = 0; i < nrows; i++) {
		for (j = 0; j < ncols; j++) {
			polys[i].coeffs[j] = M[i][j];
			polys[i].degree = zppoly_find_degree(
				&polys[i], ncols - 1);
		}
	}
}

// ----------------------------------------------------------------
void zpla_row_from_zpoly(
	int     * prow,
	zpoly_t * pzpoly,
	int       ncols)
{
	int i;
	int d = pzpoly->degree;
	for (i = 0; i <= d; i++)
		prow[i] = pzpoly->coeffs[i];
	for (i = d+1; i < ncols; i++)
		prow[i] = 0;
}

// ----------------------------------------------------------------
void zpla_transpose(
	int    ** Mout,
	int    ** Min,
	int       n)
{
	int row, col;

	for (row = 0; row < n; row++)
		for (col = 0; col < n; col++)
			Mout[row][col] = Min[col][row];
}

// ----------------------------------------------------------------
void zpla_row_echelon_form(
	int     ** M,
	int        p,
	int        nrows,
	int        ncols)
{
	zpla_row_reduce_below (M, p, nrows, ncols);
	zpla_row_reduce_above (M, p, nrows, ncols);
	zpla_make_matrix_monic(M, p, nrows, ncols);
}

// ----------------------------------------------------------------
static int find_row_with_leader_at_col(
	int ** M,
	int    nrows,
	int    ncols,
	int    clear_col,
	int    leader_row,
	int  * pswap_row)
{
	int row;
	for (row = leader_row; row < nrows; row++) {
		if (M[row][clear_col] != 0) {
			*pswap_row = row;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
static void swap_rows(
	int ** M,
	int    nrows,
	int    ncols,
	int    leader_row,
	int    swap_row)
{
	int col, temp;

	for (col = 0; col < ncols; col++) {
		temp               = M[leader_row][col];
		M[leader_row][col] = M[swap_row][col];
		M[swap_row][col]   = temp;
	}
}

// ----------------------------------------------------------------
void zpla_make_row_monic(
	int * row,
	int   ncols,
	int   leader_col,
	int   p)
{
	int col;
	int cinv = mod_recip(row[leader_col], p);
	for (col = leader_col; col < ncols; col++) {
		row[col] = mod_mul(row[col], cinv, p);
	}
}

// ----------------------------------------------------------------
void zpla_row_reduce_below(
	int     ** M,
	int        p,
	int        nrows,
	int        ncols)
{
	int clear_col, leader_row, swap_row, clear_row;

	for (clear_col = 0, leader_row = 0; clear_col < ncols; clear_col++) {
		if (!find_row_with_leader_at_col(M, nrows, ncols, clear_col,
			leader_row, &swap_row))
		{
			continue;
		}
		if (leader_row != swap_row)
			swap_rows(M, nrows, ncols, leader_row, swap_row);
		if (M[leader_row][clear_col] != 1)
			zpla_make_row_monic(M[leader_row], ncols, clear_col, p);
		for (clear_row = leader_row+1; clear_row < nrows; clear_row++) {
			int col, clear_leader, clear_leader_inv;
			clear_leader = M[clear_row][clear_col];
			if (clear_leader == 0)
				continue;
			clear_leader_inv = mod_recip(clear_leader, p);
			for (col = clear_col; col < ncols; col++) {
				M[clear_row][col] *= clear_leader_inv;
				M[clear_row][col] %= p;
				M[clear_row][col] -= M[leader_row][col];
				if (M[clear_row][col] < 0)
					M[clear_row][col] += p;
			}
		}
		leader_row++;
	}
}

// ----------------------------------------------------------------
// Assumes row_reduce_below has already been done.

void zpla_row_reduce_above(
	int     ** M,
	int        p,
	int        nrows,
	int        ncols)
{
	int row, row2, row2_leader_pos;
	int row_leader_val, row2_leader_val, mul, col;

	for (row = 0; row < nrows; row++) {
		for (row2 = row+1; row2 < nrows; row2++) {
			if (!find_row_leader_pos(M[row2], ncols,
				&row2_leader_pos))
			{
				break;
			}
			row2_leader_val = M[row2][row2_leader_pos];
			row_leader_val  = M[row][row2_leader_pos];
			if (row_leader_val == 0)
				continue;

			mul = mod_div(row_leader_val, row2_leader_val, p);
			for (col = 0; col < ncols; col++)
				// row = row - mul * row2
				M[row][col] =
					mod_sub(M[row][col],
						mod_mul(M[row2][col], mul, p),
						p);
		}
	}
}

// ----------------------------------------------------------------
// Assumes row_reduce_below has already been done.

void zpla_make_matrix_monic(
	int     ** M,
	int        p,
	int        nrows,
	int        ncols)
{
	int row, row_leader_pos, row_leader_val, col;

	for (row = 0; row < nrows; row++) {
		if (!find_row_leader_pos(M[row], ncols, &row_leader_pos))
			break;
		row_leader_val = M[row][row_leader_pos];

		for (col = 0; col < ncols; col++) {
			M[row][col] = mod_div(
				M[row][col], row_leader_val, p);
		}
	}
}

// ----------------------------------------------------------------
static int find_row_leader_pos(
	int     * prow,
	int       ncols,
	int     * ppos)
{
	int col;

	for (col = 0; col < ncols; col++) {
		if (prow[col] != 0) {
			*ppos = col;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
// Assumes the matrix is already row-reduced.
int zpla_find_rank(
	int ** M,
	int    nrows,
	int    ncols)
{
	int rv = 0;
	int row;

	for (row = 0; row < nrows; row++) {
		if (!zpla_row_is_zero(M[row], ncols))
			rv++;
	}
	return rv;
}

// ----------------------------------------------------------------
int zpla_row_is_zero(
	int      * prow,
	int        ncols)
{
	int col;
	for (col = 0; col < ncols; col++)
		if (prow[col] != 0)
			return 0;
	return 1;
}

// ----------------------------------------------------------------
// Assumes the matrix is already in row-echelon form.

void zpla_find_nullspace_basis(
	int     ** BIT,
	int        nrows,
	int        ncols,
	int        p,
	zpoly_t  * nullspace_basis,
	int        dimker)
{
	int i, j;
	unsigned char * free_flags;
	unsigned char * free_indices;
	int nfree = 0; // == dimker but I'll compute it anyway
	int rank = nrows - dimker;
	int dep_pos;

	for (i = 0; i < dimker; i++) {
		for (j = 0; j < ncols; j++)
			nullspace_basis[i].coeffs[j] = 0;
		nullspace_basis[i].degree = ncols - 1;
	}

	free_flags   = malloc_check(nrows);
	free_indices = malloc_check(nrows);

	for (i = 0; i < nrows; i++) {
		free_flags[i] = 1;
	}

	for (i = 0; i < rank; i++) {
		if (find_row_leader_pos(BIT[i], ncols, &dep_pos))
			free_flags[dep_pos] = 0;
	}

	for (i = 0; i < nrows; i++) {
		if (free_flags[i])
			free_indices[nfree++] = i;
	}

	// For each free coefficient:
	//   Let that free coefficient be 1 and the rest be zero.
	//   Also set any dependent coefficients which depend on that
	//   free coefficient.
	for (i = 0; i < dimker; i++) {
		nullspace_basis[i].coeffs[free_indices[i]] = 1;

		// (B-I)^t in row echelon form:
		//
		// 0210     c0 = ??      c0 = 1  c0 = 0
		// 1000     c1 = -2 c2   c1 = 0  c1 = 5
		// 0000     c2 = ??      c2 = 0  c2 = 1
		// 0000     c3 = 0       c3 = 0  c3 = 0

		// j  = 0,1
		// fi = 0,2

		// i = 0:
		//   j = 0  row 0 fi 0 = row 0 c0 = 0
		//   j = 1  row 1 fi 0 = row 1 c0 = 0
		// i = 1:
		//   j = 0  row 0 fi 1 = row 0 c2 = 2 dep_pos = 1
		//   j = 1  row 1 fi 1 = row 1 c2 = 0

		// 0001
		// 01?0

		for (j = 0; j < rank; j++) {
			if (BIT[j][free_indices[i]] == 0)
				continue;

			(void)find_row_leader_pos(BIT[j], ncols, &dep_pos);
			nullspace_basis[i].coeffs[dep_pos] =
				mod_sub(0, BIT[j][free_indices[i]], p);
		}
		nullspace_basis[i].degree = zppoly_find_degree(
			&nullspace_basis[i], ncols - 1);
	}
	free(free_flags);
	free(free_indices);
}

// ----------------------------------------------------------------
// For square matrices.
void zpla_mats_mul(
	int     ** out,
	int     ** in1,
	int     ** in2,
	int        p,
	int        n)
{
	int i, j, k;
	int ** temp = zpla_alloc_matrix(n, n);
	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			temp[i][j] = 0;
			for (k = 0; k < n; k++) {
				temp[i][j] += in1[i][k] * in2[k][j];
				temp[i][j] %= p;
			}
		}
	}
	zpla_mats_copy(out, temp, n);
	zpla_free_matrix(temp, n, n);
}

// ----------------------------------------------------------------
// For square matrices.
void zpla_mats_copy(
	int     ** out,
	int     ** in,
	int        n)
{
	int i, j;
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			out[i][j] = in[i][j];
}

// ----------------------------------------------------------------
// For square matrices.
int zpla_is_I(
	int     ** M,
	int        n)
{
	int i, j;

	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			if (i == j) {
				if (M[i][j] != 1)
					return 0;
			}
			else {
				if (M[i][j] != 0)
					return 0;
			}
		}
	}
	return 1;
}

// ----------------------------------------------------------------
int zpla_is_zero_matrix(
	int     ** M,
	int        nrows,
	int        ncols)
{
	int i, j;
	for (i = 0; i < nrows; i++)
		for (j = 0; j < ncols; j++)
			if (M[i][j] != 0)
				return 0;
	return 1;
}

// ----------------------------------------------------------------
int zpla_is_zero_vector(
	int      * v,
	int        length)
{
	int i;
	for (i = 0; i < length; i++)
		if (v[i] != 0)
			return 0;
	return 1;
}

// ----------------------------------------------------------------
// For square matrices.
void zpla_mats_times_vr(
	int ** M,
	int  * vin,
	int  * vout,
	int    n,
	int    p)
{
	zpla_mat_times_vr(M, vin, vout, n, n, p);
}

// ----------------------------------------------------------------
void zpla_mat_times_vr(
	int ** M,
	int  * vin,
	int  * vout,
	int    nrows, // = output vector length
	int    ncols, // = input  vector length
	int    p)
{
	int i, j;
	int prod;
	int * temp = malloc_check(nrows * sizeof(int));

	for (i = 0; i < nrows; i++) {
		temp[i] = 0;
		for (j = 0; j < ncols; j++) {
			prod = (M[i][j] * vin[j]) % p;
			temp[i] = (temp[i] + prod) % p;
		}
	}
	for (i = 0; i < nrows; i++)
		vout[i] = temp[i];
	free(temp);
}
