#include "zpnlinalg.h"
#include "mod.h"
#include "malloc_check.h"

static int find_row_leader_pos(
	zpoly_t * prow,
	int       ncols,
	int     * ppos);

// ----------------------------------------------------------------
zpoly_t ** zpnla_alloc_matrix(
	int  nrows,
	int  ncols)
{
	zpoly_t ** rv;
	int i;

	rv = malloc_check(nrows * sizeof(zpoly_t *));
	for (i = 0; i < nrows; i++)
		rv[i] = malloc_check(ncols * sizeof(zpoly_t));
	return rv;
}

// ----------------------------------------------------------------
void zpnla_free_matrix(
	zpoly_t ** M,
	int    nrows,
	int    ncols)
{
	int i;
	for (i = 0; i < nrows; i++)
		free(M[i]);
	free(M);
}

// ----------------------------------------------------------------
void zpnla_print_matrix(
	zpoly_t ** M,
	int    nrows,
	int    ncols,
	zppoly_aux_t * paux)
{
	int i, j;
	for (i = 0; i < nrows; i++) {
		printf(" ");
#if 1
		for (j = 0; j < ncols; j++)
#else
		for (j = ncols-1; j >= 0; j--)
#endif
		{
			printf(" ");
			zppolymod_print(&M[i][j], paux);
		}
		printf("\n");
	}
}

// ----------------------------------------------------------------
void zpnla_print_vector(
	zpoly_t  * v,
	int        n,
	zppoly_aux_t * paux)
{
	int i, count = 0;
#if 1
	for (i = 0; i < n; i++)
#else
	for (i = n-1; i >= 0; i--)
#endif
	{
		if (count++ > 0)
			printf(" ");
		zppolymod_print(&v[i], paux);
	}
}

// ----------------------------------------------------------------
void zpnla_matrix_from_zpnpolys(
	zpoly_t  ** M,
	int         nrows,
	int         ncols,
	zpnpoly_t * polys)
{
	int i;
	for (i = 0; i < nrows; i++)
		zpnla_row_from_zpnpoly(M[i], &polys[i], ncols);
}

// ----------------------------------------------------------------
void zpnla_zpnpolys_from_matrix(
	zpnpoly_t  * polys,
	zpoly_t ** M,
	int        nrows,
	int        ncols)
{
	int i, j;
	for (i = 0; i < nrows; i++) {
		for (j = 0; j < ncols; j++) {
			polys[i].coeffs[j] = M[i][j];
			polys[i].degree = zpnpoly_find_degree(
				&polys[i], ncols - 1);
		}
	}
}

// ----------------------------------------------------------------
void zpnla_row_from_zpnpoly(
	zpoly_t   * prow,
	zpnpoly_t * pzpoly,
	int         ncols)
{
	int i;
	int d = pzpoly->degree;
	zpoly_t zero = zppoly_from_base_rep(1, 0);
	for (i = 0; i <= d; i++)
		prow[i] = pzpoly->coeffs[i];
	for (i = d+1; i < ncols; i++)
		prow[i] = zero;
}

// ----------------------------------------------------------------
void zpnla_transpose(
	zpoly_t ** Mout,
	zpoly_t ** Min,
	int        n)
{
	int row, col;

	for (row = 0; row < n; row++)
		for (col = 0; col < n; col++)
			Mout[row][col] = Min[col][row];
}

// ----------------------------------------------------------------
void zpnla_row_echelon_form(
	zpoly_t ** M,
	zppoly_aux_t * paux,
	int        nrows,
	int        ncols)
{
	zpnla_row_reduce_below (M, paux, nrows, ncols);
	zpnla_row_reduce_above (M, paux, nrows, ncols);
	zpnla_make_matrix_monic(M, paux, nrows, ncols);
}

// ----------------------------------------------------------------
static int find_row_with_leader_at_col(
	zpoly_t ** M,
	int    nrows,
	int    ncols,
	int    clear_col,
	int    leader_row,
	int  * pswap_row)
{
	int row;
	for (row = leader_row; row < nrows; row++) {
		if (!zppoly_is_zero(&M[row][clear_col])) {
			*pswap_row = row;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
static void swap_rows(
	zpoly_t ** M,
	int    nrows,
	int    ncols,
	int    leader_row,
	int    swap_row)
{
	int col;
	zpoly_t temp;

	for (col = 0; col < ncols; col++) {
		temp               = M[leader_row][col];
		M[leader_row][col] = M[swap_row][col];
		M[swap_row][col]   = temp;
	}
}

// ----------------------------------------------------------------
void zpnla_make_row_monic(
	zpoly_t * row,
	int   ncols,
	int   leader_col,
	zppoly_aux_t * paux)
{
	int col;
	zpoly_t cinv = zppolymod_recip(&row[leader_col], paux);
	for (col = leader_col; col < ncols; col++) {
		row[col] = zppolymod_mul(&row[col], &cinv, paux);
	}
}

// ----------------------------------------------------------------
void zpnla_row_reduce_below(
	zpoly_t ** M,
	zppoly_aux_t * paux,
	int        nrows,
	int        ncols)
{
	int clear_col, leader_row, swap_row, clear_row;

	for (clear_col = 0, leader_row = 0; clear_col < ncols; clear_col++) {
		if (!find_row_with_leader_at_col(M, nrows, ncols, clear_col,
			leader_row, &swap_row))
		{
			continue;
		}
		if (leader_row != swap_row)
			swap_rows(M, nrows, ncols, leader_row, swap_row);
		if (!zppoly_is_one(&M[leader_row][clear_col]))
			zpnla_make_row_monic(M[leader_row], ncols,
				clear_col, paux);
		for (clear_row = leader_row+1; clear_row < nrows; clear_row++) {
			int col;
			zpoly_t clear_leader, clear_leader_inv;
			clear_leader = M[clear_row][clear_col];
			if (zppoly_is_zero(&clear_leader))
				continue;
			clear_leader_inv = zppolymod_recip(&clear_leader, paux);
			for (col = clear_col; col < ncols; col++) {
				M[clear_row][col] = zppolymod_mul(
					&M[clear_row][col],
					&clear_leader_inv, paux);
				M[clear_row][col] = zppolymod_sub(
					&M[clear_row][col],
					&M[leader_row][col], paux);
			}
		}
		leader_row++;
	}
}

// ----------------------------------------------------------------
// Assumes row_reduce_below has already been done.

void zpnla_row_reduce_above(
	zpoly_t ** M,
	zppoly_aux_t * paux,
	int        nrows,
	int        ncols)
{
	int row, row2, row2_leader_pos, col;
	zpoly_t row_leader_val, row2_leader_val, mul;

	for (row = 0; row < nrows; row++) {
		for (row2 = row+1; row2 < nrows; row2++) {
			if (!find_row_leader_pos(M[row2], ncols,
				&row2_leader_pos))
			{
				break;
			}
			row2_leader_val = M[row2][row2_leader_pos];
			row_leader_val  = M[row][row2_leader_pos];
			if (zppoly_is_zero(&row_leader_val))
				continue;

			mul = zppolymod_div(&row_leader_val, &row2_leader_val,
				paux);
			for (col = 0; col < ncols; col++) {
				zpoly_t prod = zppolymod_mul(&M[row2][col],
					&mul, paux);
				M[row][col] = zppolymod_sub(&M[row][col],
					&prod, paux);
			}
		}
	}
}

// ----------------------------------------------------------------
// Assumes row_reduce_below has already been done.

void zpnla_make_matrix_monic(
	zpoly_t ** M,
	zppoly_aux_t * paux,
	int        nrows,
	int        ncols)
{
	int row, row_leader_pos, col;
	zpoly_t row_leader_val;

	for (row = 0; row < nrows; row++) {
		if (!find_row_leader_pos(M[row], ncols, &row_leader_pos))
			break;
		row_leader_val = M[row][row_leader_pos];

		for (col = 0; col < ncols; col++) {
			M[row][col] = zppolymod_div(
				&M[row][col], &row_leader_val, paux);
		}
	}
}

// ----------------------------------------------------------------
static int find_row_leader_pos(
	zpoly_t * prow,
	int       ncols,
	int     * ppos)
{
	int col;

	for (col = 0; col < ncols; col++) {
		if (!zppoly_is_zero(&prow[col])) {
			*ppos = col;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
// Assumes the matrix is already row-reduced.
int zpnla_find_rank(
	zpoly_t ** M,
	int    nrows,
	int    ncols)
{
	int rv = 0;
	int row;

	for (row = 0; row < nrows; row++) {
		if (!zpnla_row_is_zero(M[row], ncols))
			rv++;
	}
	return rv;
}

// ----------------------------------------------------------------
int zpnla_row_is_zero(
	zpoly_t  * prow,
	int        ncols)
{
	int col;
	for (col = 0; col < ncols; col++)
		if (!zppoly_is_zero(&prow[col]))
			return 0;
	return 1;
}

// ----------------------------------------------------------------
// Assumes the matrix is already in row-echelon form.

void zpnla_find_nullspace_basis(
	zpoly_t ** BIT,
	int        nrows,
	int        ncols,
	zppoly_aux_t * paux,
	zpnpoly_t  * nullspace_basis,
	int        dimker)
{
	int i, j;
	unsigned char * free_flags;
	unsigned char * free_indices;
	zpoly_t zero = zppoly_from_base_rep(paux->p, 0);
	zpoly_t one  = zppoly_from_base_rep(paux->p, 1);
	int nfree = 0; // == dimker but I'll compute it anyway
	int rank = nrows - dimker;
	int dep_pos;

	for (i = 0; i < dimker; i++) {
		for (j = 0; j < ncols; j++)
			nullspace_basis[i].coeffs[j] = zero;
		nullspace_basis[i].degree = ncols - 1;
	}

	free_flags   = malloc_check(nrows);
	free_indices = malloc_check(nrows);

	for (i = 0; i < nrows; i++) {
		free_flags[i] = 1;
	}

	for (i = 0; i < rank; i++) {
		if (find_row_leader_pos(BIT[i], ncols, &dep_pos))
			free_flags[dep_pos] = 0;
	}

	for (i = 0; i < nrows; i++) {
		if (free_flags[i])
			free_indices[nfree++] = i;
	}

	// For each free coefficient:
	//   Let that free coefficient be 1 and the rest be zero.
	//   Also set any dependent coefficients which depend on that
	//   free coefficient.
	for (i = 0; i < dimker; i++) {
		nullspace_basis[i].coeffs[free_indices[i]] = one;

		// (B-I)^t in row echelon form:
		//
		// 0210     c0 = ??      c0 = 1  c0 = 0
		// 1000     c1 = -2 c2   c1 = 0  c1 = 5
		// 0000     c2 = ??      c2 = 0  c2 = 1
		// 0000     c3 = 0       c3 = 0  c3 = 0

		// j  = 0,1
		// fi = 0,2

		// i = 0:
		//   j = 0  row 0 fi 0 = row 0 c0 = 0
		//   j = 1  row 1 fi 0 = row 1 c0 = 0
		// i = 1:
		//   j = 0  row 0 fi 1 = row 0 c2 = 2 dep_pos = 1
		//   j = 1  row 1 fi 1 = row 1 c2 = 0

		// 0001
		// 01?0

		for (j = 0; j < rank; j++) {
			if (zppoly_is_zero(&BIT[j][free_indices[i]]))
				continue;

			(void)find_row_leader_pos(BIT[j], ncols, &dep_pos);
			nullspace_basis[i].coeffs[dep_pos] =
				zppolymod_sub(&zero,
				&BIT[j][free_indices[i]], paux);
		}
		nullspace_basis[i].degree = zpnpoly_find_degree(
			&nullspace_basis[i], ncols - 1);
	}
	free(free_flags);
	free(free_indices);
}

// ----------------------------------------------------------------
// For square matrices.
void zpnla_mats_copy(
	zpoly_t ** out,
	zpoly_t ** in,
	int        n)
{
	int i, j;
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			out[i][j] = in[i][j];
}

// ----------------------------------------------------------------
// Square matrix times vector
void zpnla_mx_times_vr(
	zpoly_t     ** M,
	zpoly_t      * vin,
	zpoly_t      * vout,
	int            n,
	zppoly_aux_t * paux)
{
	int i, j;
	zpoly_t prod;

	for (i = 0; i < n; i++) {
		vout[i] = zppoly_from_coeffs0(0, paux->p);
		for (j = 0; j < n; j++) {
			prod = zppolymod_mul(&M[i][j], &vin[j], paux);
			vout[i] = zppolymod_add(&vout[i], &prod, paux);
		}
	}
}

// ----------------------------------------------------------------
// Product of square matrices.
void zpnla_sq_mx_times_mx(
	zpoly_t     ** out,
	zpoly_t     ** in1,
	zpoly_t     ** in2,
	int            n,
	zppoly_aux_t * paux)
{
	int i, j, k;
	zpoly_t prod;
	// In case out points to in1 or in2:
	zpoly_t ** temp = zpnla_alloc_matrix(n, n);

	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			temp[i][j] = zppoly_from_coeffs0(0, paux->p);
			for (k = 0; k < n; k++) {
				prod = zppolymod_mul(&in1[i][k], &in2[k][j],
					paux);
				temp[i][j] = zppolymod_add(&temp[i][j], &prod,
					paux);
			}
		}
	}
	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			out[i][j] = temp[i][j];
	free(temp);
}

// ----------------------------------------------------------------
void zpnla_mx_power(
	zpoly_t     ** out,
	zpoly_t     ** in,
	int            n,
	int            power,
	zppoly_aux_t * paux)
{
	// In case out points to in1 or in2:
	zpoly_t ** temp = zpnla_alloc_matrix(n, n);
	int i, j;

	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			temp[i][j] = zppoly_from_coeffs0(
				(i == j) ? 1 : 0,
				paux->p);

	for (i = 0; i < power; i++)
		zpnla_sq_mx_times_mx(temp, temp, in, n, paux);

	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			out[i][j] = temp[i][j];
	free(temp);
}
