#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "zpnpoly.h"
#include "zpnpolymod.h"
#include "mod.h"
#include "power.h"
#include "tokenize.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
void zpnpoly_check_degree(int degree)
{
	if (degree < 0) {
		fprintf(stderr,
			"zpnpoly:  degree must be non-negative; got %d.\n",
			degree);
		exit(1);
	}
	if (degree > ZPNPOLY_MAX_DEGREE) {
		fprintf(stderr,
			"zpnpoly:  degree must be <= %d; got %d.\n",
			ZPNPOLY_MAX_DEGREE, degree);
		exit(1);
	}
}

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_add(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zppoly_aux_t * paux)
{
	zpnpoly_t   out;
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++)
			out.coeffs[i] = zppoly_add(&pin1->coeffs[i],
				&pin2->coeffs[i], paux->p);
		for (i = pin2->degree + 1; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i];
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++)
			out.coeffs[i] = zppoly_add(&pin1->coeffs[i],
				&pin2->coeffs[i], paux->p);
		for (i = pin1->degree + 1; i <= pin2->degree; i++)
			out.coeffs[i] = pin2->coeffs[i];
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were added.
	out.degree = zpnpoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_sub(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zppoly_aux_t * paux)
{
	zpnpoly_t   out;
	zpoly_t    zero = zppoly_from_base_rep(paux->p, 0);
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++)
			out.coeffs[i] = zppoly_sub(&pin1->coeffs[i],
				&pin2->coeffs[i], paux->p);
		for (i = pin2->degree + 1; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i];
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++) {
			out.coeffs[i] = zppoly_sub(&pin1->coeffs[i],
				&pin2->coeffs[i], paux->p);
		}
		for (i = pin1->degree + 1; i <= pin2->degree; i++) {
			// out.coeffs[i] = negative of pin2->coeffs[i]
			out.coeffs[i] = zppoly_sub(&zero, &pin2->coeffs[i],
				paux->p);
		}
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were added.
	out.degree = zpnpoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_mul(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zppoly_aux_t * paux)
{
	zpnpoly_t out;
	int i, j;
	zpoly_t prod;

	out.degree = pin1->degree + pin2->degree;
	zpnpoly_check_degree(out.degree);
	for (i = 0; i <= out.degree; i++)
		out.coeffs[i] = zppoly_from_base_rep(paux->p, 0);

	for (i = 0; i <= pin1->degree; i++) {
		for (j = 0; j <= pin2->degree; j++) {
			prod = zppolymod_mul(&pin1->coeffs[i], &pin2->coeffs[j],
				paux);
			out.coeffs[i+j] = zppoly_add(&out.coeffs[i+j], &prod,
				paux->p);
		}
	}
	// E.g. if one polynomial was zero.  My convention is that the zero
	// polynomial has "degree" zero -- correct as an array index, but
	// mathematically incorrect.
	out.degree = zpnpoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
zpnpoly_t zpnpoly_power(
	zpnpoly_t * pin,
	int         power,
	zppoly_aux_t * paux)
{
	zpnpoly_t in2 = *pin;
	zpnpoly_t inpower = zpnpoly_from_int(paux->p, 1);

	while (power != 0) {
		if (power & 1)
			inpower = zpnpoly_mul(&in2, &inpower, paux);
		power = (unsigned)power >> 1;
		in2 = zpnpoly_mul(&in2, &in2, paux);
	}
	return inpower;
}

// ----------------------------------------------------------------
// Assumes the input is already known to be a pth power.

zpnpoly_t zpnpoly_pth_root(
	zpnpoly_t * pin,
	zppoly_aux_t * paux)
{
	int si, di;
	int j;
	zpnpoly_t out;
	zpoly_t zero = zppoly_from_base_rep(paux->p, 0);

	for (di = 0; di <= pin->degree; di++) {
		out.coeffs[di] = zero;
	}

	for (si = 0, di = 0; si <= pin->degree; si += paux->p, di++) {
		out.coeffs[di] = pin->coeffs[si];
		for (j = 1; j < paux->p; j++) {
			if (si+j > pin->degree)
				break;
			if (!zppoly_is_zero(&pin->coeffs[si+j])) {
				printf("zpnpoly_pth_root: ");
				zpnpoly_print(pin, paux);
				printf(" is not a pth power.\n");
				exit(1);
			}
		}
	}
	out.degree = zpnpoly_find_degree(&out, pin->degree);

	return out;
}

// ----------------------------------------------------------------
// Input: a, b
// Output: GCD g, and m, n such that am + bn = d

zpnpoly_t zpnpoly_gcd(
	zpnpoly_t * pa, // 1st input.
	zpnpoly_t * pb, // 2nd input.
	zppoly_aux_t * paux)
{
	zpnpoly_t c, q, r;
	zpnpoly_t d;   // Return value.

	if (zpnpoly_is_zero(pa))
		return *pb;
	if (zpnpoly_is_zero(pb))
		return *pa;

	c = *pa;
	d = *pb;

	while (1) {

		// Divide
		// q = c / d, r = c % d
		zpnpoly_quot_and_rem(&c, &d, &q, &r, paux);
		// Note:  now c = qd + r and 0 <= r < d

		// Remainder zero?
		//if (r == zero)
		if (zpnpoly_is_zero(&r))
			break;

		// Recycle
		c = d;
		d = r;
	}
	return d;
}

// ----------------------------------------------------------------
int zpnpoly_is_zero(zpnpoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (!zppoly_is_zero(&pz->coeffs[0]))
		return 0;
	return 1;
}

// ----------------------------------------------------------------
int zpnpoly_is_one(zpnpoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (!zppoly_is_one(&pz->coeffs[0]))
		return 0;
	return 1;
}

// ----------------------------------------------------------------
void zpnpoly_make_monic(
	zpnpoly_t * pz,
	zppoly_aux_t * paux)
{
	int i;
	zpoly_t one = zppoly_from_base_rep(paux->p, 1);
	zpoly_t c   = pz->coeffs[pz->degree];
	zpoly_t cinv;

	if (zppoly_is_zero(&c))
		return;
	c = zppoly_mod(&c, &paux->r, paux->p);
	cinv = zppolymod_div(&one, &c, paux);

	for (i = 0; i <= pz->degree; i++) {
		pz->coeffs[i] = zppolymod_mul(&pz->coeffs[i], &cinv, paux);
	}
}

// ----------------------------------------------------------------
static void zpnpoly_smul(
	zpnpoly_t * pout,
	zpnpoly_t * pin, // Poly
	zpoly_t  * pscalar, // Scalar
	zppoly_aux_t * paux)
{
	zpnpoly_t  out;
	int i;

	for (i = 0; i <= pin->degree; i++)
		out.coeffs[i] = zppolymod_mul(&pin->coeffs[i], pscalar, paux);

	if (zppoly_is_zero(pscalar))
		out.degree = 0;
	else
		out.degree = pin->degree;
	*pout = out;
}

// ----------------------------------------------------------------
int zpnpoly_find_degree(
	zpnpoly_t * pin,
	int        max_deg)
{
	int i;

	for (i = max_deg; i >= 0; i--)
		if (!zppoly_is_zero(&pin->coeffs[i]))
			return i;
	return 0;
}

// ----------------------------------------------------------------
// E.g.
// dividend = 1,2,3,4 (1 + 2x + 3x^2 + 4x^3)
// divisor  = 1,1,2   (1 +  x + 2x^2)
// modulus = 7
//
//             q=4,2   r = 4,3
//        +----------
// 1,1,2  |  1,2,3,4
//        |    2,2,4 shift = 1.  4/2 mod 7 = 2.  1,1,2 * 2 = 2,2,4.
//        +----------
//        |  1 0 1
//        |  4 4 1   shift = 0.  1/2 mod 7 = 4   1,1,2 * 4 = 4,4,1
//        +----------
//        |  4 3

void zpnpoly_quot_and_rem(
	zpnpoly_t * pdividend,
	zpnpoly_t * pdivisor,
	zpnpoly_t * pquot,
	zpnpoly_t * prem,
	zppoly_aux_t * paux)
{
	zpnpoly_t quot;
	zpnpoly_t rem;
	int i;
	int max_shift;
	int shift;
	zpoly_t rem_leader;
	zpoly_t divisor_leader;
	zpoly_t dlinv;
	zpoly_t multiplier;
	zpnpoly_t divisor_with_mul;
	int stop;

	max_shift = pdividend->degree - pdivisor->degree;
	if (pdividend->degree < pdivisor->degree) {
		quot.coeffs[0] = zppoly_from_base_rep(paux->p, 0);
		quot.degree = 0;
		rem = *pdividend;

		*pquot = quot;
		*prem  = rem;
		return;
	}

	quot.degree = pdividend->degree - pdivisor->degree;
	for (i = 0; i <= quot.degree; i++)
		quot.coeffs[i] = zppoly_from_base_rep(paux->p, 0);
	rem = *pdividend;

	divisor_leader = pdivisor->coeffs[pdivisor->degree];
	dlinv = zppolymod_recip(&divisor_leader, paux);
	for (shift = max_shift; shift >= 0; shift--) {
		rem_leader = rem.coeffs[shift + pdivisor->degree];
		multiplier = zppolymod_mul(&rem_leader, &dlinv, paux);
		zpnpoly_smul(&divisor_with_mul, pdivisor, &multiplier, paux);
		quot.coeffs[shift] = multiplier;
		stop = shift + divisor_with_mul.degree;
		for (i = shift; i <= stop; i++) {
			rem.coeffs[i] = zppoly_sub(
				&rem.coeffs[i],
				&divisor_with_mul.coeffs[i-shift],
				paux->p);
		}
	}

	quot.degree = zpnpoly_find_degree(&quot, quot.degree);
	rem.degree  = zpnpoly_find_degree(&rem,  rem.degree);

	*pquot = quot;
	*prem  = rem;
}

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_div(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zppoly_aux_t * paux)
{
	zpnpoly_t quot, rem;
	zpnpoly_quot_and_rem(pin1, pin2, &quot, &rem, paux);
	return quot;
}

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_mod(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zppoly_aux_t * paux)
{
	zpnpoly_t quot, rem;
	zpnpoly_quot_and_rem(pin1, pin2, &quot, &rem, paux);
	return rem;
}

// ----------------------------------------------------------------
// Example with p=2 and hexadecimal notation:
//
//   mod 13:
//   02 =  0 . . . [  0]
//   03 =  1 . . . [  1]
//   07 =  6 7 . . [  3]
//   13 =  2 3 4 5 [ 15]
//   19 =  9 b d e [ 15]
//   1f =  8 a c f [  5]
//
// 13 is the reductor r.  b is the input z.  19 is the desired minimal
// polynomial.  Fact:  If b is a root of 19, so are all the elements
// of its Frobenius orbit mod r, e.g. b^2, b^4, b^8.  These powers yield
// b, 9, d and e.  Then, (X-b)(X-9)(X-d)(X-e) is the desired minimal
// polynomial with coefficients in Fpn[x], 11001 .  Un-extend to Fp[x]
// to get 19 (in hex).

//         1 b
// *       1 9
// ===========
//         9 c
// +     1 b
// ===========
//       1 2 c
// *       1 d
// ===========
//       d 9 3
// +   1 2 c
// ===========
//     1 f 5 3
// *       1 e
// ===========
//     e 5 3 1
// + 1 f 5 3
// ===========
//   1 1 0 0 1 = 19 in hex.

// xxx FIX ME
// NOTE:  This function currently returns the characteristic polynomial,
// not the minimal polynomial.

zpoly_t zppoly_get_min_poly(
	zpoly_t * pz,
	zppoly_aux_t * paux)
{
	int n = paux->r.degree;
	int i;
	zpoly_t one = zppoly_from_base_rep(paux->p, 1);
	zpnpoly_t extrv = zpnpoly_extend_zppoly(&one);
	zpoly_t zpower = *pz;
	zpnpoly_t factor;
	zpoly_t rv;

	for (i = 0; i < n; i++) {
		factor.degree = 1;
		factor.coeffs[1] = zppoly_from_base_rep(paux->p, 1);
		factor.coeffs[0] = zppoly_usub(&zpower, paux->p);
		extrv = zpnpoly_mul(&extrv, &factor, paux);
		zpower = zppolymod_power(&zpower, paux->p, paux);
		if (zppoly_equal(&zpower, pz, paux->p))
			break;
	}
	rv = zppoly_unextend_zpnpoly(&extrv, paux);
	return rv;
}

// ----------------------------------------------------------------
int zpnpoly_equal(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zppoly_aux_t * paux)
{
	int i;
	if (pin1->degree != pin2->degree)
		return 0;
	for (i = 0; i <= pin1->degree; i++)
		if (!zppoly_equal(&pin1->coeffs[i], &pin2->coeffs[i], paux->p))
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int zpnpoly_lt(zpnpoly_t * pa, zpnpoly_t * pb)
{
	return zpnpoly_cmp(pa, CMP_LT, pb);
}
int zpnpoly_le(zpnpoly_t * pa, zpnpoly_t * pb)
{
	return zpnpoly_cmp(pa, CMP_LE, pb);
}
int zpnpoly_gt(zpnpoly_t * pa, zpnpoly_t * pb)
{
	return zpnpoly_cmp(pa, CMP_GT, pb);
}
int zpnpoly_ge(zpnpoly_t * pa, zpnpoly_t * pb)
{
	return zpnpoly_cmp(pa, CMP_GE, pb);
}
int zpnpoly_eq(zpnpoly_t * pa, zpnpoly_t * pb)
{
	return zpnpoly_cmp(pa, CMP_EQ, pb);
}
int zpnpoly_ne(zpnpoly_t * pa, zpnpoly_t * pb)
{
	return zpnpoly_cmp(pa, CMP_NE, pb);
}

// ----------------------------------------------------------------
int zpnpoly_cmp(
	zpnpoly_t * pa,
	int         cmp,
	zpnpoly_t * pb)
{
	int direction = 0; // -1 = less, 0 = equal, +1 = greater;
	int i;

	if (pa->degree > pb->degree) {
		direction =  1;
	}
	else if (pa->degree < pb->degree) {
		direction = -1;
	}
	else {
		for (i = pa->degree; i >= 0; i--) {
			if (zppoly_lt(&pa->coeffs[i], &pb->coeffs[i])) {
				direction = -1;
				break;
			}
			else if (zppoly_gt(&pa->coeffs[i], &pb->coeffs[i])) {
				direction =  1;
				break;
			}
		}
	}

	switch (cmp) {
	case CMP_LE: return (direction <= 0); break;
	case CMP_LT: return (direction <  0); break;
	case CMP_EQ: return (direction == 0); break;
	case CMP_NE: return (direction != 0); break;
	case CMP_GE: return (direction >= 0); break;
	case CMP_GT: return (direction >  0); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); return 0; break;
	}
}

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_extend_zppoly(
	zpoly_t   * pzp)
{
	int i;
	zpnpoly_t rv;

	rv.degree = pzp->degree;
	for (i = 0; i <= pzp->degree; i++) {
		rv.coeffs[i].degree = 0;
		rv.coeffs[i].coeffs[0] = pzp->coeffs[i];
	}
	return rv;
}

// ----------------------------------------------------------------
zpoly_t zppoly_unextend_zpnpoly(
	zpnpoly_t * pznp,
	zppoly_aux_t * paux)
{
	int i;
	zpoly_t rv;

	rv.degree = pznp->degree;
	for (i = 0; i <= rv.degree; i++) {
		if (pznp->coeffs[i].degree > 0) {
			printf("zpnpoly:  Can't unextend ");
			zpnpoly_print(pznp, paux);
			printf("\n");
			exit(1);
		}
		rv.coeffs[i] = pznp->coeffs[i].coeffs[0];
	}

	return rv;
}

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_from_coeffs0(
	zpoly_t * pcoeff0)
{
	zpnpoly_t rv;
	rv.degree = 0;
	rv.coeffs[0] = *pcoeff0;
	return rv;
}

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_from_coeffs1(
	zpoly_t * pcoeff1,
	zpoly_t * pcoeff0)
{
	zpnpoly_t rv;
	rv.coeffs[1] = *pcoeff1;
	rv.coeffs[0] = *pcoeff0;
	rv.degree = zpnpoly_find_degree(&rv, 1);
	return rv;
}

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_from_coeffs2(
	zpoly_t * pcoeff2,
	zpoly_t * pcoeff1,
	zpoly_t * pcoeff0)
{
	zpnpoly_t rv;
	rv.coeffs[2] = *pcoeff2;
	rv.coeffs[1] = *pcoeff1;
	rv.coeffs[0] = *pcoeff0;
	rv.degree = zpnpoly_find_degree(&rv, 2);
	return rv;
}

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_from_int(
	int p,
	int a)
{
	zpnpoly_t rv;
	rv.degree = 0;
	rv.coeffs[0] = zppoly_from_base_rep(p, a);
	return rv;
}

// ----------------------------------------------------------------
// Example:
// * a_aux   = a^2 + a + 1 = 1,1,1 which defines an F_2^2.
// * b_red   = b^6 + b + 1 = 1,0,0,0,0,1,1 which defines an F_2^6.
// * ab_poly = x^3 + x^2 + (a+1) x + a = 1:1:1,1:1,0
//   which is a cubic polynomial in F_2^2[x].
// * ab_exp  = 21 since (2^6-1) / (2^2-1) = 21.
//
// This function substitutes x with b and a with b^exp, then reduces mod
// b_red.  The above example ab_poly reduces to 0 mod b^6 + b + 1.

zpoly_t zpnpoly_flatten(
	zpnpoly_t    * pab_poly,
	zppoly_aux_t * pa_aux,
	zpoly_t      * pb_red,
	int            ab_exp)
{
	zpoly_t flat_poly = zppoly_from_base_rep(pa_aux->p, 0);
	int j, i;

	for (j = 0; j <= pab_poly->degree; j++) {
		for (i = 0; i <= pab_poly->coeffs[j].degree; i++) {
			zpoly_t term = zppoly_monomial_from_coeff(
				pab_poly->coeffs[j].coeffs[i],
				j + ab_exp * i,
				pa_aux->p);
			flat_poly = zppoly_add(&flat_poly, &term,
				pa_aux->p);
		}
	}
	flat_poly = zppoly_mod(&flat_poly, pb_red, pa_aux->p);

	return flat_poly;
}

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_from_base_rep(
	int rep,
	int p,
	int intermed_degree)
{
	zpnpoly_t rv;
	int pd = calc_power(p, intermed_degree);

	if (rep < 0) {
		fprintf(stderr, "zpnpoly_from_base_rep: negative input %d.\n",
			rep);
		exit(1);
	}

	if (intermed_degree < 1) {
		fprintf(stderr,
			"zpnpoly_from_base_rep: need deg >= 1; got %d.\n",
			intermed_degree);
		exit(1);
	}

	rv.degree = 0;
	rv.coeffs[0] = zppoly_from_base_rep(p, 0);
	while (rep) {
		rv.coeffs[rv.degree] = zppoly_from_base_rep(p, rep % pd);
		rep /= pd;
		if (rv.degree >= ZPNPOLY_MAX_DEGREE) {
			fprintf(stderr, "zpnpoly_from_base_rep: overflow.\n");
			exit(1);
		}
		rv.degree++;
	}
	if (rv.degree > 0)
		rv.degree--;

	return rv;
}

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_deriv(
	zpnpoly_t * pz,
	zppoly_aux_t * paux)
{
	zpnpoly_t rv;
	int i;

	if (pz->degree == 0) {
		return zpnpoly_from_int(paux->p, 0);
	}

	for (i = 1; i <= pz->degree; i++)
		rv.coeffs[i-1] = zppoly_scalar_mul(&pz->coeffs[i], i % paux->p,
			paux->p);
	rv.degree = zpnpoly_find_degree(&rv, pz->degree - 1);

	return rv;
}

// ----------------------------------------------------------------
zpoly_t zpnpoly_eval(
	zpoly_t      * pa,
	zpnpoly_t    * poutr,
	zppoly_aux_t * pinnaux)
{
	zpoly_t rv;
	int d;

	rv = poutr->coeffs[poutr->degree];
	for (d = poutr->degree - 1; d >= 0; d--) {
		rv = zppolymod_mul(&rv, pa, pinnaux);
		rv = zppolymod_add(&rv, &poutr->coeffs[d], pinnaux);
	}
	return rv;
}

// ----------------------------------------------------------------
int zpnpoly_is_rootless(
	zpnpoly_t    * poutr,
	zppoly_aux_t * pinnaux)
{
	zpoly_t a;
	int i;
	int pn = calc_power(pinnaux->p, pinnaux->r.degree);
	zpoly_t out;

	for (i = 0; i < pn; i++) {
		a = zppoly_from_base_rep(pinnaux->p, i);
		out = zpnpoly_eval(&a, poutr, pinnaux);
		if (zppoly_is_zero(&out))
			return 0;
	}
	return 1;
}

// ----------------------------------------------------------------
// Assumes outr is known to be irreducible.
// See if the order of x mod r is p^m-1.  Certainly, x^(p^m-1) will be 1.  But
// check x to all powers dividing p^m-1 to make sure they are *not* 1.

// ----------------------------------------------------------------
int zpnpoly_is_primitive(
	zpnpoly_t    * poutr,
	zppoly_aux_t * pinnaux)
{
	unsigned m = poutr->degree;
	unsigned pm = calc_power(pinnaux->p, m);
	unsigned pm1 = pm - 1;
	factor_info_t finfo;

	calc_factors(pm1, &finfo);
	return zpnpoly_is_primitive_f(poutr, pinnaux, pm1, &finfo);
}

// ----------------------------------------------------------------
int zpnpoly_is_primitive_f(
	zpnpoly_t     * poutr,
	zppoly_aux_t  * pinnaux,
	unsigned        pm1,
	factor_info_t * pfinfo)
{
	zpoly_t c1 = zppoly_from_base_rep(pinnaux->p, 1);
	zpoly_t c0 = zppoly_from_base_rep(pinnaux->p, 0);
	zpnpoly_t x   = zpnpoly_from_coeffs1(&c1, &c0);
	zpnpoly_t one = zpnpoly_from_coeffs0(&c1);
	zpnpoly_t xpower;
	unsigned nf, f, k;
	int is_one;

	if (zpnpoly_equal(poutr, &x, pinnaux))
		return 0;

	nf = num_divisors(pfinfo);

	for (k = 0; k < nf; k++) {
		f = kth_divisor(pfinfo, k);
		xpower = zpnpolymod_power(&x, f, poutr, pinnaux);
		is_one = zpnpoly_equal(&xpower, &one, pinnaux);
		if ((f != pm1) && is_one) {
			return 0;
		}
		if ((f == pm1) && !is_one) {
			fprintf(stderr, "ABEND!\n");
			exit(1);
		}
	}
	return 1;
}

// ----------------------------------------------------------------
void zpnpoly_print(
	zpnpoly_t    * pin,
	zppoly_aux_t * paux)
{
	int i;

	for (i = pin->degree; i >= 0; i--) {
		if (i < pin->degree)
			printf(":");
		zppolymod_print(&pin->coeffs[i], paux);
	}
}

// ----------------------------------------------------------------
int zpnpoly_scan(
	void * pvout,
	char * pin,
	void * pvaux)
{
	return zpnpoly_vscan(pvout, pin, pvaux, sizeof(zpnpoly_t));
}

// ----------------------------------------------------------------
void zpnpoly_vprint(
	void * pvin,
	void * pvaux)
{
	zpnpoly_t * pin = pvin;
	zppoly_aux_t * paux = pvaux;
	int i;

	for (i = pin->degree; i >= 0; i--) {
		if (i < pin->degree)
			printf(":");
		//zppoly_print(&pin->coeffs[i], paux->p);
		// xxx
		zppolymod_print(&pin->coeffs[i], paux);
	}
}

// ----------------------------------------------------------------
void zpnpoly_vprintsmall(
	void * pvin,
	void * pvaux)
{
	zpnpoly_t * pin = pvin;
	zppoly_aux_t * paux = pvaux;
	int i;
	for (i = pin->degree; i >= 0; i--) {
		if (i < pin->degree)
			printf(":");
		zppoly_print(&pin->coeffs[i], paux->p);
	}
}

// ----------------------------------------------------------------
int zpnpoly_list_scan(
	void * pvout,
	char * pin,
	void * pvaux)
{
	zpnpoly_t * pout = pvout;
	zppoly_aux_t * paux = pvaux;
	int    i;
	char * coeff_stringv[ZPNPOLY_MAX_DEGREE + 1];
	char   temp_line[1024];

	pout->degree = 0;
	for (i = 0; i <= ZPNPOLY_MAX_DEGREE; i++)
		pout->coeffs[i] = zppoly_from_base_rep(paux->p, 0);

	strcpy(temp_line, pin);
	pout->degree = tokenize(temp_line, ":", coeff_stringv,
		ZPNPOLY_MAX_DEGREE + 1) - 1;
	if (pout->degree < 0) {
		fprintf(stderr, "zpnpoly_list_scan:  Couldn't scan \"%s\".\n",
			pin);
		return 0;
	}

	for (i = 0; i <= pout->degree; i++) {
		if (!zppoly_scan(&pout->coeffs[pout->degree - i],
			coeff_stringv[i], paux->p))
		{
			fprintf(stderr,
				"zpnpoly:  Couldn't scan coefficient \"%s\".\n",
				coeff_stringv[pout->degree - i]);
			return 0;
		}
	}

	// Check for leading zero coefficients, which would reduce the
	// degree.
	pout->degree = zpnpoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
int zpnpoly_exp_scan(
	void * pvout,
	char * pin,
	void * pvaux)
{
	zpnpoly_t * pout = pvout;
	zppoly_aux_t * paux = pvaux;
	int    i;
	int    num_pieces;
	char * coeff_stringv[ZPNPOLY_MAX_DEGREE + 1];
	char * coeff_and_exp[2];
	int    num_sub_pieces;
	char   temp_line[1024];
	char   temp_piece[128];
	int    exp;

	pout->degree = 0;
	for (i = 0; i <= ZPNPOLY_MAX_DEGREE; i++)
		pout->coeffs[i] = zppoly_from_base_rep(paux->p, 0);

	strcpy(temp_line, pin);
	num_pieces = tokenize(temp_line, ":", coeff_stringv,
		ZPNPOLY_MAX_DEGREE + 1);
	if (num_pieces <= 0) {
		fprintf(stderr, "zpnpoly_exp_scan:  Couldn't scan \"%s\".\n",
			pin);
		return 0;
	}

	for (i = 0; i < num_pieces; i++) {
		// Split on "^"; expect zppoly on left and int on right.
		// Bounds-check the exponent.
		strcpy(temp_piece, coeff_stringv[i]);
		num_sub_pieces = tokenize(temp_piece, "^", coeff_and_exp, 2);
		if (num_sub_pieces < 1) {
			fprintf(stderr,
				"zpnpoly:  Couldn't scan coefficient \"%s\".\n",
				temp_piece);
			return 0;
		}
		else if (num_sub_pieces == 1) {
			exp = 0;
		}
		else {
			// sscanf and bounds-check the exp
			if (sscanf(coeff_and_exp[1], "%d", &exp) != 1) {
				fprintf(stderr,
				"zpnpoly:  Couldn't scan coefficient \"%s\".\n",
					temp_piece);
				return 0;
			}
			if ((exp < 0) || (exp > ZPNPOLY_MAX_DEGREE)) {
				fprintf(stderr,
				"zpnpoly:  Exponent %d out of range 0-%d.\n",
					exp, ZPNPOLY_MAX_DEGREE);
				return 0;
			}
		}

		// Remember highest exponent encountered.
		if (exp > pout->degree)
			pout->degree = exp;

		// scan and bounds-check the coeff
		if (!zppoly_scan(&pout->coeffs[exp], coeff_and_exp[0],
			paux->p))
		{
			fprintf(stderr,
				"zpnpoly:  Couldn't scan coefficient \"%s\".\n",
				temp_piece);
			return 0;
		}
	}

	// Check for leading zero coefficients, which would reduce the
	// degree.
	pout->degree = zpnpoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
int zpnpoly_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	if (sizeof(zpnpoly_t) > max_elt_bytes) {
		fprintf(stderr, "zpnpoly:  Need %d bytes for element.\n",
			sizeof(zpnpoly_t));
		exit(1);
	}

	if (strchr(pin, '^'))
		return zpnpoly_exp_scan(pvout, pin, pvaux);
	else
		return zpnpoly_list_scan(pvout, pin, pvaux);
}
