#include <stdio.h>
#include <stdlib.h>
#include "malloc_check.h"
#include "power.h"
#include "zpnpolymod.h"
#include "zpnpoly_factor.h"
#include "zpnlinalg.h"

static void gather_multiple_factors(
	zpnpoly_t  * pirred_factors,
	zppoly_aux_t * paux,
	int        num_irred_factors,
	zpnpoly_factor_info_t * pfinfo);

static void sort_up(
	zpnpoly_t  * polys,
	int        nrows);

static void pre_berlekamppn(
	zpnpoly_t  * pr,
	zppoly_aux_t * paux,
	zpnpoly_t  * pirred_factors,
	int        * pnum_irred_factors);

static void berlekamppn(
	zpnpoly_t  * pr,
	zppoly_aux_t * paux,
	zpnpoly_t  * pirred_factors,
	int      * pnum_irred_factors);

// ----------------------------------------------------------------
void zpnpoly_factor(
	zpnpoly_t  * pr,
	zppoly_aux_t * paux,
	zpnpoly_factor_info_t * pfinfo)
{
	int num_irred_factors, i;
	zpnpoly_t * irred_factors = malloc_check(
		ZPNPOLY_MAX_NUM_FACTORS * sizeof(zpnpoly_t));
	zpoly_t leader;

	num_irred_factors = 0;
	pre_berlekamppn(pr, paux, irred_factors, &num_irred_factors);

	for (i = 0; i < num_irred_factors; i++) {
		zpnpoly_make_monic(&irred_factors[i], paux);
	}

	leader = pr->coeffs[pr->degree];

	if (!zppoly_is_zero(&leader) && !zppoly_is_one(&leader)) {
		irred_factors[num_irred_factors++] =
			zpnpoly_from_coeffs0(&leader);
	}

	sort_up(irred_factors, num_irred_factors);
	gather_multiple_factors(irred_factors, paux, num_irred_factors, pfinfo);

	free(irred_factors);
}

// ----------------------------------------------------------------
static void gather_multiple_factors(
	zpnpoly_t * pirred_factors,
	zppoly_aux_t * paux,
	int       num_irred_factors,
	zpnpoly_factor_info_t * pfinfo)
{
	int i;
	zpnpoly_t  * plast;
	int nd = 0;

	if (num_irred_factors == 0) {
		pfinfo->num_distinct = 0;
		return;
	}

	plast = &pirred_factors[0];
	pfinfo->factor_counts[nd].f = *plast;
	pfinfo->factor_counts[nd].count = 0;

	for (i = 0; i < num_irred_factors; i++) {
		if (!zpnpoly_equal(&pirred_factors[i], plast, paux)) {
			nd++;
			pfinfo->factor_counts[nd].count = 1;
			pfinfo->factor_counts[nd].f = pirred_factors[i];
			plast = &pirred_factors[i];
		}
		else {
			pfinfo->factor_counts[nd].count++;
		}
	}
	pfinfo->num_distinct = nd + 1;
}

// ----------------------------------------------------------------
static void pre_berlekamppn(
	zpnpoly_t  * pr,
	zppoly_aux_t * paux,
	zpnpoly_t  * pirred_factors,
	int        * pnum_irred_factors)
{
	zpnpoly_t d, g;

	d = zpnpoly_deriv(pr, paux);
	g = zpnpoly_gcd(pr, &d, paux);

	if (zpnpoly_is_zero(&g)) {
		if (!zpnpoly_is_zero(pr)) {
			fprintf(stderr, "CODING ERROR file %s\n", __FILE__);
			exit(1);
		}
		pirred_factors[*pnum_irred_factors] = *pr;
		(*pnum_irred_factors)++;
		return;
	}
	else if (g.degree == 0) {
		// Input is squarefree.
		berlekamppn(pr, paux, pirred_factors, pnum_irred_factors);
	}
	else if (zpnpoly_is_zero(&d)) {
		// Input is a perfect square.
		// Multiplicity is p only if degree is > 0.
		zpnpoly_t s = zpnpoly_pth_root(pr, paux);
		int i;
		int multiplicity = pr->degree > 0 ? paux->p : 1;

		for (i = 0; i < multiplicity; i++) {
			pre_berlekamppn(&s, paux, pirred_factors,
				pnum_irred_factors);
		}
	}
	else {
		zpnpoly_t q;
		q = zpnpoly_div(pr, &g, paux);
		pre_berlekamppn(&g, paux, pirred_factors, pnum_irred_factors);
		pre_berlekamppn(&q, paux, pirred_factors, pnum_irred_factors);
	}
}

// ----------------------------------------------------------------
static void berlekamppn(
	zpnpoly_t  * pr,
	zppoly_aux_t * paux,
	zpnpoly_t  * pirred_factors,
	int      * pnum_irred_factors)
{
	int n = pr->degree;
	int q = calc_power(paux->p, paux->r.degree);
	zpoly_t one  = zppoly_from_base_rep(paux->p, 1);
	zpoly_t zero = zppoly_from_base_rep(paux->p, 0);
	zpnpoly_t x  = zpnpoly_from_coeffs1(&one, &zero);
	zpnpoly_t xq;
	zpnpoly_t xqi = zpnpoly_from_int(paux->p, 1);
	int row, rank, dimker, c;
	int got_it = 0;

	zpoly_t    ** BI;
	zpoly_t    ** BIT;
	zpnpoly_t * nullspace_basis;
	zpnpoly_t   h, f1, f2;

	if (n < 2) {
		pirred_factors[*pnum_irred_factors] = *pr;
		(*pnum_irred_factors)++;
		return;
	}

	BI  = zpnla_alloc_matrix(n, n);
	BIT = zpnla_alloc_matrix(n, n);
	nullspace_basis = malloc_check(n * sizeof(zpnpoly_t));

	xq = zpnpolymod_power(&x, q, pr, paux);

	// Populate the B matrix with row i equal to x^pi mod r
	for (row = 0; row < n; row++) {
		zpnla_row_from_zpnpoly(BI[row], &xqi, n);
		xqi = zpnpolymod_mul(&xqi, &xq, pr, paux);
	}

	// Replace B with B - I
	for (row = 0; row < n; row++) {
		BI[row][row] = zppolymod_sub(&BI[row][row], &one, paux);
	}

	zpnla_transpose(BIT, BI, n);
	zpnla_row_echelon_form(BIT, paux, n, n);
	rank = zpnla_find_rank(BIT, n, n);
	dimker = n - rank;

	if (dimker == 1) {
		pirred_factors[*pnum_irred_factors] = *pr;
		(*pnum_irred_factors)++;
		return;
	}

	zpnla_find_nullspace_basis(BIT, n, n, paux, nullspace_basis, dimker);

#if 1
	// Check h^q = h.  The algorithm guarantees this, *if* implemented
	// correctly.  This computation, while not strictly necessary, is a
	// nice sanity check.
	for (row = 0; row <  dimker; row++) {
		zpnpoly_t * ph = &nullspace_basis[row];
		zpnpoly_t hp;
		hp = zpnpolymod_power(ph, q, pr, paux);
		if (!zpnpoly_equal(ph, &hp, paux)) {
			fprintf(stderr, "CODING ERROR\n");
			exit(1);
		}
	}
#endif

	got_it = 0;
	for (row = 0; row <  dimker && !got_it; row++) {
		if (zpnpoly_is_one(&nullspace_basis[row]))
			continue;

		for (c = 0; c < q; c++) {
			zpoly_t cp = zppoly_from_base_rep(paux->p, c);
			zpnpoly_t cpoly = zpnpoly_from_coeffs0(&cp);
			h = zpnpoly_sub(&nullspace_basis[row], &cpoly, paux);
			f1 = zpnpoly_gcd(pr, &h, paux);

			if ((0 < f1.degree) && (f1.degree < pr->degree)) {
				got_it = 1;
				break;
			}
		}
	}

	if (!got_it) {
		printf("CODING ERROR: no non-trivial factors found: ");
		zpnpoly_print(pr, paux);
		printf("\n");
		exit(1);
	}

	f2 = zpnpoly_div(pr, &f1, paux);

	// The nullity of B-I is the number of irreducible
	// factors of r.  If the nullity is 2, we have a
	// pair of factors which are both irreducible and
	// so we don't need to recurse.
	if (dimker == 2) {
		pirred_factors[*pnum_irred_factors] = f1;
		(*pnum_irred_factors)++;
		pirred_factors[*pnum_irred_factors] = f2;
		(*pnum_irred_factors)++;
	}
	else {
		pre_berlekamppn(&f1, paux, pirred_factors, pnum_irred_factors);
		pre_berlekamppn(&f2, paux, pirred_factors, pnum_irred_factors);
	}

	zpnla_free_matrix(BI,  n, n);
	zpnla_free_matrix(BIT, n, n);
	free(nullspace_basis);
}

// ----------------------------------------------------------------
// Ascending bubble sort
static void sort_up(
	zpnpoly_t * polys,
	int       nrows)
{
	int i, top;
	zpnpoly_t  temp;

	for (top = nrows-1; top > 0; top--) {
		for (i = 0; i < top; i++) {
			if (!zpnpoly_lt(&polys[i], &polys[i+1])) {
				temp       = polys[i];
				polys[i]   = polys[i+1];
				polys[i+1] = temp;
			}
		}
	}
}

// ----------------------------------------------------------------
// Given the prime factorization p1^m1 ... pk^mk of n, how to enumerate all
// factors of n?
//
// Example 72 = 2^3 * 3^2.  Exponent on 2 is one of 0, 1, 2, 3.  Exponent on
// 3 is one of 0, 1, 2.  Number of possibilities:  product over i of (mi + 1).
// Factors are:
//
//	2^0 3^0    1
//	2^0 3^1    3
//	2^0 3^2    9
//	2^1 3^0    2
//	2^1 3^1    6
//	2^1 3^2   18
//	2^2 3^0    4
//	2^2 3^1   12
//	2^2 3^2   36
//	2^3 3^0    8
//	2^3 3^1   24
//	2^3 3^2   72

unsigned zpnpoly_num_divisors(
	zpnpoly_factor_info_t * pfinfo)
{
	unsigned rv;
	int i;

	if ((pfinfo->num_distinct < 0)
		|| (pfinfo->num_distinct > ZPNPOLY_MAX_NUM_FACTORS))
	{
		fprintf(stderr, "zpnpoly_num_divisors:  num distinct (%d) "
			"out of bounds 0-%d\n",
			pfinfo->num_distinct, ZPNPOLY_MAX_NUM_FACTORS);
		exit(1);
	}

	rv = 1;
	for (i = 0; i < pfinfo->num_distinct; i++)
		rv *= pfinfo->factor_counts[i].count + 1;
	return rv;
}

// ----------------------------------------------------------------
// See comments to the above.  k is treated as a multibase representation
// over the bases mi+1.

zpnpoly_t zpnpoly_kth_divisor(
	zpnpoly_factor_info_t * pfinfo,
	zppoly_aux_t * paux,
	unsigned k)
{
	zpnpoly_t rv;
	zpnpoly_t fp;
	int i;
	unsigned base;
	unsigned power;

	rv = zpnpoly_from_int(paux->p, 1);
	for (i = 0; i < pfinfo->num_distinct; i++) {
		base = pfinfo->factor_counts[i].count + 1;
		power = k % base;
		k = k / base;

		fp = zpnpoly_power(&pfinfo->factor_counts[i].f, power, paux);
		rv = zpnpoly_mul(&rv, &fp, paux);
	}
	return rv;
}

// ----------------------------------------------------------------
static int zpnpoly_qcmp(const void * pv1, const void * pv2)
{
	if (zpnpoly_lt((zpnpoly_t*)pv1, (zpnpoly_t*)pv2))
		return -1;
	if (zpnpoly_gt((zpnpoly_t*)pv1, (zpnpoly_t*)pv2))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
// The caller must free the return value.

zpnpoly_t * zpnpoly_get_all_divisors(
	zpnpoly_t  * pr,
	zppoly_aux_t * paux,
	unsigned * pnum_divisors)
{
	zpnpoly_factor_info_t finfo;
	unsigned nf, k;
	zpnpoly_t * all_divisors;

	zpnpoly_factor(pr, paux, &finfo);
	nf = zpnpoly_num_divisors(&finfo);
	*pnum_divisors = nf;

	all_divisors = (zpnpoly_t *)malloc_check(nf * sizeof(zpnpoly_t));
	for (k = 0; k < nf; k++) {
		all_divisors[k] = zpnpoly_kth_divisor(&finfo, paux, k);
	}
	qsort(all_divisors, nf, sizeof(zpnpoly_t), zpnpoly_qcmp);

	return all_divisors;
}

// ----------------------------------------------------------------
// The caller must free the return value.
zpoly_t * zppoly_get_roots(
	zpoly_t      * pz,
	zppoly_aux_t * paux)
{
	zpnpoly_t zx = zpnpoly_extend_zppoly(pz);
	zpnpoly_factor_info_t finfo;
	int n = pz->degree;
	int i;
	zpoly_t * rv = malloc_check(n * sizeof(zpoly_t));

	zpnpoly_factor(&zx, paux, &finfo);

	if (finfo.num_distinct != n) {
		printf("zppoly_get_roots: coding error: split failure\n");
		exit(1);
	}

	for (i = 0; i < n; i++) {
		if (finfo.factor_counts[i].count != 1) {
			printf("zppoly_get_roots: coding error: "
				"split failure\n");
			exit(1);
		}
		if (finfo.factor_counts[i].f.degree != 1) {
			printf("zppoly_get_roots: coding error: "
				"split failure\n");
			exit(1);
		}
		rv[i] = zppoly_usub(&finfo.factor_counts[i].f.coeffs[0],
			paux->p);
	}

	return rv;
}

// ----------------------------------------------------------------
// The caller must free the return value.
zpoly_t * zppolymod_get_elements_frobord(
	zpoly_t * pr,
	int       p,
	int     * pnel)
{
	int n = pr->degree;
	int nel = calc_power(p, n);
	zpoly_t * rv = malloc_check(nel * sizeof(zpoly_t));
	zpoly_t * dirreds = malloc_check(nel * sizeof(zpoly_t));
	zpoly_t * roots, f0, fcurr;
	int nirred;
	int i, didx = 0;
	zppoly_aux_t aux;

	aux.r = *pr;
	aux.p = p;

	// Factor X^(p^n) - X into irreducible factors over Fp[x].
	zppoly_find_dirreds(p, n, dirreds, &nirred, nel);

	// For each irreducible factor of X^(p^n) - X, get its roots.
	for (i = 0; i < nirred; i++) {
		roots = zppoly_get_roots(&dirreds[i], &aux);

		// Get frobit of each root.
		f0 = roots[0];
		fcurr = f0;
		do {
			if (didx >= nel) {
				fprintf(stderr, "FROBORD ABEND!\n");
				exit(1);
			}
			rv[didx++] = fcurr;
			fcurr = zppolymod_power(&fcurr, p, &aux);
		} while (!zppoly_equal(&fcurr, &f0, p));
	}

	free(dirreds);
	*pnel = nel;
	return rv;
}
