#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "zpnpolymod.h"

// ----------------------------------------------------------------
// zpnpolymod_mul
// zpnpolymod_rarmul
// zpnpolymod_rermul
//
// Multiplication may be done (at least) two ways:  (1) Take the ring product
// of the two polynomials, then reduce mod r; (2) reduce mod r at every step.
// Testing has shown that the latter method (reduce en route) is faster
// for larger p and smaller n; the former method (ring and reduce) is faster
// for smaller p and larger n.  Furthermore, the ring-and-reduce method
// requires storage space for degree up to 2d-2, where d is the degree of the
// reductor.  E.g. with maximum degree 100, we can't multiply two degree-60
// polynomials using ring-and-reduce, but we can if we use reduce-en-route.

// ----------------------------------------------------------------
// Sample of the ring-and-reduce method:
//
// Use GF(7^4), with reductor 1,0,1,3,5.  Multiply 1,2,3,4 by 2,4,6,1.
//
// First, multiply in the ring GF(7)[x]:
//
//       1 2 3 4
//       2 4 6 1
//      =========
//       1 2 3 4
//     6 5 4 3
//   4 1 5 2
// 2 4 6 1
// ==============
// 2 1 6 5 1 6 4
//
// Second, reduce the product mod the reductor (i.e. divide and take the
// remainder):
//
//                     2 1 4
//           +---------------
// 1 0 1 3 5 | 2 1 6 5 1 6 4
//           | 2 0 2 6 3
//           +---------------
//           |   1 4 6 5 6 4
//           |   1 0 1 3 5
//           +---------------
//           |     4 5 2 1 4
//           |     4 0 4 5 6
//           +---------------
//           |       5 5 3 5

static zpnpoly_t zpnpolymod_rarmul(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zpnpoly_t * pr,
	zppoly_aux_t * paux)
{
	zpnpoly_t rprod;
	zpnpoly_t quot;
	zpnpoly_t out;

	// Take the polynomial product (degree may be higher than that
	// of the reductor).
	rprod = zpnpoly_mul(pin1, pin2, paux);

	// Reduce the polynomial product modulo the reductor.
	zpnpoly_quot_and_rem(&rprod, pr, &quot, &out, paux);
	return out;
}

// ----------------------------------------------------------------
// Sample of the reduce-en-route method:
//
// Use GF(7^4), with reductor 1,0,1,3,5.  Multiply 1,2,3,4 by 2,4,6,1.
//
//       1 2 3 4
//       2 4 6 1
//      =========
//       1 2 3 4   1,2,3,4 times 1
//
//     6 5 4 3 0   1,2,3,4 times 6x
//     6 0 6 4 2   reduce mod r
//    -----------
//       5 5 6 5
//
//   4 1 5 2 0     1,2,3,4 times 4x
//   4 0 4 5 6     reduce mod r
//  -------------
//     1 1 4 1 0   times x again
//     1 0 1 3 5   reduce mod r
//  -------------
//       1 3 5 2
//
// 2 4 6 1         1,2,3,4 times 2x
// 2 0 2 6 3       reduce mod r
// --------------
//   4 4 2 4       times x again
//   4 0 4 5 6     reduce mod r
// --------------
//     4 5 6 1     times x again
//     4 0 4 5 6   reduce mod r
// --------------
//       5 2 3 1
// ==============
//       1 2 3 4   four partial products
//       5 5 6 5
//       1 3 5 2
//       5 2 3 1
// ==============
//       5 5 3 5   final product is the sum of the four partial products
//
// ----------------------------------------------------------------
// The reduce-en-route method can be optimized a little, to not have
// to shift the coefficients when multiplying a partial product by x.
// Example:
//
//       1 2 3 4
//       2 4 6 1
//      =========
//       1 2 3 4     in1 * in2[0]
//
//
//       6 5 4 3(0)  in1 * in2[1]
//      / / / / /
//     6 0 6 4 2     mul    = in1[3]
//                   in1[3] = in1[2] - red[3] * mul
//                   in1[2] = in1[1] - red[2] * mul
//                   in1[1] = in1[0] - red[1] * mul
//                   in1[0] =     0  - red[0] * mul
//    -----------
//       5 5 6 5
//
//
//       4 1 5 2(0)  in1 * in2[2]
//      / / / / /
//     4 0 4 5 6
//    -----------
//       1 1 4 1(0)
//      / / / / /
//     1 0 1 3 5
//    -----------
//       1 3 5 2
//
//
//       2 4 6 1(0)  in1 * in2[3]
//      / / / / /
//     2 0 2 6 3
//    -----------
//       4 4 2 4(0)
//      / / / / /
//     4 0 4 5 6
//    -----------
//       4 5 6 1(0)
//      / / / / /
//     4 0 4 5 6
//    -----------
//       5 2 3 1
//
//
// ==============
//       1 2 3 4
//       5 5 6 5
//       1 3 5 2
//       5 2 3 1
// ==============
//       5 5 3 5

static zpnpoly_t zpnpolymod_rermul(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zpnpoly_t * pr,
	zppoly_aux_t * paux)
{
	zpnpoly_t  temp, out;
	int i, j, power, d1;
	zpoly_t * p1c, * p2c, * ptc, * prc;
	zpnpoly_t in1red, in2red, rmonic;
	zpoly_t elim, prod;
	zpoly_t zero = zppoly_from_base_rep(paux->p, 0);

	zpnpoly_mod(&in1red, pin1, paux);
	zpnpoly_mod(&in2red, pin2, paux);
	rmonic = *pr;
	zpnpoly_make_monic(&rmonic, paux);

	p1c = &in1red.coeffs[0];
	p2c = &in2red.coeffs[0];
	ptc = &temp.coeffs[0];
	prc = &rmonic.coeffs[0];
	d1  = paux->r.degree - 1;

	out.degree = d1;
	zpnpoly_check_degree(out.degree);

	for (i = 0; i <= out.degree; i++)
		out.coeffs[i] = zero;

	for (i = 0; i <= pin2->degree; i++) {
		// Multiply in1 by scalar coefficient on current term of in2.
		for (j = 0; j <= in1red.degree; j++)
			ptc[j] = zppolymod_mul(&p1c[j], &p2c[i], paux);
		for (j = in1red.degree + 1; j < paux->r.degree; j++)
			ptc[j] = zero;
		temp.degree = d1;

		// Now need to multiply by the power of x on the current term
		// of in2, reducing mod r en route.
		for (power = 1; power <= i; power++) {
			elim = ptc[d1];
			for (j = d1 - 1; j >= 0; j--) {
				prod = zppolymod_mul(&prc[j+1], &elim, paux);
				ptc[j+1] = zppoly_sub(&ptc[j], &prod, paux->p);
			}
			prod = zppolymod_mul(&prc[0], &elim, paux);
			ptc[0] = zppoly_usub(&prod, paux->p);
		}
		temp.degree = zpnpoly_find_degree(&temp, temp.degree);
		out = zpnpoly_add(&out, &temp, paux);
	}

	// E.g. if one polynomial was zero.  My convention is that the zero
	// polynomial has "degree" zero -- correct as an array index, but
	// mathematically incorrect.
	out.degree = zpnpoly_find_degree(&out, out.degree);
	return out;
}

zpnpoly_t zpnpolymod_mul(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zpnpoly_t * pr,
	zppoly_aux_t * paux)
{
	// Reduce-en-route allows us to use bigger polys w/o overflow.
	//if ((pin1->degree + pin2->degree) > ZPNPOLY_MAX_DEGREE)
	//if (1)
	if (0)
		return zpnpolymod_rermul(pin1, pin2, pr, paux);
	else
		return zpnpolymod_rarmul(pin1, pin2, pr, paux);
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
zpnpoly_t zpnpolymod_power(
	zpnpoly_t * px,
	unsigned    power,
	zpnpoly_t * pr,
	zppoly_aux_t * paux)
{
	zpnpoly_t x2 = *px;
	zpnpoly_t rv = zpnpoly_from_int(paux->p, 1);

	while (power != 0) {
		if (power & 1)
			rv = zpnpolymod_mul(&x2, &rv, pr, paux);
		power >>= 1;
		x2 = zpnpolymod_mul(&x2, &x2, pr, paux);
	}
	return rv;
}
