#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "zpoly.h"
#include "tokenize.h"

static int zpoly_find_degree(
	zpoly_t * pin,
	int       max_deg);

static void zpoly_check_degree(
	int degree);

static zpoly_t zpoly_scalar_mul(
	zpoly_t * pin1,
	int        in2);

// ----------------------------------------------------------------
zpoly_t zpoly_add(
	zpoly_t * pin1,
	zpoly_t * pin2)
{
	zpoly_t out;
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++)
			out.coeffs[i] = pin1->coeffs[i] + pin2->coeffs[i];
		for (i = pin2->degree + 1; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i];
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i] + pin2->coeffs[i];
		for (i = pin1->degree + 1; i <= pin2->degree; i++)
			out.coeffs[i] = pin2->coeffs[i];
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were added.
	out.degree = zpoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
zpoly_t zpoly_sub(
	zpoly_t * pin1,
	zpoly_t * pin2)
{
	zpoly_t out;
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++)
			out.coeffs[i] = pin1->coeffs[i] - pin2->coeffs[i];
		for (i = pin2->degree + 1; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i];
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++)
			out.coeffs[i] = pin1->coeffs[i] - pin2->coeffs[i];
		for (i = pin1->degree + 1; i <= pin2->degree; i++)
			out.coeffs[i] = -pin2->coeffs[i];
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were
	// subtracted.
	out.degree = zpoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
zpoly_t zpoly_usub(
	zpoly_t * pin)
{
	zpoly_t out;
	int i;

	for (i = 0; i <= pin->degree; i++)
		out.coeffs[i] = -pin->coeffs[i];
	out.degree = pin->degree;
	return out;
}

// ----------------------------------------------------------------
zpoly_t zpoly_mul(
	zpoly_t * pin1,
	zpoly_t * pin2)
{
	zpoly_t  out;
	int i, j;
	int prod;

	out.degree = pin1->degree + pin2->degree;
	zpoly_check_degree(out.degree);

	for (i = 0; i <= out.degree; i++)
		out.coeffs[i] = 0;

	for (i = 0; i <= pin1->degree; i++) {
		for (j = 0; j <= pin2->degree; j++) {
			prod = pin1->coeffs[i] * pin2->coeffs[j];
			out.coeffs[i+j] = out.coeffs[i+j] + prod;
		}
	}

	// E.g. if one polynomial was zero.  My convention is that the zero
	// polynomial has "degree" zero -- correct as an array index, but
	// mathematically incorrect.
	out.degree = zpoly_find_degree(&out, out.degree);

	return out;
}

// ----------------------------------------------------------------
zpoly_t zpoly_div(
	zpoly_t * pin1,
	zpoly_t * pin2)
{
	zpoly_t quot, rem;
	zpoly_quot_and_rem(pin1, pin2, &quot, &rem);
	return quot;
}

// ----------------------------------------------------------------
zpoly_t zpoly_mod(
	zpoly_t * pin1,
	zpoly_t * pin2)
{
	zpoly_t quot, rem;
	zpoly_quot_and_rem(pin1, pin2, &quot, &rem);
	return rem;
}

// ----------------------------------------------------------------

//        +---------------------
// 2x + 1 | 2x^3 + x^2 + 3x + 4
//        | 2x^3 + x^2
//        +---------------------
//                       3x + 4
//                       2x + 1
//        +---------------------
//                        x + 3

//      0
//   +--------
// 5 | 3x + 4
//   +--------
//   | 3x + 4

//      x
//   +--------
// 2 | 3x + 4
//   | 2x
//   +--------
//   |  x + 4

// xxx cmt best-effort altho non-euclidean; needed for cyclo;
// always can divide by monics

void zpoly_quot_and_rem(
	zpoly_t * pdividend,
	zpoly_t * pdivisor,
	zpoly_t * pquot,
	zpoly_t * prem)
{
	zpoly_t quot;
	zpoly_t rem;
	int i;
	int max_shift;
	int shift;
	int rem_leader;
	int divisor_leader;
	int leader_quot;
	int leader_rem;
	zpoly_t divisor_with_mul;
	int stop;

	max_shift = pdividend->degree - pdivisor->degree;
	if (pdividend->degree < pdivisor->degree) {
		quot.coeffs[0] = 0;
		quot.degree = 0;
		rem = *pdividend;

		*pquot = quot;
		*prem  = rem;
		return;
	}

	quot.degree = pdividend->degree - pdivisor->degree;
	for (i = 0; i <= quot.degree; i++)
		quot.coeffs[i] = 0;
	rem = *pdividend;

	divisor_leader = pdivisor->coeffs[pdivisor->degree];

	// 5 6 4 1
	// 2 2

	for (shift = max_shift; shift >= 0; shift--) {

		rem_leader = rem.coeffs[shift + pdivisor->degree];

		leader_quot = rem_leader / divisor_leader;
		leader_rem  = rem_leader % divisor_leader;

		divisor_with_mul = zpoly_scalar_mul(pdivisor, leader_quot);
		quot.coeffs[shift] = leader_quot;

		stop = shift + divisor_with_mul.degree;
		for (i = shift; i <= stop; i++) {
			rem.coeffs[i] -= divisor_with_mul.coeffs[i-shift];
		}

		if (leader_rem != 0)
			break;
	}

	quot.degree = zpoly_find_degree(&quot, quot.degree);
	rem.degree  = zpoly_find_degree(&rem,  rem.degree);

	*pquot = quot;
	*prem  = rem;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
zpoly_t zpoly_power(
	zpoly_t * px,
	unsigned  power)
{
	zpoly_t x2 = *px;
	zpoly_t rv;

	rv = zpoly_from_coeffs0(1);

	while (power != 0) {
		if (power & 1)
			rv = zpoly_mul(&x2, &rv);
		power >>= 1;
		x2 = zpoly_mul(&x2, &x2);
	}
	return rv;
}

// ----------------------------------------------------------------
int zpoly_equal(
	zpoly_t * pin1,
	zpoly_t * pin2)
{
	int i;
	if (pin1->degree != pin2->degree)
		return 0;
	for (i = 0; i <= pin1->degree; i++)
		if (pin1->coeffs[i] != pin2->coeffs[i])
			return 0;
	return 1;
}

// ----------------------------------------------------------------
static zpoly_t zpoly_scalar_mul(
	zpoly_t * pin1,
	int        in2)
{
	zpoly_t  out;
	int i;

	for (i = 0; i <= pin1->degree; i++)
		out.coeffs[i] = pin1->coeffs[i] * in2;

	if (in2 == 0)
		out.degree = 0;
	else
		out.degree = pin1->degree;
	return out;
}

// ----------------------------------------------------------------
zpoly_t zpoly_from_coeffs(
	int degree,
	int * coeffs)
{
	int i;
	zpoly_t rv;

	zpoly_check_degree(degree);
	rv.degree = degree;
	for (i = 0; i <= degree; i++)
		rv.coeffs[i] = coeffs[i];
	rv.degree = zpoly_find_degree(&rv, rv.degree);
	return rv;
}

zpoly_t zpoly_from_coeffs0(
	int   coeff0)
{
	zpoly_t rv;
	rv.degree = 0;
	zpoly_check_degree(rv.degree);
	rv.coeffs[0] = coeff0;
	rv.degree = zpoly_find_degree(&rv, rv.degree);
	return rv;
}

zpoly_t zpoly_from_coeffs1(
	int   coeff1,
	int   coeff0)
{
	zpoly_t rv;
	rv.degree = 1;
	zpoly_check_degree(rv.degree);
	rv.coeffs[1] = coeff1;
	rv.coeffs[0] = coeff0;
	rv.degree = zpoly_find_degree(&rv, rv.degree);
	return rv;
}

zpoly_t zpoly_from_coeffs2(
	int   coeff2,
	int   coeff1,
	int   coeff0)
{
	zpoly_t rv;
	rv.degree = 2;
	zpoly_check_degree(rv.degree);
	rv.coeffs[2] = coeff2;
	rv.coeffs[1] = coeff1;
	rv.coeffs[0] = coeff0;
	rv.degree = zpoly_find_degree(&rv, rv.degree);
	return rv;
}

zpoly_t zpoly_monomial_from_coeff(
	int   coeffa,
	int   dega)
{
	zpoly_t rv;
	int i;
	zpoly_check_degree(dega);
	rv.degree = dega;
	for (i = 0; i <= rv.degree; i++)
		rv.coeffs[i] = 0;
	rv.coeffs[dega] = coeffa;
	rv.degree = zpoly_find_degree(&rv, rv.degree);
	return rv;
}

zpoly_t zpoly_binomial_from_coeffs(
	int   coeffa,
	int   dega,
	int   coeffb,
	int   degb)
{
	zpoly_t rv;
	int i;
	zpoly_check_degree(dega);
	zpoly_check_degree(degb);
	if (dega > degb)
		rv.degree = dega;
	else
		rv.degree = degb;
	for (i = 0; i <= rv.degree; i++)
		rv.coeffs[i] = 0;
	rv.coeffs[dega] += coeffa;
	rv.coeffs[degb] += coeffb;
	rv.degree = zpoly_find_degree(&rv, rv.degree);
	return rv;
}

// ----------------------------------------------------------------
void zpoly_print(
	zpoly_t * pin)
{
	int i;

	for (i = pin->degree; i >= 0; i--) {
		if (i < pin->degree)
			printf(",");
		printf("%d", pin->coeffs[i]);
	}
}

// ----------------------------------------------------------------
void zpoly_x_print(
	zpoly_t * pin)
{
	int i;

	for (i = pin->degree; i >= 0; i--) {
		if (i < pin->degree)
			printf(",");
		printf("%d", pin->coeffs[i]);
	}
}

// ----------------------------------------------------------------
static int zpoly_list_scan(
	zpoly_t * pout,
	char * pin)
{
	int    i;
	char * coeff_stringv[ZPOLY_MAX_DEGREE + 1];
	char   temp_line[1024];

	pout->degree = 0;
	for (i = 0; i <= ZPOLY_MAX_DEGREE; i++)
		pout->coeffs[i] = 0;

	strcpy(temp_line, pin);
	pout->degree = tokenize(temp_line, ",", coeff_stringv,
		ZPOLY_MAX_DEGREE + 1) - 1;
	if (pout->degree < 0) {
		fprintf(stderr, "zpoly_list_scan:  Couldn't scan \"%s\".\n",
			pin);
		return 0;
	}

	for (i = 0; i <= pout->degree; i++) {
		if (sscanf(coeff_stringv[pout->degree - i], "%d",
			&pout->coeffs[i]) != 1)
		{
			fprintf(stderr,
				"zpoly:  Couldn't scan coefficient \"%s\".\n",
				coeff_stringv[pout->degree - i]);
			return 0;
		}
	}

	// Check for leading zero coefficients, which would reduce the
	// degree.
	pout->degree = zpoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
static int zpoly_exp_scan(
	zpoly_t * pout,
	char * pin)
{
	int    i;
	int    num_pieces;
	char * coeff_stringv[ZPOLY_MAX_DEGREE + 1];
	char * coeff_and_exp[2];
	int    num_sub_pieces;
	char   temp_line[1024];
	char   temp_piece[128];
	int    exp;

	pout->degree = 0;
	for (i = 0; i <= ZPOLY_MAX_DEGREE; i++)
		pout->coeffs[i] = 0;

	strcpy(temp_line, pin);
	num_pieces = tokenize(temp_line, ",", coeff_stringv,
		ZPOLY_MAX_DEGREE + 1);
	if (num_pieces <= 0) {
		fprintf(stderr, "zpoly_exp_scan:  Couldn't scan \"%s\".\n",
			pin);
		return 0;
	}

	for (i = 0; i < num_pieces; i++) {
		// Split on "^"; expect ints on both sides.
		// Bounds-check the exponent.
		strcpy(temp_piece, coeff_stringv[i]);
		num_sub_pieces = tokenize(temp_piece, "^", coeff_and_exp, 2);
		if (num_sub_pieces < 1) {
			fprintf(stderr,
				"zpoly:  Couldn't scan coefficient \"%s\".\n",
				temp_piece);
			return 0;
		}
		else if (num_sub_pieces == 1) {
			exp = 0;
		}
		else {
			// sscanf and bounds-check the exp
			if (sscanf(coeff_and_exp[1], "%d", &exp) != 1) {
				fprintf(stderr,
				"zpoly:  Couldn't scan coefficient \"%s\".\n",
					temp_piece);
				return 0;
			}
			if ((exp < 0) || (exp > ZPOLY_MAX_DEGREE)) {
				fprintf(stderr,
				"zpoly:  Exponent %d out of range 0-%d.\n",
					exp, ZPOLY_MAX_DEGREE);
				return 0;
			}
		}

		// Remember highest exponent encountered.
		if (exp > pout->degree)
			pout->degree = exp;

		// sscanf and bounds-check the coeff
		if (sscanf(coeff_and_exp[0], "%d", &pout->coeffs[exp]) != 1) {
			fprintf(stderr,
				"zpoly:  Couldn't scan coefficient \"%s\".\n",
				temp_piece);
			return 0;
		}
	}

	// Check for leading zero coefficients, which would reduce the
	// degree.
	pout->degree = zpoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
int zpoly_scan(
	zpoly_t  * pout,
	char     * pin)
{
	if (strchr(pin, '^'))
		return zpoly_exp_scan(pout, pin);
	else
		return zpoly_list_scan(pout, pin);
}

// ----------------------------------------------------------------
//     a_4 x^4 + a_3  x^3 + a_2  x^2 + a_1  x + a_0
//    (a_4 x^3 + a_3  x^2 + a_2  x   + a_1) x + a_0
//   ((a_4 x^2 + a_3  x   + a_2) x   + a_1) x + a_0
//  (((a_4 x   + a_3) x   + a_2) x   + a_1) x + a_0

int zpoly_eval(zpoly_t * pz, int x)
{
	int rv;
	int d;

	rv = pz->coeffs[pz->degree];
	for (d = pz->degree - 1; d >= 0; d--) {
		rv = rv * x;
		rv += pz->coeffs[d];
	}
	return rv;
}

// ----------------------------------------------------------------
int zpoly_is_zero(zpoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (pz->coeffs[0] != 0)
		return 0;
	return 1;
}

// ----------------------------------------------------------------
int zpoly_is_one(zpoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (pz->coeffs[0] != 1)
		return 0;
	return 1;
}

// ----------------------------------------------------------------
static int zpoly_find_degree(
	zpoly_t * pin,
	int       max_deg)
{
	int i;

	for (i = max_deg; i >= 0; i--)
		if (pin->coeffs[i] != 0)
			return i;
	return 0;
}

// ----------------------------------------------------------------
static void zpoly_check_degree(int degree)
{
	if (degree < 0) {
		fprintf(stderr,
			"zpoly:  degree must be non-negative; got %d.\n",
			degree);
		exit(1);
	}
	if (degree > ZPOLY_MAX_DEGREE) {
		fprintf(stderr,
			"zpoly:  degree must be <= %d; got %d.\n",
			ZPOLY_MAX_DEGREE, degree);
		exit(1);
	}
}
