#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "zpolymod.h"
#include "tokenize.h"

// ----------------------------------------------------------------
// zpolymod_mul
// zpolymod_rarmul
// zpolymod_rermul
//
// Multiplication may be done (at least) two ways:  (1) Take the ring product
// of the two polynomials, then reduce mod r; (2) reduce mod r at every step.

// ----------------------------------------------------------------
// Sample of the ring-and-reduce method:
//
// Use GF(7^4), with reductor 1,0,1,3,5.  Multiply 1,2,3,4 by 2,4,6,1.
//
// First, multiply in the ring GF(7)[x]:
//
//       1 2 3 4
//       2 4 6 1
//      =========
//       1 2 3 4
//     6 5 4 3
//   4 1 5 2
// 2 4 6 1
// ==============
// 2 1 6 5 1 6 4
//
// Second, reduce the product mod the reductor (i.e. divide and take the
// remainder):
//
//                     2 1 4
//           +---------------
// 1 0 1 3 5 | 2 1 6 5 1 6 4
//           | 2 0 2 6 3
//           +---------------
//           |   1 4 6 5 6 4
//           |   1 0 1 3 5
//           +---------------
//           |     4 5 2 1 4
//           |     4 0 4 5 6
//           +---------------
//           |       5 5 3 5

static zpoly_t zpolymod_rarmul(
	zpoly_t * pin1,
	zpoly_t * pin2,
	zpoly_t * paux)
{
	zpoly_t   rprod, quot, out;

	// Take the polynomial product (degree may be higher than that
	// of the reductor).
	rprod = zpoly_mul(pin1, pin2);

	// Reduce the polynomial product modulo the reductor.
	zpoly_quot_and_rem(&rprod, paux, &quot, &out);
	return out;
}

zpoly_t zpolymod_mul(
	zpoly_t * pin1,
	zpoly_t * pin2,
	zpoly_t * paux)
{
	return zpolymod_rarmul(pin1, pin2, paux);
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
zpoly_t zpolymod_power(
	zpoly_t * px,
	unsigned  power,
	zpoly_t * paux)
{
	zpoly_t x2 = *px;
	zpoly_t rv;

	rv = zpoly_from_coeffs0(1);

	while (power != 0) {
		if (power & 1)
			rv = zpolymod_mul(&x2, &rv, paux);
		power >>= 1;
		x2 = zpolymod_mul(&x2, &x2, paux);
	}
	return rv;
}
