#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "zppoly.h"
#include "mod.h"
#include "power.h"
#include "factorize.h"
#include "tokenize.h"
#include "posmod.h"

// ----------------------------------------------------------------
zpoly_t zppoly_add(
	zpoly_t * pin1,
	zpoly_t * pin2,
	int       p)
{
	zpoly_t out;
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++) {
			out.coeffs[i] = pin1->coeffs[i] + pin2->coeffs[i];
			if (out.coeffs[i] >= p)
				out.coeffs[i] -= p;
		}
		for (i = pin2->degree + 1; i <= pin1->degree; i++) {
			out.coeffs[i] = pin1->coeffs[i];
		}
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++) {
			out.coeffs[i] = pin1->coeffs[i] + pin2->coeffs[i];
			if (out.coeffs[i] >= p)
				out.coeffs[i] -= p;
		}
		for (i = pin1->degree + 1; i <= pin2->degree; i++) {
			out.coeffs[i] = pin2->coeffs[i];
		}
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were added.
	out.degree = zppoly_find_degree(&out, out.degree);
	return  out;
}

// ----------------------------------------------------------------
zpoly_t zppoly_sub(
	zpoly_t * pin1,
	zpoly_t * pin2,
	int       p)
{
	zpoly_t out;
	int i;

	if (pin1->degree > pin2->degree) {
		for (i = 0; i <= pin2->degree; i++) {
			out.coeffs[i] = pin1->coeffs[i] - pin2->coeffs[i];
			if (out.coeffs[i] < 0)
				out.coeffs[i] += p;
		}
		for (i = pin2->degree + 1; i <= pin1->degree; i++) {
			out.coeffs[i] = pin1->coeffs[i];
		}
		out.degree = pin1->degree;
	}
	else {
		for (i = 0; i <= pin1->degree; i++) {
			out.coeffs[i] = pin1->coeffs[i] - pin2->coeffs[i];
			if (out.coeffs[i] < 0)
				out.coeffs[i] += p;
		}
		for (i = pin1->degree + 1; i <= pin2->degree; i++) {
			out.coeffs[i] = -pin2->coeffs[i];
			if (out.coeffs[i] < 0)
				out.coeffs[i] += p;
		}
		out.degree = pin2->degree;
	}
	// Find out if the degree was reduced by cancellation of leading
	// coefficients, when two polynomials of the same degree were
	// subtracted.
	out.degree = zppoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
zpoly_t zppoly_usub(
	zpoly_t * pin,
	int       p)
{
	zpoly_t out;
	int i;

	for (i = 0; i <= pin->degree; i++) {
		out.coeffs[i] = -pin->coeffs[i];
		if (out.coeffs[i] < 0)
			out.coeffs[i] += p;
	}
	out.degree = pin->degree;
	return out;
}

// ----------------------------------------------------------------
zpoly_t zppoly_mul(
	zpoly_t * pin1,
	zpoly_t * pin2,
	int       p)
{
	zpoly_t  out;
	int i, j;
	int prod;

	out.degree = pin1->degree + pin2->degree;
	zppoly_check_degree(out.degree);

	for (i = 0; i <= out.degree; i++)
		out.coeffs[i] = 0;

	for (i = 0; i <= pin1->degree; i++) {
		for (j = 0; j <= pin2->degree; j++) {
			prod = (pin1->coeffs[i] * pin2->coeffs[j]) % p;
			out.coeffs[i+j] = out.coeffs[i+j] + prod;
			if (out.coeffs[i+j] >= p)
				out.coeffs[i+j] -= p;
		}
	}

	// E.g. if one polynomial was zero.  My convention is that the zero
	// polynomial has "degree" zero -- correct as an array index, but
	// mathematically incorrect.
	out.degree = zppoly_find_degree(&out, out.degree);

	return out;
}

// ----------------------------------------------------------------
zpoly_t zppoly_scalar_mul(
	zpoly_t * pin1,
	int       in2,
	int       p)
{
	zpoly_t  out;
	int i;

	for (i = 0; i <= pin1->degree; i++)
		out.coeffs[i] = (pin1->coeffs[i] * in2) % p;

	if (in2 == 0)
		out.degree = 0;
	else
		out.degree = pin1->degree;
	return out;
}

// ----------------------------------------------------------------
zpoly_t zppoly_div(
	zpoly_t * pin1,
	zpoly_t * pin2,
	int       p)
{
	zpoly_t  quot, rem;

	zppoly_quot_and_rem(pin1, pin2, &quot, &rem, p);
	return quot;
}

// ----------------------------------------------------------------
zpoly_t zppoly_mod(
	zpoly_t * pin1,
	zpoly_t * pin2,
	int       p)
{
	zpoly_t  quot, rem;

	zppoly_quot_and_rem(pin1, pin2, &quot, &rem, p);
	return rem;
}

// ----------------------------------------------------------------
// E.g.
// dividend = 1,2,3,4 (1 + 2x + 3x^2 + 4x^3)
// divisor  = 1,1,2   (1 +  x + 2x^2)
// modulus = 7
//
//             q=4,2   r = 4,3
//        +----------
// 1,1,2  |  1,2,3,4
//        |    2,2,4 shift = 1.  4/2 mod 7 = 2.  1,1,2 * 2 = 2,2,4.
//        +----------
//        |  1 0 1
//        |  4 4 1   shift = 0.  1/2 mod 7 = 4   1,1,2 * 4 = 4,4,1
//        +----------
//        |  4 3

void zppoly_quot_and_rem(
	zpoly_t * pdividend,
	zpoly_t * pdivisor,
	zpoly_t * pquot,
	zpoly_t * prem,
	int       p)
{
	zpoly_t quot;
	zpoly_t rem;
	int i;
	int max_shift;
	int shift;
	int rem_leader;
	int divisor_leader;
	int dlinv;
	int multiplier;
	zpoly_t divisor_with_mul;
	int stop;

	max_shift = pdividend->degree - pdivisor->degree;
	if (pdividend->degree < pdivisor->degree) {
		quot.coeffs[0] = 0;
		quot.degree = 0;
		rem = *pdividend;

		*pquot = quot;
		*prem  = rem;
		return;
	}

	quot.degree = pdividend->degree - pdivisor->degree;
	for (i = 0; i <= quot.degree; i++)
		quot.coeffs[i] = 0;
	rem = *pdividend;

	divisor_leader = pdivisor->coeffs[pdivisor->degree];
	dlinv = mod_recip(divisor_leader, p);
	for (shift = max_shift; shift >= 0; shift--) {
		rem_leader = rem.coeffs[shift + pdivisor->degree];
		multiplier = (rem_leader * dlinv) % p;
		divisor_with_mul = zppoly_scalar_mul(pdivisor, multiplier, p);
		quot.coeffs[shift] = multiplier;

		stop = shift + divisor_with_mul.degree;
		for (i = shift; i <= stop; i++) {
			rem.coeffs[i] -= divisor_with_mul.coeffs[i-shift];
			if (rem.coeffs[i] < 0)
				rem.coeffs[i] += p;
		}
	}

	quot.degree = zppoly_find_degree(&quot, quot.degree);
	rem.degree  = zppoly_find_degree(&rem,  rem.degree);

	*pquot = quot;
	*prem  = rem;
}

// ----------------------------------------------------------------
// Input: a, b
// Output: GCD g, and m, n such that am + bn = d

zpoly_t zppoly_ext_gcd(
	zpoly_t * pa, // 1st input.
	zpoly_t * pb, // 2nd input.
	zpoly_t * pm, // Output coefficient on a
	zpoly_t * pn, // Output coefficient on b
	int       p)
{
	zpoly_t mprime, nprime, m, n, c, q, r, t, qm, qn;
	zpoly_t d;   // Return value.

	// Initialize
	mprime = zppoly_from_base_rep(p, 1);
	n      = zppoly_from_base_rep(p, 1);
	m      = zppoly_from_base_rep(p, 0);
	nprime = zppoly_from_base_rep(p, 0);
	c = *pa;
	d = *pb;


	while (1) {

		// Divide
		// q = c / d, r = c % d
		zppoly_quot_and_rem(&c, &d, &q, &r, p);
		// Note:  now c = qd + r and 0 <= r < d

		// Remainder zero?
		//if (r == zero)
		if ((r.degree == 0) && (r.coeffs[0] == 0)) {
			break;
		}

		// Recycle
		c = d;
		d = r;

		t = mprime;
		mprime = m;
		//qm = q * m;
		qm = zppoly_mul(&q, &m, p);
		//m = t - qm;
		m = zppoly_sub(&t, &qm, p);

		t = nprime;
		nprime = n;
		//qn = q * n;
		qn = zppoly_mul(&q, &n, p);
		//n = t - qn;
		n = zppoly_sub(&t, &qn, p);
	}
	*pm = m;
	*pn = n;
	return d;
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
zpoly_t zppoly_power(
	zpoly_t * pin,
	int       power,
	int       p)
{
	zpoly_t in2 = *pin;
	zpoly_t inpower = zppoly_from_base_rep(p, 1);

	while (power != 0) {
		if (power & 1)
			inpower = zppoly_mul(&in2, &inpower, p);
		power >>= 1;
		in2 = zppoly_mul(&in2, &in2, p);
	}
	return inpower;
}

// ----------------------------------------------------------------
// Assumes the input is already known to be a pth power.  Relies on the
// fact that f(x^p) = f^p(x) over Fp[x].

zpoly_t zppoly_pth_root(
	zpoly_t * pin,
	int p)
{
	int si, di;
	int j;
	zpoly_t out;

	for (di = 0; di <= pin->degree; di++) {
		out.coeffs[di] = 0;
	}

	for (si = 0, di = 0; si <= pin->degree; si += p, di++) {
		out.coeffs[di] = pin->coeffs[si];
		for (j = 1; j < p; j++) {
			if (si+j > pin->degree)
				break;
			if (pin->coeffs[si+j] != 0) {
				printf("zppoly_pth_root: ");
				zppoly_print(pin, p);
				printf(" is not a pth power.\n");
				exit(1);
			}
		}
	}
	out.degree = zppoly_find_degree(&out, pin->degree);

	return out;
}

// ----------------------------------------------------------------
// Input: a, b
// Output: GCD g, and m, n such that am + bn = d

zpoly_t zppoly_gcd(
	zpoly_t * pa, // 1st input.
	zpoly_t * pb, // 2nd input.
	int       p)
{
	zpoly_t c, q, r;
	zpoly_t d;   // Return value.

	if (zppoly_is_zero(pa))
		return *pb;
	if (zppoly_is_zero(pb))
		return *pa;

	c = *pa;
	d = *pb;

	while (1) {

		// Divide
		// q = c / d, r = c % d
		zppoly_quot_and_rem(&c, &d, &q, &r, p);
		// Note:  now c = qd + r and 0 <= r < d

		// Remainder zero?
		//if (r == zero)
		if ((r.degree == 0) && (r.coeffs[0] == 0)) {
			break;
		}

		// Recycle
		c = d;
		d = r;
	}
	return d;
}

// ----------------------------------------------------------------
int zppoly_equal(
	zpoly_t * pin1,
	zpoly_t * pin2,
	int       p)
{
	int i;
	if (pin1->degree != pin2->degree)
		return 0;
	for (i = 0; i <= pin1->degree; i++)
		if (pin1->coeffs[i] != pin2->coeffs[i])
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int  zppoly_lt(zpoly_t * pa, zpoly_t * pb)
{
	return zppoly_cmp(pa, CMP_LT, pb);
}
int  zppoly_le(zpoly_t * pa, zpoly_t * pb)
{
	return zppoly_cmp(pa, CMP_LE, pb);
}
int  zppoly_gt(zpoly_t * pa, zpoly_t * pb)
{
	return zppoly_cmp(pa, CMP_GT, pb);
}
int  zppoly_ge(zpoly_t * pa, zpoly_t * pb)
{
	return zppoly_cmp(pa, CMP_GE, pb);
}
int  zppoly_eq(zpoly_t * pa, zpoly_t * pb)
{
	return zppoly_cmp(pa, CMP_EQ, pb);
}
int  zppoly_ne(zpoly_t * pa, zpoly_t * pb)
{
	return zppoly_cmp(pa, CMP_NE, pb);
}

// ----------------------------------------------------------------
int zppoly_cmp(
	zpoly_t * pa,
	int       cmp,
	zpoly_t * pb)
{
	int direction = 0; // -1 = less, 0 = equal, +1 = greater;
	int i;

	if (pa->degree > pb->degree) {
		direction =  1;
	}
	else if (pa->degree < pb->degree) {
		direction = -1;
	}
	else {
		for (i = pa->degree; i >= 0; i--) {
			if (pa->coeffs[i] < pb->coeffs[i]) {
				direction = -1;
				break;
			}
			else if (pa->coeffs[i] > pb->coeffs[i]) {
				direction =  1;
				break;
			}
		}
	}

	switch (cmp) {
	case CMP_LE: return (direction <= 0); break;
	case CMP_LT: return (direction <  0); break;
	case CMP_EQ: return (direction == 0); break;
	case CMP_NE: return (direction != 0); break;
	case CMP_GE: return (direction >= 0); break;
	case CMP_GT: return (direction >  0); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); return 0; break;
	}
}

// ----------------------------------------------------------------
zpoly_t zppoly_from_base_rep(
	int p,
	int base_rep)
{
	zpoly_t rv;

	for (rv.degree = 0; ; rv.degree++) {
		rv.coeffs[rv.degree] = base_rep % p;
		base_rep /= p;
		if (base_rep == 0)
			break;
		if (rv.degree > ZPOLY_MAX_DEGREE)
			zppoly_check_degree(rv.degree);
	}

	return rv;
}

// ----------------------------------------------------------------
int zppoly_to_base_rep(
	zpoly_t * pz,
	int p)
{
	int rv = 0;
	int i;

	for (i = pz->degree; i >= 0; i--) {
		rv *= p;
		rv += pz->coeffs[i];
	}

	return rv;
}

// ----------------------------------------------------------------
zpoly_t zppoly_from_coeffs(
	int   p,
	int   degree,
	int * coeffs)
{
	zpoly_t rv = zpoly_from_coeffs(degree, coeffs);
	return zppoly_from_zpoly(&rv, p);
}

zpoly_t zppoly_from_coeffs0(
	int   coeff0,
	int   p)
{
	zpoly_t rv = zpoly_from_coeffs0(coeff0);
	return zppoly_from_zpoly(&rv, p);
}

zpoly_t zppoly_from_coeffs1(
	int   coeff1,
	int   coeff0,
	int   p)
{
	zpoly_t rv = zpoly_from_coeffs1(coeff1, coeff0);
	return zppoly_from_zpoly(&rv, p);
}

zpoly_t zppoly_from_coeffs2(
	int   coeff2,
	int   coeff1,
	int   coeff0,
	int   p)
{
	zpoly_t rv = zpoly_from_coeffs2(coeff2, coeff1, coeff0);
	return zppoly_from_zpoly(&rv, p);
}


// ----------------------------------------------------------------
zpoly_t zppoly_monomial_from_coeff(
	int   coeffa,
	int   dega,
	int   p)
{
	zpoly_t rv = zpoly_monomial_from_coeff(coeffa, dega);
	return zppoly_from_zpoly(&rv, p);
}

// ----------------------------------------------------------------
zpoly_t zppoly_binomial_from_coeffs(
	int   coeffa,
	int   dega,
	int   coeffb,
	int   degb,
	int   p)
{
	zpoly_t rv = zpoly_binomial_from_coeffs(coeffa, dega, coeffb, degb);
	return zppoly_from_zpoly(&rv, p);
}

// ----------------------------------------------------------------
zpoly_t zppoly_from_zpoly(
	zpoly_t * pz,
	int       p)
{
	int i;
	zpoly_t rv = *pz;
	for (i = 0; i < rv.degree; i++)
		rv.coeffs[i] = posmod(rv.coeffs[i], p);
	rv.degree = zppoly_find_degree(&rv, rv.degree);
	return rv;
}

// ----------------------------------------------------------------
int zppoly_compact_scan(
	zpoly_t      * pout,
	char         * pin,
	int            p)
{
	int si, di;

	for (di = 0; di <= ZPOLY_MAX_DEGREE; di++)
		pout->coeffs[di] = 0;

	pout->degree = strlen(pin) - 1;
	if (pout->degree < 0) {
		fprintf(stderr, "zppoly_compact_scan:  Couldn't scan \"%s\".\n",
			pin);
		return 0;
	}

	for (si = 0, di = pout->degree; di >= 0; si++, di--) {
		if (!isdigit(pin[si])) {
		fprintf(stderr, "zppoly_compact_scan:  Non-numeric character "
			"'%c' in \"%s\".\n",
			pin[si], pin);
			return 0;
		}
		pout->coeffs[di] = pin[si] - '0';
		pout->coeffs[di] %= p;
	}

	// Check for leading zero coefficients, which would reduce the
	// degree.
	pout->degree = zppoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
int zppoly_scan(
	zpoly_t      * pout,
	char         * pin,
	int            p)
{
	int i;

	if ((p < 10) && !strchr(pin, '^') && !strchr(pin, ','))
		return zppoly_compact_scan(pout, pin, p);

	if (!zpoly_scan(pout, pin))
		return 0;
	for (i = 0; i <= pout->degree; i++)
		pout->coeffs[i] = posmod(pout->coeffs[i], p);
	pout->degree = zppoly_find_degree(pout, pout->degree);
	return 1;
}

// ----------------------------------------------------------------
void zppoly_x_print(
	zpoly_t * pin,
	int       p)
{
	zpoly_x_print(pin);
}


// ----------------------------------------------------------------
void zppoly_print(
	zpoly_t * pin,
	int       p)
{
	int i;

	for (i = pin->degree; i >= 0; i--) {
#if 1
		if (p > 9)
#endif
			if (i < pin->degree)
				printf(",");
		printf("%d", pin->coeffs[i]);
	}
}

// ----------------------------------------------------------------
void zppoly_dprint(
	zpoly_t * pin,
	int       p,
	int       min_deg)
{
	int i;
	int width;
	int printval;
	int deg;

	if (pin->degree > min_deg)
		deg = pin->degree;
	else
		deg = min_deg;

	if (p > 100)
		width = 3;
	else if (p > 10)
		width = 2;
	else
		width = 1;
	for (i = deg; i >= 0; i--) {
		if (p > 10)
			if (i < deg)
				printf(",");

		if (i > pin->degree)
			printval = 0;
		else
			printval = pin->coeffs[i];
		printf("%*d", width, printval);
	}
}

// ----------------------------------------------------------------
//     a_4 x^4 + a_3  x^3 + a_2  x^2 + a_1  x + a_0
//    (a_4 x^3 + a_3  x^2 + a_2  x   + a_1) x + a_0
//   ((a_4 x^2 + a_3  x   + a_2) x   + a_1) x + a_0
//  (((a_4 x   + a_3) x   + a_2) x   + a_1) x + a_0

int zppoly_eval(
	zpoly_t * pz,
	int p,
	int x)
{
	int rv;
	int d;

	rv = pz->coeffs[pz->degree];
	for (d = pz->degree - 1; d >= 0; d--) {
		rv = (rv * x) % p;
		rv += pz->coeffs[d];
		if (rv >= p)
			rv -= p;
	}
	return rv;
}

// ----------------------------------------------------------------
// r(x) = 2x^2 + x + 1 (p = 3)
// z(x) = x + 2
// r(z(x)) = 2(x+2)^2 + (x+2) + 1

zpoly_t zppoly_compose(
	zpoly_t * pr,
	zpoly_t * pz,
	int       p)
{
	zpoly_t rz = zppoly_from_base_rep(p, 0);
	zpoly_t zpower = zppoly_from_base_rep(p, 1);
	zpoly_t czpower;
	int d;

	for (d = 0; d <= pr->degree; d++) {
		czpower = zppoly_scalar_mul(&zpower, pr->coeffs[d], p);
		rz = zppoly_add(&rz, &czpower, p);
		zpower = zppoly_mul(&zpower, pz, p);
	}

	return rz;
}

// ----------------------------------------------------------------
zpoly_t zppoly_deriv(
	zpoly_t * pz,
	int p)
{
	zpoly_t rv;
	int i;

	if (pz->degree == 0)
		return zppoly_from_base_rep(p, 0);

	for (i = 1; i <= pz->degree; i++)
		rv.coeffs[i-1] = (pz->coeffs[i] * i) % p;
	rv.degree = zppoly_find_degree(&rv, pz->degree - 1);

	return rv;
}

// ----------------------------------------------------------------
int zppoly_is_rootless(zpoly_t * pz, int p)
{
	int x;
	for (x = 0; x < p; x++)
		if (zppoly_eval(pz, p, x) == 0)
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int zppoly_count_non_zero_terms(zpoly_t * pz)
{
	int rv = 0;
	int i;

	for (i = 0; i <= pz->degree; i++)
		if (pz->coeffs[i] != 0)
			rv++;
	return rv;
}

// ----------------------------------------------------------------
int zppoly_is_zero(zpoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (pz->coeffs[0] != 0)
		return 0;
	return 1;
}

// ----------------------------------------------------------------
int zppoly_is_one(zpoly_t * pz)
{
	if (pz->degree != 0)
		return 0;
	if (pz->coeffs[0] != 1)
		return 0;
	return 1;
}

// ----------------------------------------------------------------
void zppoly_make_monic(
	zpoly_t * pz,
	int       p)
{
	int i, c, cinv;

	c = pz->coeffs[pz->degree];
	if (c == 0)
		return;
	cinv = mod_recip(c, p);

	for (i = 0; i <= pz->degree; i++) {
		pz->coeffs[i] *= cinv;
		pz->coeffs[i] %= p;
	}
}

// ----------------------------------------------------------------
int zppoly_find_degree(
	zpoly_t * pin,
	int       max_deg)
{
	int i;

	for (i = max_deg; i >= 0; i--)
		if (pin->coeffs[i] != 0)
			return i;
	return 0;
}

// ----------------------------------------------------------------
void zppoly_check_degree(int degree)
{
	if (degree < 0) {
		fprintf(stderr,
			"zppoly:  degree must be non-negative; got %d.\n",
			degree);
		exit(1);
	}
	if (degree > ZPOLY_MAX_DEGREE) {
		fprintf(stderr,
			"zppoly:  degree must be <= %d; got %d.\n",
			ZPOLY_MAX_DEGREE, degree);
		exit(1);
	}
}

// ----------------------------------------------------------------
int zppoly_qcmp(const void * pv1, const void * pv2)
{
	if (zppoly_lt((zpoly_t*)pv1, (zpoly_t*)pv2))
		return -1;
	if (zppoly_gt((zpoly_t*)pv1, (zpoly_t*)pv2))
		return 1;
	return 0;
}
