// xxx cmt zpoly_t uses deg & rest of coeffs[] can be garbage.
// this is for efficiency -- to avoid computing a lot of needless zeroes.
// however, here we have matrices which need to be square, including
// past-degree zero terms.  Requires using different-than-usual zppoly fcns.

#include <stdio.h>
#include <stdlib.h>
#include "malloc_check.h"
#include "zppolymod.h"
#include "zppoly_factor.h"
#include "zplinalg.h"
#include "mod.h"

static void gather_multiple_factors(
	zpoly_t  * pirred_factors,
	int        p,
	int        num_irred_factors,
	zppoly_factor_info_t * pfinfo);

static void sort_up(
	zpoly_t  * polys,
	int        nrows);

static void pre_berlekampp(
	zpoly_t  * pr,
	int        p,
	zpoly_t  * pirred_factors,
	int      * pnum_irred_factors);

static void berlekampp(
	zpoly_t  * pr,
	int        p,
	zpoly_t  * pirred_factors,
	int      * pnum_irred_factors);

// ----------------------------------------------------------------
void zppoly_factor(
	zpoly_t  * pr,
	int        p,
	zppoly_factor_info_t * pfinfo)
{
	int num_irred_factors, i;
	zpoly_t * irred_factors = malloc_check(
		ZPPOLY_MAX_NUM_FACTORS * sizeof(zpoly_t));
	int leader;

	num_irred_factors = 0;
	pre_berlekampp(pr, p, irred_factors, &num_irred_factors);

	for (i = 0; i < num_irred_factors; i++)
		zppoly_make_monic(&irred_factors[i], p);

	leader = pr->coeffs[pr->degree];
	if (leader > 1) {
		irred_factors[num_irred_factors++] = zppoly_from_base_rep(p,
			leader);
	}

	sort_up(irred_factors, num_irred_factors);
	gather_multiple_factors(irred_factors, p, num_irred_factors, pfinfo);

	free(irred_factors);
}

// ----------------------------------------------------------------
static void gather_multiple_factors(
	zpoly_t * pirred_factors,
	int       p,
	int       num_irred_factors,
	zppoly_factor_info_t * pfinfo)
{
	int i;
	zpoly_t  * plast;
	int nd = 0;

	if (num_irred_factors == 0) {
		pfinfo->num_distinct = 0;
		return;
	}

	plast = &pirred_factors[0];
	pfinfo->factor_counts[nd].f = *plast;
	pfinfo->factor_counts[nd].count = 0;

	for (i = 0; i < num_irred_factors; i++) {
		if (!zppoly_equal(&pirred_factors[i], plast, p)) {
			nd++;
			pfinfo->factor_counts[nd].count = 1;
			pfinfo->factor_counts[nd].f = pirred_factors[i];
			plast = &pirred_factors[i];
		}
		else {
			pfinfo->factor_counts[nd].count++;
		}
	}
	pfinfo->num_distinct = nd + 1;
}

// ----------------------------------------------------------------
static void pre_berlekampp(
	zpoly_t  * pr,
	int        p,
	zpoly_t  * pirred_factors,
	int      * pnum_irred_factors)
{
	zpoly_t d, g;

	d = zppoly_deriv(pr, p);
	g = zppoly_gcd(pr, &d, p);

	if (zppoly_is_zero(&g)) {
		if (!zppoly_is_zero(pr)) {
			fprintf(stderr, "CODING ERROR file %s\n", __FILE__);
			exit(1);
		}
		pirred_factors[*pnum_irred_factors] = *pr;
		(*pnum_irred_factors)++;
		return;
	}
	else if (g.degree == 0) {
		// Input is squarefree.
		berlekampp(pr, p, pirred_factors, pnum_irred_factors);
	}
	else if (zppoly_is_zero(&d)) {
		// Input is a perfect square.
		// Multiplicity is p only if degree is > 0.
		zpoly_t s = zppoly_pth_root(pr, p);
		int i;
		int multiplicity = pr->degree > 0 ? p : 1;

		for (i = 0; i < multiplicity; i++) {
			pre_berlekampp(&s, p, pirred_factors,
				pnum_irred_factors);
		}
	}
	else {
		zpoly_t q;
		q = zppoly_div(pr, &g, p);
		pre_berlekampp(&g, p, pirred_factors, pnum_irred_factors);
		pre_berlekampp(&q, p, pirred_factors, pnum_irred_factors);
	}
}

//// ----------------------------------------------------------------
//// Berlekamp's algorithm requires squarefree input.  In the literature,
//// this is often described as being done by factoring out GCD(r(x), r'(x)).
//// This function takes a different approach to find squareful inputs, and also
//// often produces a complete factorization:
////
//// X^(p^n) - X factors into all irreducibles of degree d for all d dividing
//// n.  Therefore, if r (the input, of degree n) has an irreducible factor
//// of degree < n, that factor will divide X^(p^d) - X for some d < n.
//// If r has GCD = 1 with X^(p^d) - X for all d < n, then r is irreducible.
//// Reduction mod r is taken at each step to avoid overflow, with no effect
//// on the GCD.
//
//static void pre_berlekampp(
//	zpoly_t  * pr,
//	int        p,
//	zpoly_t  * pirred_factors,
//	int      * pnum_irred_factors)
//{
//	zpoly_t x = zppoly_from_base_rep(p, p);
//	zpoly_t u = zppoly_from_base_rep(p, p);
//	zpoly_t ux, g;
//	zppoly_aux_t aux;
//	int d = pr->degree;
//	int i;
//
//	aux.p = p;
//	aux.r = *pr;
//	for (i = 1; i < d; i++) {
//		// u = u**p mod r; test GCD(u - x, r)
//		u = zppolymod_power(&u, p, &aux);
//		ux = zppoly_sub(&u, &x, p);
//		if (zppoly_is_zero(&ux)) {
//			// If we got here, r is known to be reducible but this
//			// algorithm can't distinguish its factors.  However,
//			// it is known that r is squarefree, suitable for input
//			// to Berlekamp's algorithm.  (Incidentally, all of r's
//			// factors are of the same degree, but that fact won't
//			// be needed.)
//			berlekampp(pr, p, pirred_factors, pnum_irred_factors);
//			return;
//		}
//		g = zppoly_gcd(pr, &ux, p);
//		// See if the GCD is not a non-zero scalar
//		if ((g.degree > 0) || zppoly_is_zero(&g)) {
//			// If we got here, r is known to be reducible and g
//			// is a factor.
//			zpoly_t  q;
//			zppoly_div(&q, pr, &g, p);
//			pre_berlekampp(&g, p, pirred_factors,
//				pnum_irred_factors);
//			pre_berlekampp(&q, p, pirred_factors,
//				pnum_irred_factors);
//			return;
//		}
//	}
//	// If we got here, r is known to be irreducible.
//	pirred_factors[*pnum_irred_factors] = *pr;
//	(*pnum_irred_factors)++;
//}

// ----------------------------------------------------------------
static void berlekampp(
	zpoly_t  * pr,
	int        p,
	zpoly_t  * pirred_factors,
	int      * pnum_irred_factors)
{
	int       n = pr->degree;
	zpoly_t   x = zppoly_from_base_rep(p, p);
	zpoly_t   xp;
	zpoly_t   xpi = zppoly_from_base_rep(p, 1);
	zppoly_aux_t aux;
	int row, rank, dimker, c;
	int got_it = 0;

	int    ** BI;
	int    ** BIT;
	zpoly_t * nullspace_basis;
	zpoly_t   h, f1, f2;

	if (n < 2) {
		pirred_factors[*pnum_irred_factors] = *pr;
		(*pnum_irred_factors)++;
		return;
	}

	BI  = zpla_alloc_matrix(n, n);
	BIT = zpla_alloc_matrix(n, n);
	nullspace_basis = malloc_check(n * sizeof(zpoly_t));

	aux.p = p;
	aux.r = *pr;
	xp = zppolymod_power(&x, p, &aux);

	// Populate the B matrix with row i equal to x^pi mod r
	for (row = 0; row < n; row++) {
		zpla_row_from_zpoly(BI[row], &xpi, n);
		xpi = zppolymod_mul(&xpi, &xp, &aux);
	}

	// Replace B with B - I
	for (row = 0; row < n; row++) {
		BI[row][row] = mod_sub(BI[row][row], 1, p);
	}


	zpla_transpose(BIT, BI, n);
	zpla_row_echelon_form(BIT, p, n, n);
	rank = zpla_find_rank(BIT, n, n);
	dimker = n - rank;

	if (dimker == 1) {
		pirred_factors[*pnum_irred_factors] = *pr;
		(*pnum_irred_factors)++;
		return;
	}

	zpla_find_nullspace_basis(BIT, n, n, p, nullspace_basis, dimker);

#if 1
	// Check h^p = h.  The algorithm guarantees this, *if* implemented
	// correctly.  This computation, while not strictly necessary, is a
	// nice sanity check.
	for (row = 0; row <  dimker; row++) {
		zpoly_t * ph = &nullspace_basis[row];
		zpoly_t hp;
		hp = zppolymod_power(ph, p, &aux);
		if (!zppoly_equal(ph, &hp, p)) {
			fprintf(stderr, "CODING ERROR\n");
			exit(1);
		}
	}
#endif

	got_it = 0;
	for (row = 0; row <  dimker && !got_it; row++) {
		if (zppoly_is_one(&nullspace_basis[row]))
			continue;

		for (c = 0; c < p; c++) {
			zpoly_t cpoly = zppoly_from_base_rep(p, c);
			h = zppoly_sub(&nullspace_basis[row], &cpoly, p);
			f1 = zppoly_gcd(pr, &h, p);

			if ((0 < f1.degree) && (f1.degree < pr->degree)) {
				got_it = 1;
				break;
			}
		}
	}

	if (!got_it) {
		printf("CODING ERROR: no non-trivial factors found: ");
		zppoly_print(pr, p);
		printf("\n");
		exit(1);
	}

	f2 = zppoly_div(pr, &f1, p);

	// The nullity of B-I is the number of irreducible
	// factors of r.  If the nullity is 2, we have a
	// pair of factors which are both irreducible and
	// so we don't need to recurse.
	if (dimker == 2) {
		pirred_factors[*pnum_irred_factors] = f1;
		(*pnum_irred_factors)++;
		pirred_factors[*pnum_irred_factors] = f2;
		(*pnum_irred_factors)++;
	}
	else {
		pre_berlekampp(&f1, p, pirred_factors, pnum_irred_factors);
		pre_berlekampp(&f2, p, pirred_factors, pnum_irred_factors);
	}

	zpla_free_matrix(BI,  n, n);
	zpla_free_matrix(BIT, n, n);
	free(nullspace_basis);

}

// ----------------------------------------------------------------
// Ascending bubble sort
static void sort_up(
	zpoly_t * polys,
	int       nrows)
{
	int i, top;
	zpoly_t  temp;

	for (top = nrows-1; top > 0; top--) {
		for (i = 0; i < top; i++) {
			if (!zppoly_lt(&polys[i], &polys[i+1])) {
				temp       = polys[i];
				polys[i]   = polys[i+1];
				polys[i+1] = temp;
			}
		}
	}
}

// ----------------------------------------------------------------
// Given the prime factorization p1^m1 ... pk^mk of n, how to enumerate all
// factors of n?
//
// Example 72 = 2^3 * 3^2.  Exponent on 2 is one of 0, 1, 2, 3.  Exponent on
// 3 is one of 0, 1, 2.  Number of possibilities:  product over i of (mi + 1).
// Factors are:
//
//	2^0 3^0    1
//	2^0 3^1    3
//	2^0 3^2    9
//	2^1 3^0    2
//	2^1 3^1    6
//	2^1 3^2   18
//	2^2 3^0    4
//	2^2 3^1   12
//	2^2 3^2   36
//	2^3 3^0    8
//	2^3 3^1   24
//	2^3 3^2   72

unsigned zppoly_num_divisors(
	zppoly_factor_info_t * pfinfo)
{
	unsigned rv;
	int i;

	if ((pfinfo->num_distinct < 0)
		|| (pfinfo->num_distinct > ZPPOLY_MAX_NUM_FACTORS))
	{
		fprintf(stderr, "zppoly_num_divisors:  num distinct (%d) "
			"out of bounds 0-%d\n",
			pfinfo->num_distinct, ZPPOLY_MAX_NUM_FACTORS);
		exit(1);
	}

	rv = 1;
	for (i = 0; i < pfinfo->num_distinct; i++)
		rv *= pfinfo->factor_counts[i].count + 1;
	return rv;
}

// ----------------------------------------------------------------
// See comments to the above.  k is treated as a multibase representation
// over the bases mi+1.

zpoly_t zppoly_kth_divisor(
	zppoly_factor_info_t * pfinfo,
	int      p,
	unsigned k)
{
	zpoly_t rv;
	zpoly_t fp;
	int i;
	unsigned base;
	unsigned power;

	rv = zppoly_from_coeffs0(1, p);
	for (i = 0; i < pfinfo->num_distinct; i++) {
		base = pfinfo->factor_counts[i].count + 1;
		power = k % base;
		k = k / base;

		fp = zppoly_power(&pfinfo->factor_counts[i].f, power, p);
		rv = zppoly_mul(&rv, &fp, p);
	}
	return rv;
}

// ----------------------------------------------------------------
// The caller must free the return value.

zpoly_t * zppoly_get_all_divisors(
	zpoly_t  * pr,
	int        p,
	unsigned * pnum_divisors)
{
	zppoly_factor_info_t finfo;
	unsigned nf, k;
	zpoly_t * all_divisors;

	zppoly_factor(pr, p, &finfo);
	nf = zppoly_num_divisors(&finfo);
	*pnum_divisors = nf;

	all_divisors = (zpoly_t *)malloc_check(nf * sizeof(zpoly_t));
	for (k = 0; k < nf; k++) {
		all_divisors[k] = zppoly_kth_divisor(&finfo, p, k);
	}
	qsort(all_divisors, nf, sizeof(zpoly_t), zppoly_qcmp);

	return all_divisors;
}

// ----------------------------------------------------------------
void zppoly_print_factors(
	zpoly_t * pr,
	int       p)
{
	zppoly_factor_info_t finfo;
	int i;

	zppoly_factor(pr, p, &finfo);
	zppoly_print(pr, p);
	printf(": ");
	for (i = 0; i < finfo.num_distinct; i++) {
		printf(" ");
		zppoly_print(&finfo.factor_counts[i].f, p);
		if (finfo.factor_counts[i].count > 1)
			printf("^%d", finfo.factor_counts[i].count);
	}
}
