#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "zppolymod.h"
#include "power.h"
#include "posmod.h"
#include "mod.h"
#include "malloc_check.h"

static int zppolymod_element_order_irr(
	zpoly_t * pr,
	zppoly_aux_t * paux);
static int zppolymod_element_order_red(
	zpoly_t * pr,
	zppoly_aux_t * paux);

// ----------------------------------------------------------------
// zppolymod_mul
// zppolymod_rarmul
// zppolymod_rermul
//
// Multiplication may be done (at least) two ways:  (1) Take the ring product
// of the two polynomials, then reduce mod r; (2) reduce mod r at every step.
// Testing has shown that the latter method (reduce en route) is faster
// for larger p and smaller n; the former method (ring and reduce) is faster
// for smaller p and larger n.  Furthermore, the ring-and-reduce method
// requires storage space for degree up to 2d-2, where d is the degree of the
// reductor.  E.g. with maximum degree 100, we can't multiply two degree-60
// polynomials using ring-and-reduce, but we can if we use reduce-en-route.

// ----------------------------------------------------------------
// Sample of the ring-and-reduce method:
//
// Use GF(7^4), with reductor 1,0,1,3,5.  Multiply 1,2,3,4 by 2,4,6,1.
//
// First, multiply in the ring GF(7)[x]:
//
//       1 2 3 4
//       2 4 6 1
//      =========
//       1 2 3 4
//     6 5 4 3
//   4 1 5 2
// 2 4 6 1
// ==============
// 2 1 6 5 1 6 4
//
// Second, reduce the product mod the reductor (i.e. divide and take the
// remainder):
//
//                     2 1 4
//           +---------------
// 1 0 1 3 5 | 2 1 6 5 1 6 4
//           | 2 0 2 6 3
//           +---------------
//           |   1 4 6 5 6 4
//           |   1 0 1 3 5
//           +---------------
//           |     4 5 2 1 4
//           |     4 0 4 5 6
//           +---------------
//           |       5 5 3 5

static zpoly_t zppolymod_rarmul(
	zpoly_t * pin1,
	zpoly_t * pin2,
	zppoly_aux_t * paux)
{
	zpoly_t rprod;
	zpoly_t quot;
	zpoly_t out;

	// Take the polynomial product (degree may be higher than that
	// of the reductor).
	rprod = zppoly_mul(pin1, pin2, paux->p);

	// Reduce the polynomial product modulo the reductor.
	zppoly_quot_and_rem(&rprod, &paux->r, &quot, &out, paux->p);
	return out;
}

// ----------------------------------------------------------------
// Sample of the reduce-en-route method:
//
// Use GF(7^4), with reductor 1,0,1,3,5.  Multiply 1,2,3,4 by 2,4,6,1.
//
//       1 2 3 4
//       2 4 6 1
//      =========
//       1 2 3 4   1,2,3,4 times 1
//
//     6 5 4 3 0   1,2,3,4 times 6x
//     6 0 6 4 2   reduce mod r
//    -----------
//       5 5 6 5
//
//   4 1 5 2 0     1,2,3,4 times 4x
//   4 0 4 5 6     reduce mod r
//  -------------
//     1 1 4 1 0   times x again
//     1 0 1 3 5   reduce mod r
//  -------------
//       1 3 5 2
//
// 2 4 6 1         1,2,3,4 times 2x
// 2 0 2 6 3       reduce mod r
// --------------
//   4 4 2 4       times x again
//   4 0 4 5 6     reduce mod r
// --------------
//     4 5 6 1     times x again
//     4 0 4 5 6   reduce mod r
// --------------
//       5 2 3 1
// ==============
//       1 2 3 4   four partial products
//       5 5 6 5
//       1 3 5 2
//       5 2 3 1
// ==============
//       5 5 3 5   final product is the sum of the four partial products
//
// ----------------------------------------------------------------
// The reduce-en-route method can be optimized a little, to not have
// to shift the coefficients when multiplying a partial product by x.
// Example:
//
//       1 2 3 4
//       2 4 6 1
//      =========
//       1 2 3 4     in1 * in2[0]
//
//
//       6 5 4 3(0)  in1 * in2[1]
//      / / / / /
//     6 0 6 4 2     mul    = in1[3]
//                   in1[3] = in1[2] - red[3] * mul
//                   in1[2] = in1[1] - red[2] * mul
//                   in1[1] = in1[0] - red[1] * mul
//                   in1[0] =     0  - red[0] * mul
//    -----------
//       5 5 6 5
//
//
//       4 1 5 2(0)  in1 * in2[2]
//      / / / / /
//     4 0 4 5 6
//    -----------
//       1 1 4 1(0)
//      / / / / /
//     1 0 1 3 5
//    -----------
//       1 3 5 2
//
//
//       2 4 6 1(0)  in1 * in2[3]
//      / / / / /
//     2 0 2 6 3
//    -----------
//       4 4 2 4(0)
//      / / / / /
//     4 0 4 5 6
//    -----------
//       4 5 6 1(0)
//      / / / / /
//     4 0 4 5 6
//    -----------
//       5 2 3 1
//
//
// ==============
//       1 2 3 4
//       5 5 6 5
//       1 3 5 2
//       5 2 3 1
// ==============
//       5 5 3 5

static zpoly_t zppolymod_rermul(
	zpoly_t * pin1,
	zpoly_t * pin2,
	zppoly_aux_t * paux)
{
	zpoly_t  temp, out;
	int i, j, power, elim, d1;
	int * p1c, * p2c, * ptc, * prc;
	zpoly_t in1red, in2red, rmonic;

	in1red = zppoly_mod(pin1, &paux->r, paux->p);
	in2red = zppoly_mod(pin2, &paux->r, paux->p);
	rmonic = paux->r;
	zppoly_make_monic(&rmonic, paux->p);

	p1c = &in1red.coeffs[0];
	p2c = &in2red.coeffs[0];
	ptc = &temp.coeffs[0];
	prc = &rmonic.coeffs[0];
	d1  = paux->r.degree - 1;

	out.degree = d1;
	zppoly_check_degree(out.degree);

	for (i = 0; i <= out.degree; i++)
		out.coeffs[i] = 0;

	for (i = 0; i <= pin2->degree; i++) {
		// Multiply in1 by scalar coefficient on current term of in2.
		for (j = 0; j <= in1red.degree; j++)
			ptc[j] = (p1c[j] * p2c[i]) % paux->p;
		for (j = in1red.degree + 1; j < paux->r.degree; j++)
			ptc[j] = 0;
		temp.degree = d1;

		// Now need to multiply by the power of x on the current term
		// of in2, reducing mod r en route.
		for (power = 1; power <= i; power++) {
			elim = ptc[d1];
			for (j = d1 - 1; j >= 0; j--) {
				ptc[j+1] = posmod(ptc[j] - prc[j+1] * elim,
					paux->p);
			}
			ptc[0]  = posmod(-prc[0] * elim, paux->p);
		}
		temp.degree = zppoly_find_degree(&temp, temp.degree);
		out = zppoly_add(&out, &temp, paux->p);
	}

	// E.g. if one polynomial was zero.  My convention is that the zero
	// polynomial has "degree" zero -- correct as an array index, but
	// mathematically incorrect.
	out.degree = zppoly_find_degree(&out, out.degree);
	return out;
}

// ----------------------------------------------------------------
zpoly_t zppolymod_mul(
	zpoly_t * pin1,
	zpoly_t * pin2,
	zppoly_aux_t * paux)
{
	// Reduce-en-route allows us to use bigger polys w/o overflow.
	if ((pin1->degree + pin2->degree) > ZPOLY_MAX_DEGREE)
	//if (1)
		return zppolymod_rermul(pin1, pin2, paux);
	else
		return zppolymod_rarmul(pin1, pin2, paux);
}

// ----------------------------------------------------------------
zpoly_t zppolymod_div(
	zpoly_t * pin1,
	zpoly_t * pin2,
	zppoly_aux_t * paux)
{
	zpoly_t binv;

	// 1/b mod r
	binv = zppolymod_recip(pin2, paux);

	// a * (1/b) mod r
	return zppolymod_mul(pin1, &binv, paux);
}

// ----------------------------------------------------------------
// Finite-field division uses extended Euclidean GCD, with ring division
// (quotient and remainder) as a substep.
//
// a / b is a * 1/b.  To compute 1/b, find the GCD d.  (Since the reductor
// is irreducible -- although we don't test this on every division -- and
// if b is monic, the GCD had better be 1).  Then, compute 1 = m*b + n*r.
// Since r = 0 mod r, 1 = m*b and therefore m = 1/b.  The extended Euclidean
// algorithm (see Knuth vol. 1) computes d, m and n all at once.
//
// An alternate reciprocation algorithm uses the fact that a finite field
// GF(p^m) has its multiplicative group of order p^m-1.  So, by Euler's
// criterion, x^(p^n-2) = x^(-1) for all non-zero elements of the field.
//
// Testing has shown the extended-GCD reciprocation is faster for small p
// and large n; p^n-2 is faster for large p and small n.  (Small p and small
// n were too quick for a naive timing test; large p and large n both take
// too long to test.)

zpoly_t zppolymod_recip(
	zpoly_t * pin,
	zppoly_aux_t * paux)
{
#if 1
	zpoly_t g, m, n;
	int c0i;
	int i;

	g = zppoly_ext_gcd(&paux->r, pin, &m, &n, paux->p);

	// Ext. GCD actually returns a scalar multiple of 1.  Divide this out.
	c0i = mod_recip(g.coeffs[0], paux->p);
	for (i = 0; i <= n.degree; i++)
		n.coeffs[i] = (n.coeffs[i] * c0i) % paux->p;
	return n;
#else
	int pm2 = calc_power(paux->p, paux->r.degree) - 2;
	return zppolymod_power(pin, pm2, paux);
#endif
}

// ----------------------------------------------------------------
// Uses repeated squaring for efficiency.
zpoly_t zppolymod_power(
	zpoly_t * px,
	unsigned  power,
	zppoly_aux_t * paux)
{
	zpoly_t x2 = *px;
	zpoly_t rv = zppoly_from_base_rep(paux->p, 1);

	while (power != 0) {
		if (power & 1)
			rv = zppolymod_mul(&x2, &rv, paux);
		power >>= 1;
		x2 = zppolymod_mul(&x2, &x2, paux);
	}
	return rv;
}

// ----------------------------------------------------------------
int zppolymod_element_order(
	zpoly_t * pr,
	zppoly_aux_t * paux)
{
	int is_irr, rv;
	// Cache:
	static int     have_last = 0;
	static zpoly_t last_r;
	static int     last_p;
	static int     last_i;

	if (zppoly_is_zero(pr))
		return 0;
	if ((zppoly_to_base_rep(pr, paux->p) == paux->p)
	&& (zppoly_to_base_rep(&paux->r, paux->p) == paux->p))
		// x mod x, the latter being irreducible.
		return 0;

	if (have_last &&
		(paux->p == last_p) &&
		zppoly_equal(pr, &last_r, paux->p))
	{
		is_irr = last_i;
	}
	else {
		is_irr = zppoly_is_irreducible(&paux->r, paux->p);
		last_i = is_irr;
		last_r = *pr;
		last_p = paux->p;
		have_last = 1;
	}

	if (is_irr)
		rv = zppolymod_element_order_irr(pr, paux);
	else
		rv = zppolymod_element_order_red(pr, paux);

	return rv;
}

static int zppolymod_element_order_irr(
	zpoly_t * pr,
	zppoly_aux_t * paux)
{
	unsigned long long pm = calc_power_ull((unsigned long long)paux->p,
		(unsigned long long)paux->r.degree);
	unsigned long long pm1 = pm - 1;
	unsigned nd, k;
	unsigned long long * all_divisors = 0;
	unsigned long long f = 0;
	zpoly_t zpower;
	int got_it = 0;

	all_divisors = get_all_divisors_ull(pm1, &nd);

	for (k = 0; k < nd; k++) {
		f = all_divisors[k];
		zpower = zppolymod_power(pr, f, paux);
		if (zppoly_to_base_rep(&zpower, paux->p) == 1) {
			got_it = 1;
			break;
		}
	}

	free(all_divisors);
	if (got_it) {
		return f;
	}
	else {
		printf("zppolymod_element_order_irr:  Internal error:  "
			"order of ");
		zppoly_print(pr, paux->p);
		printf(" mod ");
		zppoly_print(&paux->r, paux->p);
		printf(" not found.\n");
		exit(1);
		return 0; // Not reached.
	}
}

// When r is reducible, nonetheless the invertible elements (units) of the
// polymod ring do form a multiplicative group.  An efficient order calculator
// would find the order of that multiplicative group, then try raising r only
// to powers dividing that order.  This function, though, is a brute-force
// search.
//
// How does one compute a sort of totient function for Fp[x] without factoring
// r?  Or, maybe it's worthwhile to go ahead and factor r.

static int zppolymod_element_order_red(
	zpoly_t * pr,
	zppoly_aux_t * paux)
{
	int rv;
	int max = calc_power(paux->p, paux->r.degree);
	zpoly_t rpower = *pr;

	for (rv = 1; rv < max; rv++) {
		if (zppoly_is_one(&rpower))
			return rv;
		rpower = zppolymod_mul(&rpower, pr, paux);
	}

	printf("zppolymod_element_order_red:  Internal error:  order of ");
	zppoly_print(pr, paux->p);
	printf(" mod ");
	zppoly_print(&paux->r, paux->p);
	printf(" not found.\n");
	exit(1);
	return 0; // Not reached.
}

// ----------------------------------------------------------------
zpoly_t zppolymod_find_generator(
	zppoly_aux_t * paux)
{
	int n = paux->r.degree;
	int pn1 = calc_power(paux->p, n) - 1;
	int gilo = 1;//paux->p;
	int gihi = pn1;
	int gi;
	zpoly_t g;

	if (gihi < gilo)
		gihi = gilo; // linear fields

	for (gi = gilo; gi <= gihi; gi++) {
		g = zppoly_from_base_rep(paux->p, gi);
		if (zppolymod_element_order(&g, paux) == pn1)
			return g;
	}

	printf("Generator for ");
	zppoly_print(&paux->r, paux->p);
	printf(" not found.\n");
	exit(1);
}

// ----------------------------------------------------------------
int zppolymod_log(
	zpoly_t * pz, // Input
	zpoly_t * pg, // Base, assumed (!) primitive mod r.
	zppoly_aux_t * paux) // Modulus for FF arithmetic
{
	int log;
	zpoly_t gp = *pg;
	int deg = paux->r.degree;
	int pm = calc_power(paux->p, deg);

	// find e such that g^e == z mod r
	if (zppoly_is_zero(pz))
		return 0;

	for (log = 1; log <= pm+2; log++) {
		if (zppoly_equal(pz, &gp, paux->p))
			return log;
		gp = zppolymod_mul(&gp, pg, paux);
	}

	printf("zppoly_log: ");
	zppoly_print(pg, paux->p);
	printf(" is not primitive mod ");
	zppoly_print(&paux->r, paux->p);
	printf("\n");
	exit(1);
	return 0; // not reached
}

// ----------------------------------------------------------------
zpoly_t zppolymod_add(
	zpoly_t * pin1,
	zpoly_t * pin2,
	zppoly_aux_t * paux)
{
	return zppoly_add(pin1, pin2, paux->p);
}

// ----------------------------------------------------------------
zpoly_t zppolymod_sub(
	zpoly_t * pin1,
	zpoly_t * pin2,
	zppoly_aux_t * paux)
{
	return zppoly_sub(pin1, pin2, paux->p);
}

// ----------------------------------------------------------------
// We want things like Cayley tables to line up, so we want a fixed
// width for field elements.  However, when printing irreducibles
// or intermediate products (i.e. before reducing modulo r), we
// want to see all the coefficients.

void zppolymod_print(
	zpoly_t * pin,
	zppoly_aux_t * paux)
{
	zppoly_dprint(pin, paux->p, paux->r.degree-1);
}

// ----------------------------------------------------------------
int zppolymod_auxscan(
	zppoly_aux_t *  paux,
	char * p_string,
	char * m_or_red_string)
{
	if (sscanf(p_string, "%d", &paux->p) != 1) {
		fprintf(stderr, "zppoly:  Couldn't scan \"%s\" as p.\n",
			p_string);
		return 0;
	}
	if (paux->p < 2) {
		fprintf(stderr, "zppoly:  p must be >= 2; got %d.\n", paux->p);
		return 0;
	}

	// If it doesn't start with ":", parse it as a user-specified reductor.
	// Else, parse as int m and search for a reductor of that degree.
	if (m_or_red_string[0] != ':') {
		if (!zppoly_scan(&paux->r, m_or_red_string, paux->p)) {
			// Error message already printed out.
			return 0;
		}
		if (paux->r.coeffs[paux->r.degree] != 1) {
			fprintf(stderr, "zppoly:  r must be monic; got "
				"leading coefficient %d.\n",
				paux->r.coeffs[paux->r.degree]);
			return 0;
		}
	}
	else {
		if (sscanf(m_or_red_string, ":%d", &paux->r.degree) != 1) {
			fprintf(stderr, "zppoly:  Couldn't scan \"%s\" as m.\n",
				m_or_red_string);
			return 0;
		}
		if (paux->r.degree > ZPOLY_MAX_DEGREE) {
			fprintf(stderr, "zppoly:  m must be <= %d; got %d.\n",
				ZPOLY_MAX_DEGREE, paux->r.degree);
			return 0;
		}
		if (paux->r.degree < 1) {
			fprintf(stderr, "zppoly:  m must be >= 1; got %d.\n",
				paux->r.degree);
			return 0;
		}
		*paux = zppoly_find_irreducible(paux->p, paux->r.degree);
	}
	return 1;
}

// ----------------------------------------------------------------
zppoly_aux_t zppoly_aux_from_base_rep(
	int p,
	int m,
	int base_rep)
{
	zppoly_aux_t rv;
	rv.r = zppoly_from_base_rep(p, base_rep);
	rv.p = p;
	return rv;
}

// ----------------------------------------------------------------
zppoly_aux_t zppoly_aux_from_coeffs(
	int p,
	int m,
	int * coeffs)
{
	zppoly_aux_t rv;
	rv.r = zppoly_from_coeffs(p, m, coeffs);
	rv.p = p;
	return rv;
}

// ----------------------------------------------------------------
// r(x) = 2x^2 + x + 1 (p = 3)
// z(x) = x + 2
// r(z(x)) = 2(x+2)^2 + (x+2) + 1

zpoly_t zppolymod_compose(
	zpoly_t * pr,
	zpoly_t * pz,
	zpoly_t * pm,
	int       p)
{
	zpoly_t rz = zppoly_from_base_rep(p, 0);
	zpoly_t zpower = zppoly_from_base_rep(p, 1);
	zpoly_t czpower;
	zppoly_aux_t aux;
	int d;

	aux.p = p;
	aux.r = *pm;

	for (d = 0; d <= pr->degree; d++) {
		czpower = zppoly_scalar_mul(&zpower, pr->coeffs[d], p);
		rz = zppoly_add(&rz, &czpower, p);
		zpower = zppolymod_mul(&zpower, pz, &aux);
	}

	return rz;
}

// ----------------------------------------------------------------
// X^(p^n) - X factors into all irreducibles of degree d for all d dividing
// n.  Therefore, if r (the input, of degree n) has an irreducible factor
// of degree < n, that factor will divide X^(p^d) - X for some d < n.
// If r has GCD = 1 with X^(p^d) - X for all d < n, then r is irreducible.
// Reduction mod r is taken at each step to avoid overflow, with no effect
// on the GCD.

int zppoly_is_irreducible(zpoly_t * pr, int p)
{
	zpoly_t x = zppoly_from_base_rep(p, p);
	zpoly_t u = zppoly_from_base_rep(p, p);
	zpoly_t ux, g;
	zppoly_aux_t aux;
	int d = pr->degree;
	int i;

	aux.p = p;
	aux.r = *pr;
	for (i = 1; i < d; i++) {
		// u = u**p mod r; test GCD(u - x, r).
		u = zppolymod_power(&u, p, &aux);
		ux = zppoly_sub(&u, &x, p);
		if (zppoly_is_zero(&ux))
			return 0;
		g = zppoly_gcd(pr, &ux, p);
		// See if the GCD is not a non-zero scalar
		if ((g.degree > 0) || zppoly_is_zero(&g))
			return 0;
	}
	return 1;
}

// ----------------------------------------------------------------
// This is a brute-force search, not an efficient algorithm (see above).

// This is trial division.  Just as with factoring integers, where we
// need to try only up to sqrt(n), not n, here also we need to try factors
// only up to and including degree m/2:  e.g. if z has degree 37 and we've
// tried all factors of degree up to 18, we needn't try degree 19 and above.
// If there were a degree-19 factor, it would have a degree-18 cofactor which
// we would have already found.

// How to compute the bound using base-p representation:
//
// m odd:  up to and including degree m/2-1
// E.g. p = 2, m = 5: up to     x^2+x+1.  This is  111 = 2^3-1
// Bound in base rep is p^(m/2 + 1) - 1.
//
// m even: up to and including degree m/2
// E.g. p = 2, m = 6: up to x^3+x^2+x+1.  This is 1111 = 2^4-1
// Bound in base rep is p^(m/2 + 1) - 1.

int zppoly_is_irreducible_brute(zpoly_t * pr, int p)
{
	int i;
	zpoly_t a, q, r;
	int pm2;
	int m = pr->degree;

	if (pr->degree <= 1)
		return 1;

	// has root => reducible
	// i.e. not rootless => not irreducible
	if (pr->degree > 1)
		if (!zppoly_is_rootless(pr, p))
			return 0;

	pm2 = calc_power(p, m/2 + 1);
	// Start at p rather than 1, since scalars will always divide evenly.
	for (i = p; i < pm2; i++) {
		a = zppoly_from_base_rep(p, i);
		zppoly_quot_and_rem(pr, &a, &q, &r, p);
		if ((r.degree == 0) && (r.coeffs[0] == 0)) {
			if (q.degree > 0) {
				return 0;
			}
		}
	}
	return 1;
}

// ----------------------------------------------------------------
void zppoly_list_irreducibles(int p, int m)
{
	zpoly_t r;
	int i;
	int pm, tpm;
	int nfound = 0;
	int ntried = 0;

	// xxx need to sanity-check p and m.
	printf("Irreducibles over GF(");
	if (m == 1)
		printf("%d", p);
	else
		printf("%d^%d", p, m);
	printf("):\n");

	pm = calc_power(p, m);
	tpm = pm * 2;
	for (i = pm; i < tpm; i++) {
		ntried++;
		r = zppoly_from_base_rep(p, i);
		if (zppoly_is_irreducible(&r, p)) {
			nfound++;
			printf("  ");
			zppoly_print(&r, p);
			printf("\n");
		}
	}
	printf("%6d tried; %6d found.\n", ntried, nfound);
}


// ----------------------------------------------------------------
zppoly_aux_t zppoly_find_irreducible(int p, int m)
{
	zppoly_aux_t rv;
	int i;
	int pm, tpm;

	// xxx need to sanity-check p and m.
	rv.p = p;
	pm = calc_power(p, m);
	tpm = pm * 2;
	for (i = pm; i < tpm; i++) {
		rv.r = zppoly_from_base_rep(p, i);
		if (zppoly_is_irreducible(&rv.r, p))
			if (zppoly_is_primitive(&rv.r, p, m))
				return rv;
	}
	fprintf(stderr,
	"zppoly: coding error: couldn't find any irreds in GF(%d^%d).\n",
		p, m);
	exit(1);
}

// ----------------------------------------------------------------
// Assumes z is known to be irreducible.
//
// See if the order of x mod z is p^m-1.  Certainly, x^(p^m-1) will be 1.
// But check x to all powers dividing p^m-1 to make sure they are *not* 1.

int zppoly_is_primitive(zpoly_t * pr, int p, int m)
{
	int pm = calc_power(p, m);
	unsigned pm1 = (unsigned)pm - 1;
	factor_info_t finfo;

	if (pr->degree == 0)
		return 0;

	calc_factors(pm1, &finfo);
	return zppoly_is_primitive_f(pr, p, m, pm1, &finfo);
}

// ----------------------------------------------------------------
int zppoly_is_primitive_f(
	zpoly_t * pr,
	int p,
	int m,
	unsigned pm1,
	factor_info_t * pfinfo)
{
	zpoly_t x = zppoly_from_coeffs1(1, 0, p);
	zpoly_t xpower;
	zppoly_aux_t aux;
	unsigned nf, f, k;

	if (zppoly_to_base_rep(pr, p) == p)
		return 0;

	nf = num_divisors(pfinfo);
	aux.p = p;
	aux.r = *pr;
	for (k = 0; k < nf; k++) {
		f = kth_divisor(pfinfo, k);
		xpower = zppolymod_power(&x, f, &aux);
		if ((f != pm1) && zppoly_is_one(&xpower)) {
			return 0;
		}
		if ((f == pm1) && !zppoly_is_one(&xpower)) {
			printf("ABEND!\n");

			printf("pm1=%u f=%u r=", pm1, f);
			zppoly_print(pr, p);
			printf(" deg=%d ", pr->degree);
			printf(" x=");
			zppoly_print(&x, p);
			printf(" x^f=");
			zppoly_print(&xpower, p);
			printf("\n");

			exit(1);
		}
	}
	return 1;
}

// ----------------------------------------------------------------
zpoly_t * zppolymod_get_elements(
	int p,
	int n,
	int * pnel)
{
	int nel = calc_power(p, n);
	zpoly_t * rv = malloc_check(nel * sizeof(zpoly_t));
	int i;

	for (i = 0; i < nel; i++)
		rv[i] = zppoly_from_base_rep(p, i);

	*pnel = nel;
	return rv;
}

// ----------------------------------------------------------------
void zppoly_find_dirreds(
	int        p,
	unsigned   n,
	zpoly_t  * pirred_factors,
	unsigned * pnum_irred_factors,
	unsigned   max_irred_factors)
{
	unsigned num_irred_factors = 0;
	unsigned d;
	int ilo, ihi, i;
	zpoly_t f;

	for (d = 1; d <= n; d++) {
		if ((n % d) != 0)
			continue;

		ilo = calc_power(p, d);
		ihi = ilo << 1;
		for (i = ilo; i < ihi; i++) {
			f = zppoly_from_base_rep(p, i);
			if (!zppoly_is_irreducible(&f, p))
				continue;

			if (num_irred_factors >= max_irred_factors) {
				fprintf(stderr, "Too many factors.\n");
				exit(1);
			}
			pirred_factors[num_irred_factors] = f;
			num_irred_factors++;

		}
	}

	*pnum_irred_factors = num_irred_factors;
}

// ================================================================
static void zppolymod_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zppoly_aux_t * paux = pvaux;
	zpoly_t * pout = pvout;
	zpoly_t out = zppoly_add(pvin1, pvin2, paux->p);
	*pout = out;
}

// ----------------------------------------------------------------
static void zppolymod_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zppoly_aux_t * paux = pvaux;
	zpoly_t * pout = pvout;
	zpoly_t out = zppoly_sub(pvin1, pvin2, paux->p);
	*pout = out;
}

// ----------------------------------------------------------------
static void zppolymod_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpoly_t * pout = pvout;
	zpoly_t out = zppolymod_mul(pvin1, pvin2, pvaux);
	*pout = out;
}

// ----------------------------------------------------------------
static void zppolymod_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpoly_t * pout = pvout;
	zpoly_t out = zppolymod_div(pvin1, pvin2, pvaux);
	*pout = out;
}

// ----------------------------------------------------------------
static int zppolymod_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zppoly_aux_t * paux = pvaux;
	return zppoly_equal(pvin1, pvin2, paux->p);
}

// ----------------------------------------------------------------
static int zppolymod_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpoly_t * pin1 = pvin1;
	zpoly_t * pin2 = pvin2;
	int i;
	if (pin1->degree < pin2->degree)
		return 1;
	if (pin1->degree > pin2->degree)
		return 0;
	for (i = 0; i <= pin1->degree; i++) {
		if (pin1->coeffs[i] < pin2->coeffs[i])
			return 1;
		if (pin1->coeffs[i] > pin2->coeffs[i])
			return 0;
	}
	return 0;
}

// ----------------------------------------------------------------
static void zppolymod_make_add_id(
	void * pvin,
	void * pvaux)
{
	zpoly_t * pin = pvin;
	int i;

	for (i = 0; i <= ZPOLY_MAX_DEGREE; i++)
		pin->coeffs[i] = 0;
	pin->degree = 0;
}

// ----------------------------------------------------------------
static void zppolymod_make_mul_id(
	void * pvin,
	void * pvaux)
{
	zpoly_t * pin = pvin;
	int i;

	for (i = 1; i <= ZPOLY_MAX_DEGREE; i++)
		pin->coeffs[i] = 0;
	pin->coeffs[0] = 1;
	pin->degree = 0;
}

// ----------------------------------------------------------------
static void zppolymod_vprint(
	void * pvin,
	void * pvaux)
{
	zppolymod_print(pvin, pvaux);
}

// ----------------------------------------------------------------
static void zppolymod_vauxprint(
	void * pvaux)
{
	zppolymod_print(pvaux, pvaux);
}

// ----------------------------------------------------------------
static int zppolymod_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	zppoly_aux_t * paux = pvaux;
	if (sizeof(zpoly_t) > max_elt_bytes) {
		fprintf(stderr, "zppoly:  Need %d bytes for element.\n",
			sizeof(zpoly_t));
		exit(1);
	}
	return zppoly_scan(pvout, pin, paux->p);
}

// ----------------------------------------------------------------
static int zppolymod_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	if (max_aux_bytes < sizeof(zppoly_aux_t)) {
		fprintf(stderr, "zppoly:  Need %d bytes for aux.\n",
			sizeof(zppoly_aux_t));
		exit(1);
	}

	if (argc != 2) {
		fprintf(stderr, "zppoly:  Need p and r aux arguments.\n");
		return 0;
	}

	return zppolymod_auxscan(pvaux, argv[0], argv[1]);
}

// ----------------------------------------------------------------
static void zppolymod_vindex_to_element(
	int     which,
	int     p,
	int     m,
	zpoly_t * pelement)
{
	int i;

	for (i = 0; i <= m; i++) {
		pelement->coeffs[i] = which % p;
		which /= p;
	}
	pelement->degree = zppoly_find_degree(pelement, m);
}

// ----------------------------------------------------------------
static int  zppolymod_add_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	zpoly_t * pelements = pvelements;
	zppoly_aux_t * paux = pvaux;
	int i;
	int group_size;

	group_size = calc_power(paux->p, paux->r.degree);

	if (group_size > max_group_bytes/sizeof(zpoly_t))
		return 0;
	for (i = 0; i < group_size; i++) {
		zppolymod_vindex_to_element(i, paux->p, paux->r.degree,
			&pelements[i]);
	}
	*pnum_elements = group_size;
	return 1;
}

// ----------------------------------------------------------------
static int zppolymod_mul_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	zpoly_t * pelements = pvelements;
	zppoly_aux_t * paux = pvaux;
	int i;
	int group_size;

	group_size = calc_power(paux->p, paux->r.degree) - 1;

	if (group_size > max_group_bytes/sizeof(zpoly_t))
		return 0;
	for (i = 0; i < group_size; i++) {
		zppolymod_vindex_to_element(i+1, paux->p, paux->r.degree,
			&pelements[i]);
	}
	*pnum_elements = group_size;
	return 1;
}

// ================================================================
group_spec_t zppolymod_add_group_spec = {
	element_size  : sizeof(zpoly_t),
	naux          : 2,
	pvop          : zppolymod_vadd,
	pvequal       : zppolymod_vequal,
	pvless        : zppolymod_vless,
	pvprint       : zppolymod_vprint,
	pvauxprint    : zppolymod_vauxprint,
	pvscan        : zppolymod_vscan,
	pvauxscan     : zppolymod_vauxscan,
	pfill_group   : zppolymod_add_fill_group,
	desc          : "zppolyadd"
};

group_spec_t zppolymod_mul_group_spec = {
	element_size  : sizeof(zpoly_t),
	naux          : 2,
	pvop          : zppolymod_vmul,
	pvequal       : zppolymod_vequal,
	pvless        : zppolymod_vless,
	pvprint       : zppolymod_vprint,
	pvauxprint    : zppolymod_vauxprint,
	pvscan        : zppolymod_vscan,
	pvauxscan     : zppolymod_vauxscan,
	pfill_group   : zppolymod_mul_fill_group,
	desc          : "zppolymul"
};

//ring_spec_t zppolymod_ring_spec = {
//	element_size  : sizeof(zpoly_t),
//	naux          : 1,
//	pvadd         : zppolymod_vadd,
//	pvsub         : zppolymod_vsub,
//	pvmul         : zppolymod_vmul,
//	pvdiv         : zppolymod_vdiv,
//	pvequal       : zppolymod_vequal,
//	pvless        : zppolymod_vless,
//	pvprint       : zppolymod_vprint,
//	pvauxprint    : zppolymod_vauxprint,
//	pvscan        : zppolymod_vscan,
//	pvauxscan     : zppolymod_vauxscan,
//	pfill_ring    : zppolymod_add_fill_group,
//	desc          : "zppoly"
//};

vfops_t zppolymod_vfops = {
	element_size  : sizeof(zpoly_t),
	pvadd         : zppolymod_vadd,
	pvsub         : zppolymod_vsub,
	pvmul         : zppolymod_vmul,
	pvdiv         : zppolymod_vdiv,
	pvequal       : zppolymod_vequal,
	pvless        : zppolymod_vless,
	pvprint       : zppolymod_vprint,
	pvmake_add_id : zppolymod_make_add_id,
	pvmake_mul_id : zppolymod_make_mul_id
};
