#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "z2nipoly.h"
#include "z2ilinalg.h"
#include "power.h"
#include "moebius.h"
#include "malloc_check.h"
#include "log10.h"

// ----------------------------------------------------------------
typedef struct _opts_t {
	int show_orders;
	int show_elements;
	int show_logs;
	int show_normal_rank;
	int show_trace;
	int show_norm;
	int terse_output;
} opts_t;

// ----------------------------------------------------------------
typedef struct _f2i_chart_cell_t {
	z2ipoly_t root;
	int       log;
} f2i_chart_cell_t;

typedef struct _f2i_chart_row_t {
	z2ipoly_t min_poly;
	int       num_roots;
	int       order;
	f2i_chart_cell_t * cells;
} f2i_chart_row_t;

typedef struct _f2i_chart_t {
	int       num_rows;
	z2ipoly_t g;
	f2i_chart_row_t * rows;
} f2i_chart_t;

// ----------------------------------------------------------------
static int num_irreds_deg_is(
	int       n);

static int num_irreds_deg_divides(
	int       n);

static void find_irreds(
	int       n,
	z2ipoly_t * pirreds,
	int       num_irreds);

static f2i_chart_t * alloc_f2i_chart(
	int       num_rows,
	int       num_cols);

static void free_f2i_chart(
	f2i_chart_t ** ppchart);

static int f2i_chart_row_qcmp(
	const void * pv1,
	const void * pv2);

static void do_f2i_chart(
	z2ipoly_t r,
	int       n,
	opts_t  * popts);

static void fill_f2i_chart(
	f2i_chart_t * pchart,
	z2ipoly_t r,
	int       n);

static void print_f2i_chart(
	f2i_chart_t * pchart,
	z2ipoly_t r,
	int       n,
	opts_t  * popts);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {n}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int        argi;
	int        n;
	z2ipoly_t  rj;
	int        j;
	opts_t     opts;

	opts.show_orders   = 0;
	opts.show_elements = 1;
	opts.show_logs     = 0;
	opts.show_normal_rank = 0;
	opts.show_trace    = 0;
	opts.show_norm     = 0;
	opts.terse_output  = 0;

	argi = 1;
	while ((argi < argc) && (argv[argi][0] == '-')) {
		if (strcmp(argv[argi], "-o") == 0) {
			opts.show_orders = 1;
		}
		else if (strcmp(argv[argi], "-l") == 0) {
			opts.show_elements = 0;
			opts.show_logs = 1;
		}
		else if (strcmp(argv[argi], "-L") == 0) {
			opts.show_elements = 1;
			opts.show_logs = 1;
		}
		else if (strcmp(argv[argi], "-nr") == 0) {
			opts.show_normal_rank = 1;
		}
		else if (strcmp(argv[argi], "-t") == 0) {
			opts.show_trace = 1;
		}
		else if (strcmp(argv[argi], "-n") == 0) {
			opts.show_norm = 1;
		}
		else if (strcmp(argv[argi], "-b") == 0) {
			opts.terse_output = 1;
		}
		else
			usage(argv[0]);
		argi++;
	}

	if ((argc - argi) < 1)
		usage(argv[0]);

	for (; argi < argc; argi++) {

		if (sscanf(argv[argi], ":%u", &n) == 1) {
			int num_irreds = num_irreds_deg_is(n);
			z2ipoly_t * pirreds = malloc_check(
				num_irreds * sizeof(z2ipoly_t));
			find_irreds(n, pirreds, num_irreds);
			for (j = 0; j < num_irreds; j++) {
				rj = pirreds[j];
				do_f2i_chart(rj, n, &opts);
			}
			free(pirreds);
		}
		else if (z2ipoly_scan(&rj, argv[argi])) {
			n = z2ipoly_get_degree(rj);
			do_f2i_chart(rj, n, &opts);
		}
		else {
			usage(argv[0]);
		}
	}

	return 0;
}

// ----------------------------------------------------------------
static int num_irreds_deg_is(
	int       n)
{
	int d, sum = 0;

	for (d = 1; d <= n; d++)
		if ((n % d) == 0)
			sum += calc_moebius_unsigned(n/d) * (1 << d);
	return sum / n;
}

// ----------------------------------------------------------------
static int num_irreds_deg_divides(
	int       n)
{
	int d, sum = 0;
	for (d = 1; d <= n; d++)
		if ((n % d) == 0)
			sum += num_irreds_deg_is(d);
	return sum;
}

// ----------------------------------------------------------------
static void find_irreds(
	int       n,
	z2ipoly_t * pirreds,
	int       num_irreds)
{
	int count = 0;
	z2ipoly_t f, flo, fhi;

	flo = 1LL << n;
	fhi = (flo << 1) - 1LL;
	for (f = flo; f <= fhi; f++) {
		if (!z2ipoly_is_irreducible(f))
			continue;

		if (count >= num_irreds) {
			fprintf(stderr, "Coding error:  too many factors.\n");
			exit(1);
		}
		pirreds[count] = f;
		count++;
	}
}

// ----------------------------------------------------------------
static f2i_chart_t * alloc_f2i_chart(
	int       num_rows,
	int       num_cols)
{
	f2i_chart_t * pchart;
	int i, j;

	pchart = malloc_check(sizeof(f2i_chart_t));
	pchart->num_rows = num_rows;
	pchart->rows = malloc_check(num_rows * sizeof(f2i_chart_row_t));
	for (i = 0; i < num_rows; i++)
		pchart->rows[i].cells = malloc_check(
			num_cols * sizeof(f2i_chart_cell_t));

	pchart->g = 0LL;
	for (i = 0; i < num_rows; i++) {
		pchart->rows[i].min_poly = 0LL;
		pchart->rows[i].num_roots = 0;
		pchart->rows[i].order = 0;
		for (j = 0; j < num_cols; j++)
			pchart->rows[i].cells[j].root = 0LL;
			pchart->rows[i].cells[j].log = 0;
	}

	return pchart;
}

// ----------------------------------------------------------------
static void free_f2i_chart(
	f2i_chart_t ** ppchart)
{
	f2i_chart_t * pchart = *ppchart;
	int i;
	for (i = 0; i < pchart->num_rows; i++)
		free(pchart->rows[i].cells);
	free(pchart->rows);
	free(pchart);
	*ppchart = 0;
}

// ----------------------------------------------------------------
static int f2i_chart_row_qcmp(
	const void * pv1,
	const void * pv2)
{
	const f2i_chart_row_t * prow1 = pv1;
	const f2i_chart_row_t * prow2 = pv2;
	return (prow1->min_poly > prow2->min_poly);
}

// ----------------------------------------------------------------
static void do_f2i_chart(
	z2ipoly_t r,
	int       n,
	opts_t  * popts)
{
	int num_rows = num_irreds_deg_divides(n);
	f2i_chart_t  * pchart = alloc_f2i_chart(num_rows, n);

	fill_f2i_chart(pchart, r, n);
	print_f2i_chart(pchart, r, n, popts);

	free_f2i_chart(&pchart);
}

// ----------------------------------------------------------------
static void fill_f2i_chart(
	f2i_chart_t * pchart,
	z2ipoly_t r,
	int       n)
{
	int pn1 = (1 << n) - 1;
	int i, d, pd1, sfld_log, elt_log, orbit_log, j;
	char * marks;
	z2ipoly_t g, elt;
	int    rowi = 0;
	f2i_chart_row_t * prow;

	g = z2ipolymod_find_generator(r);
	pchart->g = g;
	marks = malloc_check(pn1 * sizeof(char));
	for (i = 0; i < pn1; i++)
		marks[i] = 0;

	pchart->rows[rowi].min_poly = 2LL;
	pchart->rows[rowi].num_roots = 1;
	pchart->rows[rowi].order = 0;
	pchart->rows[rowi].cells[0].root = 0LL;
	pchart->rows[rowi].cells[0].log  = 0;
	rowi++;

	for (d = 1; d <= n; d++) {
		if ((n % d) != 0)
			continue;
		pd1 = (1 << d) - 1;
		sfld_log = pn1 / pd1;

		for (elt_log = 0; elt_log < pn1; elt_log += sfld_log) {
			if (marks[elt_log])
				continue;

			prow = &pchart->rows[rowi];

			orbit_log = elt_log;
			elt = z2ipolymod_power(g, orbit_log, r);

			prow->min_poly = z2ipoly_get_min_poly(elt, r);
			pchart->rows[rowi].num_roots = d;
			pchart->rows[rowi].order =
				z2ipolymod_element_order(elt, r);

			for (j = 0; j < d; j++) {
				marks[orbit_log] = 1;

				prow->cells[j].root = elt;
				prow->cells[j].log  = orbit_log;

				orbit_log *= 2;
				orbit_log %= pn1;
				elt = z2ipolymod_power(g, orbit_log, r);
			}
			rowi++;
		}
	}

	qsort(pchart->rows, pchart->num_rows, sizeof(pchart->rows[0]),
		f2i_chart_row_qcmp);

	free(marks);
}

// ----------------------------------------------------------------
static void print_f2i_chart(
	f2i_chart_t * pchart,
	z2ipoly_t r,
	int       n,
	opts_t  * popts)
{
	int i, j;
	f2i_chart_cell_t * pc;
	z2ipoly_t z;
	int width = calc_log10_unsigned(1<<n, IFLOOR) + 1;

	if (!popts->terse_output) {
		printf("Root chart for ");
		z2ipoly_print(r);
		if (popts->show_logs) {
			printf(", g = ");
			z2ipolymod_print(pchart->g, r);
		}
		printf("\n");
	}

	for (i = 0; i < pchart->num_rows; i++) {
		if (!popts->terse_output) {
			z2ipoly_dprint(pchart->rows[i].min_poly, n);
			printf(":");
		}
		if (popts->show_orders) {
			printf(" [o= %*d]", width, pchart->rows[i].order);
		}

		if (popts->show_normal_rank) {
			// xxx make get-rank function
			int nroots = pchart->rows[i].num_roots;
			int rank;
			z2ipoly_t * M = malloc_check(
				nroots * sizeof(z2ipoly_t));
			for (j = 0; j < nroots; j++)
				M[j] = pchart->rows[i].cells[j].root;
			z2ila_row_echelon_form(M, nroots);
			rank = z2ila_find_rank(M, nroots);
			printf(" [nr= %*d]", width, rank);
			free(M);
		}

		if (popts->show_trace) {
			// Sum of conjugates
			printf(" tr=");
			z2ipoly_print(z2ipolymod_trace(
				pchart->rows[i].cells[0].root, r));
		}

		if (popts->show_norm) {
			// Product of conjugates
			printf(" nm=");
			z2ipoly_print(z2ipolymod_norm(
				pchart->rows[i].cells[0].root, r));
		}

		for (j = 0; j < pchart->rows[i].num_roots; j++) {
			pc = &pchart->rows[i].cells[j];
			z  = pc->root;

			printf(" ");
			if (popts->show_elements) {
				z2ipolymod_print(z, r);
			}
			if (popts->show_logs) {
				if (popts->show_elements)
					printf("(");
				if (z == 0LL)
					printf("%*s", width, "-");
				else
					printf("%*d",
						width,
						pchart->rows[i].cells[j].log);
				if (popts->show_elements)
					printf(")");
			}
		}
		if (!popts->terse_output) {
			printf("\n");
		}
	}
	printf("\n");
}
