// This program is similar to f2nsplit.c (q.v.), except this one tabulates
// roots for *all* monic irreducibles, i.e. computes roots using all finite
// fields of a given degree (which are isomorphic, but not bodily identical).

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "power.h"
#include "z2nipoly_factor.h"

#define MAX_IRRED_FACTORS 4116 // See f2nsplit.c.

typedef struct _opts_t {
	int show_trace;
	int show_norm;
	int show_chpoly;
	int show_orders;
	int show_elements;
	int show_logs;
} opts_t;

static void find_irred_factors(
	unsigned    n,
	z2ipoly_t * pirred_factors,
	unsigned  * pnum_irred_factors,
	z2ipoly_t * pdirred_factors,
	unsigned  * pnum_dirred_factors,
	unsigned    max_irred_factors);

static void split_poly(
	z2ipoly_t    rj,
	z2ipoly_t    g,
	z2ipoly_t  * irred_factors,
	unsigned     num_irred_factors,
	z2ipoly_t  * dirred_factors,
	unsigned     num_dirred_factors,
	unsigned     n,
	z2ipoly_t    pn,
	opts_t     * popts);

static void split_factor(
	z2nipoly_t * pf,
	z2ipoly_t    r,
	z2ipoly_t    pn,
	int          n,
	z2ipoly_t    g,
	opts_t     * popts);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {n}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int        argi;
	z2ipoly_t  pn;
	z2ipoly_t  irred_factors[MAX_IRRED_FACTORS];
	z2ipoly_t  dirred_factors[MAX_IRRED_FACTORS];
	z2ipoly_t  rj;
	z2ipoly_t  g = 0LL;
	unsigned   num_irred_factors = 0;
	unsigned   num_dirred_factors = 0;
	unsigned   n, j;
	opts_t     opts;

	opts.show_trace  = 0;
	opts.show_norm   = 0;
	opts.show_chpoly = 0;
	opts.show_orders = 0;
	opts.show_elements = 1;
	opts.show_logs   = 0;

	if (argc < 2)
		usage(argv[0]);

	argi = 1;
	while (argv[argi][0] == '-') {
		if (strcmp(argv[argi], "-t") == 0)
			opts.show_trace = 1;
		else if (strcmp(argv[argi], "-n") == 0)
			opts.show_norm = 1;
		else if (strcmp(argv[argi], "-c") == 0)
			opts.show_chpoly = 1;
		else if (strcmp(argv[argi], "-o") == 0)
			opts.show_orders = 1;
		else if (strcmp(argv[argi], "-l") == 0) {
			opts.show_elements = 0;
			opts.show_logs = 1;
		}
		else if (strcmp(argv[argi], "-L") == 0) {
			opts.show_elements = 1;
			opts.show_logs = 1;
		}
		else
			usage(argv[0]);
		argi++;
	}

	for (; argi < argc; argi++) {
		if (sscanf(argv[argi], ":%u", &n) == 1) {
			pn = 1LL << n;
			find_irred_factors(n, irred_factors, &num_irred_factors,
				dirred_factors, &num_dirred_factors, MAX_IRRED_FACTORS);
			for (j = 0; j < num_dirred_factors; j++) {
				rj = dirred_factors[j];
				g = z2ipolymod_find_generator(rj);
				split_poly(rj, g, irred_factors, num_irred_factors,
					dirred_factors, num_dirred_factors, n, pn, &opts);
			}
			printf("\n");
		}
		else if (z2ipoly_scan(&rj, argv[argi])) {
			z2ipoly_t dummy;
			if (sscanf(argv[argi], "%llx:%llx", &dummy, &g) != 2)
				g = z2ipolymod_find_generator(rj);
			n = z2ipoly_get_degree(rj);
			pn = 1LL << n;
			find_irred_factors(n, irred_factors, &num_irred_factors,
				dirred_factors, &num_dirred_factors, MAX_IRRED_FACTORS);
			split_poly(rj, g, irred_factors, num_irred_factors,
				dirred_factors, num_dirred_factors, n, pn, &opts);
		}
		else {
			usage(argv[0]);
		}
	}

	return 0;
}

// ----------------------------------------------------------------
static void find_irred_factors(
	unsigned    n,
	z2ipoly_t * pirred_factors,
	unsigned  * pnum_irred_factors,
	z2ipoly_t * pdirred_factors,
	unsigned  * pnum_dirred_factors,
	unsigned    max_irred_factors)
{
	unsigned num_irred_factors = 0;
	unsigned num_dirred_factors = 0;
	unsigned d;
	z2ipoly_t lo, hi, f;

	for (d = 1; d <= n; d++) {
		if ((n % d) != 0)
			continue;

		lo = 1LL << d;
		hi = lo << 1;
		for (f = lo; f < hi; f++) {
			if (!z2ipoly_is_irreducible(f))
				continue;

			if (num_irred_factors >= max_irred_factors) {
				fprintf(stderr, "Too many factors.\n");
				exit(1);
			}
			pirred_factors[num_irred_factors] = f;
			num_irred_factors++;

			if (d == n) {
				pdirred_factors[num_dirred_factors] = f;
				num_dirred_factors++;
			}
		}
	}

	*pnum_irred_factors = num_irred_factors;
	*pnum_dirred_factors = num_dirred_factors;
}

// ----------------------------------------------------------------
static void split_poly(
	z2ipoly_t    rj,
	z2ipoly_t    g,
	z2ipoly_t  * irred_factors,
	unsigned     num_irred_factors,
	z2ipoly_t  * dirred_factors,
	unsigned     num_dirred_factors,
	unsigned     n,
	z2ipoly_t    pn,
	opts_t     * popts)
{
	z2ipoly_t  ri;
	z2nipoly_t ext_factor;
	unsigned   i;

	printf("mod ");
	//printf("%c ", z2ipoly_is_primitive(rj) ? 'o' : '-');
	z2ipoly_print(rj);

	if (popts->show_logs) {
		printf(" g=%llx", g);
	}
	printf(":\n");

	for (i = 0; i < num_irred_factors; i++) {
		ri = irred_factors[i];

		ext_factor = z2nipoly_extend_z2ipoly(ri);

		printf("%c ", z2ipoly_is_primitive(ri) ? 'o' : '-');
		z2ipolymod_print(ri, rj<<1);
		printf(" = ");

		split_factor(&ext_factor, rj, pn, n, g, popts);
		printf("\n");
		fflush(stdout);
	}
	printf("\n");
}

//// ----------------------------------------------------------------
//// This uses Berlekamp's algorithm on F2n[x].
//static void split_factor(
//	z2nipoly_t * pf,
//	z2ipoly_t    r,
//	z2ipoly_t    pn,
//	int          n,
//	z2ipoly_t    g,
//	opts_t     * popts)
//{
//	z2nipoly_t linear_factor, remnant, quot, rem;
//	z2ipoly_t i;
//	int root_order = 0;
//	int nfactors = 0;
//
//	// Try each possible linear factor (X - a) for a in F_2^n.
//
//	remnant = *pf;
//	linear_factor.degree = 1;
//	linear_factor.coeffs[1] = 1LL;
//	for (i = 0; i < pn; i++) {
//		linear_factor.coeffs[0] = i;
//
//		z2nipoly_quot_and_rem(&remnant, &linear_factor,
//			&quot, &rem, r);
//
//		// Try the next factor.
//		if (!z2nipoly_is_zero(&rem))
//			continue;
//
//		nfactors++;
//		printf(" ");
//
//		if (popts->show_elements) {
//			z2ipolymod_print(i, r);
//			if (popts->show_logs)
//				printf("(%2u)", z2ipolymod_log(i, g, r));
//		}
//		else {
//			if (popts->show_logs)
//				printf(" %2u", z2ipolymod_log(i, g, r));
//		}
//
//		root_order = z2ipolymod_element_order(i, r);
//
//		remnant = quot;
//
//		if (z2nipoly_is_one(&remnant)) {
//			break;
//		}
//	}
//	for (i = 0; i < (n - nfactors); i++)
//		printf(" .");
//	printf(" o= %3d", root_order);
//
//	if (!z2nipoly_is_one(&remnant)) {
//		printf("ERROR:  remnant ");
//			z2nipoly_print(&remnant, r);
//			printf("\n");
//	}
//}

// ----------------------------------------------------------------
static void split_factor(
	z2nipoly_t * pf,
	z2ipoly_t    r,
	z2ipoly_t    pn,
	int          n,
	z2ipoly_t    g,
	opts_t     * popts)
{
	z2ipoly_t root;
	static z2nipoly_factor_info_t finfo;
	int i;
	int root_order = 0;

	z2nipoly_factor(pf, r, &finfo);

	for (i = 0; i < finfo.num_distinct; i++) {
		if (finfo.factor_counts[i].count != 1) {
			fprintf(stderr,
			"fpnsplits: multiple factors in universal poly.\n");
			exit(1);
		}
		if (finfo.factor_counts[i].f.degree != 1) {
			fprintf(stderr,
			"fpnsplits: non-linear factors in universal poly.\n");
			exit(1);
		}
		// F2n is self-negating.
		root = finfo.factor_counts[i].f.coeffs[0];
		printf(" ");

		if (popts->show_elements) {
			z2ipolymod_print(root, r);
		}

		if (popts->show_logs) {
			if (popts->show_elements)
				printf("(");
			else
				printf(" ");
			if (root == 0LL)
				printf("--");
			else
				printf("%2u", z2ipolymod_log(root, g, r));
			if (popts->show_elements)
				printf(")");
		}

		if (i == 0)
			root_order = z2ipolymod_element_order(root, r);
	}

	for (i = 0; i < (n - finfo.num_distinct); i++)
		printf(" .");

	if (popts->show_trace) {
		// Sum of conjugates
		z2ipoly_t trace = 0LL;
		for (i = 0; i < finfo.num_distinct; i++) {
			root = finfo.factor_counts[i].f.coeffs[0];
			trace = z2ipoly_add(trace, root);
		}

		printf(" tr=");
		z2ipoly_print(trace);
	}
	if (popts->show_norm) {
		// Product of conjugates
		z2ipoly_t norm = 1LL;
		for (i = 0; i < finfo.num_distinct; i++) {
			root = finfo.factor_counts[i].f.coeffs[0];
			norm = z2ipolymod_mul(norm, root, r);
		}

		printf(" nm=");
		z2ipoly_print(norm);
	}
	if (popts->show_chpoly) {
		// Characteristic polynomial not implemented yet.
		printf(" ch=?");
	}

	if (popts->show_orders)
		printf(" o= %3d", root_order);
}
