// This program is similar to f2nsplit.c (q.v.), except this one tabulates
// roots for *all* monic irreducibles, i.e. computes roots using all finite
// fields of a given degree (which are isomorphic, but not bodily identical).

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "power.h"
#include "z2npoly_factor.h"

#define MAX_IRRED_FACTORS 4116 // See f2nsplit.c.

static void find_irred_factors(
	unsigned    n,
	z2poly_t  * pirred_factors,
	unsigned  * pnum_irred_factors,
	z2poly_t  * pdirred_factors,
	unsigned  * pnum_dirred_factors,
	unsigned    max_irred_factors);

static void split_poly(
	z2poly_t  * prj,
	z2poly_t  * irred_factors,
	unsigned    num_irred_factors,
	z2poly_t  * dirred_factors,
	unsigned    num_dirred_factors,
	unsigned    n,
	int         show_elements,
	int         show_logs);

static void split_factor(
	z2npoly_t * pf,
	z2poly_t  * pr,
	int         n,
	z2poly_t  * pg,
	int         show_elements,
	int         show_logs);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {n}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int       argi;
	z2poly_t  irred_factors[MAX_IRRED_FACTORS];
	z2poly_t  dirred_factors[MAX_IRRED_FACTORS];
	z2poly_t  rj;
	unsigned  num_irred_factors = 0;
	unsigned  num_dirred_factors = 0;
	unsigned  n, j;
	int       show_elements = 1;
	int       show_logs = 0;

	if (argc < 2)
		usage(argv[0]);

	for (argi = 1; argi < argc; argi++) {
		if (strcmp(argv[argi], "-l") == 0) {
			show_elements = 0;
			show_logs = 1;
			continue;
		}
		if (strcmp(argv[argi], "-L") == 0) {
			show_elements = 1;
			show_logs = 1;
			continue;
		}

		if (sscanf(argv[argi], ":%u", &n) == 1) {
			find_irred_factors(n, irred_factors, &num_irred_factors,
				dirred_factors, &num_dirred_factors, MAX_IRRED_FACTORS);
			for (j = 0; j < num_dirred_factors; j++) {
				rj = dirred_factors[j];
				split_poly(&rj, irred_factors, num_irred_factors,
					dirred_factors, num_dirred_factors, n, show_elements,
					show_logs);
			}
		}
		else if (z2poly_scan(&rj, argv[argi])) {
			n = z2poly_get_degree(&rj);
			find_irred_factors(n, irred_factors, &num_irred_factors,
				dirred_factors, &num_dirred_factors, MAX_IRRED_FACTORS);
			split_poly(&rj, irred_factors, num_irred_factors,
				dirred_factors, num_dirred_factors, n, show_elements,
				show_logs);
		}
		else {
			usage(argv[0]);
		}
	}

	return 0;
}

// ----------------------------------------------------------------
static void find_irred_factors(
	unsigned   n,
	z2poly_t * pirred_factors,
	unsigned * pnum_irred_factors,
	z2poly_t * pdirred_factors,
	unsigned * pnum_dirred_factors,
	unsigned   max_irred_factors)
{
	unsigned num_irred_factors = 0;
	unsigned num_dirred_factors = 0;
	unsigned d;
	unsigned long long lo, hi, f;
	z2poly_t z;

	for (d = 1; d <= n; d++) {
		if ((n % d) != 0)
			continue;

		lo = 1LL << d;
		hi = lo << 1;
		for (f = lo; f < hi; f++) {
			z = z2poly_from_base_rep_64(f);
			if (!z2poly_is_irreducible(&z))
				continue;

			if (num_irred_factors >= max_irred_factors) {
				fprintf(stderr, "Too many factors.\n");
				exit(1);
			}
			pirred_factors[num_irred_factors] = z;
			num_irred_factors++;

			if (d == n) {
				pdirred_factors[num_dirred_factors] = z;
				num_dirred_factors++;
			}
		}
	}

	*pnum_irred_factors = num_irred_factors;
	*pnum_dirred_factors = num_dirred_factors;
}

// ----------------------------------------------------------------
static void split_poly(
	z2poly_t  * prj,
	z2poly_t  * irred_factors,
	unsigned    num_irred_factors,
	z2poly_t  * dirred_factors,
	unsigned    num_dirred_factors,
	unsigned    n,
	int         show_elements,
	int         show_logs)
{
	z2poly_t  ri;
	z2poly_t  g = z2poly_from_base_rep(0);
	z2npoly_t ext_factor;
	unsigned  i;
	z2poly_t  print_junk;

	printf("mod ");
	//printf("%c ", z2poly_is_primitive(prj) ? 'o' : '-');
	z2poly_print(prj);

	if (show_logs) {
		g = z2polymod_find_generator(prj);
		printf(" g=");
		z2poly_print(&g);
	}
	printf(":\n");

	for (i = 0; i < num_irred_factors; i++) {
		ri = irred_factors[i];

		ext_factor = z2npoly_extend_z2poly(&ri);

		printf("%c ", z2poly_is_primitive(&ri) ? 'o' : '-');
		print_junk = *prj;
		z2poly_promote_1(&print_junk);
		z2polymod_print(&ri, &print_junk);
		printf(" = ");

		split_factor(&ext_factor, prj, n, &g, show_elements, show_logs);
		printf("\n");
		fflush(stdout);
	}
	printf("\n");
}

// ----------------------------------------------------------------
static void split_factor(
	z2npoly_t * pf,
	z2poly_t  * pr,
	int         n,
	z2poly_t  * pg,
	int         show_elements,
	int         show_logs)
{
	z2poly_t root;
	static z2npoly_factor_info_t finfo;
	int i;
	int root_order = 0;

	z2npoly_factor(pf, pr, &finfo);

	for (i = 0; i < finfo.num_distinct; i++) {
		if (finfo.factor_counts[i].count != 1) {
			fprintf(stderr,
			"fpnsplits: multiple factors in universal poly.\n");
			exit(1);
		}
		if (finfo.factor_counts[i].f.degree != 1) {
			fprintf(stderr,
			"fpnsplits: non-linear factors in universal poly.\n");
			exit(1);
		}
		// F2n is self-negating.
		root = finfo.factor_counts[i].f.coeffs[0];
		printf(" ");

		if (show_elements) {
			z2polymod_print(&root, pr);
		}

		if (show_logs) {
			if (show_elements)
				printf("(");
			else
				printf(" ");
			if (z2poly_is_zero(&root))
				printf("--");
			else
				printf("%2u", z2polymod_log(&root, pg, pr));
			if (show_elements)
				printf(")");
		}

		if (i == 0)
			root_order = z2polymod_element_order(&root, pr);
	}

	for (i = 0; i < (n - finfo.num_distinct); i++)
		printf(" .");

	printf(" [%3d]", root_order);
}
