#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "power.h"
#include "z3npoly.h"

#define MAX_IRRED_FACTORS 4116 // See f2nsplit.c.

typedef struct _opts_t {
	int show_trace;
	int show_norm;
	int show_chpoly;
	int show_orders;
	int show_elements;
	int show_logs;
} opts_t;

static void find_irred_factors(
	unsigned    n,
	z3poly_t  * pirred_factors,
	unsigned  * pnum_irred_factors,
	z3poly_t  * pdirred_factors,
	unsigned  * pnum_dirred_factors,
	unsigned    max_irred_factors);

static void split_poly(
	z3poly_t  * prj,
	z3poly_t  * irred_factors,
	unsigned    num_irred_factors,
	z3poly_t  * dirred_factors,
	unsigned    num_dirred_factors,
	int         n,
	int         pn,
	opts_t    * popts);

static void split_factor(
	z3npoly_t * pf,
	z3poly_t  * pr,
	int         n,
	z3poly_t  * pg,
	opts_t    * popts);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {n}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int       argi;
	static z3poly_t irred_factors[MAX_IRRED_FACTORS]; // xxx malloc
	static z3poly_t dirred_factors[MAX_IRRED_FACTORS];
	int        n, pn;
	unsigned   num_irred_factors = 0;
	unsigned   num_dirred_factors = 0;
	z3poly_t   rj;
	unsigned   j;
	opts_t     opts;

	opts.show_trace  = 0;
	opts.show_norm   = 0;
	opts.show_chpoly = 0;
	opts.show_orders = 0;
	opts.show_elements = 1;
	opts.show_logs   = 0;

	if (argc < 2)
		usage(argv[0]);

	argi = 1;
	while (argv[argi][0] == '-') {
		if (strcmp(argv[argi], "-t") == 0)
			opts.show_trace = 1;
		else if (strcmp(argv[argi], "-n") == 0)
			opts.show_norm = 1;
		else if (strcmp(argv[argi], "-c") == 0)
			opts.show_chpoly = 1;
		else if (strcmp(argv[argi], "-o") == 0)
			opts.show_orders = 1;
		else if (strcmp(argv[argi], "-l") == 0) {
			opts.show_elements = 0;
			opts.show_logs = 1;
		}
		else if (strcmp(argv[argi], "-L") == 0) {
			opts.show_elements = 1;
			opts.show_logs = 1;
		}
		else
			usage(argv[0]);
		argi++;
	}

	for (; argi < argc; argi++) {
		if (sscanf(argv[argi], ":%u", &n) == 1) {
			pn = calc_power(3, n);
			find_irred_factors(n, irred_factors, &num_irred_factors,
				dirred_factors, &num_dirred_factors, MAX_IRRED_FACTORS);
			for (j = 0; j < num_dirred_factors; j++) {
				rj = dirred_factors[j];
				split_poly(&rj, irred_factors, num_irred_factors,
					dirred_factors, num_dirred_factors, n, pn, &opts);
			}
		}
		else if (z3poly_scan(&rj, argv[argi])) {
			n = z3poly_get_degree(&rj);
			pn = calc_power(3, n);
			find_irred_factors(n, irred_factors, &num_irred_factors,
				dirred_factors, &num_dirred_factors, MAX_IRRED_FACTORS);
			split_poly(&rj, irred_factors, num_irred_factors,
				dirred_factors, num_dirred_factors, n, pn, &opts);
		}
		else {
			usage(argv[0]);
		}
	}

	return 0;
}

// ----------------------------------------------------------------
static void find_irred_factors(
	unsigned    n,
	z3poly_t  * pirred_factors,
	unsigned  * pnum_irred_factors,
	z3poly_t  * pdirred_factors,
	unsigned  * pnum_dirred_factors,
	unsigned    max_irred_factors)
{
	unsigned num_irred_factors = 0;
	unsigned num_dirred_factors = 0;
	unsigned d, ilo, ihi, i;
	z3poly_t f;

	for (d = 1; d <= n; d++) {
		if ((n % d) != 0)
			continue;

		ilo = calc_power(3, d);
		ihi = 2 * ilo - 1;
		for (i = ilo; i <= ihi; i++) {
			f = z3poly_from_base_rep(i);
			if (!z3poly_is_irreducible(&f))
				continue;

			if (num_irred_factors >= MAX_IRRED_FACTORS) {
				fprintf(stderr, "Too many factors.\n");
				exit(1);
			}
			pirred_factors[num_irred_factors] = f;
			num_irred_factors++;

			if (d == n) {
				pdirred_factors[num_dirred_factors] = f;
				num_dirred_factors++;
			}
		}
	}

	*pnum_irred_factors = num_irred_factors;
	*pnum_dirred_factors = num_dirred_factors;
}

// ----------------------------------------------------------------
static void split_poly(
	z3poly_t  * prj,
	z3poly_t  * irred_factors,
	unsigned    num_irred_factors,
	z3poly_t  * dirred_factors,
	unsigned    num_dirred_factors,
	int         n,
	int         pn,
	opts_t    * popts)
{
	z3poly_t   ri;
	z3poly_t   g = z3poly_from_base_rep(0);
	z3npoly_t  ext_factor;
	unsigned   i;

	printf("mod ");
	//printf("%c ", z3poly_is_primitive(prj, n) ? 'o' : '-');
	z3poly_print(prj);

	if (popts->show_logs) {
		g = z3polymod_find_generator(prj);
		printf(" g=");
		z3poly_print(&g);
	}
	printf(":\n");

	for (i = 0; i < num_irred_factors; i++) {
		ri = irred_factors[i];

		ext_factor = z3npoly_extend_z3poly(&ri);

		printf("%c ", z3poly_is_primitive(&ri, n) ? 'o' : '-');
		z3poly_dprint(&ri, n+1);

		split_factor(&ext_factor, prj, n, &g, popts);
		printf("\n");
		fflush(stdout);
	}
	printf("\n");
}

// ----------------------------------------------------------------
static void split_factor(
	z3npoly_t * pf,
	z3poly_t  * pr,
	int         n,
	z3poly_t  * pg,
	opts_t    * popts)
{
	int pn = calc_power(3, n);
	int i;
	z3npoly_t linear_factor, remnant, quot, rem;
	z3poly_t z;
	int root_order = 0;
	int nfactors = 0;

	// Try each possible linear factor (X - a) for a in F_2^n.
	remnant = *pf;
	linear_factor.degree = 1;
	linear_factor.coeffs[1] = z3poly_from_base_rep(1);
	for (i = 0; i < pn; i++) {
		z = z3poly_from_base_rep(i);
		linear_factor.coeffs[0] = z;

		z3npoly_quot_and_rem(&remnant, &linear_factor, &quot, &rem, pr);

		// Try the next factor.
		if (!z3npoly_is_zero(&rem))
			continue;

		nfactors++;
		printf(" ");

		if (popts->show_elements) {
			z3polymod_print(&z, pr);
		}

		if (popts->show_logs) {
			if (popts->show_elements)
				printf("(");
			else
				printf(" ");
			if (z3poly_is_zero(&z))
				printf("--");
			else
				printf("%2u", z3polymod_log(&z, pg, pr));
			if (popts->show_elements)
				printf(")");
		}

		root_order = z3polymod_element_order(&z, pr);

		remnant = quot;

		if (z3npoly_is_one(&remnant)) {
			break;
		}
	}
	for (i = 0; i < (n - nfactors); i++)
		printf(" .");

	if (popts->show_trace) {
		// Sum of conjugates
		printf(" tr=?");
	}
	if (popts->show_norm) {
		// Product of conjugates
		printf(" nm=?");
	}
	if (popts->show_chpoly) {
		// Characteristic polynomial not implemented yet.
		printf(" ch=?");
	}


	if (popts->show_orders)
		printf(" o= %3d", root_order);

	if (!z3npoly_is_one(&remnant)) {
		printf("ERROR:  remnant ");
			z3npoly_print(&remnant, pr);
			printf("\n");
	}
}
