#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#include "z2ipoly_factor.h"
#include "z2poly_factor.h"
#include "zppoly_factor.h"
#include "z3poly_factor.h"

#define DO_Z2I_SEQ
#define DO_Z2I_RAND
#define DO_Z2_SEQ
#define DO_Z2_RAND
#define DO_Z3_SEQ
#define DO_Z3_RAND
#define DO_ZP_SEQ
#define DO_ZP_RAND

#if (defined DO_ZP_SEQ) || (defined DO_ZP_RAND)
static int primes[] = { 2, 3, 5, 7, 11, 101 };
static int num_primes = sizeof(primes) / sizeof(primes[0]);
#endif

// ----------------------------------------------------------------
#define STEP_RAND(x) ((x) = 1664525 * (x) + 1013904223)

// ----------------------------------------------------------------
z2ipoly_t z2ipoly_unfactor(
	z2ipoly_factor_info_t * pfinfo)
{
	z2ipoly_t rv = 1LL;
	z2ipoly_t temp;
	z2ipoly_t f;
	int c;
	int i, j;

	for (i = 0; i < pfinfo->num_distinct; i++) {
		temp = 1LL;
		c = pfinfo->factor_counts[i].count;
		f = pfinfo->factor_counts[i].f;
		if (!z2ipoly_is_irreducible(f)) {
			printf("z2ipoly factorization failure: ");
			z2ipoly_print(f);
			printf(" is not irreducible.\n");
			exit(1);
		}
		for (j = 0; j < c; j++)
			temp = z2ipoly_mul(temp, f);
		rv = z2ipoly_mul(rv, temp);
	}

	return rv;
}

// ----------------------------------------------------------------
z2poly_t z2poly_unfactor(
	z2poly_factor_info_t * pfinfo)
{
	z2poly_t rv = z2poly_from_base_rep(1);
	z2poly_t temp;
	z2poly_t f;
	int c;
	int i;

	for (i = 0; i < pfinfo->num_distinct; i++) {
		c = pfinfo->factor_counts[i].count;
		f = pfinfo->factor_counts[i].f;
		if (!z2poly_is_irreducible(&f)) {
			printf("z2poly factorization failure: ");
			z2poly_print(&f);
			printf(" is not irreducible.\n");
			exit(1);
		}
		temp = z2poly_power(&f, c);
		rv = z2poly_mul(&rv, &temp);
	}

	return rv;
}

// ----------------------------------------------------------------
z3poly_t z3poly_unfactor(
	z3poly_factor_info_t * pfinfo)
{
	z3poly_t rv = z3poly_from_base_rep(1);
	z3poly_t temp;
	z3poly_t f;
	int c;
	int i;


	for (i = 0; i < pfinfo->num_distinct; i++) {
		c = pfinfo->factor_counts[i].count;
		f = pfinfo->factor_counts[i].f;
		if (!z3poly_is_irreducible(&f)) {
			printf("z3poly factorization failure: ");
			z3poly_print(&f);
			printf(" is not irreducible.\n");
			exit(1);
		}
		temp = z3poly_power(&f, c);
		rv = z3poly_mul(&rv, &temp);
	}

	return rv;
}

// ----------------------------------------------------------------
zpoly_t zppoly_unfactor(
	zppoly_factor_info_t * pfinfo,
	int p)
{
	zpoly_t rv = zppoly_from_base_rep(p, 1);
	zpoly_t temp;
	zpoly_t f;
	int c;
	int i;

	for (i = 0; i < pfinfo->num_distinct; i++) {
		c = pfinfo->factor_counts[i].count;
		f = pfinfo->factor_counts[i].f;
		if (!zppoly_is_irreducible(&f, p)) {
			printf("zppoly factorization failure: ");
			zppoly_print(&f, p);
			printf(" is not irreducible.\n");
			exit(1);
		}
		temp = zppoly_power(&f, c, p);
		rv = zppoly_mul(&rv, &temp, p);
	}

	return rv;
}

// ----------------------------------------------------------------
int test_z2ipoly_one(
	z2ipoly_t z,
	int verbose)
{
	z2ipoly_t zcheck;
	z2ipoly_factor_info_t finfo;

	if (verbose) {
		printf("%llx ", z);
		fflush(stdout);
	}

	z2ipoly_factor(z, &finfo);
	zcheck = z2ipoly_unfactor(&finfo);

	if (verbose) {
		printf("%llx %s\n",
			zcheck, z == zcheck ? "OK" : "ERROR!");
	}
	if (z == zcheck)
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
int test_z2ipoly_seq(
	int max_base_rep,
	int verbose)
{
	z2ipoly_t z;
	int rv = 1;

	for (z = 0LL; z < (z2ipoly_t)max_base_rep; z++) {
		if (!test_z2ipoly_one(z, verbose))
			rv = 0;
	}

	return rv;
}

// ----------------------------------------------------------------
int test_z2ipoly_rand(
	int niter,
	int verbose)
{
	int i, deg;
	unsigned rand;
	z2ipoly_t z;
	int rv = 1;

	rand = time(0) ^ getpid();
	STEP_RAND(rand);

	for (i = 0; i < niter; i++) {
		deg = rand % Z2IPOLY_MAX_DEGREE;
		STEP_RAND(rand);

		z = (z2ipoly_t)rand << 32;
		STEP_RAND(rand);
		z |= rand;
		STEP_RAND(rand);

		z >>= (Z2IPOLY_MAX_DEGREE - deg);

		//printf("deg %2u z=%016llx\n", deg, z);
		if (!test_z2ipoly_one(z, verbose))
			rv = 0;
	}

	return rv;
}

// ----------------------------------------------------------------
int test_z2poly_one(
	z2poly_t * pz,
	int verbose)
{
	z2poly_t zcheck;
	z2poly_factor_info_t finfo;

	if (verbose) {
		z2poly_print(pz);
		printf(" ");
		fflush(stdout);
	}

	z2poly_factor(pz, &finfo);
	zcheck = z2poly_unfactor(&finfo);

	if (verbose) {
		z2poly_print(&zcheck);
		printf(" %s\n",
			z2poly_equal(pz, &zcheck) ? "OK" : "ERROR!");
	}
	if (z2poly_equal(pz, &zcheck))
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
int test_z2poly_seq(
	int max_base_rep,
	int verbose)
{
	z2poly_t z;
	int zi, rv = 1;

	for (zi = 0; zi < max_base_rep; zi++) {
		z = z2poly_from_base_rep(zi);
		if (!test_z2poly_one(&z, verbose))
			rv = 0;
	}

	return rv;
}

// ----------------------------------------------------------------
int test_z2poly_rand(
	int niter,
	int verbose)
{
	int i, j, deg;
	unsigned rand;
	z2poly_t z;
	int rv = 1;

	rand = time(0) ^ getpid();
	STEP_RAND(rand);

	for (i = 0; i < niter; i++) {
		//deg = rand % Z2POLY_MAX_DEGREE;
		//deg = (rand >> 12) % (Z2POLY_MAX_DEGREE / 2);
		deg = (rand >> 12) % 30;
		STEP_RAND(rand);

		for (j = 0; j < Z2POLY_NUM_PARTS; j++) {
			z.parts[j] = rand;
			STEP_RAND(rand);
		}

		// xxx need a demote_n
		for (j = 0; j < (Z2POLY_MAX_DEGREE - deg); j++)
			z2poly_demote_1(&z);

		if (!test_z2poly_one(&z, verbose))
			rv = 0;
	}

	return rv;
}

// ----------------------------------------------------------------
int test_z3poly_one(
	z3poly_t * pz,
	int verbose)
{
	z3poly_t zcheck;
	z3poly_factor_info_t finfo;

	if (verbose) {
		z3poly_print(pz);
		printf(" ");
		fflush(stdout);
	}

	z3poly_factor(pz, &finfo);
	zcheck = z3poly_unfactor(&finfo);

	if (verbose) {
		z3poly_print(&zcheck);
		printf(" %s\n",
			z3poly_equal(pz, &zcheck) ? "OK" : "ERROR!");
	}
	if (z3poly_equal(pz, &zcheck))
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
int test_z3poly_seq(
	int max_base_rep,
	int verbose)
{
	z3poly_t z;
	int zi, rv = 1;

	for (zi = 0; zi < max_base_rep; zi++) {
		z = z3poly_from_base_rep(zi);
		if (!test_z3poly_one(&z, verbose))
			rv = 0;
	}

	return rv;
}

// ----------------------------------------------------------------
int test_z3poly_rand(
	int niter,
	int verbose)
{
	int i, j, deg;
	unsigned rand;
	z3poly_t z;
	int rv = 1;

	rand = time(0) ^ getpid();
	STEP_RAND(rand);

	for (i = 0; i < niter; i++) {
		deg = rand % (Z3POLY_MAX_DEGREE / 2);
		STEP_RAND(rand);

		z = z3poly_from_base_rep(0);
		for (j = 0; j <= deg; j++) {
			z.coeffs[j] = rand % 3;
			STEP_RAND(rand);
		}

		if (!test_z3poly_one(&z, verbose))
			rv = 0;
	}

	return rv;
}

// ----------------------------------------------------------------
int test_zppoly_one(
	zpoly_t * pz,
	int p,
	int verbose)
{
	zpoly_t zcheck;
	zppoly_factor_info_t finfo;

	if (verbose) {
		zppoly_print(pz, p);
		printf(" ");
		fflush(stdout);
	}

	zppoly_factor(pz, p, &finfo);
	zcheck = zppoly_unfactor(&finfo, p);

	if (verbose) {
		zppoly_print(&zcheck, p);
		printf(" %s\n",
			zppoly_equal(pz, &zcheck, p) ? "OK" : "ERROR!");
	}
	if (zppoly_equal(pz, &zcheck, p))
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
int test_zppoly_seq(
	int p,
	int max_base_rep,
	int verbose)
{
	zpoly_t z;
	int zi, rv = 1;

	for (zi = 0; zi < max_base_rep; zi++) {
		z = zppoly_from_base_rep(p, zi);
		if (!test_zppoly_one(&z, p, verbose))
			rv = 0;
	}

	return rv;
}

// ----------------------------------------------------------------
int test_zppoly_rand(
	int p,
	int niter,
	int verbose)
{
	int i, j, deg;
	unsigned rand;
	zpoly_t z;
	int rv = 1;

	rand = time(0) ^ getpid();
	STEP_RAND(rand);

	for (i = 0; i < niter; i++) {
		deg = rand % (ZPOLY_MAX_DEGREE / 2);
		//deg = rand % 30;
		STEP_RAND(rand);

		z = zppoly_from_base_rep(p, 0);
		for (j = 0; j <= deg; j++) {
			z.coeffs[j] = (rand >> 16) % p;
			STEP_RAND(rand);
		}
		z.degree = zppoly_find_degree(&z, deg);

		if (!test_zppoly_one(&z, p, verbose))
			rv = 0;
	}

	return rv;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	//int max_base_rep = 1000;
	int max_base_rep = 500;
	int verbose = 0;
	int rc = 0;
	int i = 0, p = 0;

	if (argc >= 2)
		(void)sscanf(argv[1], "%d", &max_base_rep);
	if (argc >= 3)
		(void)sscanf(argv[1], "%d", &verbose);

	p = p;
	i = i;
#ifdef DO_Z2I_SEQ
	printf("\n");
	printf("z2ipoly sequential test:\n");
	if (test_z2ipoly_seq(max_base_rep, verbose)) {
		printf("z2ipoly sequential test passed.\n");
	}
	else {
		printf("z2ipoly sequential test FAILED.\n");
		rc = 1;
	}
#endif

#ifdef DO_Z2I_RAND
	printf("\n");
	printf("z2ipoly random test:\n");
	if (test_z2ipoly_rand(max_base_rep, verbose)) {
		printf("z2ipoly random test passed.\n");
	}
	else {
		printf("z2ipoly random test FAILED.\n");
		rc = 1;
	}
#endif

#ifdef DO_Z2_SEQ
	printf("\n");
	printf("z2poly sequential test:\n");
	if (test_z2poly_seq(max_base_rep, verbose)) {
		printf("z2poly sequential test passed.\n");
	}
	else {
		printf("z2poly sequential test FAILED.\n");
		rc = 1;
	}
#endif

#ifdef DO_Z2_RAND
	printf("\n");
	printf("z2poly random test:\n");
	if (test_z2poly_rand(max_base_rep, verbose)) {
		printf("z2poly random test passed.\n");
	}
	else {
		printf("z2poly random test FAILED.\n");
		rc = 1;
	}
#endif

#ifdef DO_Z3_SEQ
	printf("\n");
	printf("z3poly sequential test:\n");
	if (test_z3poly_seq(max_base_rep, verbose)) {
		printf("z3poly sequential test passed.\n");
	}
	else {
		printf("z3poly sequential test FAILED.\n");
		rc = 1;
	}
#endif

#ifdef DO_Z3_RAND
	printf("\n");
	printf("z3poly random test:\n");
	if (test_z3poly_rand(max_base_rep, verbose)) {
		printf("z3poly random test passed.\n");
	}
	else {
		printf("z3poly random test FAILED.\n");
		rc = 1;
	}
#endif

#ifdef DO_ZP_SEQ
	for (i = 0; i < num_primes; i++) {
		p = primes[i];
		printf("\n");
		printf("zppoly sequential test, p=%d:\n", p);
		if (test_zppoly_seq(p, max_base_rep, verbose)) {
			printf("zppoly sequential test passed, p=%d.\n", p);
		}
		else {
			printf("zppoly sequential test FAILED, p=%d.\n", p);
			rc = 1;
		}
	}
#endif

#ifdef DO_ZP_RAND
	for (i = 0; i < num_primes; i++) {
		p = primes[i];
		printf("\n");
		printf("zppoly random test, p=%d:\n", p);
		if (test_zppoly_rand(p, max_base_rep, verbose)) {
			printf("zppoly random test passed, p=%d.\n", p);
		}
		else {
			printf("zppoly random test FAILED, p=%d.\n", p);
			rc = 1;
		}
	}
#endif

	printf("\n");
	if (rc == 0)
		printf("All tests passed.\n");
	else
		printf("Not all tests passed.\n");

	return rc;
}
