#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zppolymod.h"
#include "power.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {r1} {r2}, of same degree\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void print_phi(
	zpoly_t * phi,
	int       p,
	int       pn,
	zppoly_aux_t * paux)
{
	int i;
	zpoly_t z;

	for (i = 0; i < pn; i++) {
		z = zppoly_from_base_rep(p, i);
		printf("  ");
		zppolymod_print(&z, paux);
		printf(" |-> ");
		zppolymod_print(&phi[i], paux);
		printf("\n");
	}
}

// ----------------------------------------------------------------
int test_map(
	zpoly_t * pr1,
	zpoly_t * pg1,
	zpoly_t * pr2,
	zpoly_t * pg2,
	zpoly_t * phi,
	int       p,
	int       pn,
	int       verbosity)
{
	int ia, ib;
	zpoly_t a, b, ab;
	zpoly_t phi_ab;
	zpoly_t phi_a_phi_b;


	if (verbosity >= 1) {
		zppoly_aux_t aux2;
		aux2.p = p;
		aux2.r = *pr2;

		printf("Test: ");
		zppoly_print(pg1, p);
		printf(" mod ");
		zppoly_print(pr1, p);
		printf(" and ");
		zppoly_print(pg2, p);
		printf(" mod ");
		zppoly_print(pr2, p);
		printf("\n");

		print_phi(phi, p, pn, &aux2);
	}

	for (ia = 0; ia < pn; ia++) {
		a = zppoly_from_base_rep(p, ia);
		for (ib = 0; ib < pn; ib++) {
			b = zppoly_from_base_rep(p, ib);
			ab = zppoly_add(&a, &b, p);
			phi_ab = phi[zppoly_to_base_rep(&ab, p)];
			phi_a_phi_b = zppoly_add(&phi[ia], &phi[ib], p);

			if (verbosity >= 2) {
				zppoly_aux_t aux1, aux2;
				aux1.p = p;
				aux1.r = *pr1;
				aux2.p = p;
				aux2.r = *pr2;

				printf("  ");
				printf(" a=");
				zppolymod_print(&a, &aux1);
				printf(" b=");
				zppolymod_print(&b, &aux1);
				printf(" phi(a)=");
				zppolymod_print(&phi[ia], &aux2);
				printf(" phi(b)=");
				zppolymod_print(&phi[ib], &aux2);
				printf(" phi(a+b)=");
				zppolymod_print(&phi_ab, &aux2);
				if (zppoly_equal(&phi_ab, &phi_a_phi_b, p))
					printf(" ==");
				else
					printf(" !=");
				printf(" phi(a)+phi(b)=");
				zppolymod_print(&phi_a_phi_b, &aux2);
				printf("\n");

			}

			if (!zppoly_equal(&phi_ab, &phi_a_phi_b, p))
				return 0;

			// We don't need to test multiplication *if* we
			// obtained the isomorphism from multiplicative
			// generators.
		}
		if (verbosity >= 2) {
			printf("\n");
		}
	}

	return 1;
}

// ----------------------------------------------------------------
int test_generators(
	zpoly_t * pr1,
	zpoly_t * pg1,
	zpoly_t * pr2,
	zpoly_t * pg2,
	int       p,
	int       n,
	int       verbosity)
{
	int pn = calc_power(p, n);
	int i;
	zpoly_t * phi;
	zpoly_t g1p, g2p;
	zppoly_aux_t aux1, aux2;
	int rv = 0;

	aux1.p = p;
	aux1.r = *pr1;
	aux2.p = p;
	aux2.r = *pr2;

	phi = malloc_check(pn * sizeof(zpoly_t));

	phi[0] = zppoly_from_coeffs0(0, p);
	g1p = zppoly_from_coeffs0(1, p);
	g2p = zppoly_from_coeffs0(1, p);

	for (i = 1; i < pn; i++) {
		phi[zppoly_to_base_rep(&g1p, p)] = g2p;
		g1p = zppolymod_mul(&g1p, pg1, &aux1);
		g2p = zppolymod_mul(&g2p, pg2, &aux2);
	}

	rv = test_map(pr1, pg1, pr2, pg2, phi, p, pn, verbosity);
	if (rv) {

		printf("Isom: ");
		zppoly_print(pg1, p);
		printf(" mod ");
		zppoly_print(pr1, p);
		printf(" and ");
		zppoly_print(pg2, p);
		printf(" mod ");
		zppoly_print(pr2, p);
		printf("\n");

		for (i = 0; i < pn; i++) {
			zpoly_t z = zppoly_from_base_rep(p, i);
			printf("  ");
			zppolymod_print(&z, &aux1);
			printf(" |-> ");
			zppolymod_print(&phi[i], &aux1);
			printf("\n");
		}

		printf("\n");
	}

	free(phi);
	return rv;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int p, n, pn1;
	int iglo, ighi, ig1, ig2;
	zpoly_t r1, r2, g1, g2;
	zppoly_aux_t aux1, aux2;
	int ng1 = 0;
	int nisom = 0;
	int verb = 0;

	if (argc < 4)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &p) != 1)
		usage(argv[0]);
	if (!zppoly_scan(&r1, argv[2], p))
		usage(argv[0]);
	if (!zppoly_scan(&r2, argv[3], p))
		usage(argv[0]);
	if (!zppoly_is_irreducible(&r1, p)) {
		fprintf(stderr, "%s is not irreducible.\n", argv[2]);
		exit(1);
	}
	if (!zppoly_is_irreducible(&r2, p)) {
		fprintf(stderr, "%s is not irreducible.\n", argv[3]);
		exit(1);
	}
	verb = argc - 4;

	if (r1.degree != r2.degree)
		usage(argv[0]);
	n = r1.degree;

	pn1  = calc_power(p, n) - 1;
	iglo = p;
	ighi = pn1;
	if (ighi < iglo)
		ighi = iglo; // linear fields

	aux1.p = p;
	aux1.r = r1;
	aux2.p = p;
	aux2.r = r2;

	ng1 = 0;
	for (ig1 = iglo; ig1 <= ighi; ig1++) {
		g1 = zppoly_from_base_rep(p, ig1);

		if (zppolymod_element_order(&g1, &aux1) != pn1)
			continue;

		if (ng1 > 0)
			break;

		for (ig2 = iglo; ig2 <= ighi; ig2++) {
			g2 = zppoly_from_base_rep(p, ig2);
			if (zppolymod_element_order(&g2, &aux2) != pn1)
				continue;

			if (test_generators(&r1, &g1, &r2, &g2, p, n, verb))
				nisom++;
		}

		ng1++;
	}
	printf("%d isom%s\n", nisom, nisom == 1 ? "" : "s");

	return 0;
}
