#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {r1} {r2}, of same degree\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void print_phi(
	z2ipoly_t * phi,
	int pn)
{
	int i;

#if 0
	for (i = 0; i < pn; i++)
		printf(" %llx |-> %llx\n", (z2ipoly_t)i, phi[i]);
#else
	for (i = 0; i < pn; i++)
		printf(" %llx", (z2ipoly_t)i);
	printf("\n");

	for (i = 0; i < pn; i++)
		printf(" %llx", phi[i]);
	printf("\n");
#endif
}

// ----------------------------------------------------------------
int test_map(
	z2ipoly_t   r1,
	z2ipoly_t   g1,
	z2ipoly_t   r2,
	z2ipoly_t   g2,
	z2ipoly_t * phi,
	int         pn,
	int         verbosity)
{
	z2ipoly_t a, b;
	z2ipoly_t phi_ab;
	z2ipoly_t phi_a_phi_b;

	if (verbosity >= 1) {
		printf("Test: ");
		z2ipoly_print(g1);
		printf(" mod ");
		z2ipoly_print(r1);
		printf(" and ");
		z2ipoly_print(g2);
		printf(" mod ");
		z2ipoly_print(r2);
		printf("\n");
	}

	for (a = 0LL; a < (z2ipoly_t)pn; a++) {
		for (b = 0LL; b < (z2ipoly_t)pn; b++) {

			phi_ab = phi[a ^ b];
			phi_a_phi_b = phi[a] ^ phi[b];

			if (verbosity >= 2) {
				printf("  ");
				printf(" a=");
				z2ipolymod_print(a, r1);
				printf(" b=");
				z2ipolymod_print(b, r1);
				printf(" phi(a)=");
				z2ipolymod_print(phi[a], r2);
				printf(" phi(b)=");
				z2ipolymod_print(phi[b], r2);
				printf(" phi(a+b)=");
				z2ipolymod_print(phi_ab, r2);
				if (phi_ab == phi_a_phi_b)
					printf(" ==");
				else
					printf(" !=");
				printf(" phi(a)+phi(b)=");
				z2ipolymod_print(phi_a_phi_b, r2);
				printf("\n");
			}

			if (phi_ab != phi_a_phi_b) {
				return 0;
			}


			// We don't need to test multiplication *if* we
			// obtained the isomorphism from multiplicative
			// generators.
		}
	}

	return 1;
}

// ----------------------------------------------------------------
int test_generators(
	z2ipoly_t r1,
	z2ipoly_t g1,
	z2ipoly_t r2,
	z2ipoly_t g2,
	int       n,
	int       verbosity)
{
	int pn = 1 << n;
	int i;
	z2ipoly_t * phi;
	z2ipoly_t g1p, g2p;
	int rv = 0;

	phi = malloc_check(pn * sizeof(z2ipoly_t));

	//printf("Test: %llx mod %llx and %llx mod %llx\n",
		//g1, r1, g2, r2);

	phi[0] = 0LL;
	g1p = 1LL;
	g2p = 1LL;
	for (i = 1; i < pn; i++) {
		phi[g1p] = g2p;
		g1p = z2ipolymod_mul(g1p, g1, r1);
		g2p = z2ipolymod_mul(g2p, g2, r2);
	}

	rv = test_map(r1, g1, r2, g2, phi, pn, verbosity);
	if (rv) {
		printf("Isom: %llx mod %llx and %llx mod %llx\n",
			g1, r1, g2, r2);

		print_phi(phi, pn);
		printf("\n");
	}
	else {
		//printf("Not:  %llx mod %llx and %llx mod %llx\n",
			//g1, r1, g2, r2);
	}

	free(phi);
	return rv;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t r1, r2, glo, ghi, g1, g2;
	z2ipoly_t pn1;
	int n;
	int ng1 = 0;
	int nisom = 0;
	int verbosity = 0;

	if (argc < 3)
		usage(argv[0]);
	if (!z2ipoly_scan(&r1, argv[1]))
		usage(argv[0]);
	if (!z2ipoly_scan(&r2, argv[2]))
		usage(argv[0]);
	if (!z2ipoly_is_irreducible(r1)) {
		fprintf(stderr, "%s is not irreducible.\n", argv[1]);
		exit(1);
	}
	if (!z2ipoly_is_irreducible(r2)) {
		fprintf(stderr, "%s is not irreducible.\n", argv[2]);
		exit(1);
	}
	if (z2ipoly_get_degree(r1) != z2ipoly_get_degree(r2))
		usage(argv[0]);
	verbosity = argc - 3;
	n = z2ipoly_get_degree(r1);

	pn1 = (1LL << n) - 1;
	glo = 2LL;
	ghi = pn1;
	if (ghi < glo)
		ghi = glo; // linear fields

	ng1 = 0;
	for (g1 = glo; g1 <= ghi; g1++) {
		if (z2ipolymod_element_order(g1, r1) != pn1)
			continue;

		if (ng1 > 0)
			break;

		for (g2 = glo; g2 <= ghi; g2++) {
			if (z2ipolymod_element_order(g2, r2) != pn1)
				continue;

			if (test_generators(r1, g1, r2, g2, n, verbosity))
				nisom++;
		}

		ng1++;
	}
	printf("%d isom%s\n", nisom, nisom == 1 ? "" : "s");

	return 0;
}
