#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z3polymod.h"
#include "power.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {r1} {r2}, of same degree\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void print_phi(
	z3poly_t * phi,
	int        pn,
	z3poly_t * pr)
{
	int i;
	z3poly_t z;

	for (i = 0; i < pn; i++) {
		z = z3poly_from_base_rep(i);
#if 0
		printf("  ");
		z3polymod_print(&z, pr);
		printf(" |-> ");
		z3polymod_print(&phi[i], pr);
		printf("\n");
#else
		if (i > 0)
			printf(" ");
		z3polymod_print(&phi[i], pr);
#endif
	}
}

// ----------------------------------------------------------------
int test_map(
	z3poly_t * pr1,
	z3poly_t * pg1,
	z3poly_t * pr2,
	z3poly_t * pg2,
	z3poly_t * phi,
	int       pn,
	int       verbosity)
{
	int ia, ib;
	z3poly_t a, b, ab;
	z3poly_t phi_ab;
	z3poly_t phi_a_phi_b;


	if (verbosity >= 1) {
		printf("Test: ");
		z3poly_print(pg1);
		printf(" mod ");
		z3poly_print(pr1);
		printf(" and ");
		z3poly_print(pg2);
		printf(" mod ");
		z3poly_print(pr2);
		printf("\n");

		print_phi(phi, pn, pr2);
	}

	for (ia = 0; ia < pn; ia++) {
		a = z3poly_from_base_rep(ia);
		for (ib = 0; ib < pn; ib++) {
			b = z3poly_from_base_rep(ib);
			ab = z3poly_add(&a, &b);
			phi_ab = phi[z3poly_to_base_rep(&ab)];
			phi_a_phi_b = z3poly_add(&phi[ia], &phi[ib]);

			if (verbosity >= 2) {
				printf("  ");
				printf(" a=");
				z3polymod_print(&a, pr1);
				printf(" b=");
				z3polymod_print(&b, pr1);
				printf(" phi(a)=");
				z3polymod_print(&phi[ia], pr2);
				printf(" phi(b)=");
				z3polymod_print(&phi[ib], pr2);
				printf(" phi(a+b)=");
				z3polymod_print(&phi_ab, pr2);
				if (z3poly_equal(&phi_ab, &phi_a_phi_b))
					printf(" ==");
				else
					printf(" !=");
				printf(" phi(a)+phi(b)=");
				z3polymod_print(&phi_a_phi_b, pr2);
				printf("\n");

			}

			if (!z3poly_equal(&phi_ab, &phi_a_phi_b))
				return 0;

			// We don't need to test multiplication *if* we
			// obtained the isomorphism from multiplicative
			// generators.
		}
		if (verbosity >= 2) {
			printf("\n");
		}
	}

	return 1;
}

// ----------------------------------------------------------------
int test_generators(
	z3poly_t * pr1,
	z3poly_t * pg1,
	z3poly_t * pr2,
	z3poly_t * pg2,
	int       n,
	int       verbosity)
{
	int pn = calc_power(3, n);
	int i;
	z3poly_t * phi;
	z3poly_t g1p, g2p;
	int rv = 0;

	phi = malloc_check(pn * sizeof(z3poly_t));

	phi[0] = z3poly_from_base_rep(0);
	g1p = z3poly_from_base_rep(1);
	g2p = z3poly_from_base_rep(1);

	for (i = 1; i < pn; i++) {
		phi[z3poly_to_base_rep(&g1p)] = g2p;
		g1p = z3polymod_mul(&g1p, pg1, pr1);
		g2p = z3polymod_mul(&g2p, pg2, pr2);
	}

	rv = test_map(pr1, pg1, pr2, pg2, phi, pn, verbosity);
	if (rv) {

		printf("Isom: ");
		z3poly_print(pg1);
		printf(" mod ");
		z3poly_print(pr1);
		printf(" and ");
		z3poly_print(pg2);
		printf(" mod ");
		z3poly_print(pr2);
		printf("\n");

		print_phi(phi, pn, pr2);

		printf("\n");
	}

	free(phi);
	return rv;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int n, pn1;
	int iglo, ighi, ig1, ig2;
	z3poly_t r1, r2, g1, g2;
	int degr1, degr2;
	int ng1 = 0;
	int nisom = 0;
	int verb = 0;

	if (argc < 3)
		usage(argv[0]);
	if (!z3poly_scan(&r1, argv[1]))
		usage(argv[0]);
	if (!z3poly_scan(&r2, argv[2]))
		usage(argv[0]);
	if (!z3poly_is_irreducible(&r1)) {
		fprintf(stderr, "%s is not irreducible.\n", argv[1]);
		exit(1);
	}
	if (!z3poly_is_irreducible(&r2)) {
		fprintf(stderr, "%s is not irreducible.\n", argv[2]);
		exit(1);
	}
	verb = argc - 3;

	degr1 = z3poly_get_degree(&r1);
	degr2 = z3poly_get_degree(&r2);
	if (degr1 != degr2)
		usage(argv[0]);
	n = degr1;

	pn1  = calc_power(3, n) - 1;
	iglo = 3;
	ighi = pn1;
	if (ighi < iglo)
		ighi = iglo; // linear fields

	ng1 = 0;
	for (ig1 = iglo; ig1 <= ighi; ig1++) {
		g1 = z3poly_from_base_rep(ig1);

		if (z3polymod_element_order(&g1, &r1) != pn1)
			continue;

		if (ng1 > 0)
			break;

		for (ig2 = iglo; ig2 <= ighi; ig2++) {
			g2 = z3poly_from_base_rep(ig2);
			if (z3polymod_element_order(&g2, &r2) != pn1)
				continue;

			if (test_generators(&r1, &g1, &r2, &g2, n, verb))
				nisom++;
		}

		ng1++;
	}
	printf("%d isom%s\n", nisom, nisom == 1 ? "" : "s");

	return 0;
}
