#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zppolymod.h"
#include "zplinalg.h"
#include "power.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {r1} {r2}, of same degree\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void print_phi(
	zpoly_t * phi,
	int       p,
	int       pn,
	zppoly_aux_t * paux)
{
	int i;
	zpoly_t z;

	for (i = 0; i < pn; i++) {
		z = zppoly_from_base_rep(p, i);
		printf("  ");
		zppolymod_print(&z, paux);
		printf(" |-> ");
		zppolymod_print(&phi[i], paux);
		printf("\n");
	}
}

// ----------------------------------------------------------------
int test_map(
	zpoly_t * pr1,
	zpoly_t * pg1,
	zpoly_t * pr2,
	zpoly_t * pg2,
	zpoly_t * phi,
	int       p,
	int       pn,
	int       verbosity)
{
	int ia, ib;
	zpoly_t a, b, ab;
	zpoly_t phi_ab;
	zpoly_t phi_a_phi_b;


	if (verbosity >= 1) {
		zppoly_aux_t aux2;
		aux2.p = p;
		aux2.r = *pr2;

		printf("Test: ");
		zppoly_print(pg1, p);
		printf(" mod ");
		zppoly_print(pr1, p);
		printf(" and ");
		zppoly_print(pg2, p);
		printf(" mod ");
		zppoly_print(pr2, p);
		printf("\n");

		print_phi(phi, p, pn, &aux2);
	}

	for (ia = 0; ia < pn; ia++) {
		a = zppoly_from_base_rep(p, ia);
		for (ib = 0; ib < pn; ib++) {
			b = zppoly_from_base_rep(p, ib);
			ab = zppoly_add(&a, &b, p);
			phi_ab = phi[zppoly_to_base_rep(&ab, p)];
			phi_a_phi_b = zppoly_add(&phi[ia], &phi[ib], p);

			if (verbosity >= 2) {
				zppoly_aux_t aux1, aux2;
				aux1.p = p;
				aux1.r = *pr1;
				aux2.p = p;
				aux2.r = *pr2;

				printf("  ");
				printf(" a=");
				zppolymod_print(&a, &aux1);
				printf(" b=");
				zppolymod_print(&b, &aux1);
				printf(" phi(a)=");
				zppolymod_print(&phi[ia], &aux2);
				printf(" phi(b)=");
				zppolymod_print(&phi[ib], &aux2);
				printf(" phi(a+b)=");
				zppolymod_print(&phi_ab, &aux2);
				if (zppoly_equal(&phi_ab, &phi_a_phi_b, p))
					printf(" ==");
				else
					printf(" !=");
				printf(" phi(a)+phi(b)=");
				zppolymod_print(&phi_a_phi_b, &aux2);
				printf("\n");

			}

			if (!zppoly_equal(&phi_ab, &phi_a_phi_b, p))
				return 0;

			// We don't need to test multiplication *if* we
			// obtained the isomorphism from multiplicative
			// generators.
		}
		if (verbosity >= 2) {
			printf("\n");
		}
	}

	return 1;
}

// ----------------------------------------------------------------
void do_matrix_stuff(
	zpoly_t * phi,
	int       p,
	int       n,
	int       pn,
	zppoly_aux_t * paux1,
	zppoly_aux_t * paux2)
{
	int i, j, k, br = 1;
	zpoly_t a, b, Ma;
	int ** M;
	int ** Mpow;
	int ngood = 0, nbad = 0;

	M = zpla_alloc_matrix(n, n);
	Mpow = zpla_alloc_matrix(n, n);
	printf("\n");
	printf("Matrix stuff:\n");
	for (i = 0; i < n; i++) {
		a = zppoly_from_base_rep(p, br);
		b = phi[br];

		printf("  a=");
		zppolymod_print(&a, paux1);
		printf(" phi(a)=");
		zppolymod_print(&b, paux2);
		printf("\n");

		for (j = 0; j < n; j++)
			M[j][i] = 0;
		for (j = 0; j <= b.degree; j++)
			M[j][i] = b.coeffs[j];

		br *= p;
	}

	zpla_print_matrix(M, n, n);

	zpla_mats_copy(Mpow, M, n);
	for (i = 1; i <= n; i++) {
		printf("M^%d:\n", i);
		zpla_print_matrix(Mpow, n, n);
		if (zpla_is_I(Mpow, n))
			printf("M^%d=I\n", i);
		zpla_mats_mul(Mpow, Mpow, M, p, n);
	}

	// for each poly
	// check M*poly against phi of poly
	for (i = 0; i < pn; i++) {
		a = zppoly_from_base_rep(paux1->p, i);
		b = phi[i];

		// [ Ma0 ]   [ M00 M01 M02 ]   [ a0 ]
		// [ Ma1 ] = [ M10 M11 M12 ] * [ a1 ]
		// [ Ma2 ]   [ M20 M21 M22 ]   [ a2 ]

		// M00*a0 + M01*a1 + M02*a2
		// M10*a0 + M11*a1 + M12*a2
		// M20*a0 + M21*a1 + M22*a2
		for (j = 0; j < n; j++) {
			Ma.coeffs[j] = 0;
			//printf("    Ma[%d]: ", j);
			for (k = 0; k <= a.degree; k++) {
				//printf(" %d*%d=%d",
					//M[j][k],
					//k > a.degree ? 0 : a.coeffs[k],
					//(M[j][k] * a.coeffs[k]) % p);
				Ma.coeffs[j] += M[j][k] * a.coeffs[k];
				Ma.coeffs[j] %= paux1->p;
			}
			//printf("\n");
		}
		Ma.degree = zppoly_find_degree(&Ma, n - 1);

		if (zppoly_equal(&b, &Ma, paux1->p))
			ngood++;
		else
			nbad++;

		//printf("  a=");
		//zppolymod_print(&a, paux1);
		//printf(" phi(a)=");
		//zppolymod_print(&b, paux2);
		//printf(" M*a=");
		//zppolymod_print(&Ma, paux2);
		//printf("%s", zppoly_equal(&b, &Ma, paux1->p) ? " *" : "");
		//printf("\n");
	}
	printf("matrix good=%d bad=%d\n", ngood, nbad);

	zpla_free_matrix(M, n, n);
	zpla_free_matrix(Mpow, n, n);
	printf("\n");
}

// ----------------------------------------------------------------
int test_generators(
	zpoly_t * pr1,
	zpoly_t * pg1,
	zpoly_t * pr2,
	zpoly_t * pg2,
	int       p,
	int       n,
	int       verbosity)
{
	int pn = calc_power(p, n);
	int i;
	zpoly_t * phi;
	zpoly_t g1p, g2p;
	zppoly_aux_t aux1, aux2;
	int rv = 0;

	aux1.p = p;
	aux1.r = *pr1;
	aux2.p = p;
	aux2.r = *pr2;

	phi = malloc_check(pn * sizeof(zpoly_t));

	phi[0] = zppoly_from_coeffs0(0, p);
	g1p = zppoly_from_coeffs0(1, p);
	g2p = zppoly_from_coeffs0(1, p);

	for (i = 1; i < pn; i++) {
		phi[zppoly_to_base_rep(&g1p, p)] = g2p;
		g1p = zppolymod_mul(&g1p, pg1, &aux1);
		g2p = zppolymod_mul(&g2p, pg2, &aux2);
	}

	rv = test_map(pr1, pg1, pr2, pg2, phi, p, pn, verbosity);
	if (rv) {

		printf("Isom: ");
		zppoly_print(pg1, p);
		printf(" mod ");
		zppoly_print(pr1, p);
		printf(" and ");
		zppoly_print(pg2, p);
		printf(" mod ");
		zppoly_print(pr2, p);
		printf("\n");

		print_phi(phi, p, pn, &aux1);

		do_matrix_stuff(phi, p, n, pn, &aux1, &aux2);

		printf("\n");
	}

	free(phi);
	return rv;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int p, n, pn1;
	int iglo, ighi, ig2;
	zpoly_t r1, r2, g1, g2;
	zppoly_aux_t aux1, aux2;
	int nisom = 0;
	int verb = 0;

	if (argc < 4)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &p) != 1)
		usage(argv[0]);
	if (!zppoly_scan(&r1, argv[2], p))
		usage(argv[0]);
	if (!zppoly_scan(&r2, argv[3], p))
		usage(argv[0]);
	if (!zppoly_is_irreducible(&r1, p)) {
		fprintf(stderr, "%s is not irreducible.\n", argv[2]);
		exit(1);
	}
	if (!zppoly_is_irreducible(&r2, p)) {
		fprintf(stderr, "%s is not irreducible.\n", argv[3]);
		exit(1);
	}
	verb = argc - 4;

	if (r1.degree != r2.degree)
		usage(argv[0]);
	n = r1.degree;

	pn1  = calc_power(p, n) - 1;
	iglo = 1;
	ighi = pn1;

	aux1.p = p;
	aux1.r = r1;
	aux2.p = p;
	aux2.r = r2;

	g1 = zppolymod_find_generator(&aux1);
	for (ig2 = iglo; ig2 <= ighi; ig2++) {
		g2 = zppoly_from_base_rep(p, ig2);
		if (zppolymod_element_order(&g2, &aux2) != pn1)
			continue;

		if (test_generators(&r1, &g1, &r2, &g2, p, n, verb))
			nisom++;
	}
	printf("%d isom%s\n", nisom, nisom == 1 ? "" : "s");

	return 0;
}
