#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zppolymod.h"
#include "zpnpoly.h"
#include "zplinalg.h"
#include "power.h"
#include "moebius.h"
#include "malloc_check.h"
#include "log10.h"

// ----------------------------------------------------------------
typedef struct _opts_t {
	int show_orders;
	int show_elements;
	int show_logs;
	int show_normal_rank;
	int show_trace;
	int show_norm;
	int terse_output;
} opts_t;

// ----------------------------------------------------------------
typedef struct _fp_chart_cell_t {
	zpoly_t root;
	int     log;
} fp_chart_cell_t;

typedef struct _fp_chart_row_t {
	zpoly_t   min_poly;
	int       num_roots;
	int       order;
	fp_chart_cell_t * cells;
} fp_chart_row_t;

typedef struct _fp_chart_t {
	int     num_rows;
	zpoly_t g;
	fp_chart_row_t * rows;
} fp_chart_t;

// ----------------------------------------------------------------
static int num_irreds_deg_is(
	int       p,
	int       n);

static int num_irreds_deg_divides(
	int       p,
	int       n);

static void find_irreds(
	int       p,
	int       n,
	zpoly_t * pirreds,
	int       num_irreds);

static fp_chart_t * alloc_fp_chart(
	int       num_rows,
	int       num_cols);

static void free_fp_chart(
	fp_chart_t ** ppchart);

static int fp_chart_row_qcmp(
	const void * pv1,
	const void * pv2);

static void do_fp_chart(
	zpoly_t * pr,
	int       p,
	int       n,
	opts_t  * popts);

static void fill_fp_chart(
	fp_chart_t * pchart,
	zpoly_t * pr,
	int       p,
	int       n);

static void print_fp_chart(
	fp_chart_t * pchart,
	zpoly_t * pr,
	int       p,
	int       n,
	opts_t  * popts);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {n}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int        argi;
	int        p, n;
	zpoly_t    rj;
	int        j;
	opts_t     opts;

	opts.show_orders   = 0;
	opts.show_elements = 1;
	opts.show_logs     = 0;
	opts.show_normal_rank = 0;
	opts.show_trace    = 0;
	opts.show_norm     = 0;
	opts.terse_output  = 0;

	argi = 1;
	while ((argi < argc) && (argv[argi][0] == '-')) {
		if (strcmp(argv[argi], "-o") == 0) {
			opts.show_orders = 1;
		}
		else if (strcmp(argv[argi], "-l") == 0) {
			opts.show_elements = 0;
			opts.show_logs = 1;
		}
		else if (strcmp(argv[argi], "-L") == 0) {
			opts.show_elements = 1;
			opts.show_logs = 1;
		}
		else if (strcmp(argv[argi], "-nr") == 0) {
			opts.show_normal_rank = 1;
		}
		else if (strcmp(argv[argi], "-t") == 0) {
			opts.show_trace = 1;
		}
		else if (strcmp(argv[argi], "-n") == 0) {
			opts.show_norm = 1;
		}
		else if (strcmp(argv[argi], "-b") == 0) {
			opts.terse_output = 1;
		}
		else
			usage(argv[0]);
		argi++;
	}

	if ((argc - argi) < 2)
		usage(argv[0]);

	if (sscanf(argv[argi], "%u", &p) != 1)
		usage(argv[0]);
	argi++;

	for (; argi < argc; argi++) {

		if (sscanf(argv[argi], ":%u", &n) == 1) {
			int num_irreds = num_irreds_deg_is(p, n);
			zpoly_t * pirreds = malloc_check(
				num_irreds * sizeof(zpoly_t));
			find_irreds(p, n, pirreds, num_irreds);
			for (j = 0; j < num_irreds; j++) {
				rj = pirreds[j];
				do_fp_chart(&rj, p, n, &opts);
			}
			free(pirreds);
		}
		else if (zppoly_scan(&rj, argv[argi], p)) {
			n = rj.degree;
			do_fp_chart(&rj, p, n, &opts);
		}
		else {
			usage(argv[0]);
		}
	}

	return 0;
}

// ----------------------------------------------------------------
static int num_irreds_deg_is(
	int       p,
	int       n)
{
	int d, sum = 0;

	for (d = 1; d <= n; d++)
		if ((n % d) == 0)
			sum += calc_moebius_unsigned(n/d) * calc_power(p, d);
	return sum / n;
}

// ----------------------------------------------------------------
static int num_irreds_deg_divides(
	int       p,
	int       n)
{
	int d, sum = 0;
	for (d = 1; d <= n; d++)
		if ((n % d) == 0)
			sum += num_irreds_deg_is(p, d);
	return sum;
}

// ----------------------------------------------------------------
static void find_irreds(
	int       p,
	int       n,
	zpoly_t * pirreds,
	int       num_irreds)
{
	int ilo, ihi, i, count = 0;
	zpoly_t f;

	ilo = calc_power(p, n);
	ihi = 2 * ilo - 1;
	for (i = ilo; i <= ihi; i++) {
		f = zppoly_from_base_rep(p, i);
		if (!zppoly_is_irreducible(&f, p))
			continue;

		if (count >= num_irreds) {
			fprintf(stderr, "Coding error:  too many factors.\n");
			exit(1);
		}
		pirreds[count] = f;
		count++;
	}
}

// ----------------------------------------------------------------
static fp_chart_t * alloc_fp_chart(
	int       num_rows,
	int       num_cols)
{
	fp_chart_t * pchart;
	int i, j;

	pchart = malloc_check(sizeof(fp_chart_t));
	pchart->num_rows = num_rows;
	pchart->rows = malloc_check(num_rows * sizeof(fp_chart_row_t));
	for (i = 0; i < num_rows; i++)
		pchart->rows[i].cells = malloc_check(
			num_cols * sizeof(fp_chart_cell_t));

	pchart->g = zppoly_from_base_rep(1, 0);
	for (i = 0; i < num_rows; i++) {
		pchart->rows[i].min_poly = zppoly_from_base_rep(1, 0);
		pchart->rows[i].num_roots = 0;
		pchart->rows[i].order = 0;
		for (j = 0; j < num_cols; j++)
			pchart->rows[i].cells[j].root =
				zppoly_from_base_rep(1, 0);
			pchart->rows[i].cells[j].log = 0;
	}

	return pchart;
}

// ----------------------------------------------------------------
static void free_fp_chart(
	fp_chart_t ** ppchart)
{
	fp_chart_t * pchart = *ppchart;
	int i;
	for (i = 0; i < pchart->num_rows; i++)
		free(pchart->rows[i].cells);
	free(pchart->rows);
	free(pchart);
	*ppchart = 0;
}

// ----------------------------------------------------------------
static int fp_chart_row_qcmp(
	const void * pv1,
	const void * pv2)
{
	const fp_chart_row_t * prow1 = pv1;
	const fp_chart_row_t * prow2 = pv2;
	return (zppoly_gt(
		(zpoly_t *)&prow1->min_poly,
		(zpoly_t *)&prow2->min_poly));
}

// ----------------------------------------------------------------
static void do_fp_chart(
	zpoly_t * pr,
	int       p,
	int       n,
	opts_t  * popts)
{
	int num_rows = num_irreds_deg_divides(p, n);
	fp_chart_t  * pchart = alloc_fp_chart(num_rows, n);

	fill_fp_chart(pchart, pr, p, n);
	print_fp_chart(pchart, pr, p, n, popts);

	free_fp_chart(&pchart);
}

// ----------------------------------------------------------------
static void fill_fp_chart(
	fp_chart_t * pchart,
	zpoly_t * pr,
	int       p,
	int       n)
{
	int pn1 = calc_power(p, n) - 1;
	int i, d, pd1, sfld_log, elt_log, orbit_log, j;
	char * marks;
	zpoly_t g, elt;
	zppoly_aux_t aux;
	int    rowi = 0;
	fp_chart_row_t * prow;

	aux.p = p;
	aux.r = *pr;

	g = zppolymod_find_generator(&aux);
	pchart->g = g;
	marks = malloc_check(pn1 * sizeof(char));
	for (i = 0; i < pn1; i++)
		marks[i] = 0;

	pchart->rows[rowi].min_poly = zppoly_from_coeffs1(1, 0, p);
	pchart->rows[rowi].num_roots = 1;
	pchart->rows[rowi].order = 0;
	pchart->rows[rowi].cells[0].root = zppoly_from_base_rep(1, 0);
	pchart->rows[rowi].cells[0].log  = 0;
	rowi++;

	for (d = 1; d <= n; d++) {
		if ((n % d) != 0)
			continue;
		pd1 = calc_power(p, d) - 1;
		sfld_log = pn1 / pd1;

		for (elt_log = 0; elt_log < pn1; elt_log += sfld_log) {
			if (marks[elt_log])
				continue;

			prow = &pchart->rows[rowi];

			orbit_log = elt_log;
			elt = zppolymod_power(&g, orbit_log, &aux);

			prow->min_poly = zppoly_get_min_poly(&elt, &aux);
			pchart->rows[rowi].num_roots = d;
			pchart->rows[rowi].order =
				zppolymod_element_order(&elt, &aux);

			for (j = 0; j < d; j++) {
				marks[orbit_log] = 1;

				prow->cells[j].root = elt;
				prow->cells[j].log  = orbit_log;

				orbit_log *= p;
				orbit_log %= pn1;
				elt = zppolymod_power(&g, orbit_log, &aux);
			}
			rowi++;
		}
	}

	qsort(pchart->rows, pchart->num_rows, sizeof(pchart->rows[0]),
		fp_chart_row_qcmp);

	free(marks);
}

// ----------------------------------------------------------------
static void print_fp_chart(
	fp_chart_t * pchart,
	zpoly_t * pr,
	int       p,
	int       n,
	opts_t  * popts)
{
	int i, j;
	zppoly_aux_t aux;
	fp_chart_cell_t * pc;
	zpoly_t * pz;
	int pn, l, width;

	printf("%s", ""); // Workaround for apparent GCC bug.
	pn = calc_power(p, n);
	printf("%s", "");
	l = calc_log10_unsigned(pn, IFLOOR);
	printf("%s", "");
	width = l + 1;

	aux.p = p;
	aux.r = *pr;

	if (!popts->terse_output) {
		printf("Root chart for ");
		zppoly_print(pr, p);
		if (popts->show_logs) {
			printf(", g = ");
			zppolymod_print(&pchart->g, &aux);
		}
		printf("\n");
	}

	for (i = 0; i < pchart->num_rows; i++) {
		if (!popts->terse_output) {
			zppoly_dprint(&pchart->rows[i].min_poly, p, n);
			printf(":");
		}
		if (popts->show_orders) {
			printf(" [o= %*d]", width, pchart->rows[i].order);
		}

		if (popts->show_normal_rank) {
			// xxx make get-rank function
			int nroots = pchart->rows[i].num_roots;
			int rank = 0;
			int ** M = zpla_alloc_matrix(nroots, n);
			for (j = 0; j < nroots; j++)
				zpla_row_from_zpoly(M[j],
					&pchart->rows[i].cells[j].root, n);
			zpla_row_echelon_form(M, p, nroots, n);
			rank = zpla_find_rank(M, nroots, n);
			printf(" [nr= %*d]", width, rank);
			zpla_free_matrix(M, nroots, n);
		}

		if (popts->show_trace) {
			// Sum of conjugates
			int nroots = pchart->rows[i].num_roots;
			zpoly_t trace = zppoly_from_base_rep(p, 0);
			for (j = 0; j < nroots; j++) {
				trace = zppoly_add(&trace,
					&pchart->rows[i].cells[j].root, p);
			}
			printf(" tr=");
			zppoly_print(&trace, p);
		}

		if (popts->show_norm) {
			// Product of conjugates
			int nroots = pchart->rows[i].num_roots;
			zpoly_t norm = zppoly_from_base_rep(p, 1);
			for (j = 0; j < nroots; j++) {
				norm = zppolymod_mul(&norm,
					&pchart->rows[i].cells[j].root, &aux);
			}
			printf(" nm=");
			zppoly_print(&norm, p);
		}

		for (j = 0; j < pchart->rows[i].num_roots; j++) {
			pc = &pchart->rows[i].cells[j];
			pz = &pc->root;

			printf(" ");
			if (popts->show_elements) {
				zppolymod_print(pz, &aux);
			}
			if (popts->show_logs) {
				if (popts->show_elements)
					printf("(");
				if (zppoly_is_zero(pz))
					printf("%*s", width, "-");
				else
					printf("%*d",
						width,
						pchart->rows[i].cells[j].log);
				if (popts->show_elements)
					printf(")");
			}
		}
		if (!popts->terse_output) {
			printf("\n");
		}
	}
	printf("\n");
}
