#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zppolymod.h"
#include "power.h"
#include "moebius.h"
#include "malloc_check.h"

typedef struct _opts_t {
	int show_orders;
	int show_elements;
	int show_logs;
} opts_t;

typedef struct _chart_cell_t {
	zpoly_t root;
	int     log;
} chart_cell_t;

typedef struct _chart_row_t {
	zpoly_t   min_poly;
	int       num_roots;
	chart_cell_t * cells;
} chart_row_t;

typedef struct _chart_t {
	int     num_rows;
	chart_row_t * rows;
} chart_t;

static int num_irreds_deg_is(
	int       p,
	int       n);

static int num_irreds_deg_divides(
	int       p,
	int       n);

static void find_irreds(
	int       p,
	int       n,
	zpoly_t * pirreds,
	int       num_irreds);

static chart_t * alloc_chart(
	int       num_rows,
	int       num_cols);

static void free_chart(
	chart_t ** ppchart);

static void do_chart(
	zpoly_t * pr,
	int       p,
	int       n,
	opts_t  * popts);

static void fill_chart(
	chart_t * pchart,
	zpoly_t * pr,
	int       p,
	int       n);

static void print_chart(
	chart_t * pchart,
	zpoly_t * pr,
	int       p,
	int       n,
	opts_t  * popts);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {n}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int        argi;
	int        p, n;
	zpoly_t    rj;
	int        j;
	opts_t     opts;

	opts.show_orders = 0;
	opts.show_elements = 1;
	opts.show_logs   = 0;

	if (argc < 3)
		usage(argv[0]);

	argi = 1;
	while (argv[argi][0] == '-') {
		if (strcmp(argv[argi], "-o") == 0)
			opts.show_orders = 1;
		else if (strcmp(argv[argi], "-l") == 0) {
			opts.show_elements = 0;
			opts.show_logs = 1;
		}
		else if (strcmp(argv[argi], "-L") == 0) {
			opts.show_elements = 1;
			opts.show_logs = 1;
		}
		else
			usage(argv[0]);
		argi++;
	}

	if (sscanf(argv[argi], "%u", &p) != 1)
		usage(argv[0]);
	argi++;

	for (; argi < argc; argi++) {

		if (sscanf(argv[argi], ":%u", &n) == 1) {
			int num_irreds = num_irreds_deg_is(p, n);
			zpoly_t * pirreds = malloc_check(
				num_irreds * sizeof(zpoly_t));
			find_irreds(p, n, pirreds, num_irreds);
			for (j = 0; j < num_irreds; j++) {
				rj = pirreds[j];
				do_chart(&rj, p, n, &opts);
				printf("\n");
			}
			free(pirreds);
		}
		else if (zppoly_scan(&rj, argv[argi], p)) {
			n = rj.degree;
			do_chart(&rj, p, n, &opts);
		}
		else {
			usage(argv[0]);
		}
	}

	return 0;
}

// ----------------------------------------------------------------
static int num_irreds_deg_is(
	int       p,
	int       n)
{
	int d, sum = 0;

	for (d = 1; d <= n; d++)
		if ((n % d) == 0)
			sum += calc_moebius_unsigned(n/d) * calc_power(p, d);
	return sum / n;
}

// ----------------------------------------------------------------
static int num_irreds_deg_divides(
	int       p,
	int       n)
{
	int d, sum = 0;
	for (d = 1; d <= n; d++)
		if ((n % d) == 0)
			sum += num_irreds_deg_is(p, d);
	return sum;
}

// ----------------------------------------------------------------
static void find_irreds(
	int       p,
	int       n,
	zpoly_t * pirreds,
	int       num_irreds)
{
	int ilo, ihi, i, count = 0;
	zpoly_t f;

	ilo = calc_power(p, n);
	ihi = 2 * ilo - 1;
	for (i = ilo; i <= ihi; i++) {
		f = zppoly_from_base_rep(p, i);
		if (!zppoly_is_irreducible(&f, p))
			continue;

		if (count >= num_irreds) {
			fprintf(stderr, "Coding error:  too many factors.\n");
			exit(1);
		}
		pirreds[count] = f;
		count++;
	}
}

// ----------------------------------------------------------------
static chart_t * alloc_chart(
	int       num_rows,
	int       num_cols)
{
	chart_t * pchart;
	int i;

	pchart = malloc_check(sizeof(chart_t));
	pchart->num_rows = num_rows;
	pchart->rows = malloc_check(num_rows * sizeof(chart_row_t));
	for (i = 0; i < num_rows; i++)
		pchart->rows[i].cells = malloc_check(
			num_cols * sizeof(chart_cell_t));

	return pchart;
}

// ----------------------------------------------------------------
static void free_chart(
	chart_t ** ppchart)
{
	chart_t * pchart = *ppchart;
	int i;
	for (i = 0; i < pchart->num_rows; i++)
		free(pchart->rows[i].cells);
	free(pchart->rows);
	free(pchart);
	*ppchart = 0;
}

// ----------------------------------------------------------------
static void do_chart(
	zpoly_t * pr,
	int       p,
	int       n,
	opts_t  * popts)
{
	int num_rows = num_irreds_deg_divides(p, n);
	chart_t  * pchart = alloc_chart(num_rows, n);

	fill_chart(pchart, pr, p, n);
	print_chart(pchart, pr, p, n, popts);

	free_chart(&pchart);
}

// ----------------------------------------------------------------
static void fill_chart(
	chart_t * pchart,
	zpoly_t * pr,
	int       p,
	int       n)
{
	int d, i, j, pn, pd, sfld_log;
	zpoly_t * fes, g, elt_power;
	char * marks;
	int  * logs;
	int    rowi;
	zppoly_aux_t aux;

	printf("Doing: ");
	zppoly_print(pr, p);
	printf("\n");

	aux.p = p;
	aux.r = *pr;

	g     = zppolymod_find_generator(&aux);
	fes   = zppolymod_get_elements(p, n, &pn);
	marks = malloc_check(pn * sizeof(char));
	logs  = malloc_check(pn * sizeof(int));
	for (i = 0; i < pn; i++) {
		marks[i] = 0;
		logs[i] = zppolymod_log(&fes[i], &g, &aux);
	}

	rowi = 0;
	for (d = 1; d <= n; d++) {
		if ((n % d) != 0)
			continue;
		pd = calc_power(p, d);
		sfld_log = (pn - 1) / (pd - 1);

printf("-- d=%2d pd=%2d sfld log = %2d\n", d, pd, sfld_log);

		for (i = 0; i < pn; i++) {

			//if (marks[i])
				//continue;
			if ((logs[i] % sfld_log) != 0)
				continue;
			//marks[i] = 1;
//printf("MARKING i=%d ", i);
//zppolymod_print(&fes[i], &aux);
//printf("\n");

printf("d=%d log=%3d ", d, logs[i]);
zppolymod_print(&fes[i], &aux);
printf(": ");

			elt_power = fes[i];
			for (j = 0; j < d; j++) {
				int foo_log = zppolymod_log(
					&elt_power, &g, &aux);

printf(" ");
zppolymod_print(&elt_power, &aux);
printf("(%2d)", foo_log);

//printf(" ");
//zppolymod_print(&elt_power, &aux);

				//marks[foo_log] = 1;
//printf("\nMARKING l=%d ", foo_log);
//zppolymod_print(&fes[foo_log], &aux);
//printf(" >> ");
//zppolymod_print(&elt_power, &aux);
//printf("\n");
//printf("(%2d)", foo_log);

				elt_power = zppolymod_power(&elt_power, p,
					&aux);
			}
printf("\n");
		}
//printf("\n");
	}

	free(fes);
	free(marks);
	free(logs);
}

// ----------------------------------------------------------------
static void print_chart(
	chart_t * pchart,
	zpoly_t * pr,
	int       p,
	int       n,
	opts_t  * popts)
{
	printf("chart print stub.\n");
}

//
//// ----------------------------------------------------------------
//static void split_poly(
//	zpoly_t   * prj,
//	zpoly_t   * irred_factors,
//	int         num_irred_factors,
//	zpoly_t   * dirred_factors,
//	int         num_dirred_factors,
//	int         p,
//	int         n,
//	int         pn,
//	opts_t    * popts)
//{
//	zpoly_t    ri;
//	zpoly_t    g = zppoly_from_base_rep(1, 0);
//	zppoly_aux_t aux;
//	zpnpoly_t ext_factor;
//	int        i;
//
//	// Just for columnar printing.
//	aux.p = p;
//	aux.r = zppoly_from_base_rep(p, calc_power(p, n+1));
//
//	printf("mod ");
//	//printf("%c ", zppoly_is_primitive(prj, p, n) ? 'o' : '-');
//	zppoly_print(prj, p);
//
//	if (popts->show_logs) {
//		zppoly_aux_t auxj;
//		auxj.p = p;
//		auxj.r = *prj;
//		g = zppolymod_find_generator(&auxj);
//		printf(" g=");
//		zppoly_print(&g, p);
//	}
//	printf(":\n");
//
//	for (i = 0; i < num_irred_factors; i++) {
//		ri = irred_factors[i];
//
//		ext_factor = zpnpoly_extend_zppoly(&ri);
//
//		printf("%c ", zppoly_is_primitive(&ri, p, n) ? 'o' : '-');
//		zppolymod_print(&ri, &aux);
//		printf(" = ");
//
//		split_factor(&ext_factor, prj, p, n, &g, popts);
//		printf("\n");
//		fflush(stdout);
//	}
//	printf("\n");
//}
//
//// ----------------------------------------------------------------
//// This uses Berlekamp's algorithm on Fpn[x].
//static void split_factor(
//	zpnpoly_t * pf,
//	zpoly_t   * pr,
//	int         p,
//	int         n,
//	zpoly_t   * pg,
//	opts_t    * popts)
//{
//	zppoly_aux_t aux;
//	zpoly_t root;
//	static zpnpoly_factor_info_t finfo;
//	int i;
//	int root_order = 0;
//
//	aux.p = p;
//	aux.r = *pr;
//
//	zpnpoly_factor(pf, &aux, &finfo);
//
//	for (i = 0; i < finfo.num_distinct; i++) {
//		if (finfo.factor_counts[i].count != 1) {
//			fprintf(stderr,
//			"fpnsplits: multiple factors in universal poly.\n");
//			exit(1);
//		}
//		if (finfo.factor_counts[i].f.degree != 1) {
//			fprintf(stderr,
//			"fpnsplits: non-linear factors in universal poly.\n");
//			exit(1);
//		}
//		root = zppoly_usub(&finfo.factor_counts[i].f.coeffs[0], p);
//		printf(" ");
//
//		if (popts->show_elements) {
//			zppolymod_print(&root, &aux);
//		}
//
//		if (popts->show_logs) {
//			if (popts->show_elements)
//				printf("(");
//			else
//				printf(" ");
//			if (zppoly_is_zero(&root))
//				printf("--");
//			else
//				printf("%2u", zppolymod_log(&root, pg, &aux));
//			if (popts->show_elements)
//				printf(")");
//		}
//
//		if (i == 0)
//			root_order = zppolymod_element_order(&root, &aux);
//	}
//
//	for (i = 0; i < (n - finfo.num_distinct); i++)
//		printf(" .");
//
//	if (popts->show_trace) {
//		// Sum of conjugates
//		zpoly_t trace = zppoly_from_coeffs0(0, p);
//		for (i = 0; i < finfo.num_distinct; i++) {
//			root = zppoly_usub(&finfo.factor_counts[i].f.coeffs[0], p);
//			trace = zppoly_add(&trace, &root, p);
//		}
//		printf(" tr=");
//		zppoly_print(&trace, p);
//	}
//	if (popts->show_norm) {
//		// Product of conjugates
//		zpoly_t norm = zppoly_from_coeffs0(1, p);
//		for (i = 0; i < finfo.num_distinct; i++) {
//			root = zppoly_usub(&finfo.factor_counts[i].f.coeffs[0], p);
//			norm = zppolymod_mul(&norm, &root, &aux);
//		}
//		printf(" nm=");
//		zppoly_print(&norm, p);
//	}
//	if (popts->show_chpoly) {
//		// Characteristic polynomial not implemented yet.
//		printf(" ch=?");
//	}
//
//	if (popts->show_orders)
//		printf(" o= %3d", root_order);
//}
//
////// ----------------------------------------------------------------
////// This is a brute-force search.
////static void split_factor(
////	zpnpoly_t * pf,
////	zpoly_t   * pr,
////	int         p,
////	int         n,
////	zpoly_t   * pg,
////	opts_     * popts)
////{
////	int pn = calc_power(p, n);
////	int i;
////	zpnpoly_t linear_factor, remnant, quot, rem;
////	int root_order = 0;
////	int nfactors = 0;
////	zpoly_t z;
////	zppoly_aux_t aux;
////
////	aux.p = p;
////	aux.r = *pr;
////
////	// Try each possible linear factor (X - a) for a in F_2^n.
////	remnant = *pf;
////	linear_factor.degree = 1;
////	linear_factor.coeffs[1] = zppoly_from_base_rep(p, 1);
////	for (i = 0; i < pn; i++) {
////		z = zppoly_from_base_rep(p, i);
////		linear_factor.coeffs[0] = z;
////
////		zpnpoly_quot_and_rem(&remnant, &linear_factor,
////			&quot, &rem, &aux);
////
////		// Try the next factor.
////		if (!zpnpoly_is_zero(&rem))
////			continue;
////
////		nfactors++;
////		printf(" ");
////		if (popts->show_logs)
////			printf(" %2u", zppolymod_log(&z, pg, &aux));
////		else
////			zppolymod_print(&z, &aux);
////
////		root_order = zppolymod_element_order(&z, &aux);
////
////		remnant = quot;
////
////		if (zpnpoly_is_one(&remnant)) {
////			break;
////		}
////	}
////	for (i = 0; i < (n - nfactors); i++)
////		printf(" .");
////	if (popts->show_orders)
////		printf(" o= %3d", root_order);
////
////	if (!zpnpoly_is_one(&remnant)) {
////		printf("ERROR:  remnant ");
////			zpnpoly_print(&remnant, &aux);
////			printf("\n");
////	}
////}
