#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "zppolymod.h"
#include "zppoly_factor.h"
#include "zpn_convert.h"
#include "zplinalg.h"
#include "zpnlinalg.h"
#include "power.h"
#include "lcm.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
// r = 0x13
// M = 0 0 1 1
//     1 0 0 0
//     0 1 0 0
//     0 0 1 0

static zpoly_t ** make_TLI(
	zpoly_t * pcharpoly,
	zppoly_aux_t * paux,
	zpoly_t * proot,
	int chardeg,
	int do_galois_cfg)
{
	zpoly_t ** M = zpnla_alloc_matrix(chardeg, chardeg);
	int c;
	int i, j;

	// Zero-fill
	for (i = 0; i < chardeg; i++)
		for (j = 0; j < chardeg; j++)
			M[i][j] = zppoly_from_base_rep(paux->p, 0);

	// Ones on subdiagonal
	for (i = 1; i < chardeg; i++)
		M[i][i-1] = zppoly_from_base_rep(paux->p, 1);

	// Coefficients along top row
	for (i = 0; i < chardeg; i++) {
		c = (paux->p - pcharpoly->coeffs[i]) % paux->p;
		if (do_galois_cfg)
			M[i][chardeg-1] = zppoly_from_base_rep(paux->p, c);
		else
			M[0][chardeg-i-1] = zppoly_from_base_rep(paux->p, c);
	}

	// Minus lambda I
	for (i = 0; i < chardeg; i++) {
		M[i][i] = zppolymod_sub(&M[i][i], proot, paux);
	}

	return M;
}

// ----------------------------------------------------------------
static void do_one_poly(
	zpoly_t * pcharpoly,
	int p,
	int do_galois_cfg,
	int verbose)
{
	zppoly_factor_info_t finfo;
	int i, j, k;
	zppoly_aux_t factoraux;
	zpoly_t ** TLI;
	zpoly_t factorgen, splitgen;
	zpoly_t factorroot, splitroot;
	int chardeg = pcharpoly->degree;
	int factordeg;
	int splitdeg = 1;
	zppoly_aux_t splitaux;
	zpoly_t  * eigenvector = malloc_check(chardeg * sizeof(zpoly_t));
	zpoly_t  * checkvector = malloc_check(chardeg * sizeof(zpoly_t));
	zpoly_t ** eigenspace = zpnla_alloc_matrix(chardeg, chardeg);
	int check_rank;
	int num_eivrs = 0;
	int eigenrank = 0;

	zppoly_factor(pcharpoly, p, &finfo);

	printf("\n");
	printf("Charpoly ");
	zppoly_print(pcharpoly, p);
	printf(" = ");
	for (i = 0; i < finfo.num_distinct; i++) {
		printf(" ");
		zppoly_print(&finfo.factor_counts[i].f, p);
		if (finfo.factor_counts[i].count > 1)
			printf("^%d", finfo.factor_counts[i].count);
		factordeg = finfo.factor_counts[i].f.degree;
		splitdeg = calc_lcm_signed(splitdeg, factordeg);
	}
	if (chardeg == splitdeg) {
		splitaux.r = *pcharpoly;
		splitaux.p = p;
	}
	else {
		splitaux = zppoly_find_irreducible(p, splitdeg);
	}
	printf("\n");

	printf("Splitting field = F_%d^%d (", p, splitdeg);
	zppoly_print(&splitaux.r, p);
	printf(")\n");

	for (i = 0; i < finfo.num_distinct; i++) {
		if (verbose) {
			printf("Factor ");
			zppoly_print(&finfo.factor_counts[i].f, p);
			printf(":\n");
		}

		factoraux.p = p;
		factoraux.r = finfo.factor_counts[i].f;
		factordeg = factoraux.r.degree;
		factorroot = zppoly_from_coeffs1(1, 0, factoraux.p);
		// To correctly handle degree-1 fields:
		factorroot = zppoly_mod(&factorroot, &factoraux.r, p);

		zpn_convert_find_generators(&factoraux, &splitaux,
			&factorgen, &splitgen);
		zpn_convert_scalar(&factorroot, &factoraux, &factorgen,
			&splitroot, &splitaux, &splitgen);

		for (j = 0; j < factordeg; j++) {
			TLI = make_TLI(pcharpoly, &splitaux, &splitroot,
				chardeg, do_galois_cfg);

			if (verbose) {
				printf("\n");
				printf("T - ");
				zppolymod_print(&splitroot, &splitaux);
				printf(" I:\n");
				zpnla_print_matrix(TLI, chardeg, chardeg,
					&splitaux);
			}

			zpnla_row_echelon_form(TLI, &splitaux, chardeg,
				chardeg);
			if (verbose) {
				printf("Row-echelon form:\n");
				zpnla_print_matrix(TLI, chardeg, chardeg,
					&splitaux);
				printf("\n");
			}

			// Make sure the nullity is 1, so that we may
			// compute the eigenvector easily.
			check_rank = zpnla_find_rank(TLI, chardeg, chardeg);
			if (check_rank != (chardeg - 1)) {
				fprintf(stderr, "fpeisys coding error.\n");
				exit(1);
			}

			// Compute the eigenvector.
			for (k = 0; k < chardeg - 1; k++) {
				zpoly_t neg = zppoly_usub(&TLI[k][chardeg-1],p);
				eigenvector[k] = neg;
			}
			eigenvector[chardeg-1] = zppoly_from_base_rep(p, 1);

			if (verbose) {
				printf("Eigenvector = ");
				zpnla_print_vector(eigenvector, chardeg,
					&splitaux);
				printf("\n");
			}
			else {
				printf("Eigenvalue = ");
				zppolymod_print(&splitroot, &splitaux);
				printf("; ");
				printf("eigenvector = ");
				zpnla_print_vector(eigenvector, chardeg,
					&splitaux);
				printf("\n");
			}

			// Make sure this is really an eigenvector (check
			// Tv = lv).
			zpnla_mx_times_vr(TLI, eigenvector, checkvector,
				chardeg, &splitaux);
			for (k = 0; k < chardeg; k++) {
				if (!zppoly_is_zero(&checkvector[k])) {
					fprintf(stderr,
						"fpeisys coding error.\n");
					exit(1);
				}
			}

			// Concatenate eigenvector to eigenspace
			for (k = 0; k < chardeg; k++)
				eigenspace[num_eivrs][k] = eigenvector[k];
			num_eivrs++;

			zpnla_free_matrix(TLI, chardeg, chardeg);

			// Conjugates are found via the Frobenius automorphism.
			splitroot = zppolymod_power(&splitroot, p,
				&splitaux);
		}
	}
	printf("\n");

	// Compute eigenspace.
	if (verbose) {
		printf("Eigenspace before row reduction:\n");
		zpnla_print_matrix(eigenspace, num_eivrs, chardeg, &splitaux);
		printf("\n");
	}

	zpnla_row_echelon_form(eigenspace, &splitaux, num_eivrs, chardeg);
	eigenrank = zpnla_find_rank(eigenspace, num_eivrs, chardeg);

	printf("Eigenspace:\n");
	zpnla_print_matrix(eigenspace, eigenrank, chardeg, &splitaux);

	// Report on rank of eigenspace.
	printf("\n");
	printf("Char poly degree = %d; rank of eigenspace = %d.\n",
		chardeg, eigenrank);

	free(eigenvector);
	free(checkvector);
	zpnla_free_matrix(eigenspace, chardeg, chardeg);
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {r}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int p;
	zpoly_t charpoly;
	int argi, argb = 1;
	int do_galois_cfg = 0;
	int verbose = 0;

	for (argi = 1; argi < argc; argi++) {
		if (argv[argi][0] != '-')
			break;
		else if (strcmp(argv[argi], "-g") == 0)
			do_galois_cfg = 1;
		else if (strcmp(argv[argi], "-v") == 0)
			verbose = 1;
		else
			usage(argv[0]);
		argb++;
	}

	if ((argc - argb) < 1)
		usage(argv[0]);

	if (sscanf(argv[argb], "%d", &p) != 1) {
		fprintf(stderr, "%s: Couldn't scan \"%s\" as p.\n",
			argv[0], argv[argb]);
		exit(1);
	}

	for (argi = argb+1; argi < argc; argi++) {
		if (!zppoly_scan(&charpoly, argv[argi], p)) {
			fprintf(stderr,
				"%s: Couldn't scan \"%s\" as charpoly.\n",
				argv[0], argv[argi]);
			exit(1);
		}
		do_one_poly(&charpoly, p, do_galois_cfg, verbose);
		if ((argc - argb) > 2)
			printf("\n");
	}

	return 0;
}
