#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "zppolymod.h"
#include "zppoly_factor.h"
#include "zpn_convert.h"
#include "zplinalg.h"
#include "zpnlinalg.h"
#include "power.h"
#include "lcm.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void my_print_factors(
	zpoly_t * pcharpoly,
	int p,
	zppoly_factor_info_t * pfinfo)
{
	int i;

	printf("\n");
	printf("Charpoly ");
	zppoly_print(pcharpoly, p);
	printf(" = ");
	for (i = 0; i < pfinfo->num_distinct; i++) {
		printf(" ");
		zppoly_print(&pfinfo->factor_counts[i].f, p);
		if (pfinfo->factor_counts[i].count > 1)
			printf("^%d", pfinfo->factor_counts[i].count);
	}
	printf("\n");
}

// ----------------------------------------------------------------
zppoly_aux_t get_split_aux(
	zpoly_t * pcharpoly,
	int p,
	zppoly_factor_info_t *pfinfo)
{
	int i, factordeg, splitdeg = 1;
	zppoly_aux_t splitaux;

	for (i = 0; i < pfinfo->num_distinct; i++) {
		factordeg = pfinfo->factor_counts[i].f.degree;
		splitdeg = calc_lcm_signed(splitdeg, factordeg);
	}
	if (pcharpoly->degree == splitdeg) {
		splitaux.r = *pcharpoly;
		splitaux.p = p;
	}
	else {
		splitaux = zppoly_find_irreducible(p, splitdeg);
	}

	printf("Splitting field = F_%d^%d (", p, splitdeg);
	zppoly_print(&splitaux.r, p);
	printf(")\n");

	return splitaux;
}

// ----------------------------------------------------------------
// r = 0x13
// M = 0 0 1 1
//     1 0 0 0
//     0 1 0 0
//     0 0 1 0

static zpoly_t ** make_TLI(
	zpoly_t * pcharpoly,
	zppoly_aux_t * paux,
	zpoly_t * proot,
	int chardeg,
	int l,
	int do_galois_cfg)
{
	zpoly_t ** M = zpnla_alloc_matrix(chardeg, chardeg);
	int c;
	int i, j;

	// Zero-fill
	for (i = 0; i < chardeg; i++)
		for (j = 0; j < chardeg; j++)
			M[i][j] = zppoly_from_base_rep(paux->p, 0);

	// Ones on subdiagonal
	for (i = 1; i < chardeg; i++)
		M[i][i-1] = zppoly_from_base_rep(paux->p, 1);

	// Coefficients along top row
	for (i = 0; i < chardeg; i++) {
		c = (paux->p - pcharpoly->coeffs[i]) % paux->p;
		if (do_galois_cfg)
			M[i][chardeg-1] = zppoly_from_base_rep(paux->p, c);
		else
			M[0][chardeg-i-1] = zppoly_from_base_rep(paux->p, c);
	}

	// Minus lambda I
	if (proot) {
		for (i = 0; i < chardeg; i++) {
			M[i][i] = zppolymod_sub(&M[i][i], proot, paux);
		}
	}

	// To the l'th power
	zpnla_mx_power(M, M, chardeg, l, paux);

	return M;
}

// ----------------------------------------------------------------
static int is_eigenvector(
	zpoly_t ** TLI,
	zpoly_t  * eigenvector,
	int chardeg,
	zppoly_aux_t * psplitaux)
{
	zpoly_t  * checkvector = malloc_check(chardeg * sizeof(zpoly_t));
	int k;
	int rv = 1;

	zpnla_mx_times_vr(TLI, eigenvector, checkvector, chardeg, psplitaux);
	for (k = 0; k < chardeg; k++) {
		if (!zppoly_is_zero(&checkvector[k])) {
			rv = 0;
			break;
		}
	}
	free(checkvector);
	return rv;
}

// ----------------------------------------------------------------
static void append_eigenvectors(
	zpoly_t ** TLI,
	int   chardeg,
	zppoly_aux_t * psplitaux,
	zpoly_t * psplitroot,
	zpoly_t ** eigenspace,
	int * pnum_eivrs,
	int   verbose)
{
	zpoly_t ** TLIrr = zpnla_alloc_matrix(chardeg, chardeg);
	zpoly_t  * eigenvector = malloc_check(chardeg * sizeof(zpoly_t));
	int i, j, k;
	int rank;
	int curr_num_ei;
	zpoly_t zero = zppoly_from_base_rep(psplitaux->p, 0);
	zpoly_t one  = zppoly_from_base_rep(psplitaux->p, 1);

	zpnla_mats_copy(TLIrr, TLI, chardeg);
	zpnla_row_echelon_form(TLIrr, psplitaux, chardeg, chardeg);
	if (verbose) {
		printf("Row-echelon form:\n");
		zpnla_print_matrix(TLIrr, chardeg, chardeg, psplitaux);
		printf("\n");
	}

	rank = zpnla_find_rank(TLIrr, chardeg, chardeg);

	// Make sure the rank part of the matrix is identity so that we may
	// easily compute the eigenvectors.
	for (i = 0; i < rank; i++) {
		for (j = 0; j < rank; j++) {
			int ok;
			if (i == j) {
				ok = zppoly_equal(&TLIrr[i][j], &one,
					psplitaux->p);
			}
			else {
				ok = zppoly_equal(&TLIrr[i][j], &zero,
					psplitaux->p);
			}
			if (!ok) {
				fprintf(stderr, "append_eigenvectors: can't "
					"handle matrix.\n");
				exit(1);
			}
		}
	}

	//  a   b   c   d   e
	// --- --- --- --- ---
	//  1   0   0  m03 m04
	//  0   1   0  m13 m14
	//  0   0   1  m23 m24

	// 1st eigenvector: d = 1, e = 0.  [ -m03 -m13 -m23  1  0 ]
	// 2nd eigenvector: d = 0, e = 1.  [ -m04 -m14 -m24  0  1 ]

	// chardeg   = 5
	// rank      = 3
	// curr # ei = 2

	curr_num_ei = chardeg - rank;

	for (k = 0; k < curr_num_ei; k++) {
		for (j = 0; j < rank; j++) {
			zpoly_t neg = zppoly_usub(&TLIrr[j][rank+k],
				psplitaux->p);
			eigenvector[j] = neg;
		}
		for (j = rank; j < chardeg; j++) {
			if (j == rank+k)
				eigenvector[j] = one;
			else
				eigenvector[j] = zero;
		}

		printf("Eigenvalue = ");
		zppolymod_print(psplitroot, psplitaux);
		printf("; ");
		printf("eigenvector = ");
		zpnla_print_vector(eigenvector, chardeg, psplitaux);
		printf("\n");

		if (!is_eigenvector(TLIrr, eigenvector, chardeg, psplitaux)) {
			fprintf(stderr, "fpeisys coding error: not ev.\n");
			exit(1);
		}

		if (*pnum_eivrs >= chardeg) {
			fprintf(stderr,
				"fpeisys coding error: eigenspace overflow.\n");
			exit(1);
		}

		// Concatenate eigenvector to eigenspace
		for (j = 0; j < chardeg; j++)
			eigenspace[*pnum_eivrs][j] = eigenvector[j];
		(*pnum_eivrs)++;
	}

	free(eigenvector);
	zpnla_free_matrix(TLIrr, chardeg, chardeg);
}

// ----------------------------------------------------------------
static void do_one_power(
	zpoly_t * pcharpoly,
	int p,
	zppoly_factor_info_t * pfinfo,
	zppoly_aux_t * psplitaux,
	int l,
	int do_galois_cfg,
	int verbose)
{
	int i, j;
	zppoly_aux_t factoraux;
	int chardeg = pcharpoly->degree;
	int factordeg;
	zpoly_t factorroot, splitroot;
	zpoly_t factorgen, splitgen;
	zpoly_t ** TLI, ** Tl;
	zpoly_t ** eigenspace = zpnla_alloc_matrix(chardeg, chardeg);
	int num_eivrs = 0;
	int Tl_rank = 0;
	int eigenrank = 0;

	printf("------------------------------------------------ l = %d\n", l);

	for (i = 0; i < pfinfo->num_distinct; i++) {

		if (verbose) {
			printf("Factor ");
			zppoly_print(&pfinfo->factor_counts[i].f, p);
			printf(":\n");
		}

		factoraux.p = p;
		factoraux.r = pfinfo->factor_counts[i].f;
		factordeg = factoraux.r.degree;
		factorroot = zppoly_from_coeffs1(1, 0, factoraux.p);
		// To correctly handle degree-one fields:
		factorroot = zppoly_mod(&factorroot, &factoraux.r, p);

		zpn_convert_find_generators(&factoraux, psplitaux,
			&factorgen, &splitgen);
		zpn_convert_scalar(&factorroot, &factoraux, &factorgen,
			&splitroot, psplitaux, &splitgen);

		for (j = 0; j < factordeg; j++) {
			TLI = make_TLI(pcharpoly, psplitaux, &splitroot,
				chardeg, l, do_galois_cfg);

			if (verbose) {
				printf("\n");
				printf("T - ");
				zppolymod_print(&splitroot, psplitaux);
				printf(" I");
				if (l > 1)
					printf("^%d", l);
				printf(":\n");
				zpnla_print_matrix(TLI,
					chardeg, chardeg, psplitaux);
			}

			append_eigenvectors(TLI, chardeg, psplitaux,
				&splitroot, eigenspace, &num_eivrs, verbose);

			// Conjugates are found via the Frobenius automorphism.
			splitroot = zppolymod_power(&splitroot, p, psplitaux);
			zpnla_free_matrix(TLI, chardeg, chardeg);
		}
	}
	printf("\n");

	// Compute eigenspace.
	if (verbose) {
		printf("Eigenspace before row reduction:\n");
		zpnla_print_matrix(eigenspace, num_eivrs, chardeg, psplitaux);
		printf("\n");
	}

	zpnla_row_echelon_form(eigenspace, psplitaux, num_eivrs, chardeg);
	eigenrank = zpnla_find_rank(eigenspace, num_eivrs, chardeg);

	printf("Eigenspace:\n");
	zpnla_print_matrix(eigenspace, eigenrank, chardeg, psplitaux);

	// Compute rank of T^l.
	Tl = make_TLI(pcharpoly, psplitaux, 0, chardeg, l, do_galois_cfg);
	if (verbose) {
		printf("\n");
		printf("T");
		if (l > 1)
			printf("^%d", l);
		printf(":\n");
		zpnla_print_matrix(Tl, chardeg, chardeg, psplitaux);
	}
	zpnla_row_echelon_form(Tl, psplitaux, chardeg, chardeg);
	Tl_rank = zpnla_find_rank(Tl, chardeg, chardeg);

	// Report ranks.
	printf("\n");
	printf("Char poly degree = %d; rank of eigenspace = %d; "
		"rank of T",
		chardeg, eigenrank);
	if (l > 1)
		printf("^%d", l);
	printf(" = %d.\n", Tl_rank);

	zpnla_free_matrix(Tl, chardeg, chardeg);
	zpnla_free_matrix(eigenspace, chardeg, chardeg);
}

// ----------------------------------------------------------------
static void do_one_poly(
	zpoly_t * pcharpoly,
	int p,
	int do_galois_cfg,
	int verbose)
{
	zppoly_factor_info_t finfo;
	zppoly_aux_t splitaux;
	int l;

	zppoly_factor(pcharpoly, p, &finfo);
	my_print_factors(pcharpoly, p, &finfo);
	splitaux = get_split_aux(pcharpoly, p, &finfo);

	for (l = 1; l <= pcharpoly->degree; l++)
		do_one_power(pcharpoly, p, &finfo, &splitaux,
			l, do_galois_cfg, verbose);
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {r}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int p;
	zpoly_t charpoly;
	int argi, argb = 1;
	int do_galois_cfg = 0;
	int verbose = 0;

	for (argi = 1; argi < argc; argi++) {
		if (argv[argi][0] != '-')
			break;
		else if (strcmp(argv[argi], "-g") == 0)
			do_galois_cfg = 1;
		else if (strcmp(argv[argi], "-v") == 0)
			verbose = 1;
		else
			usage(argv[0]);
		argb++;
	}

	if ((argc - argb) < 1)
		usage(argv[0]);

	if (sscanf(argv[argb], "%d", &p) != 1) {
		fprintf(stderr, "%s: Couldn't scan \"%s\" as p.\n",
			argv[0], argv[argb]);
		exit(1);
	}

	for (argi = argb+1; argi < argc; argi++) {
		if (!zppoly_scan(&charpoly, argv[argi], p)) {
			fprintf(stderr,
				"%s: Couldn't scan \"%s\" as charpoly.\n",
				argv[0], argv[argi]);
			exit(1);
		}
		do_one_poly(&charpoly, p, do_galois_cfg, verbose);
		if ((argc - argb) > 2)
			printf("\n");
	}

	return 0;
}
