#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "power.h"
#include "zpnpoly_factor.h"

#define MAX_IRRED_FACTORS 4116 // See f2nsplit.c.

typedef struct _opts_t {
	int show_trace;
	int show_norm;
	int show_chpoly;
	int show_orders;
	int show_elements;
	int show_logs;
} opts_t;

static void find_irred_factors(
	unsigned    p,
	unsigned    n,
	zpoly_t   * pirred_factors,
	unsigned  * pnum_irred_factors,
	zpoly_t   * pdirred_factors,
	unsigned  * pnum_dirred_factors,
	unsigned    max_irred_factors);

static void split_poly(
	zpoly_t   * prj,
	zpoly_t   * irred_factors,
	unsigned    num_irred_factors,
	zpoly_t   * dirred_factors,
	unsigned    num_dirred_factors,
	int         p,
	int         n,
	int         pn,
	opts_t    * popts);

static void split_factor(
	zpnpoly_t * pf,
	zpoly_t   * pr,
	int         p,
	int         n,
	zpoly_t   * pg,
	opts_t    * popts);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {n}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int        argi;
	int        p, n, pn;
	static zpoly_t irred_factors[MAX_IRRED_FACTORS]; // xxx malloc
	static zpoly_t dirred_factors[MAX_IRRED_FACTORS];
	unsigned   num_irred_factors = 0;
	unsigned   num_dirred_factors = 0;
	zpoly_t    rj;
	unsigned   j;
	opts_t     opts;

	opts.show_trace  = 0;
	opts.show_norm   = 0;
	opts.show_chpoly = 0;
	opts.show_orders = 0;
	opts.show_elements = 1;
	opts.show_logs   = 0;

	if (argc < 3)
		usage(argv[0]);

	argi = 1;
	while (argv[argi][0] == '-') {
		if (strcmp(argv[argi], "-t") == 0)
			opts.show_trace = 1;
		else if (strcmp(argv[argi], "-n") == 0)
			opts.show_norm = 1;
		else if (strcmp(argv[argi], "-c") == 0)
			opts.show_chpoly = 1;
		else if (strcmp(argv[argi], "-o") == 0)
			opts.show_orders = 1;
		else if (strcmp(argv[argi], "-l") == 0) {
			opts.show_elements = 0;
			opts.show_logs = 1;
		}
		else if (strcmp(argv[argi], "-L") == 0) {
			opts.show_elements = 1;
			opts.show_logs = 1;
		}
		else
			usage(argv[0]);
		argi++;
	}

	if (sscanf(argv[argi], "%u", &p) != 1)
		usage(argv[0]);
	argi++;

	for (; argi < argc; argi++) {

		if (sscanf(argv[argi], ":%u", &n) == 1) {
			pn = calc_power(p, n);
			find_irred_factors(p, n, irred_factors, &num_irred_factors,
				dirred_factors, &num_dirred_factors, MAX_IRRED_FACTORS);
			for (j = 0; j < num_dirred_factors; j++) {
				rj = dirred_factors[j];
				split_poly(&rj, irred_factors, num_irred_factors,
					dirred_factors, num_dirred_factors, p, n, pn, &opts);
			}
		}
		else if (zppoly_scan(&rj, argv[argi], p)) {
			n = rj.degree;
			pn = calc_power(p, n);
			find_irred_factors(p, n, irred_factors, &num_irred_factors,
				dirred_factors, &num_dirred_factors, MAX_IRRED_FACTORS);
			split_poly(&rj, irred_factors, num_irred_factors, dirred_factors,
				num_dirred_factors, p, n, pn, &opts);
		}
		else {
			usage(argv[0]);
		}
	}

	return 0;
}

// ----------------------------------------------------------------
static void find_irred_factors(
	unsigned    p,
	unsigned    n,
	zpoly_t   * pirred_factors,
	unsigned  * pnum_irred_factors,
	zpoly_t   * pdirred_factors,
	unsigned  * pnum_dirred_factors,
	unsigned    max_irred_factors)
{
	unsigned num_irred_factors = 0;
	unsigned num_dirred_factors = 0;
	unsigned d, ilo, ihi, i;
	zpoly_t f;

	for (d = 1; d <= n; d++) {
		if ((n % d) != 0)
			continue;

		ilo = calc_power(p, d);
		ihi = 2 * ilo - 1;
		for (i = ilo; i <= ihi; i++) {
			f = zppoly_from_base_rep(p, i);
			if (!zppoly_is_irreducible(&f, p))
				continue;

			if (num_irred_factors >= MAX_IRRED_FACTORS) {
				fprintf(stderr, "Too many factors.\n");
				exit(1);
			}
			pirred_factors[num_irred_factors] = f;
			num_irred_factors++;

			if (d == n) {
				pdirred_factors[num_dirred_factors] = f;
				num_dirred_factors++;
			}
		}
	}

	*pnum_irred_factors = num_irred_factors;
	*pnum_dirred_factors = num_dirred_factors;
}

// ----------------------------------------------------------------
static void split_poly(
	zpoly_t   * prj,
	zpoly_t   * irred_factors,
	unsigned    num_irred_factors,
	zpoly_t   * dirred_factors,
	unsigned    num_dirred_factors,
	int         p,
	int         n,
	int         pn,
	opts_t    * popts)
{
	zpoly_t    ri;
	zpoly_t    g = zppoly_from_base_rep(1, 0);
	zppoly_aux_t aux;
	zpnpoly_t ext_factor;
	unsigned   i;

	// Just for columnar printing.
	aux.p = p;
	aux.r = zppoly_from_base_rep(p, calc_power(p, n+1));

	printf("mod ");
	//printf("%c ", zppoly_is_primitive(prj, p, n) ? 'o' : '-');
	zppoly_print(prj, p);

	if (popts->show_logs) {
		zppoly_aux_t auxj;
		auxj.p = p;
		auxj.r = *prj;
		g = zppolymod_find_generator(&auxj);
		printf(" g=");
		zppoly_print(&g, p);
	}
	printf(":\n");

	for (i = 0; i < num_irred_factors; i++) {
		ri = irred_factors[i];

		ext_factor = zpnpoly_extend_zppoly(&ri);

		printf("%c ", zppoly_is_primitive(&ri, p, n) ? 'o' : '-');
		zppolymod_print(&ri, &aux);
		printf(" = ");

		split_factor(&ext_factor, prj, p, n, &g, popts);
		printf("\n");
		fflush(stdout);
	}
	printf("\n");
}

// ----------------------------------------------------------------
// This uses Berlekamp's algorithm on Fpn[x].
static void split_factor(
	zpnpoly_t * pf,
	zpoly_t   * pr,
	int         p,
	int         n,
	zpoly_t   * pg,
	opts_t    * popts)
{
	zppoly_aux_t aux;
	zpoly_t root;
	static zpnpoly_factor_info_t finfo;
	int i;
	int root_order = 0;

	aux.p = p;
	aux.r = *pr;

	zpnpoly_factor(pf, &aux, &finfo);

	for (i = 0; i < finfo.num_distinct; i++) {
		if (finfo.factor_counts[i].count != 1) {
			fprintf(stderr,
			"fpnsplits: multiple factors in universal poly.\n");
			exit(1);
		}
		if (finfo.factor_counts[i].f.degree != 1) {
			fprintf(stderr,
			"fpnsplits: non-linear factors in universal poly.\n");
			exit(1);
		}
		root = zppoly_usub(&finfo.factor_counts[i].f.coeffs[0], p);
		printf(" ");

		if (popts->show_elements) {
			zppolymod_print(&root, &aux);
		}

		if (popts->show_logs) {
			if (popts->show_elements)
				printf("(");
			else
				printf(" ");
			if (zppoly_is_zero(&root))
				printf("--");
			else
				printf("%2u", zppolymod_log(&root, pg, &aux));
			if (popts->show_elements)
				printf(")");
		}

		if (i == 0)
			root_order = zppolymod_element_order(&root, &aux);
	}

	for (i = 0; i < (n - finfo.num_distinct); i++)
		printf(" .");

	if (popts->show_trace) {
		// Sum of conjugates
		zpoly_t trace = zppoly_from_coeffs0(0, p);
		for (i = 0; i < finfo.num_distinct; i++) {
			root = zppoly_usub(&finfo.factor_counts[i].f.coeffs[0], p);
			trace = zppoly_add(&trace, &root, p);
		}
		printf(" tr=");
		zppoly_print(&trace, p);
	}
	if (popts->show_norm) {
		// Product of conjugates
		zpoly_t norm = zppoly_from_coeffs0(1, p);
		for (i = 0; i < finfo.num_distinct; i++) {
			root = zppoly_usub(&finfo.factor_counts[i].f.coeffs[0], p);
			norm = zppolymod_mul(&norm, &root, &aux);
		}
		printf(" nm=");
		zppoly_print(&norm, p);
	}
	if (popts->show_chpoly) {
		// Characteristic polynomial not implemented yet.
		printf(" ch=?");
	}

	if (popts->show_orders)
		printf(" o= %3d", root_order);
}

//// ----------------------------------------------------------------
//// This is a brute-force search.
//static void split_factor(
//	zpnpoly_t * pf,
//	zpoly_t   * pr,
//	int         p,
//	int         n,
//	zpoly_t   * pg,
//	opts_     * popts)
//{
//	int pn = calc_power(p, n);
//	int i;
//	zpnpoly_t linear_factor, remnant, quot, rem;
//	int root_order = 0;
//	int nfactors = 0;
//	zpoly_t z;
//	zppoly_aux_t aux;
//
//	aux.p = p;
//	aux.r = *pr;
//
//	// Try each possible linear factor (X - a) for a in F_2^n.
//	remnant = *pf;
//	linear_factor.degree = 1;
//	linear_factor.coeffs[1] = zppoly_from_base_rep(p, 1);
//	for (i = 0; i < pn; i++) {
//		z = zppoly_from_base_rep(p, i);
//		linear_factor.coeffs[0] = z;
//
//		zpnpoly_quot_and_rem(&remnant, &linear_factor,
//			&quot, &rem, &aux);
//
//		// Try the next factor.
//		if (!zpnpoly_is_zero(&rem))
//			continue;
//
//		nfactors++;
//		printf(" ");
//		if (popts->show_logs)
//			printf(" %2u", zppolymod_log(&z, pg, &aux));
//		else
//			zppolymod_print(&z, &aux);
//
//		root_order = zppolymod_element_order(&z, &aux);
//
//		remnant = quot;
//
//		if (zpnpoly_is_one(&remnant)) {
//			break;
//		}
//	}
//	for (i = 0; i < (n - nfactors); i++)
//		printf(" .");
//	if (popts->show_orders)
//		printf(" o= %3d", root_order);
//
//	if (!zpnpoly_is_one(&remnant)) {
//		printf("ERROR:  remnant ");
//			zpnpoly_print(&remnant, &aux);
//			printf("\n");
//	}
//}
