#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "power.h"
#include "zpnpoly.h"

#define MAX_IRRED_FACTORS 4116 // See f2nsplit.c.

static void find_irred_factors(
	unsigned    p,
	unsigned    n,
	zpoly_t   * pirred_factors,
	unsigned  * pnum_irred_factors,
	zpoly_t   * pdirred_factors,
	unsigned  * pnum_dirred_factors,
	unsigned    max_irred_factors);

static void split_factor(
	zpnpoly_t * pf,
	zpoly_t   * pr,
	int         p,
	int         n,
	int         show_logs,
	zpoly_t   * pg);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {n}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int        p, n, pn;
	static zpoly_t    irred_factors[MAX_IRRED_FACTORS]; // xxx malloc
	static zpoly_t    dirred_factors[MAX_IRRED_FACTORS];
	unsigned   num_irred_factors = 0;
	unsigned   num_dirred_factors = 0;
	zpoly_t    ri, rj;
	zpoly_t    g = zppoly_from_base_rep(1, 0);
	zppoly_aux_t aux;
	zpnpoly_t ext_factor;
	unsigned   i, j;
	int        show_logs = 0;

	if (argc < 3)
		usage(argv[0]);
	if (argc > 3)
		show_logs = 1;

	if (sscanf(argv[1], "%u", &p) != 1)
		usage(argv[0]);
	if (sscanf(argv[2], "%u", &n) != 1)
		usage(argv[0]);

	pn = calc_power(p, n);

	// Just for columnar printing.
	aux.p = p;
	aux.r = zppoly_from_base_rep(p, calc_power(p, n+1));

	find_irred_factors(p, n, irred_factors, &num_irred_factors,
		dirred_factors, &num_dirred_factors, MAX_IRRED_FACTORS);

	for (j = 0; j < num_dirred_factors; j++) {
		rj = dirred_factors[j];

		printf("mod ");
		printf("%c ", zppoly_is_primitive(&rj, p, n) ? 'o' : '-');
		zppoly_print(&rj, p);

		if (show_logs) {
			zppoly_aux_t auxj;
			auxj.p = p;
			auxj.r = rj;
			g = zppolymod_find_generator(&auxj);
			printf(" g=");
			zppoly_print(&g, p);
		}
		printf(":\n");


		for (i = 0; i < num_irred_factors; i++) {
			ri = irred_factors[i];

			ext_factor = zpnpoly_extend_zppoly(&ri);

			printf("%c ",
				zppoly_is_primitive(&ri, p, n) ? 'o' : '-');
			zppolymod_print(&ri, &aux);
			printf(" = ");

			split_factor(&ext_factor, &rj, p, n, show_logs, &g);
			printf("\n");
			fflush(stdout);
		}
		printf("\n");
	}

	return 0;
}

// ----------------------------------------------------------------
static void find_irred_factors(
	unsigned    p,
	unsigned    n,
	zpoly_t   * pirred_factors,
	unsigned  * pnum_irred_factors,
	zpoly_t   * pdirred_factors,
	unsigned  * pnum_dirred_factors,
	unsigned    max_irred_factors)
{
	unsigned num_irred_factors = 0;
	unsigned num_dirred_factors = 0;
	unsigned d, ilo, ihi, i;
	zpoly_t f;

	for (d = 1; d <= n; d++) {
		if ((n % d) != 0)
			continue;

		ilo = calc_power(p, d);
		ihi = 2 * ilo - 1;
		for (i = ilo; i <= ihi; i++) {
			f = zppoly_from_base_rep(p, i);
			if (!zppoly_is_irreducible(&f, p))
				continue;

			if (num_irred_factors >= MAX_IRRED_FACTORS) {
				fprintf(stderr, "Too many factors.\n");
				exit(1);
			}
			pirred_factors[num_irred_factors] = f;
			num_irred_factors++;

			if (d == n) {
				pdirred_factors[num_dirred_factors] = f;
				num_dirred_factors++;
			}
		}
	}

	*pnum_irred_factors = num_irred_factors;
	*pnum_dirred_factors = num_dirred_factors;
}

// ----------------------------------------------------------------
static void split_factor(
	zpnpoly_t * pf,
	zpoly_t   * pr,
	int         p,
	int         n,
	int         show_logs,
	zpoly_t   * pg)
{
	int pn = calc_power(p, n);
	int i;
	zpnpoly_t linear_factor, remnant, quot, rem;
	int root_order = 0;
	int nfactors = 0;
	zpoly_t z, root;
	zppoly_aux_t aux;

	aux.p = p;
	aux.r = *pr;

	// Try each possible linear factor (X - a) for a in F_2^n.
	remnant = *pf;
	linear_factor.degree = 1;
	linear_factor.coeffs[1] = zppoly_from_base_rep(p, 1);
	for (i = 0; i < pn; i++) {
		z = zppoly_from_base_rep(p, i);
		linear_factor.coeffs[0] = z;

		zpnpoly_quot_and_rem(&remnant, &linear_factor,
			&quot, &rem, &aux);

		// Try the next factor.
		if (!zpnpoly_is_zero(&rem))
			continue;

		nfactors++;
		printf(" ");
		root = zppoly_usub(&z, p);
		if (show_logs)
			printf(" %2u", zppolymod_log(&root, pg, &aux));
		else
			zppolymod_print(&root, &aux);

		root_order = zppolymod_element_order(&root, &aux);

		remnant = quot;

		if (zpnpoly_is_one(&remnant)) {
			break;
		}
	}
	for (i = 0; i < (n - nfactors); i++)
		printf(" .");
	printf(" [%3d]", root_order);

	if (!zpnpoly_is_one(&remnant)) {
		printf("ERROR:  remnant ");
			zpnpoly_print(&remnant, &aux);
			printf("\n");
	}
}
