#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "group_opts.h"

// ----------------------------------------------------------------
int group_parse_opts(
	int                 argc,   // input
	char             ** argv,   // input
	group_opts_spec_t * pospec, // input
	group_opts_t      * popts)  // output
{
	int    argi;

	// Set defaults.
	popts->verbosity   = pospec->default_verbosity;
	popts->echo_inputs = 0;
	popts->cmd_paramc  = 0;
	popts->cmd_paramv  = 0;
	popts->group_spec_desc = 0;
	popts->pgspec      = 0;
	popts->auxc        = 0;
	popts->auxv        = 0;
	popts->file_namec  = 0;
	popts->file_namev  = 0;

	// Parse dash options.
	for (argi = 1; argi < argc; argi++) {
		if (argv[argi][0] != '-')
			break;
		if (strcmp(argv[argi], "--") == 0) {
			// Allow e.g. "-1" as a non-option parameter.
			argi++;
			break;
		}

		if (strcmp(argv[argi], "--help") == 0) {
			return 0;
		}
		else if (strcmp(argv[argi], "--list") == 0) {
			list_group_specs();
			exit(0);
		}
		else if (strcmp(argv[argi], "-v") == 0) {
			popts->verbosity++;
		}
		else if (strcmp(argv[argi], "-e") == 0) {
			popts->echo_inputs = 1;
		}
		else {
			fprintf(stderr,
				"%s: unrecognized option \"%s\".\n",
				argv[0], argv[argi]);
			return 0;
		}
	}

	// Command parameters, e.g. an integer k for a k-torsion subgroup,
	// or a group element g to generate a cyclic subgroup.
	//
	// argc-argi is the number of remaining arguments on the command line.
	if ((argc-argi) < pospec->num_cmd_params) {
		fprintf(stderr, "%s: Need %s.\n",
			argv[0], pospec->cmd_param_desc);
		return 0;
	}
	popts->cmd_paramc = pospec->num_cmd_params;
	popts->cmd_paramv = &argv[argi];
	argi += pospec->num_cmd_params;

	// One group spec -- for now, since all commands at this time need
	// a single group spec.  To make a direct product of different
	// types, e.g. (Z_7 x Q_8), I'd need multiple group specs.
	if ((argc-argi) < 1) {
		fprintf(stderr,
			"%s: Need 1 group spec.\n",
			argv[0]);
		list_group_specs();
		return 0;
	}
	popts->group_spec_desc = argv[argi];
	argi++;

	// Group aux (usually 0 or 1).  E.g. the m in Z_m, the polynomial
	// r for z2ipoly_t, n in S_n; none for V4, Q8, etc.
	//
	// The number of auxes (0 or 1) depends on the group spec, so we
	// need to look that up now.

	if (!find_group_spec(popts->group_spec_desc, &popts->pgspec)) {
		fprintf(stderr, "No such group class \"%s\".\n",
			popts->group_spec_desc);
		return 0;
	}
	if ((argc-argi) < popts->pgspec->naux) {
		fprintf(stderr,
			"%s: Need %d aux string%s.\n",
			argv[0],
			popts->pgspec->naux,
			popts->pgspec->naux == 1 ? "" : "s");
		return 0;
	}
	popts->auxc = popts->pgspec->naux;
	popts->auxv = &argv[argi];
	argi += popts->pgspec->naux;

	// The rest are file names.
	if ((argc-argi) == 0) {
		// No input files; default group filler should be called.
		popts->file_namec = 0;
		popts->file_namev = 0;
	}
	else if ((argc-argi) > pospec->num_file_names) {
		fprintf(stderr, "%s: Need %d file name%s.\n",
			argv[0],
			pospec->num_file_names,
			pospec->num_file_names == 1 ? "" : "s");
		return 0;
	}
	else {
		popts->file_namec = argc-argi;
		popts->file_namev = &argv[argi];
	}

	return 1;
}

// ----------------------------------------------------------------
void group_opts_usage(
	char * argv0,
	char * cmd_param_desc)
{
	fprintf(stderr,
	"Usage: %s [options] %s%s%s{group spec} [group aux] [file name]\n",
			argv0,
			cmd_param_desc ? "{" : "",
			cmd_param_desc ? cmd_param_desc : "",
			cmd_param_desc ? "} " : "");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  -v to increase verbosity (may be used more "
		"than once)\n");
	fprintf(stderr, "  -e to echo inputs\n");
	fprintf(stderr, "  --list to list available groups\n");
	fprintf(stderr, "  --help to see this message\n");

	exit(1);
}

// ----------------------------------------------------------------
void print_group_opts(
	group_opts_t * popts)
{
	int i;
	printf("verbosity   = %d\n", popts->verbosity);
	printf("echo_inputs = %d\n", popts->echo_inputs);

	printf("cmd_paramc  = %d\n", popts->cmd_paramc);
	for (i = 0; i < popts->cmd_paramc; i++)
		printf("  <<%s>>\n", popts->cmd_paramv[i]);

	printf("group_spec_desc = %s\n", popts->group_spec_desc);

	printf("auxc        = %d\n", popts->auxc);
	for (i = 0; i < popts->auxc; i++)
		printf("  <<%s>>\n", popts->auxv[i]);

	printf("file_namec  = %d\n", popts->file_namec);
	for (i = 0; i < popts->file_namec; i++)
		printf("  <<%s>>\n", popts->file_namev[i]);

}
