#include <stdio.h>
#include <stdlib.h>
#include "lexan.h"
#include "emit.h"
#include "symtab.h"
#include "stack.h"
#include "error.h"
#include "common_math.h"
#include "cmps.h"

static char * op_symbol_to_name(int t);
static int emit_print = 0;

static void pr_emit(int t, int tval, void * tbuf, symtab_t * pst,
	stack_t * ps, opts_t * popts);

// ----------------------------------------------------------------
void ev_emit(int t, int tval, void * tbuf, symtab_t * pst,
	stack_t * ps, opts_t * popts)
{
	char a[MAX_ATOM_SIZE];
	char b[MAX_ATOM_SIZE];
	char c[MAX_ATOM_SIZE];
	binary_op_func_t * pbop = 0;
	unary_op_func_t  * puop = 0;
	int is_uop = 0;

	if (emit_print) {
		pr_emit(t, tval, tbuf, pst, ps, popts);
		return;
	}

	if (t == L_NUM) {
		// Pre-rectification is inappropriate for modular types
		// (don't want exponents taken mod m in exponentiation).
		//if (popts->pops->prectifier)
			//popts->pops->prectifier(tbuf);

		// xxx if dynalloc, here we are pushing a pointer.
		(void)stack_push(ps, tbuf, popts->pops->atom_size_in_bytes);
		return;
	}

	if (t == L_LVALUE) {
		// tval is symbol-table index.
		(void)stack_push(ps, &tval, sizeof(tval));
		return;
	}

	if (t == L_ID) {
		// tval is symbol-table index.
		if (!st_get_val(pst, tval, c)) {
			fprintf(stderr, "Unassigned variable %s.\n",
				st_name(pst, tval));
			exit(1);
		}
		(void)stack_push(ps, c, popts->pops->atom_size_in_bytes);
		return;
	}

	if (t == '=') {
		// Example:  User typed "c = 3".  We already pushed c as an
		// lvalue (namely, pushed c's symbol-table index), and we
		// already pushed the three.  Here, we need to pop off the
		// 3, then pop off the stidx, then re-push the 3.
		int lval_stidx;
		(void)stack_pop(ps, c, popts->pops->atom_size_in_bytes);
		(void)stack_pop(ps, &lval_stidx, sizeof(lval_stidx));
		st_set_val(pst, lval_stidx, c);
		(void)stack_push(ps, c, popts->pops->atom_size_in_bytes);
		return;
	}

	if (t == '!') {
		(void)stack_pop(ps, a, popts->pops->atom_size_in_bytes);
		if (!common_factorial(c, a, popts)) {
			fprintf(stderr,
			"Invalid operator %s for type %s.\n",
				op_symbol_to_name(t),
				popts->pops->atom_desc);
			exit(1);
		}
		(void)stack_push(ps, c, popts->pops->atom_size_in_bytes);
		return;
	}

	is_uop = 1;
	switch (t) {
	case L_UNOT:   puop = popts->pops->punot_op; break;
	case L_UPLUS:  puop = popts->pops->puadd_op; break;
	case L_UMINUS: puop = popts->pops->pusub_op; break;
	case L_SQRT:   puop = popts->pops->psqrt_op; break;
	case L_EXP:    puop = popts->pops->pexp_op;  break;
	case L_LOG:    puop = popts->pops->plog_op;  break;
	case L_SIN:    puop = popts->pops->psin_op;  break;
	case L_COS:    puop = popts->pops->pcos_op;  break;
	case L_TAN:    puop = popts->pops->ptan_op;  break;
	case L_ATAN:   puop = popts->pops->patan_op; break;
	case L_CONJ:   puop = popts->pops->pconj_op; break;
	case L_MAG:    puop = popts->pops->pmag_op;  break;
	case L_PHZ:    puop = popts->pops->pphz_op;  break;
	case L_ORD:    puop = popts->pops->pord_op;  break;
	default:       puop = 0; is_uop = 0; break;
	}

	if (is_uop) {
		if (puop == 0) {
			fprintf(stderr,
				"Invalid operator %s for type %s.\n",
					op_symbol_to_name(t),
					popts->pops->atom_desc);
			exit(1);
		}
		(void)stack_pop(ps, a, popts->pops->atom_size_in_bytes);
		puop(c, a, popts->aux_buf);
		(void)stack_push(ps, c, popts->pops->atom_size_in_bytes);
		return;
	}

	(void)stack_pop(ps, b, popts->pops->atom_size_in_bytes);
	(void)stack_pop(ps, a, popts->pops->atom_size_in_bytes);

	if (t == L_GCD) {
		if (!common_gcd(c, a, b, popts)) {
			fprintf(stderr,
			"Invalid operator %s for type %s.\n",
				op_symbol_to_name(t),
				popts->pops->atom_desc);
			exit(1);
		}
		(void)stack_push(ps, c, popts->pops->atom_size_in_bytes);
		return;
	}

	if (t == L_CHOOSE) {
		if (!common_choose(c, a, b, popts)) {
			fprintf(stderr,
			"Invalid operator %s for type %s.\n",
				op_symbol_to_name(t),
				popts->pops->atom_desc);
			exit(1);
		}
		(void)stack_push(ps, c, popts->pops->atom_size_in_bytes);
		return;
	}

	switch(t) {

	case '|':   pbop = popts->pops->por_op ; break;
	case '^':   pbop = popts->pops->pxor_op; break;
	case '&':   pbop = popts->pops->pand_op; break;
	case L_LSH: pbop = popts->pops->plsh_op; break;
	case L_RSH: pbop = popts->pops->prsh_op; break;

	case '+':   pbop = popts->pops->padd_op; break;
	case '-':   pbop = popts->pops->psub_op; break;
	case '*':   pbop = popts->pops->pmul_op; break;
	case '/':   pbop = popts->pops->pdiv_op; break;
	case '%':   pbop = popts->pops->pmod_op; break;
	case L_POW: pbop = popts->pops->ppow_op; break;
	case ',':   pbop = popts->pops->pcomma_op; break;


	default:
		printf("token %d, tokenval %d\n", t, tval);
	}

	if (pbop == 0) {
		fprintf(stderr,
		"Invalid operator %s for type %s.\n",
			op_symbol_to_name(t),
			popts->pops->atom_desc);
		exit(1);
	}

	pbop(c, a, b, popts->aux_buf);

	(void)stack_push(ps, c, popts->pops->atom_size_in_bytes);
}

// ----------------------------------------------------------------
static void pr_emit(int t, int tval, void * tbuf, symtab_t * pst,
	stack_t * ps, opts_t * popts)
{
	switch(t) {

	case L_NUM:
		printf(" ");
		popts->pops->pprint_func(stdout, tbuf, popts->pvtopts,
			popts->aux_buf);
		break;

	case L_LVALUE:
		// tval is symbol-table index.
		printf(" %s", st_name(pst, tval));
		break;

	case L_ID:
		// tval is symbol-table index.
		printf(" %s", st_name(pst, tval));
		break;

	case L_UNOT:   printf(" ~"); break;
	case L_UPLUS:  printf(" +"); break;
	case L_UMINUS: printf(" -"); break;

	case L_SQRT:   printf(" sqrt"); break;
	case L_EXP:    printf(" exp");  break;
	case L_LOG:    printf(" log");  break;
	case L_SIN:    printf(" sin");  break;
	case L_COS:    printf(" cos");  break;
	case L_TAN:    printf(" tan");  break;
	case L_ATAN:   printf(" atan"); break;
	case L_CONJ:   printf(" conj"); break;
	case L_MAG:    printf(" mag");  break;
	case L_PHZ:    printf(" phz");  break;
	case L_ORD:    printf(" ord");  break;

	case '|': case '^': case '&': case '+': case '-':
	case '*': case '/': case '%': case '=': case ',':
		printf(" %c", t);
		break;

	case L_LSH: printf(" <<"); break;
	case L_RSH: printf(" >>"); break;
	case L_POW: printf(" **"); break;
	case L_GCD: printf(" gcd"); break;
	case L_CHOOSE: printf(" choose"); break;

	default:
		printf(" (token %d, tval %d)\n", t, tval);
	}
}

// ----------------------------------------------------------------
static char * op_symbol_to_name(int t)
{
	switch(t) {
	case '|'     : return "|";      break;
	case '^'     : return "^";      break;
	case '&'     : return "&";      break;
	case L_LSH   : return "<<";     break;
	case L_RSH   : return ">>";     break;
	case '+'     : return "+";      break;
	case '-'     : return "-";      break;
	case '*'     : return "*";      break;
	case '/'     : return "/";      break;
	case '%'     : return "%";      break;
	case ','     : return ",";      break;
	case L_UMINUS: return "unary -";break;
	case L_UPLUS : return "unary +";break;
	case L_UNOT  : return "unary ~";break;
	case L_POW   : return "**";     break;
	case L_GCD   : return "gcd";    break;
	case L_CHOOSE: return "choose"; break;
	case L_SQRT  : return "sqrt";   break;
	case L_EXP   : return "exp";    break;
	case L_LOG   : return "log";    break;
	case L_SIN   : return "sin";    break;
	case L_COS   : return "cos";    break;
	case L_TAN   : return "tan";    break;
	case L_ATAN  : return "atan";   break;
	case L_CONJ  : return "conj";   break;
	case L_MAG   : return "mag";    break;
	case L_PHZ   : return "phz";    break;
	case L_ORD   : return "ord";    break;
	case '!'     : return "!";      break;
	default      : return "??";     break;
	}
}

// ----------------------------------------------------------------
void emit_set_print(int onoff)
{
	emit_print = onoff;
}
int emit_get_print(void)
{
	return emit_print;
}
