// ================================================================
// -4**2 means -(4**2), *not* (-4)**2.  Thus exponentiation has a higher
// precedence than unary minus.
//
// a*b+c*d  = (a*b)+(c*d)    * before +
// a*b**2   = (a)*(b**2)    ** before *
// -a**2    = -(a**2)       ** before u-
// -a*b     = -(a*b)         * before u-
//
// a,b*c,d  = (a,b)*(c,d)    , before *
// a,b/c,d  = (a,b)/(c,d)    , before /
// a,b-c,d  = (a,b)-(c,d)    , before b-
// -a, b    = (-a,b)        u- before ,
//
// Grammar including bitwise operations, using C's precedence, and single
// assignments (I don't yet know how to implement right associativity):
// S  => SA ;
// SA => E | id = E
// E  => O | O \| O
// O  => X | X ^ X
// X  => A | A & A
// A  => H | H << H | H >> H
// H  => T | T + T | T - T
// T  => F | F * F | F / F | F % F
// F  => C | C , C
// C  => B | - B | + B | ~ B
// B  => U | U! (this is the factorial operator)
// U  => P | P ** P
// P  => E | (E) | num | id
// ================================================================

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "parse.h"
#include "lexan.h"
#include "symtab.h"
#include "emit.h"
#include "stack.h"
#include "error.h"

#ifdef PARSE_DEBUG
static int gl_depth = 0;
//#define ENTER printf("enter %s\n", __FUNCTION__)
//#define EXIT  printf("exit  %s\n", __FUNCTION__)
#define ENTER enter_func(__FUNCTION__)
#define EXIT  exit_func(__FUNCTION__)

static void enter_func(char * func)
{
	int i;
	for (i = 0; i < gl_depth; i++)
		printf(" ");
	printf("enter %s\n", func);
	gl_depth++;
}

static void exit_func(char * func)
{
	int i;
	gl_depth--;
	for (i = 0; i < gl_depth; i++)
		printf(" ");
	printf("exit  %s\n", func);
}

#else
#define ENTER
#define EXIT
#endif

static void SA(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs);

static void E(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs,
	int          lhs_done_already);

static void O(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs);

static void X(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs);

static void A(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs);

static void H(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs);

static void T(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs);

static void C(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs);

static void F(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs);

static void B(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs);

static void U(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs);

static void P(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs);

static void match(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          t,
	int          in_pow_rhs);

// ----------------------------------------------------------------
void parse(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts)
{
	char result[MAX_ATOM_SIZE];
	int in_pow_rhs = 0;

	ENTER;

	if (isatty(fileno(stdin))) {
		printf("kalk> ");
		fflush(stdout);
	}

	lexan(&pc->lookahead1, pst, infp, popts, in_pow_rhs);
	pc->lookahead2.token = L_DEFER;
	while (pc->lookahead1.token != L_DONE) {

		if (!popts->req_semi) {
			if (pc->lookahead1.token == '\n') {

				if (emit_get_print()) {
					printf("\n");
				}
				if (isatty(fileno(stdin))) {
					printf("kalk> ");
					fflush(stdout);
				}
				match(pc, pst, ps, infp, popts, '\n',
					in_pow_rhs);
				continue;
			}
		}

		SA(pc, pst, ps, infp, popts, in_pow_rhs);

		if (emit_get_print()) {
			continue;
		}

		(void)stack_pop(ps, result,
			popts->pops->atom_size_in_bytes);

		// Post-rectification is appropriate for modular types
		// (e.g.  user types just "13" without arguments, with
		// modulus 11).  Pre-rectification is inappropriate for
		// modular types (don't want exponents taken mod m in
		// exponentiation).

		if (popts->pops->prectifier)
			popts->pops->prectifier(result, popts->aux_buf);

		popts->pops->pprint_func(stdout, result, popts->pvtopts,
			popts->aux_buf);
		printf("\n");

		if (isatty(fileno(stdin))) {
			printf("kalk> ");
			fflush(stdout);
		}

		if (popts->req_semi)
			match(pc, pst, ps, infp, popts, ';',
				in_pow_rhs);
		else
			match(pc, pst, ps, infp, popts, '\n',
				in_pow_rhs);
	}
	EXIT;
	if (isatty(fileno(stdin)))
		printf("\n");
}

// ----------------------------------------------------------------
static void SA(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs)
{
	lexctx_t idctx;

	ENTER;

	// emit  NUM
	// match NUM
	// match op
	// emit  NUM
	// match NUM
	// emit  op

	// SA =>  E  |  id = E

	// 3       OK
	// l1 2

	// 3  +  4   OK
	// l1 l2

	// 3  =  4   not OK
	// l1 l2

	// c         OK (c is an rvalue)
	// l1 l2

	// c  (EOF)  not OK (missing statement terminator)
	// l1 l2

	// c  +  3   OK (c is an rvalue)
	// l1 l2

	// c  =  3   OK (c is an lvalue)
	// l1 l2

	if (pc->lookahead1.token != L_ID) {
		E(pc, pst, ps, infp, popts, in_pow_rhs, 0);
		EXIT;
		return;
	}

	if (pc->lookahead2.token == L_DEFER) {
		pc->lookahead2 = pc->lookahead1;
			// Retain line # & comment state
		lexan(&pc->lookahead2, pst, infp, popts, in_pow_rhs);
	}

	if (pc->lookahead2.token == '=') {
		idctx = pc->lookahead1;
		match(pc, pst, ps, infp, popts, L_ID, in_pow_rhs);

		ev_emit(L_LVALUE, idctx.tokenval, idctx.tokenbuf,
			pst, ps, popts);

		match(pc, pst, ps, infp, popts, '=', in_pow_rhs);
		E(pc, pst, ps, infp, popts, in_pow_rhs, 0);
		ev_emit('=', L_NONE, idctx.tokenbuf, pst, ps, popts);
	}
	else {
		E(pc, pst, ps, infp, popts, in_pow_rhs, 0);
	}
	EXIT;
}

// ----------------------------------------------------------------
static void E(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs,
	int          lhs_done_already)
{
	lexctx_t t;
	ENTER;

	if (!lhs_done_already)
		O(pc, pst, ps, infp, popts, in_pow_rhs);
	while (1) {
		switch (pc->lookahead1.token) {
		case '|':
			t = pc->lookahead1;
			match(pc, pst, ps, infp, popts,
				pc->lookahead1.token, in_pow_rhs);
			O(pc, pst, ps, infp, popts, in_pow_rhs);
			ev_emit(t.token, L_NONE, t.tokenbuf, pst, ps,
				popts);
			continue;
		default:
			if (lhs_done_already)
				O(pc, pst, ps, infp, popts, in_pow_rhs);
			EXIT;
			return;
		}
	}
	EXIT;
}

// ----------------------------------------------------------------
static void O(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs)
{
	lexctx_t t;
	ENTER;

	X(pc, pst, ps, infp, popts, in_pow_rhs);
	while (1) {
		switch (pc->lookahead1.token) {
		case '^':
			t = pc->lookahead1;
			match(pc, pst, ps, infp, popts,
				pc->lookahead1.token, in_pow_rhs);
			X(pc, pst, ps, infp, popts, in_pow_rhs);
			ev_emit(t.token, L_NONE, t.tokenbuf, pst, ps, popts);
			continue;
		default:
			EXIT;
			return;
		}
	}
}

// ----------------------------------------------------------------
static void X(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs)
{
	lexctx_t t;
	ENTER;

	A(pc, pst, ps, infp, popts, in_pow_rhs);
	while (1) {
		switch (pc->lookahead1.token) {
		case '&':
			t = pc->lookahead1;
			match(pc, pst, ps, infp, popts,
				pc->lookahead1.token, in_pow_rhs);
			A(pc, pst, ps, infp, popts, in_pow_rhs);
			ev_emit(t.token, L_NONE, t.tokenbuf, pst, ps, popts);
			continue;
		default:
			EXIT;
			return;
		}
	}
}

// ----------------------------------------------------------------
static void A(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs)
{
	lexctx_t t;
	ENTER;

	H(pc, pst, ps, infp, popts, in_pow_rhs);
	while (1) {
		switch (pc->lookahead1.token) {
		case L_LSH:
		case L_RSH:
			t = pc->lookahead1;
			match(pc, pst, ps, infp, popts,
				pc->lookahead1.token, in_pow_rhs);
			H(pc, pst, ps, infp, popts, in_pow_rhs);
			ev_emit(t.token, L_NONE, t.tokenbuf, pst, ps, popts);
			continue;
		default:
			EXIT;
			return;
		}
	}
}

// ----------------------------------------------------------------
static void H(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs)
{
	lexctx_t t;
	ENTER;

	T(pc, pst, ps, infp, popts, in_pow_rhs);
	while (1) {
		switch (pc->lookahead1.token) {
		case '+':
		case '-':
			t = pc->lookahead1;
			match(pc, pst, ps, infp, popts,
				pc->lookahead1.token, in_pow_rhs);
			T(pc, pst, ps, infp, popts, in_pow_rhs);
			ev_emit(t.token, L_NONE, t.tokenbuf, pst, ps, popts);
			continue;
		default:
			EXIT;
			return;
		}
	}
}

// ----------------------------------------------------------------
static void T(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs)
{
	lexctx_t t;
	ENTER;
	C(pc, pst, ps, infp, popts, in_pow_rhs);
	while (1) {
		switch (pc->lookahead1.token) {
		case '*':
		case '/':
		case '%':
			t = pc->lookahead1;
			match(pc, pst, ps, infp, popts,
				pc->lookahead1.token, in_pow_rhs);
			C(pc, pst, ps, infp, popts, in_pow_rhs);
			ev_emit(t.token, L_NONE, t.tokenbuf, pst, ps, popts);
			continue;
		default:
			EXIT;
			return;
		}
	}
}

// ----------------------------------------------------------------
static void C(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs)
{
	lexctx_t t;
	ENTER;

	F(pc, pst, ps, infp, popts, in_pow_rhs);
	while (1) {
		switch (pc->lookahead1.token) {
		case ',':
			t = pc->lookahead1;
			match(pc, pst, ps, infp, popts,
				pc->lookahead1.token, in_pow_rhs);
			F(pc, pst, ps, infp, popts, in_pow_rhs);
			ev_emit(t.token, L_NONE, t.tokenbuf, pst, ps, popts);
			continue;
		default:
			EXIT;
			return;
		}
	}
}

// ----------------------------------------------------------------
static void F(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs)
{
	lexctx_t t;
	ENTER;

	switch (pc->lookahead1.token) {

	case '+':
		t = pc->lookahead1;
		match(pc, pst, ps, infp, popts,
			pc->lookahead1.token, in_pow_rhs);
		B(pc, pst, ps, infp, popts, in_pow_rhs);
		ev_emit(L_UPLUS, L_NONE, t.tokenbuf, pst, ps, popts);
		break;

	case '-':
		t = pc->lookahead1;
		match(pc, pst, ps, infp, popts,
			pc->lookahead1.token, in_pow_rhs);
		B(pc, pst, ps, infp, popts, in_pow_rhs);
		ev_emit(L_UMINUS, L_NONE, t.tokenbuf, pst, ps, popts);
		break;

	case '~':
		t = pc->lookahead1;
		match(pc, pst, ps, infp, popts,
			pc->lookahead1.token, in_pow_rhs);
		B(pc, pst, ps, infp, popts, in_pow_rhs);
		ev_emit(L_UNOT, L_NONE, t.tokenbuf, pst, ps, popts);
		break;

	case L_SQRT:
	case L_EXP:
	case L_LOG:
	case L_SIN:
	case L_COS:
	case L_TAN:
	case L_ATAN:
	case L_CONJ:
	case L_MAG:
	case L_PHZ:
	case L_ORD:
		t = pc->lookahead1;
		match(pc, pst, ps, infp, popts,
			pc->lookahead1.token, in_pow_rhs);
		B(pc, pst, ps, infp, popts, in_pow_rhs);
		ev_emit(t.token, L_NONE, t.tokenbuf, pst, ps, popts);
		break;

	default:
		B(pc, pst, ps, infp, popts, in_pow_rhs);
		EXIT;
		return;
	}
}

// ----------------------------------------------------------------
static void B(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs)
{
	lexctx_t t;
	ENTER;

	U(pc, pst, ps, infp, popts, in_pow_rhs);

	switch (pc->lookahead1.token) {
	case '!':
		t = pc->lookahead1;
		match(pc, pst, ps, infp, popts,
			pc->lookahead1.token, in_pow_rhs);
		ev_emit(t.token, L_NONE, t.tokenbuf, pst, ps, popts);
		break;

	default:
		EXIT;
		return;
	}
}

// ----------------------------------------------------------------
static void U(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs)
{
	lexctx_t t;
	ENTER;

	P(pc, pst, ps, infp, popts, in_pow_rhs);
	while (1) {
		switch (pc->lookahead1.token) {
		case L_GCD: // xxx temp; need to find appropriate precedence.
		case L_CHOOSE: // xxx temp; need to find appropriate precedence.
		case L_POW:
			t = pc->lookahead1;
			// This match() gets the next token after the **.
			// Which might be a number (atom), or might be
			// a more complex expression.
			in_pow_rhs++;
			match(pc, pst, ps, infp, popts,
				pc->lookahead1.token, in_pow_rhs);

			P(pc, pst, ps, infp, popts, in_pow_rhs);
			ev_emit(t.token, L_NONE, t.tokenbuf, pst, ps, popts);
			in_pow_rhs--;
			continue;
		default:
			EXIT;
			return;
		}
	}
}

// ----------------------------------------------------------------
static void P(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          in_pow_rhs)
{
	ENTER;
	switch (pc->lookahead1.token) {
	case '(':
		match(pc, pst, ps, infp, popts, '(', in_pow_rhs);
		E(pc, pst, ps, infp, popts, in_pow_rhs, 0);
		match(pc, pst, ps, infp, popts, ')', in_pow_rhs);
		break;
	case L_NUM:
		ev_emit(L_NUM, pc->lookahead1.tokenval,
			pc->lookahead1.tokenbuf, pst, ps, popts);
		match(pc, pst, ps, infp, popts, L_NUM, in_pow_rhs);
		break;
	case L_ID:
		ev_emit(L_ID, pc->lookahead1.tokenval,
			pc->lookahead1.tokenbuf, pst, ps, popts);
		match(pc, pst, ps, infp, popts, L_ID, in_pow_rhs);
		break;
	default:
		fprintf(stderr, "Line %d:  syntax error, token ",
			pc->lookahead1.lineno);
		lexan_desc_token(pc->lookahead1.token, stderr);
		fprintf(stderr, ".\n");
		exit(1);
	}
	EXIT;
}

// ----------------------------------------------------------------
static void match(
	parsectx_t * pc,
	symtab_t   * pst,
	stack_t    * ps,
	FILE       * infp,
	opts_t     * popts,
	int          t,
	int          in_pow_rhs)
{
	ENTER;
	if (pc->lookahead1.token == t) {
#ifdef PARSE_DEBUG
		printf("match token ");
		lexan_desc_token(pc->lookahead1.token, stdout);
		printf("\n");
#endif
		if (pc->lookahead2.token == L_DEFER) {
			pc->lookahead2 = pc->lookahead1;
				// Retain line # & comment state
			lexan(&pc->lookahead2, pst, infp, popts, in_pow_rhs);
		}
		pc->lookahead1 = pc->lookahead2;
		pc->lookahead2.token = L_DEFER;
	}
	else {
		fprintf(stderr, "Line %d:  syntax error.  Expected ",
			pc->lookahead1.lineno);
		lexan_desc_token(t, stderr);
		fprintf(stderr, "; got ");
		lexan_desc_token(pc->lookahead1.token, stderr);
		fprintf(stderr, ".\n");
		exit(1);
	}
	EXIT;
}
