#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "gcd.h"
#include "ops.h"

typedef struct _rat64_t {
	long long numer;
	long long denom;
} rat64_t;

// ----------------------------------------------------------------
static void simplify(rat64_t * p)
{
	long long g;
	if (p->denom == 0) {
		fprintf(stderr, "rat64: Divide by zero.\n");
		exit(1);
	}
	if (p->denom < 0) {
		p->numer = -p->numer;
		p->denom = -p->denom;
	}
	g = calc_gcd_signed_ll(p->numer, p->denom);
	p->numer /= g;
	p->denom /= g;
}

// ----------------------------------------------------------------
static int rat64_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int rat64_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	rat64_t rat64;
	rat64.denom = 1;
	if (fscanf(fp, "%lld", &rat64.numer) != 1)
		return 0;
	memcpy(patom, &rat64, sizeof(rat64));
	return 1;
}

// ----------------------------------------------------------------
static void rat64_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	rat64_t x;
	memcpy(&x, patom, sizeof(x));
	fprintf(fp, "%lld", x.numer);
	if (x.denom != 1)
		fprintf(fp, "/%lld", x.denom);
}

// ----------------------------------------------------------------
static void rat64_rectifier(
	void * buf,
	void * pvaux)
{
	rat64_t c;
	memcpy(&c, buf, sizeof(c));
	simplify(&c);
	memcpy(buf, &c, sizeof(c));
}

// ----------------------------------------------------------------
//  a     c     ad     bc     ad + bc
// --- + --- = ---- + ---- = ---------
//  b     d     bd     bd       bd

static void rat64_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	rat64_t a, b, c;
	long long  g, l, r;

	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	g = calc_gcd_signed_ll(a.denom, b.denom);
	l = b.denom / g;
	r = a.denom / g;

	c.numer = a.numer * l + b.numer * r;
	c.denom = a.denom * l;

	simplify(&c);

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
//  a     c     ad     bc     ad - bc
// --- - --- = ---- - ---- = ---------
//  b     d     bd     bd       bd

static void rat64_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	rat64_t a, b, c;
	long long g, l, r;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	g = calc_gcd_signed_ll(a.denom, b.denom);
	l = b.denom / g;
	r = a.denom / g;

	c.numer = a.numer * l - b.numer * r;
	c.denom = a.denom * l;

	simplify(&c);

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void rat64_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	rat64_t a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	c.numer = a.numer * b.numer;
	c.denom = a.denom * b.denom;

	simplify(&c);

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
//  a     c     a     d     ad
// --- / --- = --- * --- = ----
//  b     d     b     c     bc

static void rat64_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	rat64_t a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(b));

	c.numer = a.numer * b.denom;
	c.denom = a.denom * b.numer;

	simplify(&c);

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void rat64_power(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	rat64_t a, b, c;
	rat64_t p2;
	long long i;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	if (b.denom != 1) {
		fprintf(stderr,
		"rat:  Can't handle fractional exponents (got %lld/%lld).\n",
			b.numer, b.denom);
		exit(1);
	}

	// xxx needs range check on b.

	c.denom  = 1;
	c.numer  = 1;
	p2 = a;

	for (i = 0; i < 64; i++) {
		if (b.numer & 1) {
			c.numer *= p2.numer;
			c.denom *= p2.denom;
		}
		b.numer >>= 1;
		p2.numer *= p2.numer;
		p2.denom *= p2.denom;
	}

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void rat64_uadd(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	memcpy(pvout, pvin, sizeof(rat64_t));
}

// ----------------------------------------------------------------
static void rat64_usub(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	rat64_t c;
	memcpy(&c, pvin, sizeof(c));
	c.numer = -c.numer;
	memcpy(pvout, &c, sizeof(c));
}

// ================================================================
ops_t rat64_ops = {
	atom_desc          : "rat64",
	plex_first_func    : rat64_lex_first,
	pscan_func         : rat64_scan,
	pprint_func        : rat64_print,
	prectifier         : rat64_rectifier,
	padd_op            : rat64_add,
	psub_op            : rat64_sub,
	pmul_op            : rat64_mul,
	pdiv_op            : rat64_div,
	ppow_op            : rat64_power,
	puadd_op           : rat64_uadd,
	pusub_op           : rat64_usub,
	atom_size_in_bytes : sizeof(rat64_t),
	aux_size_in_bytes  : 0
};
