#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "gcd.h"
#include "ops.h"

typedef struct _rat_t {
	int numer;
	int denom;
} rat_t;

// ----------------------------------------------------------------
static void simplify(rat_t * p)
{
	int g;
	if (p->denom == 0) {
		fprintf(stderr, "rat: Divide by zero.\n");
		exit(1);
	}
	if (p->denom < 0) {
		p->numer = -p->numer;
		p->denom = -p->denom;
	}
	g = calc_gcd_signed(p->numer, p->denom);
	p->numer /= g;
	p->denom /= g;
}

// ----------------------------------------------------------------
static int rat_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int rat_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	rat_t rat;
	rat.denom = 1;
	if (fscanf(fp, "%d", &rat.numer) != 1)
		return 0;
	memcpy(patom, &rat, sizeof(rat));
	return 1;
}

// ----------------------------------------------------------------
static void rat_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	rat_t x;
	memcpy(&x, patom, sizeof(x));
	fprintf(fp, "%d", x.numer);
	if (x.denom != 1)
		fprintf(fp, "/%d", x.denom);
}

// ----------------------------------------------------------------
static void rat_rectifier(
	void * buf,
	void * pvaux)
{
	rat_t c;
	memcpy(&c, buf, sizeof(c));
	simplify(&c);
	memcpy(buf, &c, sizeof(c));
}

// ----------------------------------------------------------------
//  a     c     ad     bc     ad + bc
// --- + --- = ---- + ---- = ---------
//  b     d     bd     bd       bd

static void rat_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	rat_t a, b, c;
	int  g, l, r;

	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	g = calc_gcd_signed(a.denom, b.denom);
	l = b.denom / g;
	r = a.denom / g;

	c.numer = a.numer * l + b.numer * r;
	c.denom = a.denom * l;

	simplify(&c);

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
//  a     c     ad     bc     ad - bc
// --- - --- = ---- - ---- = ---------
//  b     d     bd     bd       bd

static void rat_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	rat_t a, b, c;
	int g, l, r;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	g = calc_gcd_signed(a.denom, b.denom);
	l = b.denom / g;
	r = a.denom / g;

	c.numer = a.numer * l - b.numer * r;
	c.denom = a.denom * l;

	simplify(&c);

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void rat_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	rat_t a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	c.numer = a.numer * b.numer;
	c.denom = a.denom * b.denom;

	simplify(&c);

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
//  a     c     a     d     ad
// --- / --- = --- * --- = ----
//  b     d     b     c     bc

static void rat_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	rat_t a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(b));

	c.numer = a.numer * b.denom;
	c.denom = a.denom * b.numer;

	simplify(&c);

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void rat_power(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	rat_t a, b, c;
	rat_t p2;
	int i;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	if (b.denom != 1) {
		fprintf(stderr,
		"rat:  Can't handle fractional exponents (got %d/%d).\n",
			b.numer, b.denom);
		exit(1);
	}

	// xxx needs range check on b.

	c.denom  = 1;
	c.numer  = 1;
	p2 = a;

	for (i = 0; i < 32; i++) {
		if (b.numer & 1) {
			c.numer *= p2.numer;
			c.denom *= p2.denom;
		}
		b.numer >>= 1;
		p2.numer *= p2.numer;
		p2.denom *= p2.denom;
	}

	memcpy(pvout, &c, sizeof(a));
}


// ----------------------------------------------------------------
static void rat_uadd(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	memcpy(pvout, pvin, sizeof(rat_t));
}

// ----------------------------------------------------------------
static void rat_usub(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	rat_t c;
	memcpy(&c, pvin, sizeof(c));
	c.numer = -c.numer;
	memcpy(pvout, &c, sizeof(c));
}

// ================================================================
ops_t rat_ops = {
	atom_desc          : "rat",
	plex_first_func    : rat_lex_first,
	pscan_func         : rat_scan,
	pprint_func        : rat_print,
	prectifier         : rat_rectifier,
	padd_op            : rat_add,
	psub_op            : rat_sub,
	pmul_op            : rat_mul,
	pdiv_op            : rat_div,
	ppow_op            : rat_power,
	puadd_op           : rat_uadd,
	pusub_op           : rat_usub,
	atom_size_in_bytes : sizeof(rat_t),
	aux_size_in_bytes  : 0
};
