#include <stdio.h>
#include <stdlib.h>
#include "rat.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
// The caller must free the return value.

int * R_from_m(
	int   m,
	int   stride,
	int * poR)
{
	int ms, * rv, i;

	ms = m / stride;

	if (ms <= 0) {
		fprintf(stderr,
			"R_from_m:  Need m/s > 1; got %d.\n",
			ms);
		exit(1);
	}

	rv = malloc_check(sizeof(int) * m);

	for (i = 0; i < ms; i++)
		rv[i] = i * stride;
	*poR = ms;
	return rv;
}

// ----------------------------------------------------------------
// The caller must free the return value.

int * S_from_range(
	int   slo,
	int   shi,
	int * poS)
{
	int n, i, * rv;

	n = shi - slo + 1;
	if (n <= 0) {
		fprintf(stderr,
			"S_from_range:  Need slo < shi; got %d, %d.\n",
			slo, shi);
		exit(1);
	}

	rv = malloc_check(sizeof(int) * n);

	for (i = 0; i < n; i++)
		rv[i] = i + slo;
	*poS = n;
	return rv;
}

// ----------------------------------------------------------------
// The caller must free the return value.

int * S_from_argv(
	int     argc,
	char ** argv,
	int   * poS)
{
	int argi, * rv;

	if (argc <= 0) {
		fprintf(stderr, "S_from_argv:  Need non-empty set.\n");
		exit(1);
	}

	rv = malloc_check(sizeof(int) * argc);

	for (argi = 0; argi < argc; argi++) {
		if (sscanf(argv[argi], "%d", &rv[argi]) != 1) {
			fprintf(stderr, "Couldn't scan \"%s\".\n", argv[argi]);
			exit(1);
		}
	}
	*poS = argc;
	return rv;
}

// ----------------------------------------------------------------
int icontains_element(
	int * R,
	int   oR,
	int   s)
{
	int i;

	for (i = 0; i < oR; i++)
		if (s == R[i])
			return 1;
	return 0;
}

// ----------------------------------------------------------------
int icontains_set(
	int * R,
	int   oR,
	int * S,
	int   oS)
{
	int i;

	for (i = 0; i < oS; i++)
		if (!icontains_element(R, oR, S[i]))
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int is_multiplicative(
	int * S,
	int   oS,
	int   m)
{
	int i, j;

	for (i = 0; i < oS; i++)
		for (j = 0; j < oS; j++)
			if (!icontains_element(S, oS, (S[i] * S[j]) % m))
				return 0;
	return 1;
}

// ----------------------------------------------------------------
int rcontains_element(
	rat_t * R,
	int    oR,
	rat_t   s)
{
	int i;

	for (i = 0; i < oR; i++)
		if (rat_equal(s, R[i]))
			return 1;
	return 0;
}

// ----------------------------------------------------------------
void radd_element(
	rat_t * SR,
	int   * poSR,
	rat_t   q)
{
	if (!rcontains_element(SR, *poSR, q)) {
		SR[*poSR] = q;
		(*poSR)++;
	}
}

// ----------------------------------------------------------------
// The caller must free the return value.

rat_t * form_SR(
	int * R,
	int   oR,
	int * S,
	int   oS,
	int   m,
	int * poSR)
{
	int i, j;
	rat_t q;
	rat_t * SR = malloc_check(sizeof(rat_t) * oR * oS);
	int oSR = 0;

	for (i = 0; i < oR; i++) {
		for (j = 0; j < oS; j++) {
			if ((S[j] % m) == 0) {
				fprintf(stderr, "S contains 0.  Imagine "
					"one big equivalence class.\n");
				exit(1);
			}
			q = rat_from_ints(R[i], S[j]);
			radd_element(SR, &oSR, q);
		}
	}

	qsort(SR, oSR, sizeof(rat_t), rat_qcmp);

	*poSR = oSR;
	return SR;
}

// ----------------------------------------------------------------
void show_equiv_classes(
	rat_t * SR,
	int     oSR,
	int   * S,
	int     oS,
	int     m)
{
	int i, j, k;
	char * marks = malloc_check(oSR);
	int ri_sj, rj_si, sk_ri_sj, sk_rj_si;

	for (i = 0; i < oSR; i++)
		marks[i] = 0;

	for (i = 0; i < oSR; i++) {
		if (marks[i])
			continue;
		for (j = i; j < oSR; j++) {
			if (marks[j])
				continue;
			ri_sj = (SR[i].numer * SR[j].denom) % m;
			rj_si = (SR[j].numer * SR[i].denom) % m;
			for (k = 0; k < oS; k++) {
				sk_ri_sj = (S[k] * ri_sj) % m;
				sk_rj_si = (S[k] * rj_si) % m;
				if (((sk_ri_sj - sk_rj_si) % m) == 0) {
					//rat_print(SR[i]);
					//printf(" ~ ");
					if (i != j)
						printf(" ");
					rat_print(SR[j]);
					//printf("\n");
					marks[i] = 1;
					marks[j] = 1;
					break;
				}
			}
		}
		printf("\n");
	}
	free(marks);
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {m} {S elements ... }\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int * R, * S;
	rat_t * SR;
	int   oR, oS, oSR;
	int   slo = 0, shi = 0;
	int   m, stride;

	// Parse command line to get R, S.
	if (argc < 3)
		usage(argv[0]);

	if (sscanf(argv[1], "%d:%d", &m, &stride) == 2)
		;
	else if (sscanf(argv[1], "%d", &m) == 1)
		stride = 1;
	else
		usage(argv[0]);

	R = R_from_m(m, stride, &oR);

	if (sscanf(argv[2], "%d-%d", &slo, &shi) == 2)
		S = S_from_range(slo, shi, &oS);
	else
		S = S_from_argv(argc - 2, &argv[2], &oS);

	// Check that S is a subset of R.
	if (!icontains_set(R, oR, S, oS)) {
		fprintf(stderr, "S is not a subset of R.\n");
		exit(1);
	}

	// Check that S is multiplicative.
	if (!is_multiplicative(S, oS, m)) {
		fprintf(stderr, "S is not multiplicative.\n");
		exit(1);
	}
	SR = form_SR(R, oR, S, oS, m, &oSR);

#if 0
	{
		int i;
		printf("R: ");
		for (i = 0; i < oR; i++)
			printf(" %d", R[i]);
		printf("\n");

		printf("S: ");
		for (i = 0; i < oS; i++)
			printf(" %d", S[i]);
		printf("\n");

		printf("SR: ");
		for (i = 0; i < oSR; i++) {
			printf(" ");
			rat_print(SR[i]);
		}
		printf("\n");
		printf("\n");
	}
#endif

	show_equiv_classes(SR, oSR, S, oS, m);

	free(R);
	free(S);
	free(SR);
	return 0;
}
