// This program is an illustration of Zolotarev's Lemma, which states that
// for prime p and m non-zero mod p, the Legendre symbol (m|p) is equal to the
// parity of the permutation tau_m on Zp* induced by multiplication by m, i.e.
// tau_m(x) = mx.  That is, m is a quadratic residue mod p if and only if
// tau_m is an even permutation.
//
// This program prints (for m = 1 to p-1) the number m, the sign of tau_m,
// and the cycle decomposition of tau_m.  Quadratic residues are printed
// first, followed by all quadratic non-residues.
//
// John Kerl
// 2003/10/08

#include <stdio.h>
#include <stdlib.h>
#include "isprime.h"
#include "pmt.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr,
		"Usage: %s {p} {m}, where p must be prime.\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int p, m, x;
	pmt_t pmt;
	int   parity;
	int * parities = 0;

	if (argc != 2)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &p) != 1)
		usage(argv[0]);
	if (!isprime(p))
		usage(argv[0]);

	parities = (int *)malloc_check(p * sizeof(int));
	for (m = 1; m < p; m++) {
		for (x = 0; x < p; x++)
			pmt.images[x] = (m * x) % p;
		parities[m] = pmt_parity(&pmt, p);
	}

	// Print squares first, then non-squares.
	for (parity = 0; parity <= 1; parity++) {
		for (m = 1; m < p; m++) {
			if (parities[m] != parity)
				continue;

			for (x = 0; x < p; x++)
				pmt.images[x] = (m * x) % p;
			printf("%d (%c1): ",
				m,
				pmt_parity(&pmt, p) == 0 ? '+' : '-');
			pmt_print_cd_style(&pmt, p, PMT_CD_PARENS);
			printf("\n");
		}
	}
	printf("\n");

	free(parities);
	return 0;
}
