#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "power.h"
#include "zpnpoly.h"
#include "zpnpoly_factor.h"
#include "factorize.h"

#define DEPTH_NONE        0
#define DEPTH_ALL         1
#define DEPTH_ROOTLESS    2
#define DEPTH_IRREDUCIBLE 3
#define DEPTH_PRIMITIVE   4

#define DEFAULT_DISPLAY_DEPTH DEPTH_IRREDUCIBLE
#define DEFAULT_DO_EVAL   0
#define DEFAULT_BREAK_ON_FIRST 0

typedef struct _num_found_t {
	int num_all;
	int num_rootless;
	int num_irreducible;
	int num_primitive;
} num_found_t;

typedef struct _opts_t {
	zppoly_aux_t innaux;
	int margi;
	int display_depth;
	int do_eval;
	int show_factors;
	int show_period;
	int show_sc;
	int do_plot;
	int plot_width;
	int break_on_first;
	int brief_output;
} opts_t;

static void do_one_degree(
	zppoly_aux_t * pinnaux,
	int            m,
	num_found_t  * pnf,
	opts_t       * popts);

static int do_one_poly(
	zpnpoly_t     * poutr,
	zppoly_aux_t  * pinnaux,
	unsigned        pm1,
	factor_info_t * pifinfo,
	num_found_t   * pnf,
	opts_t        * popts);

// ----------------------------------------------------------------
static void set_defaults(
	opts_t * popts);

static int parse_command_line(
	int       argc,
	char   ** argv,
	opts_t  * popts);

static char * get_depth_desc(
	int depth);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr,
	"Usage: %s [options] {p} {innr} {mlo[-mhi]}\n", argv0);
	fprintf(stderr,
	"Lists some or all polynomials of specified degree(s) over GF(p^m).\n"
	"Options:\n"
	"-a:      Show all polynomials.\n"
	"-r:      Show rootless polynomials.\n"
	"-i:      Show irreducible polynomials (default).\n"
	"-p:      Show primitive polynomials.\n"
	"\n"
	"-1:      Show only the first polynomial matching specified criteria.\n"
	"-v:      Show polynomial evaluations in GF(p^m).\n"
	"-nv:     Do not show polynomial evaluations in GF(p^m).\n"
	"-f:      Show polynomial factorizations.  (Silly without -a "
		"or -r.)\n"
	"-e:      Show polynomial periods.\n"
	"-sc:     Show self-complementarity.\n"
	"-q:      Terse output.\n"
	"\n"
	);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	opts_t opts;
	int argi, mlo, mhi, m;
	num_found_t nf;

	nf.num_all         = 0;
	nf.num_rootless    = 0;
	nf.num_irreducible = 0;
	nf.num_primitive   = 0;

	set_defaults(&opts);
	if (!parse_command_line(argc, argv, &opts))
		usage(argv[0]);


	for (argi = opts.margi; argi < argc; argi++) {
		if (sscanf(argv[argi], "%d-%d", &mlo, &mhi) == 2)
			;
		else if (sscanf(argv[argi], "%d", &mlo) == 1)
			mhi = mlo;
		else
			usage(argv[0]);

		if (mlo < 1)
			usage(argv[0]);
		if (mhi < mlo)
			usage(argv[0]);
		for (m = mlo; m <= mhi; m++) {
			do_one_degree(&opts.innaux, m, &nf, &opts);
		}
	}


	if (!opts.break_on_first && !opts.brief_output) {
		printf("\n");
		printf("%6d found; %6d rootless; %6d irreducible; "
			"%6d primitive.\n",
			nf.num_all, nf.num_rootless,
			nf.num_irreducible, nf.num_primitive);
	}

	return 0;
}

// ----------------------------------------------------------------
static void do_one_degree(
	zppoly_aux_t * pinnaux,
	int            m,
	num_found_t  * pnf,
	opts_t       * popts)
{
	zpnpoly_t outr;
	int d = pinnaux->r.degree;
	unsigned rep, pm, pm1, tpm1; // xxx need 64-bit here
	factor_info_t finfo;

	if (!popts->break_on_first && !popts->brief_output) {
		printf("%s degree-%d polynomials over GF(%d^%d)",
			get_depth_desc(popts->display_depth),
			m, pinnaux->p, d);
		printf("):\n");
	}

	pm = calc_power(pinnaux->p, m * d);
	tpm1 = pm + pm - 1;
	pm1 = pm - 1;
	calc_factors(pm1, &finfo);

	for (rep = pm; rep <= tpm1; rep++) {
		outr = zpnpoly_from_base_rep(rep, pinnaux->p, d);
		if (!do_one_poly(&outr, pinnaux, pm1, &finfo, pnf, popts))
			break;
	}
	if (popts->brief_output)
		printf("\n");
}

// ----------------------------------------------------------------
static int do_one_poly(
	zpnpoly_t     * poutr,
	zppoly_aux_t  * pinnaux,
	unsigned        pm1,
	factor_info_t * pifinfo,
	num_found_t   * pnf,
	opts_t        * popts)
{
	int rootless = 0, irr = 0, prim = 0;
	zpnpoly_factor_info_t zfinfo;

	rootless = zpnpoly_is_rootless(poutr, pinnaux);
	zpnpoly_factor(poutr, pinnaux, &zfinfo);
	if ((zfinfo.num_distinct == 1) && (zfinfo.factor_counts[0].count == 1))
		irr = 1;
	else
		irr = 0;
	if (irr)
		prim = zpnpoly_is_primitive_f(poutr, pinnaux, pm1, pifinfo);
	else
		prim = 0;

	pnf->num_all++;
	if (irr)
		pnf->num_irreducible++;
	if (prim)
		pnf->num_primitive++;
	if (rootless)
		pnf->num_rootless++;

	if (popts->do_plot) {
//		//if (((r-pm1-1) % popts->plot_width) == 0)
//			//printf("\n");
//		printf("%c",
//			prim ? '@' :
//			irr  ? '*' :
//			rootless ? '.' :
//			'-');
//		fflush(stdout);
//		return 1;
	}
	else {
		int print_it = 0;

		if (popts->display_depth == DEPTH_ALL)
			print_it = 1;
		else if ((popts->display_depth == DEPTH_ROOTLESS)
				&& rootless)
			print_it = 1;
		else if ((popts->display_depth == DEPTH_IRREDUCIBLE)
				&& irr)
			print_it = 1;
		else if ((popts->display_depth == DEPTH_PRIMITIVE)
				&& prim)
			print_it = 1;

		if (!print_it)
			return 1;

		if (popts->brief_output)
			printf(" ");
		zpnpoly_print(poutr, pinnaux);

		if (!popts->brief_output) {

			if (popts->do_eval) {
				int inndeg = pinnaux->r.degree;
				int pn = calc_power(pinnaux->p, inndeg);
				int i;
				zpoly_t in, out;

				printf(" [");
				for (i = 0; i < pn; i++) {
					in = zppoly_from_base_rep(
						pinnaux->p, i);
					if (i > 0LL)
						printf("/");
					out = zpnpoly_eval(&in, poutr, pinnaux),
					zppolymod_print(&out, pinnaux);
				}
				printf("]");
			}

			if (popts->show_period) {
//				int ndigits = 1 + (int)log10(pm1);
//				printf(" e=");
//				if (irr) {
////					printf(" %*d", ndigits,
////					zpnpolymod_element_order(2LL, r));
//				}
//				else {
//					printf(" %*s", ndigits, "-");
//				}
			}

//			if (popts->show_sc) {
//				zpnpoly_t r1 = zpnpoly_compose(r, 3LL);
//				if (r == r1)
//					printf(" sc");
//				else
//					printf("  m");
//			}

			if (popts->show_factors) {
				int i;

				printf(" %c%c%c  = ",
					rootless ? 'r' : '-',
					irr      ? 'i' : '-',
					prim     ? 'p' : '-');

				for (i = 0; i < zfinfo.num_distinct; i++) {
					printf(" ");
					zpnpoly_print(
						&zfinfo.factor_counts[i].f,
						pinnaux);
					if (zfinfo.factor_counts[i].count > 1)
						printf("^%d",
						zfinfo.factor_counts[i].count);
				}
			}
			else {
				printf(" ");
				if (rootless)
					printf("r");
				if (irr)
					printf("i");
				if (prim)
					printf("p");
			}


			printf("\n");
			// In case output is not a tty, e.g. | less
			fflush(stdout);
		}

		if (popts->break_on_first)
			return 0;
		else
			return 1;
	}
	return 1;
}

// ----------------------------------------------------------------
static void set_defaults(
	opts_t * popts)
{
	memset(popts, 0, sizeof(*popts));

	popts->innaux.p = 2;
	popts->innaux.r = zppoly_from_base_rep(2, 0x13);
	popts->margi = 0;

	popts->display_depth  = DEFAULT_DISPLAY_DEPTH;
	popts->do_eval        = DEFAULT_DO_EVAL;
	popts->show_factors   = 0;
	popts->show_period    = 0;
	popts->show_sc        = 0;
	popts->do_plot        = 0;
	popts->plot_width     = 0;
	popts->break_on_first = 0;
	popts->brief_output   = 0;
}

// ----------------------------------------------------------------
static int parse_command_line(
	int       argc,
	char   ** argv,
	opts_t  * popts)
{
	int argi;

	for (argi = 1; argi < argc; ) {
		if (argv[argi][0] != '-')
			break;

		if (strcmp(argv[argi], "--help") == 0)
			return 0;

		else if (strcmp(argv[argi], "-a") == 0) {
			popts->display_depth = DEPTH_ALL;
			argi++;
		}
		else if (strcmp(argv[argi], "-r") == 0) {
			popts->display_depth = DEPTH_ROOTLESS;
			argi++;
		}
		else if (strcmp(argv[argi], "-i") == 0) {
			popts->display_depth = DEPTH_IRREDUCIBLE;
			argi++;
		}
		else if (strcmp(argv[argi], "-p") == 0) {
			popts->display_depth = DEPTH_PRIMITIVE;
			argi++;
		}

		else if (strcmp(argv[argi], "-1") == 0) {
			popts->break_on_first = 1;
			argi++;
		}

		else if (strcmp(argv[argi], "-v") == 0) {
			popts->do_eval = 1;
			argi++;
		}
		else if (strcmp(argv[argi], "-nv") == 0) {
			popts->do_eval = 0;
			argi++;
		}
		else if (strcmp(argv[argi], "-f") == 0) {
			popts->show_factors = 1;
			argi++;
		}
		else if (strcmp(argv[argi], "-e") == 0) {
			popts->show_period = 1;
			argi++;
		}
		else if (strcmp(argv[argi], "-sc") == 0) {
			popts->show_sc = 1;
			argi++;
		}

		else if (strcmp(argv[argi], "-plot") == 0) {
			if ((argc - argi) < 2)
				return 0;
			if (sscanf(argv[argi+1], "%d", &popts->plot_width) != 1)
				return 0;
			popts->do_plot = 1;
			argi += 2;
		}

		else if (strcmp(argv[argi], "-b") == 0) {
			popts->brief_output = 1;
			argi++;
		}

		else {
			fprintf(stderr,
				"%s: Unrecognized option \"%s\".\n",
				argv[0], argv[argi]);
			return 0;
		}
	}

	if ((argc - argi) < 3)
		return 0;

	if (!zppolymod_auxscan(&popts->innaux, argv[argi], argv[argi+1]))
		return 0;
	popts->margi = argi + 2;

	return 1;
}

// ----------------------------------------------------------------
static char * get_depth_desc(
	int depth)
{
	switch (depth) {
	case DEPTH_ALL:
		return "All";
		break;
	case DEPTH_ROOTLESS:
		return "Rootless";
		break;
	case DEPTH_IRREDUCIBLE:
		return "Irreducible";
		break;
	case DEPTH_PRIMITIVE:
		return "Primitive";
		break;
	default:
		return "???";
	}
}
