#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "z2ipolymod.h"
#include "z2ilinalg.h"
#include "z2nilinalg.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
// r = 0x13
// M = 0 0 1 1
//     1 0 0 0
//     0 1 0 0
//     0 0 1 0

static z2ipoly_t ** ext_comp_mx(
	z2ipoly_t r,
	int do_galois_cfg)
{
	int n = z2ipoly_get_degree(r);
	z2ipoly_t ** M = z2nila_alloc_matrix(n, n);
	int i, j;

	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			M[i][j] = 0LL;
	for (i = 1; i < n; i++)
		M[i][i-1] = 1LL;
	for (i = 0; i < n; i++) {
		if (r & (1LL << i)) {
			if (do_galois_cfg)
				M[i][n-1] = 1LL;
			else
				M[0][n-i-1] = 1LL;
		}
	}

	return M;
}


// ----------------------------------------------------------------
static void do_one_poly(
	z2ipoly_t r,
	int do_galois_cfg,
	int show_logs,
	int verbose)
{
	z2ipoly_t ** M = 0;
	z2ipoly_t  * ev = 0;
	z2ipoly_t  * ck0 = 0;
	z2ipoly_t root = 2LL;
	int n, k, i;
	int width = 0;
	z2ipoly_t g = 0LL;

	n = z2ipoly_get_degree(r);

	if (show_logs) {
		width = 1 + (int)log10((1 << n) - 1);
		g = z2ipolymod_find_generator(r);
	}

	ev  = malloc_check(n * sizeof(z2ipoly_t));
	ck0 = malloc_check(n * sizeof(z2ipoly_t));

	for (k = 0; k < n; k++) {

		M = ext_comp_mx(r, do_galois_cfg);
		if (verbose) {
			printf("Companion matrix:\n");
			z2nila_print_matrix(M, n, n, r);
			printf("\n");
		}

		for (i = 0; i < n; i++)
			M[i][i] ^= root;
		if (verbose) {
			printf("Minus lambda I:\n");
			z2nila_print_matrix(M, n, n, r);
			printf("\n");
		}

		z2nila_row_echelon_form(M, r, n, n);
		if (verbose) {
			printf("Row-echelon form:\n");
			z2nila_print_matrix(M, n, n, r);
			printf("\n");
		}

		// Quick and dirty:  Assume r is irreducible, so rank is 1.
		// Print out eigenvectors via last column.  A more flexible
		// eigensystem solver would use back-substitution, etc.
		for (i = 0; i < n - 1; i++) {
			ev[i] = M[i][n-1];
		}
		ev[n-1] = 1LL;

		if (show_logs) {
			for (i = 0; i < n; i++) {
				if (i > 0)
					printf(" ");
				printf("%*d",
					width,
					z2ipolymod_log(ev[i], g, r));
			}
		}
		else {
			z2nila_print_vector(ev, n, r);
		}
		printf("\n");

		z2nila_mx_times_vr(M, ev, ck0, n, r);

		// Make sure this is really an eigenvector.
		for (i = 0; i < n; i++) {
			if (ck0[i] != 0LL) {
				printf("f2eisys: coding error.\n");
				z2nila_print_vector(ck0, n, r);
				printf("\n");
				exit(1);
			}
		}

		z2nila_free_matrix(M, n, n);

		root = z2ipolymod_power(root, 2, r);
	}

	free(ev);
	free(ck0);
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {r}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t r = 0LL;
	int do_galois_cfg = 0;
	int show_logs = 0;
	int verbose = 0;
	int argi, argb = 1;

	for (argi = 1; argi < argc; argi++) {
		if (argv[argi][0] != '-')
			break;
		if (strcmp(argv[argi], "-l") == 0)
			show_logs = 1;
		else if (strcmp(argv[argi], "-g") == 0)
			do_galois_cfg = 1;
		else if (strcmp(argv[argi], "-v") == 0)
			verbose = 1;
		else
			usage(argv[0]);
		argb++;
	}

	if ((argc - argb) < 1)
		usage(argv[0]);

	for (argi = argb; argi < argc; argi++) {
		if (sscanf(argv[argi], "%llx", &r) != 1) {
			fprintf(stderr, "%s: Couldn't scan \"%s\".\n",
				argv[0], argv[argi]);
			exit(1);
		}
		do_one_poly(r, do_galois_cfg, show_logs, verbose);
		if ((argc - argb) > 1)
			printf("\n");
	}

	return 0;
}
