#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "zppolymod.h"
#include "zplinalg.h"
#include "zpnlinalg.h"
#include "power.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
// r = 0x13
// M = 0 0 1 1
//     1 0 0 0
//     0 1 0 0
//     0 0 1 0

static zpoly_t ** ext_comp_mx(
	zppoly_aux_t * paux,
	int do_galois_cfg)
{
	int n = paux->r.degree;
	zpoly_t ** M = zpnla_alloc_matrix(n, n);
	int c;
	int i, j;

	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			M[i][j] = zppoly_from_base_rep(paux->p, 0);
	for (i = 1; i < n; i++)
		M[i][i-1] = zppoly_from_base_rep(paux->p, 1);
	for (i = 0; i < n; i++) {
		c = (paux->p - paux->r.coeffs[i]) % paux->p;
		if (do_galois_cfg)
			M[i][n-1] = zppoly_from_base_rep(paux->p, c);
		else
			M[0][n-i-1] = zppoly_from_base_rep(paux->p, c);
	}

	return M;
}


// ----------------------------------------------------------------
static void do_one_poly(
	zppoly_aux_t * paux,
	int do_galois_cfg,
	int show_logs,
	int verbose)
{
	zpoly_t ** M = 0;
	zpoly_t * ev = 0;
	zpoly_t * ck0 = 0;
	zpoly_t root = zppoly_from_coeffs1(1, 0, paux->p);
	zpoly_t one = zppoly_from_coeffs0(1, paux->p);
	zpoly_t neg;
	int n, k, i;
	int width = 0;
	zpoly_t g = zppoly_from_coeffs0(0, paux->p);

	n = paux->r.degree;

	if (show_logs) {
		width = 1 + (int)log10(calc_power(paux->p, n) - 1);
		g = zppolymod_find_generator(paux);
	}

	ev  = malloc_check(n * sizeof(zpoly_t));
	ck0 = malloc_check(n * sizeof(zpoly_t));

	for (k = 0; k < n; k++) {

		M = ext_comp_mx(paux, do_galois_cfg);
		if (verbose) {
			printf("Companion:\n");
			zpnla_print_matrix(M, n, n, paux);
			printf("\n");
		}

		for (i = 0; i < n; i++)
			M[i][i] = zppolymod_sub(&M[i][i], &root, paux);
		if (verbose) {
			printf("Minus lambda I:\n");
			zpnla_print_matrix(M, n, n, paux);
			printf("\n");
		}

		zpnla_row_echelon_form(M, paux, n, n);
		if (verbose) {
			printf("Row-echelon form:\n");
			zpnla_print_matrix(M, n, n, paux);
			printf("\n");
		}

		// Quick and dirty:  Assume r is irreducible, so rank is 1.
		// Print out eigenvectors via last column.  A more flexible
		// eigensystem solver would use back-substitution, etc.
		for (i = 0; i < n - 1; i++) {
			neg = zppoly_usub(&M[i][n-1], paux->p);
			ev[i] = neg;
		}
		ev[n-1] = one;

		if (show_logs) {
			for (i = 0; i < n; i++) {
				if (i > 0)
					printf(" ");
				printf("%*d",
					width,
					zppolymod_log(&ev[i], &g, paux));
			}
		}
		else {
			zpnla_print_vector(ev, n, paux);
		}
		printf("\n");

		zpnla_mx_times_vr(M, ev, ck0, n, paux);

		// Make sure this is really an eigenvector.
		for (i = 0; i < n; i++) {
			if (!zppoly_is_zero(&ck0[i])) {
				printf("fpeisys: coding error.\n");
				zpnla_print_vector(ck0, n, paux);
				printf("\n");
				exit(1);
			}
		}

		zpnla_free_matrix(M, n, n);
		if (verbose) {
			printf("----\n");
		}

		root = zppolymod_power(&root, paux->p, paux);
	}

	free(ev);
	free(ck0);
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {r}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	zppoly_aux_t aux;
	int do_galois_cfg = 0;
	int show_logs = 0;
	int verbose = 0;
	int argi, argb = 1;

	for (argi = 1; argi < argc; argi++) {
		if (argv[argi][0] != '-')
			break;
		if (strcmp(argv[argi], "-l") == 0)
			show_logs = 1;
		else if (strcmp(argv[argi], "-g") == 0)
			do_galois_cfg = 1;
		else if (strcmp(argv[argi], "-v") == 0)
			verbose = 1;
		else
			usage(argv[0]);
		argb++;
	}

	if ((argc - argb) < 1)
		usage(argv[0]);

	for (argi = argb+1; argi < argc; argi++) {
		if (!zppolymod_auxscan(&aux, argv[argb], argv[argi])) {
			fprintf(stderr, "%s: Couldn't scan \"%s\", \"%s\".\n",
				argv[0], argv[argb], argv[argi]);
			exit(1);
		}
		do_one_poly(&aux, do_galois_cfg, show_logs, verbose);
		if ((argc - argb) > 2)
			printf("\n");
	}

	return 0;
}
