#include "factorize.h"
#include "gcd.h"
#include "iroot.h"
#include "isprime.h"
#include "isqrt.h"
#include "lcm.h"
#include "log2.h"
#include "malloc_check.h"
#include "power.h"
#include "primes16.h"

#include <stdio.h>
#include <math.h>

// ----------------------------------------------------------------
// The first step in AKS is to see if n is of the form a**b, b > 1.  We can
// just take the floor (or ceil) of the bth root of n, for all reasonable b
// (call the floor a), then raise a**b and see if == n.
//
// Up to what value of b should we test?  If n = a ** b, maximum b is when
// a = 2.  So top b = log2(n).

int is_exponent_free_32(
	unsigned n)
{
	unsigned a, b, top_b, a_to_the_b;

	top_b = calc_log2_unsigned(n, IFLOOR);

	for (b = 2; b <= top_b; b++) {
		a = calc_iroot_32(n, b, IFLOOR);
		a_to_the_b = calc_power(a, b);
		if (a_to_the_b == n)
			return 0; // n is not exponent-free.
	}
	return 1; // n is exponent-free.
}

// ----------------------------------------------------------------
// This is trivial for 32-bit integers, since it is easy to tabulate
// primes up to 2**16.
// Tabulated in primes16.h: unsigned primes16[], int numprimes16.

unsigned get_largest_prime_factor_32(
	unsigned n)
{
	int i;
	unsigned rv = n;
	unsigned p;

	for (i = 0; i < numprimes16; i++) {
		p = primes16[i];
		if (n == 1)
			return rv;
		if ((n % p) != 0)
			continue;
		rv = p;
		while ((n % p) == 0) {
			n /= p;
		}
	}
	if (n == 1)
		return rv;
	else
		return n;
}

// ----------------------------------------------------------------
unsigned mod_power_32(
	unsigned b,
	unsigned e,
	unsigned m)
{
	unsigned b2 = b % m;
	unsigned rv  = 1;

	while (e) {
		if (e & 1)
			rv = (rv * b2) % m;
		e >>= 1;
		b2 = (b2 * b2) % m;
	}

	return rv;
}

// ----------------------------------------------------------------
int AKS_is_prime_32(
	unsigned n)
{
	unsigned r, a, max_a, q, min_q, n_pow_mod_r;
	unsigned minfq;

	// See if n is of the form a**b, b > 1.
	if (!is_exponent_free_32(n))
		return 0;

printf("\n");
	r = 2;
	while (r < n) {
		if (calc_gcd_unsigned(n, r) != 1)
			return 0;

		// This isprime() is for 32-bit integers only.
		// Thus we need to make sure we always find an r < 2**32
		// (i.e. small enough to fit into a 32-bit integer), or
		// else revisit this computation.
		if (isprime(r)) {
			q = get_largest_prime_factor_32(r - 1);
			// Need q >= 4 sqrt(r) log(n)
			min_q = 4 * calc_isqrt_32(r, ICEIL)
				* calc_log2_unsigned(n, ICEIL); // xxx ceil
			minfq = 4 * sqrt((double)r) * log((double)n) / log(2.0);
			// Need n ** ((r-1)/q) !equiv 1 mod r
			n_pow_mod_r = mod_power_32(n, (r-1)/q, r);
printf("n=%4u r=%4u r-1=%4u q=%4u min_q = %4u (%4u) n_pow_mod_r =%4u\n",
	n, r, r-1, q, min_q, minfq, n_pow_mod_r);
			if ((q >= min_q) && (n_pow_mod_r != 1))
				break;
		}
//#define R_PANIC 0xfffffffe
#define R_PANIC 0x0000fffe
		if (r >= R_PANIC) {
			printf("I'm getting nervous about finding an r.\n");
			exit(1);
		}
		r++;
	}

printf("n=%4u r=%4u @@\n", n, r);

	max_a = 2 * calc_isqrt_32(r, ICEIL) * calc_log2_unsigned(n, ICEIL);
	for (a = 1; a <= max_a; a++) {
		//zppoly xr1 = x**r - 1 w/ n
		//zppoly xan = (x-a)**n mod xr1
		//zppoly xna = (x**n-a) mod xr1
		//if (xan != xna)
			//return 0;
	}
	// ...

	return 1;
}

int main(int argc, char ** argv)
{
	unsigned n;
	if (argc == 2) {
		(void)AKS_is_prime_32(atoi(argv[1]));
	}
	else if (argc == 3) {
		unsigned nlo, nhi;
		nlo = atoi(argv[1]);
		nhi = atoi(argv[2]);
		for (n = nlo; n <= nhi; n++)
			(void)AKS_is_prime_32(n);
	}
	else {
		for (n = 1; n < 1000; n++) {
			(void)AKS_is_prime_32(n);
		}
	}
	//for (n = 1; n < 1000; n++) {
	//	printf("%4u %4u\n", n, get_largest_prime_factor_32(n));
	//}
	return 0;
}
