// Given modulus m and relatively prime factors f1, f2, ..., fk.
// Search for coefficients c1, c2, ..., ck such that
//   c1 * (x mod f1)  +  c2 * (x mod f2)  +  ...  +  ck * (x mod fk) = x
// for all x in G = (Zm, *).

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "gcd.h"
#include "phi.h"

#define MAX_FACTORS 20

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {factors}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int m;
	int f[MAX_FACTORS];
	int m_over_f[MAX_FACTORS];
	int c[MAX_FACTORS];
	int r[MAX_FACTORS];
	int K;
	int x, rex;
	int i, j;
	int code, temp;

	char * members;
	int * xmods;

	int ngood, nbad;
	int ntotry;
	int print_status = isatty(fileno(stdout));

	if (argc < 2)
		usage(argv[0]);
	K = 0;
	for (i = 1; i < argc; i++) {
		if (K > MAX_FACTORS) {
			printf("Too many factors; please edit "
				"and recompile.\n");
			exit(1);
		}
		if (sscanf(argv[i], "%d", &f[K]) != 1)
			usage(argv[0]);
		K++;
	}

	m = 1;
	for (i = 0; i < K; i++) {
		m *= f[i];
	}
	printf("Modulus %d (", m);
	for (i = 0; i < K; i++)
		printf(" %d", f[i]);
	printf(")\n");
	for (i = 0; i < K; i++) {
		for (j = i+1; j < K; j++) {
			if (calc_gcd(f[i], f[j]) != 1) {
				fprintf(stderr, "Factors %d, %d are not "
					"relatively prime.\n", f[i], f[j]);
				exit(1);
			}
		}
	}
	for (i = 0; i < K; i++) {
		m_over_f[i] = m / f[i];
	}

	ntotry = 1;
	for (i = 0; i < K; i++) {
		ntotry *= f[i] - 1;
	}

	members = (char *)malloc(m);
	if (m == 0) {
		fprintf(stderr, "malloc(%d) failed.\n", m);
		exit(1);
	}

	xmods = (int *)malloc(K * m * sizeof(int));
	if (m == 0) {
		fprintf(stderr, "malloc(%d) failed.\n", K * m * sizeof(int));
		exit(1);
	}
	for (x = 0; x < m; x++) {
		if (calc_gcd(x, m) == 1) {
			members[x] = 1;
			for (i = 0; i < K; i++) {
				xmods[x*K+i] = x % f[i];
			}
		}
		else {
			members[x] = 0;
		}
	}

	// If we knew a priori the number of factors, we'd just write K
	// nested for loops.  Instead, we'll encode all the possibilities
	// into a single integer over which we can loop.
	//
	// Example:  factors are 3, 4, 5 and 7; modulus is 420.
	// * 2 possibilities for c[0]: 140 280
	// * 3 possibilities for c[1]: 105 210 315
	// * 4 possibilities for c[2]:  84 168 252 336
	// * 6 possibilities for c[3]:  60 120 180 240 300 360
	// 2 * 3 * 4 * 6 = 144.
	// In ordinary base-b notation, each place's significance is a power
	// of b.   E.g. 1234 base 10 is 1*10*10*10 + 2*10*10 + 3*10 + 1.
	// However, we can just easily use heterogeneous bases to make
	// an integer, where the "digit" in each place is an index into
	// the array of possible coefficients for the corresponding factor.

	for (code = 0; code < ntotry; code++) {
		temp = code;
		for (i = 0; i < K; i++) {
			c[i] = ((temp % (f[i] - 1)) + 1) * m_over_f[i];
			temp =   temp / (f[i] - 1);

		}

		ngood = 0;
		nbad  = 0;

		for (x = 1; x < m; x++) {
			//if (calc_gcd(x, m) != 1)
			if (!members[x])
				continue;

			for (i = 0; i < K; i++) {
				//r[i] = x % f[i];
				r[i] = xmods[x*K+i];
			}
			rex = 0;
			for (i = 0; i < K; i++) {
				rex += r[i] * c[i];
				rex %= m;
			}
			//rex %= m;
			if (x == rex)
				ngood++;
			else
				nbad++;

		}
		if (nbad == 0) {
			if (print_status)
				printf("\r");

			for (i = 0; i < K; i++)
				printf(" %5d", c[i]);
			printf(" (");

			for (i = 0; i < K; i++)
				printf(" %3d", c[i]/ m_over_f[i]);

			printf(")");
			if (print_status)
				printf("            ");
			printf("\n");
		}
		if (print_status) {
			printf("\r");
			for (i = 0; i < K; i++)
				printf(" %5d", c[i]);
			printf(" %6.2lf%%      ",
				(100.0 * code) / ntotry);
			fflush(stdout);
		}
	}
	if (print_status)
		printf("\r                                     \n");

	free(members);
	free(xmods);
	return 0;
}
