// Given modulus m and relatively prime factors f1, f2, ..., fn,
// find the Chinese coefficients of f and test them out.
//
// To solve the n equations x = r_i mod f_i, i = 1 .. n,
// use x = sum_i r_i c_i (m/f_i) mod m
// where c_i = f_i/m mod f_i.

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "gcd.h"
#include "phi.h"
#include "inv.h"

#define MAX_FACTORS 20

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {factors}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int verbose = 0;

	int m;
	int f[MAX_FACTORS];
	int n;

	int c[MAX_FACTORS];
	int C[MAX_FACTORS];

	int r[MAX_FACTORS];
	int x, rex;

	int i, j;
	int ngood, nbad;

	if (argc < 2)
		usage(argv[0]);
	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Get f
	n = 0;
	for (i = 1; i < argc; i++) {
		if (n > MAX_FACTORS) {
			printf("Too many factors; please edit "
				"and recompile.\n");
			exit(1);
		}
		if (strcmp(argv[i], "-v") == 0) {
			verbose++;
		}
		else {
			if (sscanf(argv[i], "%d", &f[n]) != 1)
				usage(argv[0]);
			n++;
		}
	}

	for (i = 0; i < n; i++) {
		for (j = i+1; j < n; j++) {
			if (calc_gcd(f[i], f[j]) != 1) {
				fprintf(stderr, "Factors %d, %d are not "
					"relatively prime.\n", f[i], f[j]);
				exit(1);
			}
		}
	}

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Compute m
	m = 1;
	for (i = 0; i < n; i++) {
		m *= f[i];
	}
	printf("m = %d\n", m);
	printf("f = (");
	for (i = 0; i < n; i++) {
		if (i != 0)
			printf(" ");
		printf("%d", f[i]);
	}
	printf(")\n");

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Compute c_i

	for (i = 0; i < n; i++) {
		// c_i = f_i/m mod f_i
		// c_i = 1/(m/f_i) mod f_i

		c[i] = calc_inv(m/f[i], f[i]);
		C[i] = c[i] * m/f[i];
	}

	ngood = 0;
	nbad  = 0;

	for (x = 1; x < m; x++) {
		if (calc_gcd(x, m) != 1)
			continue;

		// Split x into remainders.
		for (i = 0; i < n; i++) {
			r[i] = x % f[i];
		}

		// Reconstitute x in terms of its remainders.
		// Use x = sum_i r_i c_i (m/f_i) mod m.
		rex = 0;
		for (i = 0; i < n; i++) {
			//rex += r[i] * c[i] * (m/f[i]);
			rex += r[i] * C[i];
			rex %= m;
		}

		if (verbose) {
			printf("  x=%3d ", x);
			printf(" r=(");
			for (i = 0; i < n; i++) {
				printf("%3d", r[i]);
				if (i < (n-1))
					printf(",");
			}
			printf(")");
			printf(" rex=%3d ", rex);

			if (x != rex)
				printf(" !");
			printf("\n");
		}

		// Check that x has been reconstituted.
		if (x == rex)
			ngood++;
		else
			nbad++;
	}
	if (nbad != 0)
		printf("Fail: Good %d, bad %d.\n", ngood, nbad);

	printf("C = (");
	for (i = 0; i < n; i++) {
		if (i != 0)
			printf(" ");
		printf("%d", C[i]);
	}
	printf(")\n");

	printf("c = (");
	for (i = 0; i < n; i++) {
		if (i != 0)
			printf(" ");
		printf("%d", c[i]);
	}
	printf(")\n");

	return 0;
}
