#include <stdio.h>
#include <stdlib.h>
#include "factorize.h"
#include "primes16.h"
#include "power.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
void calc_factors(
	unsigned n,
	factor_info_t * pfinfo)
{
	int i;
	unsigned p;

	pfinfo->num_distinct = 0;
	if (n == 0)
		return;
	for (i = 0; i < numprimes16; i++) {
		p = primes16[i];
		if (n == 1)
			return;
		if ((n % p) != 0)
			continue;
		if (pfinfo->num_distinct >= FACTORIZE_MAX) {
			fprintf(stderr,
				"calc_factors(): coding error:  "
				"FACTORIZE_MAX too small (%u).\n",
				FACTORIZE_MAX);
			exit(1);
		}
		pfinfo->factors[pfinfo->num_distinct].p = p;
		pfinfo->factors[pfinfo->num_distinct].count = 0;
		while ((n % p) == 0) {
			pfinfo->factors[pfinfo->num_distinct].count++;
			n /= p;
		}
		pfinfo->num_distinct++;
	}
	if (n != 1) {
		pfinfo->factors[pfinfo->num_distinct].p = n;
		pfinfo->factors[pfinfo->num_distinct].count = 1;
		pfinfo->num_distinct++;
	}
}

// ----------------------------------------------------------------
void print_factors(
	unsigned n)
{
	factor_info_t finfo;
	int i;

	calc_factors(n, &finfo);

	for (i = 0; i < finfo.num_distinct; i++) {
		if (i > 0)
			printf(" ");
		printf("%d", finfo.factors[i].p);
		if (finfo.factors[i].count > 1)
			printf("^%d", finfo.factors[i].count);
	}
}

// ----------------------------------------------------------------
// Given the prime factorization p1^m1 ... pk^mk of n, how to enumerate all
// factors of n?
//
// Example 72 = 2^3 * 3^2.  Exponent on 2 is one of 0, 1, 2, 3.  Exponent on
// 3 is one of 0, 1, 2.  Number of possibilities:  product over i of (mi + 1).
// Factors are:
//
//	2^0 3^0    1
//	2^0 3^1    3
//	2^0 3^2    9
//	2^1 3^0    2
//	2^1 3^1    6
//	2^1 3^2   18
//	2^2 3^0    4
//	2^2 3^1   12
//	2^2 3^2   36
//	2^3 3^0    8
//	2^3 3^1   24
//	2^3 3^2   72

unsigned num_divisors(
	factor_info_t * pfinfo)
{
	unsigned rv;
	int i;

	if ((pfinfo->num_distinct < 0)
		|| (pfinfo->num_distinct > FACTORIZE_MAX))
	{
		fprintf(stderr,
			"num_divisors:  num distinct (%d) out of bounds 0-%d\n",
			pfinfo->num_distinct, FACTORIZE_MAX);
		exit(1);
	}

	rv = 1;
	for (i = 0; i < pfinfo->num_distinct; i++)
		rv *= pfinfo->factors[i].count + 1;
	return rv;
}

// ----------------------------------------------------------------
// See comments to the above.  k is treated as a multibase representation
// over the bases mi+1.

unsigned kth_divisor(
	factor_info_t * pfinfo,
	unsigned k)
{
	unsigned rv;
	int i;
	unsigned base;
	unsigned power;

	rv = 1;
	for (i = 0; i < pfinfo->num_distinct; i++) {
		base = pfinfo->factors[i].count + 1;
		power = k % base;
		k = k / base;
		rv *= calc_power(pfinfo->factors[i].p, power);
	}
	return rv;
}

// ----------------------------------------------------------------
static int uqcmp(const void * pv1, const void * pv2)
{
	const unsigned * p1 = pv1;
	const unsigned * p2 = pv2;
	if (*p1 < *p2)
		return -1;
	if (*p1 > *p2)
		return 1;
	return 0;
}

// ----------------------------------------------------------------
// The caller must free the return value.

unsigned * get_all_divisors(
	unsigned   n,
	unsigned * pnum_divisors)
{
	factor_info_t finfo;
	unsigned nf, k;
	unsigned * all_divisors;

	calc_factors(n, &finfo);
	nf = num_divisors(&finfo);
	*pnum_divisors = nf;

	all_divisors = (unsigned *)malloc_check(nf * sizeof(unsigned));
	for (k = 0; k < nf; k++) {
		all_divisors[k] = kth_divisor(&finfo, k);
	}
	qsort(all_divisors, nf, sizeof(unsigned), uqcmp);

	return all_divisors;
}

// ----------------------------------------------------------------
// Algorithm taken from Paul Rubin's factor.c (phr@ocf.berkeley.edu).

void calc_factors_ull(
	unsigned long long n0,
	factor_info_ull_t * pfinfo)
{
	unsigned long long n = n0, d, q;

	pfinfo->num_distinct = 0;

	if (n < 1)
		return;

	if ((n % 2) == 0) {
		pfinfo->factors[pfinfo->num_distinct].p = 2;
		pfinfo->factors[pfinfo->num_distinct].count = 0;
		while (n % 2 == 0) {
			pfinfo->factors[pfinfo->num_distinct].count++;
			n /= 2;
		}
		pfinfo->num_distinct++;
	}

	// The exit condition in the following loop is correct because
	// any time it is tested one of these 3 conditions holds:
	//   (1) d divides n
	//   (2) n is prime
	//   (3) n is composite but has no factors less than d.
	// If (1) or (2) obviously the right thing happens.
	// If (3), then since n is composite it is >= d^2.

	pfinfo->factors[pfinfo->num_distinct].count = 0;
	d = 3;
	do {
		q = n / d;
		while (n == q * d) {
			// assert (n_factors < max_n_factors);
			pfinfo->factors[pfinfo->num_distinct].p = d;
			pfinfo->factors[pfinfo->num_distinct].count++;
			n = q;
			q = n / d;
		}
		if (pfinfo->factors[pfinfo->num_distinct].count != 0) {
			pfinfo->num_distinct++;
			pfinfo->factors[pfinfo->num_distinct].count = 0;
		}
		d += 2;
	} while (d <= q);

	if (n != 1 || n0 == 1) {
		pfinfo->factors[pfinfo->num_distinct].p = n;
		pfinfo->factors[pfinfo->num_distinct].count = 1;
		pfinfo->num_distinct++;
	}
}

// ----------------------------------------------------------------
// Given the prime factorization p1^m1 ... pk^mk of n, how to enumerate all
// factors of n?
//
// Example 72 = 2^3 * 3^2.  Exponent on 2 is one of 0, 1, 2, 3.  Exponent on
// 3 is one of 0, 1, 2.  Number of possibilities:  product over i of (mi + 1).
// Factors are:
//
//	2^0 3^0    1
//	2^0 3^1    3
//	2^0 3^2    9
//	2^1 3^0    2
//	2^1 3^1    6
//	2^1 3^2   18
//	2^2 3^0    4
//	2^2 3^1   12
//	2^2 3^2   36
//	2^3 3^0    8
//	2^3 3^1   24
//	2^3 3^2   72

unsigned num_divisors_ull(
	factor_info_ull_t * pfinfo)
{
	unsigned rv;
	int i;

	if ((pfinfo->num_distinct < 0)
		|| (pfinfo->num_distinct > FACTORIZE_MAX))
	{
		fprintf(stderr,
		"num_divisors_ull:  num distinct (%d) out of bounds 0-%d\n",
			pfinfo->num_distinct, FACTORIZE_MAX);
		exit(1);
	}

	rv = 1;
	for (i = 0; i < pfinfo->num_distinct; i++)
		rv *= pfinfo->factors[i].count + 1;
	return rv;
}

// ----------------------------------------------------------------
// See comments to the above.  k is treated as a multibase representation
// over the bases mi+1.

unsigned long long kth_divisor_ull(
	factor_info_ull_t * pfinfo,
	unsigned k)
{
	unsigned long long rv;
	unsigned i;
	unsigned base;
	unsigned power;

	rv = 1;
	for (i = 0; i < pfinfo->num_distinct; i++) {
		base = pfinfo->factors[i].count + 1;
		power = k % base;
		k = k / base;
		// Avoid the overhead of the function call, if pointless.
		if (power > 0) {
			rv *= calc_power_ull(pfinfo->factors[i].p, power);
		}
	}
	return rv;
}

// ----------------------------------------------------------------
static int ull_qcmp(const void * pv1, const void * pv2)
{
	const unsigned long long * p1 = pv1;
	const unsigned long long * p2 = pv2;
	if (*p1 < *p2)
		return -1;
	if (*p1 > *p2)
		return 1;
	return 0;
}

// ----------------------------------------------------------------
// The caller must free the return value.

unsigned long long * get_all_divisors_ull(
	unsigned long long n,
	unsigned * pnum_divisors)
{
	factor_info_ull_t finfo;
	unsigned nf, k;
	unsigned long long * all_divisors;

	calc_factors_ull(n, &finfo);
	nf = num_divisors_ull(&finfo);
	*pnum_divisors = nf;

	all_divisors = (unsigned long long *)malloc_check(
		nf * sizeof(unsigned long long));
	for (k = 0; k < nf; k++) {
		all_divisors[k] = kth_divisor_ull(&finfo, k);
	}
	qsort(all_divisors, nf, sizeof(unsigned long long), ull_qcmp);

	return all_divisors;
}
