// This is a bisection search for the kth root of n, using integer arithmetic.
// Since it's bisection, it takes only log(n) steps.  One could start with
// lower bracket = 0 and upper bracket = n, then bisect.  But, as an
// easy-to-implement performance enhancement, I will make a better first
// estimate.
//
// Given n, 2^w <= n < 2^(w+1) for some w.  This is implemented using the
// log2_floor(n) function.  w = qk + r.  2^qk <= n < 2^(qk+k), so
// 2^q <= kth root of n < 2^(q+1).
//
// Example: n = 23456, k = 3.  log2(n) = w = 14.  14 = 4*3+2; q = 4, r = 2.
// qk = 12; qk+k = 15.  q = 4, q+1 = 5.  23456 is between 4096 and 32768, so
// the cube root of 23456 is between 16 and 32.
//
// Check:
// * cube root ~= 28.6253
// * 28**3 = 21952
// * 29**3 = 24389

#include <stdio.h>
#include "iroot.h"
#include "power.h"
#include "log2.h"

// ----------------------------------------------------------------
unsigned calc_iroot_32(
	unsigned n,
	unsigned k,
	int want_ceil)
{
	unsigned w, q;
	unsigned lower, upper, diff;
	unsigned half_diff, middle, middle_power;

	if (n == 0)
		return 0;
	w = calc_log2_unsigned(n, IFLOOR);
	q = w / k;
	lower = 1 << q;
	upper = lower << 1;
	diff  = upper - lower;

	while (diff > 0) {
		half_diff = diff >> 1;
		middle = lower + half_diff;
		middle_power = calc_power(middle, k);
		if (middle_power == n)
			return middle;
		else if (middle_power < n)
			lower += half_diff;
		else
			upper -= half_diff;
		diff = half_diff;
	}
	if (want_ceil == ICEIL)
		return upper;
	else
		return lower;
}

// ----------------------------------------------------------------
unsigned long long calc_iroot_64(
	unsigned long long n,
	unsigned long long k,
	int want_ceil)
{
	unsigned long long nsv, half_log;
	unsigned long long lower, upper, diff;
	unsigned long long half_diff, middle, middle_sq;

	if (n == 0)
		return 0;

	for (nsv = n, half_log = 0; nsv != 0; nsv >>= 2, half_log++)
		;
	half_log--;
	lower = 1ULL << half_log;
	upper = lower << 1;
	diff  = upper - lower;

	while (diff > 0) {
		half_diff = diff >> 1;
		middle = lower + half_diff;
		middle_sq = middle * middle;
		if (middle_sq == n)
			return middle;
		else if (middle_sq < n)
			lower += half_diff;
		else
			upper -= half_diff;
		diff = half_diff;
	}
	if (want_ceil == ICEIL)
		return upper;
	else
		return lower;
}
