#include <stdio.h>
#include "isqrt_ubig.h"

static unsigned ubig_calc_half_log(
	ubig * pn);

// ----------------------------------------------------------------
// As of this writing (2003/06/22), bisection is faster than Newton's method
// by about 3x.  Newton's method requires much fewer iterations, but each
// iteration is more expensive, apparently due to my piss-poor implementation
// of arbitrary-precision division.  When I get division working more
// efficiently, this decision should be revisited.

ubig calc_isqrt_ubig(
	ubig * pn,
	int want_ceil)
{
	//return calc_isqrt_newton_ubig(pn, want_ceil);
	return calc_isqrt_bisect_ubig(pn, want_ceil);
}

// ----------------------------------------------------------------
ubig calc_isqrt_bisect_ubig(
	ubig * pn,
	int want_ceil)
{
	ubig zero;
	unsigned half_log;
	ubig lower, upper, diff;
	ubig half_diff, middle, middle_sq;

	zero = ubig_from_unsigned(0);
	if (ubig_equals_unsigned(pn, 0))
		return zero;

	half_log = ubig_calc_half_log(pn);
	lower = zero;
	upper = zero;
	ubig_set_bit(&lower, half_log);
	ubig_set_bit(&upper, half_log + 1);
	diff = ubig_sub(&upper, &lower);

	while (ubig_gt(&diff, &zero)) {
		half_diff = ubig_irsh(&diff, 1);
		middle = ubig_add(&lower, &half_diff);
		middle_sq = ubig_mul(&middle, &middle);
		if (ubig_eq(&middle_sq, pn))
			return middle;
		else if (ubig_lt(&middle_sq, pn))
			lower = ubig_add(&lower, &half_diff);
		else
			upper = ubig_sub(&upper, &half_diff);
		diff = half_diff;
	}
	if (want_ceil == ICEIL)
		return upper;
	else
		return lower;
}

// ----------------------------------------------------------------
// Newton's method:
//
// f(x) = 0.  p is an exact solution; u is an approximation.  u ~= p.
// Taylor expansion of f(x) about u:
//   f(x) = f(u) + (x-u)f'(u) + 0.5 (x-u)^2 f''(xi(x)), xi in [x, u].
//   f(p) = 0 ~= f(u) + (p-u) f'(u)
//   0 ~= f(u) + (p-u) f'(u)
//   p ~= u - f(u)/f'(u)
// Iteration:
//   p(n+1) ~= p(n) - f(p(n))/f'(p(n))
//
// We want to take the square root of N, so f(x) = x^2 - N = 0; f'(x) = 2x.
// p(n+1) = p(n) - (p(n)^2 - N)/(2 p(n))
//        = p(n) - p(n)^2/2p(n) + N/2p(n)
//        = p(n) - p(n)/2 + N/2p(n)
//        = p(n)/2 + N/2p(n)
//        = (p(n) + N/p(n)) / 2
// p := (p + N/p) / 2
//
// When using floating-point arithmetic, we iterate until |p(n+1) - p(n)| is
// less than some tolerance.  When using integer arithmetic, we can simply
// iterate until p(n+1) == p(n), *except* for a special case: when n is one
// less than a perfect square, the sequence of approximants oscillates between
// the floor of the square root and the ceiling of the square root.
//
// For example:  N = 14, initial approximation p = 2.
//   p0 = 2
//   p1 = (2 + 14/2) / 2 = 4
//   p2 = (4 + 14/4) / 2 = 3
//   p3 = (3 + 14/3) / 2 = 3
//   p4 = (3 + 14/3) / 2 = 3
//   ...
//
// By contrast:  N = 15, initial approximation p = 2.
//   p0 = 2
//   p1 = (2 + 15/2) / 2 = 4
//   p2 = (4 + 15/4) / 2 = 3
//   p3 = (3 + 15/3) / 2 = 4
//   p4 = (4 + 15/4) / 2 = 3
//   p5 = (3 + 15/3) / 2 = 4
//   p6 = (4 + 15/4) / 2 = 3
//   ...

ubig calc_isqrt_newton_ubig(
	ubig * pn,
	int want_ceil)
{
	ubig zero;
	ubig p, q, prev1, prev2;
	ubig floor, ceil, psq;
	unsigned half_log, i;

	// The half-log algorithm doesn't do so well when its input is zero.
	zero = ubig_from_unsigned(0);
	if (ubig_equals_unsigned(pn, 0))
		return zero;

	// Initial approximation
	half_log = ubig_calc_half_log(pn);
	p = zero;
	ubig_set_bit(&p, half_log);

	for (i = 0; ; i++) {
		prev2 = prev1;
		prev1 = p;

		//p = (p + N/p) / 2;
		q = ubig_div(pn, &p);
		p = ubig_add(&p, &q);
		p = ubig_irsh(&p, 1);

		// Stopping criterion
		if (i < 2)
			continue;
		if (!ubig_eq(&p, &prev2))
			continue;

		if (ubig_eq(&p, &prev1)) {
			floor = p;
		}
		else {
			if (ubig_lt(&p, &prev1))
				floor = p;
			else
				floor = prev1;
		}
		break;
	}

	ceil = floor;
	psq = ubig_mul(&p, &p);
	if (!ubig_eq(pn, &psq))
		ubig_incr(&ceil);

	if (want_ceil == IFLOOR)
		return floor;
	else
		return ceil;
}

// ----------------------------------------------------------------
static unsigned ubig_calc_half_log(
	ubig * pn)
{
	ubig sv = *pn;
	unsigned half_log;

	half_log = 0;
	for ( ; !ubig_equals_unsigned(&sv, 0); ubig_irsh(&sv, &sv, 2))
		half_log++;
	half_log--;

	return half_log;
}
